
#include <string.h>
#include "fertilitylopfeatureset.h"

FertilityLOPFeatureSet::FertilityLOPFeatureSet(const char* ParamFile, const char * ScalingFactorString,int numberOfExp, double * expWeights)
 : FeatureSet()
{
	numberOfScalingFactors = 0;
	numberOfExperts = numberOfExp;
	expertWeights = expWeights;
	useGIZAFertilities = 0;
	GIZASourceScalingFactors = new double[numberOfExperts];
	GIZATargetScalingFactors = new double[numberOfExperts];
	numberOfGIZAFertilities = 0;
	paramFile = ParamFile;
	if(ParamFile == NULL) {
		cout << "Parameter File for Fertility Feature needed"<< endl;
		exit(-1);
	}
	LoadParamFile(ParamFile);
  
	ScalingFactors = new double[2*numberOfScalingFactors*numberOfExperts];
	Utils::Extract2DScalingFactors(ScalingFactorString,ScalingFactors,2*numberOfScalingFactors,numberOfExperts);
	if(useGIZAFertilities) {
		InitDefaultGIZAFertilities(); 
	}

}


FertilityLOPFeatureSet::~FertilityLOPFeatureSet()
{
	delete[] GIZASourceScalingFactors;
	delete[] GIZATargetScalingFactors;
	delete[] ScalingFactors;
}

  /**
 * @param  G
   */
void FertilityLOPFeatureSet::GenerateFeatures (Graph * G , int offset, int numberOfAllFeatures ) {
	SourceLength = G->GetSourceLength();
	TargetLength = G->GetTargetLength();
	sourceFeatures = new FertilityLOPFeature[SourceLength];
	targetFeatures = new FertilityLOPFeature[TargetLength];
	for(int i = 0; i < SourceLength; i++) {
		vector<HiddenNode *> links;
		vector<double> GIZAFert;
		for(int j = 0; j < TargetLength; j++) {
			links.push_back(G->GetNode(i,j));
		}
		if(useGIZAFertilities) {
			const string source = G->GetSourceWord(i);
			map<string, vector<double> >::iterator iter = GIZASourceFertilities.find(source);
			if(iter != GIZASourceFertilities.end()) {
				GIZAFert = iter->second;
			}else {
				cout << "could not find GIZA fert for \"" << source << "\"" << endl;
				GIZAFert = defaultGIZASourceFertilities;
			}
		}
		assert(numberOfGIZAFertilities <= GIZAFert.size());
		sourceFeatures[i].Init(links,ScalingFactors,numberOfScalingFactors,GIZAFert,GIZASourceScalingFactors,numberOfGIZAFertilities,2,offset,numberOfAllFeatures,offset+2*numberOfScalingFactors,numberOfExperts,expertWeights);
	}
	for(int i = 0; i < TargetLength; i++) {
		vector<HiddenNode *> links;
		vector<double> GIZAFert;
		for(int j = 0; j < SourceLength; j++) {
			links.push_back(G->GetNode(j,i));
		}
		if(useGIZAFertilities) {
			const string target = G->GetTargetWord(i);
			map<string, vector<double> >::iterator iter = GIZATargetFertilities.find(target);
			if(iter != GIZATargetFertilities.end()) {
				GIZAFert = iter->second;
			}else {
				cout << "could not find GIZA fert for " << target << endl;
				GIZAFert = defaultGIZATargetFertilities;
			}
		}
		targetFeatures[i].Init(links,&ScalingFactors[numberOfScalingFactors*numberOfExperts],numberOfScalingFactors,GIZAFert,GIZATargetScalingFactors,numberOfGIZAFertilities,3,offset + numberOfScalingFactors,numberOfAllFeatures,offset+2*numberOfScalingFactors+1,numberOfExperts,expertWeights);
	}

}
void FertilityLOPFeatureSet::RemoveFeatures() {
	delete[] sourceFeatures;
	delete[] targetFeatures;
}
  /**
 * @return FertilitySetIterator
   */
FertilityLOPFeatureSetIterator * FertilityLOPFeatureSet::GetIterator ( ) const {
	return new FertilityLOPFeatureSetIterator(SourceLength,TargetLength,sourceFeatures,targetFeatures);

};

void FertilityLOPFeatureSet::EstimateScalingFactores() {
	EstimatedDerivation.clear();
	vector<double> ExpectValue;
	vector<double> ExampleValue;
	ExpectValue.resize(numberOfExperts,0);
	ExampleValue.resize(numberOfExperts,0);
	for(int i = 0; i < SourceLength; i++) {
		vector<double> TempExpectValue = sourceFeatures[i].GetExpectationValue();
		vector<double> TempExampleValue = sourceFeatures[i].GetExampleValue();
		assert(TempExpectValue.size() == ExpectValue.size());
		assert(TempExpectValue.size() == ExampleValue.size());
		for(int j = 0; j < numberOfExperts; j++) {
			ExpectValue[j] += TempExpectValue[j];
			ExampleValue[j] += TempExampleValue[j];
/*			if(ExampleValue[j] > 100) {
				cout << TempExampleValue[j] << endl;
			}
*/		}
	}

	for(int i = 0; i < TargetLength; i++) {
		vector<double> TempExpectValue = targetFeatures[i].GetExpectationValue();
		vector<double> TempExampleValue = targetFeatures[i].GetExampleValue();
		assert(TempExpectValue.size() == ExpectValue.size());
		assert(TempExpectValue.size() == ExampleValue.size());
		for(int j = 0; j < numberOfExperts; j++) {
			ExpectValue[j] += TempExpectValue[j];
			ExampleValue[j] += TempExampleValue[j];
/*			if(ExampleValue[j] > 100) {
				cout << TempExampleValue[j] << endl;
			}
*/		}
	}

	for(int j = 0; j < numberOfExperts; j++) {
		double deltaL = ExampleValue[j] - ExpectValue[j];
   //cout << "Calculating Der[" << j << "]: " << ExampleValue[j] << " - " <<  ExpectValue[j] << " = " << deltaL << endl;
		EstimatedDerivation.push_back(deltaL);
	}
}

void FertilityLOPFeatureSet::LoadParamFile(const char* ParamFileName) {
	string GIZASourceSFString;
	string GIZATargetSFString;
	ifstream ParamFile( ParamFileName);

	if (!ParamFile) {
		cerr << "File " << ParamFileName << " could not be opened\n";
		exit(-2);
	}
  
	char *pKey = new char[ 200 ];
	char *pValue = new char[ 200 ];
// TODO: Does not work with LF.Param with empty line at end!!!!!!!!!!!!
	while ( !ParamFile.eof() ) {
		ParamFile >> pKey >> pValue;
		if(strcmp(pKey,"SourceGIZAFertilityFile") == 0) {    
			useGIZAFertilities = 1;
			Utils::LoadGIZAFertilities(pValue,GIZASourceFertilities);
		}else if(strcmp(pKey,"TargetGIZAFertilityFile") == 0) {
			useGIZAFertilities = 1;
			Utils::LoadGIZAFertilities(pValue,GIZATargetFertilities);
		}else if(strcmp(pKey,"GIZASourceScalingFactor") == 0) {
			useGIZAFertilities = 1;
			GIZASourceSFString = string(pValue);
		}else if(strcmp(pKey,"GIZATargetScalingFactor") == 0) {
			useGIZAFertilities = 1;
			GIZATargetSFString = string(pValue);
		}else if(strcmp(pKey,"NumberOfGIZAFertilities") == 0) {
			useGIZAFertilities = 1;
			numberOfGIZAFertilities = atoi(pValue);
		}else if(strcmp(pKey,"NumberOfFertilities") == 0) {
			numberOfScalingFactors = atoi(pValue);
		}
	}

	assert(numberOfScalingFactors > 0);
	assert((useGIZAFertilities == 0) || (GIZASourceFertilities.size() > 0));
	assert((useGIZAFertilities == 0) || (GIZATargetFertilities.size() > 0));
	assert((useGIZAFertilities == 0) || (numberOfGIZAFertilities > 0));
	
	if(useGIZAFertilities) {
		Utils::ExtractScalingFactors(GIZASourceSFString.c_str(),GIZASourceScalingFactors,numberOfExperts);
		Utils::ExtractScalingFactors(GIZATargetSFString.c_str(),GIZATargetScalingFactors,numberOfExperts);
	}
	ParamFile.close();
	delete[] pKey;
	delete[] pValue;    
}

void FertilityLOPFeatureSet::InitDefaultGIZAFertilities() {
	double f1 = 1/sqrt(2*PI);
	for(int i = 0; i < numberOfGIZAFertilities; i++) {
		double f = f1*exp(-0.5*(i-1)*(i-1));
		defaultGIZASourceFertilities.push_back(f);
		defaultGIZATargetFertilities.push_back(f);
	}


}


