#include <string.h>
#include "locallopfeatureset.h"
	/** 
	 *   Constructor for a LocalLOPFeature set.
	 *  Init scaling factors and features specified in the parameter file (e.g. load Lex,..)
	 *  @param ParamFile Parameter file for LocalFeatureSet
	 *  @param Factors String containing the ScalingFactors
	 *  @param numberOfExp Number of Experters used in this model
	 * @param expWeights Weights for the different experts
	 */

LocalLOPFeatureSet::LocalLOPFeatureSet(const char* ParamFile, const char * ScalingFactorsString,int numberOfExp, double * expWeights)
 : FeatureSet()
{
	numberOfExperts=numberOfExp;
	expertWeights = expWeights;
  additionalFeatures = 0;
  additionalLexFeatures = 0;
  useRelativePosition = 0;
  useIdentityFeature = 0;
  useSimilarityFeature = 0;
  usePOSFeature = 0;
  useFrequentWordFeature = 0;
  numberOfLinkFeatures = 0;
  NumberOfDynamicFeatures = 0;

	cout << "Parameter-File: " << ParamFile << endl;
	//Load Param file, done by subclass
	LoadParamFile(ParamFile);

	ScalingFactors = new double[2*(numberOfLexicons + additionalFeatures) * numberOfExperts];	  
	Utils::Extract2DScalingFactors(ScalingFactorsString,ScalingFactors,2*(numberOfLexicons + additionalFeatures),numberOfExperts);
}


LocalLOPFeatureSet::~LocalLOPFeatureSet()
{
	delete[] ScalingFactors;
	delete[] Lexicon;
	if(usePOSFeature == 1) {
		delete[] POSScalingFactors;
	}
	if(useFrequentWordFeature == 1) {
		delete[] FrequentWordScalingFactors;
	}

}


/**
 * Loads Parameter File
 * Extract info form parameter file and load addtional data, if needed
 * @param ParamFileName Filename of parameter file
 */

void LocalLOPFeatureSet::LoadParamFile(const char* ParamFileName) {
   vector<string> LexFileNames;
   vector<int> Directions;

   ifstream ParamFile( ParamFileName);

  if (!ParamFile) {
    cerr << "File " << ParamFileName << " could not be opened\n";
    exit(-2);
  }
  
  char *pKey = new char[ 200 ];
  char *pValue = new char[ 200 ];
// TODO: Does not work with LF.Param with empty line at end!!!!!!!!!!!!
  while ( !ParamFile.eof() ) {
    ParamFile >> pKey >> pValue;
//    cout << "Line: " << pKey << pValue << endl;
    if(strcmp(pKey,"Lexicon") == 0) {
	LexFileNames.push_back(string((const char*)pValue));
	Directions.push_back(0);
	SourceNormalized.push_back(0);
	TargetNormalized.push_back(0);
//	cout << "Increase number of Lex to: " << LexFileNames.size() << "with file: " << pValue << endl;
    }else if(strcmp(pKey,"Direction") == 0) {
        Directions[Directions.size() -1] = (atoi((const char*) pValue));
    }else if(strcmp(pKey,"SourceNormalized") == 0) {
        SourceNormalized[SourceNormalized.size() -1] = (atoi((const char*) pValue));
    }else if(strcmp(pKey,"TargetNormalized") == 0) {
        TargetNormalized[TargetNormalized.size() -1] = (atoi((const char*) pValue));
    }else if(strcmp(pKey, "relativePosition") == 0) {
      useRelativePosition = atoi((const char*) pValue);
    }else if(strcmp(pKey, "IdentityFeature") == 0) {
      useIdentityFeature = atoi((const char*) pValue);
    }else if(strcmp(pKey, "SimilarityFeature") == 0) {
      useSimilarityFeature = atoi((const char*) pValue);
    }else if(strcmp(pKey, "NumberOfLinkFeatures") == 0) {
      numberOfLinkFeatures = atoi((const char*) pValue);
    }else if(strcmp(pKey, "POSFeatureScalingFactorFile") == 0) {
      usePOSFeature = 1;
      POSScalingFactorFiles.push_back(string((const char*)pValue));
      
    }else if(strcmp(pKey, "POSSourceFile") == 0) {
      usePOSFeature = 1;
      Utils::LoadVoc(pValue,POSSourceVoc);
      
    }else if(strcmp(pKey, "POSTargetFile") == 0) {
      usePOSFeature = 1;
      Utils::LoadVoc(pValue,POSTargetVoc);
    }else if(strcmp(pKey, "FrequentWordFeatureScalingFactorFile") == 0) {
      useFrequentWordFeature = 1;
      FrequentWordScalingFactorFiles.push_back(string((const char*)pValue));
      
    }else if(strcmp(pKey, "FrequentWordSourceFile") == 0) {
      useFrequentWordFeature = 1;
      Utils::LoadVoc(pValue,FrequentWordSourceVoc);
      
    }else if(strcmp(pKey, "FrequentWordTargetFile") == 0) {
      useFrequentWordFeature = 1;
      Utils::LoadVoc(pValue,FrequentWordTargetVoc);
    }



  }

  assert(Directions.size() == LexFileNames.size());
  ParamFile.close();
  delete[] pKey;
  delete[] pValue;

  if(LexFileNames.size() == 0) {
	cout << "LocalFeatureParamfile contained no Lexica" << endl;
	exit(1);
  }
  Lexicon = new map<pair<string,string>,double>[LexFileNames.size()];
  numberOfLexicons = LexFileNames.size();
  for(int i = 0; i < LexFileNames.size(); i++) {
	Utils::LoadLex(Lexicon[i],LexFileNames[i].c_str(),Directions[i]);
  }
  
  for(int k = 0; k < numberOfLexicons; k++) {
	  if(SourceNormalized[k] != 0) {
		  additionalFeatures++;
		  additionalLexFeatures++;
	  }
	  if(TargetNormalized[k] != 0) {
		  additionalFeatures ++;
		  additionalLexFeatures ++;
	  }
  }
  cout << "Using " << additionalLexFeatures << " addtional lexical features" << endl;
if(useRelativePosition){
    cout << "Using relative Position" << endl;
    additionalFeatures ++;
  }
  if(useIdentityFeature) {
    cout << "Using identityfeature" << endl;
    additionalFeatures ++;
  }
  if(useSimilarityFeature) {
    cout << "Using similarity Feature" << endl;
    additionalFeatures ++;
  }
  if(numberOfLinkFeatures) {
	cout << "Using Link Features of " << numberOfLinkFeatures << " diffenrent sources" << endl;
	additionalFeatures += numberOfLinkFeatures;
  }
  if(usePOSFeature) {
    assert(POSScalingFactorFiles.size() == numberOfExperts);
    POSScalingFactors = new double[numberOfExperts*POSSourceVoc.size()*POSTargetVoc.size()];
    cout << "Using POS Feature" << endl;
    Utils::LoadScalingFactorFiles(POSScalingFactorFiles,POSScalingFactors,POSSourceVoc.size(),POSTargetVoc.size());
    cout << "Number of POS scaling factors:" << POSSourceVoc.size() * POSTargetVoc.size() << endl;
    NumberOfDynamicFeatures += POSSourceVoc.size() * POSTargetVoc.size();
 
  }
  if(useFrequentWordFeature) {
    cout << "Using Frequent Word  Feature" << endl;
    assert(FrequentWordScalingFactorFiles.size() == numberOfExperts);
    FrequentWordScalingFactors = new double [numberOfExperts * FrequentWordSourceVoc.size() * FrequentWordTargetVoc.size()];
    Utils::LoadScalingFactorFiles(FrequentWordScalingFactorFiles,FrequentWordScalingFactors,FrequentWordSourceVoc.size(),FrequentWordTargetVoc.size());
    cout << "Number of frequent word scaling factors : " << FrequentWordSourceVoc.size() * FrequentWordTargetVoc.size() << endl;
    NumberOfDynamicFeatures += FrequentWordSourceVoc.size() * FrequentWordTargetVoc.size();
  }
  NumberOfFeatures = 2*(numberOfLexicons + additionalFeatures);


}
void LocalLOPFeatureSet::GenerateFeatures (Graph * G,int offset, int numberOfAllFeatures) {
	SourceLength = G->GetSourceLength();
	TargetLength = G->GetTargetLength();
	features = new LocalLOPFeature[SourceLength * TargetLength];
	int SourceLengthForRel = max(SourceLength-1, 1);
	int TargetLengthForRel = max(TargetLength-1,1);
	int numberOfLexMatrix;
	double * lexMatrix = CreateLexMatrix(numberOfLexMatrix,G);
	double * linkFeatureValues = InitLinkFeatures(G->GetLinkFeatures());
	for(int i = 0; i < SourceLength; i++) {
		for(int j = 0; j < TargetLength; j++) {
			const string Source = G->GetSourceWord(i);
			const string Target = G->GetTargetWord(j);
			vector<double *> dynScalingFactor;
			vector<int> dynScalingIndex;
			double* p = new double[NumberOfFeatures/2];
			int k = 0;
			for(; k < numberOfLexMatrix;k++) {
				p[k] = lexMatrix[k*SourceLength*TargetLength + i * TargetLength + j];
				//cout << "Probability[" << i <<"][" << j <<"][" << k <<"] = " << p[k] << endl;
			}
			if(useRelativePosition) {
			  assert(k < NumberOfFeatures/2);
			  p[k] = (1.0*i/(SourceLengthForRel))-(1.0*j/(TargetLengthForRel));
			  //p[k] = (1.0*(i + 1)/(SourceLength))-(1.0*(j + 1)/(TargetLength));
			  if(p[k] < 0 ) {p[k] *= -1;}
			  p[k] = 1 - p[k];
			  k++;
			}
			if(useIdentityFeature) {
			  //cout << Source << "-" << Target << " ";
			  //cout << 1-abs(strcmp(Source.c_str(),Target.c_str())) << endl;
			  if((Source.length() > 1) && (strcmp(Source.c_str(),Target.c_str()) == 0))
			    p[k] = 1;
			  else
			    p[k] = 0;
			  k++;
			}
			if(useSimilarityFeature) {
			  p[k] = 1 - Utils::CacluateEditDistance(Source,Target);
			  //cout << "Edit distance for (" << Source <<" , " << Target << ") : " << p[k]<< endl;
			  k++;
			}
			if(numberOfLinkFeatures) {
				//cout << "LinkFeatures for (" << i << "," << j << ")" << endl;
				for(int l = 0; l < numberOfLinkFeatures; l++) {
					int index = i*TargetLength*numberOfLinkFeatures + j * numberOfLinkFeatures + l;
					p[k] = linkFeatureValues[index];
					//cout << p[k] << " ";
					k++;
				}
				//cout << endl;
			}
			int dynOffset = NumberOfFeatures;
			if(usePOSFeature) {
			  const string SourcePOS = G->GetSourcePOS(i);
			  const string TargetPOS = G->GetTargetPOS(j);
			  assert(POSSourceVoc.count(SourcePOS) != 0);
			  assert(POSTargetVoc.count(TargetPOS) != 0);
			  int SourcePOSIndex = POSSourceVoc[SourcePOS];
			  int TargetPOSIndex = POSTargetVoc[TargetPOS];
			  int index = SourcePOSIndex * POSTargetVoc.size() + TargetPOSIndex;
			  //cout << "Pos: ( "<< i << "," <<j<< ") : (" << Source << "," << Target <<") " << endl;
			  //cout << "Scaling Factor for " << SourcePOS << " and " << TargetPOS << " : " << POSScalingFactors[index] << endl;
			  //cout << "Index : " << SourcePOSIndex << " , " << TargetPOSIndex << endl;
			  assert(index < POSTargetVoc.size() * POSSourceVoc.size());
			  dynScalingFactor.push_back(&POSScalingFactors[index*numberOfExperts]);
			  dynScalingIndex.push_back(dynOffset + index);
			  dynOffset += POSTargetVoc.size() * POSSourceVoc.size();
			}
			if(useFrequentWordFeature) {
			  //check if source word is in
			  map<string,int>::iterator sourceIter = FrequentWordSourceVoc.find(Source);
			  map<string,int>::iterator targetIter = FrequentWordTargetVoc.find(Target);
			  if((sourceIter != FrequentWordSourceVoc.end()) &&(targetIter != FrequentWordTargetVoc.end())) {
			    int sourceIndex = sourceIter->second;
			    int targetIndex = targetIter->second;
			    int index = sourceIndex * FrequentWordTargetVoc.size() + targetIndex;
			    //cout << "Pos: ( "<< i << "," <<j<< ") : (" << Source << "," << Target <<") " << endl;
			    //cout << "Index : " << sourceIndex << " , " << targetIndex << ": " <<  FrequentWordScalingFactors[index] << endl;
			    assert(index < FrequentWordTargetVoc.size() * FrequentWordSourceVoc.size());
			    dynScalingFactor.push_back(&FrequentWordScalingFactors[index*numberOfExperts]);
			    dynScalingIndex.push_back(dynOffset + index);
			  }
			  dynOffset += FrequentWordTargetVoc.size() * FrequentWordSourceVoc.size();
			}
			//cout << "Orig LP" << p[numberOfLexicons] << endl;
			assert(i*TargetLength+j < SourceLength*TargetLength);
			assert(dynScalingFactor.size() == dynScalingIndex.size());
			features[i*TargetLength+j].Init(G->GetNode(i,j),p,ScalingFactors,NumberOfFeatures/2,dynScalingFactor,dynScalingIndex,offset, numberOfAllFeatures,numberOfExperts,expertWeights);
		}

	}
	delete[] linkFeatureValues;
	delete[] lexMatrix;

}

void LocalLOPFeatureSet::RemoveFeatures() {
	delete[] features;
}
FeatureSetIterator * LocalLOPFeatureSet::GetIterator ( ) const {
	return new LocalLOPFeatureSetIterator(SourceLength,TargetLength,features);
};


void LocalLOPFeatureSet::EstimateScalingFactores() {
  EstimatedDerivation.clear();
  vector<double> ExpectValue;
  vector<double> ExampleValue;
  ExpectValue.resize(numberOfExperts,0);
  ExampleValue.resize(numberOfExperts,0);
  
  for(int i = 0; i < SourceLength * TargetLength;i++) {
    vector<double> TempExpectValue = features[i].GetExpectationValue();
    vector<double> TempExampleValue = features[i].GetExampleValue();
    assert(TempExpectValue.size() == numberOfExperts);
    assert(TempExampleValue.size() == numberOfExperts);
    for(int j = 0; j < numberOfExperts; j++) {
      //cout << "New ExpectValue:" << ExpectValue[j] << " New ExampleValue: " << ExampleValue[j] <<endl;
      ExpectValue[j] += TempExpectValue[j];
      ExampleValue[j] += TempExampleValue[j];
    }
  }
  for(int j = 0; j < numberOfExperts; j++) {
   double deltaL = ExampleValue[j] - ExpectValue[j];
   //cout << "Calculating Der[" << j << "]: " << ExampleValue[j] << " - " <<  ExpectValue[j] << " = " << deltaL << endl;
    EstimatedDerivation.push_back(deltaL);
  }

}

double * LocalLOPFeatureSet::InitLinkFeatures(vector<vector<LinkFeatures> > & linkFeatures) {
	if(numberOfLinkFeatures == 0)
		return new double[1];
	assert(numberOfLinkFeatures == linkFeatures.size());
	double * featureMatrix = new double [SourceLength * TargetLength * numberOfLinkFeatures];
	for(int i = 0; i < SourceLength * TargetLength * numberOfLinkFeatures; i++) {
		featureMatrix[i] = 0;
	}
	for(int i = 0; i < linkFeatures.size(); i++) {
		for(int j = 0; j < linkFeatures[i].size();j++) {
			int sourceIndex = linkFeatures[i][j].source;
			int targetIndex = linkFeatures[i][j].target;
			int index = sourceIndex * (TargetLength * numberOfLinkFeatures) + targetIndex * numberOfLinkFeatures + i;
			if(index >= SourceLength * TargetLength * numberOfLinkFeatures) {
				//cout << sourceIndex << " " << targetIndex << " " << index << endl;
				//cout << SourceLength << " " << TargetLength << " " << numberOfLinkFeatures << endl;
			}
			assert(index < SourceLength * TargetLength * numberOfLinkFeatures);
			featureMatrix[index] = linkFeatures[i][j].value;
		}
	}
	
	return featureMatrix;
}

double * LocalLOPFeatureSet::CreateLexMatrix(int & numberOfLexMatrix,Graph * g) {
	SourceLength = g->GetSourceLength();
	TargetLength = g->GetTargetLength();	
	numberOfLexMatrix = numberOfLexicons + additionalLexFeatures;
	double * lexMatrix = new double[numberOfLexMatrix * SourceLength * TargetLength];
	//cout << "Original:" << endl;
	for(int i = 0; i < SourceLength; i++) {
		for(int j = 0; j < TargetLength; j++) {
			const string Source = g->GetSourceWord(i);
			const string Target = g->GetTargetWord(j);
			for(int k = 0; k < numberOfLexicons; k++) {
				map<pair<string,string>,double>::iterator iter = Lexicon[k].find(make_pair(Source,Target));
				int index = k * SourceLength * TargetLength + i * TargetLength + j;
				if(iter != Lexicon[k].end()) {
					lexMatrix[index] = (iter->second);
				}else {
					lexMatrix[index] = 0.00001;
				}
				//cout << lexMatrix[index] <<  " ";
				//cout << "Store item[" << i << "][" << j << "][" << k << "] in index " << index << ": " << lexMatrix[index] << endl;
			}
		}
		//cout << endl;

	}
	int k = numberOfLexicons;
	for(int l = 0; l < numberOfLexicons; l++) {
		if(SourceNormalized[l] != 0) {
			//cout << "Source: " << endl;
			for(int j = 0; j < TargetLength; j++) {
				double sum = 0;
				for(int i = 0; i < SourceLength; i++) {
					int index = l * SourceLength * TargetLength + i * TargetLength + j;
					sum += lexMatrix[index];
				}
				for(int i = 0; i < SourceLength; i++) {
					int index = l * SourceLength * TargetLength + i * TargetLength + j;
					int newIndex = k * SourceLength * TargetLength + i * TargetLength +j;
					lexMatrix[newIndex]= lexMatrix[index]/sum;
					//cout << lexMatrix[newIndex] << " ";
				}
				//cout << endl;
			}
			k++;
		}
	}
	for(int l = 0; l < numberOfLexicons; l++) {
		if(TargetNormalized[l] != 0) {
			//cout << "Target: " << endl;
			for(int i = 0; i < SourceLength; i++) {
				double sum = 0;
				for(int j = 0; j < TargetLength; j++) {
					int index = l * SourceLength * TargetLength + i * TargetLength + j;
					sum += lexMatrix[index];
				}
				for(int j = 0; j < TargetLength; j++) {
					int index = l * SourceLength * TargetLength + i * TargetLength + j;
					int newIndex = k * SourceLength * TargetLength + i * TargetLength +j;
					lexMatrix[newIndex]= lexMatrix[index]/sum;
					//cout << lexMatrix[newIndex] << " ";				
				}
				//cout << endl;
			}
			k++;
		}
	}
	
	return lexMatrix;

}
