#include <stdlib.h>
#include <stdio.h>
// _AccessNgramDB.cpp: implementation of the C_AccessNgramDB class.
//
//////////////////////////////////////////////////////////////////////

#include "_AccessNgramDB.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

C_AccessNgramDB::C_AccessNgramDB(char * filenameStem, int maxN)
{

	this->voc = new C_VocabularyDB(filenameStem, true, true);
	this->maxN = maxN;
	

	//Initialize DBs for n-gram frequencies
	char dbFileName[1024];
	unsigned int openFlag = DB_RDONLY;	//read-only when accessing the DB

	for(int n=0;n<this->maxN; n++){
		sprintf(dbFileName, "%s.%dgram.db", filenameStem, n+1);

		Db * ngramDb = new Db(NULL, 0);

		try{
			ngramDb->open(NULL,
				dbFileName,
				NULL,
				DB_BTREE,
				openFlag,
				0);
		}
		catch (DbException & e){
			cerr<<"Error while opening the ngram DB file "<<dbFileName<<endl;
			cerr<<"Error information: "<<e.get_errno()<<endl;
			exit(e.get_errno());
		}

		this->ngramDBs.push_back(ngramDb);
	}

}

C_AccessNgramDB::~C_AccessNgramDB()
{
	delete(this->voc);

	for(int i=0; i< this->ngramDBs.size(); i++){
		try{
			this->ngramDBs[i]->close(0);
			delete(this->ngramDBs[i]);
		}
		catch (DbException & e){
			cerr<<"Error while closing the "<<i+1<<"-gram DB file "<<endl;
			cerr<<"Error information: "<<e.get_errno()<<endl;
			exit(e.get_errno());
		}
	}
}

/**
* Search for n-gram and return the log frequency if exist
**/
bool C_AccessNgramDB::returnLogFreqOfNgram(char *ngramAsCharBuffer, int bufferSize, int sizeOfn, T_LOG_FREQ_INFO &logFreq)
{
	Dbt key(ngramAsCharBuffer,bufferSize);
	key.set_flags(DB_DBT_USERMEM);
	key.set_ulen(bufferSize);

	Dbt value(&logFreq, sizeof(T_LOG_FREQ_INFO));
	value.set_flags(DB_DBT_USERMEM);
	value.set_ulen(sizeof(T_LOG_FREQ_INFO));
	
	int result = this->ngramDBs[sizeOfn-1]->get(NULL, &key, &value, 0);

	if(result!=0){
		return false;
	}

	return true;
}

/**
* This function access the cache to speed up
**/
bool C_AccessNgramDB::returnLogFreqOfNgram(vector<IndexType> ngram, T_LOG_FREQ_INFO &logFreq)
{
	long currentTime;

	//first check the cache
	map< vector<IndexType>, S_CachedInfo>::iterator iterCache;

	iterCache = this->cache.find(ngram);
	if(iterCache!=this->cache.end()){
		//exist in cache already

		//update the last visited time		
		time(&currentTime);
		iterCache->second.lastTimeVisited = currentTime;
		logFreq = iterCache->second.logFreq;
		return true;
	}
	
	//else, not exist in cache yet
	int sizeOfn = ngram.size();

	C_NgramFreqInfo ngramInfo;

	for(int i=0;i<sizeOfn; i++){
		ngramInfo.addWord(ngram[i]);
	}

	int sizeOfBufferUsed;
	ngramInfo.convertNgramAsVocIdToCharBuffer(this->buffer1, sizeOfBufferUsed);

	bool successfullyAccess = this->returnLogFreqOfNgram(this->buffer1, sizeOfBufferUsed, sizeOfn, logFreq);

	if(successfullyAccess){
		//add the n-gram into the cache
		S_CachedInfo tmpNode;
		time(&currentTime);

		tmpNode.logFreq = logFreq;
		tmpNode.lastTimeVisited = currentTime;

		this->cache.insert(make_pair(ngram, tmpNode));
		
	}

	return successfullyAccess;
}

/**
* Input ngram as a vector of C_String, return the log freq if the ngram exists
**/
bool C_AccessNgramDB::returnLogFreqOfNgram(vector<C_String> ngram, T_LOG_FREQ_INFO &logFreq)
{
	int sizeOfn = ngram.size();
	C_NgramFreqInfo ngramInfo;
	for(int i=0;i<sizeOfn;i++){
		IndexType wordId = this->voc->returnId(ngram[i].toString(), ngram[i].length());
		ngramInfo.addWord(wordId);
	}

	int sizeOfBufferUsed;
	ngramInfo.convertNgramAsVocIdToCharBuffer(this->buffer1, sizeOfBufferUsed);

	return this->returnLogFreqOfNgram(this->buffer1, sizeOfBufferUsed, sizeOfn, logFreq);
}


/**
* Access the DB and return the vocID of a word
**/
IndexType C_AccessNgramDB::returnId(char *text, int textLen)
{
	return this->voc->returnId(text, textLen);
}

IndexType C_AccessNgramDB::returnVocSize()
{
	return this->voc->returnVocSize();
}

/**
* Purge entries in the cache that have not been visted since lastTimeCalled
**/
void C_AccessNgramDB::purgeCache(long lastTimeCalled)
{
	cerr<<"Size of cache: "<<this->cache.size()<<" ";
	map< vector<IndexType>, S_CachedInfo>::iterator iterCache1, iterCache2;

	iterCache1 = this->cache.begin();

	while(iterCache1 != this->cache.end() ){
		iterCache2=iterCache1;
		iterCache2++;

		if(iterCache1->second.lastTimeVisited<lastTimeCalled){
			this->cache.erase(iterCache1);
		}

		iterCache1=iterCache2;;
	}

	cerr<<" purged to: "<<this->cache.size()<<" entries\n";
}
