#include <memory.h>
#include <string.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
// _NgramDBLM.cpp: implementation of the C_NgramDBLM class.
//
//////////////////////////////////////////////////////////////////////

#include "_NgramDBLM.h"
#include "_NgramDBLMSentenceState.h"

#include "malloc.h"
#include "time.h"
#include <fstream>
#include <assert.h>

using namespace std;


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


/**
* Initialize the DBLM with configuration specified in the cfgFile
**/
C_NgramDBLM::C_NgramDBLM(const char *cfgFileName)
{
	
	this->sentProcessed = 0;

	fstream cfgFile;
	cfgFile.open(cfgFileName,ios::in);

	if(!cfgFile){
		fprintf(stderr,"Configuration file does not exist! quit!!\n");
		exit(0);
	}

	//-----------------------------------------------------------------------------
	//reading parameters
	char paraName[1024];

	char dbFileNameStem[1024];
	dbFileNameStem[0]=0;

	this->corpusSize = 0;
	this->maxLenOfNgramConsidered = 0;
	this->interpolationStrategy = 'e';	//default, equal weighted
	this->freshTime=10;					//default, 10 seconds, entries not used within 10 seconds will be purged
	this->numberOfSentencesProcessedBeforeCachedPurge = 10;		//purge cache every 10 sentences


	char sentenceStartString[1024];
	memset(sentenceStartString, 1024, 0);

	char sentenceEndString[1024];
	memset(sentenceEndString, 1024, 0);

	while(!cfgFile.eof()){
		cfgFile>>paraName;

		if(strcmp(paraName,"NGRAM_DB")==0){
			cfgFile>>dbFileNameStem;
		}
		else if(strcmp(paraName,"CORPUS_SIZE")==0){
			cfgFile>>this->corpusSize;
		}
		else if(strcmp(paraName,"N")==0){
			cfgFile>>this->maxLenOfNgramConsidered;
		}
		else if(strcmp(paraName,"SENT_START")==0){
			cfgFile>>sentenceStartString;
		}
		else if(strcmp(paraName,"SENT_END")==0){
			cfgFile>>sentenceEndString;			
		}
		else if(strcmp(paraName,"INTERPOLATION_STRATEGY")==0){
			cfgFile>>this->interpolationStrategy;
		}
		else if(strcmp(paraName,"FRESH_TIME")==0){
			cfgFile>>this->freshTime;
		}
		else if(strcmp(paraName, "PURGE_CACHE")==0){
			cfgFile>>this->numberOfSentencesProcessedBeforeCachedPurge;
		}

		paraName[0]='\0';
	}

	if(this->maxLenOfNgramConsidered==0){
		cerr<<"Must specify the max order of n-gram available in the database, keyword: N\n";
		exit(-1);
	}

	if(strlen(dbFileNameStem)==0){
		cerr<<"Must specify the filename stem of the n-gram DB, keyword: NGRAM_DB\n";
		exit(-1);
	}

	this->dbAccess = new C_AccessNgramDB(dbFileNameStem, this->maxLenOfNgramConsidered);

	if(this->corpusSize==0){
		cerr<<"Must specify the corpus size in cfg file, keyword: CORPUS_SIZE\n";
		exit(-1);
	}

	if(strlen(sentenceStartString)==0){
		cerr<<"Must specify the word used in the corpus for sentence start, keyword: SENT_START\n";
		exit(-1);
	}

	this->sentStartVocId = this->dbAccess->returnId(sentenceStartString, strlen(sentenceStartString));
	if(this->sentStartVocId==0){
		cerr<<"Specified sentence start symbol: "<<sentenceStartString<<" can not be found in the vocabuary!\n";
		exit(-1);
	}


	if(strlen(sentenceEndString)==0){
		cerr<<"Must specify the word used in the corpus for sentence end, keyword: SENT_END\n";
		exit(-1);
	}

	this->sentEndVocId = this->dbAccess->returnId(sentenceEndString, strlen(sentenceEndString));
	if(this->sentEndVocId==0){
		cerr<<"Specified sentence end symbol: "<<sentenceEndString<<" can not be found in the vocabuary!\n";
		exit(-1);
	}

	cerr<<"DBLM initilized\n";
}

C_NgramDBLM::~C_NgramDBLM()
{
	delete(this->dbAccess);
}

void C_NgramDBLM::setParam_interpolationStrategy(char interpolationStrategy)
{
	this->interpolationStrategy = interpolationStrategy;	
}


void C_NgramDBLM::setParam_numberOfSentencesProcessedBeforeCachePurge(int sentNumber)
{
	this->numberOfSentencesProcessedBeforeCachedPurge = sentNumber;
}

void C_NgramDBLM::setParam_freshnessOfCacheToAvoidPurge(long freshTime)
{
	this->freshTime = freshTime;
}

void C_NgramDBLM::setParam_maxLenOfNgramConsidered(int maxLenOfNgramConsidered)
{
	this->maxLenOfNgramConsidered = maxLenOfNgramConsidered;
}

////freq: 1-gram Freq, corpusSize, 2-gram freq, freq of 2-gram history
//	    3-gram freq, freq of 3-gram history
//freqTable should have length of 2*n
//
double C_NgramDBLM::LogProb(vector<LMWordIndex> & history, LMWordIndex nextWord, vector<LMWordIndex> & nextStateAsNgram)
{
	long currentTime;

	//first, check if history+nextWord already exist in the cache
	vector<LMWordIndex> historyWithNextWord = history;
	historyWithNextWord.push_back(nextWord);

	map< vector<LMWordIndex>, S_CachedNgramLogProb>::iterator iterCache;
	iterCache = this->cache.find(historyWithNextWord);

	int historyLen = history.size();
	int longest_n_matched = 0;	//longest n-gram with next word that can be matched
	double logProb;
	if(iterCache==this->cache.end()){
		//does not exist in the cache yet

		double * freqTable = (double *) malloc(sizeof(double)*2*(this->maxLenOfNgramConsidered));
		memset(freqTable, 0, 2*this->maxLenOfNgramConsidered*sizeof(double));

		//vector<LMWordIndex> ngram; (removing this; LMInterface wants vector< IndexType > -apd)
    vector< IndexType > ngram;
		bool ngramFound;
		T_LOG_FREQ_INFO ngramLogFreq;

		freqTable[1]=this->corpusSize;

		//find unigram frequency
		ngram.clear();
		ngram.push_back((IndexType)nextWord);
		ngramFound = this->dbAccess->returnLogFreqOfNgram(ngram, ngramLogFreq);
		
		if(!ngramFound){	//this word is OOV			
			longest_n_matched = 0;
			//and freq = 0 for unigram			
		}
		else{
			longest_n_matched = 1;	//longest n-gram with next word that can be matched

			freqTable[0]=exp(ngramLogFreq);

			int max_n_in_this_case = history.size()+1;
			if(max_n_in_this_case > this->maxLenOfNgramConsidered){
				max_n_in_this_case = this->maxLenOfNgramConsidered;
			}

			bool stillExist = true;
			int n=2;
			while(stillExist && (n<=max_n_in_this_case)){
				//construct ngram history
				ngram.clear();
				for(int j=(historyLen-n+1);j<historyLen;j++){	//n-1 words as history
					ngram.push_back(history[j]);
				}

				T_LOG_FREQ_INFO logFreqOfHistory;
				stillExist = this->dbAccess->returnLogFreqOfNgram(ngram, logFreqOfHistory);

				if(stillExist){
					ngram.push_back(nextWord);
					stillExist = this->dbAccess->returnLogFreqOfNgram(ngram, ngramLogFreq);
				}

				if(stillExist){
					freqTable[2*n-2]=exp(ngramLogFreq);
					freqTable[2*n-1]=exp(logFreqOfHistory);

					longest_n_matched = n;
				}

				n++;
			}
		}	
		
		logProb = this->calcLogProb(freqTable);

		//add into the cache
		S_CachedNgramLogProb tmpNode;
		time(&currentTime);
		tmpNode.lastVisitedTimeStamp = currentTime;
		tmpNode.logProb = logProb;
		tmpNode.longestMatchingHistory = longest_n_matched;
		this->cache.insert(make_pair(historyWithNextWord, tmpNode));

		free(freqTable);
	}
	else{	//history plus next word exist in the cache, update the time stamp
		time(&currentTime);
		logProb = iterCache->second.logProb;
		longest_n_matched = iterCache->second.longestMatchingHistory;
		iterCache->second.lastVisitedTimeStamp = currentTime;
	}

	//create "next state"
	//if 5-gram including the next word is matched, 
	//create a memoryistory with 3-gram from prev history plus next word as the memoryistory
	if(longest_n_matched>0){
		if(longest_n_matched>(this->maxLenOfNgramConsidered-1)){
			longest_n_matched = this->maxLenOfNgramConsidered-1;
		}
		nextStateAsNgram.clear();		
		for(int k=(historyLen-longest_n_matched+1);k<historyLen;k++){
			nextStateAsNgram.push_back(history[k]);
		}
		nextStateAsNgram.push_back(nextWord);
	}
	else{	//longest_n_matched = 0, even the nextword as unigram can not be matched
		nextStateAsNgram.clear();
	}
	return logProb;

}

double C_NgramDBLM::calcLogProb(double *freq)
{
	switch(this->interpolationStrategy){
	case 'e':
		return this->calcLogProb_equalWeightedInterpolation(freq);
		break;
	case 'i':
		return this->calcLogProb_ibmHeuristicInterpolation(freq);
		break;
	case 'm':
		return this->calcLogProb_maxProbInterpolation(freq);
		break;
	default:
		cerr<<"Unknown interpolation strategy!\n";
		exit(0);
	}
}

double C_NgramDBLM::calcLogProb_equalWeightedInterpolation(double *freq)
{
	double prob = 0.0;

	
	if(freq[0]>0){

		int i=0;
		bool stillMatched = true;

		while(stillMatched && (i<this->maxLenOfNgramConsidered)){
			if(freq[2*i]>0){
				prob+=freq[2*i]/freq[2*i+1];
			}
			else{
				stillMatched = false;
			}

			i++;
		}

		return log(prob/(double)this->maxLenOfNgramConsidered);
	}
	else{	//unknown word
		return DBLM_LOG_PROB_UNK;
	}
}

double C_NgramDBLM::calcLogProb_ibmHeuristicInterpolation(double *freq)
{
	double prob = 0.0;
	if(freq[0]==0){	//unknown word
		return DBLM_LOG_PROB_UNK;
	}

	double remainingWeightSum = 1.0;

	//find the first non-zero match
	int i = this->maxLenOfNgramConsidered - 1;

	while(freq[2*i]==0){	//will stop for sure because freq[0]!=0
		i--;
	}

	for(int j=i;j>=0;j--){
		//for (j+1)-gram
		double historyFreq = freq[2*j+1];
		double logHistoryFreq = log(historyFreq);
		if(logHistoryFreq>1){
			logHistoryFreq = 1.0;	//cap it to 1
		}

		double reliability = 0.1*logHistoryFreq+0.3;	//heuristics for reliability of the history
		double adjustedWeights = remainingWeightSum *  reliability;
		
		prob+=adjustedWeights * freq[2*i]/freq[2*i+1];

		remainingWeightSum -= adjustedWeights;
	}

	return log(prob);	
}



double C_NgramDBLM::calcLogProb_maxProbInterpolation(double *freq)
{
	double maxProb = 0.0;
	
	if(freq[0]>0){

		int i=0;
		bool stillMatched = true;

		while(stillMatched && (i<this->maxLenOfNgramConsidered)){
			if(freq[2*i]>0){
				double prob=freq[2*i]/freq[2*i+1];

				if(prob>maxProb){
					maxProb = prob;
				}
			}
			else{
				stillMatched = false;
			}

			i++;
		}

		return log(maxProb);
	}
	else{	//unknown word
		return DBLM_LOG_PROB_UNK;
	}
}

CLMSentenceState * C_NgramDBLM::NewSentence() const
{

	return new C_NgramDBLMSentenceState(this);

}


/**
* This function should be better put inside NewSentence()
* But NewSentence() isa const method 
* So we have to creat another method to be called by the SentenceState object
* Awkward, but given the interface as defined, there is no other solution
**/
void C_NgramDBLM::StartsAnotherSentenceNow()
{
	this->sentProcessed++;

	if(this->sentProcessed>=this->numberOfSentencesProcessedBeforeCachedPurge){
		this->sentProcessed=0;
		long currentTime;
		time(&currentTime);
		this->purgeCache(currentTime-this->freshTime);
	}
}

LMWordIndex C_NgramDBLM::returnSentStartVocId()
{
	return this->sentStartVocId;
}

LMWordIndex C_NgramDBLM::returnSentEndVocId()
{
	return this->sentEndVocId;
}


LMWordIndex C_NgramDBLM::GetVocabularyIndex(string word)
{
	return dbAccess->returnId((char *) word.c_str(), word.length());
}


const vector< string > & 
C_NgramDBLM::Vocabulary() const
{
	assert( 0 );
	vector< string > EmptyVector;
}

LMWordIndex C_NgramDBLM::VocabularySize() const
{
	return this->dbAccess->returnVocSize();
}

/**
* Purge those entries in the cache that have not been visited after 'lastVisitedTime'
**/
void C_NgramDBLM::purgeCache(long lastVisitedTime)
{
	this->dbAccess->purgeCache(lastVisitedTime);

	//purge the n-gram logprob cache
	map< vector<LMWordIndex>, S_CachedNgramLogProb>::iterator iter1, iter2;

	cerr<<"DBLM cache: purged from "<<this->cache.size()<<" entries to: ";
	iter1 = this->cache.begin();
	while(iter1!=this->cache.end()){
		iter2=iter1;
		iter2++;

		if(iter1->second.lastVisitedTimeStamp<lastVisitedTime){
			this->cache.erase(iter1);
		}

		iter1=iter2;
	}

	cerr<<this->cache.size()<<" entries.\n";
}
