#include <memory.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
// _VocabularyDB.cpp: implementation of the C_VocabularyDB class.
//
//////////////////////////////////////////////////////////////////////

#include "_VocabularyDB.h"
#include <iostream>
#include <fstream>
#include <string>

using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

C_VocabularyDB::C_VocabularyDB(char * filename, bool readonly, bool useCache)
{
	this->readonly = readonly;
	this->useCache = useCache;

	this->cacheText2Id.clear();

	unsigned int openFlag = DB_CREATE;	//if the DB does not exist by now, create it

	if(readonly){
		openFlag = DB_RDONLY;			//read-only mode
	}

	this->id2text_DB = new Db(NULL, 0);	//instantiate a DB object
	this->text2id_DB = new Db(NULL, 0);
	
	char id2text_db_filename[4096];
	char text2id_db_filename[4096];

	sprintf(id2text_db_filename, "%s.voc.id2text", filename);
	sprintf(text2id_db_filename, "%s.voc.text2id", filename);

	try{
		this->id2text_DB->open(NULL,
			id2text_db_filename,
			NULL,
			DB_BTREE,
			openFlag,
			0);
	}
	catch (DbException & e){
		cerr<<"Error while opening the ID2TEXT DB file "<<id2text_db_filename<<endl;
		cerr<<"Error information: "<<e.get_errno()<<endl;
		exit(e.get_errno());
	}

	try{
		this->text2id_DB->open(NULL,
			text2id_db_filename,
			NULL,
			DB_BTREE,
			openFlag,
			0);
	}
	catch (DbException & e){
		cerr<<"Error while opening the TEXT2ID DB file "<<text2id_db_filename<<endl;
		cerr<<"Error information: "<<e.get_errno()<<endl;
		exit(e.get_errno());
	}


	//check if we already have entries in the vocabulary
	memset(this->buffer, 0, MAX_TOKEN_LEN);
	sprintf(this->buffer, "__MAX_ID_IN_VOC__"); 


	int exist = this->readText2Id(this->buffer, 17, this->lastAvailabelVocId);

	if(exist==0){	//there are already entries in the voc

	}
	else{
		this->lastAvailabelVocId=1;	//reserve vocId=0 for NULL, or OOV
	}
	
}

C_VocabularyDB::~C_VocabularyDB()
{
	if(! this->readonly){
		memset(this->buffer, 0, MAX_TOKEN_LEN);
		sprintf(this->buffer, "__MAX_ID_IN_VOC__"); 
		this->writeText2Id(this->buffer, 17, this->lastAvailabelVocId);

		try{
			this->id2text_DB->close(0);
			delete(this->id2text_DB);
		}
		catch(DbException &e){
			cerr<<"Error while closing the vocabulary ID2TEXT DB\n";
			cerr<<"Error information: "<<e.get_errno()<<endl;
			exit(e.get_errno());
		}

		try{
			this->text2id_DB->close(0);
			delete(this->text2id_DB);
		}
		catch(DbException &e){
			cerr<<"Error while closing the vocabulary TEXT2ID DB\n";
			cerr<<"Error information: "<<e.get_errno()<<endl;
			exit(e.get_errno());
		}
	}
	else{	//vocabulary is read only
		try{
			this->id2text_DB->close(DB_NOSYNC);
			delete(this->id2text_DB);
		}
		catch(DbException &e){
			cerr<<"Error while closing the vocabulary ID2TEXT DB\n";
			cerr<<"Error information: "<<e.get_errno()<<endl;
			exit(e.get_errno());
		}

		try{
			this->text2id_DB->close(DB_NOSYNC);
			delete(this->text2id_DB);
		}
		catch(DbException &e){
			cerr<<"Error while closing the vocabulary TEXT2ID DB\n";
			cerr<<"Error information: "<<e.get_errno()<<endl;
			exit(e.get_errno());
		}
	}
}


int C_VocabularyDB::writeId2Text(IndexType id, char *text, int textLen)
{
	Dbt key(&id, sizeof(IndexType));
	key.set_flags(DB_DBT_USERMEM);
	key.set_ulen(sizeof(IndexType));

	memset(this->buffer, 0, MAX_TOKEN_LEN);
	sprintf(this->buffer, "%s\0", text);
	Dbt value(this->buffer, textLen+1);
	value.set_flags(DB_DBT_USERMEM);
	value.set_ulen(textLen+1);

	int result;

	try{
		result = this->id2text_DB->put(NULL, &key, &value, 0);
	}
	catch(DbException &e){
		cerr<<"Error while writing the ID2TEXT DB of the vocabulary!"<<endl;
		exit(1);
	}

	return result;
}

int C_VocabularyDB::writeText2Id(char *text, int textLen, IndexType id)
{
	if(textLen>=MAX_TOKEN_LEN){
		cerr<<"Word too long."<<endl;
		exit(1);
	}

	memset(this->buffer, 0, MAX_TOKEN_LEN);
	sprintf(this->buffer, "%s\0", text);	//to make sure that we have a correct ending

	Dbt key(this->buffer, textLen+1);
	key.set_flags(DB_DBT_USERMEM);
	key.set_ulen(textLen+1);

	Dbt value(&id, sizeof(IndexType));
	value.set_flags(DB_DBT_USERMEM);
	value.set_ulen(sizeof(IndexType));

	int result;
	
	try{
		result = this->text2id_DB->put(NULL, &key, &value, 0);
	}
	catch(DbException &e){
		cerr<<"Error while writing the TEXT2ID DB of the vocabulary!"<<endl;
		exit(1);
	}

	return result;
}

int C_VocabularyDB::readText2Id(char *text, int textLen, IndexType &id)
{
	if(textLen>=MAX_TOKEN_LEN){
		cerr<<"Word too long."<<endl;
		exit(1);
	}

	memset(this->buffer, 0, MAX_TOKEN_LEN);
	sprintf(this->buffer,"%s\0",text);
	
	Dbt key(this->buffer, textLen+1);
	key.set_flags(DB_DBT_USERMEM);
	key.set_ulen(textLen+1);

	Dbt value(&id, sizeof(IndexType));
	value.set_flags(DB_DBT_USERMEM);
	value.set_ulen(sizeof(IndexType));

	int result;
	
	try{
		result=this->text2id_DB->get(NULL, &key, &value, 0);
	}
	catch(DbException &e){
		cerr<<"Error while reading the TEXT2ID DB of the vocabulary!"<<endl;
		exit(1);
	}

	return result;
}

int C_VocabularyDB::readId2Text(IndexType id, char *text, int &textLen)
{
	Dbt key(&id, sizeof(IndexType));
	key.set_flags(DB_DBT_USERMEM);
	key.set_ulen(sizeof(IndexType));

	Dbt value(text, textLen);
	value.set_flags(DB_DBT_USERMEM);
	value.set_ulen(textLen);

	int result;
	
	try{
		result = this->id2text_DB->get(NULL, &key, &value, 0);
	}
	catch(DbException &e){
		cerr<<"Error while reading the ID2TEXT DB of the vocabulary!"<<endl;
		exit(1);
	}

	textLen = strlen(text);

	return result;
}

IndexType C_VocabularyDB::returnId(char *text, int textLen)
{
	IndexType id;

	//first check if the word already in cache
	bool needToInsertIntoCache = false;
	if(this->useCache){
		map<C_String, IndexType>::iterator iterCachedText2Id = this->cacheText2Id.find(C_String(text, textLen));
		if(iterCachedText2Id==this->cacheText2Id.end()){
			//this word does not exist in the cache yet
			needToInsertIntoCache = true;
		}
		else{	//already in cache
			id = iterCachedText2Id->second;
			return id;
		}
	}

	int result = this->readText2Id(text, textLen, id);

	if(result!=0){		//the word does not exist yet
		return 0;		//0 for OOV
	}

	if(needToInsertIntoCache){
		//word exist in DB but not in mem
		this->cacheText2Id.insert(make_pair(C_String(text, textLen), id));
	}

	return id;
}

/**
* Get the Id for word(text), if the word does not exist in the voc, assign an id and added it to the vocabulary
**/
IndexType C_VocabularyDB::getId(char *text, int textLen)
{
	IndexType id;

	id = this->returnId(text, textLen);

	if(id==0){		//the word does not exist yet
		id=this->lastAvailabelVocId;
		
		this->lastAvailabelVocId++;

		if(this->lastAvailabelVocId==(IndexType)-2){
			cerr<<"Vocabulary size has exceed the limit of "<<(IndexType)-1<<endl;
			exit(-1);
		}
		
		this->writeText2Id(text, textLen, id);
		this->writeId2Text(id, text, textLen);

		if(this->useCache){
			this->cacheText2Id.insert(make_pair(C_String(text), id));
			this->cachedId2Text.insert(make_pair(id, C_String(text)));
		}

	}

	return id;
}

/**
* Return the text of the word given the ID
* the memory of *text has already been allocated from the calling function
* If the ID does not exist, return textLen=0
**/
int C_VocabularyDB::returnText(IndexType id, char *text)
{
	bool needToInsertIntoCache = false;

	if(this->useCache){
		//then check if the word pair is already in cache
		map<IndexType, C_String>::iterator iterCachedId2Text = this->cachedId2Text.find(id);

		if(iterCachedId2Text==this->cachedId2Text.end()){	//does not exist in cache yet
			needToInsertIntoCache = true;
		}
		else{
			text=iterCachedId2Text->second.toString();			
			return iterCachedId2Text->second.length();
		}
	}

	memset(this->buffer, 0, MAX_TOKEN_LEN);
	
	int textLen;
	int result = this->readId2Text(id, text, textLen);

	if(result!=0){	//not successful
		textLen = 0;
	}
	else{
		if(needToInsertIntoCache){
			this->cachedId2Text.insert(make_pair(id, C_String(text, textLen)));
		}
	}

	return textLen;
}

IndexType C_VocabularyDB::returnVocSize()
{
	return this->lastAvailabelVocId;
}
