#ifndef __DistortionModel__
#define __DistortionModel__


/**
   Class CDistortionModel

   Base class for distortion models
   Realized the interface which allows the decoder to access specialized distortion models.
   Essentially, a distortion models is p( distortion | dependencies )
   The distortion can by a jump on source or target side, the difference between 
   source word position and target word position, or just the indication to go forward
   or backwards.  The probabilities can depend on any information from the sentences,
   but typically it will be sentence length (original IBM distortion model), on words,
   or word classes.  Here, we realize lexicalized distortion models, i.e. we allow
   dependencies on the word identities.  There can be unigram models (prob depends on 
   one word) and bigram models (prob depends on two words).
   The models typically need to apply some smoothing to deal with unseen and low
   frequency events.

   Jump models:                                                            Informaiton needed
   - Jumps on the source side, consecutive positions on the target side
     (this is what the current decoder does)
     - source word outbound                                                jump, source word left
     - source word inbound                                                 jump, source word right
     - target word outbound                                                jump, target word left
     - target word inbound                                                 jump, target word right

   Distance models:
   - Distance between source word and target word
     - first word in source phrase
     - first word in target phrase

   Absolute position
   - absolute position of target word                                      position, target word, expected length
   - ditto, but normalized by (expected) sentence length                   position, target word, expected length
   - categorized position: first, beginning, midddle, end, last            position, target word, expected length

   
   We specify the type of the distortion model in a header
   Model JumpSourceOutbound|JumpSourceInbound|JumpTargetOutbound|JumpTargetInbound
   MaxJump nnn    ( 
*/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include <SparseArray.hh>
#include <NumericalArray.hh>
#include <Vocabulary.hh>




class CDistortionModel
{
public:
  //! the type of a distortion model
  enum DMType { JumpSourceOutbound = 1,
                JumpSourceInbound,
                JumpTargetOutbound,
                JumpTargetInbound };

  //typedef DMTypes TDistortionModelType;



  //! create a distortion model and give it a vocabulary
  CDistortionModel( CVocabulary* pVocabulary );

  //! get the type of the distortion model
  DMType DistortionModelType() const;


  //! provide lexicon, which is used to convert between word and word ID
  void SetVocab( CVocabulary* pVocabulary );

  //! set smoothing factor (before loading of model, will not be effective later)
  void SetSmoothingFactor( float SmoothingFactor );


  //! load distortion model from file
  virtual bool Load( const char* FileName );

  //! store distortion model from file
  virtual bool Store( const char* FileName );


  //! return the probability for a distorion (jump)
  virtual float LogProb( int JumpWidth );

  //! return distortion probability; all words provided, distortion models knows what to use
  virtual float LogProb( int JumpWidth,
                         CWordIndex FromSourceWord, CWordIndex FromTargetWord,
                         CWordIndex ToSourceWord,   CWordIndex ToTargetWord );
  

  //! report some information about the distortion model
  virtual void Report();


protected:
  void Init();

  //! return the probability for a distorion (jump) and a word index
  virtual float LogProb( int JumpWidth, CWordIndex Word );

  //! return the probability for a distorion (jump) and a word index
  virtual float LogProb( int JumpWidth, CWordIndex Word1, CWordIndex Word2 );


  //! type of the distortion model
  DMType m_DistortionModelType;

  //! pointer to the vocabulary
  CVocabulary* m_pVoc;

  //! a matrix, which stores the log probabilities which depend on one word
  // first dimension is word index, second dimension is distorion
  CNumericalMatrix< float > *m_pLogProbs;

  //! a sparce matrix to store probabilities which depend on two words
  CSparseMatrix *m_pLogProbs2;

  //! the background model, a normal distribution or an accumulated distribution
  CNumericalArray< float > *m_pNormalDistribution;

  //! accumulated distribution, i.e. counts summed over all words
  CNumericalArray< float > *m_pGlobalDistribution;

  //! the number of time each words has been seen, i.e. the sum of the different Count( Jump )
  //CNumericalArray< float > *m_pTotalCounts;

  //! a smoothing factor
  float m_SmoothingFactor;

};





#endif














