#ifndef __NBEST__
#define __NBEST__

#include <config.h>
#include <fortify.h>


#if __GNUC__ == 3 && __GNUC_MINOR__ > 0
#include <ext/hash_map>
#include <ext/hash_set>
namespace sgi {
  using __gnu_cxx::hash;
  using __gnu_cxx::hash_map;
  using __gnu_cxx::hash_set;
}
#define SGI_HASH_NAMESPACE __gnu_cxx
#else
#include <hash_map>
#include <hash_set>

namespace sgi {
     using std::hash;
     using std::hash_map;
     using std::hash_set;

}
#define SGI_HASH_NAMESPACE std
#endif

#include <map>
#include <string>
#include <sstream>

#include <Trace.hh>
#include <param.hh>
#include <Text.hh>
#include <NumericalArray.hh>
#include <Transducer.hh>
#include <Vocabulary.hh>
#include <DecoderNB.hh>

#include <CompactTSequence.hh>

#include <Types.hh>

#include <pstream.h>


#define NSZ 4
#define MAXBUF 10000

// a 4-gram tracker
typedef double Counter[NSZ];

//holds intermediate metric statistics
typedef struct{

  Counter numCorrect;
  Counter numSuggested;
  double referenceLengths;

}RawCounts;

//holds feature data and hash elements
typedef struct{
  double score;
  float* p_scalingFactor;
  int hashValue;
}FeaturePack;

struct CTranslationItem{
  vector<CWordIndex> m_translation;
  vector<float> m_modelScores;
  RawCounts m_scoringData;
  //number of correct/suggest pairs
  int gramSize;
  
};




typedef SGI_HASH_NAMESPACE::hash_map<int, RawCounts> ProximityMap;

struct ltdbl
{
  bool operator()(const double x, const double y) const
  {
    
    return (x==y);
    

  }
};


typedef std::multimap<const double, double*,  ltdbl> UniqueFeatures;

// the feature sets of all the candidate translations
// each element contains a feature array for that particular translation
typedef std::vector<double*> FeatureBank;
//pairing of the score for this transducer item and the item itself

//old scheme
//typedef CArray< pair<double, CTransducerItem *> > EdgeList;
//old scheme for fixed tranducer list
//typedef CArray< FeaturePack > EdgeList;

typedef CArray< CWordInfo* > EdgeList;
typedef CArray< EdgeList > EdgeBank;

//using namespace std;
class CNBest;
class CNBestItem;

class CNBestItem
{

public:
  CNBestItem();
  ~CNBestItem();

  void Display() const;

  // the vocabulary object that is used to translate the words in this transducer
  static CVocabulary *m_pVoc;


  //======================================
  // Insert feature data for this path
  //======================================
  //adds all features and removes conflicts
  bool AddFeatureSet(double* fb);
  //replaces the feature set directly
  bool ReplaceFeatureSet(double* fb);
  static int m_numFeatures;
  FeatureBank* getFeatureSets();

  //======================================
  // Insert metric data for this path
  //======================================
  Counter* getNumberCorrect();
  Counter* getNumberSuggested();

  void setNGramData(int n, double correctScore, double suggestScore);
  void incrementNGramData(int n, double correctScore, double suggestScore);

  void setRefLengthData(double refLength);
  void incrementRefLengthData(double refLength);
  double* getClosestRefLength();


  //==============================================
  // Insert path representation data for this path
  //==============================================
  // sets the index representation of the candidate string
  void setCandidateVector(std::vector<CWordIndex> candidateVector);
  //convert back into the string representation
  std::string getCandidateString();
  void AddEdgeList(EdgeList eList);
  //returns the complete edge bank for this
  EdgeBank& getEdgeBank();

  //returns the confidence score of a pathway given a wordProbLookup
  double GetConfidenceScore(int pathNumber, CCompactTSequence<double>* wordProbLookup);

  //score a single phrase
  double GetPhraseConfidence(CCompactTSequence<int>* conditioned, CCompactTSequence<int>* randomVariable, CCompactTSequence<double>* wordProbLookup);

  //takes in the raw count data
    void PrintOptimizationData(const int sentenceNumber, const int NGramsConsidered,  int& startingIndex, ostream& os);
  void PrintOptimizationDataWithSentence(const int sentenceNumber, const int NGramsConsidered, int& startingIndex, vector<float>& scalingFactors, fstream& os);
  void PrintOptimizationDataWithSentenceAndSegmentation(const int sentenceNumber, const int NGramsConsidered, int& startingIndex, vector<float>& scalingFactors, fstream& os);

  // check if this candidate has already been processed for counts against the references
  bool isCalculated();
  void setCalculated(bool isCalc);

  //for confidence measuring
  vector<double> GenerateAllSequenceProbs(vector<float>& scaling);
  
  //increment with n-m gram occurrences
  void IncrementOccurrencesNGram(CCompactTSequence<double>* wordProbLookup, double prob, CCompactTSequence<double>* conditionedSum);
  void IncrementOccurrences(CCompactTSequence<double>* wordProbLookup, double prob);
  void IncrementAndSum(CCompactTSequence<int>* conditioned, CCompactTSequence<int>* randomVariable, CCompactTSequence<double>* wordProbLookup,  CCompactTSequence<double>* conditionedSum, double prob);


  //return the lowest cost pathway under the current scaling factors
  double GetMinCost(vector<float>& scalingFactors);
  //sets the feature data for all translation pathways in this nbest item to value determined
  //in this rescoring process. For now, the key operation is to calculate and set IBM model 1 
  //scores for each translation
  
  //void RescoreAllTranslations(CLexicon* pLexicon, CSentence& sourceSentence);
  CCompactTSequence<int>* CollectNGrams(vector<int> source, int maxNSize);

private:
  
  //a translation can be generated by several feature combinations
  FeatureBank m_FeatureSets;
  EdgeBank m_EdgeSets;

  // the raw counts for this translation item when compared to the reference set for the sources sentence
  Counter m_numberCorrect;
  Counter m_numberSuggested;
  double m_closestRefLength;

  // representation of the whole string that this item represents
  vector<CWordIndex> m_candidateVector;

  //have the counts for this sentence already been calculated?
  // this is on a per candidate translation level
  bool m_isCalculated;
  
  //keep track of features that have been seen by putting the product of the features
  //into a sorted map 
  UniqueFeatures m_uniqueFeatures;

  friend class CNBest;
};

class CNBest
{
public:
  CNBest();
  CNBest(int numFeatures);
  ~CNBest();
  bool Init( CVocabulary *pInputVocabulary );
  bool Init( CVocabulary *pInputVocabulary, std::vector<float> scalingFactorPtrs );

  int NextSource( CTransducerState State, CWordIndex Word, 
		  CTransducerState &NextState );

  //full translations
  CNumericalArray< CNBestItem * > m_TransducerItems;
  std::map< CStateWord, CTransducerState, CompareStateWord > m_TransitionTableSource;

  // each transducer state has one element, which contains all the analysis for this 
  // particular candidate translation string
  sgi::hash_map< CTransducerState , CNBestItem* > m_EmissionTableSource;

  //load a whole translation bank into the mapping
  // sv 051102 Decoder result should not be decoder spefic
  //int LoadFromBank(int sentenceNumber, CArray<CDecoderNBResult*>& TranslationBank, CSentence sourceSentence);
  //int LoadFromBankUnique(int sentenceNumber, CArray<CDecoderNBResult*>& TranslationBank, CSentence sourceSentence);
  int LoadFromBank(int sentenceNumber, CArray<CDecoderResult*>& TranslationBank, CSentence sourceSentence);
  int LoadFromBankUnique(int sentenceNumber, CArray<CDecoderResult*>& TranslationBank, CSentence sourceSentence);

  void ClearAll();


  // add one candidate translation to the mapping
  // returns the number of new translations found
  // the scoring process is used to generate scoring data based on the predefined metric
  
  // sv 051102 Decoder result should not be decoder spefic
  //int AddOneCandidate(int sentenceNumber, CDecoderNBResult* result, bool saveDuplicates);
  int AddOneCandidate(int sentenceNumber, CDecoderResult* result, bool saveDuplicates);
  // gets back the CNBestItem coresponding to this translation
  CNBestItem* GetNBestItem(string sourceWord);

  static CTransducerState s_NotFound;

  vector<Counter*> getNumberCorrectList();
  vector<Counter*> getNumberSuggestedList();
  vector<double*> getReferenceLengthsList();

  int ConvertCandidateToBloc(int candidateNumber);

  // print simple diagnostics to analyse the n-best list
  void PrintDiagnosticInformation(int sentenceNumber, std::ostream& os);
  void PrintOptimizationData(const int sentenceNumber, const int NGramsConsidered, ostream& osFeats, ostream& osCands);
  void PrintOptimizationDataWithSentence(const int sentenceNumber, const int NGramsConsidered, fstream& osFeats, fstream& osCands);
  void PrintOptimizationDataWithSentenceAndSegmentation(const int sentenceNumber, const int NGramsConsidered, fstream& osFeats, fstream& osCands);
  //returns the total number of candidates in this nbest list
  int getNumCands();
  int getNumDuplicates();
  int m_numFeatures;
  

  //based on globalScalingPts
  void BuildWordPosteriors();
  //void RescoreAllTranslations(CLexicon* pLexicon);

  //returns the index of the lowest cost path under the given scaling factors
  int GetModelBestIndex(vector<float>& scalingFactors);

private:


  std::vector<float> m_globalScalingPtrs;

  //source sentence that generateed this candidate translation set
  CSentence m_sourceSentence;

  CVocabulary *m_pVoc;
  //double* m_scalingFactors;
  bool m_BuildMode;
  
  int m_NumberOfStatesSource;
  double* tmpHolder;
  
};

#endif



