#ifndef DecoderNB_hh
#define DecoderNB_hh
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include <bitset>
using std::bitset;
#include <map>
using std::map;
#include <Decoder.hh>
#include <LMInterface.hh>
#include <LatticeEdgeMS.hh>
#include <Lexicon.hh>
#include <NBestListRescorer.hh>
//Sanjika 04/25
#include <SentenceLengthModel.hh>
//ARV 07/13
#include <FutureEstimator.hh>

//Nguyen, Aug 7, 2008: Cohesive Decoding
#include <tokens.hh>


#ifdef HAVE_EXT_HASH_SET
#include <ext/hash_set>
using __gnu_cxx::hash;
using __gnu_cxx::hash_set;
#else
#include <hash_set>
using std::hash;
using std::hash_set;
#endif

#define HypContainer hash_set


//Commment listing
//ARV: 04/20 Generalizing the decoder to handle a set of scores
//ARV: 04/22 for compact with PDA Devices we are moving to using all out internal classes
//Sanjika 04/25 Added UpdateLexicalSLScore, InitLexicalSLScoreEnd GetLexicalSLScoreEnd functionality


// classes defined in this header file
class CDecoderBLM;
class CDecoderBLM_Hyp;

// maximum lenght for reordering, as coverage pattern is stored in short int
const int c_MaxReorderWindowLength = COVERAGE_BIT_VECTOR_SIZE;

// the score for a word not translated
const float c_TMScoreForSourceWord = 10.0;


//used to keep track of deletions
typedef multimap< float, CDecoderBLM_Hyp * > RecombinedMap;


struct CDecoderBLM_Info {

  // hyps can be kept apart according to these features:
  bool m_UseCoverage;
  bool m_UseCoveragePattern;
  bool m_UseLMState;
  bool m_UseTargetLength;

  // for pruning hyps are kept apart according these following features
  // which should be a subset of the previous set of features
  bool m_UseCoverageForPruning;
  bool m_UseCoveragePatternForPruning;
  bool m_UseLMStateForPruning;
  bool m_UseTargetLengthForPruning;

  // how many reorderings are totally allowed
  // should be set depending on sentence length, i.e. for each sentence individually
  // default is 77;-)
  short m_MaxNumReorderings;
};



// Structure used to backtrace while generating the n-best list of translations
// Each BackPointer points to the edge which was travered and the hypothesis
// resulting after this tranversal.  All true scores, i.e. scores this hypothesis
// would have had, had it been expanded to the sentence end, are carried on.

class CDecoderBLM_BP : public CDecoder_BP
{
public:
  CDecoderBLM_BP( int numModels );
  CDecoderBLM_BP( CDecoderBLM_Hyp *pHyp );
  CDecoderBLM_BP( CDecoderBLM_BP *pBP );
  ~CDecoderBLM_BP();

  float              m_TotalScore;      // accumulated total score

  int                m_TargetLength;    // true length of translation
  //  CLatticeEdgeMS *m_pEdge;           // pointer to edge which was traversed
  CDecoderBLM_Hyp    *m_pHyp;            // pointer to hypothesis
  CDecoderBLM_BP     *m_pBP;             // pointer to next BackPointer in the list

  static long int s_NumberOfObjects; 
};




class CDecoderBLM_Hyp : public CDecoder_Hyp
{

public:
  CDecoderBLM_Hyp();
  CDecoderBLM_Hyp( int numModels );
  CDecoderBLM_Hyp( const CDecoderBLM_Hyp &Hyp );
  ~CDecoderBLM_Hyp();
  CDecoderBLM_Hyp  &operator = ( const CDecoderBLM_Hyp & );

  friend ostream   &operator << (ostream &s, const CDecoderBLM_Hyp & );

  CDecoderBLM_Info *m_pDecoderInfo; // access to decoder info which can be different 
                                   // for different decoder objects
  //CLMState          m_LMState;         // language model state
  CNumericalArray< CLMState >   m_LMStates;         // language model state

  short             m_Coverage;        // how many words from the source sentence have been covered
  short             m_CoverageStart;   // first position not covered
  bitset< COVERAGE_BIT_VECTOR_SIZE >      m_CoveragePattern; // bit-vector indicating covered postions in source sentence
  short             m_TargetLength;    // length of path over which phrase was constructed
  short             m_NumReorderings;  // number of total reorderings
  CLatticeEdgeMS   *m_pEdge;      // pointer to edge which was traversed last
  

  CDecoderBLM_Hyp  *m_pPrevHyp;   // pointer to previous hypothesis, needed for backtracking

  multimap< float, CDecoderBLM_Hyp * > *m_pRecombinedHyps;  // to store recombined hyps

  bool              m_Expand;          // flag set in histogram pruning
  bool              operator== (CDecoderBLM_Hyp &Hyp);

  // how many hyps exist, can be used to detect memory leaks
  static long int   s_NumberOfObjects; 

  // test if some of the positions between j_min and j_max are already covered
  // if yes return false else set those positions, possible shift position bits
  // and update CoverageStart, and return true
  bool TestAndSetCoverage( int j_min, int j_max );

  static bool s_TraceOn;
};



typedef CArray< multimap< float, CDecoderBLM_BP * > > NBestHyps;

// compare, if two CDecoderBLM_Hyps are equal 
bool operator == ( CDecoderBLM_Hyp &Hyp1, CDecoderBLM_Hyp &Hyp2 );


// hash fuction and compare function for the hash_set 
// storing the hypotheses

struct HypHash
{
  size_t operator()(const CDecoderBLM_Hyp *pHyp) const
  {
    //cout << "HypHash::operator() used as hash function" << endl;
    //cout << *pHyp << endl;
    
    // size_t is unsigned long
    size_t HashNumber = 0;
    
    if ( pHyp->m_pDecoderInfo->m_UseCoverage ) {
      //cout << "Use Coverage" << endl;
      HashNumber += (size_t)pHyp->m_Coverage;
      //cout << HashNumber << endl;
    }
    if ( pHyp->m_pDecoderInfo->m_UseCoveragePattern ) {
      //cout << "Use Coverage" << endl;
			       HashNumber += (size_t)pHyp->m_CoverageStart + (size_t)(BitsetHash(pHyp->m_CoveragePattern));
			//			HashNumber += (size_t)pHyp->m_CoverageStart + (size_t)(pHyp->m_CoveragePattern.to_ulong());
    }
    if ( pHyp->m_pDecoderInfo->m_UseLMState ) {
      //cout << "Use LMState" << endl;
      HashNumber += (size_t)pHyp->m_LMStates[ pHyp->m_LMStates.MinIndex() ]
        + (size_t)(pHyp->m_LMStates[ pHyp->m_LMStates.MinIndex() ] >> 32);
      //cout << HashNumber << endl;
    }
    if ( pHyp->m_pDecoderInfo->m_UseTargetLength ) {
      //cout << "Use Length" << endl;
      HashNumber += (size_t)pHyp->m_TargetLength;
      //cout << HashNumber << endl;
    }
	
    //cout << HashNumber << endl;
    return HashNumber;
	
  }
};



struct HypEqual
{
  bool operator()(const CDecoderBLM_Hyp *pHyp1, const CDecoderBLM_Hyp *pHyp2 ) const
  {
    //cout << "operator( pHyp, pHyp ) used as equal function" << endl;
    //cout << *this << endl;
    
    if ( pHyp1->m_pDecoderInfo->m_UseCoverage  && (pHyp1->m_Coverage != pHyp2->m_Coverage) ) {
      //cout << "Hyps have unequal coverage" << endl;
      return false;
    }
    if ( pHyp1->m_pDecoderInfo->m_UseCoveragePattern  && (pHyp1->m_CoverageStart != pHyp2->m_CoverageStart) ) {
      //cout << "Hyps have unequal coverage" << endl;
      return false;
    }
    if ( pHyp1->m_pDecoderInfo->m_UseCoveragePattern  && (pHyp1->m_CoveragePattern != pHyp2->m_CoveragePattern) ) {
      //cout << "Hyps have unequal coverage pattern" << endl;
      return false;
    }
    if ( pHyp1->m_pDecoderInfo->m_UseLMState 
         && ( pHyp1->m_LMStates[ pHyp1->m_LMStates.MinIndex() ] != pHyp2->m_LMStates[ pHyp2->m_LMStates.MinIndex() ]) ) {
      //cout << "Hyps have unequal LMStates" << endl;
      return false;
    }
    if ( pHyp1->m_pDecoderInfo->m_UseTargetLength  && (pHyp1->m_TargetLength != pHyp2->m_TargetLength) ) {
      //cout << "Hyps have unequal length" << endl;
      return false;
    }
    //cout << "Hyps are equal" << endl;

    return true;
  }
};



// now the hash function and equal operator used for pruning
struct HypHashForPruning
{
  size_t operator()(const CDecoderBLM_Hyp *pHyp) const
  {
    //cout << "HypHash::operator() used as hash function" << endl;
    //cout << *pHyp << endl;
	
    // size_t is unsigned long
    size_t HashNumber = 0;
    if ( pHyp->m_pDecoderInfo->m_UseCoverageForPruning ) {
      //cout << "Use Coverage" << endl;
		  HashNumber += (size_t)pHyp->m_Coverage;
      //cout << HashNumber << endl;
    }
    if ( pHyp->m_pDecoderInfo->m_UseCoveragePatternForPruning ) {
      //cout << "Use Coverage" << endl;
      
      // HashNumber += (size_t)pHyp->m_CoverageStart + (size_t)pHyp->m_CoveragePattern.to_ulong();
      HashNumber += (size_t)pHyp->m_CoverageStart + (size_t)(BitsetHash( pHyp->m_CoveragePattern ));      
      
      //cout << HashNumber << endl;
    }
    if ( pHyp->m_pDecoderInfo->m_UseLMStateForPruning ) {
      //cout << "Use LMState" << endl;
      HashNumber += (size_t)pHyp->m_LMStates[ pHyp->m_LMStates.MinIndex() ]
        + (size_t)(pHyp->m_LMStates[ pHyp->m_LMStates.MinIndex() ] >> 32);
      //cout << HashNumber << endl;
    }
    if ( pHyp->m_pDecoderInfo->m_UseTargetLengthForPruning ) {
      //cout << "Use Length" << endl;
      HashNumber += (size_t)pHyp->m_TargetLength;
      //cout << HashNumber << endl;
    }
	
    //cout << HashNumber << endl;
    return HashNumber;
	
  }

};



struct HypEqualForPruning
{
  bool operator()(const CDecoderBLM_Hyp *pHyp1, const CDecoderBLM_Hyp *pHyp2 ) const
  {
    //cout << "operator( pHyp, pHyp ) used as equal function" << endl;
    //cout << *this << endl;

    if ( pHyp1->m_pDecoderInfo->m_UseCoverageForPruning  && (pHyp1->m_Coverage != pHyp2->m_Coverage) ) {
      //cout << "Hyps have unequal coverage" << endl;
      return false;
    }
    if ( pHyp1->m_pDecoderInfo->m_UseCoveragePatternForPruning  && (pHyp1->m_CoverageStart != pHyp2->m_CoverageStart) ) {
      //cout << "Hyps have unequal coverage pattern" << endl;
      return false;
    }
    if ( pHyp1->m_pDecoderInfo->m_UseCoveragePatternForPruning  && (pHyp1->m_CoveragePattern != pHyp2->m_CoveragePattern) ) {
      //cout << "Hyps have unequal coverage pattern" << endl;
      return false;
    }
    if ( pHyp1->m_pDecoderInfo->m_UseLMStateForPruning  
         && ( pHyp1->m_LMStates[ pHyp1->m_LMStates.MinIndex() ] != pHyp2->m_LMStates[ pHyp2->m_LMStates.MinIndex() ] ) ) {
      //cout << "Hyps have unequal LMStates" << endl;
      return false;
    }
    if ( pHyp1->m_pDecoderInfo->m_UseTargetLengthForPruning  && (pHyp1->m_TargetLength != pHyp2->m_TargetLength) ) {
      //cout << "Hyps have unequal length" << endl;
      return false;
    }
    //cout << "Hyps are equal" << endl;

    return true;
  }
};



//Represents the pair of a translation CString
// and a DecoderNB_Hyp object representing the final 
// scores of this translation
/*
class CDecoderBLMResult{
public:

  CDecoderBLMResult();
  CDecoderBLMResult(CString translation, CDecoderBLM_BP * m_pBP,CArray< CLatticeEdgeMS * > edges );
  
  ~CDecoderBLMResult();
  void scaleFeatures(std::vector<double>& scalingFactors, bool divide);
  CDecoderBLMResult& CDecoderBLMResult::operator= ( const CDecoderBLMResult& arg );

  // to put DecoderNBResult objects into sorted arrays
  bool operator< ( const CDecoderBLMResult& arg );
  
  CString m_translation; 
  float m_TotalScore;
  vector<double> m_features;
  CArray< CLatticeEdgeMS * > m_edges;

  void Display(string tag, int id, vector<double>& scalingFactors, CLatticeLM* lattice=NULL);

  //CArray< pair<double, CTransducerItem *> > m_edges;

  //CDecoderBLM_BP * m_pBP;
};
*/



// ----------------------------------------------------
// 
class CDecoderBLM : public CDecoder
{
public:
  CDecoderBLM();
  CDecoderBLM( const CDecoderBLM &Decoder );
  ~CDecoderBLM();
  
  // initialize the decoder 
  // currently just sets the size  of the scaling factor array
  // will be the place where models can be turned on or off
  void InitModels( int NumberOfModels );

  //void Init( CTranslationMemory *pTranslationMemory, CLM *pLM );
  //void Init( CTranslationMemory *pTranslationMemory, CLM *pLM, CVocabulary *pVocabulary );
  //void Init( const char *ParamFileName, CVocabulary *pVocabulary );
  //void SetVocabulary( CVocabulary *pVocabulary );

  void SetLexicon( const CLexicon *pLexicon );
  void SetSourceSentence( const CSentence *pSentence );

  void SetSentenceLengthModel( CSentenceLengthModel *pSLM);

  //ARV removed, this parameter is no longer used for translation
  //void SetSLStretchFactor( float SLStretch = 1.0 );

  void SetTagsConsideredAsWords( const CString &TagsConsideredAsWords, char SplitCharacter );


  CString TranslateSentence( const CSentence *pSentence );

  CArray<CDecoderResult*> TranslateLattice( CLatticeLM *pLattice );

  void SetLengthOfReorderingWindow( int LengthOfReorderingWindow );
  void SetReorderingsPerWord( float ReorderingsPerWord );
  void SetNumOfNBest( int N );
  void SetCompareHypFeatures( bool UseCoverage = true, bool UseCoveragePattern = true,
                              bool UseLM = true, bool UseTargetLength = true );
  void SetPruneHypFeatures( bool UseCoverage = true, bool UseCoveragePattern = true,
                            bool UseLM = true, bool UseTargetLength = false );

  void SetBeamSize( unsigned long int BeamSize = 10000 );
  void SetMaxHypsPerCoverage( unsigned long int MaxHyps = 10000000 );
  void SetMaxHyps( int MaxHyps = 1000 );

  //ARV 07/13
  void SetFutureEstimator(CFutureEstimator* pFutures){
    m_pFutures = pFutures;
  }


  void SetScalingFactors( double TMScale = 1.0, double LMScale = 1.0,
                          double ALScale = 1.0, double SLScale = 1.0 );

  void SetScalingFactors( CNumericalArray< float > ScalingFactors );
  void SetScalingFactors( vector< float > ScalingFactors );

  //ARV 4/20, rescoring should be done at the n-best list level
  void SetRescore( bool Rescore = False );

  void SetRescorer( CNBestListRescorer *pRescorer );


  // get methods to be used when administering the decoding server
  void GetLengthOfReorderingWindow( int& LengthOfReorderingWindow );
  void GetReorderingsPerWord( float& ReorderingsPerWord );
  void GetNumOfNBest( int& N );
  void GetCompareHypFeatures( bool& UseCoverage, bool& UseCoveragePattern,
                              bool& UseLM, bool& UseTargetLength );
  void GetPruneHypFeatures( bool& UseCoverage, bool& UseCoveragePattern,
                            bool& UseLM, bool& UseTargetLength);

  void GetBeamSize( unsigned long int& BeamSize);
  void GetMaxHypsPerCoverage( unsigned long int& MaxHypsPerCov);


  void GetScalingFactors( float& TMScale, float &LMScale,
                          float& ALScale, float &SLScale);

  void GetScalingFactors( CNumericalArray< float > &ScalingFactors );

  void SetLabelsConsideredAsWords( CArray< CString > Labels );

  void SetLMBatchMode( int LMBatchMode = 0 );

  // the actual decoding: build translation hypothesis by taking
  // edges from the translation lattice to extend partial translations,
  // adding the LM score to the TM score and keeping track of 
  // already covered source words
  void BuildTranslationLattice( CLatticeLM *pLattice );
  void AddOverlappingPhrases( CLattice *pLattice );
  
  CArray< CDecoderResult* > ApplyLanguageModel();
  CArray< CDecoderResult* > ApplyLanguageModel( CLatticeLM *pLattice );
  CArray< CDecoderResult* > ApplyLanguageModel( CLatticeLM *pLattice, CLM *pLM );
  
  //Nguyen, AUg 7, 2008: Cohesive Decoding
  CArray< CDecoderResult* > ApplyLanguageModelWithDependencyTree( CLatticeLM *pLattice, vector<tokens>& sentDtree, int algorithm );
  CArray< CDecoderResult* > ApplyLanguageModelWithDependencyTree();  

  void WriteLattice();

  // used to clear the decoder after the last sentence
  vector< RecombinedMap* > HarvestFromHyp(CDecoderBLM_Hyp* pOneHyp);
  
  void Clear();


protected:
  virtual CDecoderBLM_Hyp * NewHypothesis();
  virtual RecombinedMap   * NewRecombinedMap();
  
  // for efficiency: convert target words into integers for use with LM
  void GetLMWordIndicesAndTMScores();
  bool LabelIsNormalWord( const CString &Word );

  void GetTMScore( CLatticeEdgeMS *pEdge );

  void GetTMScoreMS( CLatticeEdgeMS *pEdge );
  void GetALScore( CDecoderBLM_Hyp *pHyp, CLatticeEdgeMS *pEdge );
  //void GetLMScore( CDecoderBLM_Hyp *pHyp, CLatticeEdgeMS *pEdge );
  void GetPCScore( CDecoderBLM_Hyp *pHyp, CLatticeEdgeMS *pEdge );
  void GetWCScore( CDecoderBLM_Hyp *pHyp, CLatticeEdgeMS *pEdge );

  // for multiple LMs and multiple DMs
  void GetLMScores( CDecoderBLM_Hyp *pHyp, CLatticeEdgeMS *pEdge );
  void GetLMScoresEOS( CDecoderBLM_Hyp *pHyp );
  void GetDMScores( CDecoderBLM_Hyp *pHyp, CLatticeEdgeMS *pEdge );
  
  //Nguyen, Aug 4, 2008
  void GetCMScore( CDecoderBLM_Hyp *pHyp, CLatticeEdgeMS *pEdge, vector<tokens> fh1, vector<tokens> fh );  
  bool InterruptionCheck (vector<tokens>& v, vector<tokens>& fh, vector<tokens>& fh1, vector<int>& HypCoverage, tokens& r, int algorithm );
  //comparator
  //bool operator<( CovEdg a, CovEdg b );  
  
  /////// End of Nguyen ///////      
  
  int  TargetLength( CDecoderBLM_Hyp *pHyp );
  int  TargetLength( CLatticeEdgeMS *pEdge );
  void UpdateSLScore( CDecoderBLM_BP *pBP, int &n );
  void UpdateWCScore( CDecoderBLM_BP *pBP, int &n );


  //ARV 07/13 - determines which word segments are still left to translate
  // used in future cost estimation
  CArray< std::pair<int, int> >  LeftToTranslate( CDecoderBLM_Hyp *pHyp );

  //using a standard form of initializing for sentence end and then
  //running a function on each hyp
  //void InitLexicalSLScoreEnd();
  //double GetLexicalSLScoreEnd(CDecoderBLM_Hyp* pHyp);
  //void UpdateLexicalSLScore(CDecoderBLM_BP *pBP, int &n );


  int m_PruningMode;
  void SetPruningMode( int PruningMode );


  // expanding hyps of a given coverage, over an edge, over the sentence end
  virtual void ExpandAllHypsWithCoverage( int Coverage );
  virtual void ExpandAllHypsWithCoverageBeamWidthPruning( int Coverage );
  virtual void ExpandAllHypsWithCoverageBeamSizePruning ( int Coverage );
  virtual void ExpandHyp( CDecoderBLM_Hyp *pHyp );
  virtual void ExpandHypOverEdge( CDecoderBLM_Hyp *pHyp, CLatticeEdgeMS *pEdge );
  virtual void ExpandAllHypsSentenceEnd();
  
  //Nguyen, July 31, 2008: Cohesive decoding
  virtual void ExpandHypOverEdgeWithCohesion( CDecoderBLM_Hyp *pHyp, CLatticeEdgeMS *pEdge );


  // for batch LM
  bool m_LMBatchMode;

  void ExpandAllHypsWithCoverage_BatchLM( int Coverage );
  void ExpandAllHypsWithCoverage_BeamSizePruning_BatchLM( int Coverage );
  //void ExpandAllHypsWithCoverage_BeamWidth_PruningBatchLM( int Coverage );
  void ApplyLM_BatchMode( int Coverage );
  void PruneBeamWidth_BatchLM( int Coverage );
  void PruneBeamSize_BatchLM ( int Coverage );


  void GetLMLogProbsAllHypsWithCoverage( int Coverage );

  void DisplayAllHypsWithCoverage( int Coverage, int N = 100 );

  // trace back to collect the words making up the translations
  CString TraceBack( CDecoderBLM_Hyp *pHyp );
  //trace back the sequence along with the paths that it took to get here
  pair<CString, CArray< CLatticeEdgeMS * > > TraceBackWithPath( CDecoderBLM_Hyp* pHyp );

  void BuildNBestList();
  
  pair<CString, CArray< CLatticeEdgeMS * > > TraceStringAndPath( CDecoderBLM_BP * pBP, int n );
  CString TraceForward( CDecoderBLM_BP * pBP );

  virtual void StoreHypothesis();
  void StoreHypothesisTmp();

  // keep only n-best hypotheses
  void HistogramPruning( HypContainer< CDecoderBLM_Hyp *, HypHash, HypEqual > &Hyps,
                         HypContainer< CDecoderBLM_Hyp *, HypHashForPruning, HypEqualForPruning > &BestHyps );

  void Rescore();
  double Rescore( CDecoderBLM_Hyp *pHyp );

  void TraceHyp( bool TraceOn, int TraceLevel, CString Comment, CDecoderBLM_Hyp *pHyp );

  void DisplayDecoderStatistics();
  void DisplayDecoderSettings();
  void TraceHyp( bool s_TraceOn, const char * Message, CDecoderBLM_Hyp *pHyp );



  // all recombined maps are tracked in this list to ensure that they can be deleted
  // when clear is called
  vector< RecombinedMap * > m_RecombinedMapList;

  // hyps are organized according to coverage
  CArray< HypContainer< CDecoderBLM_Hyp *, HypHash, HypEqual > > m_AllHyps;

  // for pruning: store
  CArray< HypContainer< CDecoderBLM_Hyp *, HypHashForPruning, HypEqualForPruning > > m_BestHyps;

  // use multimap to store final list of hyps sorted according to total cost
  multimap< float, CDecoderBLM_Hyp *> m_FinalHyps;

  // use multimap to store final list of hyps sorted according to total cost
  // after rescoring with some rescorer
  multimap< float, CDecoderBLM_Hyp *> m_FinalHypsRescored;

  // for histogram pruning
  CMatrix< CDecoderBLM_Hyp * > m_PruningBuckets;

  // for LM batch mode, hyps are first stored in a temporary list
  CNumericalArray< CDecoderBLM_Hyp * > m_TmpHyps;


  // n-best list generation: hyps are again organized according to coverage
  // hypothesis h is expanded over edge e into hypothesis h_next
  // h_next may be recombined with best hypothesis h_best
  // for backtracing and generating an n-best list we store for each
  // coverage the best 
  CArray< multimap< float, CDecoderBLM_BP * > > m_NBestHyps;

  // use multimap to store final list of hyps sorted according to total cost
  // after rescoring with some rescorer
  multimap< float, CDecoderBLM_BP * > m_NBestHypsRescored;

  // a hypothesis object which is used to test coverage collision
  // only if no collision occurs, a memoryypothesis is created
  CDecoderBLM_Hyp *m_pNumberOneHypothesis;
  CDecoderBLM_Hyp *m_pNumberOneBestHyp;

  // when deleting the hyps we first move them into this array
  // deleting all elements in a hash_set shows strange behaviour
  CNumericalArray< CDecoderBLM_Hyp * > m_HypsForDeleting;

  // add new edges for overlapping phrases to lattice
  bool m_AddOverlappingPhrases;

  CDecoderBLM_Info *m_pDecoderInfo;

  // how many positions the decoder can advance in the source sentence 
  // to select the next word to be translated
  // the window starts with the first untranslated position
  short m_LengthOfReorderingWindow;

  // a factor to limit total number of reorderings in the tranlation
  // i.e. 0.5 means that on average each 5th words can be reordered.
  // a b c d translated as a c b d has two jumps: a-c and c-b
  // but we count each jump only as one half reordering
  float m_ReorderingsPerWord;

  // pruning factor: only expand BeamSize best recombined hyps within similarity class
  unsigned long int m_BeamSize;

  //! Maximal total number of hypotheses expanded by ExpandAllHypsWithCoverage().
  unsigned long int m_MaxHypsPerCoverage;

  int   m_MaxHyps;

  //indicates the number of models used in this decoding process
  int m_NumberOfModels;
  

  // flag, if rescoring should be done
  bool m_Rescore;
  CNBestListRescorer *m_pRescorer;
  // the lattice with all partial translations
  //CLatticeLM *m_pLattice;

  // the language model used; it's always given as parameter to the
  // ApplyLanguageModel method, but it is assigned to a member variable
  // as it is used in different methods
  //CLM *m_pLM;

  //CLMSentenceState *m_pLMSentenceState;

  // some statistics
  long m_HypsExpanded;
  long m_HypsOutOfBeam;
  long m_HypsOutOfBeamPerCov;
  long m_HypsNotExpanded;
  long m_NumNewHyps;
  long m_NumBetterHyps;
  long m_NumWorseHyps;
  long m_NumCoverageCollisions;
  
  //CString TargetString( CArray< CLatticeEdge* > &Path );
  //CString Translation( CArray< CLatticeEdge* > &Path );

  //void DisplayBestPath(  const CLattice *pLattice );  
  //void DisplayTargetString( const CLattice *pLattice );
  //void DisplayLatticeStatistics( const CLattice *pLattice );

  //void
  //ShortestPath( const CLattice *pLattice, CArray< CString > &Lattice );
  
  //CLatticeSearch m_LatticeSearch;

  int m_NumOfNBest;

  // for rescoring
  const CLexicon *m_pLexicon;
  CArray< CWordIndex > m_Source;  // the source sentence

  const CSentence *m_pSourceSentence;

  //char *m_VocFile;
  //CVocabulary *m_pVocabulary;

  CArray< CString > m_TagsConsideredAsWords;
  
  //char *m_WordLM;
  // the name of the language model to be run
  //char *m_LMName;
  //int m_HistoryLength;

  //double m_LMScale;
  //double m_TMScale;

  //int m_LatticeStatistics;

  //CParam m_Param;

  // to generate a unique n-best list we first overgenerate by that factor
  // then keep track of already seen word sequences
  static short c_NumOfNBestMultiplier;



  static bool s_TraceOn;
  //ARV 
  // tracks the number of decoder instances sharing the transmem
  static int m_NumberOfInstances;

  //Sanjika 04/25
  CSentenceLengthModel* m_pSLenModel;

  //ARV 07/13
  CFutureEstimator* m_pFutures;

};



#endif
