#include <config.h>
#include <fortify.h>
#ifndef __DecoderGT__
#define __DecoderGT__

#include <bitset>
//#include <ext/hash_set>

#include <Decoder.hh>
#include <LMInterface.hh>
#include <LatticeEdgeTrans.hh>

#if __GNUC__ == 3 && __GNUC_MINOR__ > 0
#include <ext/hash_set>
namespace sgi {
  using __gnu_cxx::hash;
  using __gnu_cxx::hash_set;
}
#define SGI_HASH_NAMESPACE __gnu_cxx
#else
#include <hash_set>

namespace sgi {
     using std::hash;
     using std::hash_set;
}
#define SGI_HASH_NAMESPACE std
#endif

#define HypContainer sgi::hash_set


// classes defined in this header file
class CDecoderGT;
class CDecoderGT_Hyp;


// maximum lenght for reordering, as coverage pattern is stored in short int
const int c_MaxReorderWindowLength = 32;

// the score for a word not translated
const float c_TMScoreForSourceWord = 10.0;



struct CDecoderGT_Info {

  // hyps can be kept apart according to these features:
  bool m_UseCoverage;
  bool m_UseCoveragePattern;
  bool m_UseLMState;
  bool m_UseTargetLength;

  // for pruning hyps are kept apart according these following features
  // which should be a subset of the previous set of features
  bool m_UseCoverageForPruning;
  bool m_UseCoveragePatternForPruning;
  bool m_UseLMStateForPruning;
  bool m_UseTargetLengthForPruning;

  // how many reorderings are totally allowed
  // should be set depending on sentence length, i.e. for each sentence individually
  // default is 77;-)
  short m_MaxNumReorderings;
};



struct CDecoderGT_Hyp {

public:
  CDecoderGT_Hyp();
  CDecoderGT_Hyp( const CDecoderGT_Hyp &Hyp );
  ~CDecoderGT_Hyp();

  friend ostream &operator << (ostream &s, const CDecoderGT_Hyp & );

  CDecoderGT_Info *m_pDecoderInfo; // access to decoder info which can be different 
                                   // for different decoder objects
  CLMState      m_LMState;         // language model state
  float         m_ACScore;         // accumulated acoustic score over partial path
  float         m_TMScore;         // accumulated translation score over partial path
  float         m_LMScore;         // accumulated language score over partial path
  float         m_ALScore;         // accumulated position alignment score
  float         m_SLScore;         // accumulated sentence length model score
  float         m_TotalScore;      // accumulated total score
  short         m_Coverage;        // how many words from the source sentence have been covered
  short         m_CoverageStart;   // first position not covered
  bitset< COVERAGE_BIT_VECTOR_SIZE >  m_CoveragePattern; // bit-vector indicating covered postions in source sentence
  short         m_TargetLength;    // length of path over which phrase was constructed
  short         m_NumReorderings;  // number of total reorderings
  CLatticeEdgeTrans *m_pEdge;      // pointer to edge which was traversed last
  CDecoderGT_Hyp    *m_pPrevHyp;   // pointer to previous hypothesis, needed for backtracking

  bool          m_Expand;          // flag set in histogram pruning
  //bool operator== (CDecoderGT_Hyp &Hyp);

  // how many hyps exist, can be used to detect memory leaks
  static long int s_NumberOfObjects; 

  // test if some of the positions between j_min and j_max are already covered
  // if yes return false else set those positions, possible shift position bits
  // and update CoverageStart, and return true
  bool TestAndSetCoverage( int j_min, int j_max );

  static bool s_TraceOn;
};


// stream operator for CDecoderGT_Hyp
//ostream &operator << ( ostream &, const CDecoderGT_Hyp );

// compare, if two CDecoderGT_Hyps are equal 
bool operator == ( CDecoderGT_Hyp &Hyp1, CDecoderGT_Hyp &Hyp2 );


// hash fuction and compare function for the hash_set 
// storing the hypotheses

struct HypHash
{
  size_t operator()(const CDecoderGT_Hyp *pHyp) const
  {
	//cout << "HypHash::operator() used as hash function" << endl;
	//cout << *pHyp << endl;
	
	// size_t is unsigned long
	size_t HashNumber = 0;

	if ( pHyp->m_pDecoderInfo->m_UseCoverage ) {
	  //cout << "Use Coverage" << endl;
	  HashNumber += (size_t)pHyp->m_CoverageStart;
	  //cout << HashNumber << endl;
	}
 	if ( pHyp->m_pDecoderInfo->m_UseCoveragePattern ) {
	  //cout << "Use Coverage" << endl;
	  HashNumber += (size_t)pHyp->m_CoverageStart + (size_t)pHyp->m_CoveragePattern.to_ulong();
	  //cout << HashNumber << endl;
	}
	if ( pHyp->m_pDecoderInfo->m_UseLMState ) {
	  //cout << "Use LMState" << endl;
	  HashNumber += (size_t)pHyp->m_LMState + (size_t)(pHyp->m_LMState >> 32);
	  //cout << HashNumber << endl;
	}
	if ( pHyp->m_pDecoderInfo->m_UseTargetLength ) {
	  //cout << "Use Length" << endl;
	  HashNumber += (size_t)pHyp->m_TargetLength;
	  //cout << HashNumber << endl;
	}
	
	//cout << HashNumber << endl;
 	return HashNumber;
	
  }

};

struct HypEqual
{
  bool operator()(const CDecoderGT_Hyp *pHyp1, const CDecoderGT_Hyp *pHyp2 ) const
  {
	//cout << "operator( pHyp, pHyp ) used as equal function" << endl;
	//cout << *this << endl;

	if ( pHyp1->m_pDecoderInfo->m_UseCoverage  && (pHyp1->m_CoverageStart != pHyp2->m_CoverageStart) ) {
	  //cout << "Hyps have unequal coverage" << endl;
	  return false;
	}
	if ( pHyp1->m_pDecoderInfo->m_UseCoveragePattern  && (pHyp1->m_CoveragePattern != pHyp2->m_CoveragePattern) ) {
	  //cout << "Hyps have unequal coverage pattern" << endl;
	  return false;
	}
	if ( pHyp1->m_pDecoderInfo->m_UseLMState  && ( pHyp1->m_LMState != pHyp2->m_LMState) ) {
	  //cout << "Hyps have unequal LMStates" << endl;
	  return false;
	}
	if ( pHyp1->m_pDecoderInfo->m_UseTargetLength  && (pHyp1->m_TargetLength != pHyp2->m_TargetLength) ) {
	  //cout << "Hyps have unequal length" << endl;
	  return false;
	}
	//cout << "Hyps are equal" << endl;

	return true;
  }
};


// now the hash function and equal operator used for pruning
struct HypHashForPruning
{
  size_t operator()(const CDecoderGT_Hyp *pHyp) const
  {
	//cout << "HypHash::operator() used as hash function" << endl;
	//cout << *pHyp << endl;
	
	// size_t is unsigned long
	size_t HashNumber = 0;
        if ( pHyp->m_pDecoderInfo->m_UseCoverageForPruning ) {
          //cout << "Use Coverage" << endl;
          HashNumber += (size_t)pHyp->m_CoverageStart;
          //cout << HashNumber << endl;
        }
 	if ( pHyp->m_pDecoderInfo->m_UseCoveragePatternForPruning ) {
	  //cout << "Use Coverage" << endl;
	  HashNumber += (size_t)pHyp->m_CoverageStart + (size_t)pHyp->m_CoveragePattern.to_ulong();
	  //cout << HashNumber << endl;
	}
	if ( pHyp->m_pDecoderInfo->m_UseLMStateForPruning ) {
	  //cout << "Use LMState" << endl;
	  HashNumber += (size_t)pHyp->m_LMState + (size_t)(pHyp->m_LMState >> 32);
	  //cout << HashNumber << endl;
	}
	if ( pHyp->m_pDecoderInfo->m_UseTargetLengthForPruning ) {
	  //cout << "Use Length" << endl;
	  HashNumber += (size_t)pHyp->m_TargetLength;
	  //cout << HashNumber << endl;
	}
	
	//cout << HashNumber << endl;
 	return HashNumber;
	
  }

};

struct HypEqualForPruning
{
  bool operator()(const CDecoderGT_Hyp *pHyp1, const CDecoderGT_Hyp *pHyp2 ) const
  {
	//cout << "operator( pHyp, pHyp ) used as equal function" << endl;
	//cout << *this << endl;

	if ( pHyp1->m_pDecoderInfo->m_UseCoverageForPruning  && (pHyp1->m_CoverageStart != pHyp2->m_CoverageStart) ) {
	  //cout << "Hyps have unequal coverage" << endl;
	  return false;
	}
	if ( pHyp1->m_pDecoderInfo->m_UseCoveragePatternForPruning  && (pHyp1->m_CoveragePattern != pHyp2->m_CoveragePattern) ) {
	  //cout << "Hyps have unequal coverage pattern" << endl;
	  return false;
	}
	if ( pHyp1->m_pDecoderInfo->m_UseLMStateForPruning  && ( pHyp1->m_LMState != pHyp2->m_LMState) ) {
	  //cout << "Hyps have unequal LMStates" << endl;
	  return false;
	}
	if ( pHyp1->m_pDecoderInfo->m_UseTargetLengthForPruning  && (pHyp1->m_TargetLength != pHyp2->m_TargetLength) ) {
	  //cout << "Hyps have unequal length" << endl;
	  return false;
	}
	//cout << "Hyps are equal" << endl;

	return true;
  }
};





// ----------------------------------------------------
// 
class CDecoderGT : public CDecoder
{
public:
  CDecoderGT();
  CDecoderGT( const CDecoderGT &Decoder );
  ~CDecoderGT();
  
  // initialize the decoder 
  void Init( CTranslationMemory *pTranslationMemory, CLM *pLM );
  void Init( const char *ParamFileName, CVocabulary *pVocabulary );

  // ARV 
  // The translate sentence is defunct, it doesn't quite work...
  CString TranslateSentence( const CSentence *pSentence );
  
  // ARV - TranslateLattice now returns the complete n-best list
  CArray<CString> TranslateLattice( CLatticeLM *pLattice );

  void SetLengthOfReorderingWindow( int LengthOfReorderingWindow );
  void SetReorderingsPerWord( float ReorderingsPerWord );
  void SetNumOfNBest( int N );
  void SetCompareHypFeatures( bool UseCoverage = true, bool UseCoveragePattern = true,
			bool UseLM = true, bool UseTargetLength = true );
  void SetPruneHypFeatures( bool UseCoverage = true, bool UseCoveragePattern = true,
			bool UseLM = true, bool UseTargetLength = false );

  void SetBeamFactor( float BeamFactor = 1.0e10 );
  void SetMaxHyps( int MaxHyps = 1000 );

  void SetSLStretchFactor( float SLStretch = 1.0 );

  void SetScalingFactors( float TMScale = 1.0, float LMScale = 1.0,
			  float ALScale = 1.0, float SLScale = 1.0 );

  // ARV introduced get methods to be used when administering the decoding server
  void GetLengthOfReorderingWindow( int& LengthOfReorderingWindow );
  void GetReorderingsPerWord( float& ReorderingsPerWord );
  void GetNumOfNBest( int& N );
  void GetCompareHypFeatures( bool& UseCoverage, bool& UseCoveragePattern,
			bool& UseLM, bool& UseTargetLength );
  void GetPruneHypFeatures( bool& UseCoverage, bool& UseCoveragePattern,
			bool& UseLM, bool& UseTargetLength);

  void GetBeamFactor( float& BeamFactor);

  void GetSLStretchFactor( float& SLStretch );

  void GetScalingFactors( float& TMScale, float& LMScale,
			  float& ALScale, float& SLScale);


  // the actual decoding: build translation hypothesis by taking
  // edges from the translation lattice to extend partial translations,
  // adding the LM score to the TM score and keeping track of 
  // already covered source words
  void BuildTranslationLattice( CLatticeLM *pLattice );
  void AddOverlappingPhrases( CLattice *pLattice );

  // ARV
  // Changed to allow the decoder to return an nbest decoding
  CArray<CString> ApplyLanguageModel();
  CArray<CString> ApplyLanguageModel( CLatticeLM *pLattice );
  CArray<CString> ApplyLanguageModel( CLatticeLM *pLattice, CLM *pLM );

  void ExpandAllHypsWithCoverage( int Coverage );
  void ExpandHyp( CDecoderGT_Hyp *pHyp );
  void ExpandHypOverEdge( CDecoderGT_Hyp *pHyp, CLatticeEdgeTrans *pEdge );
  void ExpandAllHypsSentenceEnd();

  void DisplayAllHypsWithCoverage( int Coverage, int N = 100 );

  // trace back to collect the words making up the translations
  CString TraceBack( CDecoderGT_Hyp *pHyp );

  void WriteLattice();

private:
  // delete all hyps
  void Clear();

  // for efficiency: convert target words into integers for use with LM
  void GetLMWordIndicesAndTMScores();

  void GetTMScore( CLatticeEdgeTrans *pEdge );
  void GetALScore( CDecoderGT_Hyp *pHyp, CLatticeEdgeTrans *pEdge );
  void GetLMScore( CDecoderGT_Hyp *pHyp, CLatticeEdgeTrans *pEdge );

  void StoreHypothesis();

  // keep only n-best hypotheses
  void HistogramPruning( HypContainer< CDecoderGT_Hyp *, HypHash, HypEqual > &Hyps,
						 HypContainer< CDecoderGT_Hyp *, HypHashForPruning, HypEqualForPruning > &BestHyps );

  // hyps are organized according to coverage
  CArray< HypContainer< CDecoderGT_Hyp *, HypHash, HypEqual > > m_AllHyps;

  // for pruning: store 
  CArray< HypContainer< CDecoderGT_Hyp *, HypHashForPruning, HypEqualForPruning > > m_BestHyps;

  // use multimap to store final list of hyps sorted according to total cost
  multimap<double, CDecoderGT_Hyp *> m_FinalHyps;
  
  // for histogram pruning
  CMatrix< CDecoderGT_Hyp * > m_PruningBuckets;

  // a hypothesis object which is used to test coverage collision
  // only if no collision occurs, a memoryypothesis is created
  CDecoderGT_Hyp *m_pNumberOneHypothesis;
  CDecoderGT_Hyp *m_pNumberOneBestHyp;

  // when deleting the hyps we first move them into this array
  // deleting all elements in a hash_set shows strange behaviour
  CNumericalArray< CDecoderGT_Hyp * > m_HypsForDeleting;

  // add new edges for overlapping phrases to lattice
  bool m_AddOverlappingPhrases;

  CDecoderGT_Info *m_pDecoderInfo;

  // how many positions the decoder can advance in the source sentence 
  // to select the next word to be translated
  // the window starts with the first untranslated position
  short m_LengthOfReorderingWindow;

  // a factor to limit total number of reorderings in the tranlation
  // i.e. 0.5 means that on average each 5th words can be reordered.
  // a b c d translated as a c b d has two jumps: a-c and c-b
  // but we count each jump only as one half reordering
  float m_ReorderingsPerWord;

  // pruning factor: prune hyps with Score > BeamFactor * BestScore
  float m_BeamFactor;
  int   m_MaxHyps;

  // the scaling factors for the different models
  float m_TMScale;
  float m_LMScale;
  float m_ALScale;
  float m_SLStretch;
  float m_SLScale;

  // the lattice with all partial translations
  //CLatticeLM *m_pLattice;

  // the language model used; it's always given as parameter to the
  // ApplyLanguageModel method, but it is assigned to a member variable
  // as it is used in different methods
  //CLM *m_pLM;

  //CLMSentenceState *m_pLMSentenceState;

  // some statistics
  long m_HypsExpanded;
  long m_HypsPruned;
  long m_HypsNotExpanded;
  long m_NumNewHyps;
  long m_NumBetterHyps;
  long m_NumWorseHyps;
  long m_NumCoverageCollisions;
  
  //CString TargetString( CArray< CLatticeEdge* > &Path );
  //CString Translation( CArray< CLatticeEdge* > &Path );

  //void DisplayBestPath(  const CLattice *pLattice );  
  //void DisplayTargetString( const CLattice *pLattice );
  //void DisplayLatticeStatistics( const CLattice *pLattice );

  //void
  //ShortestPath( const CLattice *pLattice, CArray< CString > &Lattice );
  
  //CLatticeSearch m_LatticeSearch;

  int m_NumOfNBest;

  //char *m_VocFile;
  //CVocabulary *m_pVocabulary;
  
  //char *m_WordLM;
  // the name of the language model to be run
  //char *m_LMName;
  //int m_HistoryLength;

  //double m_LMScale;
  //double m_TMScale;

  //int m_LatticeStatistics;

  //CParam m_Param;

  static bool s_TraceOn;
  //ARV 
  // tracks the number of decoder instances sharing the transmem
  static int m_numInstances;

};



inline
CDecoderGT_Hyp::CDecoderGT_Hyp()
{
  Trace( 3, "CDecoderGT_Hyp Constructor" );
  m_LMState = 0;
  m_ACScore = 0.0;
  m_TMScore = 0.0;
  m_LMScore = 0.0;
  m_ALScore = 0.0;
  m_SLScore = 0.0;
  m_TotalScore = 0.0;
  m_Coverage = 0;
  m_CoverageStart = 0;
  m_CoveragePattern = 0;
  m_NumReorderings = 0;
  m_TargetLength = 0;
  m_Expand = true;
  m_pEdge = NULL;
  m_pPrevHyp = NULL;
  s_NumberOfObjects++;
}


inline
CDecoderGT_Hyp::CDecoderGT_Hyp( const CDecoderGT_Hyp &Hyp )
{
  Trace( 3, "CDecoderGT_Hyp Copy Constructor" );
  m_LMState = Hyp.m_LMState;
  m_ACScore = Hyp.m_ACScore;
  m_TMScore = Hyp.m_TMScore;
  m_LMScore = Hyp.m_LMScore;
  m_ALScore = Hyp.m_ALScore;
  m_SLScore = Hyp.m_SLScore;
  m_TotalScore = Hyp.m_TotalScore;
  m_Coverage = Hyp.m_Coverage;
  m_CoverageStart = Hyp.m_CoverageStart;
  m_CoveragePattern = Hyp.m_CoveragePattern;
  m_NumReorderings = Hyp.m_NumReorderings;
  m_TargetLength = Hyp.m_TargetLength;
  m_Expand = Hyp.m_Expand;
  m_pEdge = Hyp.m_pEdge;
  m_pPrevHyp = Hyp.m_pPrevHyp;
  s_NumberOfObjects++;
}


inline
CDecoderGT_Hyp::~CDecoderGT_Hyp() {
  s_NumberOfObjects--;
}


inline void
CDecoderGT::SetLengthOfReorderingWindow( int LengthOfReorderingWindow )
{
  if ( LengthOfReorderingWindow > c_MaxReorderWindowLength ) {
	cerr << "Requested length of reordering window larger than possible" << endl;
	cerr << "Restricted to c_MaxReorderWindowLength" << endl;
	LengthOfReorderingWindow = c_MaxReorderWindowLength;
  } else {
	m_LengthOfReorderingWindow = LengthOfReorderingWindow;
  }
}



inline void
CDecoderGT::SetReorderingsPerWord( float ReorderingsPerWord )
{
  m_ReorderingsPerWord = ReorderingsPerWord;
}



inline void
CDecoderGT::SetBeamFactor( float BeamFactor )
{
  if ( BeamFactor < 1.0 ) {
	cerr << "Beam factor " << BeamFactor << " has to be >= 1.0" << endl;
	exit ( -2 );
  }

  m_BeamFactor = BeamFactor;
}



inline void
CDecoderGT::SetMaxHyps( int MaxHyps )
{
  m_MaxHyps = MaxHyps;
}



inline void
CDecoderGT::SetNumOfNBest( int N )
{
  m_NumOfNBest = N;
  cout << "SetNumOfNBest to " << m_NumOfNBest << endl;
}



inline void
CDecoderGT::SetSLStretchFactor( float SLStretch )
{
  m_SLStretch = SLStretch;
}



inline void
CDecoderGT::SetCompareHypFeatures( bool UseCoverage, bool UseCoveragePattern, 
				  bool UseLM, bool UseTargetLength )
{
  m_pDecoderInfo->m_UseCoverage = UseCoverage;
  m_pDecoderInfo->m_UseCoveragePattern = UseCoveragePattern;
  m_pDecoderInfo->m_UseLMState = UseLM;
  m_pDecoderInfo->m_UseTargetLength = UseTargetLength;

  cout << "SetCompareHypFeatures " << UseCoverage  << " " << UseCoveragePattern
	   << " " << UseLM << " " << UseTargetLength << endl;
}


inline void
CDecoderGT::SetPruneHypFeatures( bool UseCoverage, bool UseCoveragePattern, 
				  bool UseLM, bool UseTargetLength )
{
  m_pDecoderInfo->m_UseCoverageForPruning = UseCoverage;
  m_pDecoderInfo->m_UseCoveragePatternForPruning = UseCoveragePattern;
  m_pDecoderInfo->m_UseLMStateForPruning = UseLM;
  m_pDecoderInfo->m_UseTargetLengthForPruning = UseTargetLength;

  cout << "SetPruneHypFeatures " << UseCoverage  << " " << UseCoveragePattern
	   << " " << UseLM << " " << UseTargetLength << endl;
}

inline void
CDecoderGT::SetScalingFactors( float TMScale, float LMScale,
							   float ALScale, float SLScale )
{
  m_TMScale = TMScale;
  m_LMScale = LMScale;
  m_ALScale = ALScale;
  m_SLScale = SLScale;

}


inline
void CDecoderGT::GetLengthOfReorderingWindow( int& LengthOfReorderingWindow ){
  LengthOfReorderingWindow = m_LengthOfReorderingWindow;
}


inline
void CDecoderGT::GetReorderingsPerWord( float& ReorderingsPerWord ){
  ReorderingsPerWord = m_LengthOfReorderingWindow;
}


inline
void CDecoderGT::GetNumOfNBest( int& N ){
  N = m_NumOfNBest;
}


inline
void CDecoderGT::GetCompareHypFeatures( bool& UseCoverage, bool& UseCoveragePattern,
					bool& UseLM, bool& UseTargetLength ){
  UseCoverage = m_pDecoderInfo->m_UseCoverage;
  UseCoveragePattern = m_pDecoderInfo->m_UseCoveragePattern;
  UseLM = m_pDecoderInfo->m_UseLMState;
  UseTargetLength = m_pDecoderInfo->m_UseTargetLength;
}


inline
void CDecoderGT::GetPruneHypFeatures( bool& UseCoverage, bool& UseCoveragePattern,
				      bool& UseLM, bool& UseTargetLength){
  UseCoverage = m_pDecoderInfo->m_UseCoverageForPruning;
  UseCoveragePattern = m_pDecoderInfo->m_UseCoveragePatternForPruning;
  UseLM = m_pDecoderInfo->m_UseLMStateForPruning;
  UseTargetLength = m_pDecoderInfo->m_UseTargetLengthForPruning;
}


inline
void CDecoderGT::GetBeamFactor( float& BeamFactor){
  BeamFactor = m_BeamFactor;
}


inline
void CDecoderGT::GetSLStretchFactor( float& SLStretch ){
  SLStretch = m_SLStretch; 
}


inline
void CDecoderGT::GetScalingFactors( float& TMScale, float& LMScale,
				    float& ALScale, float& SLScale){
  TMScale = m_TMScale;
  LMScale = m_LMScale;
  ALScale = m_ALScale;
  SLScale = m_SLScale;
}

#endif
