#include <FutureEstimator.hh>
#include <iomanip>

CFutureEstimator::CFutureEstimator(CNumericalArray<double>& scalingFactors, CNumericalArray<double>& srcFeatures, SriLM* pLM){

  m_srcFeatures = srcFeatures;
  m_scalingFactors = scalingFactors;  
  //initialize the default src word cost
  SetSrcModels(scalingFactors, srcFeatures);
  m_pLM = pLM;

}

void CFutureEstimator::Init(CSentence* pSentence, CLattice* pLattice){


  m_pSourceSentence = pSentence;

  //clear out the memory if any is used in the estimator
  m_minCostStructure.Clear();
  m_minCostStructure.SetBoundaries(1, m_pSourceSentence->Size(), 1, m_pSourceSentence->Size());

  //initialize the cost matrix max costs to be i*m_srcWordCost
  for (int i=1; i<=m_pSourceSentence->Size(); i++){
    for (int j=1; j<=m_pSourceSentence->Size()-i+1; j++){
      m_minCostStructure.GetRow(i)[j] = i*m_srcWordCost;
    }
  }
  //the cells that are not accessed by the loop above are never considered
  //since the j+i > # src words

  
  //go thru the nodes of the lattice
  CArray<CLatticeNode*> Nodes = pLattice->Nodes();
  for (int n=Nodes.MinIndex(); n<=Nodes.MaxIndex(); n++){
    CLatticeNode* pNode = Nodes[n];
    const CArray< CLatticeEdge * > & Edges = pNode->OutEdges();
    for (int e=Edges.MinIndex(); e<=Edges.MaxIndex(); e++){
      CLatticeEdge* pEdge = Edges[e];
      int wordsTranslated = pEdge->ToNode()->Number() - pEdge->FromNode()->Number();
      int startingAt = n+1; 
      //cout<<"startingAt: "<<startingAt<<" wordsTranslated: "<<wordsTranslated<<endl;
      //pEdge->Display();
      
      //cout<<" about to call cost: "<<endl;
      CheckMin(CostFromEdge((CLatticeEdgeMS*)pEdge), wordsTranslated, startingAt);
    }
  }
}

void CFutureEstimator::SetSrcModels(CNumericalArray<double>& scalingFactors, CNumericalArray<double>& srcFeatures){
  m_srcWordCost = 0.0;
  int offset = 0;
  
  for (int i=srcFeatures.MinIndex(); i<=srcFeatures.MaxIndex(); i++){
    m_srcWordCost += srcFeatures[i]*scalingFactors[scalingFactors.MinIndex()+offset];
    offset++;
  }
    
  cout<<"m_srcWordCost: "<<m_srcWordCost<<endl;
}

void CFutureEstimator::ParseMinCost(){
 
  //the parsing process used to fill each cell of the m_minCostStructure table
  for (int i=1; i<=m_pSourceSentence->Size(); i++){
    for (int j=1; j<=m_pSourceSentence->Size()-i+1; j++){
      //try to apply the fundamental rule between the cell (i,j) and the cell (j+i, *)
      
      //see cyk+ paper for definitions of i,j,k
      for (int k=1; k<i; k++){

	//cout<<"Checking: ("<<k<<","<<j<<") with ("<<i-k<<","<<j+k<<") ="<<GetFutureCostEstimate(k,j)<<" + "<<GetFutureCostEstimate(i-k, j+k)<<" with ("<<i<<","<<j<<") = "<<GetFutureCostEstimate(i,j)<<endl;	
	CheckMin(GetFutureCostEstimate(k,j)+GetFutureCostEstimate(i-k, j+k), i, j);

      }
      
    }
  }
}


void CFutureEstimator::CheckMin(double cost, int i, int j){
  if (cost<m_minCostStructure.GetRow(i)[j]){
    m_minCostStructure.GetRow(i)[j] = cost;
  }
}


double
CFutureEstimator::CostFromEdge(CLatticeEdgeMS* pEdge)
{

  //cout<<"About to check scores: "<<endl;

  CWordInfo *pWordInfo =  ((CLatticeEdgeMS *)pEdge)->WordInfo();
  //cout<<" pWordInfo: "<<pWordInfo<<endl;
  double cost = 0.0;


  //fill this with target words if possible

  CArray<CString> Words;
  //calculate TM cost if there is a target sequence here, otherwise use the default values
  if (pWordInfo->HasTargetString()){

    //setup the sequence
    pWordInfo->GetTargetWords(Words);
    //fill the word sequence vector for the LM cost
    vector<string> wordSequence;
    for (int w=Words.MinIndex(); w<=Words.MaxIndex(); w++){
      //cout<<"Filling: "<<Words[w]<<endl;
      wordSequence.push_back(string((const char*)Words[w]));
    }

    // do each model in order
    int offset =0;

    // language model cost
    double lmCost = -1*(m_pLM->IsolatedLogProb(wordSequence));
    cost += lmCost*m_scalingFactors[m_scalingFactors.MinIndex()+offset];
    //cout<<"lmCost: "<<lmCost<<" scaled: "<<lmCost*m_scalingFactors[m_scalingFactors.MinIndex()+offset]<<endl;
    offset++;

    //AL cost (not done at this stage)
    //cout<<" alCost: not done at this stage"<<endl;
    offset++;

    //SL cost = number of words translated on this edge
    int numWordsAtEdge = pWordInfo->TargetStringLength();
    cost += numWordsAtEdge*m_scalingFactors[m_scalingFactors.MinIndex()+offset];
    //cout<<"slCost: "<<numWordsAtEdge<<" scaled: "<<numWordsAtEdge*m_scalingFactors[m_scalingFactors.MinIndex()+offset]<<endl;
    offset++;

    //PL cost = number of edges used -- DO THIS NOW!!!!!!!!!!!
    cost += 1.0*m_scalingFactors[m_scalingFactors.MinIndex()+offset];
    //cout<<"plCost: "<<1.0<<" scaled: "<<1.0*m_scalingFactors[m_scalingFactors.MinIndex()+offset]<<endl;
    offset++;

    CNumericalArray<float> scores = pEdge->Scores();    
    //cout<<"Inside tgt section: "<<endl;
    for (int i=scores.MinIndex(); i<=scores.MaxIndex(); i++){
      cost += scores[i]*m_scalingFactors[m_scalingFactors.MinIndex()+offset];
      //cout<<"tmCost: "<<i<<"  tmCost: "<<scores[i]<<" scaled: "<<scores[i]*m_scalingFactors[m_scalingFactors.MinIndex()+offset]<<endl;
      offset++;
    }

  }else{

    //default m_srcWordCost calculatd in Init()
    int numWords = pEdge->ToNode()->Number() - pEdge->FromNode()->Number();
    cost = numWords*m_srcWordCost;
    //cout<<"TotalCost: "<<cost<<endl;
  }

  return cost;
}


double CFutureEstimator::GetFutureCostEstimate(int i, int j){
  return m_minCostStructure.GetRow(i)[j];
}

void CFutureEstimator::Display(){
  for (int i=1; i<=m_pSourceSentence->Size(); i++){
    for (int j=1; j<=m_pSourceSentence->Size()-i+1; j++){
      cout<<setprecision(3)<<"("<<i<<","<<j<<") = "<<m_minCostStructure.GetRow(i)[j]<<" ";
    }
    cout<<endl;
  }
}
