#include <LatticeRescorer.hh>


void
CLatticeRescorer::SetBoostFactor( float BoostFactor )
{
  m_BoostFactor = BoostFactor;
}



void
CLatticeRescorer::Rescore( CLatticeLM *pLattice, CLexicon *pLexicon )
{

  Rescore( pLattice, pLexicon, NULL, 1.0 );

  return;
}




void
CLatticeRescorer::Rescore( CLatticeLM *pLattice, CLexicon *pLexicon,
                           CLexicon *pReverseLexicon, float Lambda )
{
//  if ( pReverseLexicon ) {
  //  cout << "Rescorer - rescore with two lexicons" << endl;
  //} else {
  //  cout << "Rescorer - rescore with one lexicon" << endl;
 // }
 
  CArray< CLatticeEdge * > Edges = pLattice->Edges();

  // array to collect words attached to one edge for LM-scoring
  CArray< CString > Words;
  CArray< LMWordIndex > WordIndices;

  // first get the words
  CString Word;
  string word;
  CLatticeEdgeTrans *pEdge;


  for ( int i = Edges.MinIndex(); i <= Edges.MaxIndex(); i++ ) {
	pEdge = (CLatticeEdgeTrans *)Edges[ i ];
	
	Word = pEdge->Word();
	word = Word;
	CWordInfo *pWordInfo =  ((CLatticeEdgeTrans *)pEdge)->WordInfo();
	
	Trace( 2, "WordInfo: " << *pWordInfo );
	Trace( 2, "SourceString: " << pWordInfo->SourceString() );
	Trace( 2, "TargetString: " << pWordInfo->TargetString() );
	
	// if edge was not generated by a transducer item
	// then set TM score to TMScoreForSourceWord
	if ( Word[ 0 ] != '@' ) {
	  //	  pEdge->WordInfo()->Score() = c_TMScoreForSourceWord;
	  pEdge->WordInfo()->Score() = 10.0;
	  //	  TraceIf( s_TraceOn, 0, "      Translation Score for source language edge set to : " 
	  //	   << c_TMScoreForSourceWord );
	} else {

	  CSentence &SentenceE = pEdge->WordInfo()->TargetWords();
	  CSentence &SentenceF = pEdge->WordInfo()->SourceWords();

	  double LogP1, LogP2, LogP;
	  LogP1 = CalcLogProb( SentenceF, SentenceE, pLexicon );
	  if ( pReverseLexicon ) {
		LogP2 = CalcLogProb( SentenceE, SentenceF, pReverseLexicon );
		// this score is taken as average per word, as otherwise 
		// long translations will be heavily penalized
		//cout << "LogP1 = " << LogP1 << " divided by length of source phrase = " << SentenceF.Length() 
		//	 << " = " << LogP1 / SentenceF.Length() << endl;
		//cout << "LogP2 = " << LogP2 << " divided by length of target phrase = " << SentenceE.Length() 
		//	 << " = " << LogP2 / SentenceE.Length() << endl << endl;
		LogP1 /= SentenceF.Length();
		LogP2 /= SentenceE.Length();

		LogP = ( 1 - Lambda ) * LogP1 + Lambda * LogP2;
		//cout << "LogP1/w = " << LogP1 << " LogP2/w = " << LogP2 << " LogP/w = " << LogP << endl;
		LogP *= SentenceF.Length();
		//cout << "Interpolated LogP = " << LogP << endl;

	  } else {
		LogP = LogP1;
		//cout << "LogP1 = " << LogP1 << " LogP = " << LogP << endl;
	  }

	  // boost the probability
	  int I = SentenceE.Length();
	  
	  double BoostedLogP = ( 1 / (1 + m_BoostFactor * (I - 1 )) ) * LogP;
	  //cout << "LogP = " << -LogP << " boosted LogP = " << -BoostedLogP << endl;
	  
	  // set the phrase probability for this edge to the LogProb we just got
	  // note: this is the negative log prob.
	  pEdge->WordInfo()->Score() = -BoostedLogP;
	}

  }
}


double
CLatticeRescorer::CalcLogProb( const CSentence & SentenceF,
                               const CSentence & SentenceE, 
                               const CLexicon *pLexicon )
{
  // NOTE: multiply over the F's (sourcewords) 
  //       and sum over the E's (targetwords) inside each product term 

  double LogP = 0.0;
  double SumP = 0.0;
  CWordIndex FwordIndex, EwordIndex;
  
  // Go over all words in the source sentence 
  for( int j = SentenceF.MinIndex(); j <= SentenceF.MaxIndex(); j++ ){

	// get the Lexicon index for the source word
	FwordIndex = pLexicon->SourceVocabulary()->GetIndex( SentenceF.Word( j ) );
	Trace( 3, "Calculating prob for source word " << SentenceF.Word( j ) 
		   <<" at index " << FwordIndex << endl );
	
	// For each target word, sum all the probs of the source word 
	// given this target word
	SumP = 0.0;
	for( int i = SentenceE.MinIndex(); i <= SentenceE.MaxIndex(); i++){
	  
	  // get the Lexicon index for the target word and add Pr to the Sum
	  //
	  EwordIndex = pLexicon->TargetVocabulary()->GetIndex( SentenceE.Word( i ) );
	  SumP += pLexicon->Pr( FwordIndex, EwordIndex );
	  Trace( 3, " next corresponding target word " << SentenceE.Word( i ) << " at index " << EwordIndex << endl );
	  
	  // debugging information //
	  double p;
	  p = pLexicon->Pr( FwordIndex, EwordIndex );
	  //cout << "Adding Lexical prob for " << SentenceF.Word( j ) << " " << FwordIndex << " " 
	  // << ", " << SentenceE.Word( i ) << " " << EwordIndex << ": " << p << endl;

	}
	// multiply the prob of the phrase by the prob of this src word
	// (by adding the Logprob)
	LogP += log( SumP );
  }
  
  return LogP;
}

