//
// Evaluator Base Class (CEvaluator), NGram class (CEvaluatorNGram)
// 
// Programming: Anthony D'Auria <apd@andrew.cmu.edu>
//
// 03/2006
// 


#ifndef __Evaluator__
#define __Evaluator__


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sstream> 

#include <NumericalArray.hh>
#include <Vocabulary.hh>
#include <Text.hh>
#include <TextFilterBoostRegex.hh>



//====================================================================
//      CEvaluatorNGram
//
//   Represents ngram as CWordIndex*
//   - Gives signature to CWordIndex* as a sum of products of word indices with 
//     primes, with a distinct primes for every n in the ngram.
//     - Note: this signature is not unique when an ngram contains the right 
//       combination of word indices that are equivalent to prime numbers used 
//       in the function. Experimentally, this happens very infrequently. 
//       Therefore, equivalence can't be assumed when signature value is equal 
//       between two ngrams.
//   - Hash value is generated once

class CEvaluatorNGram
{
public:
  
  CEvaluatorNGram ();
  
  //! Creates a new CEvaluatorNGram with one word
  CEvaluatorNGram ( const CWordIndex word );
  
  //! Creates a new NGram with contents of passed ngram. 
  //  Last element in new ngram is the CWordIndex word
  CEvaluatorNGram ( CEvaluatorNGram &ngram , size_t size, const CWordIndex word );
  
  CEvaluatorNGram ( CEvaluatorNGram* p_prevNGram, size_t newSize);
  
  ~CEvaluatorNGram();
  
  unsigned int getSignature() const;  
  size_t       getSize() const;
  CWordIndex*  getWords() const;
  void         Display();
  CString      ToString(CVocabulary* pVoc) const;
  CString      ToString() const;
  
  //! Used by NGram_hash_map to compare two CEvaluatorNGram keys
  struct equalCEvaluatorNGram
  {
    bool operator()(const CEvaluatorNGram* p_n1, const CEvaluatorNGram* p_n2) const
    {
    
    //return (p_n1->getSignature() == p_n2->getSignature()); // this isn't guaranteed

    bool eq = true;
    
    CWordIndex* pp_n1Words = p_n1->getWords();
    CWordIndex* pp_n2Words = p_n2->getWords();
    
    int size = p_n1->getSize();           
    
    if(size != p_n2->getSize()){
      eq = false;
    } else {
      for(int i=0; eq && i < size; ++i)
        if(pp_n1Words[i] != pp_n2Words[i]) 
          eq = false;
    }
    /*
    // debug: hash eq check
    if((p_n1->getSignature() == p_n2->getSignature()) == eq ){
      //cout << " HASH correct (" << (p_n1->getSignature() == p_n2->getSignature()) << "," << (eq) << ") ";
    } else {
      //cout << " HASH ***INCORRECT*** (" << (p_n1->getSignature() == p_n2->getSignature()) << "," << (eq) << ") ";
    }
     */
    return eq; 
    

  }
  };  
  
private:
  //! array of ngram elements
  CWordIndex* m_p_words;             
  
  //! size of ngram
  size_t m_size;
  
  //! Signature of ngram used in hashing. Not guaranteed to be unique.
  mutable unsigned int m_signature;
};

//--makes CEvaluatorNGrams a hashable key in __gnu_cxx::hash_map
//  default hash function is passed a signature of the ngram
namespace __gnu_cxx {
  template<> struct hash< CEvaluatorNGram*> {
    size_t operator()(const CEvaluatorNGram* ngram) const {
      return hash<int>()( ngram->getSignature() );
    }
  };
}

//-- hash map <CEvaluatorNGram*,int>
typedef __gnu_cxx::hash_map< CEvaluatorNGram*, int, __gnu_cxx::hash< CEvaluatorNGram* >,
  CEvaluatorNGram::equalCEvaluatorNGram > NGram_hash_map;

//-- NGramToDouble_hash_map
//   Maps ngrams to info gain
typedef __gnu_cxx::hash_map< CEvaluatorNGram*, double, __gnu_cxx::hash< CEvaluatorNGram* >, CEvaluatorNGram::equalCEvaluatorNGram > NGramToDouble_hash_map;

//-- NGramToFloat_hash_map
//   Maps ngrams to float, when higher precision than double is necessary
typedef __gnu_cxx::hash_map< CEvaluatorNGram*, float, __gnu_cxx::hash< CEvaluatorNGram* >, CEvaluatorNGram::equalCEvaluatorNGram > NGramToFloat_hash_map;

//====================================================================
//      CEvaluatorNGramMatchStats
//
//    needs to be on main function level
//    could be replaced with Cumulative Scores (which can be renamed to CEvaluatorScores)

class CEvaluatorNGramMatchStats
{
public:
  CEvaluatorNGramMatchStats();         // TODO: pass max ngram size
  ~CEvaluatorNGramMatchStats();
  
  size_t closestRefSentLength;         // BLEU
  size_t totalRefWords;
  size_t hypSentLength;                // gives numNGramsInHyp
  unsigned short* matchingNGrams;      // [ngram size+1]
  float* infoGain;

  // hypothesis that generate this ngram match object
  CSentence* pHypSentence;
  
  short hypSentenceIndex;
  
  float  avgRefSentLength;             // NIST
 
};

//====================================================================
//      CEvaluatorSharedData
//
//      Contains data that can be reused by evaluation metrics
//      

class CEvaluatorSharedData
{
public:

  CEvaluatorSharedData();
  ~CEvaluatorSharedData();
  
  bool ToComputeMatchCounts();
  bool ToComputeMatchGain();
  bool ToComputeNistLengths();
  bool ToComputeBleuLengths();
  bool ToComputeNistBleuLengths();
  bool ToBuildMaxRefHash();
  bool ToBuildGainHash();
  
  //-- Get methods ------
  CText*                          GetTranslations();
  CArray< CArray< CSentence* > >* GetNBestTranslations();
  CArray< CArray< CSentence* > >* GetReferences();
  NGram_hash_map**                GetMaxRefHash();
  NGramToDouble_hash_map*         GetGainHash();
  CArray< CArray< CEvaluatorNGramMatchStats* > >* GetHypMatchStats();  
  CTextFilterBase*                GetTextFilter();
  
  //-- Set methods ------
  void SetToComputeMatchCounts(bool b);
  void SetToComputeMatchGain(bool b);
  void SetToComputeNistLengths(bool b);
  void SetToComputeNistBleuLengths(bool b);
  void SetToComputeBleuLengths(bool b);
  void SetReferences(CArray< CArray< CSentence* > >* p_references);
  void SetMaxRefHash(NGram_hash_map** pp_maxRefHash);
  void SetGainHash(NGramToDouble_hash_map* p_gainHash);
  void SetHypMatchStats(CArray< CArray< CEvaluatorNGramMatchStats* > >* p_hypMatchStats);
  
  //! Reset all computed metrics, delete p_nBestTranslations and p_translations
  //! Called by CEvaluator::LoadNBestTranslations
  void SetNBestTranslations( CArray< CArray< CSentence* > >* p_nBestTranslations );
  
  //! Reset all computed metrics, delete p_translations and p_nBestTranslations
  //! Called by CEvaluator::LoadAsciiText
  void SetTranslations( CText* p_translations );
  
  //! Set TextFilterBase object that will normalize and tokenize text loaded
  //! via the Evaluator instance.
  void SetTextFilter(CTextFilterBase*);
  
protected:
  //! Deletes translation container and resets flags and runs ResetStats()
  void ResetTranslations();
  
  //! Reset hyp and ref stats flags and container
  void ResetStats();
  
  //! Clear reference data
  void ResetReferences();
  
  //! stats flags checked by evaluation metrics that need match stats
  bool m_toComputeMatchGain;
  bool m_toComputeMatchCounts;
  bool m_toComputeNistLengths;
  bool m_toComputeBleuLengths;
  bool m_toComputeNistBleuLengths;
  
  //! 1-best translations [src index][ref sent index]
  CText* m_p_translations;
  
  //! n-best translations [src index][ref sent index]
  CArray< CArray< CSentence* > >* m_p_nBestTranslations;

  //! reference sentences [src index][ref sent index]
  CArray< CArray< CSentence* > >* m_p_references;
  
  //! For each [src index], hash of ngram->max count in set or refs
  NGram_hash_map** m_pp_maxRefHash;                            // BLEU & NIST
  
  //! ngram -> double gain score, used by NIST
  NGramToDouble_hash_map* m_p_gainHash; 
  
  //! stores match count objects for BLEU and NIST metrics
  CArray< CArray< CEvaluatorNGramMatchStats* > >* m_p_hypStats; // BLEU & NIST
  
  //! Text normalizing and tokenizing object that can be used by Evaluators
  //! on text they load from files
  CTextFilterBase* m_pTextFilter;
  
};


//====================================================================
//      CEvaluator
//
//      Base class for NIST, BLEU, WER evaluation
//

class CEvaluator
{
public:
  // construct an evaluator; give the vocabulary which will be used 
  // to map any word to a word ID
  // pass shared data object that will point to precomputed data, like 
  // ngram matches
  CEvaluator( CVocabulary *pVocabulary, CEvaluatorSharedData *pSharedData );
  CEvaluator( CVocabulary *pVocabulary );
  virtual ~CEvaluator();

  // load references, can be multiple:
  // Format positive n: there are always n references
  // Format zero: variable number of references, numbered as n m
  // where n >= 1 is the number of the corresponding source sentence
  // and m >= 1 is the number of the reference
  // the load references can build some initial statistics required
  // in the actual evaluaton
  virtual bool LoadReferences( const char *pFileName, int Format = 0, int LoadOnlySetNumber = -1 );
  
  // load NBestList
  // all Evaluator subclasses will load the n-best list the same way
  // Format positive n: there are always n hypotheses per source sentence
  // Format zero: variable number of hypotheses per source sentence
  // places p_NBestTranslations into SharedData object
  CArray< CArray< CSentence* > >* LoadNBestTranslations( const char *pFileName, 
                                                         int Format = 0 );
    
  // calculate the score or scores for the translation
  // some evaluation metrics have several internal scores, plus one total score
  // put total score at index 0 (1-dimensional array with results in a 
	// particular order)
  virtual bool Evaluate( const char *pFileWithTranslations, 
                         CNumericalArray< float > & Scores ) = 0;
  virtual bool Evaluate( CArray< CString > &Translations, 
                         CNumericalArray< float > & Scores ) = 0;
  virtual bool Evaluate( CText *pTranslations, 
                         CNumericalArray< float > & Scores ) = 0;
  virtual bool Evaluate( const CArray< CNumericalArray< CWordIndex > > *pTanslations, 
                         CNumericalArray< float > & Scores ) = 0;
  virtual bool Evaluate( int sourceSentenceIndex, CSentence* p_hypSentence,
                         CNumericalArray< float >& Scores ) = 0;
  virtual bool Evaluate( CSentence* pSentence,
                         CArray< CSentence* > &References,
                        CNumericalArray< float >& Scores ) = 0;
  

  // calculate the oracle score for a nbest list; Format: n m word sequence
  // oracle scores means: take those translations in nbest lists which optimize the scores
  // no matter if read from file, or getting the translations in arrays
  virtual bool GetOracleScore( const char *pFileWithNBestTranslations, 
                               CNumericalArray< float > & Scores ) = 0;
  virtual bool GetOracleScore( CArray< CString > *pNBestTranslations, 
                               CNumericalArray< float > & Scores ) = 0;
  virtual bool GetOracleScore( CText *pNBestTranslations, 
                               CNumericalArray< float > & Scores ) = 0;
  virtual bool GetOracleScore( CArray< CNumericalArray< CWordIndex > > *pNBestTranslations, 
                               CNumericalArray< float > & Scores ) = 0; // to replace
  virtual bool GetOracleScore( CArray< CArray< CSentence* > >* p_NBestTranslations,
                               CNumericalArray< float >& Scores ) = 0;
  
  //! Compute 95% confidence interval using bootstrapping 
  //! returns pair < score at 2.5 percentile, 97.5 percentile >
  //! uses CArray< CArray< CEvaluatorNGramMatchStats* > >* from SharedData
  virtual std::pair< float, float > GetConfidenceInterval() = 0;
  
  //! output report to standard out
  virtual void ReportScores( CNumericalArray< float > & Scores ) = 0;

  //! verbosity level
  //! 0 = none, 1 = only final result, 2 = results per sentence
  void SetVerbosityLevel( int VerbosityLevel );
  
  //! Write to file gain scores generated by OracleScore
  void SetWriteGainToFile( bool flag );
  void SetGainFileName( char* gainFileName );
  
  //! Set nbest search depth limit
  void SetNBestSearchDepth( size_t limit );
  size_t GetNBestSearchDepth();
  
  //! Set TextFilterBase object that will normalize and tokenize text loaded
  //! via the Evaluator instance.
  void SetTextFilter(CTextFilterBase* pTextFilter);
  CTextFilterBase* GetTextFilter();

  //! Display oracle-best sentences after computing the oracle-best score
  void SetDisplayOracleBestSentences( bool display );
  
  //! Display detailed information during evaluation process
  void SetVerboseOutput( bool verbose );

protected:
    
  NGram_hash_map* GetNGramsFromSentence( CSentence* p_sentence );

  //! Validate hypothesis sentences against sets of references
  //!  # of hyp sentences must equal number of reference sets
  //!    # of reference sentences per set may vary  
  bool validateData( CText *pTranslations );

  // references are for n source sentences, m references
  // each reference is an array of word IDs
		
	CVocabulary* m_pVocabulary;
  
  //! Normalizes and tokenizes according to perl-style regexps listed in a file
  //! that was passed as an argument to the CTextFilterBoostRegex constructor
  CTextFilterBase* m_pTextFilter;
	
  CEvaluatorSharedData* m_p_sharedData;
  
  int m_VerbosityLevel;

  //! maximum size of ngram. 9 by default, set in constructor
  int m_maxNGram; 
  
  //! name of file in which gain scores for nbest hyps will be stored
  char* m_gainFileName;
  
  //! flag which should be removed. m_gainFileName == NULL when not writing
  bool  m_writeGainToFile;
  
  //! limits how far to search for a better hypothesis within an nbest list 
  size_t m_nbestSearchDepth;
  
  //! Flag for verbose output
  bool m_verboseOutput;
  
  //! Flag for displaying oracle-best sentences, prefixed by ORACLE_BEST <sentence index>
  bool m_displayOracleBestSent;
  
};





#endif
