#include <stdlib.h>
#include <EvaluatorBLEU.hh>
#include <math.h>      
#include <fstream>
#include <typeinfo>


//====================================================================
//      CEvaluatorBLEU
//

CEvaluatorBLEU::CEvaluatorBLEU( CVocabulary *pVocabulary, 
                                CEvaluatorSharedData *pSharedData )
  : CEvaluator( pVocabulary, pSharedData )
{
  m_p_sharedData = pSharedData;
  m_N_PRECISION  = 4;                  //!< n-gram precision for final BLEU score
}

CEvaluatorBLEU::~CEvaluatorBLEU()
{
  
}


//------------------------------------------------------------------
//      CEvaluatorBLEU:: Evaluate
//

bool
CEvaluatorBLEU::Evaluate( CArray< CEvaluatorNGramMatchStats* >* p_hypMatchStats,
                          CNumericalArray< float > &Scores )
{
  
  CumulativeScores cumScores(m_maxNGram);                       // track aggregate values
  
  int sourceSentCount = p_hypMatchStats->Size();   // iterate over p_hypMatchStats
 
  for(int s=0; s < sourceSentCount; s++)
  {
    CEvaluatorNGramMatchStats* p_hypStats = (*p_hypMatchStats)[s];

    //-- add hyp length and closest ref length
    cumScores.hypSentLength += p_hypStats->hypSentLength;
    cumScores.closestRefSentLength += p_hypStats->closestRefSentLength;
    
    //-- add # of possible n-grams in hyp only
    for(int n=1; n <= m_maxNGram; n++) {
      cumScores.numNGramsInHyp[n] += p_hypStats->hypSentLength-n+1;
      cumScores.matchingNGrams[n] += p_hypStats->matchingNGrams[n];
    }
  } // end source sent loop
  
  
  //-- length penalty
  double bleuLengthPenalty 
    = std::max( ((double)cumScores.closestRefSentLength) 
                / ((double)cumScores.hypSentLength) -1, (double) 0); 
  
  //-- cumulative scores
  double runningScore = 0;
  
  for(int n=1; n <= m_maxNGram; n++)
  {
    if( cumScores.matchingNGrams[n] != 0) 
    {
      double matchRatio = ( (double)(cumScores.matchingNGrams[n]) 
                            / (double)(cumScores.numNGramsInHyp[n]) );
      
      runningScore += log(matchRatio);
      
      Scores[n] = exp(runningScore/n - bleuLengthPenalty);
      
    } else {
      Scores[n] = 0;
    }
    
  } // end n loop
  
  Scores[10] = cumScores.hypSentLength;
  Scores[11] = cumScores.closestRefSentLength;
  
  Scores [0] = Scores[m_N_PRECISION];
  
  return true;
  
}


bool
CEvaluatorBLEU::Evaluate( CText* pTranslations,
                          CNumericalArray< float > &Scores )
{
  m_p_sharedData->SetTranslations(pTranslations); // will not reset if same ptr
  
  if ( m_p_sharedData->ToComputeMatchCounts() 
       || m_p_sharedData->ToComputeBleuLengths() )
  {

    // convert pTranslations to nBestList format
    
    CArray< CArray< CSentence* > >* p_1BestList = new CArray< CArray< CSentence* > >;
    
    for(int i=pTranslations->MinIndex(); i <= pTranslations->MaxIndex(); i++){
      (*p_1BestList)[i][0] = pTranslations->GetSentence(i);
    }
    
    ComputeNGramMatchStats(p_1BestList);
  }
  
  CArray< CArray< CEvaluatorNGramMatchStats* > >* p_hypMatchStats 
    = m_p_sharedData->GetHypMatchStats();
  
  
  CArray< CEvaluatorNGramMatchStats* >* p_baselineStats      
    = new CArray< CEvaluatorNGramMatchStats* >; 

  
  for(int i=p_hypMatchStats->MinIndex(); i <= p_hypMatchStats->MaxIndex(); i++){
    (*p_baselineStats)[i] = (*p_hypMatchStats)[i][0];
  }

  bool evalSuccess = Evaluate( p_baselineStats, Scores );
  
  delete p_baselineStats;
  
  return evalSuccess;
  
}


bool
CEvaluatorBLEU::Evaluate( const char *pFileWithTranslations, 
                          CNumericalArray< float > & Scores )
{
}


bool 
CEvaluatorBLEU::Evaluate( CArray< CString > &Translations, 
                          CNumericalArray< float > & Scores )
{
}

/**
 * @brief Score one sentence with the given set of references
 *
 *
 */
bool 
CEvaluatorBLEU::Evaluate( CSentence* pHypSentence, CArray< CSentence* > &References,
                          CNumericalArray< float >& Scores )
{
  ;
}

bool
CEvaluatorBLEU::Evaluate( int sourceSentenceIndex, CSentence* p_hypSentence,
                          CNumericalArray< float >& Scores )
{

/*  
  //-- init cumulative scores
  CumulativeScores cumScores;
  cumScores.numNGramsInHyp = new int[m_maxNGram+1]; //1..9
  cumScores.matchingNGrams = new int[m_maxNGram+1]; //1..9
  
  for(int n=1; n <= m_maxNGram; n++) {
    cumScores.numNGramsInHyp[n] = 0;
    cumScores.matchingNGrams[n] = 0;  
  }
  cumScores.closestRefSentLength = 0;
  cumScores.hypSentLength = 0;
  
  //-- store max reference ngram counts once
  bool createMaxRefHash = (m_pp_maxRefHash == 0);
  
  if (createMaxRefHash) {
    m_pp_maxRefHash = new NGram_hash_map* [1];    // 1 hypothesis sentence
  }
  
  NGram_hash_map* p_maxRefHash;
  
  //-- hash ngram->{freq in sentence}
  NGram_hash_map* p_hypHash = GetNGramsFromSentence(p_hypSentence);
  
  //-- hash ngram->{max freq out of all references for hypothesis}
  if(createMaxRefHash)
    p_maxRefHash = new NGram_hash_map();
  
  //-- Select reference sentence closest in length to hypothesis
  //   of two equidistant ref sentences, choose shortest (IBM implementation)
  int hypSentLength         = p_hypSentence->Size();
  int smallestDiff          = 10000000;
  int smallestDiffRefLength = 10000000;
  int currRefLength, diff, refSentCount; 
  
  refSentCount = (int)m_refSentences[sourceSentenceIndex].Size();   // member of Evaluator
  
  //-- Select reference sentence closest in length to hypothesis
  CSentence* p_currRefSentence;
  for(int r=0; r < refSentCount; r++)
  {
    p_currRefSentence = m_refSentences[sourceSentenceIndex][r];
    
    // populate m_pp_maxRefHash[h] if not previously done
    if(createMaxRefHash)
    {
      NGram_hash_map* p_currRefHash = GetNGramsFromSentence(p_currRefSentence); // new
      p_maxRefHash = MaxValueHash(p_maxRefHash, p_currRefHash);
    }
    
    //-- Track closest length of ref sentences
    currRefLength = m_refSentences[sourceSentenceIndex][r]->Size();
    diff = std::abs( currRefLength - hypSentLength );
    
    if( diff <= smallestDiff )
    {
      if( diff == smallestDiff ) {
        if( currRefLength < smallestDiffRefLength ) {
          smallestDiffRefLength   = currRefLength;
        }
      } else {
        smallestDiff            = diff;
        smallestDiffRefLength   = currRefLength;
      }
    }
    
  } // end ref loop
  
  //-- save or retrieve hash of max # of ngram occurrences in all ref sentences
  if(createMaxRefHash) {
    m_pp_maxRefHash[sourceSentenceIndex] = p_maxRefHash;
  } else{
    p_maxRefHash = m_pp_maxRefHash[sourceSentenceIndex];
  }
  
  //-- add hyp length and closest ref length
  cumScores.hypSentLength        += hypSentLength;
  cumScores.closestRefSentLength += smallestDiffRefLength;
  
  //-- add # of possible n-grams in hyp only
  for(int n=1; n <= m_maxNGram; n++) {
    cumScores.numNGramsInHyp[n] += hypSentLength-n+1;
  }
  
  //-- count n-gram co-occurrences
  for( NGram_hash_map::iterator hypHash_iter = p_hypHash->begin(); 
       hypHash_iter != p_hypHash->end(); 
       hypHash_iter++) 
  {
    NGram_hash_map::iterator maxRefHash_iter = p_maxRefHash->find(hypHash_iter->first);
    
    if(maxRefHash_iter != p_hypHash->end())
    {
      int n = (hypHash_iter->first)->getSize();
      
      cumScores.matchingNGrams[n] 
        += std::min( hypHash_iter->second, maxRefHash_iter->second );
    }
    
    delete hypHash_iter->first;
    
    p_hypHash->clear();
    delete p_hypHash;
  }
  
  //-- length penalty
  double bleuLengthPenalty 
    = std::max( ((double)cumScores.closestRefSentLength) 
                / ((double)cumScores.hypSentLength) -1, (double) 0); 
  
  
  //-- cumulative scores
  Scores[0] = 0;
  double runningScore = 0;
  
  for(int n=1; n <= m_maxNGram; n++)
  {
    if( cumScores.matchingNGrams[n] != 0) 
    {
      double matchRatio = ( (double)(cumScores.matchingNGrams[n]) 
                            / (double)(cumScores.numNGramsInHyp[n]) );
      
      runningScore += log(matchRatio);
      
      Scores[n] = exp(runningScore/n - bleuLengthPenalty);
      
    } else {
      Scores[n] = 0;
    }
    
  } // end n loop
  
  Scores[10] = cumScores.hypSentLength;
  Scores[11] = cumScores.closestRefSentLength;
  
  // Scores[10]  
  // hypSentLength
  // closestRefLength
  
  delete [] cumScores.numNGramsInHyp;
  delete [] cumScores.matchingNGrams;
  
  return true;
*/  
}

  /*! Runs BLEU evaluation metric on translations in the 
      CArray< CNumericalArray< CWordIndex> > container.
    
      \note Not implemented.
    
      \todo Currently not needed. Should be in abstract class. Can run virtual
        Evaluate function.
  */

bool
CEvaluatorBLEU::Evaluate( const CArray< CNumericalArray< CWordIndex > > *pTanslations,
                          CNumericalArray< float > & Scores )
{
}


//------------------------------------------------------------------
//      CEvaluatorBLEU:: GetOracleScore
//

bool
CEvaluatorBLEU::GetOracleScore( const char *pFileWithNBestTranslations, 
                                CNumericalArray< float > & Scores )
{
  
  CArray< CArray< CSentence* > >* p_NBestList 
    = LoadNBestTranslations(pFileWithNBestTranslations);
  
  cout << "Loaded NBest Translations from " << pFileWithNBestTranslations << endl;
  
  GetOracleScoreJoy1(p_NBestList, Scores);  // updates SharedData
  
  //GetOracleScoreAnthony1(p_NBestList, Scores);
  
  return true;
  
}



bool
CEvaluatorBLEU::GetOracleScore( CArray< CString > *pNBestTranslations, 
                                CNumericalArray< float > & Scores )
{
}



bool
CEvaluatorBLEU::GetOracleScore( CText *pNBestTranslations, 
                                CNumericalArray< float > & Scores )
{
}


bool
CEvaluatorBLEU::GetOracleScore( CArray< CNumericalArray< CWordIndex > > *pNBestTranslations,
                                CNumericalArray< float > & Scores )
{
}


bool
CEvaluatorBLEU::GetOracleScore( CArray< CArray< CSentence* > >* p_NBestTranslations,
                                CNumericalArray< float >& Scores )
{
}

bool
CEvaluatorBLEU::GetOracleScoreJoy1( CArray< CArray< CSentence* > >* p_NBestTranslations,
                                    CNumericalArray< float >& Scores )
{

  // search control vars
  int   maxSearches     = 10; 
  float baselineGain    = 0;           // baselinePrime's gain over baseline
  float baselineGainMin = 0.5;         // can be used to stop searching
  
  //-- prepare to create gain file, if necessary
  std::ofstream gainTableOutFile;    
  if( m_writeGainToFile )
  {
    maxSearches = 1;                   // perform only 1 search to save gain scores
    gainTableOutFile.open( m_gainFileName );
    if( !gainTableOutFile ) {
      std::cerr << "Unable to open gain file \"" << m_gainFileName 
                << "\" for writing! Exiting..."  << endl;
      exit(1);
    }
  }

  //-- Update SharedData object. If p_NBestTranslations has changed, SharedData
  //   will reset ngram match stats
  m_p_sharedData->SetNBestTranslations(p_NBestTranslations);
  
  std::string thisClassName = typeid(*this).name();
  
  if ( m_verboseOutput ) {
    cout << "Computing ngram match stats..." << endl;
    //cout << "#### TYPE of this object: " << typeid(*this).name() << endl;  
  }
  
  
  
  //-- Compute NGram Matches, if necessary
  if( m_p_sharedData->ToComputeMatchCounts() 
      || ( m_p_sharedData->ToComputeBleuLengths() && thisClassName == "14CEvaluatorBLEU" )
      || ( m_p_sharedData->ToComputeNistLengths() && thisClassName == "14CEvaluatorNIST" )
      )
  {
    ComputeNGramMatchStats(p_NBestTranslations);
  }
  
  // 2d array of match scores [sys][hyp]
  CArray< CArray< CEvaluatorNGramMatchStats* > >* p_hypStats = m_p_sharedData->GetHypMatchStats();
  
  //-- set baseline translation to first-best
  CArray< CEvaluatorNGramMatchStats* >* p_baselineStats      
    = new CArray< CEvaluatorNGramMatchStats* >; 
  
  CArray< CEvaluatorNGramMatchStats* >* p_baselinePrimeStats 
    = new CArray< CEvaluatorNGramMatchStats* >;
  
  for(int i=p_hypStats->MinIndex(); i <= p_hypStats->MaxIndex(); i++){
    (*p_baselineStats)     [i] = (*p_hypStats)[i][0];
    (*p_baselinePrimeStats)[i] = (*p_hypStats)[i][0];
  }
  CNumericalArray< float > baselineScores;
  Evaluate(p_baselineStats, baselineScores);
   
  cout << "Baseline Scores: " << endl;
  ReportScores(baselineScores);
  
  //-- begin search iterations over baseline/baselinePrime
  for ( int searchCount = 1; 
           ( searchCount <= maxSearches )
        && ( searchCount == 1 || baselineGain > 0 ) ; searchCount++ )
  {
    cout << endl << "Starting search iteration " << searchCount << "..." << endl;
    // find hyp with maximum gain for each source sentence
    CNumericalArray< float > baselinePrimeScores;
            
    for(int s = p_hypStats->MinIndex(); s <= p_hypStats->MaxIndex(); s++)
    {
      float maxHypGain   = 0;
      int   bestHypIndex = 0;                // if no gain over baseline
      
      CEvaluatorNGramMatchStats* p_baselineHyp = (*p_baselineStats)[s];

      // save baseline in gain file
      if( m_writeGainToFile ) {
        gainTableOutFile << s << " " << 0 << " " << 0 << endl;
      }
      
      
      // h=0 is baseline, so start with h=1
      for(int h = 1; h<=(*p_hypStats)[s].MaxIndex() && h < m_nbestSearchDepth; h++)
      {
        // temporarily modify baseline with current hypothesis
        (*p_baselineStats)[s] = (*p_hypStats)[s][h];
        Evaluate(p_baselineStats, baselinePrimeScores);
        
        //use 4-gram for BLEU score
        float currHypGain = baselinePrimeScores[m_N_PRECISION] - baselineScores[m_N_PRECISION];
        
        // save gain to a file here
        if( m_writeGainToFile ) {
          gainTableOutFile << s << " " << h << " " << currHypGain << endl;
        }

        if( currHypGain > maxHypGain ) {
          bestHypIndex = h;
          maxHypGain = currHypGain;
        }
        
        // restore baseline's hypothesis
        (*p_baselineStats)[s] = p_baselineHyp;
        
      } // hyp sent loop
      
      
      // check if there's gain over baseline
      if ( maxHypGain > 0 )
      {
        // modify baselinePrime with stats of greatest hyp
        (*p_baselinePrimeStats)[s] = (*p_hypStats)[s][bestHypIndex];

        if( ! m_writeGainToFile && m_verboseOutput ) { 
          cout << "picked s: " << s << " bestHypIndex: " << bestHypIndex << " gain: " << maxHypGain << endl; 
        }
        
      }
      
      if( m_writeGainToFile && m_verboseOutput ) {
        cout << "Source sent: " << s << " Num of gains written: " 
             << (*p_hypStats)[s].MaxIndex() << endl;
      }
      
      
    } // source sent loop
    
    
    if( ! m_writeGainToFile ) 
    {
      // evaluate baselinePrime with best hypotheses and compare with baseline
      Evaluate(p_baselinePrimeStats, baselinePrimeScores);
      baselineGain = baselinePrimeScores[m_N_PRECISION] - baselineScores[m_N_PRECISION];
      
      cout << "Gain over baseline: " << baselineGain << endl;    
      
      if ( baselineGain > 0 )
      {
        cout << "Changing baseline to baselinePrime. Continuing search." << endl;
        for(int s = 0; s <= p_baselinePrimeStats->MaxIndex(); s++) {
          (*p_baselineStats)[s] = (*p_baselinePrimeStats)[s];      
        }
        baselineScores = baselinePrimeScores;
        ReportScores(baselineScores);

      } else {
        cout << "No gain from iteration. Stopping search." << endl;
        
        if ( m_displayOracleBestSent ) {
          cout << endl << "Printing Oracle-best sentences..." << endl;
          PrintOracleBestSentences( p_baselineStats );
        }
      }
      
    }

  } // search loop
  
  if( m_writeGainToFile ) {
    gainTableOutFile.close();
  } 

  Scores = baselineScores;

  delete p_baselineStats;
  delete p_baselinePrimeStats;
  
  return true;
}


bool
CEvaluatorBLEU::GetOracleScoreAnthony1( CArray< CArray< CSentence* > >* p_NBestTranslations,
                                        CNumericalArray< float >& Scores )
{
  
  // TODO: check if BLEU or NIST is calling this function
  
  m_p_sharedData->SetNBestTranslations(p_NBestTranslations);
  if( m_p_sharedData->ToComputeMatchCounts()
      || m_p_sharedData->ToComputeBleuLengths() )
  {
    ComputeNGramMatchStats(p_NBestTranslations);
  }
  
  // 2d array of match scores [sys][hyp]
  CArray< CArray< CEvaluatorNGramMatchStats* > >* 
    p_hypStats = m_p_sharedData->GetHypMatchStats();
  
  // search control vars
  int   maxSearches     = 10; 
  float baselineGain    = 0;             // baselinePrime's gain over baseline
  float baselineGainMin = 0.5;           // can be used to stop searching
  
  //-- set baseline translation to first-best
  CArray< CEvaluatorNGramMatchStats* >* 
    p_baselineStats      = new CArray< CEvaluatorNGramMatchStats* >; 
  CArray< CEvaluatorNGramMatchStats* >* 
    p_baselinePrimeStats = new CArray< CEvaluatorNGramMatchStats* >;
  for(int i=p_hypStats->MinIndex(); i <= p_hypStats->MaxIndex(); i++){
    (*p_baselineStats)     [i] = (*p_hypStats)[i][0];
    (*p_baselinePrimeStats)[i] = (*p_hypStats)[i][0];
  }
  CNumericalArray< float > baselineScores;
  Evaluate(p_baselineStats, baselineScores);
  
  cout << "Starting with Baseline Scores: " << endl;
  ReportScores(baselineScores);
  
  //-- begin search iterations over baseline/baselinePrime
  for(int searchCount = 0; searchCount < maxSearches; searchCount++)
  {
    cout << "Search iteration: " << searchCount << endl;

    // find hyp with maximum gain for each source sentence

    CNumericalArray< float > baselinePrimeScores;

    for(int s = p_hypStats->MinIndex(); s <= p_hypStats->MaxIndex(); s++)
    {
      float maxHypGain   = 0;
      int   bestHypIndex = 0;                // if no gain over baseline
      
      CEvaluatorNGramMatchStats* p_baselineHyp = (*p_baselinePrimeStats)[s];
      
      // h=0 is baseline, so start with h=1
      for(int h = 1; h<=(*p_hypStats)[s].MaxIndex() && h < m_nbestSearchDepth; h++)
      {
        // temporarily modify baseline with current hypothesis
        (*p_baselinePrimeStats)[s] = (*p_hypStats)[s][h];
        Evaluate(p_baselinePrimeStats, baselinePrimeScores);
        
        //use 4-gram for BLEU score
        float currHypGain = baselinePrimeScores[m_N_PRECISION] - baselineScores[m_N_PRECISION];
        
        if( currHypGain > maxHypGain ) {
          bestHypIndex = h;
          maxHypGain = currHypGain;
        } 
        
      } // hyp sent loop
      
      // modify baselinePrime with stats of greatest hyp
      if(maxHypGain > 0) 
      {
        (*p_baselinePrimeStats)[s] = (*p_hypStats)[s][bestHypIndex];
        cout << "picked s: " << s << " bestHypIndex: " << bestHypIndex 
             << " gain: " << maxHypGain << endl;
      } else {
        (*p_baselinePrimeStats)[s] = p_baselineHyp;
      }
      
    } // source sent loop
    
    
    // evaluate baselinePrime with best hypotheses and compare with baseline
    Evaluate(p_baselinePrimeStats, baselinePrimeScores);
    baselineGain = baselinePrimeScores[m_N_PRECISION] - baselineScores[m_N_PRECISION];
    
    cout << "Gain over baseline: " << baselineGain << endl;    
    
    if(baselineGain > 0) 
    {
      cout << "changing baseline to baselinePrime" << endl;
      for(int s = 0; s <= p_baselinePrimeStats->MaxIndex(); s++) {
        (*p_baselineStats)[s] = (*p_baselinePrimeStats)[s];      
      }
      
      baselineScores = baselinePrimeScores;
    } else {
      cout << "not changing baseline to baselinePrime" << endl;
    }
    
    ReportScores(baselineScores);
    cout << "reached end of search iteration" << endl;
    
  } // search loop
  
  Scores = baselineScores;
  
  delete p_baselineStats;
  delete p_baselinePrimeStats;
  
  return true;
}



//------------------------------------------------------------------
//      CEvaluator:: GetConfidenceInterval
//

std::pair< float, float >
CEvaluatorBLEU::GetConfidenceInterval()
{
  
  CArray< CArray< CEvaluatorNGramMatchStats* > >* p_hypMatchStats 
    = m_p_sharedData->GetHypMatchStats();
  assert( p_hypMatchStats );

  int sampleSpaceSize = 2000;
  std::vector< float > sampleSpace (sampleSpaceSize);
  
  CArray< CEvaluatorNGramMatchStats* >* p_randomBaseline 
    = new CArray< CEvaluatorNGramMatchStats* >;

  // Generate 2000 artificial baselines
  for (int r=0; r < sampleSpaceSize; r++)
  {
    for (int s = p_hypMatchStats->MinIndex(); s<= p_hypMatchStats->MaxIndex(); s++)
    {
      // seed rand with current time
      //srand(time(0));
      int randomSource = rand() % (p_hypMatchStats->MaxIndex() + 1);
      
      (*p_randomBaseline)[s] = (*p_hypMatchStats)[randomSource][0];
      
    } // source sent loop

    CNumericalArray< float > Scores;
    Evaluate(p_randomBaseline, Scores);
    
    sampleSpace[r] = Scores[m_N_PRECISION];
    
  } // sample space loop
  
  std::sort( sampleSpace.begin(), sampleSpace.end() );

  double topPercentile    = .975;
  int    topPercentileIndex = (int)sampleSpaceSize * topPercentile;       // round off
  float  topPercentileScore = sampleSpace[topPercentileIndex];
  
  double bottomPercentile = .025;
  int    bottomPercentileIndex = (int)sampleSpaceSize * bottomPercentile; // round off
  float  bottomPercentileScore = sampleSpace[bottomPercentileIndex];

  
  delete p_randomBaseline;  
  
  return std::pair< float, float>(bottomPercentileScore, topPercentileScore);
  
}



//------------------------------------------------------------------
//      CEvaluatorBLEU:: ReportScores
//

void
CEvaluatorBLEU::ReportScores( CNumericalArray< float > & Scores )
{
  cout << "BLEU Metric Scores" << endl;
  
  for(int n=1; n<=m_maxNGram; n++){
    cout<<n<<"-gram Bleu score: " << Scores[n] << endl;
  }
  
}

//------------------------------------------------------------------
//      CEvaluatorBLEU:: PrintOracleBestSentences
//
void
CEvaluatorBLEU::PrintOracleBestSentences( CArray< CEvaluatorNGramMatchStats* >* pOracleBestStats )
{

  int numHyps = pOracleBestStats->Size();
  for (int s = pOracleBestStats->MinIndex(); s < numHyps; s++ )
  {
    cout << "ORACLE_BEST " << pOracleBestStats->GetAt(s)->hypSentenceIndex << " ";
    pOracleBestStats->GetAt(s)->pHypSentence->Display( m_pVocabulary ); 
    cout << endl;
  }

}

//------------------------------------------------------------------
//      CEvaluatorBLEU:: ComputeNGramStats
//
//      creates reference ngram max hash if needed
//      creates memoryypMatchStats
//
//      BLEU needs match stats only
//      called after LoadReferences
//             after LoadNBestList or LoadAsciiText
//      or, Evaluate sees empty p_hypMatchStats
//
//      creates m_pp_maxRefHash if it needs to be created
//          this can also be reused by other metrics
//

bool
CEvaluatorBLEU::ComputeNGramMatchStats(CArray< CArray< CSentence* > >* p_NBestTranslations)
{
  
  // retrieve shared data
  CArray< CArray< CSentence* > >* p_refSentences = m_p_sharedData->GetReferences();

  assert(p_refSentences);
  
  CArray< CArray< CEvaluatorNGramMatchStats* > >* p_hypMatchStats 
    = m_p_sharedData->GetHypMatchStats();
  
  bool toBuildMaxRefHash = m_p_sharedData->ToBuildMaxRefHash();
  bool toComputeMatchCounts = m_p_sharedData->ToComputeMatchCounts();
  bool toComputeBleuLengths = m_p_sharedData->ToComputeBleuLengths();
  
  int sourceSentCount = p_NBestTranslations->Size();
  
  NGram_hash_map** pp_maxRefHash;      // store max reference ngram counts once
  NGram_hash_map* p_maxRefHash;        // per source sentence
  
  if (toBuildMaxRefHash) {
    pp_maxRefHash = new NGram_hash_map* [sourceSentCount];    
    m_p_sharedData->SetMaxRefHash(pp_maxRefHash);
  } else if (toComputeMatchCounts) {
    pp_maxRefHash = m_p_sharedData->GetMaxRefHash();
  }
  
  if ( m_verboseOutput ) {
    cout << "source sent count: " << sourceSentCount << endl;
  }
  
  for(int s=0; s < sourceSentCount; s++)
  {
    int hypSentCount = (*p_NBestTranslations)[s].Size();
    
    if ( m_verboseOutput ) {
      cout << "s: " << s << " hyp sent count: " << hypSentCount << endl;
    }
    
    // loop over every ref sentence for this system
    int refSentCount = (int)((*p_refSentences)[s].Size());   // p_refSentences member of Evaluator

    if (toBuildMaxRefHash) 
    {
      //-- hash ngram->{max freq out of all references for hypothesis}
      p_maxRefHash = new NGram_hash_map();

      for(int r=0; r < refSentCount; r++)
      {
        CSentence* p_currRefSentence = (*p_refSentences)[s][r];


        // populate m_pp_maxRefHash[h] if not previously done
        NGram_hash_map* p_currRefHash = GetNGramsFromSentence(p_currRefSentence); // new
        p_maxRefHash = MaxValueHash(p_maxRefHash, p_currRefHash);
      }
    }
    
    //-- save or retrieve hash of max # of ngram occurrences in all ref sentences
    if (toBuildMaxRefHash) {
      pp_maxRefHash[s] = p_maxRefHash;
    } else if (toComputeMatchCounts) {
      p_maxRefHash = pp_maxRefHash[s];
    }
    
    
    for(int h=0; h < hypSentCount; h++)
    {
      
      CSentence* p_currHypSentence = (*p_NBestTranslations)[s][h];
      
      //-- hash ngram->{freq in sentence}
      NGram_hash_map* p_hypHash;
      if (toComputeMatchCounts) {
        p_hypHash = GetNGramsFromSentence(p_currHypSentence);
      }
      
      //-- Select reference sentence closest in length to hypothesis
      //   of two equidistant ref sentences, choose shortest (IBM implementation)
      int hypSentLength         = p_currHypSentence->Size();
      int smallestDiff          = 10000000;
      int smallestDiffRefLength = 10000000;
      
      //-- Select reference sentence closest in length to hypothesis    
      for(int r=0; r < refSentCount; r++)
      {
         //-- Track closest length of ref sentences
        int currRefLength = (*p_refSentences)[s][r]->Size();
        int diff = std::abs( currRefLength - hypSentLength );
        
        if( diff <= smallestDiff )
        {
          if( diff == smallestDiff ) {
            if( currRefLength < smallestDiffRefLength ) {
              smallestDiffRefLength   = currRefLength;
            }
          } else {
            smallestDiff            = diff;
            smallestDiffRefLength   = currRefLength;
          }
        }
        
      } // end ref loop
      
      
      //-- count n-gram co-occurrences and save stats for BLEU evaluation
      CEvaluatorNGramMatchStats* p_currHypStats;
      
      if(toComputeMatchCounts) {
        p_currHypStats = new CEvaluatorNGramMatchStats();

        // save hyp sentence in ngram match object to display oracle best later
        p_currHypStats->pHypSentence = p_currHypSentence;
        p_currHypStats->hypSentenceIndex = h;
        (*p_hypMatchStats)[s][h] = p_currHypStats;
      } else if (toComputeBleuLengths) {
        // p_currHypStats already computed partially
        p_currHypStats = (*p_hypMatchStats)[s][h];
      } else {
        // shouldn't be in this function!
      }
      
      p_currHypStats->hypSentLength = hypSentLength;
      p_currHypStats->closestRefSentLength = smallestDiffRefLength;
      
      if( toComputeMatchCounts )
      {
        for( NGram_hash_map::iterator hypHash_iter = p_hypHash->begin(); 
             hypHash_iter != p_hypHash->end(); 
             hypHash_iter++) 
        {
          NGram_hash_map::iterator maxRefHash_iter = p_maxRefHash->find(hypHash_iter->first);
          
          if(maxRefHash_iter != p_hypHash->end())
          {
            int n = (hypHash_iter->first)->getSize();
            
            p_currHypStats->matchingNGrams[n] 
              += std::min( hypHash_iter->second, maxRefHash_iter->second );
          }
          
          delete hypHash_iter->first;
        }
        
        p_hypHash->clear();
        delete p_hypHash;
      }
    } // end hyp sent loop
  } // end source sent loop

  m_p_sharedData->SetToComputeBleuLengths(false);
  m_p_sharedData->SetToComputeMatchCounts(false);
  
  Trace( 2, "Ngram match stats computed on nBestTranslations\n");

}

bool
CEvaluatorBLEU::ComputeNGramMatchStats(CText* pTranslations)
{
}  

//------------------------------------------------------------------
//      CEvaluatorBLEU:: MaxValueHash
//
//      Builds hash_map of {reference ngram}->{max frequency}

NGram_hash_map* 
CEvaluatorBLEU::MaxValueHash(NGram_hash_map* p_maxRefHash, 
                            NGram_hash_map* p_refHash)
{
  for(NGram_hash_map::iterator refHash_iter = p_refHash->begin();
      refHash_iter != p_refHash->end(); refHash_iter++)
  {
    NGram_hash_map::iterator maxRefHash_iter 
                        = p_maxRefHash->find(refHash_iter->first);
    
    if(maxRefHash_iter != p_maxRefHash->end())
    {
      // ngram found in max
      maxRefHash_iter->second = std::max( refHash_iter->second,
                                     maxRefHash_iter->second) ;    
      // no use for curr instance of ngram
      delete refHash_iter->first;

    } else {
      //save in max
      p_maxRefHash->insert( std::pair< CEvaluatorNGram*, int> 
                            (refHash_iter->first, refHash_iter->second) ); 
    }
    
  }
  
  p_refHash->clear();
  delete p_refHash;
  
  return p_maxRefHash;
}
