//
// BLEU Evaluation Metric
//
// Anthony D'Auria <apd@andrew.cmu.edu>
// 3/19/06
//
// (Comments in Doxygen style)
//

#ifndef __EvaluatorBLEU__
#define __EvaluatorBLEU__

#include "Evaluator.hh"


//====================================================================
//      CumulativeScores
//!     Stores aggregate values used by the BLEU metric
struct CumulativeScores
{
  int* numNGramsInHyp;                 //!< [max ngram size+1]
  int* matchingNGrams;                 //!< [max ngram size+1]
  int  closestRefSentLength;
  int  hypSentLength;

  CumulativeScores(int maxNGram)
  {
    numNGramsInHyp = new int[maxNGram+1]; // index 1..9, so size=10
    matchingNGrams = new int[maxNGram+1]; // index 1..9

    for(int n=1; n <= maxNGram; n++) {
      numNGramsInHyp[n] = 0;
      matchingNGrams[n] = 0;  
    }
    closestRefSentLength = 0;
    hypSentLength = 0;
  }
  
  ~CumulativeScores() 
  {
    if(numNGramsInHyp != NULL){
      delete [] numNGramsInHyp;
    }
    if(matchingNGrams != NULL) {
      delete [] matchingNGrams;
    }
  }
  
};


//====================================================================
//      CEvaluatorBLEU
//

//! An implementation of the IBM Bleu translation evaluation metric
/*! A subclass of CEvaluator, from which it inherits the following:
    - variables: m_pVocabulary, m_maxNGram (= 9)
    - methods: LoadReferences( const char *pFileName, int Format )
               GetNGramsFromSentence( CSentence* p_sentence )
*/


class CEvaluatorBLEU : public CEvaluator
{
public:

/*! Creates an evaluator that can compute the BLEU evaluation metric on 1-best
    and n-best lists, with one or more references translations for each source
    sentence. Constructor sets a pointer to the CEvaluatorSharedData class and 
    a member variable m_N_PRECISION that is the n-gram precision at which the
    final BLEU score is computed.

    \param pVocabulary maps any word to a word id
 
    \param pSharedData contains pointers to containers that can be used by
      different evaluators. This class computes ngram matches between hypotheses and 
      references and stores them in CEvaluatorSharedData so that other instances of
      CEvaluatorBLEU or its subclass CEvaluatorNIST do not perform redundant
      calculations.
 
    \author Anthony D'Auria (apd@andrew.cmu.edu)
 */
  CEvaluatorBLEU( CVocabulary *pVocabulary, CEvaluatorSharedData *pSharedData);
  
  //! Empty destructor
  ~CEvaluatorBLEU();
  
  //! Computes BLEU score on translations
  /*! Computes BLEU metric for baseline only (best translation for each source sentence),
      passed as a CText. Converts CText to n-best list format to reuse 
      ComputeNGramMatchStats.
   
      \param pTranslations is a CText, an array of CSentenceS
      \param Scores contains ngram precision scores. Scores[1] for 1-gram, etc.
        Bleu result stored in Scores[0]
   
      \return bool, marking success or failure
   
      \author Anthony D'Auria (apd@andrew.cmu.edu)
  */
  bool Evaluate( CText *pTranslations, CNumericalArray< float > & Scores);

  /*! Loads translations from file and computes BLEU metric.
   
      \param pFileWithTranslations defines file name with hypotheses.
      \param Scores contains scores at different ngram precisions
        - Scores[m_SCORE_KEY] contains final BLEU score
   
      \note Currently not needed and not implemented. Should be in abstract class since 
        it can run CEvaluator::LoadAsciiText() and a virtual Evaluate().
  */
  bool Evaluate( const char *pFileWithTranslations, 
                 CNumericalArray< float > & Scores );
  
  /*! Runs BLEU evaluation metric on translations in the CArray< CString > 
      container.
    
      \note Not implemented.
   
      \todo Currently not needed. Should be in abstract class. Can run virtual
        Evaluate function.
  */
  bool Evaluate( CArray< CString > &Translations, 
                 CNumericalArray< float > & Scores );
  
  /*! Runs BLEU evaluation metric on one source sentence.
   
      \note Not used by any methods. Deprecated and commented out.
   
      \todo Update to use CEvaluatorNGramMatchStats, if the need for this fn arises.
  */
  bool Evaluate( int sourceSentenceIndex, CSentence* pHypSentence, 
                 CNumericalArray< float >& Scores);
  
  /*! Runs BLEU evaluation metric on translations in the 
      CArray< CNumericalArray< CWordIndex> > container.
    
      \note Not implemented.
    
      \todo Currently not needed. Should be in abstract class. Can run virtual
        Evaluate function.
  */
  bool Evaluate( const CArray< CNumericalArray< CWordIndex> > *pTanslations, 
                 CNumericalArray< float > & Scores );

  bool Evaluate( CSentence* pHypSentence, CArray< CSentence* > &References,
                CNumericalArray< float >& Scores );
  
  //! Computes BLEU OracleScore for nbest lists
  /*! Loads nbest translations from file and calls an OracleScore method. Oracle 
      scores is the Bleu score on best hypotheses per source sentence in nbest lists.
    
      \note Does not update CEvaluatorSharedData with nbest list.
 
      \param pFileWithNBestTranslations defines file name
      \param Scores contains Oracle scores at various ngram precisions
 
      \return bool value for success
 
      \author Anthony D'Auria (apd@andrew.cmu.edu)
  */
  bool GetOracleScore( const char *pFileWithNBestTranslations, 
                       CNumericalArray< float > & Scores );
  
  bool GetOracleScore( CArray< CString > *pNBestTranslations, 
                       CNumericalArray< float > & Scores );
  bool GetOracleScore( CText *pNBestTranslations, 
                       CNumericalArray< float > & Scores );
  bool GetOracleScore( CArray< CNumericalArray< CWordIndex > > *pNBestTranslations, 
                       CNumericalArray< float > & Scores );
  bool GetOracleScore( CArray< CArray < CSentence* > > *pNBestTranslations,
                       CNumericalArray< float > & Scores );
  
  //! Iteratively approximate the best possible set of translations
  /*! Construct a new baseline by selecting the best hypothesis for every source 
      sentence with respect to the current baseline. Modify baseline after all 
      best hypotheses are known, repeat until gain is <= 0.
  
      \note This will not result in the best set of translations because with 
        every selection, the global length penalty changes. Previous hypothesis 
        choices may no longer be best.
 
      \param p_NBestTranslations is a 2D CArray of hypotheses, [source][hypothesis]
 
      \return bool signifying success
 
      \author Implementation: Anthony D'Auria (apd@andrew.cmu.edu)
        Concept: Joy (joy@cs.cmu.edu)
  */
  bool GetOracleScoreJoy1( CArray< CArray < CSentence* > > *pNBestTranslations,
                           CNumericalArray< float > & Scores );                     

  //! Iteratively approximate the best possible set of translations
  /*! Construct a new baseline by selecting the best hypothesis for a current sentence, 
      modify the baseline, then pick the best hypothesis for the next source sentence
      with respect to the modified baseline. Hypothetically reaches approximation of 
      optimal baseline more quickly than Joy1 method.
   
      \note This will not result in the best set of translations because with 
        every selection, the global length penalty changes. Previous hypothesis 
        choices may no longer be best.
   
      \param p_NBestTranslations is a 2D CArray of hypotheses, [source][hypothesis]
   
      \return bool signifying success
   
      \author Implementation: Anthony D'Auria (apd@andrew.cmu.edu)
  */
  bool GetOracleScoreAnthony1( CArray< CArray < CSentence* > > *pNBestTranslations,
                               CNumericalArray< float > & Scores );                     
  //! Print scores to screen
  void ReportScores( CNumericalArray< float > & Scores );
  
  //! Print oracle-best sentences
  void PrintOracleBestSentences( CArray< CEvaluatorNGramMatchStats* >* pOracleBestStats );
    
  //! Builds hash_map of {reference ngram}->{max frequency}
  //! within set of reference sentences per hypothesis sentence
  NGram_hash_map* MaxValueHash(NGram_hash_map* p_refMaxHash, 
                               NGram_hash_map* p_refHash);
  
  //! builds [source][hyp] arr of CEvaluatorNGramStats that contain values 
  //! needed for Bleu scoring of each source/hyp pair, and saves to m_p_hypMatchStats
  virtual bool ComputeNGramMatchStats( CArray< CArray< CSentence* > >* pNBestTranslations );
  //! deprecated, to remove
  virtual bool ComputeNGramMatchStats( CText* pTranslations );
  
  //! Get 95% confidence interval for baseline score
  /*! With boostrapping, create many (2000) artificial baselines, score them, sort 
      Bleu scores. Scores between 2.75 percentile and 97.5 percentile of all scores 
      contain 95% of all scores for artificial baselines. Return this 95% 
      confidence interval.
   
      \return a pair< float, float > (lower bound, upper bound)
   
      \author Anthony D'Auria (apd@andrew.cmu.edu)
   
      \todo use srand()
   */
  std::pair< float, float > GetConfidenceInterval();

protected:
    //! Computes BLEU score for translations
    /*! Generate the BLEU score for a 1-best set of translations only 
    (for each source sentence, 1 hypothesis translation).
    
    \note Assumes ngram match stats are computed by the function ComputeNGramMatchStats
    before this function is run
    
    \param p_hypMatchStats are precomputed ngram statistics used by this metric
    
    \param Scores stores ngram precision scores. Scores[1] for 1-gram, etc. 
    Scores[0] for final Bleu score
    
    \return bool to signify success or failure
    
    \author Anthony D'Auria (apd@andrew.cmu.edu)
    */
  virtual bool Evaluate( CArray< CEvaluatorNGramMatchStats* >* p_hypMatchStats,
                         CNumericalArray< float > &Scores );

  int m_N_PRECISION;  //!< ngram at which to measure Bleu score

};





#endif
