/*
 *  EvaluatorCER.cc
 *  STTK_Lattice_Evaluator_SVN
 *
 *  Created by Anthony D'Auria on 10/26/06.
 *  Character Error Rate
 *
 */

#include "EvaluatorCER.hh"

CEvaluatorCER::CEvaluatorCER( CVocabulary *pVocabulary, 
                              CEvaluatorSharedData *pSharedData )
: CEvaluatorWER( pVocabulary, pSharedData )
{
  m_pAlteredRefSentences = NULL;
  m_pOrigRefSentences = NULL;

  
}

CEvaluatorCER::~CEvaluatorCER()
{
  ResetAlteredReferences();
  delete m_pAlteredRefSentences;
}


//------------------------------------------------------------------
//      CEvaluatorCER:: TokenizeReferences
//

void
CEvaluatorCER::TokenizeReferences( CArray< CArray< CSentence* > >* pRefSentences )
{
 
  // store a local copy of the pointer to original references in case they were changed
  if ( m_pOrigRefSentences != pRefSentences ) 
  {

    m_pOrigRefSentences = pRefSentences;
    
    ResetAlteredReferences();
    
    m_pAlteredRefSentences = new CArray< CArray< CSentence* > >;
    
    for ( int s = pRefSentences->MinIndex(); s <= pRefSentences->MaxIndex(); s++ )
    {
      CArray< CSentence* > Sentences = (*pRefSentences)[ s ];
      
      for ( int r = Sentences.MinIndex(); r <= Sentences.MaxIndex(); r++ )
      {
        
        (*m_pAlteredRefSentences)[ s ][ r ] = TokenizeSentence( Sentences[ r ] );

      } // ref sent loop
    } // source sent loop
  }
  
}


//------------------------------------------------------------------
//      CEvaluatorCER:: TokenizeSentence
//

CSentence*
CEvaluatorCER::TokenizeSentence( CSentence* pSentence )
{
  // this method is about twice as fast as s/(\S)/$1 /g with TextFilterBoostRegex
  
  std::string origStr = (std::string)pSentence->AsString( m_pVocabulary );

  std::string tokenizedStr = "";
  
  for ( int c = 0; origStr[ c ] != '\0'; c++ ) { 
    if ( origStr[ c ] != ' ' ) {
      tokenizedStr += std::string(1, origStr[ c ]) + " ";
    }
  }

  CSentence* pTokenizedSentence = new CSentence();
  pTokenizedSentence->ReadFromBuffer( tokenizedStr.c_str(), m_pVocabulary);
  
  return pTokenizedSentence;

}



//------------------------------------------------------------------
//      CEvaluatorCER:: ResetAlteredReferences
//

void
CEvaluatorCER::ResetAlteredReferences()
{
  if ( m_pAlteredRefSentences != NULL )
  {
    for ( int s = m_pAlteredRefSentences->MinIndex(); 
              s <=m_pAlteredRefSentences->MaxIndex(); s++) 
    {
      for ( int r = (*m_pAlteredRefSentences)[s].MinIndex(); 
                r <=(*m_pAlteredRefSentences)[s].MaxIndex(); r++) 
      {
        delete (*m_pAlteredRefSentences)[ s ][ r ];
        (*m_pAlteredRefSentences)[ s ][ r ] = NULL;
      }
      (*m_pAlteredRefSentences)[s].Clear();
    }
    m_pAlteredRefSentences->Clear();
  }
}


/**
 * @brief Score one sentence with the given set of references
 *
 *
 */
bool 
CEvaluatorCER::Evaluate( CSentence* pHypSentence, CArray< CSentence* > &References,
                        CNumericalArray< float >& Scores )
{
  ;
}



//------------------------------------------------------------------
//      CEvaluatorCER:: Evaluate (one sentence)
//

bool CEvaluatorCER::Evaluate( int sourceSentIndex, CSentence* pHypSentence, 
                              CNumericalArray< float >& Scores)
{
  
  // modify pHypSentence
  CSentence* pTokenizedHypSentence = TokenizeSentence( pHypSentence );
  
  // convert references
  CArray< CArray< CSentence* > >* pRefSentences = m_p_sharedData->GetReferences();
  
  if ( m_pOrigRefSentences != pRefSentences || m_pAlteredRefSentences == NULL )
  {
    TokenizeReferences( pRefSentences );
  } 
  

  // Scores[] used to track WER and to count Levenshtein path steps
  Scores[WER] = 0;    
  Scores[INSERTION] = 0;
  Scores[DELETION] = 0;   
  Scores[SUBSTITUTION] = 0;
  Scores[MATCH] = 0;      
  Scores[HYP_SENTENCE_COUNT] = 1;
  Scores[REF_WORD_COUNT] = 0;         
  Scores[LEAST_DISTANCE] = 0;
  
  int hypWordCount = pTokenizedHypSentence->Size();
  
  int refSentCount = (int)((*m_pAlteredRefSentences)[sourceSentIndex].MaxIndex()+1);
  
  // track lowest wer score for each ref sentence per hyp
  float leastWerScore = 200.0;
  float leastDistance = 0;
  int leastWerScoreRefSize = 0;
  CArray< CArray< int > > leastWerBacktrack;
  
  for(int r = 0; r < refSentCount; r++)  // loop over ref sentences
  {
    
    CSentence* currRefSentence = (*m_pAlteredRefSentences)[ sourceSentIndex ][r];
    
    int refWordCount = currRefSentence->Size();
    
    // Levenshtein distance array
    float lDistance[hypWordCount+1][refWordCount+1];
    
    // backtrack Levenshtein path
    CArray< CArray< int > > backtrack;
    
    // Levenshtein init first row and first column
    for( int i = 0; i < hypWordCount+1; i++)
      lDistance[i][0] = i;
    
    for( int j = 0; j < refWordCount+1; j++)
      lDistance[0][j] = j;
    
    // Levenshtein fill rest of array with 0s
    for( int i = 1; i < hypWordCount+1; i++)
      for( int j = 1; j < refWordCount+1; j++)
        lDistance[i][j] = 0;
    
    
    // Build Levenshtein matrix
    
    CWordIndex refWord;
    CWordIndex hypWord;
    
    for(int i=1; i <= hypWordCount; i++) 
    {
      hypWord = pTokenizedHypSentence->GetAt(i);  
      
      for(int j=1; j <= refWordCount; j++) 
      {
        refWord = currRefSentence->GetAt(j);  
        
        if( hypWord == refWord )
        {
          lDistance[i][j] = lDistance[i-1][j-1];
          backtrack[i][j] = MATCH;
        } else {
          lDistance[i][j] = lDistance[i-1][j-1] + 1;
          backtrack[i][j] = SUBSTITUTION;
        }
        
        if( lDistance[i][j] > lDistance[i-1][j] + 1 ){
          lDistance[i][j] = lDistance[i-1][j] + 1;
          backtrack[i][j] = INSERTION;
        }
        
        if( lDistance[i][j] > lDistance[i][j-1] + 1 ) {
          lDistance[i][j] = lDistance[i][j-1] +1;
          backtrack[i][j] = DELETION;
        }
        
      }
    }
    
    if(leastWerScore > lDistance[hypWordCount][refWordCount]/refWordCount)
    {
      leastWerScore = lDistance[hypWordCount][refWordCount]/refWordCount;
      leastWerScoreRefSize = refWordCount;
      leastDistance = lDistance[hypWordCount][refWordCount];
      leastWerBacktrack = backtrack;
    }
    
  } // end of ref sentence loop
  
  // trace backwards Levenshtein path of fewest modifications
  int b_x = leastWerScoreRefSize, b_y = hypWordCount;
  
  while(b_x > 0 || b_y > 0)
  {
    int step = leastWerBacktrack[b_y][b_x];
    
    // check if hit edge of matrix
    if( b_x == 0 ) {    
      
      leastWerBacktrack[b_y][1] = -1;
      step = INSERTION;
      
    } else if( b_y == 0 ) {
      
      leastWerBacktrack[1][b_x] = -1;
      step = DELETION;
      
    } else {
      
      leastWerBacktrack[b_y][b_x] = -1;
      
    }
    
    Scores[step]++;     
    
    if( step == MATCH ) {
      // move backward diagonally
      b_x--;
      b_y--;
    }
    if( step == INSERTION ) {
      // move up
      b_y--;
    }
    if( step == SUBSTITUTION ) {
      // move backward diagonally
      b_x--;
      b_y--;
    }
    if( step == DELETION ) {
      // move to the left
      b_x--;
    }
  }
  
  // ref sentences with lowest WER score are counted in Scores[]
  Scores[HYP_WORD_COUNT] = hypWordCount;
  Scores[REF_WORD_COUNT] = leastWerScoreRefSize; 
  Scores[LEAST_DISTANCE] = leastDistance;
  
  Scores[WER] = (float)leastDistance / (float)leastWerScoreRefSize;
  
  // clean up
  delete pTokenizedHypSentence;
  
  return true;
  
}


void
CEvaluatorCER::ReportScores( CNumericalArray< float > & Scores )
{
	// Scores should contain WER(0), insertions (1), deletions (2), substitutions (3), matches (4),length translations, length references
	
	cout << "CER=" << Scores[WER] << " INS=" << Scores[INSERTION] 
  << " DEL=" << Scores[DELETION] << " SUB=" << Scores[SUBSTITUTION] 
  << " MATCH=" << Scores[MATCH] << " HypSent=" 
  << Scores[HYP_SENTENCE_COUNT] << " RefWords=" 
  << Scores[REF_WORD_COUNT] << " HypWords="
  << Scores[HYP_WORD_COUNT] << endl;
	
}

