//
// NIST Evaluation Metric
// Anthony D'Auria <apd@andrew.cmu.edu>
// 3/19/06
//
//
//

#ifndef __EvaluatorNIST__
#define __EvaluatorNIST__

//#include <Evaluator.hh>
#include <EvaluatorBLEU.hh>

//====================================================================
//      CumulativeNISTScores
//
//!      Stores aggregate values used in the NIST metric

struct CumulativeNISTScores
{
  int*    numNGramsInHyp;               //!< [ngram size]
  int*    matchingNGrams;               //!< [ngram size]
  double* hypInfoGain;                  //!< [hypothesisIndex]

  double  avgRefSentLength;
  int     hypSentLength;
  
  CumulativeNISTScores(int maxNGram)
  {
    numNGramsInHyp = new int[maxNGram+1];
    matchingNGrams = new int[maxNGram+1];
    hypInfoGain = new double[maxNGram+1];
    
    for(int n=0; n<= maxNGram; n++){
      numNGramsInHyp[n] = 0;
      matchingNGrams[n] = 0;
      hypInfoGain   [n] = 0;
    }
    
    avgRefSentLength = 0;
    hypSentLength    = 0;
  }
  
  ~CumulativeNISTScores() 
  {
    if(numNGramsInHyp != NULL){ 
      delete [] numNGramsInHyp;
    }
    if(matchingNGrams != NULL) {
      delete [] matchingNGrams;
    }
    if(hypInfoGain != NULL) {
      delete [] hypInfoGain;
    }
  }
};

//====================================================================
//      CEvaluatorNIST

//! An implementation of the NIST translation metric
/*! A subclass of CEvaluator, from which it inherits:
    - variables: m_pVocabulary, m_maxNGram (= 9)
    - methods: LoadReferences( const char *pFileName, int Format )
               GetNGramsFromSentence( CSentence* p_sentence )
*/
class CEvaluatorNIST : public CEvaluatorBLEU
{
public:

  /*! Creates an evaluator that can compute the NIST evaluation metric on 1-best
  and n-best lists, with one or more references translations for each source
  sentence. Constructor sets a pointer to the CEvaluatorSharedData class and 
  a member variable m_N_PRECISION that is the n-gram precision at which the
  final NIST score is computed (5-gram).
  
  \param pVocabulary maps any word to a word id
  
  \param pSharedData contains pointers to containers that can be used by
  different evaluators. This class computes ngram matches between hypotheses and 
  references and stores them in CEvaluatorSharedData so that other instances of
  CEvaluatorBLEU or its subclass CEvaluatorNIST do not perform redundant
  calculations.
  
  \author Anthony D'Auria (apd@andrew.cmu.edu)
  */
  CEvaluatorNIST( CVocabulary *pVocabulary, CEvaluatorSharedData *pSharedData );
  
  //! Empty destructor
  ~CEvaluatorNIST();

  // overload methods to calculate the score or scores for the translation
  // scores to be returned: cumulative NIST scores 1-maxNGram, Length Penalty, 
  // # of hypothesis senteces, sum of avg # of words in reference translations
  bool Evaluate( const char *pFileWithTranslations, 
                 CNumericalArray< float > & Scores );
  bool Evaluate( CArray< CString > &Translations, 
                 CNumericalArray< float > & Scores );
  bool Evaluate( int sourceSentenceIndex, CSentence* pHypSentence, 
                 CNumericalArray< float >& Scores);
  //! Computes NIST score on translations
  /*! Computes NIST metric for baseline only (best translation for each source sentence),
    passed as a CText. Converts CText to n-best list format to reuse 
    ComputeNGramMatchStats.
    
    \param pTranslations is a CText, an array of CSentenceS
    \param Scores contains ngram precision scores. Scores[1] for 1-gram, etc.
    Bleu result stored in Scores[0]
    
    \return bool, marking success or failure
    
    \author Anthony D'Auria (apd@andrew.cmu.edu)
    */
  bool Evaluate( CText *pTranslations, CNumericalArray< float > & Scores );
  bool Evaluate( const CArray< CNumericalArray< CWordIndex > > *pTanslations, 
                 CNumericalArray< float > & Scores );
  bool Evaluate( CSentence* pHypSentence, CArray< CSentence* > &References,
                CNumericalArray< float >& Scores );
  // calculate the oracle score for a nbest list; Format: n m word sequence
  // oracle scores means: take those translations in nbest lists which optimize the scores
  // no matter if read from file, or getting the translations in arrays
  bool GetOracleScore( const char *pFileWithNBestTranslations, 
                       CNumericalArray< float > & Scores );
  bool GetOracleScore( CArray< CString > *pNBestTranslations, 
                       CNumericalArray< float > & Scores );
  bool GetOracleScore( CText *pNBestTranslations, 
                       CNumericalArray< float > & Scores );
  bool GetOracleScore( CArray< CNumericalArray< CWordIndex > > *pNBestTranslations,
                       CNumericalArray< float > & Scores );
  bool GetOracleScore( CArray< CArray< CSentence* > >* p_NBestTranslations,
                       CNumericalArray< float >& Scores );

  void ReportScores( CNumericalArray< float > & Scores );

  //-- Called on every reference sentence and updates two hashtables: 1. a hash 
  //   of ngrams to total ref ngram counts, 2. has of ngrams to max ref ngram counts
  //-- The two hashes could be merged into one, but now their keys are pointers
  //   to the same memory. The loss: hash_map overhead.  
  void BuildCountsHashes(int hypSent, 
                         NGram_hash_map** pp_maxRefHash,
                         NGram_hash_map* p_refHash,
                         NGram_hash_map* p_currRefHash,
                         bool buildTotalRefCountHash,
                         bool buildMaxRefHash);

  //-- Build a ngram->double hash_map that contains info gain for every ngram 
  //   that occurs in all reference sentences
  NGramToDouble_hash_map* BuildInfoGainHash(NGram_hash_map* p_refHash, 
                                            int totalRefWords);

  //-- NIST formula for length penalty. Argument ratio is 
  //   # of reference words / sum of avg # of reference words per reference set
  double ComputeLengthPenalty(double ratio);
  
  // builds [source][hyp] arr of CEvaluatorNGramStats, saves to m_p_hypMatchStats
  virtual bool ComputeNGramMatchStats( CText* pTranslations );
  virtual bool ComputeNGramMatchStats( CArray< CArray< CSentence* > >* pNBestTranslations);
  
//  virtual std::pair<float, float> GetConfidenceInterval();
  // get stats and ngram hashes
  //CArray< CArray< CEvaluatorNGramMatchStats > >* GetNGramMatchInfoStats();
  //NGram_hash_map**        GetReferenceNGramMaxCountHash();
  //NGramToDouble_hash_map* GetGainHash();

  // set stats and ngram hashes
  //void SetReferenceNGramMaxCountHash(NGram_hash_map** pp_maxRefHash);
  //void SetGainHash(NGramToDouble_hash_map* p_gainHash);
  //void SetNGramMatchInfoStats(CArray< CArray< CEvaluatorNGramMatchStats > >*
 //                             p_hypMatchInfoStats);
  


protected:

  virtual bool Evaluate( CArray< CEvaluatorNGramMatchStats* >* p_hypMatchStats,
                         CNumericalArray< float > &Scores );
    
    
  //NGram_hash_map** m_pp_maxRefHash;       //[hypSent]->{ref ngram*}->{max count}
  
  //NGramToDouble_hash_map* m_p_gainHash;   //{ref ngram*}->{info gain}

  //CArray< CArray< CEvaluatorNGramMatchStats > >* m_p_hypMatchInfoStats;

};





#endif
