/*
 *  EvaluatorNistBLEU.cc
 *  STTK_Anthony2
 *
 *  Created by Anton D'Auria on 7/23/06.
 *  Copyright 2006 __MyCompanyName__. All rights reserved.
 *
 */


//====================================================================
//      CEvaluatorNistBLEU::
//        

#include "EvaluatorNistBLEU.hh"

CEvaluatorNistBLEU::CEvaluatorNistBLEU( CVocabulary *pVocabulary,
                                        CEvaluatorSharedData *pSharedData )
: CEvaluatorBLEU( pVocabulary, pSharedData )
{
  
}

CEvaluatorNistBLEU::~CEvaluatorNistBLEU()
{
}


//------------------------------------------------------------------
//      CEvaluatorBLEU:: ComputeNGramStats
//
//      creates reference ngram max hash if needed
//      creates memoryypMatchStats
//
//      BLEU needs match stats only
//      called after LoadReferences
//             after LoadNBestList or LoadAsciiText
//      or, Evaluate sees empty p_hypMatchStats
//
//      creates m_pp_maxRefHash if it needs to be created
//          this can also be reused by other metrics
//

bool
CEvaluatorNistBLEU::ComputeNGramMatchStats(CArray< CArray< CSentence* > >* p_NBestTranslations)
{
  
  // retrieve shared data
  CArray< CArray< CSentence* > >* p_refSentences = m_p_sharedData->GetReferences();

  assert(p_refSentences);
  
  CArray< CArray< CEvaluatorNGramMatchStats* > >* p_hypMatchStats 
    = m_p_sharedData->GetHypMatchStats();
  
  bool toBuildMaxRefHash = m_p_sharedData->ToBuildMaxRefHash();
  bool toComputeMatchCounts = m_p_sharedData->ToComputeMatchCounts();
  bool toComputeNistBleuLengths = m_p_sharedData->ToComputeNistBleuLengths();
  
  int sourceSentCount = p_NBestTranslations->Size();
  
  NGram_hash_map** pp_maxRefHash;      // store max reference ngram counts once
  NGram_hash_map* p_maxRefHash;        // per source sentence
  
  if (toBuildMaxRefHash) {
    pp_maxRefHash = new NGram_hash_map* [sourceSentCount];    
    m_p_sharedData->SetMaxRefHash(pp_maxRefHash);
  } else if (toComputeMatchCounts) {
    pp_maxRefHash = m_p_sharedData->GetMaxRefHash();
  }
  
  cout << "source sent count: " << sourceSentCount << endl;
  
  for(int s=0; s < sourceSentCount; s++)
  {
    int hypSentCount = (*p_NBestTranslations)[s].Size();
    cout << "s: " << s << " hyp sent count: " << hypSentCount << endl;
    
    // loop over every ref sentence for this system
    int refSentCount = (int)((*p_refSentences)[s].Size());   // p_refSentences member of Evaluator

    if (toBuildMaxRefHash) 
    {
      //-- hash ngram->{max freq out of all references for hypothesis}
      p_maxRefHash = new NGram_hash_map();

      for(int r=0; r < refSentCount; r++)
      {
        CSentence* p_currRefSentence = (*p_refSentences)[s][r];


        // populate m_pp_maxRefHash[h] if not previously done
        NGram_hash_map* p_currRefHash = GetNGramsFromSentence(p_currRefSentence); // new
        p_maxRefHash = MaxValueHash(p_maxRefHash, p_currRefHash);
      }
    }
    
    //-- save or retrieve hash of max # of ngram occurrences in all ref sentences
    if (toBuildMaxRefHash) {
      pp_maxRefHash[s] = p_maxRefHash;
    } else if (toComputeMatchCounts) {
      p_maxRefHash = pp_maxRefHash[s];
    }
    
    
    for(int h=0; h < hypSentCount; h++)
    {
      
      CSentence* p_currHypSentence = (*p_NBestTranslations)[s][h];
      
      //-- hash ngram->{freq in sentence}
      NGram_hash_map* p_hypHash;
      if (toComputeMatchCounts) {
        p_hypHash = GetNGramsFromSentence(p_currHypSentence);
      }
      
      //-- Select reference sentence closest in length to hypothesis
      //   of two equidistant ref sentences, choose shortest (IBM implementation)
      int hypSentLength         = p_currHypSentence->Size();
      
      int smallestRefLength = 10000000;
      
      //-- Select reference sentence closest in length to hypothesis    
      for(int r=0; r < refSentCount; r++)
      {
         //-- Track closest length of ref sentences
        int currRefLength = (*p_refSentences)[s][r]->Size();
     
        if(currRefLength < smallestRefLength) {
          smallestRefLength = currRefLength; // DIFFERENCE between IBM-BLEU and NIST-BLEU
        }
        
      } // end ref loop
      
      
      //-- count n-gram co-occurrences and save stats for BLEU evaluation
      CEvaluatorNGramMatchStats* p_currHypStats;
      
      if(toComputeMatchCounts) {
        p_currHypStats = new CEvaluatorNGramMatchStats();
        (*p_hypMatchStats)[s][h] = p_currHypStats;
      } else if (toComputeNistBleuLengths) {
        // p_currHypStats already computed partially
        p_currHypStats = (*p_hypMatchStats)[s][h];
      } else {
        // shouldn't be in this function!
      }
      
      p_currHypStats->hypSentLength = hypSentLength;
      // DIFFERENCE between IBM-BLEU and NIST-BLEU
      p_currHypStats->closestRefSentLength = smallestRefLength; 
      
      if( toComputeMatchCounts )
      {
        for( NGram_hash_map::iterator hypHash_iter = p_hypHash->begin(); 
             hypHash_iter != p_hypHash->end(); 
             hypHash_iter++) 
        {
          NGram_hash_map::iterator maxRefHash_iter = p_maxRefHash->find(hypHash_iter->first);
          
          if(maxRefHash_iter != p_hypHash->end())
          {
            int n = (hypHash_iter->first)->getSize();
            
            p_currHypStats->matchingNGrams[n] 
              += std::min( hypHash_iter->second, maxRefHash_iter->second );
          }
          
          delete hypHash_iter->first;
        }
        
        p_hypHash->clear();
        delete p_hypHash;
      }
    } // end hyp sent loop
  } // end source sent loop

  m_p_sharedData->SetToComputeNistBleuLengths(false);
  m_p_sharedData->SetToComputeMatchCounts(false);
  
  Trace( 2, "Ngram match stats computed on nBestTranslations\n");

}

//------------------------------------------------------------------
//      CEvaluatorBLEU:: ReportScores
//

void
CEvaluatorNistBLEU::ReportScores( CNumericalArray< float > & Scores )
{
  cout << "NIST-BLEU Metric Scores" << endl;
  
  for(int n=1; n<=m_maxNGram; n++){
    cout<<n<<"-gram Nist-Bleu score: " << Scores[n] << endl;
  }
  
}












