#include <EvaluatorWER.hh>

CEvaluatorWER::CEvaluatorWER( CVocabulary *pVocabulary, 
                              CEvaluatorSharedData *pSharedData )
  : CEvaluator( pVocabulary, pSharedData ) 
{	
  // Scores[] constants
	WER = 0;
	MATCH = 1;
	SUBSTITUTION = 2;
	INSERTION = 3;
	DELETION = 4;
	HYP_SENTENCE_COUNT = 5;
	REF_WORD_COUNT = 6;
	HYP_WORD_COUNT = 7;
  
  LEAST_DISTANCE = 8;
}


CEvaluatorWER::~CEvaluatorWER()
{
}



bool
CEvaluatorWER::Evaluate( const char *pFileWithTranslations, 
												 CNumericalArray< float > & Scores )
{
}



bool 
CEvaluatorWER::Evaluate( CArray< CString > &Translations, 
												 CNumericalArray< float > & Scores )
{
}


bool CEvaluatorWER::Evaluate( int sourceSentIndex, CSentence* pHypSentence, 
                              CNumericalArray< float >& Scores)
{
  CArray< CArray< CSentence* > >* p_refSentences = m_p_sharedData->GetReferences();
  
 
  
  // Scores[] used to track WER and to count Levenshtein path steps
  Scores[WER] = 0;    
  Scores[INSERTION] = 0;
  Scores[DELETION] = 0;   
  Scores[SUBSTITUTION] = 0;
  Scores[MATCH] = 0;      
  Scores[HYP_SENTENCE_COUNT] = 1;
  Scores[REF_WORD_COUNT] = 0;         
  Scores[LEAST_DISTANCE] = 0;
  
  int hypWordCount = pHypSentence->Size();

  int refSentCount = (int)((*p_refSentences)[sourceSentIndex].Size());;
  
  // track lowest wer score for each ref sentence per hyp
  float leastWerScore = 200.0;
  float leastDistance = 0;
  int leastWerScoreRefSize = 0;
  CArray< CArray< int > > leastWerBacktrack;
   
  for(int r = 0; r < refSentCount; r++)  // loop over ref sentences
  {
      
    CSentence* currRefSentence = (*p_refSentences)[ sourceSentIndex ][r];
    
    int refWordCount = currRefSentence->Size();
      
    // Levenshtein distance array
    float lDistance[hypWordCount+1][refWordCount+1];
      
    // backtrack Levenshtein path
    CArray< CArray< int > > backtrack;
      
    // Levenshtein init first row and first column
    for( int i = 0; i < hypWordCount+1; i++)
      lDistance[i][0] = i;
    
    for( int j = 0; j < refWordCount+1; j++)
      lDistance[0][j] = j;
      
    // Levenshtein fill rest of array with 0s
    for( int i = 1; i < hypWordCount+1; i++)
      for( int j = 1; j < refWordCount+1; j++)
        lDistance[i][j] = 0;
      
      
    // Build Levenshtein matrix
      
    CWordIndex refWord;
    CWordIndex hypWord;
      
    for(int i=1; i <= hypWordCount; i++) 
    {
      hypWord = pHypSentence->GetAt(i);  
  
      for(int j=1; j <= refWordCount; j++) 
      {
        refWord = currRefSentence->GetAt(j);  
  
        if( hypWord == refWord )
        {
          lDistance[i][j] = lDistance[i-1][j-1];
          backtrack[i][j] = MATCH;
        } else {
          lDistance[i][j] = lDistance[i-1][j-1] + 1;
          backtrack[i][j] = SUBSTITUTION;
        }
          
        if( lDistance[i][j] > lDistance[i-1][j] + 1 ){
          lDistance[i][j] = lDistance[i-1][j] + 1;
          backtrack[i][j] = INSERTION;
        }
          
        if( lDistance[i][j] > lDistance[i][j-1] + 1 ) {
          lDistance[i][j] = lDistance[i][j-1] +1;
          backtrack[i][j] = DELETION;
        }
          
      }
    }
      
    if(leastWerScore > lDistance[hypWordCount][refWordCount]/refWordCount)
    {
      leastWerScore = lDistance[hypWordCount][refWordCount]/refWordCount;
      leastWerScoreRefSize = refWordCount;
      leastDistance = lDistance[hypWordCount][refWordCount];
      leastWerBacktrack = backtrack;
    }
      
  } // end of ref sentence loop
  
  // trace backwards Levenshtein path of fewest modifications
  int b_x = leastWerScoreRefSize, b_y = hypWordCount;
  
  while(b_x > 0 || b_y > 0)
  {
    int step = leastWerBacktrack[b_y][b_x];
      
    // check if hit edge of matrix
    if( b_x == 0 ) {    
      
      leastWerBacktrack[b_y][1] = -1;
      step = INSERTION;
      
    } else if( b_y == 0 ) {
      
      leastWerBacktrack[1][b_x] = -1;
      step = DELETION;
      
    } else {
      
      leastWerBacktrack[b_y][b_x] = -1;
      
    }
      
    Scores[step]++;     
            
    if( step == MATCH ) {
      // move backward diagonally
      b_x--;
      b_y--;
    }
    if( step == INSERTION ) {
      // move up
      b_y--;
    }
    if( step == SUBSTITUTION ) {
      // move backward diagonally
      b_x--;
      b_y--;
    }
    if( step == DELETION ) {
      // move to the left
      b_x--;
    }
  }
    
  // ref sentences with lowest WER score are counted in Scores[]
  Scores[HYP_WORD_COUNT] = hypWordCount;
  Scores[REF_WORD_COUNT] = leastWerScoreRefSize; 
  Scores[LEAST_DISTANCE] = leastDistance;
  
  Scores[WER] = (float)leastDistance / (float)leastWerScoreRefSize;
  
  return true;

}

// under construction
bool
CEvaluatorWER::Evaluate( CText* pTranslations,
                         CNumericalArray< float > & Scores )
{
  // make a call to Evaluate( int sourceSentIndex, pHypSentence, Scores)
  
  // since Evaluate is virtual, subclasses will get called for the real thing
  
  CArray< CArray< CSentence* > >* p_refSentences = m_p_sharedData->GetReferences();
	
	CSentence *currHypSentence, *currRefSentence;
		
	int hypSentCount =  pTranslations->NumberOfSentences();
	int refSentCount;
	
	// Scores[] used to track WER and to count Levenshtein path steps
	Scores[WER] = 0;		
  
	Scores[MATCH] = 0;			
	Scores[INSERTION] = 0;
	Scores[DELETION] = 0;		
	Scores[SUBSTITUTION] = 0;
  
  Scores[HYP_WORD_COUNT] = 0;		
	Scores[HYP_SENTENCE_COUNT] = hypSentCount;

	Scores[REF_WORD_COUNT] = 0;		
  Scores[LEAST_DISTANCE] = 0;
	
	// count size of ref and hyp sentences, # of ref sentences per hyp
	int refWordCount, hypWordCount;

	// track lowest wer score for each ref sentence per hyp
	float leastWerScore = 0;
  float leastDistance = 0;
	int leastWerScoreRefSize = 0;
	CArray< CArray< int > > leastWerBacktrack;
	
	for(int h=0; h < hypSentCount; h++)
	{
  
    currHypSentence = pTranslations->GetSentence(h);
    
    CNumericalArray< float > SentenceScores;
    
    Evaluate( h, currHypSentence, SentenceScores );
    
    // count edits
    Scores[MATCH]          += SentenceScores[MATCH];
    Scores[INSERTION]      += SentenceScores[INSERTION];
    Scores[DELETION]       += SentenceScores[DELETION];
    Scores[SUBSTITUTION]   += SentenceScores[SUBSTITUTION];

    Scores[HYP_WORD_COUNT] += SentenceScores[HYP_WORD_COUNT];

    // wer vars
		Scores[LEAST_DISTANCE] += SentenceScores[LEAST_DISTANCE];
    Scores[REF_WORD_COUNT] += SentenceScores[REF_WORD_COUNT];

    
  } // of of hyp sentence loop
	
	// Average out scores
  Scores[ WER ] = Scores[LEAST_DISTANCE] / Scores[REF_WORD_COUNT];
  
}

/**
 * @brief Score one sentence with the given set of references
 *
 *
 */
bool 
CEvaluatorWER::Evaluate( CSentence* pHypSentence, CArray< CSentence* > &References,
                        CNumericalArray< float >& Scores )
{
  ;
}

/*

bool
CEvaluatorWER::Evaluate( CText *pTranslations, 
												 CNumericalArray< float > & Scores )
{
		
  CArray< CArray< CSentence* > >* p_refSentences = m_p_sharedData->GetReferences();
	
	CSentence *currHypSentence, *currRefSentence;
		
	int hypSentCount =  pTranslations->NumberOfSentences();
	int refSentCount;
	
	// Scores[] used to track WER and to count Levenshtein path steps
	Scores[WER] = 0;		
	Scores[INSERTION] = 0;
	Scores[DELETION] = 0;		
	Scores[SUBSTITUTION] = 0;
	Scores[MATCH] = 0;			
	Scores[HYP_SENTENCE_COUNT] = hypSentCount;
	Scores[REF_WORD_COUNT] = 0;		
  Scores[LEAST_DISTANCE] = 0;
	
	// count size of ref and hyp sentences, # of ref sentences per hyp
	int refWordCount, hypWordCount;

	// track lowest wer score for each ref sentence per hyp
	float leastWerScore = 0;
  float leastDistance = 0;
	int leastWerScoreRefSize = 0;
	CArray< CArray< int > > leastWerBacktrack;
	
	for(int h=0; h < hypSentCount; h++)
	{
		currHypSentence = pTranslations->GetSentence(h);
		hypWordCount = currHypSentence->Size();
		refSentCount = (int)((*p_refSentences)[h].Size());
		
		leastWerScore = 2.0;									 // WER <= 1

		for(int r = 0; r < refSentCount; r++)  // loop over ref sentences
		{
			
			currRefSentence = (*p_refSentences)[h][r]; //inherited from Evaluator
			
			refWordCount = currRefSentence->Size();
			
			// Levenshtein distance array
			float lDistance[hypWordCount+1][refWordCount+1];
			
			// backtrack Levenshtein path
			CArray< CArray< int > > backtrack;
			
			//float substCost;
			
			// Levenshtein init first row and first column
			for( int i = 0; i < hypWordCount+1; i++)
				lDistance[i][0] = i;
      
			for( int j = 0; j < refWordCount+1; j++)
				lDistance[0][j] = j;
			
			// Levenshtein fill rest of array with 0s
			for( int i = 1; i < hypWordCount+1; i++)
				for( int j = 1; j < refWordCount+1; j++)
					lDistance[i][j] = 0;
			
			
			// Build Levenshtein matrix
      
			CWordIndex refWord;
			CWordIndex hypWord;
			
			for(int i=1; i <= hypWordCount; i++) 
			{
				hypWord = currHypSentence->GetAt(i);	

				for(int j=1; j <= refWordCount; j++) 
				{
					refWord = currRefSentence->GetAt(j);	

					if( hypWord == refWord )
					{
						lDistance[i][j] = lDistance[i-1][j-1];
						backtrack[i][j] = MATCH;
					} else {
						lDistance[i][j] = lDistance[i-1][j-1] + 1;
						backtrack[i][j] = SUBSTITUTION;
					}
					
					if( lDistance[i][j] > lDistance[i-1][j] + 1 ){
						lDistance[i][j] = lDistance[i-1][j] + 1;
						backtrack[i][j] = INSERTION;
					}
					
					if( lDistance[i][j] > lDistance[i][j-1] + 1 ) {
						lDistance[i][j] = lDistance[i][j-1] +1;
						backtrack[i][j] = DELETION;
					}
					
				}
			}
			
			if(leastWerScore > lDistance[hypWordCount][refWordCount]/refWordCount)
      {
				leastWerScore = lDistance[hypWordCount][refWordCount]/refWordCount;
				leastWerScoreRefSize = refWordCount;
				leastDistance = lDistance[hypWordCount][refWordCount];
				leastWerBacktrack = backtrack;
			}
      
		} // end of ref sentence loop

		// trace backwards Levenshtein path of fewest modifications
		int b_x = leastWerScoreRefSize, b_y = hypWordCount;

		while(b_x > 0 || b_y > 0)
		{
			int step = leastWerBacktrack[b_y][b_x];
			
      // check if hit edge of matrix
			if(b_x == 0) {                   
				leastWerBacktrack[b_y][1] = -1;
				step = INSERTION;
			} else if(b_y == 0) {
				leastWerBacktrack[1][b_x] = -1;
				step = DELETION;
			} else {
				leastWerBacktrack[b_y][b_x] = -1;
			}
			
			Scores[step]++;			
						
			if(step == MATCH) {
				// move backward diagonally
				b_x--;
				b_y--;
			}
			if(step == INSERTION) {
				// move up
				b_y--;
			}
			if(step == SUBSTITUTION) {
				// move backward diagonally
				b_x--;
				b_y--;
			}
			if(step == DELETION) {
				// move to the left
				b_x--;
			}
		}
		
		// ref sentences with lowest WER score are counted in Scores[]
		Scores[REF_WORD_COUNT] += leastWerScoreRefSize; 
		Scores[LEAST_DISTANCE] += leastDistance;
		
		Scores[HYP_WORD_COUNT] += hypWordCount;

	} // of of hyp sentence loop
	
	// Average out scores
	Scores[ WER ] = Scores[LEAST_DISTANCE] / Scores[REF_WORD_COUNT];

	return true;
}

*/

bool
CEvaluatorWER::Evaluate( const CArray< CNumericalArray< CWordIndex > > 
												 *pTanslations, 
												 CNumericalArray< float > & Scores )
{
}

//------------------------------------------------------------------
//      CEvaluatorWER:: GetConfidenceInterval
//        

std::pair<float, float>
CEvaluatorWER::GetConfidenceInterval()
{
  
}



  // calculate the oracle score for a nbest list; Format: n m word sequence
  // oracle scores means: take those translations in nbest lists which optimize 
	// the scores no matter if read from file, or getting the translations in arrays
bool
CEvaluatorWER::GetOracleScore( const char *pFileWithNBestTranslations, 
															 CNumericalArray< float > & Scores )
{
}



bool
CEvaluatorWER::GetOracleScore( CArray< CString > *pNBestTranslations, 
															 CNumericalArray< float > & Scores )
{
}



bool
CEvaluatorWER::GetOracleScore( CText *pNBestTranslations, 
															 CNumericalArray< float > & Scores )
{
}


bool
CEvaluatorWER::GetOracleScore( CArray< CNumericalArray< CWordIndex > > *pNBestTranslations, 
															 CNumericalArray< float > & Scores )
{
}


bool
CEvaluatorWER::GetOracleScore( CArray< CArray< CSentence* > >* p_NBestTranslations,
                                CNumericalArray< float >& Scores )
{
  // p_NBestTranslations is just like p_refSentences
  
  // run eval for every source/hyp
  // call special version of Evaluate
  // create a local version of ScoredSentence
  // 
}

void
CEvaluatorWER::ReportScores( CNumericalArray< float > & Scores )
{
	// Scores should contain WER(0), insertions (1), deletions (2), substitutions (3), matches (4),length translations, length references
	cout << "WER Metric Scores -------------------------------------------------" << endl;
  
	cout <<  "WER="      << Scores[WER]      
       << " INS="      << Scores[INSERTION] 
			 << " DEL="      << Scores[DELETION] 
       << " SUB="      << Scores[SUBSTITUTION] 
			 << " MATCH="    << Scores[MATCH]  
       << " HypSent="  << Scores[HYP_SENTENCE_COUNT]  
       << " RefWords=" << Scores[REF_WORD_COUNT]
       << " HypWords=" << Scores[HYP_WORD_COUNT]      << endl;
	
}
