#ifndef INCLUDE_BLOOM_FILTER_H
#define INCLUDE_BLOOM_FILTER_H

#include <iostream>
#include <vector>
#include <sstream>
#include <LMInterface.hh>
#include <Vocab.h>
#include <math.h>

#include "GeneralHashFunctions.h"


const unsigned int char_size = 0x08;    // 8 bits in 1 char(unsigned)

const unsigned char bit_mask[8] = {
                                   0x01, //00000001
                                   0x02, //00000010
                                   0x04, //00000100
                                   0x08, //00001000
                                   0x10, //00010000
                                   0x20, //00100000
                                   0x40, //01000000
                                   0x80  //10000000
                                 };

class bloom_filter
{

	public:

  bloom_filter(unsigned long tbl_size, unsigned int hash_functions)
	{
		//cout << tbl_size << endl;
		table_size = tbl_size;
		tbl_char=table_size*char_size;
		number_hash_functions=hash_functions;
		hash_table = new unsigned char[table_size];
		register_default_hash_functions(hash_functions);
		cout << "BF created with size " << table_size << " byte and " << hash_functions << " hash functions" << endl;
	}

   ~bloom_filter()
   {
      delete[] hash_table;
   }

   void register_default_hash_functions(unsigned int hash_functions)
   {

      if (hash_functions>0) {hash_function.push_back(RSHash  );}
      if (hash_functions>1) {hash_function.push_back(JSHash  );}
      if (hash_functions>2) {hash_function.push_back(PJWHash );}
      if (hash_functions>3) {hash_function.push_back(BKDRHash);}
      if (hash_functions>4) {hash_function.push_back(SDBMHash);}
      if (hash_functions>5) {hash_function.push_back(DJBHash );}
      if (hash_functions>6) {hash_function.push_back(DEKHash );}
      if (hash_functions>7) {hash_function.push_back(APHash  );}
   }
  
	void insert(vector<LMWordIndex> key)
	{
		//insert 1 in bitarray for each hash function
		for(unsigned int i = 0; i < number_hash_functions; i++)
		{
			unsigned int hash = hash_function[i](key) % (tbl_char);
			hash_table[hash / char_size] |= bit_mask[hash % char_size];
		}
		//cout << "inserted" << endl;
	}
   
	void insertNgram(vector<LMWordIndex> key, int frequency)
   {
		//frequency=1+roundl(log((double)frequency)*5);
		int f=1;
		if (frequency>2) {f=2;}
		if (frequency>5) {f=3;}
		if (frequency>10) {f=4;}
		if (frequency>25) {f=5;}
		if (frequency>50) {f=6;}
		if (frequency>100) {f=7;}	
		if (frequency>200) {f=8;}
		if (frequency>500) {f=9;}
		if (frequency>1000) {f=10;}
	
		frequency=f;
		//insert key,1...key,frequency
		for (int i=1; i<=frequency; i++) 
		{
			key.push_back(i);
			insert(key);
			key.pop_back();
		}
		/*
		key.push_back(1);
		insert(key);
		key.pop_back();
		key.push_back(frequency);
		insert(key);
		*/
	}
    
	double wordProb(VocabIndex word, const VocabIndex *context)
	{
		
		//cout << "Checking: " << word << endl;
		unsigned int clen = Vocab::length(context);
		//cout << "CLEN:" << clen << endl;
		for (int i=0; i<clen;i++){
		//cout << context[i] << endl;
		}

	        cont.erase(cont.begin(), cont.end());
		cont.insert(cont.begin(),word);
		
		double probs;
	
		long freq1, freq2;
		
		double probsum;
		int sumcounter=1;
		//cout << "f: " << getfrequency(cont) << endl;

		probsum=((double)getfrequency(cont))/1000000000;
		//cout << probsum << endl;
		if (probsum==0) { return 1/1000000000; } //unknown word
		
		
		for (int i=clen-1; i>=0; i--) 
		{
			cont.insert(cont.begin(), context[i]);
			
			//get frequency at this gammity
				freq1=getfrequency(cont);
				cont.pop_back();
				freq2=getfrequency(cont);
				cont.push_back(word);
				
				//cout << "Context length: " << i << " f1: " << freq1 << " f2: " << freq2 << endl;
			   probs=((double)freq1)/((double)freq2);
			//cout << probsum << endl;   
			if (freq2==0) {return probsum/sumcounter;} else { probsum+=probs; sumcounter++;}
		}
		//cout << "RETURRN" << endl;
		return probs;
		
	}
	
	bool contains(vector<LMWordIndex> key)
   {
		//test for each hashfunction
      for(std::size_t i = 0; i < number_hash_functions; i++)
      {
         unsigned int hash = hash_function[i](key) % (tbl_char);
         unsigned int bit  = hash % char_size;
         if ((hash_table[hash / char_size] & bit_mask[bit]) != bit_mask[bit])
         {
            return false;
         }
      }
      return true;
   }
	
	int getfrequency(vector<LMWordIndex> key) 
	{
		//Find frequency for Ngram 
		//Check frequencies 1 ...
		//cout << "check frequency" << endl;
		int frequency=1;
		key.push_back(1);
		while (contains(key)) 
		{
			//cout << "check frequency" << frequency << endl;
			
			frequency++;
			key.pop_back();
			key.push_back(frequency);
		}
		key.pop_back();
		frequency--;
		if (frequency==0) {return 0;}

		//Double Checks
		//Check included Ngrams
		/*for (unsigned int g=1;g<key.size(); g++)
		{
			for (unsigned int i=0; i<=key.size()-g; i++)
			{
				vector<LMWordIndex> test;
				for (unsigned int k=0; k<g; k++)
				{
					test.push_back(key.at(i+k));
					//std::cout << key.at(i+k) << std::endl;
				}
				test.push_back(frequency);
				//std::cout << test << std::endl;
				if (!contains(test)) {frequency=-2;}
			}
		}*/
		
		//frequency=round(exp((frequency-1)/5));
		
		switch (frequency)
		{
		case 1: 
			return 1;
			break;
		case 2: 
			return 3;
			break;
		case 3: 
			return 7;
			break;
		case 4: 
			return 18;
			break;
		case 5:
			return 40;
			break;
		case 6: 
			return 75;
			break;
		case 7: 
			return 150;
			break;
		case 8: 
			return 350;
			break;
		case 9: 
			return 800;
			break;
		case 10: 
			return 2000;
			break;
		}
		return 0;
	}
   
	private:

	std::vector < HashFunction > hash_function;
	int 									number_hash_functions;
	unsigned char*        		hash_table;
	unsigned long               table_size;	
	unsigned long  				tbl_char;
	vector<LMWordIndex> 	cont;

};

#endif
