#include <config.h>

#ifndef __SRILM__
#define __SRILM__

#include <iostream>
#include <fstream>

#include <Vocabulary.hh>

#include <LMInterface.hh>
#include <option.h>
#include <File.h>
#include <Ngram.h>
#include <LM.h>
#include <Vocab.h>
#include <Prob.h>


using namespace std;

class SriLM;
class SriLMSentenceState;

#define NN 3


// sv 071025:  extension to have ngrams longer 4
// define NGram and compare operator for Ngrams
/*
typedef struct {
  vector< LMWordIndex > Ngram;
} NgramType;
*/


typedef vector< LMWordIndex > NGramType;

struct LessThenForNGrams
{
  bool operator()( NGramType a, NGramType b ) const 
  {
	  int lenA = a.size();
	  int lenB = b.size();

    // if length not equal, consider shorter as smaller
		if ( lenA != lenB ) {
			return lenA < lenB;
		}

		// two n-grams are of the same length		
    // search for unqual wordindex
		for ( int i = 0; i < lenA; i++ ) {
			if ( a[i] != b[i] ) {
				return a[i] < b[i];
			}
		}

		//a==b
		return false;
	}
};


class SriLMSentenceState : public CLMSentenceState
{
public:
  SriLMSentenceState( const SriLM* theLM );
  SriLMSentenceState();
  ~SriLMSentenceState();

  //! generate an lmstate for start of sentence <s>
  virtual CLMState BeginOfSentenceState();
  virtual CLMState BeginOfSentenceState2();  // for testing of long ngrams

  /** returns the negative difference of the param state and
	  the state that results after the merge */
  virtual double LogProb( const CLMState LMState, const LMWordIndex WordIndex );

  //! Get the logarithmic probability of a single word, also return the new state.
  virtual double LogProb( CLMState LMState, const LMWordIndex WordIndex, CLMState &NewLMState );


  /** caclulate the probability for a sequence of words and return two LM states
	  one for the left side, one for the right side; start of sentence (<s>) is implicitily added */
  virtual double LogProb( vector< string > Words, CLMState& LMStateLeft, CLMState& LMStateRight );

  /** Get the logarithmic probability of extending lmstate by words
	  which led to the creation of lmstate2. Also return the new LM state
	  in lmstate2 */
  virtual double LogProbMerge( CLMState &LMState_Left_Left, CLMState LMState_Left_Right,
                               CLMState LMState_Right_Left, CLMState &LMState_Right_Right );

  //! return log prob for adding sentence end </s>
  virtual double LogProbEnd( CLMState LMState );

  //! generate new lmstate from old lmstate and word
  virtual CLMState AppendWord( CLMState LMState, const LMWordIndex WordIndex );

  //! delete the LM state
  virtual void DeleteState( CLMState LMState );


private:
  SriLM* lm;
  VocabIndex *wids;

  // sv 071025:  extension to have ngrams longer 4
  typedef vector< LMWordIndex > NGramType;

  //! vector to store all seen Ngrams
	vector< NGramType > m_AllNGrams;

  //! vector to store all seen Ngrams
	vector< NGramType > m_AllNGramsForDump;

  //! map to map from Ngram to LMState, as we need to make sure that same LMState means same history
	map< NGramType, int, LessThenForNGrams > m_NGramToLMState;


  //! number of LM calls for on sentence
  long m_NumberOfCalls;

};




class SriLM : public CLM
{
  friend class SriLMSentenceState;

public:

  //! create the LM from a parameter file
  SriLM( char* paramFile, unsigned order = defaultNgramOrder, int UnkIsWord = 0 );

  //! create the LM with a given vocabulary to filter the n-grams during loading
  SriLM( char* File, int Order, const CVocabulary &Vocabulary, bool RestrictVocabulary = 1, bool UnkIsWord = 1 );

  //! destructor
  virtual ~SriLM();


  //! return the order ( = history length + 1 ) of the LM
  unsigned GetOrder() { return ngramLM->setorder(); }



  //! create a new sentence state
  virtual SriLMSentenceState* NewSentence() const;

  //! delete the sentence state
  virtual void DeleteSentence(CLMSentenceState* st) const;



  //! return the size of the vocabulary
  virtual LMWordIndex VocabularySize() const;

  // pure virtual functions that must be implemented, but are currently
  // not used anywhere in the system

  //! return the vocabulary as a vector of strings
  virtual const vector< string > & Vocabulary() const;

  //! compare a vocabulary to the vocabulary given in the parameter file
  virtual int Compare( char ** Vocab, char *Paramfile );

  //! get the LM internal word index for word
  virtual LMWordIndex GetVocabularyIndex( string word ) const;
  virtual LMWordIndex GetVocabularyIndex( string word );



  //ARV 07/11 used for future cost estimation
  virtual double IsolatedLogProb( vector<string> wordSequence );


private:
  //! the SRI LM
  Ngram *ngramLM;

  //! the SRI LM vocabulary
  Vocab *vocab;

  // sv 060604: this is never used except for returning it, which means it is always empty
  vector< string > m_vocabVector;

  //! load only n-grams which are covered by the vocabulary
  bool m_RestrictVocabulary;


};





inline const vector<string> & 
SriLM::Vocabulary() const
{
    return m_vocabVector;
}



inline int 
SriLM::Compare(char ** Vocab, char *Paramfile)
{
    return 0;
}



#endif
