#include <string.h>
#include <stdlib.h>
#include <config.h>
#include <fortify.h>
#include <TrieNode.hh>
#define NN 5
#define BETA -4.22

TrieNode* NistScore::m_gains;


NistScore::NistScore(NistScore& ns){
  m_nTotal = ns.m_nTotal;
  m_nist = new NistMap();
  for (NistMap::iterator iter = ns.m_nist->begin(); iter != ns.m_nist->end(); ++iter){
    (*m_nist)[(*iter).first] = (*iter).second;
  }
  m_nTotal = ns.m_nTotal;
}

NistScore::NistScore(int n){

  m_nist = new NistMap();
  for (int i=1; i<=n; i++){
    (*m_nist)[i] = pair<double,double>(0.0, 0.0);
  }
  m_nTotal = n;
}

NistScore::~NistScore(){
  // we don't delete the static gain map
  m_nist->clear();
  delete m_nist;
}

void NistScore::extend(NGram* nextGram){
  
  for (int i=1; i<=nextGram->size(); i++){
    pair<double, double> tmp = (*m_nist)[i];
    tmp.second = tmp.second + 1.0;
    (*m_nist)[i] = tmp;
  }
}

void NistScore::incrementScore(NGram* history){

  // get the score for this path from the gains trie
  double score = NistScore::m_gains->getScore(history->begin(), history->end(), 0.0);
  //cout<<"Score in increment"<<score<<endl;
  (*m_nist)[history->size()].first += score;

}

double NistScore::calculateScore(int Lref, int Lhyp){
  double sum = 0.0;
  for (int i=1; i<=m_nTotal; i++){
    pair<double, double> p = (*m_nist)[i];
    //cout<<i<<" "<<p.first<<":"<<p.second<<endl;
    if (p.second!=0.0){
      double ratio = p.first / p.second;
      sum += ratio;
    }
  }
  // calculate the length penalty
  double lrat = (double)(Lhyp)/(double)(Lref);
  if (lrat>1.0){
    lrat = 1.0;
  }
  double pen = exp(BETA*pow(log(lrat),2.0));
  cout<<"NistSum:"<<sum<<endl;
  double final = sum*pen;
  return final;
  
}

void NistScore::loadGains(map<string, LMWordIndex> GetVocabularyIndexMap, const char* fileName){
  
  //MGAINS = new TrieNode();
  NistScore::m_gains = new TrieNode();

  ifstream fp(fileName);
  char str[ 1000 ];
  
  if (!fp){
    cerr<<"Unable to load gains:"<<fileName<<endl;
    exit(-1);
  }

  for ( ; !fp.eof(); fp.getline( str, 1000 ) ) {
    //cout<<str<<endl;
    if (strlen(str)>0){
      
      vector<char*> tmpVec;
      //strcpy(starLine, str);
      char* c = strtok(str, " ");
      tmpVec.push_back(c);
      char* nextC;
      while ((nextC=strtok(NULL, " "))!=NULL){
	c = nextC;
	tmpVec.push_back(c);      
      }

      double score = atof(tmpVec.back()); 
      tmpVec.pop_back();
      if (score != 0){
	// now fill the proper gram array
	NGram* oneGram = new NGram();
	// tracks if this ngram has any out of vocab words
	// if it does, then it is useless for our purposes...
	bool valid = true;
	for (vector<char*>::iterator iter = tmpVec.begin(); iter!=tmpVec.end(); ++iter){
	  string tmpString = string(*iter);
	  if (GetVocabularyIndexMap.count(tmpString)==0){
	    valid = false;
	    break;
	  }
	  oneGram->push_back(GetVocabularyIndexMap[tmpString]);
	}
	tmpVec.clear();
	if (valid){
	  //cout<<TrieNode::printNGram(oneGram)<<"---"<<score<<endl;
	  NistScore::m_gains->addNGram(oneGram, score);
	}
	// delete it
	oneGram->clear();
	delete oneGram;
      }
    }
  }
  
  fp.close();
  //cout<<"GLOBAL:"<<NistScore::m_gains->toString()<<endl;
}

void NistScore::deleteGains(){
  delete NistScore::m_gains;
}


string NistScore::toString(){
  string str="";
  for (int i=1; i<=m_nTotal; i++){
    pair<double, double> tmp = (*m_nist)[i];
    cout<<tmp.first<<"/"<<tmp.second<<endl;
  }
  return str;
}


TrieNode::TrieNode(){
  m_numInstances = 0;
  m_branch = new Branches();
}

TrieNode::TrieNode(TrieNode& t){
  m_numInstances = t.m_numInstances;
  // copy over the branching tree
  m_branch = new Branches();
  for (Branches::iterator iter = t.m_branch->begin(); iter!= t.m_branch->end(); ++iter){
    LMWordIndex key = (*iter).first;
    TrieNode* value = (*iter).second;
    (*m_branch)[key] = new TrieNode(*value);
  }
}

// builds a complete tree by extracting all NGrams from
// the sentence represented in the input argument
TrieNode::TrieNode(NGram* sentence){

  //cout<<"INITIAL:"<<printNGram(sentence)<<endl;
  m_branch = new Branches();
  // adding all the 5 grams adds all the grams below as well
  // all grams will contain all ngrams upto size NN
  NGramList* allGrams = collectNGrams(sentence, sentence->size(), NN);
  // if the sentence is shorter than NN, we use that as the cap
  int cap = NN;
  if (sentence->size()<NN){
    cap = sentence->size();
  }

  for (NGramList::iterator iter = allGrams->begin(); iter!=allGrams->end(); ++iter){
    NGram* oneNGram = (*iter);
    // ngram initialized with scoring of -1, implies incremental counting
    //cout<<"One:"<<printNGram(oneNGram)<<endl;
    addNGram(oneNGram, 1.0);
    
  }
  // this is the root node, 0 instances
  m_numInstances = 0;
  for (NGramList::iterator iter=allGrams->begin(); iter!=allGrams->end(); ++iter){
    NGram* l = (*iter);
    l->clear();
    delete l;
  }
  allGrams->clear();
  delete allGrams;
  
}

TrieNode::~TrieNode(){
  for (Branches::iterator iter = m_branch->begin(); iter!= m_branch->end(); ++iter){
    TrieNode* childNode = (*iter).second;
    delete childNode;
  }
  delete m_branch;
}

// increments the count that refers the the number
// of instances have led to this node
void TrieNode::incrementCount(){
  m_numInstances++;
}  

// adds a multiple gram pathway into a tree
void TrieNode::addNGram(NGram* oneNGram, double score){
  
  if (oneNGram->size()==0){
    if (score!=-1){
      m_numInstances += score;
    }
    return;
  }else{

    LMWordIndex oneWord = oneNGram->front();
    //cout<<"Popped:"<<oneWord<<endl;
    TrieNode* nextNode=NULL;
    if (m_branch->count(oneWord)==0){
      //cout<<"Creating leaf:"<<oneWord<<endl;
      nextNode = new TrieNode();	
      (*m_branch)[oneWord] = nextNode;
    }else{
      //cout<<"Following leaf:"<<oneWord<<endl;
      nextNode = (*m_branch)[oneWord];
    }
    // increment the counter indicating that this 
    // path has been seen if score==-1
    // otherwise, just save this as the end score
    if (score==-1){
      nextNode->incrementCount();
    }
    oneNGram->erase(oneNGram->begin());
    //cout<<"Recursing"<<endl;
    nextNode->addNGram(oneNGram, score);
  }
}

// returns the score associated with a particular history path
// at the leaf of this path, the final score can be decremented if possible
// returns the score before the decrement
double TrieNode::getScore(NGram::iterator pathBegin, NGram::iterator pathEnd, double decrement){

  //cout<<"INSCORE:"<<this->toString()<<endl;
  if (pathBegin==pathEnd){
    double saved = m_numInstances;
    double leftOver = m_numInstances - decrement;
    //cout<<"Left:"<<leftOver;
    if (leftOver >= 0.0){
      //cout<<"Setting:"<<m_numInstances<<" to "<<leftOver<<endl;
      m_numInstances = leftOver;
    }
    return saved;
  }else{
    LMWordIndex word = (*pathBegin);
    if (m_branch->count(word)==0){
      return 0.0;
    }else{
      pathBegin++;
      TrieNode* nextNode = (*m_branch)[word];
      return nextNode->getScore(pathBegin, pathEnd, decrement);
    }
  }
}

// collects all n-grams in history, starting at the last word in the history
// this results in the n-grams arranged by those starting with the last word,
// then those starting with the second last etc...
// the last param specifies the max n-gram size
NGramList* TrieNode::collectNGrams(NGram* source, int depth, int n){
  
  NGramList* nList = new NGramList();
  int size = source->size();
  
  int cutoff = size - depth;
  for (int i=size-1; i>=cutoff; i--){
    int nlen = i;
    if ((n-1)<nlen){
      nlen = n-1;
    }
    for (int len=nlen; len>=0; len--){
      NGram* oneNGram = new NGram();
      //fill the ngram
      for (int j=len; j>=0; j--){
	//cout<<"IJ"<<(i+j)<<endl;
	LMWordIndex oneWord = (*source)[i-j];
	//cout<<oneWord<<" ";
	oneNGram->insert(oneNGram->end(), oneWord);

      }
      //cout<<endl;
      nList->insert(nList->end(), oneNGram);
    }
  }
  return nList;
}



// converts this trie to a string to be used for hashing
string TrieNode::toString(){
  
  if (m_branch->size()==0){
    return "#";
  }
  string finalString = "";
  for (Branches::iterator iter = m_branch->begin(); iter!=m_branch->end(); ++iter){
    LMWordIndex key = (*iter).first;
    TrieNode* value = (*iter).second;
    
    char str[1000];
    char dumb[1000];
    
    string tmpString = value->toString();
    strcpy(dumb, tmpString.c_str());

    sprintf(str, "%d:%1.1f(%s)", key, value->m_numInstances, dumb);
    //cout << key<<"+"<<value->m_numInstances<<":"<<value->toString()<<"-";
    finalString += str;//string(out);
  }
  return finalString;
  
}

TrieNode* TrieNode::scoreArc(LMWordIndex nextWord, NGram* history, NistScore* score){

  TrieNode* clone = new TrieNode(*this);
  //cout<<"Cloned:"<<clone->toString()<<endl;
  history->push_back(nextWord);
  if (history->size() > NN){
    // pop the front
    history->erase(history->begin());
  }

  // increases the number of hypothesis ngrams
  score->extend(history);
  
  NGramList* extendedGrams = collectNGrams(history, 1, NN);
  // see which ngrams could be score
  for (NGramList::iterator iter = extendedGrams->begin(); iter!=extendedGrams->end(); ++iter){
    NGram* oneCandidate = (*iter);    
    //cout<<"CandSize:"<<oneCandidate->size()<<endl;
    //cout<<"Evaluating:"<<printNGram(oneCandidate)<<endl;
    // get the score, and decrement the number of instances at this node!
    double value = clone->getScore(oneCandidate->begin(), oneCandidate->end(), 1.0);
    //cout<<"value:"<<value<<endl;
    if (value >= 1.0){
      //cout<<"Found opp"<<endl;
      // the decrement was made, add to the nist score
      score->incrementScore(oneCandidate);
    }else{
      // cout<<"No opps"<<endl;
    }
  }
  // delete the ngram list
  for (NGramList::iterator iter = extendedGrams->begin(); iter!=extendedGrams->end(); ++iter){
    NGram* l = (*iter);
    l->clear();
    delete l;
  }
  delete extendedGrams;

  return clone;
}

PathState::PathState(TrieNode* root, NGram* history, NistScore* score, int numDecoded){
  m_root = root; 
  m_history = history;
  m_score = score;
  m_numDecoded = numDecoded;
}
PathState::PathState(NGram* sentence){
  m_root = new TrieNode(sentence);
  m_history = new NGram();
  m_numDecoded = 0;
  m_score = new NistScore(NN);
}

PathState::~PathState(){
  delete m_root;
  m_history->clear();
  delete m_history;
  delete m_score;
}

string PathState::toString(){
  
  string treeState = m_root->toString();
  string historyString = TrieNode::printNGram(m_history);
  //cout<<"History:"<<historyString<<endl;
  char str[1000];
  char dumb[1000];  
  char hist[1000];  

  strcpy(dumb, treeState.c_str());
  strcpy(hist, historyString.c_str());
    
  sprintf(str, "%dT%sH%s", m_numDecoded, dumb, hist);
  return str;
}

PathState* PathState::scoreExtension(LMWordIndex nextWord){
  NistScore* clonedScore = new NistScore(*m_score);
  NGram* clonedHistory = new NGram(*m_history);
  TrieNode* clonedTrie = m_root->scoreArc(nextWord, clonedHistory, clonedScore);
  PathState* clonedState = new PathState(clonedTrie, clonedHistory, clonedScore, m_numDecoded+1);
  return clonedState;
}

double PathState::calculateScore(int Lref){
 return m_score->calculateScore(Lref, m_numDecoded);
}

/*int main(int argc, char* argv[]){

  NistScore::loadGains("gains.txt");
  
  NGram* ref = new NGram();
  ref->insert(ref->end(), 1);
  ref->insert(ref->end(), 2);
  ref->insert(ref->end(), 1);
  ref->insert(ref->end(), 2);

  //cout<<"TEST:"<<TrieNode::printNGram(ref)<<endl;

  PathState* p0 = new PathState(ref);
  cout<<"P0:"<<p0->toString()<<endl;
  PathState* p1 = p0->scoreExtension(1);
  cout<<"P1:"<<p1->toString()<<endl;
  PathState* p2 = p1->scoreExtension(2);
  cout<<"P2:"<<p2->toString()<<endl;
  PathState* p3 = p2->scoreExtension(1);
  cout<<"P3:"<<p3->toString()<<endl;
  PathState* p4 = p3->scoreExtension(2);
  cout<<"P4:"<<p4->toString()<<endl;

  //PathState* p5 = p4->scoreExtension(1);
  //cout<<"P5:"<<p5->toString()<<endl;
  //PathState* p6 = p5->scoreExtension(2);
  //cout<<"P6:"<<p6->toString()<<endl;
  
  TrieNode* mroot = new TrieNode();
  NGram* test1 = new NGram();
  test1->insert(test1->end(), 1);
  test1->insert(test1->end(), 2);
  test1->insert(test1->end(), 3);
  mroot->addNGram(test1, -1);
  test1->clear();
  test1->insert(test1->end(), 1);
  test1->insert(test1->end(), 2);
  test1->insert(test1->end(), 4);
  mroot->addNGram(test1, -1);


  string junk = mroot->toString();
  cout<<"j:"<<junk<<endl;
  
}

*/





