#ifndef LATTICE_H
#define LATTICE_H

// include autoconf header
#ifdef HAVE_CONFIG_H
  #include <config.h>
#endif


#include <iostream>
#include <fstream>
#include <stdio.h>

#include <map>
#include <vector>
using std::vector;

#include <String.hh>
#include <Array.hh>
//#include "SortedArray.hh"
#include <Text.hh>

using namespace std;

// ------------------------------
// Exports
// ------------------------------
class CLatticeNode;
class CLatticeEdge;
class CLattice;



// ------------------------------
// CLatticeNode
// Basisklasse f¸r die Knoten einer Lattice. 
// ------------------------------
class CLatticeNode
{
public:
  // Knotenkonstruktor 
  CLatticeNode( int Number );

  virtual ~CLatticeNode() {;}

  static long s_NumberOfObjects;

  // Selektor f¸r die Liste ausgehender Kanten 
  const CArray< CLatticeEdge * > & OutEdges();

  // Selektor f¸r die Liste eingehender Kanten 
  CArray< CLatticeEdge * > & InEdges();

  // Selektor f¸r die logische Knotennummer 
  int Number();

  // Vergleichsoperator "liegt logisch nach" 
  bool operator>( CLatticeNode *pNode );

  // Vergleichsoperator "liegt logisch vor" 
  bool operator<( CLatticeNode *pNode );

  // Einstelliger Ausgabeoperator f¸r Lattice-Knoten.
  //  @param OutStream Ausgabestrom 

  virtual std::ostream& operator<<( std::ostream& OutStream);

  void Display();


protected:
  // ausgehende Kante hinzuf¸gen 
  void AddOutEdge( CLatticeEdge *pOutEdge );

  // eingehende Kante hinzuf¸gen 
  void AddInEdge( CLatticeEdge *pInEdge );


  // remove outgoing edge
  void RemoveOutEdge( CLatticeEdge *pOutEdge );

  // remove ingoing edge
  void RemoveInEdge( CLatticeEdge *pInEdge );



  // Liste eingehender Kanten 
  CArray< CLatticeEdge * > m_InEdges;

  // Liste ausgehender Kanten 
  CArray< CLatticeEdge * > m_OutEdges;

  // Logische Nummer eines Knotens (>=0) 
  int m_Number;

  friend class CLattice;
};




// Basisklasse f¸r alle Kanten einer Lattice.
class CLatticeEdge
{
public:
  // Default-Konstruktor
  CLatticeEdge();


  //! Copy constructor
  CLatticeEdge( const CLatticeEdge &Edge );

  //Konstruktor ohne Prosodie
  CLatticeEdge( CLatticeNode *f, CLatticeNode *t );

  //Konstruktor ohne Prosodie
  CLatticeEdge( CLatticeNode *f, CLatticeNode *t, char *c );

  // Konstruktor mit Prosodie
  //LatticeEdge(NodeID f,NodeID t,char *c,char *prosInfo):wrd(c,prosInfo) { FromNode=f; ToNode=t; }

  // Destruktor
  virtual ~CLatticeEdge();

  static long s_NumberOfObjects;


  // Kopie erzeugen
  virtual CLatticeEdge * Clone() const;

// copy the content of this edge to given edge
  virtual void CopyFrom( const CLatticeEdge *pEdge );

  // copy the content of given edge to this edge
  virtual void CopyTo( CLatticeEdge *pEdge ) const;

  // make a new edge wich runs from FromNode of this edge 
  // to ToNode of pEdge; it is tested if this->ToNode == pEdge->FromNode
  virtual CLatticeEdge * Concatenate( const CLatticeEdge * ) { return NULL; }


  // 
  //virtual void Init( CLatticeNode *f, CLatticeNode *t );
  virtual void Init( CLatticeNode *f, CLatticeNode *t, const char *c );

  // Selektor f¸r Startknoten
  CLatticeNode * FromNode();

  // Selektor f¸r Zielknoten
  CLatticeNode * ToNode();

  void SetFromNode( CLatticeNode *pFromNode );
  void SetToNode( CLatticeNode *pToNode );

  //! set word
  void SetWord( const CString &Word );

  // Selektor f¸r ¸berspanntes Wort
  inline CString Word() const
  { return m_Word; }

  // Kosten fuer eine Kante
  virtual float Score() const;
  virtual void  SetScore(float Score);

  virtual float TotalScore()  const
  { return 0; }

  virtual bool ShouldBeRemoved() const;

  // einstelliger Ausgabeoperator
  inline virtual std::ostream& operator<<( std::ostream& os );

  virtual void WriteTo( std::ostream& os );

  virtual void Display() const;

  virtual bool operator == ( CLatticeEdge &Edge );
  virtual bool operator < ( CLatticeEdge &Edge );


protected:
  // Anfangsknoten der Kante
  CLatticeNode * m_pFromNode;

  // Endknoten der Kante
  CLatticeNode * m_pToNode;

  // ‹berspanntes Wort
  CString m_Word;

  friend class CLattice;

};





// Lattice-Objekt. Eine Lattice besteht aus einer Menge von 
// Knoten N und einer Menge von Kanten E. Jede Kante aus E 
// besitzt einen Start- und Endpunkt aus N.

class CLattice
{

public:
  // Konstruktor. Erzeugt eine leere Lattice.
  CLattice( CLatticeEdge * pEdgePrototype );

  // Destruktor. Lˆscht enthaltene Knoten und Kanten mit.
  virtual ~CLattice();

  static long s_NumberOfObjects;
  
  //virtual inline long NumberOfObjects() const
  //{ return s_NumberOfObjects; }

  // Lattice "leeren". Knoten und Kanten werden entfernt.
  virtual void Clear();

  // Selektor f¸r Startknoten
  inline CLatticeNode * StartNode() const
  { return m_Nodes[ m_Nodes.MinIndex() ]; }

  // Selektor f¸r Endknoten
  inline CLatticeNode * EndNode() const
  { return m_Nodes[ m_Nodes.MaxIndex() ]; }

  // return node for nodenumber
  inline CLatticeNode * Node( int NodeNumber ) const
  { return m_Nodes[ NodeNumber ]; }

  // list of nodes
  inline const CArray< CLatticeNode * > & Nodes() const
  { return m_Nodes; }

  // list of edges
  inline const CArray< CLatticeEdge * > & Edges() const
  { return m_Edges; }


  // neuen Knoten in das Lattice einf¸gen.
  // @param n Identifikationsmerkmal des neuen Knotens
  void AddNode( CLatticeNode *pNode );

  // neue Kante in das Lattice einf¸gen. Redundanzcheck findet
  // aus Performanzgr¸nden nicht statt.
  // @param e Identifikationsmerkmal der neuen Kante
  void AddEdge( CLatticeEdge *pEdge );

  // add new edge only if it is new (!==) or better (<) then an old one
  // return value: TRUE if edge was added, FALSE otherwise
  bool AddEdgeUniq( CLatticeEdge *pEdge );


  // add edge spanning from StartNodeNumber to EndNodeNumber
  // creates these nodes if necessary
  bool AddEdge( int StartNodeNumber, int EndNodeNumber, CLatticeEdge *pEdge );

  // remove one edge, updating information in nodes
  void RemoveEdge( CLatticeEdge *pEdge );

  // removes edges with <epsilon> as word
  //  copies all incoming edges and connects them to next node, with score
  //  equal to incoming edge->Score * <epsilon> edge->Score
  //  removes isolated nodes, which can only exist if first node has one edge
  //  which is <epsilon>; the resulting lattice remains connected, or empty, if 
  //  all edges are <epsilon>
  void RemoveSpecialEdgesConnectLattice( int nodeJumpMax = 10000 );
  
  // removes edges below threshold
  // if all edges between two nodes are below threshold, keep edge with highest
  // score, thereby keeping the lattice connected
  void RemoveEdgesBelowThreshold(float threshold);
  
  // remove all in-edge of one node
  inline void RemoveInEdges( CLatticeNode *pNode );

  // remove all out-edge of one node
  inline void RemoveOutEdges( CLatticeNode *pNode );

  // remove edges for which ShouldBeRemoved() is true
  // return TRUE if lattice is still connected, i.e. there
  // exists a path from first to last node
  bool RemoveSpecialEdges();

  // copy all InEdges of Node to ToNode of OutEdge
  void CopyInEdges( CLatticeNode *pNode, CLatticeEdge *pOutEdge );

  // copy all OutEdges of Node to FromNode of InEdge
  void CopyOutEdges( CLatticeNode *pNode, CLatticeEdge *pInEdge );


  // Lattice aus Datei einlesen. Beim Format wird KEINE 
  // Fehlerbehandlung durchgef¸hrt.
  // @param Filename Dateiname (voller Name, kompletter Pfad)
  // @return Konnte die angegebene Datei im Filesystem nicht 
  // lokalisiert werden, gibt die Funktion den Wert 1 zur¸ck. 
  // Anderenfalls exakt den durch addVMLine(char *) returnierten Wert.
  // @see addVMLine(char *buffer)
  virtual int ReadFromFile( const CString &Filename );
  int WriteToFile( const CString &Filename );
  virtual int ReadFromBuffer( const CString &Buffer );
  int ConvertFromText( const CArray< CString > &Sentence );
  int AddEdgesFromText( const CArray< CString > &Sentence );

  // Ausgabeoperatoren
  //virtual std::ostream& operator<<(std::ostream&);
  //std::ostream& operator<<(std::ostream& os, const CLattice<TNode, TEdge> & Lattice);

  void Display() const;
  void DisplayEdges() const;

  void DisplayStatistics() const;

  void Traverse();

  // some statistics of the lattice
  int NumberOfNodes() const;
  int NumberOfIsolatedNodes() const;
  int NumberOfEdges() const;

  int MaxPathLength() const;
  int MinPathLength() const;
  float AveragePathLength() const;


  // find the shortest path, i.e. the one which runs over
  // a minimal number of edges
  // return length of that path
  int ShortestPath( CArray< CLatticeEdge* > &ShortestPath ) const;

  // collect the N shortest pathes and their length
  // return the number (<=N) of found pathes
  int ShortestPathes( int N, multimap< int, CArray< CLatticeEdge* > > &ShortestPathes ) const;

  // find the best path through the lattice according
  // to the scores at the edges
  float BestPath( CArray< CLatticeEdge* > &ShortestPath ) const;

  // collect the N-best pathes through the lattice and their scores
  // return the number (<=N) of found pathes
  virtual int BestPathes( int N, multimap< float, CArray< CLatticeEdge* > > &BestPathes ) const;

  long long int NumberOfPathes() const;

  // Information einer Verbmobil-Dateizeile in die Lattice eintragen.
  // @param buffer Zeiger auf gef¸llten Puffer, der eine 
  // Dateizeile enth‰lt.
  // @return Falls kein Fehler aufgetreten ist, liefert die Funktion 
  // 0 zur¸ck. Falls der Speicher zur Durchf¸hrung der Operation zu 
  // klein ist, eine 2.
  virtual int ReadLine( const CString & Buffer );
  
  virtual int ReadLineASR( const CString & Buffer );
  
  //! Reads ASR edges with multple edge scores
  virtual int ReadLineMS( const CString & Buffer );
  
  virtual int ReadLineSRI( const CString & Buffer );
  
  enum LatticeFileFormat { DEFAULT, ASR, SRI, ASRMS };
  
  int SetLatticeFileFormat( char* Format );
  
  void SetLatticeFileFormat( LatticeFileFormat Format);
  LatticeFileFormat GetLatticeFileFormat();

protected:
  // Liste aller Knoten
  CArray< CLatticeNode * > m_Nodes;
  
  CLatticeEdge *m_pEdgePrototype;

  // Liste aller Kanten
  CArray< CLatticeEdge * > m_Edges;
  
  LatticeFileFormat m_FileFormat;

  
  // Verweis auf den logisch ersten Knoten der Lattice.
  //TNode *m_FirstNode;
  
  // Verweis auf den logisch letzten Knoten der Lattice.
  //TNode *m_LastNode;

  // Konstante. Puffergrˆﬂe f¸r das Einlesen von Dateizeilen.
  //static const int m_MaxLineLength;
};


#include <LatticeNode_inl.hh>
#include <LatticeEdge_inl.hh>
#include <Lattice_inl.hh>

#endif


