#include <config.h>

#ifndef __LatticeEdgeTrans__
#define __LatticeEdgeTrans__

#include <sstream>

#include <Prosody.hh>
#include <Transducer.hh>
#include <Lattice.hh>
#include <Trace.hh>




// ----------------------------------------------------------------
// CLatticeEdgeTrans
// ----------------------------------------------------------------


class CLatticeEdgeTrans : public CLatticeEdge
{
public:
  CLatticeEdgeTrans();
  CLatticeEdgeTrans( CLatticeNode *FromNode, CLatticeNode *ToNode );

  virtual ~CLatticeEdgeTrans();
  
  // create a new edge which is a copy of the current edge
  virtual CLatticeEdge * Clone() const;

  // create a new edge which spans this edge and pEdge
  // it is tested if the to edges are adjacent
  virtual CLatticeEdge * Concatenate( const CLatticeEdge *pEdge );

  //virtual void Init( CLatticeNode *f, CLatticeNode *t );
  virtual void Init( CLatticeNode *f, CLatticeNode *t, const char *c );
  virtual void Init( CLatticeNode *f, CLatticeNode *t, int FromNode, int ToNode,
	     CString Word, float Score, int FromTime, int ToTime, const char *Buffer );

  // should edge be removed from the lattice; i.e. is Pause, NIB, etc
  virtual bool ShouldBeRemoved() const;


  // word at the edge
  virtual const CString & Word();
  virtual void SetWord( CString Word );

  void SetWordIndex( CWordIndex WordIndex );
  CWordIndex WordIndex();


  // interpolation factor between acoustic and LM score
  static void SetLMScale( float LMScale );

  // acoustic score
  virtual float Score() const;
  virtual void SetScore( float Score );

  // error from weighted edit distance
  virtual float ErrorScore() const;
  virtual void SetErrorScore( float ErrorScore );

  // benefit for longer edges
   virtual float ConcatScore() const;
  virtual void SetConcatScore( float ConcatScore );

  // complete score
  virtual float TotalScore() const;

  // time frames
  unsigned long Startframe();
  unsigned long Endframe();

  // prosody
  CProsody * Prosody();
  CProsody & GetProsody();

  // wordinfo
  void SetWordInfo( CWordInfo *pWordInfo ) { 
    if ( m_pWordInfo != NULL ) {
      delete m_pWordInfo;
    } 
    m_pWordInfo = pWordInfo; 
  }
  
  CWordInfo * WordInfo();
  const CWordInfo * WordInfo() const { return m_pWordInfo; }

  // Array with word indices for LM
  //CNumericalArray< long > & LMWordIndices();

  // Array with word indices for LM
  CNumericalMatrix< long > & WordIndicesForLMs();


  // edges from which new edge was constructed through transducer
  void SetGeneratingEdges( CArray< CLatticeEdgeTrans * > *pWhgEdges );
  CArray< CLatticeEdgeTrans * > * GeneratingEdges();

  // transducer
  const CTransducer * Transducer();
  void SetTransducer( const CTransducer *pTransducer );
  // transducer item
  CTransducerItem * TransducerItem();
  void SetTransducerItem( CTransducerItem *TransducerState );

  void SetMatchFlag();
  bool MatchFlag();

  virtual std::ostream& operator<<( std::ostream& os ) const;
  virtual void WriteTo( std::ostream& os ) const;

  virtual void Display() const;

  virtual bool operator == ( CLatticeEdge &Edge );
  virtual bool operator < ( CLatticeEdge &Edge );

  //static long int s_NumberOfObjects;

protected:
  // word at the edge
  CString m_Word;
  CWordIndex m_WordIndex;

  // acoustic score
  float m_Score;

  // time information
  unsigned long m_Startframe, m_Endframe;

  // prosodic information
  CProsody m_Prosody;
  
  // additional word information: source sequence used to construct
  // this edge, translation, score for translation
  CWordInfo *m_pWordInfo;

  // array of word indices to store LM word indices for target words
  //CNumericalArray< long > m_LMWordIndices;

  // array of word indices to store LM word indices for target words
  CNumericalMatrix< long > m_WordIndicesForLMs;

  // score for errors from weighted edit distance
  float m_ErrorScore;

  // score for concatenation, i.e. one edge is result of spanning several
  // edges, the ConcatScore (which is negative) gives the benefit
  float m_ConcatScore;

  // ???
  float m_WordScore;

  // scaling factors for the different scores
  // target language model
  static float m_LMScale;
  // errors
  //static float m_ErrorScale;

  // information from which edges current edge was generated
  CArray< CLatticeEdgeTrans * > *m_pGeneratingEdges;

  // information which transducer and which final state
  // generated this edge (set for newly generated edges)
  const CTransducer *m_pTransducer;
  CTransducerItem *m_pTransducerItem;

  // match-flag is used in the test, if edges in source and target lattice
  // have been create from same transducer and transducer item
  bool m_MatchFlag;

};




// ----------------------------------------------------------------
// CLatticeEdgeTrans - Inlines
// ----------------------------------------------------------------


inline
CLatticeEdgeTrans::CLatticeEdgeTrans()
{
  //  Trace( 2, "CLatticeEdgeTrans::CLatticeEdgeTrans" );
  m_Score = 0.0;
  m_Startframe = 0;
  m_Endframe = 0;
  //m_pWordInfo = new CStatTransInfo;
  m_pWordInfo = new CWordInfo;
  m_ErrorScore = 0.0;
  m_ConcatScore = 0.0;
  m_WordIndex = 0;
  m_pGeneratingEdges = NULL;
  m_pTransducer = NULL;
  m_pTransducerItem = NULL;
  m_MatchFlag = FALSE;
  //s_NumberOfObjects++;
}


inline
CLatticeEdgeTrans::CLatticeEdgeTrans( CLatticeNode *FromNode, CLatticeNode *ToNode )
: CLatticeEdge( FromNode, ToNode )
{
  m_Score = 0.0;
  m_Startframe = 0;
  m_Endframe = 0;
  //m_pWordInfo = new CStatTransInfo;
  m_pWordInfo = new CWordInfo;
  m_ErrorScore = 0.0;
  m_ConcatScore = 0.0;
  m_WordIndex = 0;
  m_pGeneratingEdges = NULL;
  m_pTransducer = NULL;
  m_pTransducerItem = NULL;
  m_MatchFlag = FALSE;
  //s_NumberOfObjects++;
}


inline
CLatticeEdgeTrans::~CLatticeEdgeTrans()
{
  delete m_pWordInfo;
  delete m_pGeneratingEdges;
  //s_NumberOfObjects--;
}




inline bool 
CLatticeEdgeTrans::ShouldBeRemoved() const
{
  if ( (m_Word[ 0 ] == '#') || (m_Word == CString( "<UNK>" )) )
    return true;

  return false;
}



// word at the edge
inline const CString & 
CLatticeEdgeTrans::Word()
{
  return m_Word;
}


inline void
CLatticeEdgeTrans::SetWord( CString Word )
{
  m_Word = Word;
}

inline void
CLatticeEdgeTrans::SetScore( float Score )
{
  m_Score = Score;
}


inline void
CLatticeEdgeTrans::WriteTo( ostream& os ) const
{
  os << m_Word << "\t" << m_Score << 
     "\t" << m_Startframe << "\t" << m_Endframe << "\t";
  
  //m_Prosody.WriteTo(os);
  
  if (m_pWordInfo != NULL)
    //m_pWordInfo->WriteTo(os);
    os << m_pWordInfo;
}







inline void
CLatticeEdgeTrans::SetWordIndex( CWordIndex WordIndex )
{
  m_WordIndex = WordIndex;
}


inline CWordIndex
CLatticeEdgeTrans::WordIndex()
{
  return m_WordIndex;
}

  // interpolation factor between acoustic and LM score
inline void 
CLatticeEdgeTrans::SetLMScale( float LMScale )
{
  m_LMScale = LMScale;
}

// acoustic score
inline float 
CLatticeEdgeTrans::Score() const
{ 
  //  Trace( 2, "CLatticeEdgeTrans::Score = " << m_Score );
  return m_Score;
}

// error from weighted edit distance
inline float 
CLatticeEdgeTrans::ErrorScore() const
{ 
  //  Trace( 3, "CLatticeEdgeTrans::ErrorScore = " << m_ErrorScore );
  return m_ErrorScore;
}

inline void 
CLatticeEdgeTrans::SetErrorScore( float ErrorScore )
{ 
  //  Trace( 3, "CLatticeEdgeTrans::SetErrorScore = " << ErrorScore );
  m_ErrorScore = ErrorScore;
}

// benefit for longer edges
inline float 
CLatticeEdgeTrans::ConcatScore() const
{ 
  //  Trace( 3, "CLatticeEdgeTrans::ConcatScore = " << m_ConcatScore );
  return m_ConcatScore;
}

inline void 
CLatticeEdgeTrans::SetConcatScore( float ConcatScore )
{ 
  //  Trace( 3, "CLatticeEdgeTrans::SetConcatScore = " << ConcatScore );
  m_ConcatScore = ConcatScore;
}

// complete score
inline float 
CLatticeEdgeTrans::TotalScore() const
{ 
//     Trace( 2, "CLatticeEdgeTrans::TotalScore = " 
// 	 << ((1.0  - m_LMScale) * m_Score) << " "
// 	 << (m_LMScale * m_pWordInfo->Score() ) );
  return ( ( 1.0 - m_LMScale ) * m_Score 
	   + m_LMScale * m_pWordInfo->Score() );
  //return m_pWordInfo->Score();
}



// time frames
inline unsigned long 
CLatticeEdgeTrans::Startframe() 
{ 
  return m_Startframe;
}


inline unsigned long 
CLatticeEdgeTrans::Endframe()
{
  return m_Endframe;
}


  // prosody
inline CProsody * 
CLatticeEdgeTrans::Prosody() 
{
  return &m_Prosody;
}


inline CProsody & 
CLatticeEdgeTrans::GetProsody()
{
  return m_Prosody;
}


// wordinfo
inline CWordInfo * 
CLatticeEdgeTrans::WordInfo()
{ 
  assert( m_pWordInfo );
  return m_pWordInfo; 
}


/*
inline CNumericalArray< long > &
CLatticeEdgeTrans::LMWordIndices()
{ 
  return m_LMWordIndices; 
}
*/


inline CNumericalMatrix< long > &
CLatticeEdgeTrans::WordIndicesForLMs()
{ 
  return m_WordIndicesForLMs;
}



inline void 
CLatticeEdgeTrans::SetGeneratingEdges( CArray< CLatticeEdgeTrans * > *pWhgEdges )
{
  m_pGeneratingEdges = pWhgEdges;
}


inline   CArray< CLatticeEdgeTrans * > *
CLatticeEdgeTrans::GeneratingEdges()
{
  return m_pGeneratingEdges;
}


// transducer
inline const CTransducer * 
CLatticeEdgeTrans::Transducer()
{
  return m_pTransducer;
}


inline void 
CLatticeEdgeTrans::SetTransducer( const CTransducer *pTransducer )
{
  m_pTransducer = pTransducer;
}


// transducer state
inline CTransducerItem *
CLatticeEdgeTrans::TransducerItem()
{
  return m_pTransducerItem;
}


inline void
CLatticeEdgeTrans::SetTransducerItem( CTransducerItem *pTransducerItem )
{
  m_pTransducerItem = pTransducerItem;
}


inline void
CLatticeEdgeTrans::SetMatchFlag()
{
  m_MatchFlag = TRUE;
}

inline bool
CLatticeEdgeTrans::MatchFlag()
{
  return m_MatchFlag;
}


#endif


