#include <config.h>

#ifndef __LatticeEdgeVM_hh__
#define __LatticeEdgeVM_hh__

#include <strstream.h>

#include <Lattice.hh>
#include <Text.hh>



// ----------------------------------------------------------------
// CProsody
// ----------------------------------------------------------------
class CProsody
{
public:
  CProsody();
  CProsody( const CProsody &Prosody );

  CProsody & operator = ( const CProsody &Prosody );

  void ReadFromBuffer( const char *Buffer );
  virtual void WriteTo( ostream& os );
  void ToString( CString& String ) const;

  double GetBoundary2() { return m_Boundary2;}
  double LogNoBoundary2();

  void GetBoundaries( double &Boundary0, double &Boundary1, double &Boundary2, 
		      double &Boundary3, double &Boundary4 );

  void SetBoundaries( double Boundary0, double Boundary1, double Boundary2, 
		      double Boundar3, double Boundary4 );

  void GetMode( double &Mode0, double &Mode1, double &Mode2 );
  void SetMode( double Mode0, double Mode1, double Mode2 );

  double GetMode2();


private:
  void Init();

  double m_Accent0, m_Accent1;
  double m_Boundary0, m_Boundary1, m_Boundary2, m_Boundary3, m_Boundary4;
  double m_Mode0, m_Mode1, m_Mode2;

};



// ----------------------------------------------------------------
// CStatTransInfo
// ----------------------------------------------------------------

class CStatTransInfo
{
public:
  CStatTransInfo();
  CStatTransInfo( const CStatTransInfo * );
  virtual ~CStatTransInfo() {;}

  void ReadFromBuffer( const char *Buffer );
  virtual void WriteTo( ostream& os );

  const CString & FirstWordTarget() const;
  const CString & LastWordTarget() const;

  CString SourceString() const;
  CString TargetString() const;

  int SourceStringLength() const;
  int TargetStringLength() const;

  int NumberOfEdges() const;
  void SetNumberOfEdges( int NumberOfEdges );


private:
  inline void SplitUpToTokens( char *, CArray< CString > &TokArr, const char *delimiters );

  CArray< CString > m_SentenceF;
  CArray< CString > m_SentenceE;

  // a edge may replace several old edges (pause, nib, etc.)
  int m_NumberOfEdges;
};



// ----------------------------------------------------------------
// CLatticeEdgeVM
// ----------------------------------------------------------------


class CLatticeEdgeVM : public CLatticeEdge
{
public:
  CLatticeEdgeVM();

  // -- Konstruktor mit Anfangs- und Endknoten
  CLatticeEdgeVM( CLatticeNode *FromNode, CLatticeNode *ToNode );

  virtual ~CLatticeEdgeVM();
  
  virtual CLatticeEdge * Clone() const;

  // make a new edge wich runs from FromNode of this edge 
  // to ToNode of pEdge; it is tested if this->ToNode == pEdge->FromNode
  virtual CLatticeEdge * Concatenate( const CLatticeEdge *pEdge );

  //virtual void Init( CLatticeNode *f, CLatticeNode *t );
  virtual void Init( CLatticeNode *f, CLatticeNode *t, const char *c );

  // -- should edge be removed from the lattice; i.e. is Pause, NIB, etc
  virtual bool ShouldBeRemoved() const;


  // -- Selektor fuer ueberspanntes Wort
  const CString & Word();

  // interpolation factor between acoustic and LM score
  static void SetLMScale( double LMScale )
  {
    m_LMScale = LMScale;
  }

  // -- akustische Bewertung
  virtual double Score() const
  { 
    Trace( 3, "CLatticeEdgeVM::Score = " << m_Score );
    return m_Score;
  }

  // -- komplette
  virtual double TotalScore() const
  { 
    Trace( 3, "CLatticeEdgeVM::TotalScore = " 
	   << ((1.0  - m_LMScale) * m_Score) << " "
	   << (m_LMScale * m_pWordInfo->Score() ) );
    return ( ( 1.0 - m_LMScale ) * m_Score 
	     + m_LMScale * m_pWordInfo->Score() );
    //return m_pWordInfo->Score();
  }

  // -- time frames
  unsigned long Startframe() { return m_Startframe; }
  unsigned long Endframe() { return m_Endframe; }

  // -- prosody
  CProsody * Prosody() { return &m_Prosody; }

  // -- additional Info for statistical translation
  //CStatTransInfo * StatTransInfo() { 
  //  assert( m_pWordInfo );
  //  return m_pWordInfo; 
  //}
  CWordInfo * StatTransInfo() { 
    assert( m_pWordInfo );
    return m_pWordInfo; 
  }

  virtual ostream& operator<<( ostream& os ) const;
  virtual void WriteTo( ostream& os );

  virtual void Display() const;

  virtual bool operator == ( CLatticeEdge &Edge );
  virtual bool operator < ( CLatticeEdge &Edge );

protected:
  // -- ueberspanntes Wort
  CString m_Word;

  // -- akustische Bewertung
  double m_Score;
  static double m_LMScale;
  // OLI
  unsigned long m_Startframe, m_Endframe;

  // -- prosodischen Informationen
  CProsody m_Prosody;
  
  // -- Informationen zum $phrase-Label
  //CStatTransInfo *m_pWordInfo;
  CWordInfo *m_pWordInfo;

  double m_WordScore;
};




// ----------------------------------------------------------------
// CProsody - Inlines
// ----------------------------------------------------------------

inline
CProsody::CProsody()
{
  Init();
}



inline
CProsody::CProsody( const CProsody &Prosody )
{
  *this = Prosody;
}


inline CProsody &
CProsody::operator = ( const CProsody &Prosody )
{
  if ( this == &Prosody ) 
    //-- Kopie auf sich selber kann ignoriert werden
    return *this;

  m_Accent0   = Prosody.m_Accent0; 
  m_Accent1   = Prosody.m_Accent1;
  m_Boundary0 = Prosody.m_Boundary0;
  m_Boundary1 = Prosody.m_Boundary1;
  m_Boundary2 = Prosody.m_Boundary2; 
  m_Boundary3 = Prosody.m_Boundary3;
  m_Boundary4 = Prosody.m_Boundary4;
  m_Mode0     = Prosody.m_Mode0;
  m_Mode1     = Prosody.m_Mode1;
  m_Mode2     = Prosody.m_Mode2;

  return *this;
}



inline void
CProsody::Init()
{
  m_Accent0 = 1.0; 
  m_Accent1 = 0.0;
  m_Boundary0 = 1.0;
  m_Boundary1 = 0.0;
  m_Boundary2 = 0.0; 
  m_Boundary3 = 0.0;
  m_Boundary4 = 0.0;
  m_Mode0 = 1.0;
  m_Mode1 = 0.0;
  m_Mode2 = 0.0;  
}





inline double
CProsody::LogNoBoundary2()
{ 
  if ( m_Boundary2 >= 1.0 )
    return log( 0.99999 );
  else
    return log( 1.0 - m_Boundary2 );
}


inline void
CProsody::GetBoundaries( double &Boundary0, double &Boundary1, double &Boundary2,
			 double &Boundary3, double &Boundary4 )
{
  // make shure we have no 0.0 values
  Boundary0 = m_Boundary0;
  Boundary1 = m_Boundary1;
  Boundary2 = m_Boundary2;
  Boundary3 = m_Boundary3;
}



inline void
CProsody::SetBoundaries( double Boundary0, double Boundary1,  double Boundary2,
			 double Boundary3, double Boundary4 )
{
  // make shure we have no 0.0 values
  if ( Boundary0 == 0.0 )
    m_Boundary0 = 1.0e-10;
  else
    m_Boundary0 = Boundary0;

  if ( Boundary1 == 0.0 )
    m_Boundary1 = 1.0e-10;
  else
    m_Boundary1 = Boundary1;

  if ( Boundary2 == 0.0 )
    m_Boundary2 = 1.0e-10;
  else
    m_Boundary2 = Boundary2;

  if ( Boundary3 == 0.0 )
    m_Boundary3 = 1.0e-10;
  else
    m_Boundary3 = Boundary3;

}



inline void
CProsody::GetMode( double &Mode0, double &Mode1, double &Mode2 )
{
  Mode0 = m_Mode0;
  Mode1 = m_Mode1;
  Mode2 = m_Mode2;
}



inline void
CProsody::SetMode( double Mode0, double Mode1, double Mode2 )
{
  m_Mode0 = Mode0;
  m_Mode1 = Mode1;
  m_Mode2 = Mode2;
}


inline double
CProsody::GetMode2()
{
  return m_Mode2;
}


inline void
CProsody::ReadFromBuffer( const char *Buffer )
{
  // p is of the form: (PR (A  0.28 0.28) (G  0.28 0.99 0.28 0.99) (M  0.44 0.44 0.44))

  Trace( 1, "CProsody::ReadFromBuffer " << Buffer );

  if( ( strncmp( Buffer, "(PR ", 4 ) == 0 ) 
      && ( strstr( Buffer, "(S ") != NULL ) 
      && ( strstr( Buffer, "(F ") != NULL )
      && ( strstr( Buffer, "(A ") != NULL ) ) {
    // Prosodie scheint ok zu sein
    char buffer[ 4096 ];
    char *po;
    
    // Accent
    strcpy( buffer, strstr( Buffer, "(A " ) );
    po = buffer;
    while ( *po != ')' ) {
    po++;
    }
    po++;
    *po = '\0';
    //Trace( 1, "Accent: " << buffer );
    sscanf( buffer, "(A%lf%lf)" , &m_Accent0, &m_Accent1 );
    
    // Mode
    strcpy( buffer, strstr( Buffer, "(F " ) );
    po = buffer;
    while ( *po != ')' ) {
      po++;
    }
    po++;
    *po = '\0';
    //Trace( 1, "Mode: " << buffer );
    sscanf( buffer, "(F%lf%lf%lf)", &m_Mode0, &m_Mode1, &m_Mode2 );
    
    // Boundaries
    strcpy( buffer, strstr( Buffer,"(S " ) );
    po = buffer;
    while ( *po != ')' ) {
      po++;
    }
    po++;
    *po = '\0';
    //Trace( 1, "Boundaries: " << buffer );
    sscanf( buffer,"(S%lf%lf%lf%lf%lf)",&m_Boundary0, &m_Boundary1, &m_Boundary2, &m_Boundary3, &m_Boundary4 );
  }

  Trace(1, "Boundaries: " << m_Boundary0 << " " << m_Boundary1 << " " 
	<< m_Boundary2 << " " << m_Boundary3 << " " << m_Boundary4 );

  //Trace(1, "Prosody Found " << m_Accent0  << " " << m_Accent1 << " " 
  //<< m_Mode0 << " " << m_Mode1 << " " << m_Mode2 << " "
  //<< m_Boundary0 << " " << m_Boundary1 << " " << m_Boundary2 << " " << m_Boundary3 );
}


inline void
CProsody::WriteTo( ostream& os )
{
  // p is of the form: (PR (A  0.28 0.28) (G  0.28 0.99 0.28 0.99) (M  0.44 0.44 0.44))

  os <<  "(PR (A " << m_Accent0  << " " << m_Accent1 << ") (G " 
     << m_Boundary0 << " " << m_Boundary1 << " " << m_Boundary2 << " " << m_Boundary3 << ") (M "
	<< m_Mode0 << " " << m_Mode1 << " " << m_Mode2 << "))";

}



inline void
CProsody::ToString( CString& String ) const
{
  // p is of the form: (PR (A  0.28 0.28) (G  0.28 0.99 0.28 0.99) (M  0.44 0.44 0.44))

  String += CString( "(PR (A " ) + CString( m_Accent0, "%3.2f" ) 
    + CString( " " ) + CString( m_Accent1, "%3.2f" ) 
    + CString( ") (G " ) + CString( m_Boundary0, "%3.2f" ) 
    + CString( " " ) + CString( m_Boundary1, "%3.2f" )
    + CString( " " ) + CString( m_Boundary2, "%3.2f" ) 
    + CString( " " ) + CString( m_Boundary3, "%3.2f" )
    + CString( ") (M " ) + CString( m_Mode0, "%3.2f" ) 
    + CString( " " ) + CString( m_Mode1, "%3.2f" )
    + CString( " " ) + CString( m_Mode2, "%3.2f" ) 
    + CString( "))" );
}




// ----------------------------------------------------------------
// CStatTransInfo - Inlines
// ----------------------------------------------------------------

inline
CStatTransInfo::CStatTransInfo()
{
  m_NumberOfEdges = 1;
}


inline 
CStatTransInfo::CStatTransInfo( const CStatTransInfo *pStatTransInfo )
{
  // copy of string arrays missing !!!
  //CString Source = pStatTransInfo->SourceString();
  //if ( Source != "" ) 
  //  cout << "StatTransInfo not completely copied" << endl;
  //Trace( 0, "CStatTransInfo - Copy Constructor" );
  m_NumberOfEdges = pStatTransInfo->m_NumberOfEdges;
}



inline CString
CStatTransInfo::SourceString() const
{
  CString Total;
  int i;

  for ( i = m_SentenceF.MinIndex(); i <= m_SentenceF.MaxIndex(); i++ ) {
    Total += m_SentenceF[ i ];
    Total += " ";
  }

  return Total;
}


inline CString
CStatTransInfo::TargetString() const
{
  CString Total;
  int i;

  for ( i = m_SentenceE.MinIndex(); i <= m_SentenceE.MaxIndex(); i++ ) {
    Total += m_SentenceE[ i ];
    Total += " ";
  }

  return Total;
}



inline void
CStatTransInfo::ReadFromBuffer( const char *Buffer )
{
  CString dummy;

  // p is of the form: (ST (SF  wie waere es denn ) (SE what about ))

  Trace( 3, "CStatTransInfo::ReadFromBuffer " << Buffer );

  if( strncmp( Buffer, "(ST ", 4 ) == 0 ) {

    char buffer[ 4096 ];
    char *po;
    
    // number of edges
    if ( (po = strstr( Buffer, "(NE ")) != NULL) {
      sscanf( po,"(NE %i", &m_NumberOfEdges );
    }
    
    // French sentence
    if ( strstr( Buffer, "(SF ") != NULL) {
      strcpy( buffer, strstr( Buffer, "(SF " ) );
      po = buffer;
      while ( *po != ')' ) {
	po++;
      }
      //po++;
      *po = '\0';
      Trace( 3, "French sentence: " << buffer );
      
      SplitUpToTokens(buffer+3, m_SentenceF, " ");
    }
    
    if ( strstr( Buffer, "(SE ") != NULL) {
      strcpy( buffer, strstr( Buffer, "(SE " ) );
      po = buffer;
      while ( *po != ')' ) {
	po++;
      }
      //po++;
      *po = '\0';
      Trace( 3, "English sentence: " << buffer );
      
      SplitUpToTokens( buffer + 3, m_SentenceE, " " );
    }
  }
}


inline void
CStatTransInfo::WriteTo( ostream& os )
{
  // opening
  os <<  "(ST ";

  // number of original edges this edge spans over
  os << "(NE " << m_NumberOfEdges << ")";

  // phrase: original and translation
  CString Source = SourceString();
  CString Target = TargetString();

  if ( Source.GetLength() > 0 ) {
    os <<  "(SF " << Source << ") (SE " << Target << ")";
  }

  // closing
  os << ") ";
}


inline void
CStatTransInfo::SplitUpToTokens(char *TokStr, CArray<CString> &TokArr, const char *delimiters)
{
  char *ptr = NULL;
  char token[4096];
  int len = 0;
  int i = 0;

  ptr = TokStr;
  Trace( 5, "SplitUpToTokens:" << TokStr );

  while ( *(ptr += strspn( ptr, delimiters )) )  { // Find first word after delimiters
    len = strcspn(ptr, delimiters);                // Find first delimiter after word
    strncpy( token, ptr, len );
    token[ len ] = '\0';
    TokArr[ i++ ] = CString( token );
    Trace0( 5, " " << token );
    ptr += len;
  }
  Trace( 5, "" );
}


inline const CString &
CStatTransInfo::FirstWordTarget() const
{
  return m_SentenceE[ m_SentenceE.MinIndex() ];
}



inline const CString &
CStatTransInfo::LastWordTarget() const
{
  return m_SentenceE[ m_SentenceE.MaxIndex() ];
}



inline int
CStatTransInfo::SourceStringLength() const
{
  return m_SentenceE.Size();
}


inline int
CStatTransInfo::TargetStringLength() const
{
  return m_SentenceF.Size();
}


inline int
CStatTransInfo::NumberOfEdges() const
{
  return m_NumberOfEdges;
}



inline void
CStatTransInfo::SetNumberOfEdges( int NumberOfEdges )
{
  m_NumberOfEdges = NumberOfEdges;
}







// ----------------------------------------------------------------
// CLatticeEdgeVM - Inlines
// ----------------------------------------------------------------


inline
CLatticeEdgeVM::CLatticeEdgeVM()
{
  m_Score = 0.0;
  m_Startframe = 0;
  m_Endframe = 0;
  //m_pWordInfo = new CStatTransInfo;
  m_pWordInfo = new CWordInfo;
}


inline
CLatticeEdgeVM::CLatticeEdgeVM( CLatticeNode *FromNode, CLatticeNode *ToNode )
: CLatticeEdge( FromNode, ToNode )
{
  m_Score = 0.0;
  m_Startframe = 0;
  m_Endframe = 0;
  //m_pWordInfo = new CStatTransInfo;
  m_pWordInfo = new CWordInfo;
}


inline
CLatticeEdgeVM::~CLatticeEdgeVM()
{
  Trace( 3, "CLatticeEdgeVM::~CLatticeEdgeVM()" );
  delete m_pWordInfo;
}


inline void
CLatticeEdgeVM::Init( CLatticeNode *f, CLatticeNode *t, const char *Buffer )
{
  Trace( 3, "CLatticeEdgeVM::Init " << Buffer );
  m_pFromNode = f;
  m_pToNode = t;

  char WordBuffer[ MaxLineLength ];
  WordBuffer[ 0 ] = '\0';
  unsigned long From, To;
  unsigned long fframe, tframe;
  double sc;

  sscanf( Buffer,"%lu%lu%s%lf%lu%lu", &From, &To, WordBuffer, &sc, &fframe, &tframe );
  Trace( 3, "From: " << From << " To: " << To << "  " << WordBuffer << " " 
	 << sc << " " << fframe << " " << tframe );

  m_Word = CString( WordBuffer );
  m_Score = sc;

  m_Startframe = fframe;
  m_Endframe = tframe;

  char *Prosody = strstr( Buffer, "(PR " );
  Trace( 3, Prosody );
  
  if (Prosody != NULL)
    m_Prosody.ReadFromBuffer( Prosody );

  char *StatTrans = strstr( Buffer, "(ST {");
  
  if (StatTrans != NULL) {
    Trace( 3, "StatTrans: " << StatTrans );
    Trace( 3, "StatTrans: " << StatTrans + 6);
    if (m_pWordInfo == NULL)
      //m_pWordInfo = new CStatTransInfo;
      m_pWordInfo = new CWordInfo;

    //m_pWordInfo->ReadFromBuffer( StatTrans );
    istrstream Stream( StatTrans + 6 );
    Stream >> *m_pWordInfo;
    Trace( 3, "StatTrans: " << *m_pWordInfo );
  }
}



inline CLatticeEdge *
CLatticeEdgeVM::Clone() const
{
  Trace( 4, "CLatticeEdgeVM::Clone()" );

  CLatticeEdgeVM *pNewEdge = new CLatticeEdgeVM( m_pFromNode, m_pToNode );

  pNewEdge->m_Word = m_Word;
  pNewEdge->m_Score = m_Score;
  pNewEdge->m_Startframe = m_Startframe;
  pNewEdge->m_Endframe = m_Endframe;
  //pNewEdge->m_pWordInfo = new CStatTransInfo( m_pWordInfo );
  //pNewEdge->m_pWordInfo = new CWordInfo( m_pWordInfo );
  pNewEdge->m_Prosody = m_Prosody;

  return pNewEdge;
}


inline CLatticeEdge *
CLatticeEdgeVM::Concatenate( const CLatticeEdge *pEdge )
{
  Trace( 4, "CLatticeEdgeVM::Concatenate()" );

  CLatticeEdgeVM *pNewEdge = NULL;
  CLatticeEdgeVM *pOtherEdge = (CLatticeEdgeVM *)pEdge;

  if ( TraceLevel >= 4 ) {
    Display();
    pOtherEdge->Display();
  }

  // have the two edges one node in common
  if ( m_pToNode == pOtherEdge->m_pFromNode ) {        // concat to the right
    Trace( 4, "Span over right edge - left edge is dominant" );
    pNewEdge = new CLatticeEdgeVM( m_pFromNode, pOtherEdge->m_pToNode );
    pNewEdge->m_Word = m_Word;
    pNewEdge->m_Startframe = m_Startframe;
    pNewEdge->m_Endframe = pOtherEdge->m_Endframe;
    pNewEdge->m_Score = m_Score + pOtherEdge->m_Score;
    //pNewEdge->m_pWordInfo = new CStatTransInfo( m_pWordInfo );
    pNewEdge->m_pWordInfo = new CWordInfo( m_pWordInfo );
    //pNewEdge->m_pWordInfo->SetNumberOfEdges( m_pWordInfo->NumberOfEdges()
    //					  + pOtherEdge->m_pWordInfo->NumberOfEdges() );
    pNewEdge->m_Prosody = m_Prosody;
  } else if ( m_pFromNode == pOtherEdge->m_pToNode ) {      // concat to the left
    Trace( 4, "Span over left edge - right is dominant" );
    pNewEdge = new CLatticeEdgeVM( pOtherEdge->m_pFromNode, m_pToNode );
    pNewEdge->m_Word = m_Word;
    pNewEdge->m_Startframe = pOtherEdge->m_Startframe;
    pNewEdge->m_Endframe = m_Endframe;
    pNewEdge->m_Score = m_Score + pOtherEdge->m_Score;
    //pNewEdge->m_pWordInfo = new CStatTransInfo( m_pWordInfo );
    pNewEdge->m_pWordInfo = new CWordInfo( m_pWordInfo );
    //pNewEdge->m_pWordInfo->SetNumberOfEdges( m_pWordInfo->NumberOfEdges() 
    //					  + pOtherEdge->m_pWordInfo->NumberOfEdges() );
    pNewEdge->m_Prosody = m_Prosody;
  } else { 
    Trace( 4, "Edges not connected" );
  }


  return pNewEdge;

}



inline bool 
CLatticeEdgeVM::ShouldBeRemoved() const
{
  if ( (m_Word[ 0 ] == '#') || (m_Word == CString( "<UNK>" )) )
    return true;

  return false;
}



// Selektor fr berspanntes Wort
inline const CString &
CLatticeEdgeVM::Word() 
{
  return m_Word;
}


// einstelliger Ausgabeoperator
inline ostream& 
CLatticeEdgeVM::operator<<( ostream& os ) const
{
  Trace( 5, "CLatticeEdgeVM::operator << ()" );

  int FromNodeNumber, ToNodeNumber;

  if (m_pFromNode)
    FromNodeNumber = m_pFromNode->Number();
  else
    FromNodeNumber = -1;

  if (m_pToNode)
    ToNodeNumber = m_pToNode->Number();
  else
    ToNodeNumber = -1;

  return (os << "edge " //<< this 
	  // << " (" << m_FromNode << '-' << m_ToNode << " | " 
	  << "(" << FromNodeNumber << "-" << ToNodeNumber
	  << ") \"" << m_Word << " " << m_Score << " " <<
	  m_Startframe << " " << m_Endframe
	  << "\"" ) << endl;
}

inline void
CLatticeEdgeVM::WriteTo( ostream& os )
{
  os << m_Word << "\t" << m_Score << 
     "\t" << m_Startframe << "\t" << m_Endframe << "\t";
  
  m_Prosody.WriteTo(os);
  
  if (m_pWordInfo != NULL)
    //m_pWordInfo->WriteTo(os);
    os << m_pWordInfo;
}



inline void
CLatticeEdgeVM::Display() const
{
  Trace( 5, "CLatticeEdgeVM::Display" );

  int FromNodeNumber, ToNodeNumber;

  if (m_pFromNode)
    FromNodeNumber = m_pFromNode->Number();
  else
    FromNodeNumber = -1;

  if (m_pToNode)
    ToNodeNumber = m_pToNode->Number();
  else
    ToNodeNumber = -1;

  cout << "edge " 
       << "(" << FromNodeNumber << "-" << ToNodeNumber
       << ") " << m_Word << " " << m_Score << " "
       <<  m_Startframe << " " << m_Endframe << " ";

  CString Prosody;
  m_Prosody.ToString( Prosody );
  cout << Prosody << " ";
  //m_pWordInfo->WriteTo( cout );
  cout << *m_pWordInfo;
  cout << endl << flush;

}



inline bool
CLatticeEdgeVM::operator==( CLatticeEdge &Edge )
{
  Trace( 3, "CLatticeEdgeVM::operator==" );
  return ( ( FromNode() == Edge.FromNode() ) &&
	   ( ToNode() == Edge.ToNode() ) &&
	   ( m_Word == ((CLatticeEdgeVM &)Edge).m_Word ) &&
	   ( m_pWordInfo->TargetString() 
	     == ((CLatticeEdgeVM &)Edge).m_pWordInfo->TargetString() ) );
}


inline bool
CLatticeEdgeVM::operator<( CLatticeEdge &Edge )
{
  Trace( 1, "CLatticeEdgeVM::operator<" );

  return ( (*this == Edge) && (m_Score < ((CLatticeEdgeVM &)Edge).m_Score) );
}


#endif

