#ifndef LatticeSearch_hh
#define LatticeSearch_hh

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include <Array.hh>
#include <Lattice.hh>
#include <LatticeEdgeMS.hh>

#include <LatticeEdgeTrans.hh>
#include <Vocabulary.hh>
#include <Transducer.hh>


using namespace std;


// Module for searching phrases in the Whg

#define MAX_ACTIVE_HYPS 100


// CPhraseHyp stores all neccessary information to descripe a partially matched phrase
// that is:
// 1. a descriptor to identify the phrase ( which one and which word position
//    within the phrase); depending on the organization of the phrases (e.g. tree-like)
//    on descriptor can reference several partial phrases
// 2. from which node started the phrase, the accumulated acoustic score
//    and the accumulated costs for error tolerant matching
// 3. for reconstructing the traversed path, backtrace information is stored

struct CLatticeSearchHyp {

public:
  CLatticeSearchHyp() {
    m_State = -1;
    m_StartNode = -1;
    m_AcuScore = 0.0;
    m_Cost = 0.0;
    m_WhgPathLength = 0;
    m_pBackPointerToEdge = NULL;
    m_BackPointerToEntry = -1;
    s_NumberOfObjects++;
  }

  CLatticeSearchHyp(CLatticeSearchHyp &Hyp) {
    m_State = Hyp.m_State;
    m_StartNode = Hyp.m_StartNode;
    m_AcuScore = Hyp.m_AcuScore;
    m_Scores = Hyp.m_Scores;
    m_Cost = Hyp.m_Cost;
    m_WhgPathLength = Hyp.m_WhgPathLength;
    m_pBackPointerToEdge = Hyp.m_pBackPointerToEdge;
    m_BackPointerToEntry = Hyp.m_BackPointerToEntry;
    s_NumberOfObjects++;
  }

  ~CLatticeSearchHyp() {
    s_NumberOfObjects--;
  }

  friend std::ostream &operator << (std::ostream &s, const CLatticeSearchHyp & );

  CTransducerState m_State;           // descriptor for the phrase(s) matched so far
  int              m_StartNode;       // node, from which the partial phrase starts
  double           m_AcuScore;        // accumulated acoustic score over partial path
  //CNumericalArray< float > m_Scores;  // accumulated scores over partial path, apd 10/26/06
  vector< float >  m_Scores;  // accumulated scores over partial path, apd 10/26/06
  double           m_Cost;            // accumulated costs for error-tolerant matching
  int              m_WhgPathLength;   // length of path over which phrase was constructed
                                      // to calculate length-dependent cost in error-tolerant match

  // backtrace information: from which node we came, along which edge, and from what hyp
  CLatticeEdgeTrans *m_pBackPointerToEdge;
  //int                m_BackPointerToNode;   // redundant
  int                m_BackPointerToEntry;

  //bool operator== (CLatticeSearchHyp &Hyp);

  static long int s_NumberOfObjects; 
};


// stream operator for CLatticeSearchHyp
//ostream &operator << ( ostream &, const CLatticeSearchHyp );

// compare, if two CLatticeSearchHyps are equal 
bool operator == ( CLatticeSearchHyp &Hyp1, CLatticeSearchHyp &Hyp2 );

//! Multimap of score to edge, used to prune translation edges
//typedef multimap< float, CLatticeEdgeMS*, less< float > > ScoreToEdgeMultimap;
typedef multimap< float, CLatticeEdgeMS*, greater< float > > ScoreToEdgeMultimap;


class CLatticeSearch
{
public:
  CLatticeSearch();

  // normally, we work with word indices for efficiency
  // the vocabulary is required to convert between index and string
  void SetVocabulary( const CVocabulary *pVocabulary );

  // set the lattice for which transducers are applied
  void SetLattice( CLattice *pLattice );

  // attach empty search hypothesis to each lattice node
  void InitSearch();

  // run through lattice, find sequences from transducer and insert new
  // edges into the lattice
  // return: number of found phrases
  int ApplyTransducer( const CTransducerAccess *pTransducer );

  // added by Joy, joy@cs.cmu.edu
  // Mar 26, 2003
  // Simply apply the transducer on the word level
  // do not consider the Append to hierarchical
  // this will run faster than the standard version of ApplyTransducer
  // since it do not go over all the in-edges of the current node
  // just check the hyps from the previous node
  int ApplyTransducerSimple( const CTransducerAccess *pTransducer );

  
  // for error tolerant match: totally allowed errors
  void SetMaxCost( double MaxCost );
  // or length dependent, i.e. proportional to length of path in Whg
  void SetMaxCostPerWord( double ErrorsPerWord );

  void SetScalingFactors( CNumericalArray< float >* pScalingFactors );
  
  void SetLatticePruning( float LatticePruneBeamSize, 
                          float LatticePruneBeamSizeEpsilon,
                          int LatticePruneMaxTranslations );

private:

  //ARV 04/28 - versions of all the functions below need to exist for both the MS, and the regular versions
  // in time we should decrecate the versions which cannot handle multiple scores
  //apd 12/15/06 - added lattice pruning 
  void CreateNewEdgesMS( CLatticeSearchHyp* pPhraseHyp, CLatticeNode* pEndNode, 
                         CMatrix< ScoreToEdgeMultimap* > &BestEdgeScoresMatrix );
  
  float m_LatticePruneBeamSize;
  int   m_LatticePruneMaxTranslations;
  float m_LatticePruneBeamSizeEpsilon;
  
  void ReconstructPathMS( const CLatticeSearchHyp *pHyp, CLatticeNode *pNode,
			  CArray < CLatticeEdgeMS * > *pWhgEdges );
  void WriteTargetStringMS( ostringstream & Stream, CSentence *pSentenceE, 
			    CArray< CLatticeEdgeMS * > *pWhgEdges,
			    double &NewScore );
  
  // create new edges for all translations referenced by hypothesis pHyp and 
  // append them to list NewEdges
  void CreateNewEdges( CLatticeSearchHyp *pHyp, CLatticeNode *pNode,
		       CArray< CLatticeEdgeTrans * > &NewEdges );

  // trace back over the path which lead to hypothesis pHyp
  // and collect edges in WhgEdges
  void ReconstructPath( const CLatticeSearchHyp *pHyp, CLatticeNode *pNode,
			CArray < CLatticeEdgeTrans * > *pWhgEdges );

  // ARV compatibility, changed ostringstream
  void WriteTargetString( ostringstream & Stream, CSentence *pSentenceE, 
			  CArray< CLatticeEdgeTrans * > *pWhgEdges,
			  double &NewScore );

  // pointer to the search patterns, which are stored in a tree
  const CTransducerAccess *m_pTransducerAccess;
  
  //! the scaling factors for the different models
  CNumericalArray< float >* m_pScalingFactors;

  // pointer to the lattice, in which the patterns are searched
  CLattice *m_pLattice;

  // pointer to source vocabulary
  const CVocabulary *m_pVocabulary;

  // 2-dim array to store for each node all partial search hypotheses
  CArray2D< CLatticeSearchHyp > m_Hyps;


  // threshold for error tolerant match: prune hypotheses which exeed threshhold
  double m_MaxCost; // absolut error threshhold
  double m_ErrorsPerWord; // length dependent error threshold
  BOOL m_LengthDependentCosts; // flag which indicates that length dependent
                               // error threshold is active 

  uint m_NumberOfActiveHypothesis;

  // for statistics
  uint m_NumberOfStartedPhrases;
  uint m_NumberOfExpandedPhrases;
  uint m_NumberOfNIBMoves;
  uint m_NumberOfDeletedHypotheses;

  uint m_NumberOfPhrasesFromPhraseTree;
  uint m_NumberOfCompletePhrases;


};


#endif


