#ifndef __PHRASE_ALIGN__
#define __PHRASE_ALIGN__

// find optimal target phrase for given source phrase in 
// a sentence pair

#include <SortedArray.hh>
#include <Text.hh>
#include <Lexicon.hh>

#ifndef MAX_SENTENCE_LENGTH
#define MAX_SENTENCE_LENGTH 128
#endif

class CPhraseAligner
{
public:
  CPhraseAligner();
  virtual ~CPhraseAligner();

  // the lexicons to be used for calculating the alignment score
  void Init( CLexicon *pLexiconSourceTarget, CLexicon *pLexiconTargetSource );

  // use alignment probabilities from both directions?
  void SetAlignBidirectional( bool AlignBidirectional = true );

  // factors for phrase alignment:
  // StretchFactor: expected TargetLength = StretchFactor * SourceLength
  // FactorShorter: TargetLength >= FactorShorter * LengthSource
  // FactorLonger:  TargetLength <= FactorLonger  * LengthSource
  void SetStretchFactor( float StretchFactor, float m_FactorShorter, float m_FactorLonger );

  // interpolation factor when aligning in both directions
  void SetInterpolationFactor( float Lambda );

  // find best target phrase for source phrase in a sentence pair
  virtual bool AlignPhrase( CSentence *pSourceSentence, CSentence *pTargetSentence,
					int StartSource, int EndSource,
					CNumericalArray< int > & Min_i, CNumericalArray< int > & Max_i,
					CNumericalArray< double > & Scores );


protected:
  //float CalcAlignLogProbSourceTarget( int j1, int j2, int i1, int i2 );
  //float CalcAlignLogProbTargetSource( int j1, int j2, int i1, int i2 );

  void Renormalize();

  // length of sentences
  int m_J, m_I;

  // the lexica
  CLexicon *m_pLexiconSourceTarget;
  CLexicon *m_pLexiconTargetSource;

  // matrices to store the lexical probabilities
  // we use static array, i.e. can deal only with sentences up to MAX_SENTENCE_LENGTH
  float m_ProbsSourceTarget[ MAX_SENTENCE_LENGTH ][ MAX_SENTENCE_LENGTH ];
  float m_ProbsTargetSource[ MAX_SENTENCE_LENGTH ][ MAX_SENTENCE_LENGTH ];

  // should probabilities be renormalized?
  bool m_Renormalize;

  // use alignment probabilities from both directions?
  bool m_AlignBidirectional;

  // interpolation factor ( 1 - l ) p( f | e ) + l p( e | f )
  float m_Lambda;

  // general stretch factor: length target corpus / length source corpus
  float m_StretchFactor;

  // factor, by which target can be shorter or longer than expected length
  float m_FactorShorter;
  float m_FactorLonger;
};



inline void 
CPhraseAligner::SetAlignBidirectional( bool AlignBidirectional )
{
  m_AlignBidirectional = AlignBidirectional;
}


inline void 
CPhraseAligner::SetStretchFactor( float StretchFactor, float FactorShorter, float FactorLonger )
{
  StretchFactor = m_StretchFactor;
  FactorShorter = m_FactorShorter;
  FactorLonger = m_FactorLonger;
}


inline void
CPhraseAligner::SetInterpolationFactor( float Lambda )
{
  m_Lambda = Lambda;
}




// =============================================================================
class CPhraseAlignerIBM1:public CPhraseAligner
{
public:
  CPhraseAlignerIBM1();
  virtual ~CPhraseAlignerIBM1();

  // the lexicons to be used for calculating the alignment score
  //void Init( CLexicon *pLexiconSourceTarget, CLexicon *pLexiconTargetSource );

  // use alignment probabilities from both directions?
  //void SetAlignBidirectional( bool AlignBidirectional = true );

  // interpolation factor when aligning in both directions
  //void SetInterpolationFactor( float Lambda );

  // find best target phrase for source phrase in a sentence pair
  virtual bool AlignPhrase( CSentence *pSourceSentence, CSentence *pTargetSentence,
							int StartSource, int EndSource,
							CNumericalArray< int > & Min_i, CNumericalArray< int > & Max_i,
							CNumericalArray< double > & Scores );


private:
  float CalcAlignLogProbSourceTarget( int j1, int j2, int i1, int i2 );
  float CalcAlignLogProbTargetSource( int j1, int j2, int i1, int i2 );

};





// ========================================================================================
class CPhraseAlignerHMM:public CPhraseAligner
{
public:
  CPhraseAlignerHMM();
  virtual ~CPhraseAlignerHMM();

  // the jump probabilities to be used for calculating the alignment score
  //void Init( CAlignmentProbRel *pAlignmentProb );

  virtual bool AlignPhrase( CSentence *pSourceSentence, CSentence *pTargetSentence,
							int StartSource, int EndSource,
							CNumericalArray< int > & Min_i, CNumericalArray< int > & Max_i,
							CNumericalArray< double > & Scores );


private:
  float CalcAlignLogProbSourceTarget( int j1, int j2, int i1, int i2 );
  float CalcAlignLogProbTargetSource( int j1, int j2, int i1, int i2 );

};



#endif
