#ifndef __PHRASE_ALIGN_MS__
#define __PHRASE_ALIGN_MS__

// find optimal target phrase for given source phrase in 
// a sentence pair

#include <SortedArray.hh>
#include <Text.hh>
#include <Lexicon.hh>
#include <Trace.hh>
#include <PhraseAlignProbCalc.hh>

#ifndef MAX_SENTENCE_LENGTH
#define MAX_SENTENCE_LENGTH 201
#endif

#define DEBUG 1

class CPhraseAlignerMS;
class CPhraseAlignerMSIBM1;


class CPhraseAlignItemMS
{
public:
  CPhraseAlignItemMS();
  void Init();

  int   SentenceNumber  ();
  short StartPosInSource();
  short EndPosInSource  ();
  short StartPosInTarget();
  short EndPosInTarget  ();

  void  SetSentenceNumber  ( int SentenceNumber   );
  short SetStartPosInSource( int StartPosInSource );
  short SetEndPosInSource  ( int EndPosInSource   );
  short SetStartPosInTarget( int StartPosInTarget );
  short SetEndPosInTarget  ( int EndPosInTerget   );

  // source and target phrase in word IDs and string
  CNumericalArray< CWordIndex > &SourceWordIDs();
  CString SourcePhrase() const;
  void SetSourcePhrase( CString SourcePhrase );

  CNumericalArray< CWordIndex > &TargetWordIDs();
  CString TargetPhrase() const;
  void SetTargetPhrase( CString TargetPhrase );

  const CNumericalArray< float > & Scores() const;
  const CNumericalArray< float > & ScaledScores() const;

  float Score() const;
  void SetScore( float Score );

  float ScaledScore() const;

  // assign operators
   CPhraseAlignItemMS & operator=( const CPhraseAlignItemMS & Item );

  // these operators compare the translation, i.e. the TargetWordIDs
  bool operator==( const CPhraseAlignItemMS & Item ) const;
  bool operator<( const CPhraseAlignItemMS & Item ) const;
  bool operator>( const CPhraseAlignItemMS & Item ) const;

  friend ostream& operator<<(ostream& OutStream, const CPhraseAlignItemMS& PhraseAlignItem );

protected:
  friend class CPhraseAlignerMS;
  friend class CPhraseAlignerMSTag;
  friend class CPhraseAlignerMSTag222;
  friend class CPhraseSearchMS;

  // sentence number and start end end position (index) of phrase pair
  // in source and target sentence
  int   m_SentenceNumber;
  short m_StartPosInSource;
  short m_EndPosInSource;
  short m_StartPosInTarget;
  short m_EndPosInTarget;

  // source and target phrase in word IDs and string
  CNumericalArray< CWordIndex > m_SourceWordIDs;
  CString m_SourcePhrase;
  CNumericalArray< CWordIndex > m_TargetWordIDs;
  CString m_TargetPhrase;

  // flag for pruning
  bool m_Keep;

  //  float m_ScoreAdditive;
  //float m_ScoreMultiplicative;

  //different scores
  // total scores in m_Scores[0], i.e. combination of individual scores
  CNumericalArray< float > m_Scores;
  //Some more Combined Scores for Pruning
  CNumericalArray< float > m_CombinedScores;

  // now all scores scaled, i.e. multiplied by individual scaling factor
  // total scaled score in m_ScaledScores[0]
  CNumericalArray< float > m_ScaledScores;
  //Some more Combined Scores for Pruning
  CNumericalArray< float > m_CombinedScaledScores;

  // count how often one translation has been seen
  int m_TranslationCount;
};



//=======================================================
class CPhraseAlignerMS
{
public:
  CPhraseAlignerMS();
  virtual ~CPhraseAlignerMS();

  // the lexicons to be used for calculating the alignment score
  void Init( CLexicon *pLexiconSourceTarget, CLexicon *pLexiconTargetSource, CPhraseAlignProbCalc *pCalc );
  void Init( CLexicon *pLexiconSourceTarget, CLexicon *pLexiconTargetSource );

  void SetDisplayScaledScores( bool DisplayScaledScores = true );


  // use alignment probabilities from both directions?
  // deprecated: void SetAlignBidirectional( bool AlignBidirectional );

  // use empty word
  void SetUseEmptyWord( bool UseEmptyWord );

  // find center of target phrase using sum or max of probabilities p( f | . )
  enum { FINDCENTER_MAX = 0, FINDCENTER_SUM, FINDCENTER_BLOCK };
  // deprecated: void SetFindCenterWithSum( bool UseSum = True );
  void SetFindCenterMethod( int PhraseAlignFindCenterMethod = FINDCENTER_BLOCK );

  // pruning strategy
  enum { PRUNE_MULTIPLICATIVE = 0, PRUNE_ADDITIVE, PRUNE_ADDITIVE_PERWORD };

  void SetPruningStrategy( int PruningStrategy = PRUNE_ADDITIVE );

  void SetBeam( float BeamFactor, int MaxTranslations );

  // factors for phrase alignment:
  // StretchFactor: expected TargetLength = StretchFactor * SourceLength
  // FactorShorter: TargetLength >= FactorShorter * LengthSource
  // FactorLonger:  TargetLength <= FactorLonger  * LengthSource
  void SetStretchFactors( float StretchFactor, float m_FactorShorter, float m_FactorLonger );
  
  void SetStretchFactors( float StretchFactor, float FactorShorter, float FactorLonger,
                          float AddConstShorter, float AddConstLonger );

  // interpolation factor when aligning in both directions
  // deprecated: void SetInterpolationFactor( float Lambda );

  // scaling factors for the different scores calculated for the alginment
  CNumericalArray<float> &GetScalingFactors();
  void SetScalingFactors( CNumericalArray<float> ScalingFactors );
  
  int GetNumberOfScores();
  int GetNumberOfCombinedScores();

  // find best target phrase for source phrase in a sentence pair
  // virtual bool AlignPhrase( CSentence *pSourceSentence, CSentence *pTargetSentence,
  //					int StartSource, int EndSource,
  //					CNumericalArray< int > & Min_i, CNumericalArray< int > & Max_i,
  //					CNumericalArray< float > & Scores );

  virtual bool AlignPhrase( CSentence *pSourceSentence, CSentence *pTargetSentence,
							int StartSource, int EndSource,
							CSortedArray< CPhraseAlignItemMS > & PhraseAlignItems );

  bool FindBestPhrasePair( CSentence *pSourceSentence, CSentence *pTargetSentence,
                           int MinPhraseLength, int MaxPhraseLength,
                           CPhraseAlignItemMS & PhraseAlignItem );

  bool FindAllPhrasePair( CSentence *pSourceSentence, CSentence *pTargetSentence,
                          int MinPhraseLength, int MaxPhraseLength,
                          CArray< CPhraseAlignItemMS > & PhraseAlignItems );

  // align only some of the phrases in the source sentence
  // the PhraseAlignItems specify which ones
  // this can be used to find phrase pairs according to frequency, length, etc
  bool FindSomePhrasePairs( CSentence *pSourceSentence, CSentence *pTargetSentence,
                            CArray< CPhraseAlignItemMS > & PhraseAlignItems );

  // as before, but multiple translations within one sentence
  bool FindSomePhrasePairsMultiTrans( CSentence *pSourceSentence, CSentence *pTargetSentence,
                                      CArray< CPhraseAlignItemMS > & PhraseAlignItems );



protected:

  //For Target->Source Probabilis, you must use TargetSentence as SourceSentence and the other way around!!!!
  // put lexicon probability into matrix
  //Probs is an array of size MAX_SENTENCE_LENGTH * MAX_SENTENCE_LENGTH;
  void GetProbabilities( CSentence *pSourceSentence, CSentence *pTargetSentence, CLexicon *pLexicon, float *Probs );

  // and renormalize them
  //Probs and RenProbs are an array of size MAX_SENTENCE_LENGTH * MAX_SENTENCE_LENGTH;
  void Renormalize( int SourceLength, int TargetLength,
					float *Probs, float *RenProbs );
  
  // find 'center' of target phrase for source phrase
  // by only using the maximum probability
  float FindCenterMax( int StartSource, int EndSource );
  // by calculating the mean value
  float FindCenterSum( int StartSource, int EndSource );
  // by calculating the phrase alignment probability p( f | e )
  float FindCenterBlock( int StartSource, int EndSource );

  //! calculate minimal target phrase length based on source phrase length and stretch factors
  int MinTargetPhraseLength( int LengthSourcePhrase );
  //! calculate maximum target phrase length based on source phrase length and stretch factors
  int MaxTargetPhraseLength( int LengthSourcePhrase );



  // calculate all scores and store them in the AlignItem
  float CalcAllAlignLogProbs( int StartSource, int EndSource, int StartTarget, int EndTarget, CPhraseAlignItemMS &AlignItem );
  float CalcAllAlignLogProbs( CPhraseAlignItemMS &AlignItem );

  // from raw scores, calculate scaled and combined scores
  void AddScaledAndCombinedScores( CPhraseAlignItemMS & AlignItem );

  // adding word IDs and strings is expensive, do only when needed
  void AddWordIDsAndStrings( CPhraseAlignItemMS & AlignItem );

  bool ItemIsGoodEnough( CPhraseAlignItemMS &AlignItem );

  bool ItemWithinBeamOfBestItem( CPhraseAlignItemMS &AlignItem, CPhraseAlignItemMS &BestItems );

  //Display Probabilities
  void DisplayProbs( float* SourceTarget, float* TargetSource ); 
  void DisplaySourceSentence( int maxLen );
  int GetMaxWordLength();



protected:  // the member variable

  // the sentences
  CSentence *m_pSourceSentence;
  CSentence *m_pTargetSentence;

  // length of sentences
  int m_J, m_I;

  // the lexica
  CLexicon *m_pLexiconSourceTarget;
  CLexicon *m_pLexiconTargetSource;

  // matrices to store the lexical probabilities
  // we use static array, i.e. can deal only with sentences up to MAX_SENTENCE_LENGTH
  float m_ProbsSourceTarget[ MAX_SENTENCE_LENGTH ][ MAX_SENTENCE_LENGTH ];
  float m_ProbsTargetSource[ MAX_SENTENCE_LENGTH ][ MAX_SENTENCE_LENGTH ];

  // renormalized probs
  float m_RenProbsSourceTarget[ MAX_SENTENCE_LENGTH ][ MAX_SENTENCE_LENGTH ];
  float m_RenProbsTargetSource[ MAX_SENTENCE_LENGTH ][ MAX_SENTENCE_LENGTH ];

  // should probabilities be renormalized?
  bool m_Renormalize;

  // should empty word be used?
  bool m_UseEmptyWord;

  // use alignment probabilities from both directions?
  bool m_AlignBidirectional;

  // interpolation factor ( 1 - l ) p( f | e ) + l p( e | f )
  float m_Lambda;

  // how to predict center of target phrase
  // deprecated bool m_FindCenterWithSum;
  int m_FindCenterMethod;

  //number of Scores
  int m_numberOfScores;
  //number of additional Scores by combining the elemntary Scores (without the allover Score Scores[0])
  int m_numberOfCombinedScores;


  CNumericalArray<float> m_ScalingFactors;

  // general stretch factor: length target corpus / length source corpus
  float m_StretchFactor;

  // factor, by which target can be shorter or longer than expected length, default is 3/2 and 2/3
  float m_FactorShorter;
  float m_FactorLonger;
  float m_AddConstShorter;
  float m_AddConstLonger;

  // for pruning
  int m_PruningStrategy;
  float m_BeamFactor;
  int m_MaxTranslations;

  // this object calculats phrase scores from table
  // e.g. with IBM1 or HMM
  CPhraseAlignProbCalc * m_pCalc;

};


inline void
CPhraseAlignItemMS::SetSentenceNumber  ( int SentenceNumber   )
{
  m_SentenceNumber = SentenceNumber;
}



inline short
CPhraseAlignItemMS::SetStartPosInSource( int StartPosInSource )
{
  m_StartPosInSource = StartPosInSource;
}


inline short
CPhraseAlignItemMS::SetEndPosInSource  ( int EndPosInSource   )
{
  m_EndPosInSource = EndPosInSource;
}



inline short
CPhraseAlignItemMS::SetStartPosInTarget( int StartPosInTarget )
{
  m_StartPosInTarget = StartPosInTarget;
}



inline short
CPhraseAlignItemMS::SetEndPosInTarget  ( int EndPosInTarget   )
{
  m_EndPosInTarget = EndPosInTarget;
}


inline int
CPhraseAlignItemMS::SentenceNumber()
{
  return m_SentenceNumber;
}


inline short
CPhraseAlignItemMS::StartPosInSource()
{
  return m_StartPosInSource;
}


inline short
CPhraseAlignItemMS::EndPosInSource()
{
  return m_EndPosInSource;
}


inline short
CPhraseAlignItemMS::StartPosInTarget()
{
  return m_StartPosInTarget;
}


inline short
CPhraseAlignItemMS::EndPosInTarget()
{
  return m_EndPosInTarget;
}



inline CNumericalArray< CWordIndex > &
CPhraseAlignItemMS::TargetWordIDs()
{
  return m_TargetWordIDs;
}


inline const CNumericalArray< float > &
CPhraseAlignItemMS::Scores() const
{
  return m_Scores;
}


inline const CNumericalArray< float > &
CPhraseAlignItemMS::ScaledScores() const
{
  return m_ScaledScores;
}



inline float 
CPhraseAlignItemMS::ScaledScore() const
{
  return m_ScaledScores[ 0 ];
}


inline float 
CPhraseAlignItemMS::Score() const

{
  return m_Scores[ 0 ];
}


inline void
CPhraseAlignItemMS::SetScore( float Score )
{
  //  m_ScoreAdditive = Score;
  m_Scores[ 0 ] = Score;
}


inline CString
CPhraseAlignItemMS::SourcePhrase() const
{
  return m_SourcePhrase;
}


inline CString
CPhraseAlignItemMS::TargetPhrase() const
{
  return m_TargetPhrase;
}


inline void
CPhraseAlignItemMS::SetTargetPhrase( CString TargetPhrase )
{
  m_TargetPhrase = TargetPhrase;
}

#endif
