/*
  PhraseAlignMS
  Find optimal target phrase for given source phrases in a sentence pair.
*/


#ifndef __PHRASE_ALIGN_MS__
#define __PHRASE_ALIGN_MS__


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <fortify.h>

#include <Trace.hh>
#include <SortedArray.hh>
#include <Text.hh>
#include <Lexicon.hh>
#include <PhraseAlignProbCalc.hh>

#ifdef DEBUG
#undef DEBUG
#endif

#define DEBUG 0


// imports
// from PhraseAlignProbCalc: define MAX_SENTENCE_LENGTH 201


// forward declarations
class CPhraseAlignItemMS;
class CPhraseAlignerMS;


//===============================================================================
/*
  CPhraseAlignItemMS
  Holds the relavant information for one aligned source - target phrase pair.
  I.e. in which sentence and at which positions, different scores, the phrases
  themselves as arrays of word IDs and as strings.
*/

class CPhraseAlignItemMS
{
public:
  // constructor
  CPhraseAlignItemMS();

  // init function to set all member variable to reasonable values
  // can also be used to reset values to defaults
  void Init();

  // assign operators
   CPhraseAlignItemMS & operator=( const CPhraseAlignItemMS & Item );

  // access functions
  int   SentenceNumber();
  void  SetSentenceNumber( int SentenceNumber );
  short StartPosInSource();
  short EndPosInSource();
  short StartPosInTarget();
  short EndPosInTarget();

  // access functions for source and target phrase as arrays of word IDs and as strings
  CNumericalArray< CWordIndex > &SourceWordIDs();
  CNumericalArray< CWordIndex > &TargetWordIDs();

  CString SourcePhrase() const;
  CString TargetPhrase() const;

  void SetSourcePhrase( CString SourcePhrase );
  void SetTargetPhrase( CString TargetPhrase );

  // the different alignment scores, as calculated
  const CNumericalArray< float > & Scores() const;

  // and scaled, i.e. multiplied by scaling factores
  const CNumericalArray< float > & ScaledScores() const;

  // the total score as sum of all individual scores
  float Score() const;
  void SetScore( float Score );  // ??? where is this used

  // and the total score a linear interpolation of individual scores
  float ScaledScore() const;

  // these operators compare the translation, i.e. the TargetWordIDs
  bool operator==( const CPhraseAlignItemMS & Item ) const;
  bool operator<( const CPhraseAlignItemMS & Item ) const;
  bool operator>( const CPhraseAlignItemMS & Item ) const;

  friend ostream& operator<<(ostream& OutStream, const CPhraseAlignItemMS& PhraseAlignItem );


protected:
  // these classes have direct access to the member variable of the CPhraseAlignItemMS objects
  friend class CPhraseAlignerMS;
  friend class CPhraseAlignerMS2;
  friend class CPhraseAlignerMSTag;
  //friend class CPhraseAlignerMSTag222;
  friend class CPhraseSearchMS;

  // sentence number and start end end position (index) of phrase pair
  // in source and target sentence
  int   m_SentenceNumber;
  short m_StartPosInSource;
  short m_EndPosInSource;
  short m_StartPosInTarget;
  short m_EndPosInTarget;

  // source and target phrase in word IDs and string
  CNumericalArray< CWordIndex > m_SourceWordIDs;
  CNumericalArray< CWordIndex > m_TargetWordIDs;
  CString m_SourcePhrase;
  CString m_TargetPhrase;

  // flag used in pruning
  bool m_Keep;

  float m_TotalScore;
  float m_TotalScaledScore;

  //different scores
  // total score, i.e. sum of individual scores, is in m_Scores[0]
  CNumericalArray< float > m_Scores;

  //Some more Combined Scores for Pruning
  CNumericalArray< float > m_CombinedScores;

  // now all scores scaled, i.e. multiplied by individual scaling factor
  // total scaled score, i.e. weighted sum of individual scores, is in m_ScaledScores[0]
  CNumericalArray< float > m_ScaledScores;

  // some more combined scores for pruning
  CNumericalArray< float > m_CombinedScaledScores;

  // count how often one translation has been seen
  int m_TranslationCount;
};



//===============================================================================
/*
  CPhraseAlignerMS
  Provides the functionality to align phrase within one sentence pair.
  Can align one specific phrase, a number of given phrases, or all phrases,
  can also find the overall best phrase pair in the given sentence pair.
  The actual calculation of alignment probabilities is done in CPhraseAlignProbCalc.
  Here, the preparation of the matrices with the lexical probabilities is done,
  and the looping over source phrases and target phrase candidates is done.
*/

class CPhraseAlignerMS
{
public:
  CPhraseAlignerMS();
  CPhraseAlignerMS(bool useRelativeFeatures);
  virtual ~CPhraseAlignerMS();

  // initialize with the lexicons to be used for calculating the alignment score
  void Init( CLexicon *pLexiconSourceTarget, CLexicon *pLexiconTargetSource );

  // we have different ways to calculate the phrase alignments,
  // e.g. using IBM1 alignment or HMM alignment
  void Init( CLexicon *pLexiconSourceTarget, CLexicon *pLexiconTargetSource, CPhraseAlignProbCalc *pCalc );


  //jn outdated in .cc by sv

  // use alignment probabilities from both directions
  //void SetAlignBidirectional( bool AlignBidirectional );

  // use empty word
  void SetUseEmptyWord( bool UseEmptyWord );

 //use relative features
  bool UseRelativeFeatures();


  // to speed up the alignment first a rough estimation of the expected position of the target phrase is made.
  // MAX    for each source pos j: center = max_i p(f_j|e_i)
  // SUM    for each source pos j: center = sum_i p(f_j|e_i) * i
  // BLOCK  estimate target phrase length l, center = argmax_i prod_j1^j2 sum_i^i+l p(f_j|e_i)
  // NONE   don't find center but do full alignment for all target positions
  enum { MAX = 0, SUM, BLOCK, NONE };
  void SetFindCenterMethod( int PhraseAlignFindCenterMethod = MAX );
  // sv 060318 outdated: was used when we only had MAX and SUM
  // void SetFindCenterWithSum( bool UseSum = True );

  // some factors for phrase alignment:
  // StretchFactor: ExpectedTargetLength = StretchFactor * SourceLength
  // FactorShorter and FactorLonger:
  // FactorShorter * LengthSource <= ExpectedTargetLength <= FactorLonger  * LengthSource
  void SetStretchFactors( float StretchFactor, float m_FactorShorter, float m_FactorLonger );

  // scaling factors for the different scores calculated for the alginment
  CNumericalArray<float> &GetScalingFactors();
  void SetScalingFactors( CNumericalArray<float> ScalingFactors );
  
  // for each source phrase return all alternative which are within Beamfactor to the best one
  // i.e. if Cost( Candidate ) <= BeamFactor * Cost( BestCandidate ) then use
  // but only up to MaxTranslations, i.e. if too many candidates are within beam, then
  // restrict to MaxTranslations
  void SetBeam( float BeamFactor, int MaxTranslations );

  // number of scores calculated (more scores may be added in the future
  inline int GetNumberOfScores() { return m_numberOfScores;};

  // we combine for example the f->e and e->f scores, other combination are possible and 
  // may be added in the future
  inline int GetNumberOfCombinedScores(){return m_numberOfCombinedScores;};

  //number of scores added in Phrase Search to describe the relative frequences
  inline int GetNumberOfRelativeScores() {return m_numberOfRelativeScores;};


  // align the source phrase given by StartInSource and EndInSource
  // multiple alignments are possible and the target phrases are returned sorted 
  // according to their total phrase alignment cost
  // todo: pruning on this level, i.e. return only target phrases for which the 
  // total alignment cost is within a given range to the best target phrase
  virtual bool AlignPhrase( CSentence *pSourceSentence, CSentence *pTargetSentence,
							int StartInSource, int EndInSource,
							CSortedArray< CPhraseAlignItemMS > & PhraseAlignItems );

  // align all source phrases whoses length is between MinPhraseLength and MaxPhraseLength
  // but then return only the best one (i.e. no alternative candidates considered)
  bool FindBestPhrasePair( CSentence *pSourceSentence, CSentence *pTargetSentence,
						   int MinPhraseLength, int MaxPhraseLength,
						   CPhraseAlignItemMS & PhraseAlignItem );

  // align all source phrases whoses length is between MinPhraseLength and MaxPhraseLength
  // currently this is going over the source phrases, extend to going over target phrases as well
  // currently only the best for each source phrase is returned, extend to return all within beam
  bool FindAllPhrasePairs( CSentence *pSourceSentence, CSentence *pTargetSentence,
						   int MinPhraseLength, int MaxPhraseLength,
						   CArray< CPhraseAlignItemMS > & PhraseAlignItems );
  // align all source phrases whoses length is between MinPhraseLength and MaxPhraseLength
  // currently this is going over the source phrases, extend to going over target phrases as well
  // return only phrase pair which are resonable goodm
  // currently reurns pairs, if second best is not in beam
  bool FindGoodPhrasePairs( CSentence *pSourceSentence, CSentence *pTargetSentence,
						   int MinPhraseLength, int MaxPhraseLength,
						   CArray< CPhraseAlignItemMS > & PhraseAlignItems );



protected:

  //For Target->Source Probabilis, you must use TargetSentence as SourceSentence and the other way around!!!!
  // put lexicon probability into matrix
  //Probs is an array of size MAX_SENTENCE_LENGTH * MAX_SENTENCE_LENGTH;
  void GetProbabilities( CSentence *pSourceSentence, CSentence *pTargetSentence, CLexicon *pLexicon, float *Probs );

  // and renormalize them
  //Probs and RenProbs are an array of size MAX_SENTENCE_LENGTH * MAX_SENTENCE_LENGTH;
  void Renormalize( int SourceLength, int TargetLength,
					float *Probs, float *RenProbs );
  
  // find 'center' of target phrase for source phrase
  // by only using the maximum probability
  float FindCenterMax( int StartSource, int EndSource );
  // by calculating the mean value
  float FindCenterSum( int StartSource, int EndSource );
  // by calculating the phrase alignment probability p( f | e )
  float FindCenterBlock( int StartSource, int EndSource );


  // calculate all scores and store them in the AlignItem
  float CalcAllAlignLogProbs( int StartSource, int EndSource, int StartTarget, int EndTarget,
							  CPhraseAlignItemMS &AlignItem );

  // from raw scores, calculate scaled and combined scores
  void AddScaledAndCombinedScores( CPhraseAlignItemMS & AlignItem );

  // adding word IDs and strings is expensive, do only when needed
  void AddWordIDsAndStrings( CPhraseAlignItemMS & AlignItem );


  //Display Probabilities
  void DisplayProbs( float* SourceTarget, float* TargetSource ); 
  void DisplaySourceSentence( int maxLen );
  int GetMaxWordLength();

  // -------------------------------------------------------------
  // member variables
  // -------------------------------------------------------------

  // this object calculats phrase scores from table
  // e.g. with IBM1 or HMM
  CPhraseAlignProbCalc * m_pCalc;

  // the sentences
  CSentence *m_pSourceSentence;
  CSentence *m_pTargetSentence;

  // length of sentences
  int m_J, m_I;

  // the lexica
  CLexicon *m_pLexiconSourceTarget;
  CLexicon *m_pLexiconTargetSource;

  // matrices to store the lexical probabilities
  // for speed we use static array, i.e. can deal only with sentences up to MAX_SENTENCE_LENGTH
  float m_ProbsSourceTarget[ MAX_SENTENCE_LENGTH ][ MAX_SENTENCE_LENGTH ];
  float m_ProbsTargetSource[ MAX_SENTENCE_LENGTH ][ MAX_SENTENCE_LENGTH ];

  // renormalized probs
  float m_RenProbsSourceTarget[ MAX_SENTENCE_LENGTH ][ MAX_SENTENCE_LENGTH ];
  float m_RenProbsTargetSource[ MAX_SENTENCE_LENGTH ][ MAX_SENTENCE_LENGTH ];

  // should probabilities be renormalized?
  // sv 060318 outdated: we use unnormalized and renormalized scores
  // bool m_Renormalize;

  // should empty word be used?
  bool m_UseEmptyWord;

  //use relative features
  bool m_UseRelativeFeatures;

  // use alignment probabilities from both directions?
  // sv 060318 outdated: can be done through scaling factors
  // bool m_AlignBidirectional;

  // how to predict center of target phrase
  // sv 060318 outdated: have now more ways to find the center of the target phrase
  // bool m_FindCenterWithSum;

  // which method should be used to estimate the location of the target phrase
  int m_FindCenterMethod;

  //number of Scores
  int m_numberOfScores;

  //number of additional Scores by combining the elemntary Scores (without the allover Score Scores[0])
  int m_numberOfCombinedScores;

  //number of Relative Scores
  int m_numberOfRelativeScores;

  // array to hold te different scaling factors (feature weights)
  // the model scores are stored in the align items
  CNumericalArray<float> m_ScalingFactors;

  // general stretch factor: should be length target corpus / length source corpus
  float m_StretchFactor;

  // factor, by which target can be shorter or longer than expected length, default is 3/2 and 2/3
  float m_FactorShorter;
  float m_FactorLonger;

  // for pruning
  float m_BeamFactor;
  int   m_MaxTranslations;

};



#endif
