#ifndef __PhraseSearchMS__
#define __PhraseSearchMS__


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <Hash.hh>

#include <PhraseAlignMS.hh>

#include <LatticeEdgeMS.hh>
#include <Lattice.hh>

#include <Text.hh>
#include <Alignment.hh>
#include <Lexicon.hh>
#include <NumericalArray.hh>
#include <_SuffixArraySearch.h>
#include <SentenceInfo.hh>
#include <PhrasePairStats.hh>


class CPhraseSearchMS
{
public:
  CPhraseSearchMS();
  ~CPhraseSearchMS();

  void Init( C_SuffixArraySearch *pSA, CText *pTargetText, CVocabulary *pVoc, CAlignment *pAlignment, CLexicon *pLexicon );

  void Init( C_SuffixArraySearch *pSA, CText *pSourceText, CText *pTargetText, CVocabulary *pVoc,
             CAlignment *pAlignment, CLexicon *pLexicon );

  void Init( C_SuffixArraySearch *pSA, CText *pSourceText, CText *pTargetText, CVocabulary *pVoc,
             CAlignment *pAlignment, CLexicon *pLexicon, CLexicon *pReverseLexicon, double LambdaLexicon );

  void LoadSentenceInfo( const char *pFileWithLogProbPerWord );

  void SetPhrasePairStats( CPhrasePairStats *pPhrasePairStats );

  void SetPhraseAligner( CPhraseAlignerMS *pPhraseAligner );

  void SetUseEmptyWord( bool UseEmptyWord = true );

  void SetBeam( float BeamFactor = 1.0, int MaxTranslations = 10 );
  void SetMaxAlignments( int PhraseAlignMaxAlignments = 100 );

//   void SetScalingFactors( float Lambda1FE, float Lambda1EF,
// 			  float Lambda2FE, float Lambda2EF,
// 			  float Lambda3FE, float Lambda3EF );

  //! accumulation mode
  typedef enum { MAX, SUM } TAccumulationMode;

  //! when the same translation is found in multiple sentences, take best score or average scores
  void SetAccumulationMode( TAccumulationMode AccumulationMode = MAX );

  //! a table into which information can be written about which phrases have already translations
  CMatrix< short > * GetExistingTranslationsTable();

  //! find for all phrases in pSentence translations and insert them into the lattice
  void FindTranslations( CSentence *pSentence, CLattice *pLattice );

  //! search for phrases from all sentences but not the forbidden one
  void FindTranslations( CSentence *pSentence, CLattice *pLattice, int ForbiddenSentence );
  //void SearchAlign( CSentence *pSentence, CLattice *pLattice );
  //void FindTranslations( CSentence *pSentence, CMatrix< CWordIndex > &Translations );

  void SearchAlign( CSentence *pSentence, CLattice *pLattice );
  
  //phrase search in training mode, used when translating a training
  //set, and the SA matches this same data
  //void SearchAlignTraining( const int SentenceNumber, CSentence *pSentence, CLattice *pLattice );

  void FindTranslations( CSentence *pSentence, CMatrix< CWordIndex > &Translations );

  //void Search( CSentence *pSentence, CLattice *pLattice );
  //void Search1( CSentence *pSentence, CLattice *pLattice );



private:  // private methods

  // these functions are used by FindTranslation( Sentence, Lattice )
  void FindOccurences( CSentence *pSentence );

  //! align a single phrase in one sentence pair
  void AlignPhrase( S_phraseLocationElement &PhraseLocation );

  //! add relative phrase frequencies (currently based on co-occurence counts)
  void AddRelativeFrequencies();

  //! create edges for the phrase pairs and insert them into the lattice
  void CreateEdges();

  //
  void PruneAdditive();
  void PruneMultiplicative();


  void FindTranslation( int SentenceID, int PosInSentence, CMatrix< CWordIndex > TargetWordIDs );

  void FindBestLocations( const vector< S_phraseLocationElement > &PhraseLocations,
                          vector< S_phraseLocationElement > &BestPhraseLocations );

  float PhraseTranslationLogProb( const CArray< CWordIndex > &Source, 
                                  const CArray< CWordIndex > &Target, 
                                  const CLexicon *pLexicon );
  /*
  void AddEdge( CLattice *pLattice, int From, int To, double LogProb,
                const CArray< CWordIndex > &SourceWordIDs,
				        const CArray< CWordIndex > &TargetWordIDs );
  */

  void AddEdge( CLattice *pLattice, int From, int To,
                double LogProb, CNumericalArray< float > Scores,
                const CArray< CWordIndex > &SourceWordIDs,
                const CArray< CWordIndex > &TargetWordIDs );

  void DisplayPhraseLocation(  const S_phraseLocationElement &PhraseLocations ) const;
  void DisplayPhraseLocations(  const vector< S_phraseLocationElement > &PhraseLocations ) const;


private:  // private member variable

  //! the suffix array which indexes the source corpus
  C_SuffixArraySearch *m_pSA;

  //! the source side of the bilingual corpus
  //  redundant, as this is also in the suffix array
  CText *m_pSourceText;

  //! the target side of the bilingual corpus
  CText *m_pTargetText;

  //! the vocabulary covering source and target words
  CVocabulary *m_pVoc;

  //! in case we ever want to use a give alignment
  CAlignment *m_pAlignment;

  //! the source->target lexicon
  CLexicon *m_pLexicon;

  //! the target->source lexicon
  CLexicon *m_pReverseLexicon;


  //! phrase pair stats to get phrase counts and co-occurence counts
  CPhrasePairStats *m_pPhrasePairStats;


  double m_LambdaLexicon;

  int m_ForbiddenSentence;

  bool m_UseEmptyWord;

  int   m_MaxAlignments;
  float m_BeamFactor;
  int   m_MaxTranslations;

  int   m_AccumulationMode;

  // scaling factors for the different scores
//   float m_Lambda1FE;
//   float m_Lambda1EF;
//   float m_Lambda2FE;
//   float m_Lambda2EF;
//   float m_Lambda3FE;
//   float m_Lambda3EF;

  //! the sentence for which we do the phrase alignment
  CSentence *m_pSentence;

  //! the lattice into which we insert the found phrase pairs as edges
  CLattice *m_pLattice;

  //?
  //CPhraseAlignItemMS PA;

  // number of translations from transducer are recorded here
  // for those segments we skip running the phrase alignment
  CMatrix< short > *m_pExistingTranslations;

  // for all phrases in the given sentence a list of occurences in the corpus
  vector< S_phraseLocationElement > m_PhraseLocations;
  //  CSortedArray< CPhraseAlignItemMS > *m_PhraseTranslations;

  //! this table stores all the phrase pairs for the entire sentence
  CMatrix< CSortedArray< CPhraseAlignItemMS > > *m_pPhraseTranslationTable;

  //! a pointer to the actual prhase aligner, which aligns one phrase in one sentence pair
  CPhraseAlignerMS *m_pPhraseAligner;

  //! sentence info object, which provides LogProbPerWord for each sentence pair
  CSentenceInfo *m_pSentenceInfo;
};



#endif
