#include <string.h>
/*
* LHash.h --
*	Hash table with linear search.
* 
* LHash<KeyT,DataT> implements Map<KeyT,DataT> as a hash table with
* open addressing, i.e., entries are stored at the computed hash index,
* or in the next free entry, which is located with linear search.
* Lookups, insertions, and deletions are all done in linear expected
* time, with a constant factor that depends on the occupancy rate of
* the table.  No space is wasted with pointers or bucket structures,
* but the size of the table is at least the number of entries rounded to
* the next power of 2.
*
* LHashIter<KeyT,DataT> performs iterations (in random key order) over
* the entries in a hash table.
*
* Copyright (c) 1995-1998 SRI International.  All Rights Reserved.
*
* @(#)$Header$
*
*/

#ifndef _LHash_h_
#define _LHash_h_

#include <assert.h>
#include "Map.h"

//#ifdef __GNUG__
//static
//#endif
extern int (*LHash_thisKeyCompare)();		/* used by LHashIter() to communicate
									* with compareIndex() */
//#ifdef __GNUG__
//static
//#endif

extern void *LHash_thisBody;			/* ditto */


/* 
* These constants are used below to pack the number of entries 
* and the total allocated size into a single word.
*/

const unsigned minHashBits = 3;		/* minimum no. bits for hashing
									* tables smaller than this use linear
									* search to save space */
const float fillRatio = 0.8;		/* fill ration at which the table is
									* expanded and rehashed */

/*
* Key Comparison functions
*/
#define hashSize(nbits) (1<<(nbits))	/* allocated size of data array */
#define hashMask(nbits) (~((~0)<<(nbits)))
/* bitmask used to compute hash
* code modulo maxEntries */
/*
* Compute the actual minimum size required for a given number of entries
*/
static inline unsigned
roundSize(unsigned size)
{
	if (size < hashSize(minHashBits)) {
		return size;
	} else {
		return (unsigned)((size + 1)/ fillRatio);
	}
}

#define LHASH_MAXBIT_NBITS		5
#define LHASH_MAXENTRY_NBITS		27

const unsigned LHash_maxBitLimit = ((1<<LHASH_MAXBIT_NBITS)-1);
const unsigned LHash_maxEntriesLimit = ((1<<LHASH_MAXENTRY_NBITS)-1);

template <class KeyT, class DataT> class LHash;		// forward declaration
template <class KeyT, class DataT> class LHashIter;	// forward declaration

#undef INSTANTIATE_LHASH
#define INSTANTIATE_LHASH(KeyT, DataT) \
	DataT *LHash<KeyT, DataT>::removedData = 0; \
	template class LHash<KeyT, DataT>; \
	template class LHashIter<KeyT, DataT>





#define BODY(b)	((LHashBody<KeyT,DataT> *)b)

template <class KeyT, class DataT>
class LHashBody
{
	friend class LHash<KeyT,DataT>;
	friend class LHashIter<KeyT,DataT>;

	unsigned maxBits:LHASH_MAXBIT_NBITS;	/* number of bits in hash code
											*  = log2 (maxEntries) */
	unsigned nEntries:LHASH_MAXENTRY_NBITS;	/* number of entries */

	MapEntry<KeyT,DataT> data[1];	/* hashed array of key-value pairs */
};

/*
* If we define the class as a subclass of Map<KeyT,DataT> we
* incur an extra word of memory per instance for the virtual function
* table pointer.
*/
template <class KeyT, class DataT>
class LHash
#ifdef USE_VIRTUAL
	: public Map<KeyT,DataT>
#endif
{
	friend class LHashIter<KeyT,DataT>;

public:
	LHash(unsigned size = 0): body(0)
	{
		if (size != 0) {
			/*
			* determine actual initial size
			*/
			alloc(roundSize(size));
		}
	}
	~LHash(){
		clear(0);
	}

	LHash(const LHash<KeyT,DataT> &source){
		std::cerr << "WARNING: LHash copy contructor called\n";
		body = source.body;
	}
	LHash<KeyT,DataT> &operator= (const LHash<KeyT,DataT> &source){
		std::cerr << "WARNING: LHash assignment operator called\n";
		body = source.body;
		return *this;
	}

	DataT *find(KeyT key, Boolean &foundP = _Map::foundP) const{
		unsigned index;

		if (foundP = locate(key, index)) {
			return &(BODY(body)->data[index].value);
		} else {
			return 0;
		}
	}
	KeyT getInternalKey(KeyT key, Boolean &foundP = _Map::foundP) const{
		unsigned index;
		static KeyT zeroKey;

		if (foundP = locate(key, index)) {
			return BODY(body)->data[index].key;
		} else {
			return zeroKey;
		}
	}
	DataT *insert(KeyT key, Boolean &foundP = _Map::foundP){
		unsigned index;

		assert(!(Map_noKeyP(key)));

		/*
		* Make sure there is room for at least one entry
		*/
		if (body == 0) {
			alloc(1);
		}

		if (foundP = locate(key, index)) {
			return &(BODY(body)->data[index].value);
		} else {
			unsigned maxEntries = hashSize(BODY(body)->maxBits);
			unsigned nEntries = BODY(body)->nEntries;

			/*
			* Rehash table if necessary
			*/
			unsigned minSize = roundSize(nEntries + 1);

			if (minSize > maxEntries) {
				LHashBody<KeyT,DataT> *oldBody = BODY(body);
				unsigned i;

				/*
				* Since LHash_maxEntriesLimit is a power of two minus 1
				* we need to check this only when the array is enlarged
				*/
				assert(nEntries < LHash_maxEntriesLimit);

				alloc(minSize);
				BODY(body)->nEntries = nEntries;

				if (BODY(body)->maxBits < minHashBits) {
					/*
					* Just copy old entries to new storage, no reindexing
					* required.
					*/
					memcpy(BODY(body)->data, oldBody->data,
						sizeof(oldBody->data[0]) * nEntries);
				} else {
					/*
					* Rehash
					*/
					for (i = 0; i < maxEntries; i++) {
						KeyT key = oldBody->data[i].key;

						if (! Map_noKeyP(key)) {
							(void)locate(key, index);
							memcpy(&(BODY(body)->data[index]), &(oldBody->data[i]),
								sizeof(oldBody->data[0]));
						}
					}
				}
				free(oldBody);

				/*
				* Entries have been moved, so have to re-locate key
				*/
				(void)locate(key, index);
			}

			BODY(body)->data[index].key = Map_copyKey(key);

			/*
			* Initialize data to zero, but also call constructors, if any
			*/
			memset(&(BODY(body)->data[index].value), 0,
				sizeof(BODY(body)->data[index].value));
			new (&(BODY(body)->data[index].value)) DataT;

			BODY(body)->nEntries++;

			return &(BODY(body)->data[index].value);
		}
	}
	DataT *remove(KeyT key, Boolean &foundP = _Map::foundP){
		unsigned index;

		/*
		* Allocate pseudo-static memory for removed objects (returned by
		* remove()). We cannot make this static because otherwise 
		* the destructor for DataT would be called on it at program exit.
		*/
		if (removedData == 0) {
			removedData = (DataT *)malloc(sizeof(DataT));
			assert(removedData != 0);
		}

		if (foundP = locate(key, index)) {
			Map_freeKey(BODY(body)->data[index].key);
			Map_noKey(BODY(body)->data[index].key);
			memcpy(removedData, &BODY(body)->data[index].value, sizeof(DataT));

			if (BODY(body)->maxBits < minHashBits) {
				/*
				* Linear search mode -- Just move all entries above the
				* the one removed to fill the gap.
				*/
				unsigned nEntries = BODY(body)->nEntries;

				memmove(&BODY(body)->data[index],
					&BODY(body)->data[index+1],
					(nEntries - index - 1) * sizeof(BODY(body)->data[0]));
				Map_noKey(BODY(body)->data[nEntries - 1].key);
			} else {
				/*
				* The entry after the one being deleted could actually
				* belong to a prior spot in the table, but was bounced forward due
				* to a collision.   The invariant used in lookup is that
				* all locations between the hash index and the actual index are
				* filled.  Hence we examine all entries between the deleted
				* position and the next free position as whether they need to
				* be moved backward.
				*/
				while (1) {
					unsigned newIndex;

					index = (index + 1) & hashMask(BODY(body)->maxBits);

					if (Map_noKeyP(BODY(body)->data[index].key)) {
						break;
					}

					/* 
					* If find returns false that means the deletion has
					* introduced a hole in the table that would prevent
					* us from finding the next entry. Luckily, find 
					* also tells us where the hole is.  We move the 
					* entry to its rightful place and continue filling
					* the hole created by this move.
					*/
					if (!locate(BODY(body)->data[index].key, newIndex)) {
						memcpy(&(BODY(body)->data[newIndex]),
							&(BODY(body)->data[index]),
							sizeof(BODY(body)->data[0]));
						Map_noKey(BODY(body)->data[index].key);
					}
				}
			}
			BODY(body)->nEntries--;
			return removedData;
		} else {
			return 0;
		}
	}
	void clear(unsigned size = 0){
		if (body) {
			unsigned maxEntries = hashSize(BODY(body)->maxBits);
			unsigned i;

			for (i = 0; i < maxEntries; i++) {
				if (! Map_noKeyP(BODY(body)->data[i].key)) {
					Map_freeKey(BODY(body)->data[i].key);
				}
			}
			free(body);
			body = 0;
		}
		if (size != 0) {
			alloc(roundSize(size));
		}
	}
	unsigned numEntries() const
	{ return body ? ((LHashBody<KeyT,DataT> *)body)->nEntries : 0; } ;
	void dump() const{
		if (body) {
			unsigned maxEntries = hashSize(BODY(body)->maxBits);
			unsigned i;

			for (i = 0; i < maxEntries; i++) {
				std::cerr << " " << i << ": ";
				if (Map_noKeyP(BODY(body)->data[i].key)) {
					std::cerr << "FREE";
				} else {
					std::cerr << BODY(body)->data[i].key
#ifdef DUMP_VALUES
						/* 
						* Only certain types can be printed.
						*/
						<< "->" << BODY(body)->data[i].value
#endif /* DUMP_VALUES */
						;
				}
			}
		} else {
			std::cerr << "EMPTY";
		}
		std::cerr << std::endl;
	}

	void memStats(MemStats &stats) const{
		stats.total += sizeof(*this);
		if (body) {
			unsigned maxEntries = hashSize(BODY(body)->maxBits);

			stats.total += sizeof(*BODY(body)) +
				sizeof(BODY(body)->data[0]) *
				(maxEntries - 1);
			stats.wasted += sizeof(BODY(body)->data[0]) *
				(maxEntries - BODY(body)->nEntries);
		}
	}

private:
	void *body;				/* handle to the above -- this keeps
							* the size of an empty table to
							* one word */
	Boolean locate(KeyT key, unsigned &index) const{
		assert(!Map_noKeyP(key));

		if (body) {
			unsigned maxBits = BODY(body)->maxBits;
			register MapEntry<KeyT,DataT> *data = BODY(body)->data;

			if (maxBits < minHashBits) {
				/*
				* Do a linear search
				*/
				unsigned nEntries = BODY(body)->nEntries;
				register unsigned i;

				for (i = 0; i < nEntries; i++) {
					if (LHash_equalKey(data[i].key, key)) {
						index = i;
						return true;
					}
				}
				index = i;
				return false;
			} else {
				/*
				* Do a hashed search
				*/
				unsigned hash = LHash_hashKey(key, maxBits);
				unsigned i; 

				for (i = hash; ; i = (i + 1) & hashMask(maxBits))
				{
					if (Map_noKeyP(data[i].key)) {
						index = i;
						return false;
					} else if (LHash_equalKey(data[i].key, key)) {
						index = i;
						return true;
					}
				}
			}
		} else {
			return false;
		}
	}
	/* locate key in *data */
	void alloc(unsigned size){
		unsigned maxBits, maxEntries;
		unsigned i;

		/*
		* round up to power of two
		*/
		maxBits = 0;
		while (hashSize(maxBits) < size) {
			assert(maxBits < LHash_maxBitLimit);
			maxBits++;
		}

		maxEntries = hashSize(maxBits);

		//std::cerr << "LHash::alloc: current " << (body ? BODY(body)->nEntries : 0)
		//	 << ", requested " << size 
		//	 << ", allocating " << maxEntries << " (" << maxBits << ")\n";

		body = (LHashBody<KeyT,DataT> *)malloc(sizeof(*BODY(body)) +
			(maxEntries - 1) * sizeof(BODY(body)->data[0]));
		assert(body != 0);

		BODY(body)->maxBits = maxBits;
		BODY(body)->nEntries = 0;

		for (i = 0; i < maxEntries; i++) {
			Map_noKey(BODY(body)->data[i].key);
		}
		//std::cerr << "memory for header = " <<
		//		((void *)&(BODY(body)->data[0]) - (void *)BODY(body)) << std::endl;
		//std::cerr << "memory per entry = " <<
		//		((void *)&(BODY(body)->data[3]) - (void *)&(BODY(body)->data[2])) << std::endl;
	}



	static DataT *removedData;		/* temporary buffer for removed item */
};

template <class KeyT, class DataT>
class LHashIter
{
	typedef int (*compFnType)(KeyT, KeyT);

public:
	LHashIter(const LHash<KeyT,DataT> &lhash, int (*sort)(KeyT, KeyT) = 0): myLHashBody(BODY(lhash.body)), current(0),
		numEntries(lhash.numEntries()), sortFunction(sort)
	{
		/*
		* Note: we access the underlying LHash through the body pointer,
		* not the top-level LHash itself.  This allows the top-level object
		* to be moved without affecting an ongoing iteration.
		* XXX: This only works because
		* - it is illegal to insert while iterating
		* - deletions don't cause reallocation of the data
		*/
		if (sortFunction && myLHashBody) {
			sortKeys();
		} else {
			sortedKeys = 0;
		}
	}
	~LHashIter(){
		delete [] sortedKeys;
	}

	void init(){
		delete [] sortedKeys;

		current = 0;

		{
			/*
			* XXX: fake LHash object to access numEntries()
			*/
			LHash<KeyT,DataT> myLHash(0);

			myLHash.body = myLHashBody;
			numEntries = myLHash.numEntries();
			myLHash.body = 0;
		}

		if (sortFunction && myLHashBody) {
			sortKeys();
		} else {
			sortedKeys = 0;
		}
	}
	DataT *next(KeyT &key){

		if (myLHashBody == 0) {
			return 0;
		} else {
			unsigned index;

			if (sortedKeys) {
				/*
				* Sorted enumeration -- advance to next key in sorted list
				*/
				if (current == numEntries) {
					return 0;
				}

				/*
				* XXX: fake LHash object to access locate()
				*/
				LHash<KeyT,DataT> myLHash(0);

				myLHash.body = myLHashBody;
				myLHash.locate(sortedKeys[current++], index);
				myLHash.body = 0;;
			} else {
				/*
				* Detect deletions by comparing old and current number of 
				* entries.
				* A legal deletion can only affect the current entry, so
				* adjust the current position to reflect the next entry was
				* moved.
				*/
				if (myLHashBody->nEntries != numEntries) {
					assert(myLHashBody->nEntries == numEntries - 1);

					numEntries --;
					current --;
				}

				/*
				* Random enumeration mode
				*/
				unsigned maxBits = myLHashBody->maxBits;

				if (maxBits < minHashBits) {
					/*
					* Linear search mode - advance one to the next entry
					*/
					if (current == numEntries) {
						return 0;
					}
				} else {
					unsigned maxEntries = hashSize(maxBits);

					while (current < maxEntries &&
						Map_noKeyP(myLHashBody->data[current].key))
					{
						current++;
					}

					if (current == maxEntries) {
						return 0;
					}
				}

				index = current ++;
			}

			key = myLHashBody->data[index].key;
			return &(myLHashBody->data[index].value);
		}
	}

private:
	LHashBody<KeyT,DataT> *myLHashBody;	/* data being iterated over */
	unsigned current;			/* current index into data
								* or sortedKeys */
	unsigned numEntries;		/* number of entries */
	int (*sortFunction)(KeyT, KeyT);	/* key sorting function,
										* or 0 for random order */
	KeyT *sortedKeys;			/* array of sorted keys,
								or 0 for random order */
	void sortKeys()
	{
		/*
		* Store keys away and sort them to the user's orders.
		*/
		unsigned maxEntries = hashSize(myLHashBody->maxBits);

		unsigned *sortedIndex = new unsigned[numEntries];
		assert(sortedIndex != 0);

		unsigned i;

		unsigned j = 0;
		for (i = 0; i < maxEntries; i++) {
			if (!Map_noKeyP(myLHashBody->data[i].key)) {
				sortedIndex[j++] = i;
			}
		}
		assert(j == numEntries);

		/*
		* Due to the limitations of the qsort interface we have to 
		* pass extra information to compareIndex in these global
		* variables - yuck. 
		*/
		LHash_thisKeyCompare = (int(*)())sortFunction;
		LHash_thisBody = myLHashBody;
		qsort(sortedIndex, numEntries, sizeof(*sortedIndex), compareIndex);

		/*
		* Save the keys for enumeration.  The reason we save the keys,
		* not the indices, is that indices may change during enumeration
		* due to deletions.
		*/
		sortedKeys = new KeyT[numEntries];
		assert(sortedKeys != 0);

		for (i = 0; i < numEntries; i++) {
			sortedKeys[i] = myLHashBody->data[sortedIndex[i]].key;
		}

		delete [] sortedIndex;
	}

	static int compareIndex(const void *idx1, const void *idx2){
		return (*(compFnType)LHash_thisKeyCompare)
			(BODY(LHash_thisBody)->data[*(unsigned *)idx1].key,
			BODY(LHash_thisBody)->data[*(unsigned *)idx2].key);
	}
	/* callback function for qsort() */
};



template <class KeyT>
static inline Boolean
LHash_equalKey(KeyT key1, KeyT key2)
{
	return (key1 == key2);
}

static inline Boolean
LHash_equalKey(const char *key1, const char *key2)
{
	return (strcmp(key1, key2) == 0);
}

/*
* Hashing functions
* (We provide versions for integral types and char strings;
* user has to add more specialized definitions.)
*/
static inline unsigned
LHash_hashKey(unsigned key, unsigned maxBits)
{
	return (((key * 1103515245 + 12345) >> (30-maxBits)) & hashMask(maxBits));
}

static inline unsigned
LHash_hashKey(const char *key, unsigned maxBits)
{
	/*
	* The string hash function was borrowed from the Tcl libary, by 
	* John Ousterhout.  This is what he wrote:
	* 
	* I tried a zillion different hash functions and asked many other
	* people for advice.  Many people had their own favorite functions,
	* all different, but no-one had much idea why they were good ones.
	* I chose the one below (multiply by 9 and add new character)
	* because of the following reasons:
	*
	* 1. Multiplying by 10 is perfect for keys that are decimal strings,
	*    and multiplying by 9 is just about as good.
	* 2. Times-9 is (shift-left-3) plus (old).  This means that each
	*    character's bits hang around in the low-order bits of the
	*    hash value for ever, plus they spread fairly rapidly up to
	*    the high-order bits to fill out the hash value.  This seems
	*    works well both for decimal and non-decimal strings.
	*/

	unsigned i = 0;

	for (; *key; key++) {
		i += (i << 3) + *key;
	}
	return LHash_hashKey(i, maxBits);
}

//#ifndef __GNUG__
template <class KeyT, class DataT>
DataT *LHash<KeyT,DataT>::removedData = 0;
//#endif /* __GNUG__ */

#endif /* _LHash_h_ */

