/*
 * NgramStats.cc --
 *	N-gram counting
 *
 */

#ifndef _NgramStats_cc_
#define _NgramStats_cc_

#ifndef lint
static char NgramStats_Copyright[] = "Copyright (c) 1995-2002 SRI International.  All Rights Reserved.";
static char NgramStats_RcsId[] = "@(#)$Header$";
#endif


#include <iostream>



#include "NgramStats.h"

#include "Trie.h"

#define INSTANTIATE_NGRAMCOUNTS(CountT) \
	INSTANTIATE_TRIE(VocabIndex,CountT); \
	template class NgramCounts<CountT>






static inline const char *
countToString(unsigned count)
{
    sprintf(ctsBuffer, "%u", count);
    return ctsBuffer;
}

static inline const char *
countToString(int count)
{
    sprintf(ctsBuffer, "%d", count);
    return ctsBuffer;
}


static inline Boolean
stringToCount(const char *str, unsigned &count)
{
    /*
     * scanf("%u") doesn't check for a positive sign, so we have to ourselves.
     */
    return (*str != '-' && sscanf(str, "%u", &count) == 1);
}

static inline Boolean
stringToCount(const char *str, int &count)
{
    return (sscanf(str, "%d", &count) == 1);
}


/*
 * Helper function to record a set of meta-counts at a given prefix into
 * the ngram tree
 */
template <class CountT>
void
NgramCounts<CountT>::addCounts(const VocabIndex *prefix,
				const LHash<VocabIndex, CountT> &set)
{
    NgramNode *node = counts.insertTrie(prefix);

    LHashIter<VocabIndex, CountT> setIter(set);
    VocabIndex word;
    CountT *count;
    while (count = setIter.next(word)) {
	*node->insert(word) += *count;
    }
}

/*
 * Read a counts file discarding counts below given minimum counts
 * Assumes that counts are merged, i.e., ngram order is generated by a
 * pre-order traversal of the ngram tree.
 */
template <class CountT>
Boolean
NgramCounts<CountT>::readMinCounts(File &file, unsigned order,
						unsigned *minCounts)
{
    VocabString words[maxNgramOrder + 1];
    VocabIndex prefix[maxNgramOrder + 1];

    /*
     * Data for tracking deletable prefixes
     */
    LHash<VocabIndex, CountT> metaCounts[order];
    Boolean haveCounts[order];
    VocabIndex *lastPrefix[order];
    for (unsigned i = 0; i < order; i ++) {
	lastPrefix[i] = new VocabIndex[order + 1];
	assert(lastPrefix[i] != 0);
	lastPrefix[i][0] = Vocab_None;
	haveCounts[i] = false;
    }

    Vocab::compareVocab = 0;		// do comparison by word index

    CountT count;
    unsigned int howmany;

    while (howmany = readNgram(file, words, maxNgramOrder + 1, count)) {
	/*
	 * Skip this entry if the length of the ngram exceeds our 
	 * maximum order
	 */
	if (howmany > order) {
	    continue;
	}

	VocabIndex metaTag = Vocab_None;
	Boolean haveRealCount = false;

	/*
	 * Discard counts below mincount threshold
	 */
	if (count < minCounts[howmany - 1]) {
	    if ((unsigned)count == 0) {
		continue;
	    } else {
		/*
		 * See if metatag is defined, and if so, keep going
		 */
	    	metaTag = vocab.metaTagOfType((unsigned)count);

		if (metaTag == Vocab_None) {
		    continue;
		}
	    }
	}

	/* 
	 * Map words to indices (includes both N-gram prefix and last word)
	 */
	if (openVocab) {
	    vocab.addWords(words, prefix, maxNgramOrder);
	} else {
	    vocab.getIndices(words, prefix, maxNgramOrder, vocab.unkIndex);
	}

	/* 
	 * Truncate Ngram for prefix comparison
	 */
	VocabIndex lastWord = prefix[howmany - 1];
	prefix[howmany - 1] = Vocab_None;

	/*
	 * If current ngram prefix differs from previous one assume that 
	 * we are done with all ngrams of that prefix (this assumption is 
	 * true for counts in prefix order).
	 * In this case, output saved meta-counts for the previous prefix,
	 * but only if there also were some real counts -- this saves a lot
	 * of space in the counts tree.
	 */
	if (Vocab::compare(prefix, lastPrefix[howmany-1]) != 0) {
	    if (haveCounts[howmany-1]) {
		addCounts(lastPrefix[howmany-1], metaCounts[howmany-1]);
		haveCounts[howmany-1] = false;
	    }
	    metaCounts[howmany-1].clear();
	    Vocab::copy(lastPrefix[howmany-1], prefix);
	}

	/*
	 *  Update the count (either meta or real)
	 */
	if (metaTag != Vocab_None) {
	    *metaCounts[howmany-1].insert(metaTag) += 1;
	} else {
	    *insertCount(prefix, lastWord) += count;
	    haveCounts[howmany-1] = true;
	}
    }

    /*
     * Record final saved meta-counts and deallocate prefix buffers
     */
    for (unsigned i = order; i > 0; i --) {
	if (haveCounts[i-1]) {
	    addCounts(lastPrefix[i-1], metaCounts[i-1]);
	}
	delete [] lastPrefix[i-1];
    }

    /*
     * XXX: always return true for now, should return false if there was
     * a format error in the input file.
     */
    return true;
}


template <class CountT>
unsigned int
NgramCounts<CountT>::writeNgram(File &file,
		       const VocabString *words,
		       CountT count)
{
    unsigned int i;

    if (words[0]) {
	fprintf(file, "%s", words[0]);
	for (i = 1; words[i]; i++) {
	    fprintf(file, " %s", words[i]);
	}
    }
    fprintf(file, "\t%s\n", countToString(count));

    return i;
}

template <class CountT>
CountT
NgramCounts<CountT>::sumNode(NgramNode *node, unsigned int level, unsigned int order)
{
    /*
     * For leaf nodes, or nodes beyond the maximum level we are summing,
     * return their count, leaving it unchanged.
     * For nodes closer to the root, replace their counts with the
     * the sum of the counts of all the children.
     */
    if (level > order || node->numEntries() == 0) {
	return node->value();
    } else {
	NgramNode *child;
	TrieIter<VocabIndex,CountT> iter(*node);
	VocabIndex wid;

	CountT sum = 0;

	while (child = iter.next(wid)) {
	    sum += sumNode(child, level + 1, order);
	}

	node->value() = sum;

	return sum;
    }
}

template <class CountT>
CountT
NgramCounts<CountT>::sumCounts(unsigned int order)
{
    return sumNode(&counts, 1, order);
}

#endif /* _NgramStats_cc_ */
