/*
* NgramStats.h --
*	N-gram statistics
*
* Copyright (c) 1995-2002 SRI International.  All Rights Reserved.
*
* @(#)$Header$
*
*/

#ifndef _NgramStats_h_
#define _NgramStats_h_

#include <stdio.h>

//#define USE_SARRAY_TRIE

#include "LMStats.h"

#include "Trie.h"

const unsigned maxLineLength = 10000;
const unsigned int	maxNgramOrder = 100;	/* Used in allocating various
											* data structures.  For all
											* practical purposes, this
											* should infinite. */

#define NgramNode	Trie<VocabIndex,CountT>
/*
 * Type-dependent count <--> string conversions
 */
static char ctsBuffer[100];

template <class CountT>
static inline const char *
countToString(CountT count)
{
    sprintf(ctsBuffer, "%lg", (double)count);
    return ctsBuffer;
}

template <class CountT>
static inline Boolean
stringToCount(const char *str, CountT &count)
{
	double x;
	if (sscanf(str, "%lf", &x) == 1) {
		count = x;
		return true;
	} else {
		return false;
	}
}
template <class CountT> class NgramCountsIter;	// forward declaration

template <class CountT>
class NgramCounts: public LMStats
{
	friend class NgramCountsIter<CountT>;

public:
	NgramCounts(Vocab &vocab, unsigned int maxorder) : LMStats(vocab), order(maxorder)
	{
	}

	virtual ~NgramCounts() {};

	unsigned getorder() { return order; };

	/*
	* Individual word/ngram lookup and insertion
	*/
	CountT *findCount(const VocabIndex *words)
	{ return counts.find(words); };
	CountT *findCount(const VocabIndex *words, VocabIndex word1)
	{ NgramNode *node = counts.findTrie(words);
	return node ? node->find(word1) : 0; }
	CountT *insertCount(const VocabIndex *words)
	{ return counts.insert(words); };
	CountT *insertCount(const VocabIndex *words, VocabIndex word1)
	{ NgramNode *node = counts.insertTrie(words);
	return node->insert(word1); };
	CountT *removeCount(const VocabIndex *words)
	{ return counts.remove(words); };
	CountT *removeCount(const VocabIndex *words, VocabIndex word1)
	{ NgramNode *node = counts.findTrie(words);
	return node ? node->remove(word1) : 0; };

	virtual unsigned countSentence(const VocabString *word)
	{ return countSentence(word, (CountT)1); };
	virtual unsigned countSentence(const VocabString *words, CountT factor){
		static VocabIndex wids[maxWordsPerLine + 3];
		unsigned int howmany;

		if (openVocab) {
			howmany = vocab.addWords(words, wids + 1, maxWordsPerLine + 1);
		} else {
			howmany = vocab.getIndices(words, wids + 1, maxWordsPerLine + 1,
				vocab.unkIndex);
		}

		/*
		* Check for buffer overflow
		*/
		if (howmany == maxWordsPerLine + 1) {
			return 0;
		}

		/*
		* update OOV count
		*/
		if (!openVocab) {
			for (unsigned i = 1; i <= howmany; i++) {
				if (wids[i] == vocab.unkIndex) {
					stats.numOOVs ++;
				}
			}
		}

		/*
		* Insert begin/end sentence tokens if necessary
		*/
		VocabIndex *start;

		if (wids[1] == vocab.ssIndex) {
			start = wids + 1;
		} else {
			wids[0] = vocab.ssIndex;
			start = wids;
		}
		if (wids[howmany] != vocab.seIndex) {
			wids[howmany + 1] = vocab.seIndex;
			wids[howmany + 2] = Vocab_None;
		}

		return countSentence(start, factor);
	}
	virtual unsigned countSentence(const VocabIndex *word)
	{ return countSentence(word, (CountT)1); };
	virtual unsigned countSentence(const VocabIndex *words, CountT factor){
		unsigned int start;

		for (start = 0; words[start] != Vocab_None; start++) {
			incrementCounts(words + start, 1, factor);
		}

		/*
		* keep track of word and sentence counts
		*/
		stats.numWords += start;
		if (words[0] == vocab.ssIndex) {
			stats.numWords --;
		}
		if (start > 0 && words[start-1] == vocab.seIndex) {
			stats.numWords --;
		}

		stats.numSentences ++;

		return start;
	}

	Boolean read(File &file) { return read(file, order); };
	Boolean read(File &file, unsigned int order){
		VocabString words[maxNgramOrder + 1];
		VocabIndex wids[maxNgramOrder + 1];
		CountT count;
		unsigned int howmany;

		while (howmany = readNgram(file, words, maxNgramOrder + 1, count)) {
			/*
			* Skip this entry if the length of the ngram exceeds our 
			* maximum order
			*/
			if (howmany > order) {
				continue;
			}
			/* 
			* Map words to indices
			*/
			if (openVocab) {
				vocab.addWords(words, wids, maxNgramOrder);
			} else {
				vocab.getIndices(words, wids, maxNgramOrder, vocab.unkIndex);
			}

			/*
			*  Update the count
			*/
			*counts.insert(wids) += count;
		}
		/*
		* XXX: always return true for now, should return false if there was
		* a format error in the input file.
		*/
		return true;
	}
	Boolean readMinCounts(File &file, unsigned order, unsigned *minCounts);
	void write(File &file) { write(file, order); };
	void write(File &file, unsigned int order, Boolean sorted = false){
		static char buffer[maxLineLength];
		writeNode(&counts, file, buffer, buffer, 1, order, sorted);
	}


	static unsigned int parseNgram(char *line,
		VocabString *words, unsigned int max,
		CountT &count){
			unsigned howmany = Vocab::parseWords(line, words, max);

			if (howmany == max) {
				return 0;
			}

			/*
			* Parse the last word as a count
			*/
			if (!stringToCount(words[howmany - 1], count)) {
				return 0;
			}

			howmany --;
			words[howmany] = 0;

			return howmany;
		}
		static unsigned int readNgram(File &file,
			VocabString *words, unsigned int max,
			CountT &count){
				char *line;

				/*
				* Read next ngram count from file, skipping blank lines
				*/
				line = file.getline();
				if (line == 0) {
					return 0;
				}

				unsigned howmany = parseNgram(line, words, max, count);

				if (howmany == 0) {
					file.position() << "malformed N-gram count or more than " << max - 1 << " words per line\n";
					return 0;
				}

				return howmany;
			}
		/* read one ngram count from file */
		static unsigned int writeNgram(File &file, const VocabString *words,
			CountT count);
		/* write ngram count to file */

		CountT sumCounts() { return sumCounts(order); };
		CountT sumCounts(unsigned int order);
		/* sum child counts on parent nodes */

		void dump(){
			std::cerr << "order = " << order << std::endl;
			counts.dump();
		}
		void memStats(MemStats &stats){
			stats.total += sizeof(*this) - sizeof(counts) - sizeof(vocab);
			vocab.memStats(stats);
			counts.memStats(stats);
		}

protected:
	unsigned int order;
	NgramNode counts;
	void incrementCounts(const VocabIndex *words,
		unsigned minOrder = 1, CountT factor = 1){
			NgramNode *node = &counts;

			for (int i = 0; i < order; i++) {
				VocabIndex wid = words[i];

				/*
				* check of end-of-sentence
				*/
				if (wid == Vocab_None) {
					break;
				} else {
					node = node->insertTrie(wid);
					if (i + 1 >= minOrder) {
						node->value() += factor;
					}
				}
			}
		}
		void addCounts(const VocabIndex *prefix,
			const LHash<VocabIndex, CountT> &set);
		void writeNode(NgramNode *node, File &file, char *buffer, char *bptr,
			unsigned int level, unsigned int order, Boolean sorted){
				NgramNode *child;
				VocabIndex wid;

				TrieIter<VocabIndex,CountT> iter(*node, sorted ? vocab.compareIndex() : 0);

				/*
				* Iterate over the child nodes at the current level,
				* appending their word strings to the buffer
				*/
				while (!file.error() && (child = iter.next(wid))) {
					VocabString word = vocab.getWord(wid);

					if (word == 0) {
						std::cerr << "undefined word index " << wid << "\n";
						continue;
					}

					unsigned wordLen = strlen(word);

					if (bptr + wordLen + 1 > buffer + maxLineLength) {
						*bptr = '0';
						std::cerr << "ngram ["<< buffer << word
							<< "] exceeds write buffer\n";
						continue;
					}

					strcpy(bptr, word);

					/*
					* If this is the final level, print out the ngram and the count.
					* Otherwise set up another level of recursion.
					*/
					if (order == 0 || level == order) {
						fprintf(file, "%s\t%s\n", buffer, countToString(child->value()));
					} 

					if (order == 0 || level < order) {
						*(bptr + wordLen) = ' ';
						writeNode(child, file, buffer, bptr + wordLen + 1, level + 1,
							order, sorted);
					}
				}
			}
			CountT sumNode(NgramNode *node, unsigned int level, unsigned int order);
};

/*
* Iteration over all counts of a given order
*/
template <class CountT>
class NgramCountsIter
{
public:
	NgramCountsIter(NgramCounts<CountT> &ngrams, VocabIndex *keys, 
		unsigned order = 1,
		int (*sort)(VocabIndex, VocabIndex) = 0)
		: myIter(ngrams.counts, keys, order, sort) {};
	/* all ngrams of length order, starting
	* at root */

	NgramCountsIter(NgramCounts<CountT> &ngrams, const VocabIndex *start,
		VocabIndex *keys, unsigned order = 1,
		int (*sort)(VocabIndex, VocabIndex) = 0)
		: myIter(*(ngrams.counts.insertTrie(start)), keys, order, sort) {};
	/* all ngrams of length order, rooted
	* at node indexed by start */

	void init() { myIter.init(); };
	CountT *next()
	{ NgramNode *node = myIter.next();
	return node ? &(node->value()) : 0; };

private:
	TrieIter2<VocabIndex,CountT> myIter;
};


/*
* Instantiate the count trie for integer and float count types
*/

typedef unsigned int NgramCount;

class NgramStats: public NgramCounts<NgramCount>
{
public:
	NgramStats(Vocab &vocab, unsigned int order)
		: NgramCounts<NgramCount>(vocab, order) {};
	virtual ~NgramStats() {};
};

class NgramsIter: public NgramCountsIter<NgramCount>
{
public:
	NgramsIter(NgramStats &ngrams, VocabIndex *keys, unsigned order = 1,
		int (*sort)(VocabIndex, VocabIndex) = 0)
		: NgramCountsIter<NgramCount>(ngrams, keys, order, sort) {};
	NgramsIter(NgramStats &ngrams, const VocabIndex *start,
		VocabIndex *keys, unsigned order = 1,
		int (*sort)(VocabIndex, VocabIndex) = 0)
		: NgramCountsIter<NgramCount>(ngrams, start, keys, order, sort) {};
};

#endif /* _NgramStats_h_ */

