#include <string.h>
#include <config.h>
#include <fortify.h>
#include "msocket.h"
#include <iostream>

// Common error strings, to save some space
static char ErrNotInit[] = "Socket not initialized";
static char ErrConTerm[] = "Connection terminated";

// m_szErrMsg is not part of the class, because of the const functions throwing exceptions
// might modify it.  Size is probably overkill.  Unfortunately, my g++ compiler doesn't
// support the sstream class.
static char m_szErrMsg[256];

// Statics for the class
bool MSocket::m_bThrowException = true;
std::map<int,int> MSocket::RefCount;

// Start of the MSocket class functions
MSocket::MSocket( int nFamily, int nType, int nProtocol )
{
m_nTimeOut=0;
// Try to intialize the MSocket, using the socket system call
m_nSocket = ::socket( nFamily, nType, nProtocol );
if( m_nSocket < 0 && m_bThrowException ) // If an error, and we should throw it
ThrowErrMsg( "socket() call failed: %d", m_nSocket );
if( m_nSocket >= 0 )
IncRef();
}

MSocket::MSocket( const MSocket& msrSrc )  // Copy Constructor
{
m_nTimeOut=0;
m_nSocket = msrSrc.m_nSocket;
IncRef();
}

MSocket::~MSocket()  //Destructor
{
if( m_nSocket >= 0 )
Close();
}

int MSocket::Close()
{
int Ret = m_nSocket;
if( Ret < 0 && m_bThrowException )
ThrowErrMsg( "Close called on a closed socket" );
if( Ret > 0 )
{
DecRef(); // Check reference count, if we are the last 
if( !GetRef() ) // using the socket, then close it.
close( m_nSocket );
m_nSocket = -1;
}
return( Ret );
}

int MSocket::Connect( struct sockaddr * sapSrvrAdrs, int nLenAddr ) const
{
if( m_nSocket < 0 )
{
if( m_bThrowException ) // If not setup correctly
ThrowErrMsg( ErrNotInit );
return(-1);
}

// Call the connect system function to actually do connect.
int Ret = ::connect( m_nSocket, (struct sockaddr*)sapSrvrAdrs, nLenAddr );
if( Ret < 0 && m_bThrowException ) // If there was an error, and we should throw it.
ThrowErrMsg( "connect() failed: %d", Ret );
return( Ret );
}

int MSocket::Connect( const char* cpHostAddr, short sPort ) const
{
// Setup and init the sock_addr_in struct, needed by real connect function
struct sockaddr_in SrvrAdrs;
memset( &SrvrAdrs, 0, sizeof( SrvrAdrs ) );

// If szHostAddr looks like '127.0.0.1' then it's easy, just
// call inet_addr to convert the string into a netword address ID integer
if( isdigit(cpHostAddr[0]) )
SrvrAdrs.sin_addr.s_addr = inet_addr( cpHostAddr );
else
{
// Otherwise, it may be a host name.  Get it's IP address and use that.
// For example, szHostAddr may be www.acme.com
struct hostent * pHostEnt = gethostbyname( cpHostAddr );
                if( pHostEnt == NULL )
{
if( m_bThrowException )
ThrowErrMsg( "Unable to determine IP for %s", cpHostAddr );
return( -1 );
}
SrvrAdrs.sin_addr.s_addr =  *(long*)pHostEnt->h_addr_list[0];
}
SrvrAdrs.sin_family = AF_INET;
// Call htons, to convert our local pc short to format compatible with the 'net'
SrvrAdrs.sin_port = htons( sPort );

// finally, call the other version of MSocket::Connect, with the struct we set up
return( Connect( (struct sockaddr*) &SrvrAdrs , (int)sizeof( SrvrAdrs ) ) );
}

int MSocket::Bind( struct sockaddr* sapMyAddr, int nAddrLen ) const
{
if( m_nSocket < 0 )  // Some called Bind with a closed socket?
{
if( m_bThrowException ) // Should we throw exception on an error?
ThrowErrMsg( ErrNotInit );
return(-1);
}
// Call the bind system function to actually do connect.
int Ret = ::bind( m_nSocket, sapMyAddr, nAddrLen );
if( Ret < 0 && m_bThrowException ) // If there was an error, and we should throw it.
ThrowErrMsg( "bind() failed: %d", Ret );
return( Ret );
}

int MSocket::Bind( int nPort, int nFamily, int nAddr ) const
{
struct sockaddr_in SrvrAdrs;
memset( &SrvrAdrs, 0, sizeof( SrvrAdrs ) );
SrvrAdrs.sin_family = nFamily;
SrvrAdrs.sin_addr.s_addr = htonl( nAddr );
SrvrAdrs.sin_port = htons( nPort );
// Here, we call our 'other' Bind member function
return( Bind( (struct sockaddr*)&SrvrAdrs, (int)sizeof( SrvrAdrs ) ) );
}

int MSocket::Listen( int nBackLog ) const
{
if( m_nSocket < 0 )
{
if( m_bThrowException )
ThrowErrMsg( ErrNotInit );
return( -1 );
}
// Call the system 'listen' function
int Ret = ::listen( m_nSocket, nBackLog );
if( Ret < 0 && m_bThrowException )
ThrowErrMsg( "listen() failed: %d", Ret );
return( Ret );
}

int MSocket::Server( int nPort ) const // Just a convenient function
{
int Ret;
Ret = Bind( nPort );
if( Ret >= 0 )
Ret = Listen();
return( Ret );
}

int MSocket::Accept( MSocket& msrRet, struct sockaddr* sapClientAddr, int* npLen ) const
{
int MyLen;
struct sockaddr_in Client;
if( sapClientAddr == 0 ) // To make things easier on our caller, we can create the struct
{
sapClientAddr = (sockaddr*)& Client;
npLen = &MyLen;
}
//if( m_nTimeOut && TimedOut() )
//return( -1 );
// Call the system 'accept' function
//int nRet = ::accept( m_nSocket, sapClientAddr, (unsigned int*)npLen );
int nRet = ::accept( m_nSocket, sapClientAddr, (socklen_t*)npLen );
if( nRet < 0 && m_bThrowException )
ThrowErrMsg( "accept failed: %d", nRet );
msrRet = nRet;
return( nRet );
}

MSocket& MSocket::operator=( MSocket& msrSrc )
{
if( &msrSrc != this ) // Make sure caller didn't to X = X;
*this = msrSrc.m_nSocket;
return( *this );
}

MSocket& MSocket::operator=( int nSocket )
{
if(  nSocket < 0 && m_bThrowException )
throw "operator= called with bad socket";

if( m_nSocket >= 0 )
Close();
m_nSocket = nSocket;
IncRef();
return( *this );
}

int MSocket::TimedOut() const
{
  fd_set fdReadSet;
  fd_set fdWriteSet;
  fd_set fdExceptSet;
  struct timeval TimeToWait;
  //std::cout<<"inTimer:"<<m_nTimeOut<<std::endl;
  //TimeToWait.tv_sec = m_nTimeOut / 10000;
  //TimeToWait.tv_usec = m_nTimeOut % 10000;
  TimeToWait.tv_sec = m_nTimeOut;
  //std::cout<<"TimeToWait.tv_sec"<<TimeToWait.tv_sec<<std::endl;
  TimeToWait.tv_usec = 0;
  FD_ZERO( &fdReadSet );
  FD_ZERO( &fdWriteSet );
  FD_ZERO( &fdExceptSet );
  FD_SET( m_nSocket, &fdReadSet );

  //std::cout<<"FD:"<<m_nSocket<<std::endl; 
  //int Ret = select( m_nSocket+1, &fdSet, &fdSet, &fdSet, &TimeToWait );
int Ret = select( m_nSocket+1, &fdReadSet, &fdWriteSet, &fdExceptSet, &TimeToWait );
//std::cout<<"RetValue:"<<Ret<<std::endl; 
  if( Ret <=0 && m_bThrowException )
    ThrowErrMsg( Ret==0?"Time out":"Select error: %d", Ret );
  
 if(FD_ISSET(m_nSocket, &fdReadSet)){
   return 0;
 }else{
   return 1;
 }
 /*
 if (Ret > 1){
   return 0;
 }else{
   return 1;
   }*/

}

bool MSocket::SetExceptions( bool bMode )
{
bool Ret;
Ret = m_bThrowException;
m_bThrowException = bMode;
return( Ret );
}
int MSocket::SetTimeOut( int nMicroSeconds )
{
  int Ret = m_nTimeOut;
  m_nTimeOut = nMicroSeconds;
  std::cout<<"m_nTimeOutSetting:"<<m_nTimeOut<<std::endl;
  return( Ret );
}

int MSocket::ReadLine( char *cpDest, int nSize, char Term ) const
{
  int ReadIn=0, Stat;
  if( m_nSocket < 0 )
    {
      if( m_bThrowException )
	ThrowErrMsg( ErrNotInit );
      return( -1 );
    }
  while( ReadIn < nSize-2 )
    {
      //std::cout<<"m_nTimeOut:"<<m_nTimeOut<<std::endl;
      //std::cout<<"TimedOut"<<TimedOut()<<std::endl;      
      if( (m_nTimeOut!=0) && TimedOut() ){
	//std::cout<<"Both conditions true"<<std::endl;
	return( -1 );
      }
      
      Stat = read( m_nSocket, cpDest+ReadIn, 1 );
      //std::cout<<"Char:"<<cpDest[ReadIn]<<std::endl;
      if( Stat <= 0 )
	{
	  if( m_bThrowException )
	    ThrowErrMsg( Stat==0?ErrConTerm:"ReadLine error: %d", Stat );
	  return( Stat );
	}
      if( cpDest[ReadIn]==Term )
	break;
      ReadIn++;
    }
  if( Term != '\0' )
    ReadIn++;;
  cpDest[ReadIn]='\0';
}

int MSocket::WriteLine( const char *cpSrc, char Term ) const
{
int Written=0, Len;
Len = strlen(cpSrc);
if( Term=='\0' ) // If terminator is '\0', include that in data out
Len++;

if( m_nSocket < 0 )
{
if( m_bThrowException )
ThrowErrMsg( ErrNotInit );
return( -1 );
}
while( Len>0 )
{
Written=write( m_nSocket, cpSrc, Len );
if( Written <=0 )
{
if( m_bThrowException )
ThrowErrMsg( Written==0?ErrConTerm:"Writing socket: %d", Written );
return( Written );
}
cpSrc += Written;
Len -= Written;
}
return( 0 );
}

void MSocket::ThrowErrMsg( const char* cpMsg, int nCode ) const
{
  sprintf( m_szErrMsg, cpMsg, nCode );
  throw m_szErrMsg;
}

bool MSocket::WasConTermErr( const char* ErrMsg )
{
return( strcmp(ErrMsg,ErrConTerm)==0);
}

// A cout 'look alike' function, for simplicty.  ostream was not really used.
MSocket& operator<<( MSocket& msrOut, const char* cpStr )
{
msrOut.WriteLine( cpStr );
return( msrOut );
}

// A cin 'look alike' function, for simplicity.  istream is not really used.
MSocket& operator>>( MSocket& msrIn, char* cpStr )
{
msrIn.ReadLine( cpStr, 0x0FFFF );  // No good number here, really
return( msrIn );
}

