// This file implements one and two dimensional dynamic arrays
//
// This class was inspired by array implementation in
//     Martin Aupperle
//     Programmierhandbuch Visual C++
//     Vieweg, Braunschweig 1994
//
// 1997.09.08 - Version 1.0 (Vogel)
// somewhen   - as template
// 2005.10.30 - adjusted to compile with g++ 3.4.4
// 2004.11.05 - comments now English


#ifndef __Array_hh__
#define __Array_hh__


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif



//========================= Includes ===============

#include <iostream>
using std::ostream;


//========================= Export ==================

// aliases for 1- and 2-dim arrays
#define CArray  CArray1D
#define CMatrix CArray2D


template <class T> class CArray1D;
template <class T> class CArray2D;

template <class T> bool operator == (const CArray1D<T> &Array, const CArray1D<T> &Array2 );
template <class T> ostream &operator << ( ostream &Stream, const CArray1D<T> &Array );



//--------------------------------------------------------------------
//       class CArray1D
//

template <class T>
class CArray1D
{

public:

  CArray1D();  // create empty array, boundaries are [0, -1]
  ~CArray1D(); 

  // copy constructor: creates flat copy, i.e. in case of pointers, 
  // only the pointers, not the objects are copied
  CArray1D              ( const CArray1D<T> & Array );

  // assignment operator: also flat copy
  CArray1D & operator = ( const CArray1D<T> & Array );

  // create string representation of the array content
  //virtual operator const char*() const;

  // delete all elements and set boundaries to [0, -1], free memory
  void Clear();

  // delete all elements and set boundaries to [0, -1], but do not free memory
  void RemoveAll();

  // make the array object invalid, free memory
  void Invalidate();
  
  // return true if object is valid, false otherwise
  bool IsValid() const;

  // return true if Index is valid, false otherwise
  bool IsValidIndex( int Index ) const;

  // set size of increment for increasing the allocated memory
  void SetIncrement( int Increment = 8 );

  // make the array static, i.e. it can not grow
  void MakeStatic();

  // make the array dynamic, i.e. it can grow (this is default behaviour)
  void MakeDynamic();

  // for array containing basic data types and pointers the array 
  // will be initialized to zero
  void InitToZero();

  // return size of the array, i.e. number of elements in the array
  size_t Size() const;

  // return true if array is empty, false otherwise
  bool IsEmpty() const;

  // return minimum and maximum index
  int MinIndex() const;
  int MaxIndex() const;

  // set minimum and maximum index
  bool SetMinIndex( int MinIndex );
  bool SetMaxIndex( int MaxIndex );
  bool SetBoundaries( int MinIndex, int MaxIndex );

  // make sure array has at least given boundaries, only increases the size,
  // never makes the array smaller
  bool AssureBoundaries( int MinIndex, int MaxIndex );

  // return reference to object a Index; assertion if index is invalid
  const T & GetAt( int Index ) const;

  // store copy of Element at Index
  // if index out of bounds and array is dynamic, array will grow
  // return true if element is stored, false otherwise
  bool SetAt( int Index, const T & Element );

  // store copies of Element at all positions in the array
  // return false if array is invalid, true otherwise
  bool SetAll( const T & Element );

  // return reference to element at Index
  // as rvalue: only read access
  // as lvalue: store element
  T & operator[] ( int Index );

  // return const reference to element at Index
  const T & operator[] ( int Index ) const;

  // append copy of element to the array, i.e. set it at index MaxIndex + 1
  // return bool if successful
  bool Append( const T & Element);

  // prepend copy of Element to the array, i.e. set it at index MinIndex - 1
  // return bool if successful
  bool Prepend( const T & Element);

  //insert copy of Element at Index, move all Elements at and right to index
  // by one position
  // return true if successful
  bool Insert( int Index, const T & Element );
  //bool Insert( int Index, const CArray1D<T> & Array );

  // remove Count Elements at index, move all Elements at and right to index + 1
  // one position to the left
  bool Remove( int Index, int Count = 1 );

  // stream operator: write array to Stream
  friend ostream &operator << <T> ( ostream &Stream, const CArray1D<T> &Array );

  //friend bool operator== <T> ( const CArray1D<T> &Array1, const CArray<T> &Array2 );

  // write array to cout
  void Display() const;

  // write information about array to cout
  void Debug() const;


protected:
  // initialize array, MinIndex = 0, MaxIndex = -1; called in constructors
  void Init();

  // the following functions assume an initialized, but possibly invalid object

  // check if index is within array boundaries
  //bool CheckIndex( int Index ) const;

  int m_MinIndex;  // minimal index in array
  int m_MaxIndex;  // maximal Index in array

  int m_MinAlloc;  // minimal index for which memory is allocated
  int m_MaxAlloc;  // maximal index for which memory is allocated

  T *m_pData;      // pointer to allocated memory for storing the elements

  T *m_pElements;  // pointer to memory where element with index 0 lies (or would lay)

  int m_Increment; // number of elements for which new memory is allocated

  bool m_IsStatic; // flag if array is static or dynamic

  bool m_InitToZero;  // flag to indicate if array should be initialize to zero

};





//--------------------------------------------------------------------
//       class CArray2D
//
// 2-dim array is array of arrays. 
//  _
// |_|     _ _ _ _ _ _
// |_|--->|_|_|_|_|_|_| - Row
// |_|--->|_|_|_|_|
// |_|       |
//          Col
//
// first index (Index1) is index in the master array
// second index (Index2) is index in the row arrays
// each row is expanded only as far a neccessary


template <class T>
class CArray2D {

public:
  CArray2D();  // create empty array
  ~CArray2D(); 

  // copy constructor and assignment operator
  CArray2D              ( const CArray2D<T> & Array );
  CArray2D & operator = ( const CArray2D<T> & Array );

  // return string representation of array in static buffer
  //virtual operator const char*() const;

  // delete all elements and set boundaries to [0, -1][0, -1]
  // i.e. no element exists and array is invalid
  void Clear();

  // return true if obect is valid, false otherwise
  bool IsValid() const;

  // return true if Index is valid, false otherwise
  bool IsValidIndex( int Index1, int Index2 ) const;

  // set size of increment for increasing the allocated memory in master array
  void SetIncrement( int Increment = 8 );

  // set size of increment for increasing the allocated memory in rows
  void SetRowIncrement( int Increment = 8 );

  // make the array static, i.e. it can not grow
  void MakeStatic();

  // make the array dynamic, i.e. it can grow (this is default behaviour)
  void MakeDynamic();

  //int GetSize() const;

  // return minimum and maximum index for dimension 1
  int MinIndex1() const;
  int MaxIndex1() const;

  // return minimum and maximum index for dimension 2, i.e. for rows
  int MinIndex2() const;
  int MaxIndex2() const;

  // set minimum and maximum index, array size can be increased or decreased
  bool SetMinIndex1( int MinIndex1 );
  bool SetMaxIndex1( int MaxIndex1 );
  //bool SetMinIndex2( int MinIndex2 );
  //bool SetMaxIndex2( int MaxIndex2 );
  bool SetBoundaries( int MinIndex1, int MaxIndex1 );
  bool SetBoundaries( int MinIndex1, int MaxIndex1, 
					  int MinIndex2, int MaxIndex2 );

  // make sure array has at least given boundaries, only increases the size,
  // never makes the array smaller
  bool AssureBoundaries( int MinIndex1, int MaxIndex1 );

  // return reference to object a [Index1, Index2]; assertion if index is invalid
  const T & GetAt( int Index1, int Index2 ) const;

  // store copy of Element at [Index1, Index2]
  // if index out of bounds and array is dynamic, array will grow
  // return true if element is stored, false otherwise
  bool SetAt( int Index1, int Index2, const T & Element );

  // store copies of Element at all positions in the array
  // return false if array is invalid, true otherwise
  bool SetAll( const T & Element );

  // return reference to element at [Index1, Index2]
  // as rvalue: only read access
  // as lvalue: store element
  T & operator() ( int Index1, int Index2 );

  // return const reference to element at [Index1, Index2]
  const T & operator() ( int Index1, int Index2 ) const;

  //bool Append( const T & Element);
  //bool Insert( T & Element );
  //bool Insert( const CArray1D<T> & Array );
  //bool Remove( int Index, int Count );


  //-- GetRow liefert eine Zeile der Matrix als Refernz auf den Vektor
  // return reference to row array [Index1, ...]
  CArray1D<T> & GetRow( int Index );

  //-- operator +=, << alias for append
  //CArray1D &operator += ( CContainerItem * );
  //CArray1D &operator << ( CContainerItem * );
  
  // write array to cout
  void Display() const;

  // write information about array to cout
  void Debug() const;


protected:
  // initialize array, MinIndex = 0, MaxIndex = -1; called in constructors
  void Init();

  // the following functions assume an initialized, but possibly invalid object

  // make the array object invalid, free memory
  void Invalidate();
  
  // check if index is within array boundaries
  //bool CheckIndex( int Index ) const;

  //int m_MinIndex;
  //int m_MaxIndex;

  //int m_MinAlloc;
  //int m_MaxAlloc;

  CArray1D< CArray1D<T> > m_Data; // pointer to allocated memory for storing the elements

  bool m_IsStatic;    // flag if array is static or dynamic
  bool m_InitToZero;  // flag to indicate if array should be initialize to zero

};




// Implementation includieren wegen Instantiierung
#include <Array_inl.hh>

#endif
