/**
   @file ChunkAllocator.hh

   Provides fast allocator class for STTK's object allocation scheme.

   CVS id: $Id: ChunkAllocator.hh 1148 2005-04-13 17:53:33Z dietmar $
 */
#ifndef ChunkAllocator_hh
#define ChunkAllocator_hh

// autoconf's symbols
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>

// base class
#include <ObjectAllocator.hh>

/**
   @brief An allocator that allocates a large chunk of objects at once.

   This allocator allocates a whole bunch of objects at once, hopefully
   speeding up the allocation process. It does so by calling the hook
   AllocateObjects(). Subclasses may overwrite this in order to call alternate
   constructors etc. AllocateObjects() must return an array of m_ChunkSize objects,
   a constant that does not change over the life time of a CChunkAllocator.

   The life cycle of an object allocated by this allocator is as follows:

   The user requests a new object using the Create() method. The allocator
   returns an object from the current chunk, which becomes "active".
   If the current chunk is empty, a new chunk of m_ChunkSize object is being allocated,
   which all have the state "allocated" (but not active yet). Consequently,
   if the last object of the current chunk has been activated, the current chunk
   is set to empty.

   When deleting an object, the object becomes "deleted". The
   corresponding chunk will be looked up and its counter for allocated objects
   will be decreased. If the counter is zero, the whole chunk will be freed,
   causing all objects in this chunk to become "deleted and freed".
   An object deleted, but not yet freed is a "zombie".

   Note that only the current chunk can contain allocated but inactive objects.

   Chunks are organized in a balanced binary tree for fast lookup.
 */
template<class T>
class CChunkAllocator : public CObjectAllocator<T> {
private:
  /**
     @brief Data structure for a chunk of objects.

     The purpose of this data structure is twofold: Firstly, associate the pure array
     of objects in a chunk with meta data, such as allocation counters, secondly, provide
     pointers to construct a balanced binary tree of chunks.
   */
  struct TChunk {
    //! Pointer on the array of objects
    T* objects;
    //! Number of deleted objects.
    unsigned long int cZombies;

    //! Parent node in tree.
    TChunk* parent;
    //! The subtree of chunks having smaller adresses.
    TChunk* smallChild;
    //! The subtree of chunks having bigger adresses.
    TChunk* bigChild;

    /**
       @brief Size of the tree, including the current node itself.

       This is 1 iff the node is a leaf node and always the sum of the sizes of the child nodes plus 1.
    */
    int treeSize;
  }; // struct TChunk

  //! Number of elements in a chunk
  const unsigned long m_ChunkSize;

  //! Current chunk.
  TChunk* m_pCurrentChunk;
  //! Number of activated objects in current chunk.
  unsigned long m_iNextObject;

  //! Total number of allocated chunks.
  unsigned long m_cChunks;

  //! Number of objects deleted, but not yet freed.
  unsigned long m_cZombies;

  //! The root of the balanced binary tree of chunks.
  TChunk* m_pChunks;

  //! Insert a new chunk into the balanced tree.
  void InsertNode(TChunk* pChunk);
  //! Remove a chunk from the balanced tree.
  void DeleteNode(TChunk*pChunk);
  //! Check the current node/chunk of the tree and balance, if necessary.
  bool CheckAndBalance(TChunk* pChunk);
  //! Find the chunk responsible for an object
  TChunk* Lookup(const T* object);

protected:
  /**
     @brief Allocate chunk hook.
     @return An array of m_ChunkSize objects of template type T.

     This is a hook called by the class in order to allocate a chunk of objects.
     The value returned must be a pointer on an array of objects, containing
     GetChunkSize() objects.

     Subclasses may overwrite this in order to call non-default constructors etc.
     This method just allocates the array of objects, Create() deals with the corresponding
     data structures (TChunk) in order to administrate chunks etc.
  */
  inline virtual T* AllocateChunk() { assert(m_ChunkSize > 0); return new T[m_ChunkSize]; }
  /**
     @brief Delete chunk hook.
     
     Delete a chunk previously allocated by AllocateChunk().
  */
  void FreeChunk(T* pChunk) { assert(pChunk); delete[] pChunk; }

public:
  /**
     @brief Constructor.
     @param ChunkSize (greater than zero) The chunk size associated with an allocator object will stay
     constant over the allocators life time.
  */
  inline CChunkAllocator(unsigned long ChunkSize = 2048)
    : m_ChunkSize(ChunkSize), m_pCurrentChunk(NULL), m_pChunks(NULL), m_cChunks(0), m_cZombies(0)
  { assert(ChunkSize > 0); }
  //! Destruktor.
  ~CChunkAllocator();


  //! Factory method: Allocate object.
  virtual T* Create();
  //! Factory method: Copy object.
  virtual T* Copy( const T& obj );
  //! Factory method: Delete object.
  virtual void Del( T* deletable );
  //! Delete all objects allocated by this allocator at once.
  void Clear();
  //! Get number if allocated objects.
  virtual long int GetNumberOfAllocatedObjects() const { return m_cChunks * m_ChunkSize; }
  //! Get number of objects that will be allocated at once.
  inline unsigned long GetChunkSize() const { return m_ChunkSize; }
}; // CChunkAllocator

// template class, must include definitions
#include <ChunkAllocator_inl.hh>

#endif // header already included?
