/**
   @file

   @brief Inline definitions for template class CChunkAllocator.

   Never include this file directly, use ChunkAllocator.hh instead.

   CVS file ID: $Id: ChunkAllocator_inl.hh 1149 2005-04-13 18:09:25Z dietmar $
 */
#ifndef ChunkAllocator_inl_hh
#define ChunkAllocator_inl_hh

#include <assert.h>

// The using module should already have included Trace.hh, so either
// TRACE_ENABLED or TRACE_DISABLED and the trace macros should be defined.
#ifndef Trace // check for trace macro

// No trace macros! This is an error, however instead of complaining, default to "no tracing".
#ifndef NO_TRACE    // avoid double definition
#define NO_TRACE 1
#endif
#include <Trace.hh> // If already included, this does nothing

#endif // Trace defined?

// check if this header was included correctly.
#ifndef ChunkAllocator_hh
#error ERROR: Header was not included by ChunkAllocator.hh. This is an internal header. Do not attempt to include it directly. Use ChunkAllocator.hh instead.
#endif

#include <vector>
using std::vector;

/**
   @brief Insert a new chunk into the balanced tree.
   @param pNode The new chunk to be inserted.

   Insert a new chunk into the tree, keeping the tree balanced.

   The child pointers and parent pointer of the chunk object will be overwritten.
   This will not update the m_cChunks member, as that is a counter for allocated chunks,
   and not chunks in the tree (although they should be kept in sync).

   Algo: Find position to insert
         Insert node, possibly violating balancedness
         Go up to the root, checking and balancing the tree
*/
template < class T >
void CChunkAllocator<T>::InsertNode(TChunk* pNode){
  assert(pNode);
  // we only insert full chunks
  assert(pNode->cZombies == 0);

  // reset child pointers of new chunk
  pNode->smallChild = pNode->bigChild = NULL;
  pNode->treeSize = 1;

  // Simple case: first chunk
  if (m_pChunks == NULL) {
    m_pChunks = pNode;
    pNode->parent = NULL;
  } else {
    // Search position in tree for insertion
    TChunk *pNext, *pCurrent;
    pNext = m_pChunks;
    assert(pNext);
    while (pNext != NULL) {
      pCurrent = pNext;
      // pick smaller or bigger child
      if (pNode->objects < pCurrent->objects)
	pNext = pCurrent->smallChild;
      else {
	pNext = pCurrent->bigChild;
	// new chunk cannot lie within current chunk!
	assert(pNode->objects - pCurrent->objects >= m_ChunkSize);
      } // endif: new chunk's adress is bigger than current chunk's
    } // endwhile: find insertion position

    // pCurrent is now the father of the new chunk
    pNode->parent = pCurrent;
    // current node cannot be unbalanced by this insertion
    assert(pCurrent->treeSize <= 2);
    ++pCurrent->treeSize;
    if (pNode->objects < pCurrent->objects) {
      assert(pCurrent->smallChild == NULL);
      pCurrent->smallChild = pNode;
    } else {
      assert(pCurrent->bigChild == NULL);
      pCurrent->bigChild = pNode;
    } // endif: insert new chunk as left or right child of current chunk

    // re-balance tree
    pCurrent = pCurrent->parent;
    // all the way up the root
    while (pCurrent != NULL) {
      // update treeSize
      pCurrent->treeSize = ( (pCurrent->smallChild == NULL ? 0 : pCurrent->smallChild->treeSize) +
			     (pCurrent->bigChild   == NULL ? 0 : pCurrent->bigChild->treeSize) +
			     1 );
      // balance node
      if (CheckAndBalance(pCurrent))
	// tree had been reorganized, extra go-up
	pCurrent = pCurrent->parent;
      assert(pCurrent);
      pCurrent = pCurrent->parent;
    } // endwhile go up to the root
  } // endif: special treatment of first chunk

  // check balancedness of root
  assert(m_pChunks);
  assert(abs( (m_pChunks->smallChild == NULL ? 0 : m_pChunks->smallChild->treeSize) -
	      (m_pChunks->bigChild   == NULL ? 0 : m_pChunks->bigChild->treeSize) ) <= 1);
  // check pointer structure at root
  assert(m_pChunks->parent == NULL);
  assert(m_pChunks->smallChild == NULL || m_pChunks->smallChild->parent == m_pChunks);
  assert(m_pChunks->bigChild   == NULL || m_pChunks->bigChild->parent   == m_pChunks);
} // void CChunkAllocator<T>::InsertNode(TChunk* pNode)

/**
   @brief Remove a chunk from the balanced tree.
   @param pNode The chunk to be removed.

   Remove a chunk from the tree, keeping the tree balanced.

   The recursive algorithm downwards is as follows, replace the chunk to be removed
   by a copy of the child with the bigger treeSize, then remove that child
   to re-establish the data structure assertions.

   Upwards: size of subtree that replaced current node might have changed and lead to
   unbalancedness: Go up the tree, updating treeSize, checking and balancing.
*/
template < class T >
void CChunkAllocator<T>::DeleteNode(TChunk* pNode) {
  // check balancedness of pNode
  assert(pNode);
  assert(abs( (pNode->smallChild == NULL ? 0 : pNode->smallChild->treeSize) -
	      (pNode->bigChild   == NULL ? 0 : pNode->bigChild->treeSize) ) <= 1);
  // check balancedness of root
  assert(m_pChunks);
  assert(abs( (m_pChunks->smallChild == NULL ? 0 : m_pChunks->smallChild->treeSize) -
	      (m_pChunks->bigChild   == NULL ? 0 : m_pChunks->bigChild->treeSize) ) <= 1);

  // we only remove empty chunks
  assert(pNode->cZombies == m_ChunkSize);

  // allocate a copy, the chunk given as parameter does not have to be touched
  TChunk *pCurrent = new TChunk(*pNode);
  // unlink removed node from children
  if (pCurrent->smallChild != NULL) pCurrent->smallChild->parent = pCurrent;
  if (pCurrent->bigChild   != NULL) pCurrent->bigChild->parent   = pCurrent;
  if (pCurrent->parent     != NULL) {
    if (pCurrent->objects < pCurrent->parent->objects) {
      assert(pCurrent->parent->objects - pCurrent->objects >= m_ChunkSize);
      assert(pCurrent->parent->smallChild == pNode);
      pCurrent->parent->smallChild = pCurrent;
    } else {
      assert(pCurrent->objects - pCurrent->parent->objects >= m_ChunkSize);
      assert(pCurrent->parent->bigChild == pNode);
      pCurrent->parent->bigChild = pCurrent;
    } // endif: current node is small child or big child of parent
  } else {
    // we are replacing the root
    m_pChunks = pCurrent;
  } // endif: parent is NULL?
  // reset pointers of removed node
  pNode->smallChild = pNode->bigChild = NULL;
  pNode->treeSize = 1;
  pNode->parent = NULL;

  // from now on, we may modify and overwrite the members of pNode

  // going DOWNWARDS

  // pointer on the deeper child
  assert(pCurrent);
  bool blSmallIsHeavier = pCurrent->bigChild == NULL || ( pCurrent->smallChild != NULL &&
							 pCurrent->smallChild->treeSize > pCurrent->bigChild->treeSize );
  TChunk *pDeeperChild = (blSmallIsHeavier ? pCurrent->smallChild : pCurrent->bigChild );

  // loop to avoid recursion
  while (pDeeperChild != NULL) {
    assert(pCurrent);
    // check balancedness
    assert(abs( (pCurrent->smallChild == NULL ? 0 : pCurrent->smallChild->treeSize) -
		(pCurrent->bigChild   == NULL ? 0 : pCurrent->bigChild->treeSize  ) ) <= 1);

    // replace current chunk by deeper child
    pCurrent->objects = pDeeperChild->objects;
    pCurrent->cZombies = pDeeperChild->cZombies;

    if (blSmallIsHeavier) {
      // Replaced by smaller child!

      // check sorting
      assert(pCurrent->bigChild == NULL ||
	     pCurrent->objects + m_ChunkSize <= pCurrent->bigChild->objects );
      // update treeSize, it's the smaller one
      pCurrent->treeSize = (pCurrent->bigChild == NULL ? 0 : pCurrent->bigChild->treeSize) + 1;
    } else {
      // Replaced by bigger child!

      // check sorting
      assert(pCurrent->smallChild == NULL ||
	     pCurrent->objects > pCurrent->smallChild->objects );
      // update treeSize, it's the smaller one
      pCurrent->treeSize = (pCurrent->smallChild == NULL ? 0 : pCurrent->smallChild->treeSize) + 1;
    } // endif: smaller child is the deeper one

    // next
    pCurrent = pDeeperChild;
    blSmallIsHeavier = pCurrent->bigChild == NULL || ( pCurrent->smallChild != NULL &&
						      pCurrent->smallChild->treeSize > pCurrent->bigChild->treeSize );
    pDeeperChild = (blSmallIsHeavier ? pCurrent->smallChild : pCurrent->bigChild );
  } // endwhile pDeeperChild != NULL

  // Now free leaf node. Together with allocation of pCopy, total number of allocated chunks is hence not modified.
  // Not that ojects in objects array are still available in either parent or original pNode.
  assert(pCurrent);
  // pCurrent is leaf
  assert(pCurrent->bigChild == NULL);
  assert(pCurrent->smallChild == NULL);
  assert(pCurrent->treeSize == 1);
  if (pCurrent->parent == NULL) {
    // remove root node
    assert(pCurrent == m_pChunks);
    delete m_pChunks;
    m_pChunks = NULL;
  } else {
    // update parent's child pointer
    TChunk *pDeletable = pCurrent;
    pCurrent = pDeletable->parent;
    if (pDeletable == pCurrent->smallChild) {
      pCurrent->smallChild = NULL;
    } else {
      assert(pDeletable == pCurrent->bigChild);
      pCurrent->bigChild = NULL;
    } // endif: deletable is small of big child of parent
    delete pDeletable;

    // parent might need update of its treeSize counter
    // furthermore, parent might have become unbalanced
    // => re-balance tree

    // all the way up the root
    while (pCurrent != NULL) {
      // update treeSize
      pCurrent->treeSize = ( (pCurrent->smallChild == NULL ? 0 : pCurrent->smallChild->treeSize) +
			     (pCurrent->bigChild   == NULL ? 0 : pCurrent->bigChild->treeSize) +
			     1 );
      // balance node
      if (CheckAndBalance(pCurrent))
	// tree had been reorganized, extra go-up
	pCurrent = pCurrent->parent;
      assert(pCurrent);
      pCurrent = pCurrent->parent;
    } // endwhile go up to the root

  } // endif: removed last node

  // check balancedness of root
  assert(m_pChunks);
  assert(abs( (m_pChunks->smallChild == NULL ? 0 : m_pChunks->smallChild->treeSize) -
	      (m_pChunks->bigChild   == NULL ? 0 : m_pChunks->bigChild->treeSize) ) <= 1);
  // check pointer structure at root
  assert(m_pChunks->parent == NULL);
  assert(m_pChunks->smallChild == NULL || m_pChunks->smallChild->parent == m_pChunks);
  assert(m_pChunks->bigChild   == NULL || m_pChunks->bigChild->parent   == m_pChunks);
} // void CChunkAllocator<T>::DeleteNode(TChunk* pNode)

/**
   @brief Check the current node/chunk of the tree and balance, if necessary.
   @param pNode The root of the subtree to be balanced.
   @return Root of now balanced subtree (same as argument if unchanged)

   If the current node is unbalanced, that is, on child is by two or more larger than
   the other (two is actually also the maximum by which the children's treeSizes may mismatch),
   find the new root for the subtree. One can find this
   by following the deeper child (let's say that's the left, small child) and then
   stick to the right. The new root candidate is hence a leaf.

   The right child of the left child must exist. Otherwise, as the left subtree was balanced
   itself (!), the left child can have a maximal treeSize of 2 (in this case, the left child
   has another left child which must be a leaf).
   As the current node is unbalanced (!) the right child, if existent, must have a treeSize of
   at most 0, which is impossible. Hence, the right child does not exist and the left child
   does indeed have a left child (otherwise the current node must be balanced, contradiction).
   This is a very special case, which can be treated on its own:

\verbatim
       C
      /       B
     B   =>  / \
    /       A   C
   A

A is a leaf
\endverbatim

   So we can now assume that the root candidate exists, is a leaf and not a child of the current node.

   Remove that root candidate and insert it as new root, having the old root as
   bigChid (smallChild, if deeper child was the big one) and the old deeper child
   as small child (big child respectively).

   Removing the root candidate can cause <i>his</i> parent to become unbalanced, in which cause
   we start to balance recursively. Note that balancind terminates as the next node is a true child
   of the current node and the treeSize of the tree is limited.
   
   IMPORTANT: This function only modifies the subtree given as a parameter, it does not touch the parent
   node and nodes above (except for updating the child pointer of the parent). However, if the subtree
   was reconfigured (success indicator), the treeSize of the subtree must have changed and the parent's deth
   and all treeSizes up to the root node must be updated (possibly triggering further need for balancing).
   So you will usually call this method within a loop up to the root.

   The old root (in the above scenario) becomes the leftmost child of the right child. Rebalancing might become
   necessary in the right subtree as well.

   In total we move one node from the left subtree to the right.

   Prerequesite: The children of pNode must be balanced, all treeSizes with in the subtree must be up-to-date.
   Postrequesite: The subtree is balanced, all treeSizes in subtree are up-to-date
 */
template < class T >
bool CChunkAllocator<T>::CheckAndBalance(TChunk* pNode) {
  assert(pNode);
  // boolean varible which child is the deeper one
  bool blSmallIsHeavier = pNode->bigChild == NULL || ( pNode->smallChild != NULL &&
						      pNode->smallChild->treeSize > pNode->bigChild->treeSize );

  // the new root candidate
  TChunk* pNewRoot;
  // new parent for the old root
  TChunk* pNewParentOfOldRoot = NULL;
  // the parent of the new root
  TChunk* pOldParentOfNewRoot = NULL;
  // The parent of pNode, stays untouched
  TChunk* pParent = pNode->parent;
  // boolean variable which of parent's children pNode is (also true of parent is NULL)
  bool blIsParentsSmallerChild = (pParent == NULL || pParent->smallChild == pNode);
  assert(blIsParentsSmallerChild || pParent->bigChild == pNode);

  // distinguish which child is the deeper one
  if (blSmallIsHeavier) {
    // left child node is truely deeper

    // check balancedness (exit if successful)
    if (pNode->smallChild == NULL || ( pNode->smallChild->treeSize <=
				       (pNode->bigChild == NULL ? 0 : pNode->bigChild->treeSize) + 1 ))
      return false;
    assert(pNode->smallChild);

    // check special case: big child of small child does not exist
    if (pNode->smallChild->bigChild == NULL) {
      assert(pNode->smallChild->smallChild != NULL);
      assert(pNode->smallChild->smallChild->treeSize == 1);
      assert(pNode->smallChild->treeSize == 2);
      assert(pNode->bigChild == NULL);
      pNewRoot = pNode->smallChild;
      // make old root right child of left child
      pNewRoot->bigChild = pNode;
      pNode->parent = pNewRoot;
      // make old root a leaf
      pNode->treeSize = 1;
      pNode->smallChild = NULL;
    } else {
      // no "special case"

      // find new root candidate
      pNewRoot = pNode->smallChild;
      while (pNewRoot->bigChild != NULL) pNewRoot = pNewRoot->bigChild;

      // remove root candidate ...
      pOldParentOfNewRoot = pNewRoot->parent;
      assert(pOldParentOfNewRoot);
      assert(pOldParentOfNewRoot->bigChild == pNewRoot);
      pOldParentOfNewRoot->bigChild = pNewRoot->smallChild;
      if (pOldParentOfNewRoot->bigChild != NULL) pOldParentOfNewRoot->bigChild->parent = pOldParentOfNewRoot;
      //      pNewRoot->smallChild = NULL;
      // do not update treeSizes yet

      // ... and make it new root
      pNewRoot->smallChild = pNode->smallChild;
      if (pNewRoot->smallChild != NULL) pNewRoot->smallChild->parent = pNewRoot;
      pNewRoot->bigChild = pNode->bigChild;
      if (pNewRoot->bigChild != NULL) pNewRoot->bigChild->parent = pNewRoot;
      // old root will become leaf
      pNode->smallChild = pNode->bigChild = NULL;
      pNode->treeSize = 1;

      // find new parent for old root and insert old root
      if (pNewRoot->bigChild == NULL) {
	pNewRoot->bigChild = pNode;
	pNode->parent = pNewRoot;
      } else {
	pNewParentOfOldRoot = pNewRoot->bigChild;
	while (pNewParentOfOldRoot->smallChild != NULL)
	  pNewParentOfOldRoot = pNewParentOfOldRoot->smallChild;
	pNewParentOfOldRoot->smallChild = pNode;
	pNode->parent = pNewParentOfOldRoot;
      } // endif: not-so-deep child is empty 

    } // endif "special case"
  } else {
    // right child is deeper or equal

    // check balancedness (exit if successful)
    if (pNode->bigChild == NULL || ( pNode->bigChild->treeSize <=
				     (pNode->smallChild == NULL ? 0 : pNode->smallChild->treeSize) + 1 ))
      return false;
    assert(pNode->bigChild);

    // check special case: small child of big child does not exist
    if (pNode->bigChild->smallChild == NULL) {
      assert(pNode->bigChild->bigChild != NULL);
      assert(pNode->bigChild->bigChild->treeSize == 1);
      assert(pNode->bigChild->treeSize == 2);
      assert(pNode->smallChild == NULL);
      pNewRoot = pNode->bigChild;
      // make old root right child of left child
      pNode->parent = pNewRoot;
      pNewRoot->smallChild = pNode;
      // make old root a leaf
      pNode->treeSize = 1;
      pNode->bigChild = NULL;
    } else {
      // no "special case"

      // find new root candidate
      pNewRoot = pNode->bigChild;
      while (pNewRoot->smallChild != NULL) pNewRoot = pNewRoot->smallChild;

      // remove root candidate ...
      pOldParentOfNewRoot = pNewRoot->parent;
      assert(pOldParentOfNewRoot);
      assert(pOldParentOfNewRoot->smallChild == pNewRoot);
      pOldParentOfNewRoot->smallChild = pNewRoot->bigChild;
      if (pOldParentOfNewRoot->smallChild != NULL) pOldParentOfNewRoot->smallChild->parent = pOldParentOfNewRoot;
      //      pNewRoot->bigChild = NULL;
      // do not update treeSizes yet

      // ... and make it new root
      pNewRoot->smallChild = pNode->smallChild;
      if (pNewRoot->smallChild != NULL) pNewRoot->smallChild->parent = pNewRoot;
      pNewRoot->bigChild = pNode->bigChild;
      if (pNewRoot->bigChild != NULL) pNewRoot->bigChild->parent = pNewRoot;
      // old root will become leaf
      pNode->smallChild = pNode->bigChild = NULL;
      pNode->treeSize = 1;

      // find new parent for old root and insert old root
      if (pNewRoot->smallChild == NULL) {
	pNewRoot->smallChild = pNode;
	pNode->parent = pNewRoot;
      } else {
	pNewParentOfOldRoot = pNewRoot->smallChild;
	while (pNewParentOfOldRoot->bigChild != NULL)
	  pNewParentOfOldRoot = pNewParentOfOldRoot->bigChild;
	pNewParentOfOldRoot->bigChild = pNode;
	pNode->parent = pNewParentOfOldRoot;
      } // endif: not-so-deep child is empty 

    } // endif "special case"
  } // endif: which child is deeper

  // update parent
  pNewRoot->parent = pParent;
  if (pParent == NULL) {
    // update root!
    m_pChunks = pNewRoot;
  } else {
    if (blIsParentsSmallerChild)
      pParent->smallChild = pNewRoot;
    else
      pParent->bigChild = pNewRoot;
  } // endif: update root

  // now there's one thing left: balance and update treeSize from root candidate's old parent
  TChunk* pCurrent;
  if (pOldParentOfNewRoot != NULL) {
    pCurrent = pOldParentOfNewRoot;
    while (pCurrent != pNewRoot) {
      // update treeSize
      pCurrent->treeSize = ( (pCurrent->smallChild == NULL ? 0 : pCurrent->smallChild->treeSize) +
			     (pCurrent->bigChild   == NULL ? 0 : pCurrent->bigChild->treeSize) +
			     1 );
      // balance node
      if (CheckAndBalance(pCurrent))
	// tree had been reorganized, extra go-up
	pCurrent = pCurrent->parent;
      assert(pCurrent);
      pCurrent = pCurrent->parent;
      assert(pCurrent);
    } // endwhile: rebalance nodes from new root's parent to here
  } // endif: rebalance old parent of new root

  // the same for old root's new parent to here
  if (pNewParentOfOldRoot != NULL) {
    pCurrent = pNewParentOfOldRoot;
    while (pCurrent != pNewRoot) {
      // update treeSize
      pCurrent->treeSize = ( (pCurrent->smallChild == NULL ? 0 : pCurrent->smallChild->treeSize) +
			     (pCurrent->bigChild   == NULL ? 0 : pCurrent->bigChild->treeSize) +
			     1 );
      // balance node
      if (CheckAndBalance(pCurrent))
	// tree had been reorganized, extra go-up
	pCurrent = pCurrent->parent;
      assert(pCurrent);
      pCurrent = pCurrent->parent;
      assert(pCurrent);
    } // endwhile: rebalance nodes from new root's parent to here
  } // endif: rebalance new parent of old root

  // update treeSize
  pNewRoot->treeSize = ( (pNewRoot->smallChild == NULL ? 0 : pNewRoot->smallChild->treeSize) +
			 (pNewRoot->bigChild   == NULL ? 0 : pNewRoot->bigChild->treeSize) +
			 1 );
  
  // check if balanced  
  assert(abs( (pNewRoot->smallChild == NULL ? 0 : pNewRoot->smallChild->treeSize) -
	      (pNewRoot->bigChild   == NULL ? 0 : pNewRoot->bigChild->treeSize) ) <= 1);
  // check pointer structure
  assert(pNewRoot->parent == pParent);
  assert(pNewRoot->smallChild == NULL || pNewRoot->smallChild->parent == pNewRoot);
  assert(pNewRoot->bigChild   == NULL || pNewRoot->bigChild->parent   == pNewRoot);
      
  // done, subtree was reconfigured
  return true;
} // bool CChunkAllocator<T>::CheckAndBalance(TChunk* pNode)

/**
   @brief Find the chunk responsible for an object.
   @param object An object allocated by this allocator.
   @return The chunk that contains the object or NULL if failure.
   
   This performs a binary search and is therefore O(log(n)).
*/
template < class T >
typename CChunkAllocator<T>::TChunk * CChunkAllocator<T>::Lookup(const T* object) {
  // do binary search
  TChunk* result = m_pChunks;
  while (result != NULL) {
    if (object < result->objects)
      result = result->smallChild;
    else if (object < result->objects + m_ChunkSize)
      // found!
      return result;
    else
      result = result->bigChild;
  } // endwhile: lookup chunk

  // not found
  assert(result == NULL);
  return result;
} // CChunkAllocator<T>::Lookup(const T* object)

// ============= End of balanced tree functions =======================

/**
   @brief Destructor.

   Frees objects that are still in buffer.

   Also prints out some statistics for debugging, if tracing is enabled.
   As this is a template class, it depends on the instanciating class whether
   tracing is enabled or not. As tracing only occurs in the destructor which
   is rarely called (for allocators usually once per program execution), this
   should be OK.

   The statistics are interpreted as follows:

<ul>
<li><b>Objects created</b>

Number of calls to Create().</li>
<li><b>Objects copied</b>

Number of calls to Copy().</li>
<li><b>Objects deleted</b>

Number of calls to Del().
The sum of Create() calls ("Objects created") and Copy() calls ("Objects copied") should equal this number
on allocator destruction, unless you use "Clear()" to free objects.</li>
<li><b>Zombies</b>

Number of objects deleted, but not yet freed (still allocated).
<li><b>Objects currently allocated</b>

Number of objects allocated, but not yet activated and deleted, but not yet freed.
<li><b>Objects still active</b>

Number of objects that are still active, that is, objects that have been allocated using this allocator
but not been deleted using Del().
Should be 0 on allocator destruction and equals the difference between the sum of "Objects created"
plus "Objects copied" and the number of "Objects deleted".
</ul>
 */
template< class T >
CChunkAllocator<T>::~CChunkAllocator() {
  Trace(1,"CChunkAllocator Destructor.");

  // trace messages
#ifndef NDEBUG
#ifdef TRACE_ENABLED
  if (TraceLevel >= 2) {
    clog << "CChunkAllocator Statistics:" << endl
	 << "  Chunk size                  : " << m_ChunkSize << endl
	 << "  Objects created             : " << GetNumberOfCreateCalls() << endl
	 << "  Objects copied              : " << GetNumberOfCopyCalls() << endl;
	 << "  Objects deleted             : " << GetNumberOfDelCalls() << endl
	 << "  Zombies                     : " << m_cZombies << endl
	 << "  Objects currently allocated : " << GetNumberOfAllocatedObjects() << endl
	 << "  Objects still active        : " << GetNumberOfObjects() << endl;
  } // endif: TraceLevel >= 2
#endif // TRACE_ENABLED
#endif

  // clean up
  Clear();
} // CChunkAllocator destructor

/**
   @brief Get new object.

   This is the object factory method to create an object.

   The new object is taken from the "current chunk". If necessary, a whole
   new chunk will be allocated.

   @return A "new" object.
 */
template< class T >
T* CChunkAllocator<T>::Create() {
  // allocate new current chunk, if necessary
  if (m_pCurrentChunk == NULL) {
    m_pCurrentChunk = new TChunk;
    m_pCurrentChunk->objects = AllocateChunk();
    assert(m_pCurrentChunk->objects);
    m_pCurrentChunk->cZombies = 0;
    m_iNextObject = 0;
    ++m_cChunks;
    assert(m_cChunks > 0);
    // insert new chunk into balanced tree
    InsertNode(m_pCurrentChunk);
  } // endif: allocate new current chunk
  
  // current chunk is now allocated and still has allocated objects
  assert(m_pCurrentChunk);
  assert(m_pCurrentChunk->cZombies <= m_iNextObject);
  assert(m_iNextObject < m_ChunkSize);

  // return next free object from current chunk
  T* result = &m_pCurrentChunk->objects[m_iNextObject++];

  // was this the last inactive object?
  if (m_iNextObject >= m_ChunkSize) {
    assert(m_iNextObject == m_ChunkSize);
    m_pCurrentChunk = NULL;
  } // current chunk is burnt out

  ++m_cCreateCalls;
  ++m_cObjects;
  return result;
} // CChunkAllocator<T>::Create()

/**
   @brief Factory method to copy an object.

   This is the same as Create(), but does an assignement before returning
   the result. Note that this allocator never uses the template classes
   copy contructor.

   @param obj Object to be copied.
   @return A copy of obj.
 */
template<class T>
T* CChunkAllocator<T>::Copy(const T& obj) {
  T* result = Create();

  // assign
  *result = obj;

  assert(m_cCreateCalls > 0);
  --m_cCreateCalls;
  ++m_cCopyCalls;
  return result;
} // CChunkAllocator<T>::Copy(const T& obj)

/**
   @brief Delete object.

   Mark object as deleted.

   The chunk responsible for the object will not be freed until all its objects have
   been marked deleted.

   The class does not keep track of the deleted objects itself but merely of the number
   of deleted objects for a chunk. Double deletion of an object might therefore not be
   detected and lead to undesired behaviour.

   As always, do only delete objects with the very same allocator you allocated the
   object with.

   @param deletable (Not NULL) The object to be deleted.
                    Consider the object gone after function call..
 */
template< class T >
void CChunkAllocator<T>::Del( T* deletable )
{
  // Get the chunk responsible for the object
  TChunk* pChunk = Lookup(deletable);
  assert(pChunk);
  assert(pChunk->objects <= deletable && deletable - pChunk->objects < m_ChunkSize);

  // update statistics
  assert(pChunk->cZombies < m_ChunkSize);
  ++pChunk->cZombies;
  ++m_cZombies;

  // update statistics
  --m_cObjects;
  ++m_cDelCalls;

  // free chunk, if possible
  if (pChunk->cZombies >= m_ChunkSize) {
    assert(pChunk->cZombies == m_ChunkSize);

    DeleteNode(pChunk);
    FreeChunk(pChunk->objects);

    // update statistics
    assert(m_cChunks > 0);
    --m_cChunks;
    assert(m_cZombies >= m_ChunkSize);
    m_cZombies -= m_ChunkSize;

    delete pChunk;
  } // endif: free chunk
} // CChunkAllocator<T>::Del(...)

/**
   @brief Clear all objects.

   Frees all objects allocated and not yet freed by this allocator.
 */
template< class T >
void CChunkAllocator<T>::Clear() {
  // do search of balanced tree
  vector<TChunk* > searchQueue;
  TChunk* deletable;
  if (m_pChunks != NULL)
    searchQueue.push_back(m_pChunks);

  while (!searchQueue.empty()) {
    deletable = searchQueue.back();
    searchQueue.pop_back();
    if (deletable->smallChild != NULL)
      searchQueue.push_back(deletable->smallChild);
    if (deletable->bigChild != NULL)
      searchQueue.push_back(deletable->bigChild);

    // delete chunk itself
    FreeChunk(deletable->objects);
    delete deletable;
  } // endwhile: do depth first search over balanced tree

  // reset root
  m_pChunks = NULL;

  // reset current chunk
  m_pCurrentChunk = NULL;

  // reset statistics
  m_cChunks = 0;
  m_cZombies = 0;
} // CChunkAllocator<T>::Clear()

#endif // already included?
