#include <stdlib.h>
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif



// ------------------------------------------------------------------------------
//
// Implementation of CSparseArray and CSparseMatrix
//
// ------------------------------------------------------------------------------


// CVS id: $Id: SparseArray.cc 1729 2006-06-26 20:04:33Z vogel $

#include <set>
using std::set;
#include <limits.h>

// enable / disable tracing for this module here
// #undef NO_TRACE
#define NO_TRACE 1
#include <Trace.hh>

#include <SparseArray.hh>


using namespace std;

using std::cout;
using std::endl;

//static char rcsid_SparseArray_cc[] = "$Id:";



CSparseArray::CSparseArray()
{
  Trace( 5, "CSparseArray Constructor" );
  
  m_MinIndex = 0;
  m_MaxIndex = -1;

  m_DefaultValue = 0.0;

  m_Indices.SetIncrement( 8 );
  m_Values.SetIncrement( 8 );
}



CSparseArray::CSparseArray( const CSparseArray & Array )
{
  Trace( 5, "CSparseArray Copy-Constructor" );

  *this = Array;
}



CSparseArray::~CSparseArray()
{
  Trace( 5, "CSparseArray Destructor" );

}



CSparseArray &
CSparseArray::operator = ( const CSparseArray & Array )
{
  //Trace( 0, "CSparseArray::operator = " );

  //Array.Display();

  if ( this == &Array ) 
    // Don't copy on yourself
    return *this;

  m_MinIndex = Array.m_MinIndex;
  m_MaxIndex = Array.m_MaxIndex;

  m_DefaultValue = Array.m_DefaultValue;
  //Trace( 0, "DefaultValues old and new " << Array.m_DefaultValue << m_DefaultValue );

  // copy index and value arrays
  m_Indices = Array.m_Indices;
  m_Values = Array.m_Values;


  return *this;
}



int
CSparseArray::SetAt( int Index, float Value )
{
//cout << "CSparseArray::SetAt " << Index << endl;

  // new minimal or maximal index?
  if ( Index < m_MinIndex )
    m_MinIndex = Index;

  if ( Index > m_MaxIndex )
    m_MaxIndex = Index;

  // now store the new key-value pair
  // if index is new, insert value, otherwise store value in old position
  
  bool IsNew;
  int Position = m_Indices.Insert( Index, IsNew );

  if ( IsNew )
    m_Values.Insert( Position, Value );
  else
    m_Values[ Position ] = Value;

  return Position;
}



float
CSparseArray::GetAt( int Index ) const
{
  int Position;

  // look, if Index is in Index-Array
  // if yes, we know the position and can fetch the corresponding value
  // otherwise, we return the DefaultValue

  if ( m_Indices.Find( Index, Position ) ) {
    if ( m_Values.GetAt( Position ) <= 0.0 ) {
      Trace( 5, "CSparseArray::GetAt " << Index << " <= 0.0" );
    }
    Trace( 5, "CSparseArray::GetAt " << Index << " = " << m_Values.GetAt( Position ) );
    return m_Values.GetAt( Position );
  } else {
    if ( m_DefaultValue <= 0.0 ) {
      Trace( 5, "CSparseArray::GetAt - DefaultValue <= 0.0" );
    }
    Trace( 5, "CSparseArray::GetAt return DefaultValue " << m_DefaultValue );
    return m_DefaultValue;
  }
}


/**
   @brief Fetch an index-value pair.
   @param Position Number if the index-value pair. Must be between 0 and NumberOfItems()-1.
   @param Index Call by reference variable to return the index.
   @param Value Call by reference variable to return the value.

   This method provides the possibility to iterate over the items of a sparse
   array that actually contain data.
*/
bool
CSparseArray::GetItem( int Position, int &Index, float &Value ) const
{
  if ( ( Position >= m_Indices.MinIndex() ) && 
       ( Position <= m_Indices.MaxIndex() ) ) {
    Index = m_Indices.CArray<int>::operator[]( Position );
    Value = m_Values[ Position ];
    Trace( 5, Index << " " << Value );
    return TRUE;
  } else {
    Index = 0;
    Value = m_DefaultValue;
    return FALSE;
  }
}



float
CSparseArray::operator[]( int Index ) const
{
  // for read only, i.e. only r-value
  // compare GetAt()

  int Position;
  
  if ( m_Indices.Find( Index, Position ) ) {
    return m_Values.GetAt( Position );
  } else {
    return m_DefaultValue;
  }
}



float&
CSparseArray::operator [] ( int Index )
{
  // for writing, i.e. l-value;
  // if Index not stored, we first have to insert the new
  // Index-Value-Pair; then we can return a reference to 
  // the new Value-Item

  // new minimal or maximal index?
  if ( Index < m_MinIndex )
    m_MinIndex = Index;

  if ( Index > m_MaxIndex )
    m_MaxIndex = Index;

  // now store the new index-value-pair
  bool IsNew;
  
  int Position = m_Indices.Insert( Index, IsNew );
  if ( IsNew )
    m_Values.Insert( Position, m_DefaultValue );

  return m_Values[ Position ];
}



CSparseArray
CSparseArray::operator * ( float Value ) const
{
  Trace( 5, "CSparseArray::operator * " );

  CSparseArray A;
  int Index;
  int Position;

  for ( Position = 0; Position <= m_Indices.MaxIndex(); Position++ ) {
    Index = m_Indices.CArray<int>::operator[]( Position );
    A.m_Indices.CArray<int>::operator[]( Position ) = Index;
    A.m_Values[Position] = m_Values[ Position ] * Value;

    //A.(*m_Indices)[Position] = (*m_Indices)[Position];
    //A.(*m_Values)[Position] = (*m_Values)[Position] * Value;
  }
  
  A.m_DefaultValue = m_DefaultValue * Value;
  //A.Display();

  return A;
}



CSparseArray
CSparseArray::operator + ( const CSparseArray & Array ) const
{
  Trace( 5, "CSparseArray::operator +  ( const CSparseArray & Array ) " );

  int ThisPos = 0;
  int ArrayPos = 0;
  int NewPos = 0;

  int MaxThisPos = m_Indices.MaxIndex();
  int MaxArrayPos = Array.m_Indices.MaxIndex();

  int ThisIndex = m_Indices.CArray<int>::operator[]( ThisPos );
  int ArrayIndex = Array.m_Indices.CArray<int>::operator[]( ArrayPos );

  int MaxThisIndex = m_Indices.CArray<int>::operator[]( MaxThisPos );
  int MaxArrayIndex = Array.m_Indices.CArray<int>::operator[]( MaxArrayPos );

  int MaxIndex = ( MaxThisIndex > MaxArrayIndex ? MaxThisIndex : MaxArrayIndex );

  CSparseArray New;

  while ( ( ThisPos <= MaxThisPos ) || ( ArrayPos <= MaxArrayPos ) ) {
    Trace( 5, "" << ThisPos << " " << MaxThisPos << " "  << ThisIndex << " "
	   << ArrayPos << " " << MaxArrayPos  << " " << ArrayIndex );

    // Index in both Arrays
    if ( ThisIndex == ArrayIndex ) {
      New.m_Indices.CArray<int>::operator[]( NewPos ) = ThisIndex;
      New.m_Values[ NewPos ] = m_Values[ ThisPos ] + Array.m_Values[ ArrayPos ];
      ThisPos++;
      ArrayPos++;
    } else { 
      if ( ThisIndex < ArrayIndex ) {
	New.m_Indices.CArray<int>::operator[]( NewPos ) = ThisIndex;
	New.m_Values[ NewPos ] = m_Values[ ThisPos ];
	ThisPos++;
      } else {
	New.m_Indices.CArray<int>::operator[]( NewPos ) = ArrayIndex;
	New.m_Values[ NewPos ] = Array.m_Values[ ArrayPos ];
	ArrayPos++;
      }
    }

    NewPos++;

    if ( ThisPos <= MaxThisPos ) {
      ThisIndex = m_Indices.CArray<int>::operator[]( ThisPos );
    } else {
      ThisIndex = MaxIndex + 1;
    }

    if ( ArrayPos <= MaxArrayPos ) {
      ArrayIndex = Array.m_Indices.CArray<int>::operator[]( ArrayPos );
    } else {
      ArrayIndex = MaxIndex + 1;
    }
  }
  
  New.Display();

  return New;
}


/**
   @brief Remove last element.
   @return Success indicator. FALSE iff array was already empty.

   Removes the last element of the sparse array. Updates MaxIndex to the next smallest element that contains data.
 */
bool CSparseArray::DeleteLast()
{
  assert(m_Values.Size() == m_Indices.Size());
  // array already empty?
  if (m_Values.Size() == 0)
    // already empty
    return FALSE;

  // remove last item
  if (m_Values.Size() == 1) {
    // remove last item
    m_Values.Clear();
    m_Indices.Clear();
    m_MinIndex=0;
    m_MaxIndex=-1;
  } else {
    assert(m_Indices.Size() > 1);
    m_Values.SetMaxIndex(m_Values.MaxIndex()-1);
    ((CArray<int> &)m_Indices).SetMaxIndex(m_Indices.MaxIndex()-1);
    assert(m_Indices.Size() > 0);
    m_MaxIndex=m_Indices.GetAt(m_Indices.MaxIndex());
    assert(m_Values.Size() == m_Indices.Size());
  } // endif: remove last item?

  return TRUE;
} // BOOL CSparseArray::DeleteLast()



void
CSparseArray::Add( const CSparseArray & Array )
{
  Trace( 3, "CSparseArray::Add" );
  Trace( 4, "Size This:  " << m_Indices.Size() );
  Trace( 4, "Size Array: " << Array.m_Indices.Size() );

  // we add the default values
  m_DefaultValue += Array.m_DefaultValue;
  Trace( 5,  "DefaultValue = " << m_DefaultValue );

  // if Array is empty, then nothing to do
  if ( Array.m_Indices.Size() <= 0 ) {
    Trace( 5, "Array is empty" );
    return;
  }

  // this empty
  if ( m_Indices.Size() <= 0 ) {
    Trace( 5, "this is empty" );
    *this = Array;
    return;
  }

  // cast of SortedArray to normal Array, to have direct access
  CArray<int> & Indices = m_Indices;
  const CArray<int> & ArrayIndices = Array.m_Indices;

  int ThisPos = 0;
  int ArrayPos = 0;

  int ThisMaxPos = m_Indices.MaxIndex();
  int ArrayMaxPos = Array.m_Indices.MaxIndex();

  int ThisIndex = Indices[ ThisPos ];
  int ArrayIndex = ArrayIndices[ ArrayPos ];

  Trace( 5, "ThisIndex " << ThisIndex 
	 << "  ArrayIndex " << ArrayIndex );

  // use new arrays to avoid insertions, as they need repeated copying
  CNumericalArray<int> NewIndices;
  CNumericalArray<float> NewValues;
  
  //Trace( 1, "SetBoundaries: " << m_Indices.Size() + ArrayIndices.Size() );
  NewIndices.SetIncrement( m_Indices.Size() + ArrayIndices.Size() );
  NewValues.SetIncrement( m_Indices.Size() + ArrayIndices.Size() );

  // run over both arrays; if indices are equal, add values;
  // otherwise take the value from the array with the smaller index

  while ( (ThisPos <= ThisMaxPos) && (ArrayPos <= ArrayMaxPos) ) {
    ThisIndex = Indices[ ThisPos ];
    ArrayIndex = ArrayIndices[ ArrayPos ];

    if ( ThisIndex == ArrayIndex ) {
      Trace( 5, "ThisIndex == ArrayIndex " << ThisIndex );
      NewIndices.Append( Indices[ ThisPos ] );
      NewValues.Append( m_Values[ ThisPos ] + Array.m_Values[ ArrayPos ] );
      ThisPos++;
      ArrayPos++;
    } else if ( ThisIndex < ArrayIndex ) {
      Trace( 5, "ThisIndex < ArrayIndex " << ThisIndex << " " << ArrayIndex);
      NewIndices.Append( Indices[ ThisPos ] );
      NewValues.Append( m_Values[ ThisPos ] );
      ThisPos++;
    } else {
      Trace( 5, "ThisIndex > ArrayIndex " << ThisIndex << " " << ArrayIndex);
      NewIndices.Append( ArrayIndices[ ArrayPos ] );
      NewValues.Append( Array.m_Values[ ArrayPos ] );
      ArrayPos++;
    }

    Trace( 5, "Indices are now: " << ThisIndex << " " << ArrayIndex );
  }

  // if any items are left in one of the arrays, copy them to the new one
  // note: at most one of the two while-conditions will hold

  while ( ThisPos <= ThisMaxPos ) {
    ThisIndex = Indices[ ThisPos ];
    Trace( 5, "ThisIndex  " << ThisIndex );
    NewIndices.Append( Indices[ ThisPos ] );
    NewValues.Append( m_Values[ ThisPos ] );
    ThisPos++;
  }
 
  while ( ArrayPos <= ArrayMaxPos ) {
    ArrayIndex = ArrayIndices[ ArrayPos ];
    Trace( 5, "ArrayIndex " << ArrayIndex );
    NewIndices.Append( ArrayIndices[ ArrayPos ] );
    NewValues.Append( Array.m_Values[ ArrayPos ] );
    ArrayPos++;
  }
 
  // now, copy back to m_Indices and m_Values
  Trace( 4, "Now, copy back" );
  int i;

  for ( i = NewIndices.MinIndex(); i <= NewIndices.MaxIndex(); i++ ) {
    Trace( 5, "i=" << i <<  "  " << NewIndices[ i ] << " " << NewValues[ i ] );
    Indices[ i ] = NewIndices[ i ];
    m_Values[ i ] = NewValues[ i ];
  }

}



float
CSparseArray::SumOfStoredItems() const
{
  float Sum = 0.0;
  int Position;

  Trace( 5, "SumOfStoredItems" );
  for (Position = 0; Position <= m_Values.MaxIndex(); Position++) {
    Trace( 5, Position << " " << m_Values.GetAt( Position ) );
    Sum += m_Values.GetAt( Position );
  }

  return Sum;
}



float
CSparseArray::Sum( int NumOfItems ) const
{
  Trace( 5, "CSparseArray::Sum" );
  if ( NumOfItems < 0 ) {
    NumOfItems = NumberOfItems();
  }

  int NumOfStoredItems = m_Indices.Size();

  Trace( 5, NumOfItems << " " << NumOfStoredItems  << " "
	 << SumOfStoredItems()  << " " << m_DefaultValue );

  return SumOfStoredItems() + (NumOfItems - NumOfStoredItems) * m_DefaultValue;
}



void
CSparseArray::Clear()
{
  // return allocated memory
  m_Indices.RemoveAll();
  m_Values.RemoveAll();

  m_MinIndex = 0;
  m_MaxIndex = -1;

  m_DefaultValue = 0.0;
}


void
CSparseArray::Display() const
{
  int T = TraceLevel;
  TraceLevel = 0;

  std::cout << "MinIndex = " << m_Indices.MinIndex() << "   ";
  std::cout << "MaxIndex = " << m_Indices.MaxIndex() << "   ";
  std::cout << "DefaultValue = " << m_DefaultValue << std::endl;
  for (int k = m_Indices.MinIndex(); k <= m_Indices.MaxIndex(); k++){
    std::cout << k << " - " << m_Indices.GetAt( k ) 
	 << " : " << m_Values.GetAt( k ) << std::endl;
  }

  TraceLevel = T;
}
  

bool
CSparseArray::Store( std::ofstream &Stream ) const
{
  Trace( 5, "CSparseArray::Store( ofstream &Stream )" );

  int Index;
  float Value;

  Stream << NumberOfItems() << std::endl;
  Stream << m_DefaultValue << std::endl;

  int k;
  for ( k = 0; k <= m_Indices.MaxIndex(); k++ ) {
    Index = m_Indices.GetAt( k );
    Value = m_Values[ k ];
    assert( Index >= 0 );
    //assert( Index <= 100000 );
    assert( Value >= 0.0 );
    assert( Value <= 1.0 );

    Stream << Index << " " << Value << std::endl;
    //Stream << m_Indices.GetAt( k ) << " " << m_Values[ k ] << std::endl;
  }

  return TRUE;
}



bool
CSparseArray::Store( const char * pFileName ) const
{
  Trace( 5, "CSparseArray::Store( const char * pFileName )" );

  std::ofstream OutStream( pFileName );

  if ( !OutStream ) {
    Trace( 0,  "File " << pFileName << " could not be opened" );
    return FALSE;
  }
  Store( OutStream );
  OutStream.close();

  return TRUE;
}



bool
CSparseArray::Load( ifstream &Stream )
{
  int NumberOfStoredItems;

  Stream >> NumberOfStoredItems;
  // hack to detect corrupted files
  Trace( 1, "CSparseArray::Load: NumberOfStoredItems = " << NumberOfStoredItems );
  if ( NumberOfStoredItems > 100000 )
    return FALSE;

  Stream >> m_DefaultValue;
  //std::cout << NumberOfStoredItems << " " << m_DefaultValue << std::endl;

  Trace( 5, "CSparseArray::Load: DefaultValue = " << m_DefaultValue );

  //std::cout << "CSparseArray::Load: NumberOfStoredItems = " << NumberOfStoredItems
  //     << std::endl << std::flush;

  m_Indices.Clear();
  m_Values.Clear();

  m_Indices.SetIncrement( NumberOfStoredItems );
  m_Values.SetIncrement( NumberOfStoredItems );

  int Index;
  float Value;
  int k;
  for ( k = 0; k < NumberOfStoredItems; k++ ) {
    Stream >> Index >> Value;
    Trace( 2, k << " " << Index << " " << Value );
    // treat SortedArray as normal Array to prevent binary search
    // when writing to the Index-Array
    m_Indices.CArray< int >::operator[]( k ) = Index;
    m_Values[ k ] = Value;
  }
  //std::cout.std::flush();
  return TRUE;
}
  



bool
CSparseArray::Load( const char * pFileName )
{
  ifstream InStream( pFileName );

  if ( !InStream ) {
    Trace( 0,  "File " << pFileName << " could not be opened" );
    return FALSE;
  }
  bool rc = Load( InStream );
  InStream.close();

  return rc;
}



//void
//CSparseArray::RemoveSmallValues()
//{
  /*
  int i, j = 0;
  for (i = 0; i < m_Size; i++) {
    if (m_Values[i] >= m_DefaultValue) {
      m_Indices[j] = m_Indices[i];
      m_Values[j] = m_Values[i];
      j++;
    }
  }
  //if (m_Size != j)
  //  std::cout << "old size " << m_Size << "  new size " << j << std::endl;

  m_Size = j;

  */
//}

/**
   @brief "Insert" an empty element.
   @param idx Index of the "new" element.

   This method actually does not insert anything. It just shifts all the elements
   with index idx and greater one back and updates the dimensions. However, if idx
   lies beyond the dimensions of the array (in the sense of "greater than the maximal
   index"), this method does not do anything, that is, dimensions will not be updated
   so they contain idx. Correspondingly, if idx is smaller than the minimal index,
   the minimal index will be increased by one.
 */
void CSparseArray::Insert(int idx) {
  if (idx <= MaxIndex()) {
    // shift indices after idx by one
    for ( int i = m_Indices.MaxIndex(); i >= m_Indices.MinIndex() && i >= idx; --i )
      // treat SortedArray as normal Array to prevent binary search
      // when writing to the Index-Array
      m_Indices.CArray< int >::operator[]( i ) += 1;
    //      m_Indices[i] += 1;

    // update dimensions
    m_MaxIndex += 1;
    if (idx <= m_MinIndex)
      ++m_MinIndex;
  } // endif: idx lies within array dimensions
} // CSparseArray::Insert(int idx) 





// -------------------------------------------------------
//
// CSparseMatrix
//
// -------------------------------------------------------

CSparseMatrix::CSparseMatrix()
{
  Trace( 5, "CSparseMatrix - Constructor" );

  m_DefaultValue = 0.0;
  m_MinCol = 0;
  m_MaxCol = -1;     // because we don't have a column yet
}


CSparseMatrix::CSparseMatrix( int NumOfRows, int NumOfColumns )
{
  Trace( 5, "CSparseMatrix - Constructor(NumOfRows, NumOfColumns)" );

  m_DefaultValue = 0.0;

  m_MinCol = 0;
  m_MaxCol = NumOfColumns - 1;
	
  m_Rows.SetBoundaries( 0, NumOfRows - 1 );
}



CSparseMatrix::~CSparseMatrix()
{
  Trace( 5, "CSparseMatrix - Destructor" );
}



bool 
CSparseMatrix::SetBoundaries( int MinRow, int MaxRow, 
			      int MinCol, int MaxCol )
{
  if ( !m_Rows.SetBoundaries( MinRow, MaxRow ) ) {
    return FALSE;
  }

  // boundaries for each row must be set !!!

  // we create all the rows, still empty
  
  m_MinCol = MinCol;
  m_MaxCol = MaxCol;

  return TRUE;
}



void
CSparseMatrix::SetDefaultValue( float Value )
{ 
  m_DefaultValue = Value;
  // set default value for each row
  int Row;
  for ( Row = m_Rows.MinIndex(); Row <= m_Rows.MaxIndex(); Row++ ) {
    m_Rows[ Row ].SetDefaultValue( Value );
  }  
}



void
CSparseMatrix::SetDefaultValue( int Row, float Value )
{ 
  if ( m_Rows.IsValidIndex( Row ) )
    m_Rows[ Row ].SetDefaultValue( Value );
}



void
CSparseMatrix::SetIncrement( int Increment )
{
  if ( Increment <= 0 )
    Increment = 1;
  for ( int Row = m_Rows.MinIndex(); Row <= m_Rows.MaxIndex(); Row++ ) {
    m_Rows[ Row ].SetIncrement( Increment );
  }  

  m_Rows.SetIncrement( Increment );
}




/*
bool
CSparseMatrix::Exists( int , int ) const
{
  return FALSE;
}
*/


// modify all cells in the matrix
void
CSparseMatrix::Add( float Value )
{
  for (int Row = m_Rows.MinIndex(); Row <= m_Rows.MaxIndex(); Row++ ) {
    m_Rows[ Row ].Add( Value );
  }
}


void
CSparseMatrix::Sub( float Value )
{
  for (int Row = m_Rows.MinIndex(); Row <= m_Rows.MaxIndex(); Row++) {
    m_Rows[ Row ].Sub( Value );
  }
}


void
CSparseMatrix::Mul( float Value )
{
  for ( int Row = m_Rows.MinIndex(); Row <= m_Rows.MaxIndex(); Row++ ) {
    m_Rows[ Row ].Mul( Value );
  }
}



void
CSparseMatrix::Div( float Value )
{
  for ( int Row = m_Rows.MinIndex(); Row <= m_Rows.MaxIndex(); Row++ ) {
    m_Rows[ Row ].Div( Value );
  }
}



void
CSparseMatrix::Mul( const CNumericalArray< float > &Values )
{
  // multiply each row in the matrix with the corresponding
  // component in the vector
  // do this only for overlapping indices

  Trace( 5, "CSparseArray::Mul( CArray )" );

  int MinIndex, MaxIndex;

  MinIndex = MinRow();
  if ( Values.MinIndex() > MinIndex )
    MinIndex = Values.MinIndex();

  MaxIndex = MaxRow();
  if ( Values.MaxIndex() < MaxIndex )
    MaxIndex = Values.MaxIndex();

  Trace( 5, "Boundaries: " << MinIndex << " " << MaxIndex );

  int Index;
  for ( Index = MinIndex; Index <= MaxIndex; Index++ ) {
    Trace( 5, "Index: " << Index << " " << Values[ Index ] );
    m_Rows[ Index ].Mul( Values[ Index ] );
  }  

}



// add to matrices
void
CSparseMatrix::Add( const CSparseMatrix & Matrix )
{
  Trace( 5, "CSparseMatrix::Add" );
  Trace( 5, "This:   MinRow = " << MinRow() << "  MaxRow = " << MaxRow() );
  Trace( 5, "Matrix: MinRow = " << Matrix.MinRow()
	 << "  MaxRow = " << Matrix.MaxRow() );

  // adapt the boundaries
  m_Rows.AssureBoundaries( Matrix.MinRow(), Matrix.MaxRow() );

  int Row = Matrix.MinRow();
  while ( Row <= Matrix.MaxRow() ) {
    Trace( 5, "Row = " << Row );
    m_Rows[Row].Add( ((CSparseMatrix &)Matrix).GetRow( Row ) );
    Row++;
  }
}



void
CSparseMatrix::Log()
{
  Trace( 5, "starting convertion to Log" );

  if ( NumOfRows() == 0 )
    return;

  int Row;
  for ( Row = m_Rows.MinIndex(); Row <= m_Rows.MaxIndex(); Row++ ) {
    Trace( 5, "converting row: " << Row << " to Log" );
    m_Rows[ Row ].Log();
  }
}



void
CSparseMatrix::Flip( const CSparseMatrix &Matrix )
{
  Trace( 3, "CSparseMatrix::Flip" );
  Trace( 5, Matrix.m_DefaultValue );

  int Row, Col, i;
  float Value;

  // remove all items in this matrix
  Clear();

  // set boundaries: rows <-> cols
  Trace( 5, "MinCol = " << Matrix.MinCol() << " MaxCol = " <<  Matrix.MaxCol() );
  SetBoundaries( Matrix.MinCol(), Matrix.MaxCol(), Matrix.MinRow(), Matrix.MaxRow() ); 
  //m_Rows.SetBoundaries( Matrix.MinCol(), Matrix.MaxCol() );

  for ( Row = Matrix.m_Rows.MinIndex(); Row <= Matrix.m_Rows.MaxIndex(); Row++) {
    Trace( 5, "Row: " << Row << " Items: " << Matrix.m_Rows[ Row ].NumberOfItems() );
    for ( i = 0; i < Matrix.m_Rows[ Row ].NumberOfItems(); i++ ) {
      Matrix.m_Rows[ Row ].GetItem( i, Col, Value );
      m_Rows[ Col ].SetAt( Row, Value );
      //Trace( 0; "e - f: " << 
    }
  }

  // set global default value, no individual default values for rows
  SetDefaultValue( Matrix.m_DefaultValue );

}



void
CSparseMatrix::Display() const
{
  Trace( 3, "CSparseMatrix::Display()" );

  // occasion to use rcsid
  //std::cout << std::endl << rcsid_matrix_cc;
  //std::cout << "\n";
  if ( NumOfRows() == 0 )
    return;

  int Row;
  for ( Row = m_Rows.MinIndex(); Row <= m_Rows.MaxIndex(); Row++ ) {
    std::cout << "row " << Row << ":" ;
    m_Rows[ Row ].Display();
    std::cout << "\n";
  }  
}


void
CSparseMatrix::Clear()
{
  //clog << "CSparseMatrix::Clear()" << std::endl;
  // delete the rows
  m_DefaultValue = 0.0;

  int Row;
  for ( Row = m_Rows.MinIndex(); Row <= m_Rows.MaxIndex(); Row++ ) {
    m_Rows[ Row ].Clear();
  }
}



bool
CSparseMatrix::Store( std::ofstream &Stream ) const
{
  //std::cout << "CSparseMatrix::Store" << std::endl << std::flush;


  Stream << m_Rows.MinIndex() << std::endl;
  Stream << m_Rows.MaxIndex() << std::endl;
  Stream << m_MinCol << std::endl;
  Stream << m_MaxCol << std::endl;

  if ( m_Rows.Size() == 0 )
    return TRUE;
  
  int Row;
  for ( Row = m_Rows.MinIndex(); Row <= m_Rows.MaxIndex(); Row++ ) {
    Trace( 3, "Store " << Row << " " << m_Rows[ Row ].NumberOfItems() );
    m_Rows[ Row ].Store( Stream );
  }  
  return TRUE;
}


bool
CSparseMatrix::Load( ifstream &Stream )
{
  // read boudaries
  int MinRow, MaxRow;

  Stream >> MinRow;
  Stream >> MaxRow;
  Stream >> m_MinCol;
  Stream >> m_MaxCol;

  Trace(3, "Rows: " << MinRow << " - " << MaxRow << " "
	<< "Cols: " << m_MinCol << " - " << m_MaxCol << std::endl );
  
  // remove old content
  Clear();

  m_Rows.SetBoundaries( MinRow, MaxRow );

  // read the lexicon entries
  int Row;
  for ( Row = MinRow; Row <= MaxRow; Row++ ) {
    Trace( 5, "Load Row " << Row );
    //std::cout << Row << " ";
    if ( !(m_Rows[ Row ].Load( Stream )) ) {
      cerr << "Error in loading data in row " << Row << std::endl;
      exit( -2 );
    }
  }
  //Print();
  return TRUE;
}

/**
   Get the indices of the rows and columns of the matrix which acutally contain data.

   As this is a <em>sparse</em> matrix, not necessarily all cells contain data.
   This method collects the indices of all rows and columns that contain at least one
   non-default entry.
   The resulting indices are not necessarily sorted.

   @param rowResult A container object where the inidces of the used row indices are stored.
   Will be cleared before use.
   @param colResult A container object where the inidces of the used columns indices are stored.
   Will be cleared before use.
 */
void CSparseMatrix::CollectUsedIndices( CSparseMatrix::CIndexContainer& rowResult,
					CSparseMatrix::CIndexContainer& colResult ) const
{
  Trace(1,"CSparseMatrix::collectUsedIndices()");

  // clear result containers
  rowResult.clear();
  colResult.clear();
  // container of already seen column indices
  typedef set<int> CSetOfIndices;
  CSetOfIndices seenColIndices;

  // dummy variables for indices
  int idxRow, idxCol;
  // number of non-empty sparse column in current row
  int nrCol;

  // iterate over all row indices
  for (idxRow = m_Rows.MinIndex(); idxRow <= m_Rows.MaxIndex(); ++idxRow) {
    // skip empty rows
    if (m_Rows[idxRow].NumberOfItems() > 0) {
      // pushed non-empty row into row result container
      rowResult.push_back(idxRow);
      // get row
      const CSparseArray &row = m_Rows[idxRow];
      // assert(row.Size() > 0);
      // now collect column indices
      for (nrCol = 0; nrCol < row.NumberOfItems(); ++nrCol) {
	// get index of current column
	idxCol = row.GetIndexOfItem(nrCol);
	// already seen?
	if (seenColIndices.find(idxCol) == seenColIndices.end()) {
	  // new column index
	  colResult.push_back(idxCol);
	  seenColIndices.insert(idxCol);
	} // col index already seen?
      } // endfor: all columns of current row
    } // endif: current row empty?
  } // endfor: all rows
} // CSparseMatrix::collectRowIndices(...)

/**
   Update min/max column/row number.
   This might be necessary if someone accessed the internal structures
   e.g. via GetRow directly (e.g. a store or load function).
*/
void CSparseMatrix::UpdateDimensions() {
  Trace(1,"CSparseMatrix::UpdateDimensions()");

  // dummy variables for indices
  int idxRow, idxCol;
  // number of non-empty sparse column in current row
  int nrCol;

  // reset
  m_MinCol = INT_MAX;
  m_MaxCol = INT_MIN;

  // iterate over all row indices
  for (idxRow = m_Rows.MinIndex(); idxRow <= m_Rows.MaxIndex(); ++idxRow) {
    // get row
    const CSparseArray &row = m_Rows[idxRow];
    // now collect column indices
    for (nrCol = 0; nrCol < row.NumberOfItems(); ++nrCol) {
      // get index of current column
      idxCol = row.GetIndexOfItem(nrCol);
      // update column minimum and maximum
      if (idxCol < m_MinCol)
	m_MinCol = idxCol;
      if (idxCol > m_MaxCol)
	m_MaxCol = idxCol;
    } // endfor: all columns of current row
  } // endfor: all rows
} // CSparseMatrix::UpdateDimensions()

/**
   @brief Insert a new empty row into the matrix.
   @param idx The index of the new empty row.

   If row idx already exists, row number idx and following will be
   shifted by one. Otherwise the boundaries of the matrix will be updated
   to assure that idx is a valid index.
 */
void CSparseMatrix::InsertRow(int idx) {
  m_Rows.Insert(idx,CSparseArray());
} // CSparseMatrix::InsertRow(int)

/**
   @brief Insert a new empty column into the matrix.
   @param idx The index of the new column.

   This will influence the matrix in the way that data in columns after
   and including idx will be shifted by one. However, if the index of the new column
   lies outside the current boundaries of the matrix , this function will <i>not</i>
   update the dimensions of the matrix to assure that they contain idx. This is due
   to the fact that this is a sparse matrix, the biggest / smallest column is the
   biggest / smallest column that actually contains data.
 */
void CSparseMatrix::InsertColumn(int idx) {
  if (idx <= m_MaxCol) {
    for (int i = m_Rows.MinIndex(); i <= m_Rows.MaxIndex(); ++i)
      m_Rows[i].Insert(idx);

    // update dimensions
    ++m_MaxCol;
    if (idx <= m_MinCol)
      ++m_MinCol;
  } // endif: idx is withing MaxCol
} // CSparseMatrix::InsertColumn(int idx)
