#include <stdlib.h>

// ------------------------------------------------------------------------------
//
// CSparseArray and CSparseMatrix
//
// Container to store index - value pairs
// and to perform mathematical operations
//
// ------------------------------------------------------------------------------


// $Id: SparseArray.hh 1729 2006-06-26 20:04:33Z vogel $

#ifndef SparseArray_hh
#define SparseArray_hh

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include <fstream>
using std::ifstream;
using std::ofstream;

#include <vector>
using std::vector;

#include <stdlib.h>   // added by Joy, Sep 17, 4:28pm, to make exit( -2 ) legal

#include <NumericalArray.hh>
#include <SortedArray.hh>


class CSparseArray
{
public:
  CSparseArray();
  CSparseArray( const CSparseArray & Array );

  ~CSparseArray();

  CSparseArray & operator = ( const CSparseArray & Array );

  // Remove all entries
  void Clear();

  // Increment
  void SetIncrement( int Increment );

  // Boundaries and Size of the Array
  int MinIndex() const;
  int MaxIndex() const;
  int Size() const;           // MaxIndex - MinIndex + 1
  int NumberOfItems() const;  // number of actually stored items

  //! Index of item number pos in sparse array.
  inline int GetIndexOfItem(int pos) const
    { assert(pos>=m_Indices.MinIndex() && pos<=m_Indices.MaxIndex()); return m_Indices.GetAt(pos); }

  // Set boundaries - does not affect allocated memory
  bool SetMinIndex( int MinIndex );
  bool SetMaxIndex( int MaxIndex );

  bool SetBoundaries( int MinIndex, int MaxIndex );

  //! Remove last element.
  bool DeleteLast();

  // Test, if a value is stored in a specified cell
  //bool Exists( int Index ) const;

  // Access to items in the array; read is only allowed within boundaries
  // write can change boundaries
  float GetAt( int Index ) const;
  int SetAt( int Index, float Value );

  float operator[]( int Index ) const;  // read only
  float& operator[]( int Index );       // write

  //! Fetch an index-value pair
  // Position = 0 .. NumberOfItems() - 1
  bool GetItem( int Position, int  &Index, float &Value ) const;

  // Defaultvalue
  float DefaultValue() const;
  void SetDefaultValue( float DefaultValue );

  // Add Value to each value in the array, ditto for Sub, Mul, Div
  void Add( float Value );
  void Sub( float Value );
  void Mul( float Value );
  void Div( float Value );

  // Get the Sum of all NumOfItems items in the array, i.e. the sum
  // of the explicitly stored and those only stored per default value
  float Sum( int NumOfItems = -1 ) const;
  // Sum only of explicitly stored items
  float SumOfStoredItems() const;


  // Add Value to value at index Index in the array, ditto for Sub, Mul, Div
  void Add( int Index, float Value );
  void Sub( int Index, float Value );
  void Mul( int Index, float Value );
  void Div( int Index, float Value );

  // Increment or Decrement the value at Index by 1
  void Inc( int Index );
  void Dec( int Index );

  // Add two arrays
  void Add( const CSparseArray & Array );

  // Convert each value in the array to its logarithm
  void Log();

  // NewArray = Array * Value
  // expensive: constructs new Array and then copies it to return value
  CSparseArray operator * ( float Value ) const;

  // Add two arrays
  CSparseArray operator + ( const CSparseArray & Array ) const;

  //! "Insert" an empty element.
  void Insert(int idx);

  void Display() const;

  bool Load( std::ifstream &Stream );
  bool Store( std::ofstream &Stream ) const;

  bool Load( const char * pFileName );
  bool Store( const char * pFileName ) const;

protected:
  //void RemoveSmallValues();

  int m_MinIndex;         // smallest index stored
  int m_MaxIndex;         // largest index stored

  float m_DefaultValue;   // value of items not stored

  CSortedArray< int >    m_Indices;   // array, where indices are stored
  CNumericalArray< float > m_Values;    // array, where values are stored

  //friend class CDblMatrix;
};




// ----------------------------------------------------------------
//
// class CSparseMatrix
//
// ----------------------------------------------------------------

class CSparseMatrix
{
public:
  CSparseMatrix();
  CSparseMatrix( int NumOflines, int NumOfColumns );

  ~CSparseMatrix();

  // Remove items in the matrix
  void Clear();

  // Increment
  void SetIncrement( int Increment );

  // boundaries of the matrix
  int MinRow() const;
  int MaxRow() const;
  int MinCol() const;
  int MaxCol() const;

  // Size of the matrix
  int NumOfRows() const;
  int NumOfCols() const;

  // Set boundaries of the array, no elements are stored
  bool SetBoundaries( int MinRow, int MaxRow, 
		      int MinCol, int MaxCol );

  // Value for cells not stored in the matrix
  void SetDefaultValue( float Value );
  float DefaultValue() const;

  void SetDefaultValue( int Row, float Value );
  float DefaultValue( int Row ) const;

  // Test, if a value is stored in a specified cell
  //bool Exists( int Row, int Col ) const;

  // Set and get a value
  int SetAt( int Row, int Col, float Value );
  float GetAt( int Row, int j ) const;

  // Modify all cells in the matrix
  void Add( float Value );
  void Sub( float Value );
  void Mul( float Value );
  void Div( float Value );

  // Multiply i-th row with i-th component of vector
  void Mul( const CNumericalArray< float > &Values );

  // Modify a single cells in the matrix
  void Add( int Row, int Col, float Value );
  void Sub( int Row, int Col, float Value );
  void Mul( int Row, int Col, float Value );
  void Div( int Row, int Col, float Value );

  void Inc( int Row, int Col );
  void Dec( int Row, int Col );

  // Add two matrices
  void Add( const CSparseMatrix & Array );

  // Logarithm of all entries
  void Log();

  // Flip the organisation from x - y to y - x
  // default values for all rows is set to default value of Matrix
  void Flip( const CSparseMatrix & Matrix );

  // Sum of a row
  float SumOfRow( int Row ) const;

  // Get a row
  CSparseArray & GetRow( int Row );
  const CSparseArray & GetRow( int Row ) const;

  // Get number of items in a line
  int SizeOfRow( int Row ) const;

  //! Insert a new empty row into the matrix.
  void InsertRow(int idx);
  //! Insert a new empty column into the matrix.
  void InsertColumn(int idx);

  void Display() const;

  bool Load( ifstream &Stream );
  bool Store( ofstream &Stream ) const;

  //! A result type for collectRowIndices().
  typedef vector< int > CIndexContainer;
  //! Get the indices of the rows and columns of the matrix which acutally contain data.
  void CollectUsedIndices( CIndexContainer& rowResult, CIndexContainer& colResult ) const;

  //! Update min/max column/row number.
  void UpdateDimensions();

private:
  int m_MinCol;
  int m_MaxCol;

  float m_DefaultValue;

  CArray< CSparseArray > m_Rows;
};



#include <SparseArray_inl.hh>

#endif
