#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


// ------------------------------------------------------------------------------
//
// Inline functions for CSparseArray and CSparseMatrix
//
// ------------------------------------------------------------------------------


#include<assert.h>
#include<math.h>

#include <SparseArray.hh>

//static char rcsid_SparseArray_cc[] = "$Id:";


inline void
CSparseArray::SetDefaultValue( float DefaultValue )
{
  //assert( DefaultValue > 0.0 );
  m_DefaultValue = DefaultValue;
}


inline float
CSparseArray::DefaultValue() const
{
  return m_DefaultValue;
}


inline void
CSparseArray::SetIncrement( int Increment )
{
  m_Indices.SetIncrement( Increment );
  m_Values.SetIncrement( Increment );
}


/**
   @brief Return the size of the array.
   @return The array size / dimension.

   This may be greater than the actual number of allocated elements.
   \todo As {Min/Max}Index() is probably also inline, replacing the local variables
   by "function calls" is more efficient.
 */
inline int
CSparseArray::Size() const
{
  int MinPosition = m_Indices.MinIndex();
  int MaxPosition = m_Indices.MaxIndex();
  // if the array is empty, min is greater than max
  return (MinPosition > MaxPosition ?
	  0 :
	  m_Indices.GetAt(MaxPosition) - m_Indices.GetAt(MinPosition) + 1);
} // CSparseArray::Size()


inline int
CSparseArray::MinIndex() const
{
  return (m_Indices.Size() > 0 ? m_Indices.GetAt( m_Indices.MinIndex() ) : 0);
}


inline int
CSparseArray::MaxIndex() const
{
  return (m_Indices.Size() > 0 ? m_Indices.GetAt( m_Indices.MaxIndex() ) : -1);
}


inline int
CSparseArray::NumberOfItems() const
{
  return m_Indices.MaxIndex() - m_Indices.MinIndex() + 1;
}


inline void
CSparseArray::Add( float Value )
{
  m_Values.Add( Value );
  m_DefaultValue += Value;
}


inline void
CSparseArray::Sub( float Value )
{
  m_Values.Sub( Value );
  m_DefaultValue -= Value;
}


inline void
CSparseArray::Mul( float Value )
{
  m_Values.Mul( Value );
  m_DefaultValue *= Value;
  //assert( m_DefaultValue > 0.0 );
}


inline void
CSparseArray::Div( float Value )
{
  m_Values.Div( Value );
  m_DefaultValue /= Value;
}


inline void
CSparseArray::Add( int Index, float Value )
{
  //Trace( 5, "CSparseArray::Add " << Index << " " << Value );
  float OldValue = GetAt( Index );
  //Trace( 5, "OldValue " << OldValue );
  SetAt( Index, OldValue + Value );
}


inline void
CSparseArray::Sub( int Index, float Value )
{
  float OldValue = GetAt( Index );
  SetAt( Index, OldValue - Value );
}


inline void
CSparseArray::Mul( int Index, float Value )
{
  float OldValue = GetAt( Index );
  SetAt( Index, OldValue * Value );
}


inline void
CSparseArray::Div( int Index, float Value )
{
  float OldValue = GetAt( Index );
  SetAt( Index, OldValue / Value );
}


inline void
CSparseArray::Inc( int Index )
{
  float OldValue = GetAt( Index );
  SetAt( Index, OldValue + 1.0 );
}


inline void
CSparseArray::Dec( int Index )
{
  float OldValue = GetAt( Index );
  SetAt( Index, OldValue - 1.0 );
}


inline void
CSparseArray::Log()
{
  //assert( m_DefaultValue > 0.0 );
  //if ( m_DefaultValue < 10^-20 ) {
  // m_DefaultValue = 10^-20;
  //}

  m_DefaultValue = log( m_DefaultValue );

  m_Values.Log();
}





// -------------------------------------------------------
//
// CSparseMatrix - Inlines
//
// -------------------------------------------------------

inline int
CSparseMatrix::NumOfRows() const
{
  return m_Rows.Size();
}


inline int
CSparseMatrix::NumOfCols() const
{
  return m_MaxCol - m_MinCol + 1;
}


// boundaries of the matrix
inline int
CSparseMatrix::MinRow() const
{
  return m_Rows.MinIndex();
}


inline int
CSparseMatrix::MaxRow() const
{
  return m_Rows.MaxIndex();
}


inline int
CSparseMatrix::MinCol() const
{ 
  return m_MinCol;
}
  

inline int
CSparseMatrix::MaxCol() const
{ 
  return m_MaxCol;
}


inline float
CSparseMatrix::DefaultValue() const
{ 
  return m_DefaultValue;
}


inline float
CSparseMatrix::DefaultValue( int Row ) const
{
  if ( m_Rows.IsValidIndex( Row ) )
    return m_Rows[Row].DefaultValue();
  else
    return m_DefaultValue;
}


inline int
CSparseMatrix::SetAt( int Row, int Col, float Value )
{
  //Trace( 5, "CSparseMatrix::SetAt" << Row << " " << Col << " " << Value );

  if ( Col < m_MinCol )
    m_MinCol = Col;

  if ( Col > m_MaxCol )
    m_MaxCol = Col;
  
  //m_Rows.AssureBoundaries( Row, Row );
  return m_Rows[ Row ].SetAt( Col, Value );
  //return m_Rows[Row][Col] = Value;
}


inline float
CSparseMatrix::GetAt( int Row, int Col ) const
{
  //assert(Row < m_NumOfRows);
  //clog << Row << " " << Col << " " << m_NumOfRows << endl;
  return m_Rows[ Row ].GetAt( Col );
}




// modify a single cell in the matrix
inline void
CSparseMatrix::Add( int Row, int Col, float Value )
{
  //AssureSize(Row);
  m_Rows[ Row ].Add( Col, Value );
}


inline void
CSparseMatrix::Sub( int Row, int Col, float Value )
{
  //AssureSize(Row);
  m_Rows[ Row ].Sub( Col, Value );
}


inline void
CSparseMatrix::Mul( int Row, int Col, float Value )
{
  //AssureSize(Row);
  m_Rows[ Row ].Mul( Col, Value );
}


inline void
CSparseMatrix::Div( int Row, int Col, float Value )
{
  //AssureSize(Row);
  m_Rows[ Row ].Div( Col, Value );
}


inline void
CSparseMatrix::Inc( int Row, int Col )
{
  //AssureSize(Row);
  m_Rows[ Row ].Inc( Col );
}


inline void
CSparseMatrix::Dec(int Row, int Col)
{
  //AssureSize(Row);
  m_Rows[ Row ].Dec( Col );
}


inline float
CSparseMatrix::SumOfRow( int Row ) const
{
  //AssureSize(Row);
  return m_Rows[ Row ].Sum( NumOfCols() );
}


inline CSparseArray &
CSparseMatrix::GetRow( int Row )
{
  //AssureSize(Row);
  return m_Rows[ Row ];
}


inline const CSparseArray &
CSparseMatrix::GetRow( int Row ) const
{
  //AssureSize(Row);
  return m_Rows[ Row ];
}


inline int
CSparseMatrix::SizeOfRow( int Row ) const
{
  //AssureSize(Row);
  return m_Rows[ Row ].NumberOfItems();
}



