#ifndef Trace_hh
#define Trace_hh

/**
   @file Trace.hh
   @brief Define trace macros (or placebo).

   You should define or undefine NO_TRACE immedeately before loading this header to
   enable or disable tracing for a module.

   To ensure that you are the first to load this header, load it before any other STTK header.
   You may include it after any out-of-package header (e.g. STL headers), including
   config.h (generated by configure).

   This header should never be included in a header file, as this leads to an unpredictable
   including sequence. If trace macros are used in a header file (hence, inline), e.g.
   in a template class definition, be sure the including module definition file (.cc)
   has properly included Trace.hh. As any module may include especially template definitions
   directly or indirectly, <em>any</em> module should include Trace.hh in the way described
   above. Tracing of a template class then depends on the module that instanciates the class.

   To check if tracing has been enabled or not, this header defines / undefines
   TRACE_ENABLED and TRACE_DISABLED respectively (you can use either of them, as you like).
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif



#include <iostream>
using std::cout;
using std::endl;
using std::flush;

// static char rcsid_trace_hh[] = "$Id: Trace.hh 1729 2006-06-26 20:04:33Z vogel $";

//! Globale variable for tracelevel, should be set in main program
extern int TraceLevel;

#define Fatal 0
#define Error 1

// no debug, no trace
#ifdef NDEBUG

#ifndef NO_TRACE
#define NO_TRACE 1
#endif

#endif

#ifdef NO_TRACE

// Trace disabled!
#ifndef TRACE_DISABLED
//! If the trace macros are only placebos, this is defined and TRACE_ENABLED is undefined.
#define TRACE_DISABLED 1
#endif
#undef TRACE_ENABLED

#define Trace( Level, Output )
#define TraceIf( Condition, Level, Output )
#define TraceIf0( Condition, Level, Output )
#define Trace0( Level, Output )
#define TRACE( Level, Output )
#define TRACE0( Level, Output )

#else

// Trace enabled!
#ifndef TRACE_ENABLED
//! If the trace macros actually output tracing information, this is defined and TRACE_DISABLED is undefined.
#define TRACE_ENABLED 1
#endif
#undef TRACE_DISABLED

//! Trace macro with trailing newline.
#define Trace(Level, Output) \
  if (Level <= TraceLevel) { \
    cout << Output << endl << flush; \
  }

//! Conditioned trace macro with newline.
#define TraceIf( Condition, Level, Output) \
  if ((Condition) && (Level <= TraceLevel) ) { \
    cout << Output << endl << flush; \
  }

//! Conditioned trace w/o newline.
#define TraceIf0( Condition, Level, Output) \
  if ((Condition) && (Level <= TraceLevel) ) { \
    cout << Output << flush; \
  }

//! Trace w/o trailing newline
#define Trace0(Level, Output) \
  if (Level <= TraceLevel) { \
    cout << Output << flush; \
  }


//! Trace with filename and line, trailing newline
#define TRACE(Level, Output) \
  if (Level <= TraceLevel) { \
    cout << __FILE__ << ":" << __LINE__ << ": " << Output << endl << flush; \
  }

//! Trace with filename and line, no trailing newline
#define TRACE0(Level, Output) \
  if (Level <= TraceLevel) { \
    cout << __FILE__ << ":" << __LINE__ << ": " << Output << flush; \
  }

#endif // enable tracing?

#endif // Trace_hh


