#include <config.h>

#include <iostream>
#include <fstream>

#include <Trace.hh>
#include <AlignProbRel.hh>

using namespace std;

CAlignProbRel::CAlignProbRel()
{
  Trace( 1, "CAlignProbRel Constructor" );

  m_NegJumpLimit = -30;
  m_PosJumpLimit = 30;

  m_NegJumpMargin = -30;
  m_PosJumpMargin = 30;

  SetJumpLimits( m_NegJumpLimit, m_PosJumpLimit );
  SetJumpMargins( m_NegJumpMargin, m_PosJumpMargin );

  m_SourceSLen = -1;
  m_TargetSLen = -1;

  m_Epsilon = 1.0e-10;

  Init();
}


CAlignProbRel::~CAlignProbRel()
{
}


void
CAlignProbRel:: SetJumpLimits( int NegJumpLimit, int PosJumpLimit )
{
  assert( NegJumpLimit <= 0 );
  assert( PosJumpLimit >= 1 );

  // keep jump margins within jump limits
  if ( NegJumpLimit > m_NegJumpMargin )
    SetJumpMargins( NegJumpLimit, m_PosJumpMargin );

  if ( PosJumpLimit < m_PosJumpMargin )
    SetJumpMargins( m_NegJumpMargin, PosJumpLimit );


  m_NegJumpLimit = NegJumpLimit;
  m_PosJumpLimit = PosJumpLimit;

}


void
CAlignProbRel::SetJumpMargins( int NegJumpMargin, int PosJumpMargin )
{
  assert( NegJumpMargin <= 0 );
  assert( PosJumpMargin >= 0 );

  // jump limits must not be smaller than jump margins
  if ( NegJumpMargin < m_NegJumpLimit )
    SetJumpLimits( NegJumpMargin, m_PosJumpLimit );

  if ( PosJumpMargin > m_PosJumpLimit )
    SetJumpLimits( m_NegJumpLimit, PosJumpMargin );


  m_NegJumpMargin = NegJumpMargin;
  m_PosJumpMargin = PosJumpMargin;

  // - 1 and + 1 to store probabilities and counts for jumps 
  // outside of margin
  m_Pr.SetBoundaries( m_NegJumpMargin - 1, m_PosJumpMargin + 1);
  m_PrSOS.SetBoundaries( 1, m_PosJumpMargin + 1);
  m_PrEOS.SetBoundaries( 1, m_PosJumpMargin + 1);

  m_Counts.SetBoundaries( m_NegJumpMargin - 1, m_PosJumpMargin + 1 );
  m_CountsSOS.SetBoundaries( 1, m_PosJumpLimit + 1 );
  m_CountsEOS.SetBoundaries( 1, m_PosJumpLimit + 1 );
}



void
CAlignProbRel::Init()
{
  Trace( 0, "CAlignProbRel::Init" );
  Trace( 3, "NegJumpMargin " << m_NegJumpMargin << endl
	 << "NegJumpLimit  " << m_NegJumpLimit << endl
	 << "PosJumpMargin " << m_PosJumpMargin << endl
	 << "PosJumpLimit  " << m_PosJumpLimit << endl );

  int i;

  // uniform probability distribution
  
  // 1. probabilities for jumps in the middle

  // total number of jump distances: m_PosJumpLimit + m_NegJumpLimit + 1
  double InitValue = 1.0 / double( m_PosJumpLimit - m_NegJumpLimit + 1 );
  Trace( 3, "InitValue " << InitValue );

  for ( i = m_NegJumpMargin - 1; i <= m_PosJumpMargin + 1; i++ ) {
    m_Pr[ i ] = InitValue;
    m_Counts[ i ] = 0;
  }


  // 2. probabilities for jumps from SOS to first word and 
  // for jumps from last word to EOS

  // total number of jump distances: m_PosJumpLimit
  InitValue = 1.0 / double( m_PosJumpLimit )                       ;
  
  for ( i = 1; i <= m_PosJumpMargin + 1; i++ ) {
    m_PrSOS[ i ] = InitValue;
    m_PrEOS[ i ] = InitValue;
    m_CountsSOS[ i ] = 0;
    m_CountsEOS[ i ] = 0;
  }

  //for ( int i = - 20; i <= 20; i++ ) {
  //  Trace( 1, i << " " << Pr( i, Start ) << " " << Pr( i, Middle )
  //	   << " " << Pr( i, End ) );
  //}
}
  


double
CAlignProbRel::Pr( int JumpWidth, TPosition Position )
{

  if ( JumpWidth < m_NegJumpLimit )
    return m_Epsilon;

  if ( JumpWidth < m_NegJumpMargin )
    JumpWidth = m_NegJumpMargin - 1;
  
  if ( JumpWidth > m_PosJumpLimit )
    return m_Epsilon;

  if ( JumpWidth > m_PosJumpMargin )
    JumpWidth = m_PosJumpMargin + 1;
  
  double Prob = m_Epsilon;

  switch ( Position ) {
  case Start:
    Prob = m_PrSOS[ JumpWidth ];
    break;
  case Middle:
    Prob = m_Pr[ JumpWidth ];
    break;
  case End:
    Prob = m_PrEOS[ JumpWidth ];
    break;
  default:
    Trace( 0, "Wrong Case Label in CAlignProbRel::Pr" );
    Prob = m_Pr[ JumpWidth ];
  }

  if ( Prob <= 0.0 )
    Prob = m_Epsilon;

  return Prob;

}


double
CAlignProbRel::LogPr( int JumpWidth, TPosition Position )
{
  return log( Pr( JumpWidth, Position )  );
}




void
CAlignProbRel::PreparePr( int SourceSentenceLength, int TargetSentenceLength )
{
  Trace( 3, "CAlignProbRel::PreparePr " 
	 << SourceSentenceLength << " " << TargetSentenceLength );

  m_SourceSLen = SourceSentenceLength;

  // if already calculated for this sentence length
  // the nothing to do
  if ( TargetSentenceLength == m_TargetSLen )
    return;

  m_TargetSLen = TargetSentenceLength;

  int Current, Prev;

  double LogEpsilon;
  if ( m_Epsilon > 0.0 )
    LogEpsilon = log( m_Epsilon );
  else
    LogEpsilon = -1.0e10;

  Trace( 3, "Epsilon = " << m_Epsilon << "  LogEpsilon = " << LogEpsilon );

  // 1. jumps from (SOS, SOS) to (1, i) with i = 1..I

  // set all entries to epsilon resp. log( epsilon )
  for ( Current = 1; Current <= m_TargetSLen; Current++ ) {
    m_RenormPrSOS[ Current ] = m_Epsilon;
    m_PenaltySOS[ Current ] = LogEpsilon;
  }

  // calculate sum
  double PrSum = 0.0;
  double Prob;

  int MaxJump = m_TargetSLen;
  if ( m_TargetSLen > m_PosJumpLimit )
    MaxJump = m_PosJumpLimit;

  int JumpWidth;
  for ( JumpWidth = 1; JumpWidth <= MaxJump; JumpWidth++ ) {
    PrSum += Pr( JumpWidth, Start );
  }
   
  assert( PrSum > 0.0 );
  
  // now redistribute
  for ( JumpWidth = 1; JumpWidth <= MaxJump; JumpWidth++ ) {
    Prob = Pr( JumpWidth, Start ) / PrSum;
    assert( Prob > 0.0 );
    m_RenormPrSOS[ JumpWidth ] = Prob;
    m_PenaltySOS[ JumpWidth ] = log( Prob );
  }


  // 2. jumps from (j-1, prev) -> (j, current)

  // init all with small epsilon
  for ( Prev = 1; Prev <= m_TargetSLen; Prev++ ) {
    for ( Current = 1; Current <= m_TargetSLen; Current++ ) {
      m_RenormPr( Current, Prev ) = m_Epsilon;
      m_Penalty( Current, Prev ) = LogEpsilon;
    }
  }

  int MinCurrent, MaxCurrent;
  for ( Prev = 1; Prev <= m_TargetSLen; Prev++ ) {
     // make shure we keep within valid region for position (j, current)
    MaxCurrent = Prev + m_PosJumpLimit;
    if ( MaxCurrent > m_TargetSLen )
      MaxCurrent = m_TargetSLen;

    MinCurrent = Prev + m_NegJumpLimit;
    if ( MinCurrent < 1 )
      MinCurrent = 1;

    PrSum = 0.0;
    for ( Current = MinCurrent; Current <= MaxCurrent; Current++ ) {
      PrSum += Pr( Current - Prev, Middle );   // Current - Prev = JumpWidth
    }
    
    assert( PrSum > 0.0 );

    // now redistribute
    for ( Current = MinCurrent; Current <= MaxCurrent; Current++ ) {
      Prob = Pr( Current - Prev, Middle ) / PrSum;
      assert( Prob > 0.0 );
      m_RenormPr( Current, Prev ) = Prob;
      m_Penalty( Current, Prev ) = log( Prob );
    }

  } // loop current


  // 3. jumps from (J, i) to (SOS, SOS) with i = 1..I
  // set all entries to 1.0 resp. log( 1.0 ) as there is no choice
  for ( Prev = 1; Prev <= m_TargetSLen; Prev++ ) {
    m_RenormPrEOS[ Prev ] = 1.0;
    m_PenaltyEOS[ Prev ] = log( 1.0 );
  }

  //DisplayPenalties();
  /*
  for ( JumpWidth = 1; JumpWidth <= MaxJump; JumpWidth++ ) {
    cout << m_RenormPrSOS[ JumpWidth ] << "    ";
  }
  cout << endl;

  for ( Prev = 1; Prev <= m_TargetSLen; Prev++ ) {
    for ( Current = 1; Current <= m_TargetSLen; Current++ ) {
      cout << m_RenormPr( Current, Prev ) << "    ";
    }
    cout << endl;
  }
  */

  //  return TRUE;
}

void
CAlignProbRel::IncCount( int Current, int Previous, int PosInSentence )
{
  Trace( 5, "CAlignProbRel::IncCount " << Current << " " 
	 << Previous << " " << PosInSentence );

  Trace( 5, "m_SourceSLen " << m_SourceSLen
	 << " m_TargetSLen " << m_TargetSLen );

  assert( PosInSentence >= 1 && PosInSentence <= m_SourceSLen + 1 );

  int JumpWidth = Current - Previous;

  if ( JumpWidth < m_NegJumpMargin )
    JumpWidth = m_NegJumpMargin - 1;
  
  if ( JumpWidth > m_PosJumpMargin )
    JumpWidth = m_PosJumpMargin + 1;
  
  // modification for alignment of graphs
  // edge may run from node > 1 to node 0
  // therefore, quick hack
  if ( Previous == 0 )
    PosInSentence = 1;

  // is Penalty for Start, Middle or End of source sentence requested
  if ( PosInSentence == 1 ) {
    assert( Current >= 1 && Current <= m_TargetSLen );
    assert( Previous == 0 );
    m_CountsSOS.Inc( JumpWidth );
  }
  else if ( PosInSentence == m_SourceSLen + 1 ) {
    assert( Current == m_TargetSLen + 1 );
    assert( Previous >= 1 && Previous <= m_TargetSLen );
    m_CountsEOS.Inc( JumpWidth );
  }
  else {
    assert( Current >= 1 && Current <= m_TargetSLen );
    assert( Previous >= 1 && Previous <= m_TargetSLen );
    m_Counts.Inc( JumpWidth );
  }

}


void
CAlignProbRel::ClearCounts()
{
  m_Counts.SetAll( 0 );
  m_CountsSOS.SetAll( 0 );
  m_CountsEOS.SetAll( 0 );
		
}



////////////////////////////////////////////////////////////////////////
// Renormalization
////////////////////////////////////////////////////////////////////////


void
CAlignProbRel::Renormalize()
{
  cout << "Renormalize\n";

  // copy counts to probability arrays
  m_Pr = m_Counts;
  m_PrSOS = m_CountsSOS;
  m_PrEOS = m_CountsEOS;

  // renormalize
  double Sum;
  Sum = m_Pr.Sum();
  if ( Sum > 0 )
    m_Pr.Div( Sum );

  Sum = m_PrSOS.Sum();
  if ( Sum > 0 )
    m_PrSOS.Div( Sum );

  Sum = m_PrEOS.Sum();
  if ( Sum > 0 )
    m_PrEOS.Div( Sum );

  // if we have large jumps, divide probability by number of jump distances
  int LargeJumpDistances =  m_PosJumpLimit - m_PosJumpMargin;
  if ( LargeJumpDistances > 0 ) {
    m_Pr[ m_PosJumpMargin ] /= LargeJumpDistances;
    m_PrSOS[ m_PosJumpMargin ] /= LargeJumpDistances;
    m_PrEOS[ m_PosJumpMargin ] /= LargeJumpDistances;
  }

  LargeJumpDistances =  m_NegJumpLimit - m_NegJumpMargin;
  if ( LargeJumpDistances > 0 ) {
    m_Pr[ m_NegJumpMargin ] /= LargeJumpDistances;
    m_PrSOS[ m_NegJumpMargin ] /= LargeJumpDistances;
    m_PrEOS[ m_NegJumpMargin ] /= LargeJumpDistances;
  }

  // make shure, we have no zero probabilities
  int i;

  for ( i = m_Pr.MinIndex(); i <= m_Pr.MaxIndex(); i++ ) {
    if ( m_Pr[ i ] <= 0.0 ) {
      Trace( 5, "m_Pr[ i ] = m_Epsilon" << i );
      m_Pr[ i ] = m_Epsilon;
    }
  }

  for ( i = m_PrSOS.MinIndex(); i <= m_PrSOS.MaxIndex(); i++ ) {
    if ( m_PrSOS[ i ] <= 0.0 ) {
      Trace( 5, "m_PrSOS[ i ] = m_Epsilon" << i);
      m_PrSOS[ i ] = m_Epsilon;
    }
  }

  for ( i = m_PrEOS.MinIndex(); i <= m_PrEOS.MaxIndex(); i++ ) {
    if ( m_PrEOS[ i ] <= 0.0 ) {
      Trace( 5, "m_PrEOS[ i ] = m_Epsilon" << i);
      m_PrEOS[ i ] = m_Epsilon;
    }
  }

}


bool
CAlignProbRel::TestProbabilities()
{
  return FALSE;
}


////////////////////////////////////////////////////////////////////////
// Display - Methods
////////////////////////////////////////////////////////////////////////

void
CAlignProbRel::DisplayCounts()
{
  int i;

  // 1. probabilities for jumps from SOS to first word

  // normal probabilities
  cout << "normal counts for SOS position" << endl;
  for ( i = 1; i <= m_PosJumpMargin; i++ ) {
    cout << i << " " <<  m_CountsSOS[ i ] << endl;
  }

  // number of jump distances below m_NegJumpMargin: none
  // number of jump distances above m_PosJumpMargin
  cout << "count for jump distances above m_PosJumpMargin" << endl;
  cout << (m_PosJumpMargin + 1) << " " 
       <<  m_CountsSOS[ m_PosJumpMargin + 1 ] << endl;
  
  // 2. probabilities for jumps in the middle

  // normal probabilities
  cout << "normal counts for middle positions" << endl;
  for ( i = m_NegJumpMargin; i <= m_PosJumpMargin; i++ ) {
    cout << i << " " << m_Counts[ i ] << endl;
  }

  // probability for jump distances below m_NegJumpMargin
  cout << "count for jump distances below m_NegJumpMargin" << endl;
  cout << (m_NegJumpMargin - 1) << " " << m_Counts[ m_NegJumpMargin - 1 ] << endl;
  

  // probability for jump distances above m_PosJumpMargin
  cout << "count for jump distances above m_PosJumpMargin" << endl;
  cout << (m_PosJumpMargin + 1) << " " <<  m_Counts[ m_PosJumpMargin + 1 ] << endl;
  

  // 3. probabilities for jumps from last word to EOS

  // normal probabilities
  cout << "normal counts for EOS position" << endl;
  for ( i = 1; i <= m_PosJumpMargin; i++ ) {
    cout << i << " " <<  m_CountsEOS[ i ] << endl;
  }

  // number of jump distances below m_NegJumpMargin: none
  // number of jump distances above m_PosJumpMargin
  cout << "count for jump distances above m_PosJumpMargin" << endl;
  cout << (m_PosJumpMargin + 1) << " " 
       <<  m_CountsEOS[ m_PosJumpMargin + 1 ] << endl;
  


}



void
CAlignProbRel::DisplayProb()
{
  int i;

  // 1. probabilities for jumps from SOS to first word

  // normal probabilities
  cout << "normal probabilities for SOS position" << endl;
  for ( i = 1; i <= m_PosJumpMargin; i++ ) {
    cout << i << " " <<  m_PrSOS[ i ] << endl;
  }

  // number of jump distances below m_NegJumpMargin: none
  // number of jump distances above m_PosJumpMargin
  cout << "probability for jump distances above m_PosJumpMargin" << endl;
  cout << (m_PosJumpMargin + 1) << " " 
       <<  m_PrSOS[ m_PosJumpMargin + 1 ] << endl;
  
  // 2. probabilities for jumps in the middle

  // normal probabilities
  cout << "normal probabilities for middle positions" << endl;
  for ( i = m_NegJumpMargin; i <= m_PosJumpMargin; i++ ) {
    cout << i << " " << m_Pr[ i ] << endl;
  }

  // probability for jump distances below m_NegJumpMargin
  cout << "probability for jump distances below m_NegJumpMargin" << endl;
  cout << (m_NegJumpMargin - 1) << " " << m_Pr[ m_NegJumpMargin - 1 ] << endl;
  

  // probability for jump distances above m_PosJumpMargin
  cout << "probability for jump distances above m_PosJumpMargin" << endl;
  cout << (m_PosJumpMargin + 1) << " " <<  m_Pr[ m_PosJumpMargin + 1 ] << endl;
  

  // 3. probabilities for jumps from last word to EOS

  // normal probabilities
  cout << "normal probabilities for EOS position" << endl;
  for ( i = 1; i <= m_PosJumpMargin; i++ ) {
    cout << i << " " <<  m_PrEOS[ i ] << endl;
  }

  // number of jump distances below m_NegJumpMargin: none
  // number of jump distances above m_PosJumpMargin
  cout << "probability for jump distances above m_PosJumpMargin" << endl;
  cout << (m_PosJumpMargin + 1) << " " 
       <<  m_PrEOS[ m_PosJumpMargin + 1 ] << endl;
  
}



void
CAlignProbRel::DisplayLogPr()
{
  Trace( 1, "CAlignProbRel::DisplayPenalties " 
	 << m_SourceSLen << " " << m_TargetSLen );

  cout << endl;
  cout << "jumps from (SOS, SOS) to (1, i) with i = 1..I" << endl;

  int Current, Prev;

  cout << "prev = " << 0 << ": ";
  for ( Current = 1; Current <= m_TargetSLen; Current++ ) {
    cout << LogPr( Current, 0, 1 ) << " ";
  }
  cout << endl << flush;

  cout << "jumps from (j-1, prev) -> (j, current)" << endl;
  for ( Prev = 1; Prev <= m_TargetSLen; Prev++ ) {
    cout << "prev = " << Prev << ": ";
    for ( Current = 1; Current <= m_TargetSLen; Current++ ) {
      cout << LogPr( Current, Prev, 2 ) << " ";
    }
    cout << endl << flush;
  }


  cout << "jumps from (J, i) to (SOS, SOS) with i = 1..I" << endl;

  cout << "prev = " << m_SourceSLen << ": ";
  for ( Prev = 1; Prev <= m_TargetSLen; Prev++ ) {
    cout << LogPr( m_TargetSLen + 1, Prev, m_SourceSLen + 1 ) << " ";
  }
  cout << endl;
  cout << endl << flush;

}



bool
CAlignProbRel::Store(const char *pFileName)
{
  ofstream TFile (pFileName);
  if (!TFile) {
    cout << "File " << pFileName << " could not be opened for writing\n";
    return FALSE;
  }

  TFile.setf(ios::scientific);
  TFile.precision(6);

  TFile << "# Alignment probabilities: jumps, i.e. parameterized, relative\n";
  //TFile << "# format:\n";
  //TFile << "# largest jump for which individual probability is stored\n";
  //TFile << "# jump probability for large jumps\n";
  //TFile << "# jump probability for small jumps\n";
  //TFile << "# with one jump width - jump probability pair per line \n";
  //TFile << "#" << endl;
  //TFile << m_SmallJumpLimit << endl;
  //TFile << m_JBigJump << endl;

  //for (int dist = -m_SmallJumpLimit; dist <= m_SmallJumpLimit; dist++) {
  //  if (dist < 0)
  //    TFile << dist << " "<< m_JNeg[-dist][0] << endl;
  //  else
  //    TFile << dist << " "<< m_JPos[dist][0] << endl;
  //}

  int i;

  TFile << m_NegJumpLimit << " ";
  TFile << m_NegJumpMargin << " ";
  TFile << m_PosJumpMargin << " ";
  TFile << m_PosJumpLimit << " ";
  TFile << endl;


  // 1. probabilities for jumps from SOS to first word

  // normal probabilities
  TFile << "# probabilities for SOS position" << endl;
  for ( i = 1; i <= m_PosJumpMargin + 1; i++ ) {
    TFile <<  m_PrSOS[ i ] << endl;
  }

  // number of jump distances below m_NegJumpMargin: none
  // number of jump distances above m_PosJumpMargin
  //cout << "probability for jump distances above m_PosJumpMargin" << endl;
  //cout << (m_PosJumpMargin + 1) << " " 
  //TFile <<  m_PrSOS[ m_PosJumpMargin + 1 ] << endl;
  //TFile << endl;


  // 2. probabilities for jumps in the middle

  // normal probabilities
  TFile << "# probabilities for middle positions" << endl;
  for ( i = m_NegJumpMargin - 1; i <= m_PosJumpMargin + 1; i++ ) {
    TFile << m_Pr[ i ] << endl;
  }

  // probability for jump distances below m_NegJumpMargin
  //cout << "probability for jump distances below m_NegJumpMargin" << endl;
  //TFile << m_Pr[ m_NegJumpMargin - 1 ] << endl;
  

  // probability for jump distances above m_PosJumpMargin
  //cout << "probability for jump distances above m_PosJumpMargin" << endl;
  //TFile << m_Pr[ m_PosJumpMargin + 1 ] << endl;
  

  // 3. probabilities for jumps from last word to EOS

  // normal probabilities
  TFile << "# probabilities for EOS position" << endl;
  for ( i = 1; i <= m_PosJumpMargin + 1; i++ ) {
    TFile << m_PrEOS[ i ] << endl;
  }

  // number of jump distances below m_NegJumpMargin: none
  // number of jump distances above m_PosJumpMargin
  //cout << "probability for jump distances above m_PosJumpMargin" << endl;
  //TFile <<  m_PrEOS[ m_PosJumpMargin + 1 ] << endl;
 
  return TRUE;
}



bool
CAlignProbRel::Load(const char *pFileName)
{

  Trace( 0, "CAlignProbRel::Load" );

  ifstream TFile (pFileName);

  if (!TFile) {
    cout << "File " << pFileName << " could not be opened for reading\n";
    return FALSE;
  }

  // read comment line
  char String[100];
  TFile.getline(String, 100);
  //TFile.getline(String, 100);
  //TFile.getline(String, 100);
  //TFile.getline(String, 100);
  //TFile.getline(String, 100);
  //TFile.getline(String, 100);
  //TFile.getline(String, 100);

  TFile >> m_NegJumpLimit 
	>> m_NegJumpMargin 
	>> m_PosJumpMargin 
	>> m_PosJumpLimit;

  Trace( 0, m_NegJumpLimit << " "
	 << m_NegJumpMargin << " "
	 << m_PosJumpMargin << " "
	 << m_PosJumpLimit );

  int i;

  // 1. probabilities for jumps from SOS to first word

  // normal probabilities
  //cout << "normal probabilities for SOS position" << endl;
  TFile.getline(String, 100);
  //Trace( 0, String );
  TFile.getline(String, 100);
  //Trace( 0, String );
  for ( i = 1; i <= m_PosJumpMargin + 1; i++ ) {
    TFile >> m_PrSOS[ i ];
    //Trace( 0, i << " " << m_PrSOS[ i ] );
  }

  //TFile >> m_PrSOS[ m_PosJumpMargin + 1 ];


  // 2. probabilities for jumps in the middle

  // normal probabilities
  //cout << "normal probabilities for middle positions" << endl;
  TFile.getline(String, 100);
  //Trace( 0, String );
  TFile.getline(String, 100);
  //Trace( 0, String );
  for ( i = m_NegJumpMargin - 1; i <= m_PosJumpMargin + 1; i++ ) {
    TFile >> m_Pr[ i ];
    //Trace( 0, i << " " << m_Pr[ i ] );
  }

  // probability for jump distances below m_NegJumpMargin
  //cout << "probability for jump distances below m_NegJumpMargin" << endl;
  //TFile >> m_Pr[ m_NegJumpMargin - 1 ];
  

  // probability for jump distances above m_PosJumpMargin
  //cout << "probability for jump distances above m_PosJumpMargin" << endl;
  //TFile >> m_Pr[ m_PosJumpMargin + 1 ];
  

  // 3. probabilities for jumps from last word to EOS

  // normal probabilities
  //cout << "normal probabilities for EOS position" << endl;
  TFile.getline(String, 100);
  //Trace( 0, String );
  TFile.getline(String, 100);
  //Trace( 0, String );
  for ( i = 1; i <= m_PosJumpMargin + 1; i++ ) {
    TFile >> m_PrEOS[ i ];
    //Trace( 0, i << " " << m_PrEOS[ i ] );
  }

  // number of jump distances below m_NegJumpMargin: none
  // number of jump distances above m_PosJumpMargin
  //cout << "probability for jump distances above m_PosJumpMargin" << endl;
  //TFile >> m_PrEOS[ m_PosJumpMargin + 1 ];
  


  TFile.close();

  //DisplayProb();

  return TRUE;
}


