#ifndef AlignProbRel_hh
#define AlignProbRel_hh

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


// CAlignProbRel

// Class for alignment propabilities, parameterized, relative


#include <Types.hh>
#include <NumericalArray.hh>
#include <AlignProb.hh>


class CAlignProbRel : public CAlignProb
{
public:
  CAlignProbRel();
  ~CAlignProbRel();

  // set limits (which jumps have probabilities > 0) 
  // and margins (which jumps are individually modelled)
  // NegJumpLimit <= NegJumpMargin <= 0 
  void SetJumpLimits( int NegJumpLimit, int PosJumpLimit );
  void SetJumpMargins( int NegJumpMargin, int PosJumpMargin );

  // init with uniform probabilities, according to limits and margins
  virtual void Init();

  // position type within the sentence
  typedef enum{ Start, Middle, End } TPosition;

  // jump probability
  double Pr   ( int JumpWidth, TPosition Position );
  double LogPr( int JumpWidth, TPosition Position );

  // jump probability renormalized for given sentence length I
  // and positions (Previous, PosInSentence - 1) and
  // (Current, PosInSentence)
  double Pr   ( int Current, int Previous, int PosInSentence );
  double LogPr( int Current, int Previous, int PosInSentence );

  // ditto for alignment of a sequence of positions
  // from (IndexSource - NumberOfWords + 1, IndexTarget)
  // to (IndexSource, IndexTarget)
  double Pr   ( int Current, int Previous, 
		int NumberOfWords, int PosInSentence );
  double LogPr( int Current, int Previous, 
		int NumberOfWords, int PosInSentence );

  // calculate the renormalized probabilities for sentence length
  // TargetSentenceLength; SourceSentenceLength is only required
  // to know from PosInSentence in Pr and LogPr if probability 
  // for jump from last word to EOS is requested
  virtual void PreparePr( int SourceSentenceLength, int TargetSentenceLength );
  virtual void PrepareLogPr( int SourceSentenceLength, int TargetSentenceLength );

  // update count for jump from (Previous, PosInSentence - 1) to 
  // (Current, PosInSentence)
  void IncCount( int Current, int Previous, int PosInSentence );

  void ClearCounts();

  bool Log();

  // renormalize and test probability distribution
  void Renormalize();
  bool TestProbabilities();
  


  // ----- Display Methods -----
  void DisplayProb();
  void DisplayCounts();

  void DisplayLogPr();

  // ----- Store & Load alignment probabilities -----
  bool Store( const char *pFileName );
  bool Load( const char *pFileName ); 


protected:
  // boundaries for valid jumps
  int m_NegJumpLimit;
  int m_PosJumpLimit;

  // boundaries for jumps which are modelled explicitely
  int m_NegJumpMargin;
  int m_PosJumpMargin;

  // probability for invalid jumps
  double m_Epsilon;

  // probabilities for jumps
  CNumericalArray<double> m_Pr;     // jump from word to word
  CNumericalArray<double> m_PrSOS;  // jump from SOS to first word
  CNumericalArray<double> m_PrEOS;  // jump from last word to EOS

  // counts for jumps, used in training (double to ease array assignment)
  CNumericalArray<double> m_Counts;
  CNumericalArray<double> m_CountsSOS;
  CNumericalArray<double> m_CountsEOS;

  // renormalized probabilities for sentence length m_CurrentSLen
  CMatrix<double> m_RenormPr;     // for jump from (j-1, prev) -> (j, current)
  CArray<double> m_RenormPrSOS;   // for jump from (SOS, SOS) -> (1, current)
  CArray<double> m_RenormPrEOS;   // for jump from (J, prev) -> to (EOS, EOS)

  // log( RenormPr )
  CMatrix<double> m_Penalty;
  CArray<double> m_PenaltySOS;
  CArray<double> m_PenaltyEOS;

};



// --------------------------------------------------------------------
// Inlines
// --------------------------------------------------------------------


inline void
CAlignProbRel::PrepareLogPr( int SourceSentenceLength, int TargetSentenceLength )
{
  PreparePr (SourceSentenceLength, TargetSentenceLength);
}



// jump probability renormalized for given sentence length I
// and positions (Previous, PosInSentence - 1) and
// (Current, PosInSentence)
inline double
CAlignProbRel::Pr( int Current, int Previous, int PosInSentence )
{
  assert( PosInSentence >= 1 && PosInSentence <= m_SourceSLen + 1 );

  // modification for alignment of graphs
  // edge may run from node > 1 to node 0
  // therefore, quick hack
  if ( Previous == 0 )
    PosInSentence = 1;

  // is Penalty for Start, Middle or End of source sentence requested
  if ( PosInSentence == 1 ) {
    assert( Current >= 1 && Current <= m_TargetSLen );
    assert( Previous == 0 );
    return m_RenormPrSOS[ Current ];
  }
  else if ( PosInSentence == m_SourceSLen + 1 ) {
    assert( Current == m_TargetSLen + 1 );
    assert( Previous >= 1 && Previous <= m_TargetSLen );
    return m_RenormPrEOS[ Previous ];
  }
  else {
    assert( Current >= 1 && Current <= m_TargetSLen );
    assert( Previous >= 1 && Previous <= m_TargetSLen );
    return m_RenormPr( Current, Previous );
  }
}


inline double
CAlignProbRel::LogPr( int Current, int Previous, int PosInSentence )
{
  // Trace( 5, "CAlignProbRel::LogPr " << Current << " " 
  //	 << Previous << " " << PosInSentence );

  // Trace( 5, "m_SourceSLen " << m_SourceSLen
  //	 << " m_TargetSLen " << m_TargetSLen );

  assert( PosInSentence >= 1 && PosInSentence <= m_SourceSLen + 1 );

  // modification for alignment of graphs
  // edge may run from node > 1 to node 0
  // therefore, quick hack
  if ( Previous == 0 )
    PosInSentence = 1;

  // is Penalty for Start, Middle or End of source sentence requested
  if ( PosInSentence == 1 ) {
    assert( Current >= 1 && Current <= m_TargetSLen );
    assert( Previous == 0 );
    return m_PenaltySOS[ Current ];
  }
  else if ( PosInSentence == m_SourceSLen + 1 ) {
    assert( Current == m_TargetSLen + 1 );
    assert( Previous >= 1 && Previous <= m_TargetSLen );
    return m_PenaltyEOS[ Previous ];
  }
  else {
    assert( Current >= 1 && Current <= m_TargetSLen );
    assert( Previous >= 1 && Previous <= m_TargetSLen );
    return m_Penalty( Current, Previous );
  }
}

inline double
CAlignProbRel::Pr( int Current, int Previous, 
		   int NumberOfWords, int PosInSentence )
{
  assert( PosInSentence >= 1 && PosInSentence <= m_SourceSLen + 1 );
  assert( NumberOfWords >= 1 );

  double PrAkkumulated = 0.0;
  int Index;

  for ( Index = Current - NumberOfWords + 1; Index <= Current; Index++ ) 
    {
      PrAkkumulated *= Pr( Index, Previous, PosInSentence );
    }

  return PrAkkumulated;
}


inline double
CAlignProbRel::LogPr( int Current, int Previous, 
		      int NumberOfWords, int PosInSentence )
{
   // Trace( 5, "CAlignProbRel::LogPr " << Current << " " 
  //	  << Previous << " " << NumberOfWords << " " << PosInSentence );

  assert( PosInSentence >= 1 && PosInSentence <= m_SourceSLen + 1 );
  assert( NumberOfWords >= 1 );

  double SumLogPr = 0.0;
  int Index;

  for ( Index = Current - NumberOfWords + 1; 
	Index <= Current; 
	Index++ ) {
    SumLogPr += LogPr( Index, Previous, PosInSentence );
  }

  return SumLogPr;
}
#endif

