// CLexicon

// CVS file ID: $Id: Lexicon.hh 1685 2006-06-14 15:36:55Z vogel $


#ifndef Lexicon_hh
#define Lexicon_hh

#include <string>
using std::string;
#include <iostream>
using std::istream;

#include <Types.hh>
#include <Vocabulary.hh>
#include <SparseArray.hh>
#include <Cognates.hh>

/**
   CLexcion: base class to store probabilities for a statistical lexicon.
   The lexicon has access to a source language vocabulary and a target language
   vocabulary, to map between word IDs and strings.

   CLexicon is organized for quick access to all f for given e.  The probabilities 
   are stored in a sparse matrix, which is an array of sparse arrays.  We have 
   for each target word e one sparse array, which contains pairs ( f, p( f | e ) ),
   where the f is stored in form of the word index. During training a second 
   sparse matrix is used to collect the counts N( f, e).  Renormalizing the counts into
   probabilities runs then over all entries in one line (i.e. one sparse array),
   sums the counts, and sets the updated probabilies as p( f | e ) = N( f, e) / Sum N( f', e ).
   The updating of counts and renormalization is implemented in the derived class 
   CLexiconForTraining (see LexiconTrain.hh).

   When a lexicon is used during decoding, then a preferred organization is over the 
   source words.  I.e. for each source word f we have a sparse array which contains
   the lexicon entries ( f, p( f | e ) ).  This organization was implemented in the 
   derived class CLexiconForSearch, which is currently not used and not part of STTK.

   The lexicon has the following additional features:

   - smoothing: part of the probability mass is reserved for unseen events, which are
   not explicitely stored in the lexicon.  Probabities, which are smaller then the 
   probability for unseen events are not stored in the lexicon, but pruned.  This means
   that such a word pair is treated as an unseen pair.

   - expansion of sparse arrays into full arrays for speed-up.  As the sparse array
   requires binary search to find the index, this slows down access to the lexicon.
   Training runs faster if this search is avoided for high frequency target words.

   - probabilities for cognates can be boosted (has not been used for a long time, 
   so beware!)

   - easy interpolation of two lexicons using the funcktions to multiply a lexicon
   with a constant and add two lexicon.


*/



class CLexicon
{
public:
  // ----- Construktor & Destruktor -----
  CLexicon();
  ~CLexicon();

  // ----- Initialisation -----
  //BOOL Init( const CVocabulary *pVocF, const CVocabulary *pVocE );
  // BOOL Init( int SizeVocF, int SizeVocE );

  //! Init the lexicon with pointers to source and target vocabulary,
  //! and optionally a pointer to a list of cognates
  BOOL Init( CVocabulary *pVocF, CVocabulary *pVocE,
             CCognates *pCognates = NULL );

  //! Return pointer to the source vocabulary
  inline const CVocabulary * SourceVocabulary() const;

  //! Return pointer to the target vocabulary
  inline const CVocabulary * TargetVocabulary() const;

  //! Return pointer to the underlying sparse matrix containing the probabilities
  inline const CSparseMatrix & Probabilities() const;

  //! Return pointer on modifiable underlying sparse matrix containing the probabilities.
  inline CSparseMatrix & Probabilities() { return m_Pr; }

  //! Expand p(.|e), log(p(.|e)) and count(.,e) to full arrays for all words in target vocabulary
  BOOL ExpandAll();

  //! Expand lexicon for all target vocabulary words in file FileName
  BOOL Expand( const char *FileName );

  //! Expand lexicon for all target vocabulary words given in array
  BOOL Expand( const CArray<int> &WordIndices );

  //! Undo all expansions, i.e. write p(.|e), log(p(.|e)) and count(.,e) back into sparse arrays
  BOOL Shrink();

  //! Set the smoothing factor used in renormalization
  void SetLambdaSmooth( float LambdaSmooth );

  //! Get probability for word pair ( f, e);
  inline float Pr( CWordIndex f, CWordIndex e ) const;

  //! Calculate probability for phrase pair as product over source words,
  //! sum over target words (IBM1 alignment model style)
  inline float Pr( const CArray< CWordIndex > *pPhraseF,
                   const CArray< CWordIndex > *pPhraseE );

  //! Get default probability p( f | e), for unseen pairs ( . | e )
  inline float DefaultPr( CWordIndex e ) const;

  //! Get log probability for word pair ( f, e);
  inline float LogPr( CWordIndex f, CWordIndex e ) const;

  //! Calculate log probability for phrase pair as product over source words,
  //! sum over target words (IBM1 alignment model style)
  inline float LogPr( const CArray< CWordIndex > *pPhraseF,
                      const CArray< CWordIndex > *pPhraseE );

  //! Get log of default probability p( f | e), for unseen pairs ( . | e )
  inline float DefaultLogPr( CWordIndex e ) const;

  //! Return the size of the lexicon
  int Size() const;

  //! Get number of items (f, p(f|e)) stored in the lexicon for given e
  inline int NumberOfItems( CWordIndex e ) const ;

  //! Get n-th item (f, p(f|e)) for given e
  inline void GetItem( CWordIndex e, int Position, CWordIndex &f, float &Value ) const;

  //! Get n-th item (f, log p(f|e)) for given e
  inline void GetLogItem( CWordIndex e, int Position, CWordIndex &f, float &LogValue ) const;

  //! Compute log p(f|e) from p(f|e) for all f and e, 
  //! log probabilities are stored seperately from probabilities
  BOOL Log();

  //! Renormalize probabilities
  void Renormalize();

  //! Test, if probabilities add up to 1
  BOOL TestProbabilities() const;
  
  //! Multiply all lexicon items with Value
  inline void Mul( float Value );

  //! Multiply p(.|e) with e-th component in array
  inline void Mul( const CNumericalArray< float > *pValues );

  //! Add Value to all lexicon items
  inline void Add( float Value );

  //! Add second lexicon to current lexicon
  inline void Add( const CLexicon *pLexicon );


  //! Display the lexicon
  void Display() const;


  //! Store lexicon to file.
  BOOL Store( const char *pFileName, int dataVersion = 5 );

  //! Load lexicon from file.
  BOOL Load( const char *pFileName );


  // Error constants for IO
  //! No error. Is not 0, so if(!open()) still works.
  static const int IOERR_OK = 1;
  //! Error while opening file for read.
  static const int IOERR_OPENREAD = 2;
  //! Error while opening file for write.
  static const int IOERR_OPENWRITE = 3;
  //! General error while writing file.
  static const int IOERR_WRITE = 4;
  //! Syntax error while reading "lexicon" file, dataversion 5.
  static const int IOERR_SYNTAX_V5 = 5;
  //! Unsupported or unknown version, wrong subversion etc.
  static const int IOERR_BAD_VERSION = 6;

  //! Store word translation probabilities, "dataversion 5". The underslash is there on purpose.
  int Store_Dataversion5( ostream& out);
  //! Load a lexicon which was previously stored with store_dataversion5(). Please leave the underslash.
  int Load_Dataversion5( istream& in );

  //! Store lexicon as triplets  f  e  p( f | e), f and e as strings,
  //!but restrict to entries with  probability above given threshold
  BOOL StoreWithWords( const char *pFileName, float PruneLevel = 0.000001 ) const;

  //! Store lexicon as triplets  e  f  p( f | e), f and e as strings,
  //!but restrict to entries with  probability above given threshold
  BOOL StoreWithWordsGIZA( const char *pFileName, float PruneLevel = 0.000001 ) const;

  //! Load lexicon from file which contains triplets  f  e  p( f | e), f and e 
  //! as strings
  BOOL LoadWithWords( const char *pFileName );

  //! Load lexicon from file which contains triplets  e  f  p( f | e), f and e 
  //! as strings;  this is the format written by GIZA
  BOOL LoadWithWordsGIZA( const char *pFileName );

  //! Store counts as triplets  f  e  p( f | e), f and e as strings,
  //!but restrict to entries with  probability above given threshold
  BOOL StoreCountsWithWords( const char *pFileName, float PruneLevel = 0.001 ) const;


protected: // ----- METHODS -----

  void Init();



protected: // ----- VARIABLES -----


  CSparseMatrix m_Pr;                //!< translation probabilities ( = lexicon)
  CSparseMatrix m_LogPr;             //!< logarithm of translation probabilities

  CArray< BOOL > m_IsExpanded;       //!< for which words e is p(f|e) full array

  CMatrix< float > m_ExpandedPr;     //!< translation probabilities in full arrays
  CMatrix< float > m_ExpandedLogPr;  //!< logarithm thereof

  CSparseMatrix m_Counts;            //!< translation counts
  CMatrix<float> m_ExpandedCounts;   //!< translation counts in full arrays

  //! size of source vocabulary associated with lexicon,
  //! can be smaller then vocabulary used for display
  //! but needs to be consistent with that vocabulary
  CWordIndex m_VocSizeF;

  //! size of source vocabulary associated with lexicon,
  //! can be smaller then vocabulary used for display
  //! but needs to be consistent with that vocabulary
  CWordIndex m_VocSizeE;

  //! pointer to source vocabulary, for display of lexicon
  CVocabulary *m_pVocF;
  //! pointer to target vocabulary, for display of lexicon
  CVocabulary *m_pVocE;

  //bool m_StoreTWithWords;

  //! Parameter for smoothing the probabilities
  float m_LambdaSmooth;

  //! Pinter to list of cognates
  CCognates *m_pCognates;

};






// ------------------------------------------------------------------
// Inline Functions for CLexicon
// ------------------------------------------------------------------


inline const CVocabulary *
CLexicon::SourceVocabulary() const
{
  return m_pVocF;
}



inline const CVocabulary *
CLexicon::TargetVocabulary() const
{
  return m_pVocE;
}



inline const CSparseMatrix &
CLexicon::Probabilities() const
{
  return m_Pr;
}



inline float
CLexicon::Pr( CWordIndex f, CWordIndex e ) const
{
  //  Trace0( 3, "CLexicon::Pr  " << "f: " << f << " e: " << e );
  //  if ( (f <= m_VocSizeF) && (e <= m_VocSizeE) ) {
  if ( ( e >= m_Pr.MinRow() ) && ( e <= m_Pr.MaxRow() ) ) {
    if ( m_IsExpanded[ e ] ) {
      //  Trace( 3, " CLexicon::Pr Expanded " );
      return m_ExpandedPr.GetAt( e, f );
    } else {
      // Trace( 3, " CLexicon::Pr Not Expanded " );
      return m_Pr.GetAt( e, f );
    }
  } else {
    return 1.0e-10;
  }
}


inline float
CLexicon::Pr( const CArray< CWordIndex > *pPhraseF,
              const CArray< CWordIndex > *pPhraseE )
{
  float P = 1.0;
  float SumP = 0.0;
  
  for ( int j = pPhraseF->MinIndex(); j <= pPhraseF->MaxIndex(); j++ ) {
    SumP = 0.0;
    for ( int i = pPhraseE->MinIndex(); i <= pPhraseE->MinIndex(); i++ ) {
      SumP += Pr( (*pPhraseF)[ j ], (*pPhraseE)[ i ] );
    }
    P *= SumP;
  }      
  return P;
}



inline float
CLexicon::DefaultPr( CWordIndex e ) const
{
  return m_Pr.DefaultValue( e );
}



inline float
CLexicon::LogPr( CWordIndex f, CWordIndex e ) const
{
  //  if ( (f <= m_VocSizeF) && (e <= m_VocSizeE) ) {
  if ( ( e >= m_LogPr.MinRow() ) && ( e <= m_LogPr.MaxRow() ) ) {
    if ( m_IsExpanded[ e ] ) {
      //	  Trace( 3, "CLexicon::LogPr Expanded "   << f << " " << e );
      return m_ExpandedLogPr.GetAt( e, f );
    } else {
      //	  Trace( 3, "CLexicon::LogPr Not Expanded "  << f << " " << e );
      return m_LogPr.GetAt( e, f );
    }
  } else {
    return log( 1.0e-10 );
  }
}



inline float
CLexicon::LogPr( const CArray< CWordIndex > *pPhraseF,
                 const CArray< CWordIndex > *pPhraseE )
{
  float LogP = 0.0;
  float SumP = 0.0;

  for ( int j = pPhraseF->MinIndex(); j <= pPhraseF->MaxIndex(); j++ ) {
    SumP = 0.0;
    for ( int i = pPhraseE->MinIndex(); i <= pPhraseE->MaxIndex(); i++ ) {
      SumP = Pr( (*pPhraseF)[ j ], (*pPhraseE)[ i ] );
    }
    LogP += log( SumP );
  }
  return LogP;
}



inline float
CLexicon::DefaultLogPr( CWordIndex e ) const
{
  if ( e <= m_VocSizeE ) {
    return m_LogPr.DefaultValue( e );
  } else {
    return log( 1.0e-10 );
  }
}



inline int
CLexicon::NumberOfItems( CWordIndex e ) const
{
  // Trace( 5, m_Pr.NumOfRows() << " " <<  m_Pr.NumOfCols() );
  return m_Pr.SizeOfRow( e );
}


inline void
CLexicon::GetItem( CWordIndex e, int Position, CWordIndex &f, float &Value ) const
{
  // doesn't work without assignment to local variable - mistery

  CWordIndex F = 0;

  m_Pr.GetRow( e ).GetItem( Position, F, Value );
  //m_Pr.GetRow( e ).GetItem( Position, F, Value );
  //   Trace( 5, "Position = " << Position << "  f = " << F
  // 	 << "  Value = " << Value );
  f = F;
}


inline void
CLexicon::GetLogItem( CWordIndex e, int Position, CWordIndex &f, float &LogValue ) const
{
  // doesn't work without assignment to local variable - mistery

  CWordIndex F = 0;

  m_LogPr.GetRow( e ).GetItem( Position, F, LogValue );
  //   Trace( 5, "Position = " << Position << "  f = " << F
  // 	 << "  LogValue = " << LogValue );
  f = F;
}



inline int
CLexicon::Size() const
{
  return m_Pr.NumOfRows();
}



inline void
CLexicon::Add( float Value )
{
  m_Pr.Add( Value );
}



inline void
CLexicon::Mul( const CNumericalArray< float > *pValues )
{
  m_Pr.Mul( *pValues );
}



inline void
CLexicon::Mul( float Value )
{
  m_Pr.Mul( Value );
}



inline void
CLexicon::Add( const CLexicon *pLexicon )
{
  m_Pr.Add( pLexicon->m_Pr );
}


#endif
