// CTM-IBM1

// $Id: TM_IBM1.hh 1743 2006-06-27 02:31:01Z vogel $


/**
   CTM-IBM1: Translation Model IBM1, Mixture Densities without alignments

   This implementation realizes the training algorithm of the IBM1 
   alignment model.

   Additional features:

   - Sentences with high perplexity can be pruned.  Of all sentence pairs
   in the corpus those are removed, which have highest perplexity.

   - Sentences can be split, based on split words.  Of all possible 
   pairs of split points in a sentence pair the one is selected,
   which reduces alignment perplexity most.  Of all sentences in the 
   corpus those are split, for which the perplexity reduction is highest.

*/   


#ifndef __tm_IBM1__
#define __tm_IBM1__

#include <config.h>

#include <Types.hh>
#include <Text.hh>
#include <TM.hh>
#include <Segmenter.hh>


class CTM_IBM1 : public CTranslationModel
{
public:
  CTM_IBM1();
  ~CTM_IBM1();

  //! Switch sentence pruning on
  void SetSentencePrune( int SentencePrune );
  //! Number of iteration in which pruning should start.
  //! We first need to train the lexicon for some iterations.
  void SetFirstPruningIteration( int FirstPruningIteration );

  //! Number of iteration in which sentence segmentation should start.
  //! We first need to train the lexicon for some iterations.
  void SetFirstSegmentingIteration( int FirstSegmentingIteration );
  //! Number of sentences which should be split per iteration.
  void SetNumberOfSplits( int NumberOfSplits );
  //! Set the words which function as split points (typically punctuation).
  void SetSplitWords( const CString &SplitWords );
  //! Set the threshhold for perplexity reduction required to split a sentence.
  void SetSplitThreshold( double SplitThreshold );
  //! How many split points are required: both sides (=2), one side (=1),
  //! none at all, i.e. all words can be split points (=0).
  void SetRequiredSplitPoints( int RequiredSplitPoints );

  //! Train the alignment model for specified number of iterations.
  bool Train();
  //! Calculate the perplexity on the training corpus.
  double Perplexity();
  //! Calcluate the perplixity for the Viterbi alignment on the traning corpus.
  double PerplexityForMaximum();

  //! Test if probabilities add up to 1.
  virtual bool TestProbabilities();

  //! Store the text modified by pruning and/or sentence splitting.
  bool StoreBiText( const char *pFileName );

protected:
  bool NextIteration();
  //void RenormalizeT();

  void DisplayBiText();
  void DisplayBadSentences();
  void RemoveBadSentences();

  void CheckBoundaries( const CSentence *pSentence );

protected:
  //double m_LambdaSmooth;

  //CArray<double> m_pLambda;
  //CDblMatrix *m_pT1;

  //! Array to store perplexity results from different interations.
  //! AlignmentPerplexity is constant for IBM1, but should be displayed
  CArray<double> m_AlignPerplexity;

  int m_SentencePrune;
  int m_FirstPruningIteration;
  CSortedArray<double> m_BadSentenceValues;
  CArray<int> m_BadSentenceNumbers;

  // to split sentences
  int m_FirstSegmentingIteration;
  CArray< CString > *m_pBoundaryWords;
  // array over wordindices, 1 = is boundary word 
  CNumericalArray< int > *m_pIsBoundaryWord;
  
  CSegmenter *m_pSegmenter;
  // number of splits in each iteration
  // default: 0
  int m_NumberOfSplits;

  //! Gain in perplexity required to split sentence pair;
  //! default: 0.0.
  double m_SplitThreshold;

  //! Number of splitpoints required to trigger calculation of perplexity gain.
  //! 2: split word in both sentences, 1: in at least one sentene, 0: none required, 
  //! i.e. all positions in the sentence pair near the diagonal are considered.
  //! default: 2
  int m_RequiredSplitPoints;
};

#endif


