/**
   @file Text.hh

   Provide base class for text handling

   CText

   ----------------------------------------------------------------\n
   Imports\n
   ----------------------------------------------------------------\n
   type CWordIndex\n
   class CVocabularyItem\n
   class CVocabualry\n
   all imported from Vocabulary.hh\n
   \n
   ----------------------------------------------------------------\n
   Exports\n
   ----------------------------------------------------------------\n
   class CWordInfo;\n
   class CSentence;\n
   class CText;\n
   typedef long TPosition;\n

   CVS id: $Id: Text.hh 1742 2006-06-27 02:28:05Z vogel $ 
*/

#ifndef __TEXT_DT_H__
#define __TEXT_DT_H__


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include <iostream>
#include <map>
using std::map;

#include <Array.hh>
#include <Types.hh>
#include <Vocabulary.hh>


#include <TextFilterBoostRegex.hh>
//using namespace std;






// typedef size_t TPosition;

//! A position in a text
typedef long TPosition;

// a map containing the number of occurencies for each word ID in the map.
typedef long int TWordCount;
typedef map<CWordIndex, TWordCount> CWordIdx2CountMap;

class CWordInfo;
class CSentence;
class CText;


/**
   CSentence: Data structure to handle one line of a text file, 
   which typically is one sentence.

   A sentence is an array of word indices, and optionally 
   an array of strings and an array of additional word information,
   stored as CWordInfo structures.

   The words in a sentence run from position 1 to sentence length.
   Notice: Sentence[ i ] is word index at position i.  As CSentence
   is derived from CArray< CWordIndex > the array operations can
   be applied to a CSentence object.  Esp. the MinIndex() and 
   MaxIndex() functions to determine start and end point of a 
   sentence.
*/


class CSentence : public CArray< CWordIndex >
{
public:
  CSentence();
  CSentence( const CSentence &S );
  CSentence( const CString &S, CVocabulary *pVocabulary );
  ~CSentence();

  CSentence & operator = ( const CSentence &Sentence );

  //! Append a word to the current sentence
  void Append( CWordIndex WordIndex );

  //! Append another sentence to the current sentence
  void Append( CSentence &Sentence );

  //! Length of the sentence, i.e. number of words.
  inline int Length() const;

  //! Clear sentence, length is then 0.
  void Clear();

  //! Return a string representation of the sentence
  //! at the moment only words, no additional word information.
  CString AsString( const CVocabulary *pVocabulary ) const;

  //! Display entire sentence on screen or write it into some stream.
  //! Default: only the words without additional word information
  //! if WithWordInfo == true, then with additional word information.
  void Display( const CVocabulary *pVocabulary, bool WithWordInfo = false ) const;

  //! Write sentence as string to OutStream, converting word indices
  //! according to the given vocabulary.
  void Write( std::ostream &OutStream, const CVocabulary *pVocabulary,
              bool WithWordInfo = false ) const;

  //! Convert sentence into a string, using the given vocabulary.
  //! If words have annotations in attached CWordInfo structures,
  //! ouput the source side of the word info.
  void GetSourceString( CString &SourceString,
                        const CVocabulary *pVocabulary = NULL ) const;


  //! Convert sentence into a string, using the given vocabulary.
  //! If words have annotations in attached CWordInfo structures,
  //! ouput the target side of the word info.
  void GetTargetString( CString &TargetString,
                        const CVocabulary *pVocabulary = NULL ) const;


  //! Read sentence from buffer and convert into indices
  //! using the given vocabulary.  Returns the number of words.
  int ReadFromBuffer( const char *pBuffer, CVocabulary *pVocabulary );

  friend std::istream &operator >> ( std::istream &InStream, CSentence &Sentence );
  friend std::ostream &operator << ( std::ostream &OutStream, const CSentence &Sentence );

  //! Convert sentence to CWordIndex *.  
  //! ATTENTION: returns 0-based array (don't ask why).
  inline operator CWordIndex const *();

  //! Access to word stored as wordindex or as string.
  const CString & Word( int Index, const CVocabulary *pVocabulary = NULL ) const;

  //! Set the word pWord at the given Index ( 1 <= Index <= current length + 1 )
  //! using the given vocabulary.
  void SetWord( const char *pWord, int Index,
                CVocabulary *pVocabulary = NULL );

  //! Access to word info, return NULL if no wordinfo at given index exists.
  inline CWordInfo * WordInfo( int Index ) const;

  //! Set wordinfo at given index
  bool SetWordInfo( CWordInfo *WordInfo, int Index );

  //!  Calculate counts for the words in the sentence.
  void GetWordCounts( CWordIdx2CountMap &Result ) const;

  //!  Calculate counts for the words in the sentence.
  void GetWordCounts( CWordIdx2CountMap &Result, const CVocabulary &Vocab) const;

private:
  //! For words not in the vocabulary we can store the string.
  CArray< CString > *m_pWords;

  //! Here, we store the additional word infos.
  //! For minimal memory overhead we use a pointer to 
  //! an array of pointers
  CArray< CWordInfo * > *m_pWordInfos;
}
;




/**
   CWordInfo: To handle additional information for each word.

   At the moment disigned towards labeling, i.e. a label
   may be followed by a list of source words and optionally
   a list of target words in the following format:
   $label { s1 ... sn # t1 ... tm  }.  
   Can be nested: $label { $label1 { ... # ... } s # $label1 t }
*/

class CWordInfo
{
public:
  CWordInfo();
  CWordInfo( const CWordInfo * );

  //char * ReadFromBuffer( char *Buffer );
  //void Write( std::ostream& os );

  friend std::istream &operator >> ( std::istream &Stream, CWordInfo &Info );
  friend std::ostream &operator << ( std::ostream &Stream, const CWordInfo &Info );

  inline void AddSourceWord( const CString &Word );
  inline void AddTargetWord( const CString &Word );

  inline const CString & FirstWordTarget() const;
  inline const CString & LastWordTarget() const;

  inline CSentence & SourceWords();
  inline CSentence & TargetWords();
  inline double & Score();

  CString SourceString() const;
  CString TargetString() const;

  int GetSourceWords( CArray< CString > &Words ) const;
  int GetTargetWords( CArray< CString > &Words ) const;

  inline int SourceStringLength() const;
  inline int TargetStringLength() const;

  inline bool HasSourceString() const;
  inline bool HasTargetString() const;
  inline bool HasScore() const;

  void SwapSourceTarget();


private:
  CSentence m_SourceWords;
  CSentence m_TargetWords;

  double m_Score;
};






/**
   @class CText

   A text is an array of sentences.

   The sentences are numbered from 1 to NumberOfSentences.

   A text can be loaded in two formats:

   - indexed: file contains vocabulary plus wordindices
              each sentence end is marked with index of EOS (&lt;/s>).

   - ascii:   file contains plain text, but each word can have
              additional information in the form of

	      { source_side_info # target_side_info # score }.

   A text can be loaded into the memory all at once, or
   once sentence at a time.  The FirstSentence(), NextSentence()
   functions support both ways.
*/

class CText
{
public:
  CText ();
  ~CText ();

  void Clear ();

  inline bool Allocate( int MaxNumOfSentences );
  inline bool AllocateSentenceLength( int Increment );

  // load and store text

  /**
	 Read one sentence at a time, i.e. when NextSentence() is used,
	 Load then only opens the file, or load the entire text into memory at load time.
	 Default is false, i.e. load the entire text.

	 Notice: this fuctionality has been added to train on very large corpora.
	 Some functions of CText make sense only when the entire text is in memory.
  **/
  void ReadOneByOne( bool Flag = false );

  //! Open a plain text file.
  void OpenAsciiText( const char *name );

  //! Load an indexed text file.
  int LoadIndexedText( const char *name );

  //! Load a plain text file
  int LoadAsciiText( const char *name );

  //! Add a sentence to the end of the sentence.
  bool AppendSentence( CSentence *pSentence );

  //! Number of words (without sentence end markers)
  inline int Size () const;

  inline int NumberOfSentences() const;

  //! Index of first sentence in text.
  inline int MinIndex() const;

  //! Index of last sentence in text.
  inline int MaxIndex() const;

  //! Fetch the first sentence and set the position, where the sentence is located in the text.
  //! Return value is EOS (end of sentence is used both at start and of sentence), if a first
  //! sentence exists, otherwise EOT (end of text).
  CWordIndex FirstSentence( TPosition &PositionOfSentenceInText );

  //! Fetch the next sentence after current PositionOfSentenceInText and set the position
  //! of the next sentence, where it is located in the text.
  //! Return value is EOS if a first sentence exists, otherwise EOT (end of text).
  CWordIndex NextSentence( TPosition &PositionOfSentenceInText );

  //! Fetch first word in a sentence given by PositionOfSentenceInText, set the position of the word
  //! within that sentence, and return the WordIndex for that word.  Return value is EOS if PositionOfWordInSentence
  //! is larger than the sentence, and EOT if sentence is not in the text, i.e. PositionOfSentenceInText out of bounds.
  inline CWordIndex FirstWord ( TPosition PositionOfSentenceInText, TPosition &PositionOfWordInSentence ) const;

  //! In a sentence given by PositionOfSentenceInText fetch next word after PositionOfWordInSentence,
  //! set the position of the returned word to PositionOfWordInSentence, and return the WordIndex for that word.
  //! Return value is EOS if PositionOfWordInSentence is larger than the sentence, 
  //!and EOT if sentence is not in the text, i.e. PositionOfSentenceInText out of bounds.
  inline CWordIndex NextWord( TPosition PositionOfSentenceInText, TPosition &PositionOfWordInSentence ) const;

  //! Return a pointer to the sentence at PositionOfSentenceInText.
  inline CSentence *GetSentence( TPosition PositionOfSentenceInText );

  //! Fill the array Words with the word indices of the sentence at 
  //! ATTENTION: Words is a 0-based array (no, you shouldn't ask why)
  inline int GetSentence( TPosition PositionOfSentenceInText, 
			   CWordIndex const *&WordIndices ) const;

  //void       IndexToTPosition ( int index, TPosition &possent ) const;

  //CWordIndex GetWordAt ( TPosition posword ) const; 

  //! Direct access to a word at PositionOfWordInSentence in sentence PositionOfSentenceInText.
  //! Return word index if positions are valid, otherwise asserts or crashes.
  CWordIndex GetWordAt( TPosition PositionOfSentenceInText, int PositionOfWordInSentence ) const; 

  //! Return the number of words in sentence at PositionOfSentenceInTextt
  inline int SentenceLength( TPosition PositionOfSentenceInText ) const;

  // functions to modify text

  //! Put a sentence at position PositionOfSentenceInText, overwrites existing sentence.
  void PutSentence ( TPosition PositionOfSentenceInText, const CSentence &Sentence );

  //! Insert a sentence at position PositionOfSentenceInText, shifting existing sentences.
  //! /ToDo: Make parameters for PutSentence() and InsertSentence() consisten.
  void InsertSentence ( TPosition PositionOfSentenceInText, CSentence *pSentence );

  //! Delete the words in sentence at PositionOfSentenceInText; does not remove it from the text.
  void ClearSentence ( TPosition PositionOfSentenceInText );
  
  //! Set word given by WordIndex to position into PositionOfWordInSentence in
  //! sentence at PositionOfSentenceInText, overwriting an existing word or expanding the sentence.
  void SetWordAt ( TPosition PositionOfSentenceInText, int PositionOfWordInSentence, CWordIndex WordIndex ); 

  //! Set pointer to TextFilter object for normalization of text during loading
  void SetTextFilter( CTextFilterBase *pTextFilter );

  //! Find next sentence after PositionOfSentenceInText containing the word with wordindex WordIndex.
  TPosition FindSentence( TPosition PositionOfSentenceInText, CWordIndex WordIndex ) const;

  //! Set the vocabulary, CText does not own the vocabulary.
  inline bool SetVocabulary ( CVocabulary * Vocabulary );
  //! Get a pointer to the vocabulary.
   inline CVocabulary * Vocabulary();
   
  //! Display the total text with/without additional word information.
  void Display( bool WithWordInfo = false ) const;

  //! Write the total text with/without additional word information to OutStream.
  void Write( std::ostream &OutStream, bool WithWordInfo = false ) const;

  //! Display a single sentences with/without additional word information
  void DisplaySentence( TPosition PositionOfSentenceInText, bool WithWordInfo = false ) const;

  //! Write a single sentence with/without additional word information to OutStream.
  void WriteSentence( TPosition PositionOfSentenceInText, std::ostream &OutStream, bool WithWordInfo = false ) const;



protected:
  //! Test, if sentence index is within bounds.
  inline bool IsValidIndex( int Index ) const;

  //! File name of file from which text is loaded.
  CString m_FileName;

  //! The infile stream from which text is loaded.
  std::ifstream *m_pTextStream;

  //! Vocabulary to index the text.
  CVocabulary * m_pVocabulary;

  //! Text is array of sentences.
  CArray< CSentence * > m_Sentences;

  //! Text size (number of words).
  int m_TextSize;

  //! For memory allocation in the array of the sentences when loading a text. 
  int m_SentenceLengthIncrement;

  //! Flag: if true then read one line at a time for the text file,
  //!       if false load the entire text in Load().
  bool m_ReadOneByOne;
  
  //! For postprocessing of text
  CTextFilterBase* m_pTextFilter;
};


#include <Text_inl.hh>


#endif
