#include <VocabMapper.hh>

/**
   @brief Set source vocab. Only called during initialization.

   Allocate the mapping array, set some relevant members.
*/
void CVocabMapper::SetSrcVoc(const CVocabulary* pSrcVoc) {
  assert(pSrcVoc);
  m_pSrcVoc        = pSrcVoc;

  // Note we map EOS, which is not covered by First() ... Last()
  m_IndexMapSize   = pSrcVoc->Last() - pSrcVoc->First() + 1 + 1;
  assert(m_IndexMapSize > 0);
  assert(CVocabulary::UNK == 0);
  m_IndexMapOffset = 0;
  // reallocate mapping array
  if (m_pIndexMap != NULL) free(m_pIndexMap);
  m_pIndexMap      = (CWordIndex *)calloc(m_IndexMapSize,sizeof(CWordIndex));
} // CVocabMapper::SetSrcVoc(const CVocabulary* pSrcVoc)

/**
   @brief Map, insert words if necessary.

   Set the target vocab, calculate the index map, insert words into target vocab if necessary.

   @param pTgtVoc (Not NULL)
   Pointer on the target vocabulary where indices are mapped to.
*/
void CVocabMapper::MapOrInsert(CVocabulary* pTgtVoc) {
  assert(m_pIndexMap);

  assert(m_pSrcVoc);
  assert(pTgtVoc);
  assert(m_pSrcVoc->Last() - m_pSrcVoc->First() + 2 == m_IndexMapSize);

  m_pTgtVoc = pTgtVoc;

  // map UNK and EOS
  assert(CVocabulary::UNK == 0);
  m_pIndexMap[CVocabulary::UNK] == CVocabulary::UNK;
#ifdef VOCAB_STATIC_EOS
  assert(CVocabulary::EOS == 1);
  m_pIndexMap[CVocabulary::EOS] == CVocabulary::EOS;
#else
  assert(m_pSrcVoc->EOS == m_IndexMapSize - 1);
  m_pIndexMap[m_IndexMapSize-1] == pTgtVoc->EOS;
#endif

  // the current word (yes, a string)
  CString w;
  // The target index
  CWordIndex iTgt;

  // Loop over source vocab and map
  for (CWordIndex iSrc = m_pSrcVoc->First(); iSrc <= m_pSrcVoc->Last(); ++iSrc) {
    // get current word
    w = m_pSrcVoc->GetWord(iSrc);
    // get application wide index (source)
    iTgt = pTgtVoc->GetIndex(w);
    // if source index is UNK so must be target
    assert(!(iSrc == CVocabulary::UNK) || iTgt == CVocabulary::UNK);

    // insert if necessary
    if (iTgt == CVocabulary::UNK && iSrc != CVocabulary::UNK) {
      // insert!
      iTgt = pTgtVoc->AddWord(w);
    } // endif: insert vocab index into target vocab
    // if source index is not UNK, so can't be target now
    assert(!(iTgt == CVocabulary::UNK) || iSrc == CVocabulary::UNK);

    // store mapping
    assert(iSrc >= m_IndexMapOffset && iSrc-m_IndexMapOffset < m_IndexMapSize);
    m_pIndexMap[iSrc-m_IndexMapOffset] = iTgt;
  } // endfor: loop over source vocab
} // CVocabMapper::MapOrInsert(CVocabulary& TgtVocab)

/**
   @brief Map, to UNK if necessary.

   Set the target vocab, calculate the index map, map to UNK if necessary.

   @param pTgtVoc
   Pointer on the target vocabulary where indices are mapped to. If NULL,
   all indices are mapped to UNK.
*/
void CVocabMapper::MapOrUnk(const CVocabulary* pTgtVoc) {
  assert(m_pIndexMap);
  assert(m_pSrcVoc->Last()-m_pSrcVoc->First()+1 == m_IndexMapSize);

  m_pTgtVoc = pTgtVoc;

  // map UNK and EOS
  assert(CVocabulary::UNK == 0);
  m_pIndexMap[CVocabulary::UNK] == CVocabulary::UNK;
#ifdef VOCAB_STATIC_EOS
  assert(CVocabulary::EOS == 1);
  m_pIndexMap[CVocabulary::EOS] == CVocabulary::EOS;
#else
  assert(m_pSrcVoc->EOS == m_IndexMapSize - 1);
  m_pIndexMap[m_IndexMapSize-1] == pTgtVoc->EOS;
#endif

  // if tgt vocab is NULL ...
  if (pTgtVoc == NULL) {
    // ... set all to UNK
    for (unsigned int iSrc = 0; iSrc < m_IndexMapSize; ++iSrc)
      m_pIndexMap[iSrc] = CVocabulary::UNK;

  } else {
    // target vocab is NOT null

    // the current word (yes, a string)
    CString w;
    // The target index
    CWordIndex iTgt;

    // Loop over source vocab and map
    for (CWordIndex iSrc = m_pSrcVoc->First(); iSrc <= m_pSrcVoc->Last(); ++iSrc) {
      // get current word
      w = m_pSrcVoc->GetWord(iSrc);
      // get application wide index (source)
      iTgt = pTgtVoc->GetIndex(w);
      // if source index is UNK so must be target
      assert(!(iSrc == CVocabulary::UNK) || iTgt == CVocabulary::UNK);
      
      // store mapping
      assert(iSrc >= m_IndexMapOffset && iSrc-m_IndexMapOffset < m_IndexMapSize);
      m_pIndexMap[iSrc-m_IndexMapOffset] = iTgt;
    } // endfor: loop over source vocab
  } // endif: target vocab is NULL?
} // CVocabMapper::MapOrUnk(const CVocabulary* pTgtVoc)
