/**
   @file

   Provide a class to map vocabulary indices from one vocabulary to another.

   CVS file ID: $Id: VocabMapper.hh 891 2005-02-11 03:53:47Z dietmar $
 */

#include <Vocabulary.hh>

/**
   A class to map word indices from one CVocabulary object to another.

   When constructing a vocabulary mapper, you must at least define the source vocabulary,
   which stays constant during the whole lifetime of a vocabulary mapper. The target
   vocabulary can be provided by one of the constructors or later, using MapOrInsert()
   or MapOrUnk() (which also invalidates any previous mapping).

   If you provide a constant CVocabulary object to the constructor or call MapOrUnk()
   explicitely, words from the source vocabulary not known to the target vocabulary
   are mapped to CVocabulary::UNK.

   Providing a mutable CVocabulary object or calling MapOrInsert(), unknown words
   from the source vocabulary are inserted into the target vocabulary.

   Neither source nor target vocabulary are freed during destruction
   (which is what you probably want).

   Source and target vocabularies are only used when the index map is calculated
   (during MapOrInsert(), MapOrUnk() or one of the constructors). So if you modify one
   the vocabularies afterwards, this does not affect the index map.
   On the other hand, mapping an index is very fast.

   This class is NOT thread safe.
 */
class CVocabMapper {
private:
  //! The source vocabulary.
  const CVocabulary* m_pSrcVoc;
  //! The target vocabulary.
  const CVocabulary* m_pTgtVoc;
  //! The size of the map.
  unsigned int m_IndexMapSize;
  //! Source vocab's index of m_IndexMap[0].
  CWordIndex   m_IndexMapOffset;
  //! The map of word indices.
  CWordIndex  *m_pIndexMap;

  //! Set source vocabulary. Only called during initialization.
  void SetSrcVoc(const CVocabulary* pSrcVoc);
public:
  //! Constructor, insert words into target vocab if necessary.
  inline CVocabMapper(const CVocabulary* pSrcVoc, CVocabulary* pTgtVoc)
    : m_pIndexMap(NULL) { SetSrcVoc(pSrcVoc); MapOrInsert(pTgtVoc); }
  //! Constructor, set mapped index to UNK if not in target vocab or target vocab is NULL.
  inline CVocabMapper(const CVocabulary* pSrcVoc, const CVocabulary* pTgtVoc = NULL)
    : m_pIndexMap(NULL) { SetSrcVoc(pSrcVoc); MapOrUnk(pTgtVoc); }
  //! Destructor: free map.
  inline ~CVocabMapper() { if (m_pIndexMap != NULL) free(m_pIndexMap); }

  //! Map, insert words if necessary.
  void MapOrInsert(CVocabulary* pTgtVoc);
  //! Map, to UNK if necessary.
  void MapOrUnk(const CVocabulary* pTgtVoc = NULL);

  //! Map an index. Return UNK if out-of-bounds.
  inline const CWordIndex& operator[](const CWordIndex& idx) const
  { assert(m_pSrcVoc); return (idx >= m_IndexMapOffset && idx-m_IndexMapOffset < m_IndexMapSize ?
			       m_pIndexMap[idx - m_IndexMapOffset] :
			       CVocabulary::UNK); }

  //! Get associated source vocabulary.
  inline const CVocabulary& GetSrcVocabulary() const
  { assert(m_pSrcVoc); return *m_pSrcVoc; }
  //! Get associated target vocabulary.
  inline const CVocabulary& GetTgtVocabulary() const
  { assert(m_pTgtVoc); return *m_pTgtVoc; }
}; // class CVocabMapper
