#include <string.h>
#include <stdlib.h>
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


// CVocabulary
// Base class for vocabulary handling
//

#include <stdlib.h>
#include <stdio.h>
#include <string>
#include <fstream>
using std::ifstream;
using std::ofstream;

// enable / disable tracing for this module here
// #undef NO_TRACE
#define NO_TRACE 1
#include <Trace.hh>

#include <utils.hh>
#include <Vocabulary.hh>

using namespace std;

//static char rcsid_vocabulary_cc[] = "$Id: ";

CWordIndex CVocabulary::EOT       = -1;
CWordIndex CVocabulary::UNK       = 0;
CWordIndex CVocabulary::EMPTYWORD = 1;
CWordIndex CVocabulary::EOS       = 2;


CVocabularyItem::CVocabularyItem()
{
  m_IsContentWord = false;
  m_IsMultiWord = false;
  m_pWordIndices = NULL;
}


CVocabularyItem::CVocabularyItem( const CString &Word )
{
  //Trace( 0, "constructor CVocabularyItem::CVocabularyItem()" );
  m_Word = Word;
  m_IsContentWord = false;

  if ( Word.Pos( "_" ) >= 0 ) {
    m_IsMultiWord = true;
    m_pWordIndices = new CArray< CWordIndex >;
  } else {
    m_IsMultiWord = false;
    m_pWordIndices = NULL;
  }
}

CVocabularyItem::~CVocabularyItem()
{
  //Trace( 0, "destructor CVocabularyItem::~CVocabularyItem()" );
  delete m_pWordIndices;
}


void
CVocabularyItem::SplitWords()
{
}


/**
   @brief Constructor.

   Allocates entries for &lt;UNK> and sentence end $lt;/s> automatically.
 */
CVocabulary::CVocabulary()
{
  m_VocItems.SetIncrement( 1000 );

  m_StringForEmptyWord = CString( "<EMPTYWORD>" );

  // insert UNK
  m_VocItems[ UNK ] = new CVocabularyItem( "<UNK>" );
  m_VocIndices[ m_VocItems[ UNK ]->PointerToWord() ] = UNK;

  // insert EOS
  m_VocItems[ EOS ] = new CVocabularyItem( "</s>" );
  m_VocIndices[ m_VocItems[ EOS ]->PointerToWord() ] = EOS;

  // insert EMPTYWORD
  m_VocItems[ EMPTYWORD ] = new CVocabularyItem( m_StringForEmptyWord );
  m_VocIndices[ m_VocItems[ EMPTYWORD ]->PointerToWord() ] = EMPTYWORD;
} // CVocabulary


/**
   @brief Destructor.

   Frees <i>all</i> vocabulary items (including &lt;UNK> and &lt;/s>).
*/
CVocabulary::~CVocabulary()
{
  // delete the VocItems
  Clear();
  // delete EOS and UNK and NULL
  assert(m_VocItems[ UNK ]);
  delete m_VocItems[ UNK ];
  assert(m_VocItems[ EOS ]);
  delete m_VocItems[ EOS ];
  assert(m_VocItems[ EMPTYWORD ]);
  delete m_VocItems[ EMPTYWORD ];

  // don't know if this is really clean ... see Clear()
  m_VocIndices.clear();
} // ~CVocabulary



bool
CVocabulary::Allocate( int MaxNumOfVocItems )
{
  m_VocItems.SetIncrement( MaxNumOfVocItems );
  return true;
}


/**
   @brief Clear vocabulary.

   All vocabulary item except EOS, UNK, and NULL are deleted.

   \todo There is probably a memory leak here ...
 */
void
CVocabulary::Clear()
{
  Trace( 3, "CVocabulary::Clear" );
  // delete the VocItems
  for ( int i = m_VocItems.MinIndex(); i <= m_VocItems.MaxIndex(); i++ ) {
    //Trace( 0, "Delete item " << i );
    delete m_VocItems[ i ];
    m_VocItems[ i ] = NULL;
  } // endfor
  // delete the array itself
  m_VocItems.RemoveAll();

  // delete string->index mapping
  // TODO: what's that? first you clear it, then you delete[] it
  m_VocIndices.clear();
  // delete strings in hashmap
  sgi::hash_map< const char *, CWordIndex, sgi::hash<const char*>, eqstr >::const_iterator Iter;
  for ( Iter = m_VocIndices.begin(); Iter != m_VocIndices.end(); ++Iter ) {
      // TODO: this should be TRACE
    //    Trace(4, "CVocabulary::Clear(): delete " << Iter->first << " " << Iter->second << endl);
    delete [] Iter->first;
  }
  //for ( Iter = m_VocIndices.begin(); Iter != m_VocIndices.end(); ++Iter ) {
  //cout << "delete " << Iter->first << " " << Iter->second << endl;
    //delete [] Iter->first;
  //}

  // re-insert UNK and EOS and NULL
  m_VocItems[ UNK ] = new CVocabularyItem( "<UNK>" );
  m_VocIndices[ m_VocItems[ UNK ]->PointerToWord() ] = UNK;

  m_VocItems[ EOS ] = new CVocabularyItem( "</s>" );
  m_VocIndices[ m_VocItems[ EOS ]->PointerToWord() ] = EOS;

  m_VocItems[ EMPTYWORD ] = new CVocabularyItem( m_StringForEmptyWord );
  m_VocIndices[ m_VocItems[ EMPTYWORD ]->PointerToWord() ] = EMPTYWORD;

} // CVocabulary::Clear()



bool
CVocabulary::Load( const char *FileName )
{
  assert(FileName);
  ifstream InStream( FileName );

  if ( !InStream.good() ) {
      cerr <<  "CVocabulary::Load(): Vocabulary file " << FileName << " was not opened for read" << endl << flush;
      return false;
  } // endif: error opening file

  bool Result = Load( InStream );
  InStream.close();

  return Result;
} // Load


/**
   @brief Load vocabulary from input stream.

   The format for a vocabulary file (although Load() can load vocabulary from any input
   stream) is one word per line. Words that are already known to the vocabulary are skipped.

   Comment lines starting with '#' are skipped <i>at the beginning</i>. Words are read from the input
   stream until EOF or the word "</s>" is encountered.
 */
bool
CVocabulary::Load( istream &InStream )
{
  static const int BUFLEN = 2560;
  char str[ BUFLEN ];

  // remove old items
  Clear();

  if ( !InStream.good() ) {
      cerr <<  "CVocabulary::Load(): Could not read from ifstream" << endl << flush;
      return false;
  } // endif: error reading stream
  
  // skip comment lines
  InStream.getline( str, BUFLEN );
  Trace( 1, str );
  while ( (str[ 0 ] == '#') && !InStream.eof() ) {
    InStream.getline( str, BUFLEN );
  }

  if ( InStream.eof() ) {
    Trace( 0, "No vocabulary item read, empty file" );
    return false;
  }

  CVocabularyItem *pVocItem;
  CWordIndex w;

  // read lines and make from strings new vocabulary items
  // until EOF or EOS (word "</s>")

  // we assume that vocabulary is started by <UNK> and ended by </s>,
  // so these entries are ignored (already added by constructor).
  
  // index is the first free index available for insertion
  int index = m_VocItems.MaxIndex() + 1;
  assert( index >= EMPTYWORD );
  do {
    // str contains last line read

    // check if word already exists
    w = GetIndex( str );
    // word exists iff ( index(w) == UNK => w == "<UNK>" )
    if ( w != UNK || strcmp(str,m_VocItems[ UNK ]->PointerToWord()) == 0 ) {
      assert(strcmp(str,GetWord(w))==0);
      Trace(4,"CVocabulary::Load(): " << str << " already exists, skipping.");
    } else {
      // new word

      // new item
      Trace( 5, "CVocabulary::Load(): create new CVocabularyItem" );
      pVocItem = new CVocabularyItem( str );
      if ( ! pVocItem ) {
        cerr << "CVocabulary::Load(): ERROR: Out of memory while reading vocabulary item" << endl;
        exit( -1 );
      }
	
      // append to list
      Trace( 4, "CVocabulary::Load(): " << index << "\t" << str );
      m_VocItems[ index ] = pVocItem;
	
      // insert into map for quick string to index conversion
      m_VocIndices[pVocItem->PointerToWord()]= index;
      ++index;
    } // endif: word already seen?
  
    // read next line
    InStream.getline( str, BUFLEN );
    //Trace( 1, index << "\t" << str );
	
  } while ( !InStream.eof() && (strcmp( str, m_VocItems[ EOS ]->PointerToWord()) != 0) );
  
  Trace( 2, "CVocabulary::Load(): " << Size() << " proper vocabulary items read" );
  
  // test output of string to index map
  //sgi::hash_map< const char *, CWordIndex, hash< const char *>, eqstr >::const_iterator Iter;
  //for ( Iter = m_VocIndices.begin(); Iter != m_VocIndices.end(); Iter++ ) {
  //cout << Iter->first << " " << Iter->second << endl;
  //}
  

  // split multi-word items
  UpdateMultiWords();

  return true;
} // Load



bool
CVocabulary::Store( const char *FileName, bool WithSGML ) const
{
  ofstream OutStream( FileName );

  if ( !OutStream.good() ) {
    cerr <<  "CVocabulary::Store( " << FileName << " ): File was not opened for write\n";
    exit (-1);
  }

  bool Result = Store( OutStream, WithSGML );
  OutStream.close();

  return Result;
} // Store


/**
   @brief Store vocabulary to stream.
   @param OutStream The stream to serialize the vocabulary to.
   @param WithSGML Flag whether to output UNK and EOS or not.

   This store method simply outputs one word per line to the output
   stream OutStream.

   If WithSGML is "true", "&lt;UNK>" is output at the beginning and
   "&lt;/s>" at the end.
 */
bool
CVocabulary::Store( ostream &OutStream, bool WithSGML ) const
{
  if ( !OutStream.good() ) {
    cerr << "CVocabulary::Store(): File could not be opened for write\n";
    exit (-1);
  }
  
  CWordIndex index = 0;

  if ( WithSGML ) {
    OutStream << m_VocItems[ UNK ]->Word() << endl;
  }

  for ( index = First(); index <= Last(); index++ ) {
    OutStream << m_VocItems[ index ]->Word() << endl;
  }

  if ( WithSGML ) {
    OutStream << m_VocItems[ EOS ]->Word() << endl;
  }

  return true;
} // Store



// ********************************************************************
// Search-Methods

/**
 @brief Quickly find the index of a given word.
 
 This method uses a hashmap to lookup an index.
 */
CWordIndex
CVocabulary::GetIndex( const char* word ) const
{
  Trace( 2, "Vocabulary:FindIndex " << *word );
  
  if ( m_VocIndices.count( word ) == 0 ) {
    return 0;
  }else{
    
    sgi::hash_map< const char *, CWordIndex, sgi::hash<const char*>, eqstr >::const_iterator FindIter;
    FindIter = m_VocIndices.find( word );
    
    if ( FindIter != m_VocIndices.end() ) {
      return (*FindIter).second;
    } else {
      return UNK;
    }
  }
  
} // GetIndex



/**
   @brief Quickly find the index of a given word.

   This method uses a hashmap to lookup an index.
 */
CWordIndex
CVocabulary::GetIndex( const CString & word ) const
{
  Trace( 2, "Vocabulary:FindIndex " << (const char *)word );

  if ( m_VocIndices.count( (const char*)word ) == 0 ) {
    return 0;
  }else{

    sgi::hash_map< const char *, CWordIndex, sgi::hash<const char*>, eqstr >::const_iterator FindIter;
    FindIter = m_VocIndices.find( (const char *)word );

    if ( FindIter != m_VocIndices.end() ) {
      return (*FindIter).second;
    } else {
      return UNK;
    }
  }
  
} // GetIndex



/**
   @brief Quickly find the index of a given word.

   This method uses a hashmap to lookup an index.
 */
CWordIndex
CVocabulary::GetIndexAddWord( const CString & word )
{
  Trace( 2, "Vocabulary:FindIndex " << (const char *)word );

  sgi::hash_map< const char *, CWordIndex, sgi::hash<const char*>, eqstr >::const_iterator FindIter;
  FindIter = m_VocIndices.find( (const char *)word );
  
  if ( FindIter != m_VocIndices.end() ) {
    return (*FindIter).second;
  } else {
    // we add new words
    return AddWord( word );
  }
  
} // GetIndexAddWord




/**
   @brief Search the vocabulary sequentially for a word.

   \deprecated
   This is rather for debugging purposes only. Use GetIndex() instead.
 */
CWordIndex
CVocabulary::SearchWord( const char *word ) const
{
  CWordIndex i;
  
  for ( i = m_VocItems.MaxIndex(); i >= m_VocItems.MinIndex(); i-- ) {
    if( strcmp( word, m_VocItems[ i ]->Word() ) == 0 )
      {
        return i;
      }
  }

  // smaller than <UNK>, so return <UNK>, i.e. index 0
  return 0;
} // SearchWord

CWordIndex CVocabulary::AddWord(string word)
{
  return AddWord((CString)word.c_str());
}



const bool CVocabulary::IsInRange( CWordIndex WordIndex ) const
{
  if ((WordIndex>=m_VocItems.MinIndex()) && (WordIndex<=m_VocItems.MaxIndex())){
    return true;
  }else{
    return false;
  }
}


/**
   @brief Add a word to the vocabulary.

   This method does NOT check if the word already exists.

   \todo What happens if a word is inserted twice? Especially with that
   implementation of the hashmap comparing char pointers ...
 */
CWordIndex CVocabulary::AddWord( const CString & word )
{

  // okay, don't insert <UNK>
  if ( word == CString( "<UNK>" ) ) {
    return 0;
  }

  CVocabularyItem* pVocItem = new CVocabularyItem(word);
  if (! pVocItem ) {
    Trace( 0, "CVocabulary::AddWord(): Out of memory while reading vocabulary item" );
    exit( -1 );
  }
	
  // append to list
  int currentLastIndex = m_VocItems.MaxIndex()+1;
  m_VocItems[ currentLastIndex ] = pVocItem;
  
  // insert into map for quick string to index conversion
  //char *NewStr = strdup( str );
  m_VocIndices[pVocItem->PointerToWord()]= currentLastIndex;

  //  CString wtemp2 = word.c_str();
  // CWordIndex wi = GetIndex( wtemp2 );	
  // if ( wi == 0 ) {
  //  cout << "ERROR: " << wtemp2 << " " << wi << " not found" << endl << flush;
  //  exit(-1);
  //}else{
  //  cout<<"WRecovered:"<<wi<<endl;
  //}

  return currentLastIndex;
}

// convert string into CWordIndices
/*
BOOL 
CVocabulary::ConvertToIndices( char * const str, CArray<CWordIndex> *pSent, int *len)
{
  char *help, *word;
  uint  index;
  BOOL  valid=TRUE;
  
  help = str;

  int lengthOfString = strlen( str );

  EXIT_IF (!(help = new char[ lengthOfString+1 ]),
	   "\n could not allocate memory", -1);

  strcpy( help, str );

  word = strtok(help, " ");
  for (index=0; valid && (word != 0); index++)
    {
      (*pSent)[ index ] = SearchWord( word );
      if ((*pSent)[ index ] == EOT)
	{
	  cout << "\n CVocabulary::ConvertToIndices: unknown word: " << word << flush;
	  valid = FALSE;
	}
      else
	word = strtok (NULL, " ");
    }

  *len = index;

  free( help );

  return valid;  
}
*/



// ********************************************************************
// Display-Methods

void
CVocabulary::Display( BOOL Interactive ) const
{
  CWordIndex i;

  cout << "Vocabulary:\n";
  cout << " Index Word\n";
  for( i = m_VocItems.MinIndex(); i <= m_VocItems.MaxIndex(); i++ ) {
      cout.width(6);
      cout << i;
      cout << " |" << m_VocItems[ i ]->m_Word << "|" << endl;
      //      DisplayWord( cont );
      //cout << "\n";
      if( !(i % 20) && Interactive )
	cin.get();
    }
} // DisplayVocabulary



void
CVocabulary::DisplayItem( CWordIndex WordIndex ) const
{
  m_VocItems[ WordIndex ]->Display();
} // DisplayItem



void
CVocabulary::WriteWord( CWordIndex WordIndex, ostream &OutStream ) const
{
  OutStream << m_VocItems[ WordIndex ]->m_Word;
} // WriteWord



void
CVocabulary::DisplayWord( CWordIndex WordIndex ) const
{
  WriteWord( WordIndex, cout );
} // DisplayWord



void
CVocabulary::WriteSentence( const CWordIndex *WordIndices, uint Length,
			    ostream &OutStream ) const
{
  uint i = 0;

  // first word
  if ( Length > 0 ) {
    WriteWord( WordIndices[ i ], OutStream ); 
    
    // remaining words seperated by blanks
    for ( i = 1; i < Length; i++ ) {
      OutStream << " ";
      WriteWord( WordIndices[ i ], OutStream ); 
    }
  }
}



void
CVocabulary::DisplaySentence( const CWordIndex *WordIndices, uint Length ) const
{
  WriteSentence( WordIndices, Length, cout );
}




void
CVocabulary::LoadContentWords( const char *filename )
{
  assert( 0 );
/*
  char str[200];
  CWordIndex Index;

  ifstream fp( filename );

  EXIT_IF ( (!fp), "Content Word file could not be opened:" << filename, -1);
  
  fp.getline( str, 200 ); // Kommentarzeile

  fp.getline( str, 200 );
  while ( !fp.eof() ) {
    Index = SearchWord( str );
    //cout << str << " " << Index << endl << flush;
    if (Index > 0)
      m_VocItems[ Index ]->IsContentWord( TRUE );

    fp.getline( str, 200 );
  }
*/
}


void
CVocabulary::UpdateMultiWords()
{
  int i;
  int l;
  CWordIndex WordIndex;
  CString MultiWord, Word;

  for ( i = m_VocItems.MinIndex(); i <= m_VocItems.MaxIndex(); i++ ) {
    if ( m_VocItems[ i ]->m_IsMultiWord ) {
      MultiWord = m_VocItems[ i ]->m_Word;
      Trace( 5, "MultiWord " << MultiWord );

      l = MultiWord.Pos( "_" );
      while ( l >= 0 ) {
	Word = MultiWord.SubString( 0, l );
	WordIndex = GetIndex( Word );
	Trace0( 5,  "word: " << Word << " Index " << WordIndex << " " );
	//if ( WordIndex >= 0 )
	//  DisplayWord( WordIndex );
	MultiWord = MultiWord.SubString( l + 1, 1000 );
	//cout << "tail: " << String << endl;
	l = MultiWord.Pos( "_" );
	//cout << l << endl;
      }
      WordIndex = GetIndex( Word );
      Trace( 5,  "word: " << Word << " Index " << WordIndex << " " );
      //if ( WordIndex >= 0 )
      //DisplayWord( WordIndex );
    }
  }

}


