#ifndef Vocabulary_hh
#define Vocabulary_hh

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


// CVocabulary
// Base class for vocabulary handling
//

// CVS file ID:
// static char rcsid_Vocabulary_hh[] = "$Id: Vocabulary.hh 1742 2006-06-27 02:28:05Z vogel $"

#include <iostream>
using std::istream;
using std::ostream;

#include <Types.hh>
#include <Array.hh>
#include <String.hh>

#ifdef HAVE_EXT_HASH_MAP
#include <ext/hash_map>
// Hey, what's that? seems like overkill to me. --db
namespace sgi {
  using __gnu_cxx::hash;
  using __gnu_cxx::hash_map;
}
#define SGI_HASH_NAMESPACE __gnu_cxx
#else
#include <hash_map>

namespace sgi {
     using std::hash;
     using std::hash_map;
}
#define SGI_HASH_NAMESPACE std
#endif

struct eqstr
{
  bool operator()(const char* s1, const char* s2) const
  {
    return strcmp(s1, s2) == 0;
  }
};

typedef int CWordIndex;

class CVocabulary;
class CVocabularyItem
{
public:
  CVocabularyItem();
  CVocabularyItem( const CString &Word );
  ~CVocabularyItem();

  inline const CString & Word() const;
  inline void Word( const CString & Word);

  inline const char * PointerToWord() const;

  inline bool IsContentWord() const;
  inline void IsContentWord( bool Value );

  inline bool IsMultiWord() const;

  inline uint Length() const;

  inline CWordIndex FirstWord();
  inline CWordIndex LastWord();

  inline void Display();


protected:

  void SplitWords();

  CString m_Word;
  bool    m_IsContentWord;
  bool    m_IsMultiWord;

  CArray< CWordIndex > *m_pWordIndices;

  friend class CVocabulary;
};


/**
   @brief Indicate if the word index of EOS is static.

   In older versions of CVocabulary, EOS used to be the last word in the vocabulary.
   However, this made it hard to add new words later. Now EOS, the word index of
   the end-of-sentence symbol, is a static constant member of CVocabulary (with value 1)
   and hence the same for all vocabularies.

   This constant is supposed to indicate only whether or not the CVocabulary code uses
   a static member for EOS, it is not supposed to be setable by the user:
   Dynamic EOS is DEPRECATED, the dynamic EOS code is gone, so this will stay
   defined in the future.
 */
#define VOCAB_STATIC_EOS 1

class CVocabulary
{
public:
  CVocabulary();
  ~CVocabulary();

  // allocate memory for MaxNumOfVocItems vocabulary items
  // does not affect the size of the vocabulary, but allows faster loading
  bool Allocate( int MaxNumOfVocItems );

  // remove all entries in the vocabulary
  void Clear();

  // load vocabulary with/without comment lines/<UNK>/</s> 
  bool Load( const char *FileName );
  bool Load( istream &InStream ); 

  // store vocabulary with/without <UNK>/</s>
  // no comment line is written
  bool Store( const char *FileName, bool WithSGML = false ) const;
  bool Store( ostream &OutStream, bool WithSGML = false ) const;

  void LoadContentWords( const char *filename );

  inline CWordIndex First() const;
  inline CWordIndex Last() const;

  // number of stored vocabulary items (excluding <UNK> or </s>)
  inline int Size() const;

  inline const CVocabularyItem & GetItem( CWordIndex WordIndex ) const;
  inline const CString &         GetWord( CWordIndex WordIndex ) const;

  inline bool IsContentWord( CWordIndex WordIndex ) const;
  inline bool IsMultiWord( CWordIndex WordIndex ) const;

  // Get the word ptr back if the index is valid, else returns null, which is 
  // something that GetWord() cannot indicate
  const CString * GetWordPtr( CWordIndex WordIndex ) const;

  // return true if this word index you are looking for is within the range of this vocabulary
  const bool IsInRange( CWordIndex WordIndex ) const;

  // for multi-word items, get index of first and last word
  inline CWordIndex FirstWord( CWordIndex WordIndex ) const;
  inline CWordIndex LastWord( CWordIndex WordIndex ) const;

  // return index of Word if in vocabulary, and EOT if not found
  CWordIndex GetIndex( const CString & Word ) const;

  // return index of Word if in vocabulary, and EOT if not found
  CWordIndex GetIndex( const char * Word ) const;
  
  // search for the nearest vocabulary item <= Word
  CWordIndex SearchWord( const char * Word ) const;

  // add a word to the vocabulary
  CWordIndex AddWord( string Word );
  CWordIndex AddWord( const CString & Word );

  // get index, if word is new, add it
  CWordIndex GetIndexAddWord( const CString & Word );

  // convert string into CWordIndices
  //BOOL ConvertToIndices( char * const str, CArray<CWordIndex> *pSent, int *lenn);

  // display methods
  void Display( BOOL Interactive = FALSE ) const;
  void DisplayItem( CWordIndex WordIndex ) const;
  void DisplayWord( CWordIndex WordIndex ) const;
  void DisplaySentence( const CWordIndex *WordIndices, uint Length ) const;

  void WriteWord( CWordIndex WordIndex, ostream &OutStream ) const;
  void WriteSentence( const CWordIndex *WordIndices, uint Length, 
                      ostream &OutStream ) const;

  // special vocabulary entries
  // EOT = -1, UNK = 0, EMPTYWORD = 1, EOS = 2
  static CWordIndex EOT, UNK, EMPTYWORD, EOS;

  CString GetStringForUNK();
  CString GetStringForEOS();

  //! set the string used for the empty word, default is "EMPTYWORD"
  void SetStringForEmptyWord( CString StringForEmptyWord );

  //! get the string used for the empty word
  CString GetStringForEmptyWord();


protected:
  void UpdateMultiWords();

  CString m_StringForEmptyWord;

  CArray< CVocabularyItem * > m_VocItems;
  sgi::hash_map< const char *, CWordIndex, sgi::hash< const char* >, eqstr > m_VocIndices;
};



#include <Vocabulary_inl.hh>

#endif
