#include <string.h>
#include <stdlib.h>
// Header file for CTranslationMemory

// Translation memory manages transducers, i.e. a list of transducers
// which can be on different levels, to allow for hierarchical transducers


// File ID:
// static char ID_TransMem_cc[] = "$Id: TransMem.cc 1759 2006-07-02 18:12:58Z vogel $"

// Programming: S. Vogel
// Last Update: July 1, 2006


// History: see header file


/* ToDo:
- Need to check that calculation of neg log probs when loading the transducers works
  under all conditions.

*/



#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#define NO_TRACE 1
#include <Trace.hh>

#include <TransMem.hh>

using namespace std;

CTranslationMemory::CTranslationMemory()
{
  m_Direction = 0;
  m_pLattice = NULL;
  m_pLM = NULL;
  m_LatticeStatistics = 0;
  m_DisplayLattice = 0;
  m_TransducerMatchingMode = 0;
  m_UseProsody = 0;
  m_pLatticeEdgeTransPrototype = new CLatticeEdgeTrans;
  m_BeamFactor = 0.0;
  m_pScalingFactors = NULL;
}


CTranslationMemory::~CTranslationMemory()
{
  //delete m_pLattice;

  // leave freeing memory to OS, deleting SRI LM takes very, very long
  //delete m_pLM;

  int i;
  for ( i = m_Transducers.MinIndex(); i <= m_Transducers.MaxIndex(); i++ ) {
    delete m_Transducers[ i ];
  }

  for ( i = m_TransducersSource.MinIndex(); 
	i <= m_TransducersSource.MaxIndex(); i++ ) {
    delete m_TransducersSource[ i ];
  }

  for ( i = m_TransducersTarget.MinIndex(); 
	i <= m_TransducersTarget.MaxIndex(); i++ ) {
    delete m_TransducersTarget[ i ];
  }

  delete m_pLatticeEdgeTransPrototype;
}



void
CTranslationMemory::Init( const char *ParamFile, CVocabulary *pVocabulary )
{
  InitParams( ParamFile );

  m_pVocabulary = pVocabulary;
  
  //cerr<< "Loading Transducers ..." <<endl;

  ifstream fp( m_TransducerList );
  char str [ 1000 ];
  str[ 0 ] = 0;
  //char *strptr = str;
  int index;

  if ( !fp.good() ) {
      cerr << "CTranslationMemory::Init(): File: " << m_TransducerList;
      const char *source = m_Param.getSourceFromKey("TransducerList");
      if (source != NULL) {
	  // wrong parameter comes from param file
	  cerr << ", referred to by param file " << source << ",";
      } // endif: wrong parameter came from param file
      cerr << " could not be opened for reading" << endl
	   << "CTranslationMemory::Init(): Exiting." << endl << flush;
      exit( -1 );
  } // endif: Transducer List not loaded
  
  Trace( 1, "Load list of transducer parameter files" << m_TransducerList );
  //char *buffer;
  //int I = 1;

  string Source, Target;

  for ( index = 0; !fp.eof(); index++, fp.getline( str, 1000 ) ) {
    Trace( 1, "Transducer parameter file: " << str );
    //TraceIf( ( index % 1000 == 0 ), 1, index );

    if ( (str[ 0 ] != ';')  && strlen( str ) > 0 ) {
      Trace( 1, "Transducer param file: " << str );
      m_TransducerParamFiles[ index ] = CString( str );
    }
  }


  Trace( 1, "Loading transducers" );

  for ( int i = m_TransducerParamFiles.MinIndex();
        i <= m_TransducerParamFiles.MaxIndex(); 
        i++ ) {
    
    if ( InitTransducer( m_TransducerParamFiles[ i ], pVocabulary ) == NULL ) {
      cerr << "CTranslationMemory::Init(): Parameter file for transducer "
           << m_TransducerParamFiles[ i ] << ", referred to by transducer list "
           << m_TransducerList << "could not be loaded." << endl
           << "CTranslationMemory::Init(): Exiting.\n" << flush;
      exit(-1);
    } // endif: error occured loading transducer
  }    
  
  m_Transducers.MakeStatic();


  // load language models
  m_pLM = NULL;

  /*
    sv 2006-05-25: LM no longer used in here

  cerr << "Loading Language Model..." << endl;
  if ( m_WordLM ) {
    Trace( 0, "Load LM" << " "<< m_LMName<< "-" << m_WordLM );
  
    // arv : to allow loading of several lm's
    //if (strcmp(m_LMName, "NistLM")==0){
    //  Trace( 1, "Loading NistLM");
    //  m_pLM = new NistLM( m_WordLM);
    //}else 
    if (strcmp(m_LMName, "LMWordSA")==0){
      Trace( 0, "Loading LMWordSA");
      m_pLM = new CLMwordSA( m_WordLM );
    }
    else if (strcmp(m_LMName, "SriLM")==0){
      Trace( 0, "Loading SriLM");
      m_pLM = new SriLM( m_WordLM );
      
    }
    if (strcmp(m_LMName, "LMWordSA")==0){
      if ( m_LMScale == 0.0 ) {
		((CLMwordSA *)m_pLM)->setgrammity( 0 );
      } else {
		((CLMwordSA *)m_pLM)->setgrammity( m_HistoryLength + 1 );
      }
    }

    Trace( 1, "History Length = " << m_HistoryLength << 
	   "  Grammity = " << m_HistoryLength + 1 );
    Trace( 1, "Load LM done" );
  }

  */

}



void
CTranslationMemory::Init( CVocabulary *pVocabulary )
{
  m_pVocabulary = pVocabulary;
}



void 
CTranslationMemory::SetBeamFactor( float BeamFactor )
{
  m_BeamFactor = BeamFactor;
}



void 
CTranslationMemory::InitParams( const char *ParamFileName )
{

  Trace( 1, "CTranslationMemory::InitParams with Param-File " << ParamFileName );
  
  m_Param.AddParamDef("TransducerList", 'T' , &m_TransducerList, NULL , 
                    "name of file with list of transducer parameter files", 
                    True);
  // arv : to allow picking a LM, with default as LMWordSA
  m_Param.AddParamDef("LM", 'L', &m_LMName, "LMWordSA",
		    "The kind of language model that you will use",
		    True);

  m_Param.AddParamDef("WordLM", 'w', &m_WordLM, NULL,
		    "Corpus filename for generation of word based LM", true);

  m_Param.AddParamDef("HistoryLength", 'H', &m_HistoryLength, 3,
		    "length of history in suffix-array LM",
		    True);
  
  m_Param.AddParamDef("BigramLM", '2', &m_BigramLMParamFile, NULL,
		    "Parameter file for standard bigram LM", true);
  
  m_Param.AddParamDef("TrigramLM", '3', &m_TrigramLMParamFile, NULL,
		    "Parameter file for standard trigram LM", true);

  //Added by Joy, joy@cs.cmu.edu
  //Mar 29, 2003
  //to handle applyTransducerSimpe
  m_Param.AddParamDef("ApplyTdSimple", 's', &m_TransducerMatchingMode, 0,
                      "Apply transducers to the lattice in a non-hierarchical mode", true);


  // DEPRECATED
  m_VocFile = NULL;
  m_Param.AddParamDef("Vocabulary", 'V' , &m_VocFile, NULL , 
                    "DEPRECATED", 
                    True);

  m_Param.AddParamDef("LMScale", 'a', &m_LMScale, 0.0,
		    "Factor between acoustic score and LM score",
		    True);

  m_Param.AddParamDef("TMScale", 'b', &m_TMScale, 0.0,
		    "Factor between acoustic score and TM score",
		    True);

  m_Param.AddParamDef("ERScale", 'c', &m_ERScale, 0.0,
		    "Factor between acoustic score and error score",
		    True);

  m_Param.AddParamDef("CCScale", 'd', &m_CCScale, 0.0,
		    "Factor between acoustic score and concatenation score",
		    True);

  m_Param.AddParamDef("NBest", 'N', &m_NumOfNBest, 1,
		    "Display n best translations",
		    True);


  int Result = m_Param.read_parameter_file( (char *)ParamFileName );
  if ( Result < 0 ) {
    cerr << "\n\n ... exiting!!\n";
    exit(-1);
  }

  // check for deprecated parameters
  if ( m_VocFile != NULL ) {
      cerr << "CTranslationMemory::InitParams(): The vocabulary parameter is DEPRECATED. ";
      const char *fileName = m_Param.getSourceFromKey("Vocabulary");
      if (fileName == NULL) {
	  // does not come from any param file
	  cerr << "Maybe you should check your command-line options." << endl;
      } else {
	  cerr << "Review your parameter file " << fileName << endl;
      } // endif: can we compute the name of the errorous parameter file?
      cerr << "CTranslationMemory::InitParams(): Exiting." << endl << flush;
      exit(-1);
  } // endif: Deprecated parameters
}



void
CTranslationMemory::SetMatchingMode( int MatchingMode )
{
  //cout << "CTranslationMemory::SetMatchingMode " << MatchingMode << endl;
  m_TransducerMatchingMode = MatchingMode;
}



CLM *
CTranslationMemory::GetLM()
{
  return m_pLM;
}



CLatticeLM *
CTranslationMemory::Lattice()
{
  return m_pLattice;
}



void 
CTranslationMemory::SetDirection( int Direction )
{
  assert( (m_Direction == 0) || (m_Direction == 1) );
  m_Direction = Direction;
}



int
CTranslationMemory::FirstLevel()
{
  return m_Transducers.MinIndex();
}



int
CTranslationMemory::LastLevel()
{
  return m_Transducers.MaxIndex();
}



CString
CTranslationMemory::TranslateSentence( const CSentence *pSentence )
{

  // Convert sentence into lattice
  delete m_pLattice;
  CLatticeLM *pLattice = new CLatticeLM( m_pLatticeEdgeTransPrototype, m_pLM );

  CArray< CString > Words;
  ConvertToLatticeRepresentation( pSentence, m_pVocabulary, Words );
 

  pLattice->ConvertFromText( Words );
  if ( m_DisplayLattice ) {
	pLattice->Display();
	cout << endl;
  }
  

  CString Translation = TranslateLattice( pLattice );

  if ( m_DisplayLattice ) {
	pLattice->Display();
	cout << endl;
  }
  
  return Translation;
}



CString
CTranslationMemory::TranslateLattice( CLatticeLM *pLattice )
{

  Trace( 1, "CTranslationMemory::TranslateLattice" );

  //  delete m_pLattice;
  //pLattice->Display();
  m_pLattice = pLattice;
  //m_pLattice->Display();

  BuildTranslationLattice( pLattice );

  return ApplyLanguageModel( pLattice );
}



void
CTranslationMemory::BuildTranslationLattice( CLatticeLM *pLattice )
{

  Trace( 1, "CTranslationMemory::BuildTranslationLattice" );

  m_pLattice = pLattice;

  const CArray< CLatticeNode * > &NodeList = m_pLattice->Nodes();
  Trace( 2, "Nodes ranging from: " << NodeList.MinIndex() << " to" << NodeList.MaxIndex() <<" in this lattice" );

  // give lattice to PhraseWhg
  m_LatticeSearch.SetLattice( m_pLattice );
  
  // give source vocabulary to PhraseWhg
  m_LatticeSearch.SetVocabulary( m_pVocabulary );
  
  m_LatticeSearch.SetScalingFactors( m_pScalingFactors );
  
  //commented by Joy, joy@cs.cmu.edu
  //Mar 29, 2003
  //this is a bug, since the hyp added by previous td application will remain in the lattice
  //and will affect the following td applications. because the state for each td are unique
  //state 4 in td1 can be different as state4 in td3
  // init search by setting an empty search hypothesis for each lattice node
  //m_LatticeSearch.InitSearch();

  
  Trace( 1, "Apply transducers" );

  //cout << "Number of Search Hypotheses = " 
  //   << CLatticeSearchHyp::s_NumberOfObjects << endl;
  //cout << "Number of Edges = " 
  //     << CLatticeEdgeTrans::s_NumberOfObjects << endl;
  
  assert( (m_Direction == 0) || (m_Direction == 1) );

  for ( int i = m_Transducers.MinIndex(); i <= m_Transducers.MaxIndex(); i++ ) {

    //added by Joy, joy@cs.cmu.edu
    //Mar 29, 2003
    //to fix the bug
    //moving the LatticeSearch.initSearch(), so that each time the lattic will start as a new empty one
    //hyp from previous td applications will not affect the following td applications
    m_LatticeSearch.InitSearch();

    Trace ( 0, "Transducer Number " << i );
    if ( m_Direction == 0 ) {
      Trace( 1, "Apply source transducer" );
      //m_TransducersSource[ i ]->Statistics();
      //m_TransducersSource[ i ]->DisplayTransitionTable();
      //m_TransducersSource[ i ]->DisplayEmissionTable();
      ApplyTransducer( m_TransducersSource[ i ] );
    } else {
      Trace( 1, "Apply target transducer" );
      //m_TransducersTarget[ i ]->Statistics();
      //m_TransducersTarget[ i ]->DisplayTransitionTable();
      //m_TransducersTarget[ i ]->DisplayEmissionTable();
      ApplyTransducer( m_TransducersTarget[ i ] );
    }
  }    

  
  if ( m_LatticeStatistics )
    DisplayLatticeStatistics( pLattice );

  // now we do some reordering
  //((CLatticeLM *)pLattice)->LocalReordering();
}


  
CString
CTranslationMemory::ApplyLanguageModel( CLatticeLM *pLattice )
{

  Trace( 1, "CTranslationMemory::ApplyLanguageModel" );

  //((CLatticeLM *)pLattice)->SetLMScale( LMScale );
  ((CLatticeLM *)pLattice)->SetScalingFactors( 0.0, m_LMScale, m_TMScale, m_ERScale, m_CCScale );
  ((CLatticeLM *)pLattice)->SetBeamFactor( m_BeamFactor );

  //Rescore( pLattice, m_pLM );

  Trace( 1, "Find n-best pathes in tranlation lattice using LM" );

  int NumPathes;

  multimap< float, CArray< CLatticeEdge* > > BestPathes;
  NumPathes = ((CLatticeLM *)pLattice)->BestPathes( m_NumOfNBest, BestPathes );
  
  
  // TODO: This should be TRACE
  // so far, modified by Joy, joy@cs.cmu.edu
  // using tracelevel to turn this off for the interactive mode where tracelevel is -1
  if(TraceLevel >= 0){
    cout << "PartialPath: ";
    DisplayPath( BestPathes.begin()->second );
    cout << endl;
  }  

  return Translation( BestPathes.begin()->second );
}



// -----------------------------------------------------------------------------------
CTransducer *
CTranslationMemory::InitTransducer( CString &TransducerParamFile, CVocabulary *pVocabulary )
{
  if ( TransducerParamFile.GetLength() == 0 ) {
      cerr << "CTranslationMemory::InitTransducer(): No Parameter File given" << endl;
      return NULL;
  }

  CTransducer *pTransducer = new CTransducer;
  if ( !pTransducer->Init( pVocabulary, pVocabulary ) ) {
      // error occured
      cerr << "CTranslationMemory::InitTransducer(): Could not allocate transducer." << endl;
      delete pTransducer;
      return NULL;
  } // endif: error init transducer
  if ( !pTransducer->LoadParam( TransducerParamFile ) ) {
      // error occured
      cerr << "CTranslationMemory::InitTransducer(): Could not load transducer parameter file "
	   << TransducerParamFile << endl;
      delete pTransducer;
      return NULL;
  } // endif: error load transducer param file
  m_Transducers.Append( pTransducer );

  // set level of transducer
  int L = m_Transducers.MaxIndex();
  pTransducer->SetLevel( L + 1 ); 
  // append sets first entry at index 0
  // but level zero is reseved for lexicon probabilities in alignment


  //arv added to indicate that by default, value in phrase table are probs
  pTransducer->SetValuesInFileAreProbabilities();


  // access objects
  CTransducerAccess *pTransducerAccess = new CTransducerAccess( pTransducer, 0 );
  m_TransducersSource.Append( pTransducerAccess );
  pTransducerAccess = new CTransducerAccess( pTransducer, 1 );
  m_TransducersTarget.Append( pTransducerAccess );

  Trace( 1, "Transducer " << pTransducer->Name() << " initialized "
	 << "has index " << m_Transducers.MaxIndex() );

  return pTransducer;
}



// -----------------------------------------------------------------------------------
CTransducer *
CTranslationMemory::LoadTransducer( const char *pFileName )
{
  //cout << "Directly Loading Transducer ..." << pFileName << endl;

  if ( !pFileName ) {
      cerr << "CTranslationMemory::LoadTransducer(): No File given" << endl;
      return NULL;
  }

  CTransducer *pTransducer = new CTransducer;

  if ( !pTransducer->Init( m_pVocabulary, m_pVocabulary ) ) {
      // error occured
      cerr << "CTranslationMemory::LoadTransducer(): Could not initialize transducer." << endl;
      delete pTransducer;
      return NULL;
  } // endif: error init transducer

  if ( !pTransducer->Load( pFileName ) ) {
      // error occured
      cerr << "CTranslationMemory::LoadTransducer(): Could not load transducer "
	   << pFileName << endl;
      delete pTransducer;
      return NULL;
  } // endif: error load transducer param file
  m_Transducers.Append( pTransducer );

  // set level of transducer
  int L = m_Transducers.MaxIndex();
  pTransducer->SetLevel( L + 1 ); 
  // append sets first entry at index 0
  // but level zero is reseved for lexicon probabilities in alignment


  // access objects
  CTransducerAccess *pTransducerAccess = new CTransducerAccess( pTransducer, 0 );
  m_TransducersSource.Append( pTransducerAccess );
  pTransducerAccess = new CTransducerAccess( pTransducer, 1 );
  m_TransducersTarget.Append( pTransducerAccess );

  Trace( 1, "Transducer " << pTransducer->Name() << " initialized "
	 << "has index " << m_Transducers.MaxIndex() );

  return pTransducer;
}

void
CTranslationMemory::ClearTransducer()
{
  Trace( 1, "CTranslationMemory::ClearTransducer" );
  // clean m_Transducers
  for ( int t = m_Transducers.MinIndex(); t <= m_Transducers.MaxIndex(); t++ ) {
    delete m_Transducers[ t ];
    m_Transducers[ t ] = NULL;
  }
  m_Transducers.Clear();
  
  // clean m_TransducersSource
  for ( int t = m_TransducersSource.MinIndex(); t <= m_TransducersSource.MaxIndex(); t++ ) {
    delete m_TransducersSource[ t ];
    m_TransducersSource[ t ] = NULL;
  }
  m_TransducersSource.Clear();

  // clean m_TransducersTarget
  for ( int t = m_TransducersTarget.MinIndex(); t <= m_TransducersTarget.MaxIndex(); t++ ) {
    delete m_TransducersTarget[ t ];
    m_TransducersTarget[ t ] = NULL;
  }
  m_TransducersTarget.Clear();
  
}

void
CTranslationMemory::ApplyTransducer( CLatticeLM *pLattice, int Level )
{
  Trace( 1, "CTranslationMemory::ApplyTransducer" );

  assert( (m_Direction == 0) || (m_Direction == 1) );

  m_pLattice = pLattice;

  // give lattice to PhraseWhg
  m_LatticeSearch.SetLattice( pLattice );
  
  // give source vocabulary to PhraseWhg
  m_LatticeSearch.SetVocabulary( m_pVocabulary );
  
  // init search by setting an empty search hypothesis for each lattice node
  if ( Level == m_Transducers.MinIndex() ) {
    m_LatticeSearch.InitSearch();
  }

  Trace( 1, "Transducer Number " << Level );

  //cout << "Number of Search Hypotheses = " 
  //     << CLatticeSearchHyp::s_NumberOfObjects << endl;
  //cout << "Number of Edges = " 
  //     << CLatticeEdgeTrans::s_NumberOfObjects << endl;

  if ( m_Direction == 0 ) {
    Trace( 1, "Apply source transducer" );
    ApplyTransducer( m_TransducersSource[ Level ] );
  } else {
    Trace( 1, "Apply target transducer" );
    ApplyTransducer( m_TransducersTarget[ Level ] );
  }    

  if ( m_LatticeStatistics )
    DisplayLatticeStatistics( pLattice );

  //cout << "Number of Search Hypotheses = " 
  //     << CLatticeSearchHyp::s_NumberOfObjects << endl;
  //cout << "Number of Edges = " 
  //     << CLatticeEdgeTrans::s_NumberOfObjects << endl;
  //cout << "Number of BackPointers = " 
  //     << CBackPointerLM::s_NumberOfObjects << endl;

}



void
CTranslationMemory::ApplyTransducer( const CTransducerAccess *pTransducerAccess )
{
  //cout << "ApplyTransducer - Access " << endl;
  if ( pTransducerAccess == NULL ) {
    //cout << "TransducerAccess == NULL" << endl;
    return;
  }

  //assert( m_LatticeSearch );

  //const CPhraseTree & PhraseTree = pTransducer->GetPhraseTree();
  int Iteration = 1;
  int NumberOfFoundPhrases = 0;

  Trace( 1, "Apply Transducer " << pTransducerAccess->Transducer()->Name() );
  if ( pTransducerAccess->Transducer()->ShouldBeIterated() )
    Trace( 1, "Will be iterated" );

  //Trace( 3, "SetMaxCost: " << pTransducer->MaxCost() );
  //PhraseWhg.SetMaxCost( pTransducer->MaxCost() );
  
  Trace( 1, "SetMaxCostPerWord: " << pTransducerAccess->Transducer()->MaxCostPerWord() );
  m_LatticeSearch.SetMaxCostPerWord( pTransducerAccess->Transducer()->MaxCostPerWord() );
  
  do {
    Trace( 1, "Transducer " << pTransducerAccess->Transducer()->Name() 
	   << ": Iteration " << Iteration );
    Iteration++;

    //modified by Joy, joy@cs.cmu.edu
    //Mar 26, 2003
    //commented the regular ApplyTransducer routine
    //instead, using the simplified ApplyTransducer method which considers only the souce
    //cout << "Matching mode = " << m_TransducerMatchingMode << endl;
    if ( m_TransducerMatchingMode == 1  ) {
      cout << "Apply transducer in full mode" << endl;
      CString Translation = m_LatticeSearch.ApplyTransducer( pTransducerAccess );
    } else {
      cout << "Apply Transducer in simple mode" << endl;
      CString Translation = m_LatticeSearch.ApplyTransducerSimple( pTransducerAccess );
    }


    //pTransducerAccess->Transducer()->Report();

    //NumberOfFoundPhrases = m_PhraseWhg.FindAndInsertPhrases( &PhraseTree );
    Trace( 1, "Transducer " << pTransducerAccess->Transducer()->Name() );

#ifdef TRACE_ENABLED
    if ( TraceLevel > 0 ) {
      m_pLattice->Display();
      DisplayBestPath( m_pLattice );
      DisplayTargetString( m_pLattice );
    }
#endif

  } while ( ( pTransducerAccess->Transducer()->ShouldBeIterated() 
	      && (NumberOfFoundPhrases > 0) )
        && ( Iteration <= 1 ) );    // restrict to 1 (apd 12/19/06)
  
#ifdef TRACE_ENABLED
  if ( TraceLevel > 0 ) {
    cout << "Best Path: ";
    DisplayBestPath( m_pLattice );
  }
#endif
  
}



CString
CTranslationMemory::GetBoundary( CLatticeEdgeTrans *pEdge )
{
  float Boundary = pEdge->Prosody()->GetBoundary2();
  float Mode = pEdge->Prosody()->GetMode2();

  if ( Boundary > 0.7 ) {
    if ( Mode > 0.7 ) {
      return "?";
    } else {
      return ".";
    }
  } else if ( Boundary > 0.3 ) {
    return ",";
  }

  return "";

}



CTransducer *
CTranslationMemory::Transducer( int Number )
{
  return m_Transducers[ Number ];
}



// --------------------------------------------------------
// Display Methods
// --------------------------------------------------------
void
CTranslationMemory::DisplayBestPath( const CLattice *pLattice )
{
  Trace( 1, "CTranslationMemory::DisplayBestPath" );

  CArray< CLatticeEdge * > Edges;
  //CString SourceString, TargetString;
  CWordInfo *pWordInfo;
  CString Boundary;
  
  //Edges.RemoveAll();
  //pLattice->ShortestPath( Edges );
  pLattice->BestPath( Edges );
  
  Trace0( 0, "BEST PATH: " );
  // we use best path to generate sentence
  for ( int e = Edges.MinIndex(); e <= Edges.MaxIndex(); e++ ) {
    if ( e > Edges.MinIndex() )
      cout << " ";
    
    cout << ((CLatticeEdgeTrans *)Edges[ e ])->Word();
    pWordInfo = ((CLatticeEdgeTrans *)Edges[ e ])->WordInfo();
    if ( pWordInfo->HasSourceString() ) {
      cout << " " << *pWordInfo;
    }
        
    // Boundary
    Boundary = GetBoundary( (CLatticeEdgeTrans *)Edges[ e ] );
    if ( Boundary.GetLength() > 0 ) {
      cout << " " << Boundary;
    }
  }
  cout << endl << endl << flush;
  
}



void
CTranslationMemory::DisplayTargetString( const CLattice *pLattice )
{
  CArray< CLatticeEdge * > Edges;
  //CString SourceString, TargetString;
  CWordInfo *pWordInfo;
  CString Boundary;
  
  //Edges.RemoveAll();
  //pLattice->ShortestPath( Edges );
  pLattice->BestPath( Edges );
  
  // we use shortest path to generate sentence
  for ( int e = Edges.MinIndex(); e <= Edges.MaxIndex(); e++ ) {
    if ( e > Edges.MinIndex() )
      cout << " ";
    
    //cout << ((CLatticeEdgeTrans *)Edges[ e ])->Word();
    pWordInfo = ((CLatticeEdgeTrans *)Edges[ e ])->WordInfo();
    if ( pWordInfo->HasTargetString() ) {
      cout << " " << pWordInfo->TargetString();
    } else {
      cout << ((CLatticeEdgeTrans *)Edges[ e ])->Word();
    }
    
    // Boundary
    Boundary = GetBoundary( (CLatticeEdgeTrans *)Edges[ e ] );
    if ( Boundary.GetLength() > 0 ) {
      cout << " " << Boundary;
    }
  }
  cout << endl << endl << flush;
  
}



void
CTranslationMemory::DisplayPath( CArray< CLatticeEdge* > &Path )
{
  CString String, SourceString, TargetString;
  CString Word;
  //CLatticeEdge * pEdge;
  CLatticeEdgeTrans * pEdgeTrans;
  CWordInfo * pWordInfo;

  //BOOL AddLabelInfo = TRUE;

  Trace( 1, "Display Path" );


  for ( int i = Path.MinIndex();
	i<= Path.MaxIndex();
	i++ ) {
    if ( i > Path.MinIndex() )
      cout << " ";
 
    pEdgeTrans = (CLatticeEdgeTrans *)(Path[ i ]);
    //Word = pEdgeTrans->Word();
    cout << pEdgeTrans->Word();

    //    if ( Word[0] != '#' ) {
    //String += Word;
    // String += CString( " " );
      
    pWordInfo = pEdgeTrans->WordInfo();
    if ( pWordInfo->HasSourceString() ) {
      cout << " " << *pWordInfo;
    }

    /*
      if ( AddLabelInfo ) {
      // get additional information
      SourceString = pEdgeTrans->StatTransInfo()->SourceString();
      if ( SourceString.GetLength() > 0 ) {
      String += CString( "{ " );
      String += SourceString;
      }
      
      TargetString = pEdgeTrans->StatTransInfo()->TargetString();
      if ( TargetString.GetLength() > 0 ) {
      String += " # ";
      String += TargetString;
      
      float Score = pEdgeTrans->StatTransInfo()->Score();
      String += " # ";
      String += CString( Score );
      
      String += CString( " } " );
      }
      
      }
    */
    /*
      // get sentence boundary and mode
      Boundary = pEdgeTrans->Prosody()->GetBoundary2();
      Mode = pEdgeTrans->Prosody()->GetMode();
      
      if ( Boundary > 0.8 ) {
      if ( Mode > 0.8 ) {
      String += CString( "? " );
      } else {
      String += CString( ". " );
      }
      } else if ( Boundary > 0.4 ) {
      String += CString( ", " );
      }
    */
    
    // Boundary
    //Boundary = GetBoundary( (CLatticeEdgeTrans *)Edges[ e ] );
    //if ( Boundary.GetLength() > 0 ) {
    //  cout << " " << Boundary;
    //  }
  }
  // }
  //cout << " Hypothesis from WHG: " << String << endl;
  //cout << String << endl;
  
}



void
CTranslationMemory::WritePath( ofstream &Stream, CArray< CLatticeEdge* > &Path )
{
  //CString String, SourceString, TargetString;
  //CString Word;
  CLatticeEdgeTrans * pEdgeTrans;
  CWordInfo * pWordInfo;

  Trace( 1, "Write Path to stream" );

  for ( int i = Path.MinIndex(); i<= Path.MaxIndex(); i++ ) {
    if ( i > Path.MinIndex() )
      Stream << " ";
 
    pEdgeTrans = (CLatticeEdgeTrans *)(Path[ i ]);
    Stream << pEdgeTrans->Word();

    pWordInfo = pEdgeTrans->WordInfo();
    if ( pWordInfo->HasSourceString() ) {
      Stream << " " << *pWordInfo;
    }

  }
}



void
CTranslationMemory::DisplayTransducer( int Number )
{
  if ( Number == -1 ) {
    Trace( 1, "Display transducer counts for all transducers" );
    for ( int i = m_Transducers.MinIndex(); i <= m_Transducers.MaxIndex(); i++ ) {
      //if ( m_Transducers[ Number ] )
      m_Transducers[ i ]->Display();
    }
  } else {
    //    if ( m_Transducers[ Number ] )
    m_Transducers[ Number ]->Display();
  }
}



void
CTranslationMemory::RenormalizeTransducer( int Number )
{
  Trace( 1, "CTranslationMemory::RenormalizeTransducer" );

  if ( Number == -1 ) {
    Trace( 1, "Renormalize transducer counts for all transducers" );
    for ( int i = m_Transducers.MinIndex(); i <= m_Transducers.MaxIndex(); i++ ) {
      //if ( m_Transducers[ Number ] )
      m_Transducers[ i ]->Renormalize();
    }
  } else {
    //    if ( m_Transducers[ Number ] )
    m_Transducers[ Number ]->Renormalize();
  }
}



void 
CTranslationMemory::ClearTransducerCounts( int Number )
{
  if ( Number == -1 ) {
    Trace( 1, "Clear transducer counts for all transducers" );
    for ( int i = m_Transducers.MinIndex(); i <= m_Transducers.MaxIndex(); i++ ) {
      //      if ( m_Transducers[ i ] )
	m_Transducers[ i ]->ClearCounts();
    }
  } else {
    //    if ( m_Transducers[ Number ] )
      m_Transducers[ Number ]->ClearCounts();
  }

}



void 
CTranslationMemory::SetScalingFactors( float scalingFactor, int Number )
{
  if ( Number == -1 ) {
    Trace( 1, "Clear transducer counts for all transducers" );
    for ( int i = m_Transducers.MinIndex(); i <= m_Transducers.MaxIndex(); i++ ) {
      m_Transducers[ i ]->SetScalingFactors(scalingFactor);
    }

  } else {
    m_Transducers[ Number ]->SetScalingFactors(scalingFactor);
  }

}

//! propagates scaling factors to LatticeSearch
void
CTranslationMemory::SetScalingFactors( CNumericalArray< float >* pScalingFactors )
{
  if ( m_pScalingFactors != 0 && (m_pScalingFactors != pScalingFactors) ) {
    delete m_pScalingFactors;
  }
  
  m_pScalingFactors = pScalingFactors;
  m_LatticeSearch.SetScalingFactors( m_pScalingFactors );
}

void
CTranslationMemory::SetLatticePruning( float LatticePruneBeamSize, 
                                       float LatticePruneBeamSizeEpsilon,
                                       int LatticePruneMaxTranslations )
{
  m_LatticeSearch.SetLatticePruning( LatticePruneBeamSize, 
                                     LatticePruneBeamSizeEpsilon,
                                     LatticePruneMaxTranslations );
}


void 
CTranslationMemory::StoreTransducer( int Number )
{
  if ( Number == -1 ) {
    Trace( 1, "Store all transducers" );

    for ( int i = m_Transducers.MinIndex(); i <= m_Transducers.MaxIndex(); i++ ) {
      //      if ( m_Transducers[ i ] )
	if ( !m_Transducers[ i ]->Store() ) {
	    cerr << "CTranslationMemory::StoreTransducer: Warning: Could not store Transducer "
		 << m_Transducers[ i ]->Name() << endl << flush;
	} // endif: error
    }
  } else {
    //    if ( m_Transducers[ Number ] )
      if ( !m_Transducers[ Number ]->Store() ) {
	  cerr << "CTranslationMemory::StoreTransducer: Warning: Could not store Transducer "
	       << m_Transducers[ Number ]->Name() << endl << flush;
      } // endif: error
  }


}



CString
CTranslationMemory::TargetString( CArray< CLatticeEdge* > &Path )
{
  CString String, TargetString;
  CLatticeEdgeTrans * pEdgeTrans;
  //CWordInfo * pWordInfo;

  Trace( 1, "Target String" );

  for ( int i = Path.MinIndex();
	i<= Path.MaxIndex();
	i++ ) {
 
    pEdgeTrans = (CLatticeEdgeTrans *)(Path[ i ]);
    TargetString = pEdgeTrans->WordInfo()->TargetString();
    if ( TargetString.GetLength() > 0 ) {
      String += " ";
      String += TargetString;
    }

    /*
      // get sentence boundary and mode
      Boundary = pEdgeTrans->Prosody()->GetBoundary2();
      Mode = pEdgeTrans->Prosody()->GetMode();
      
      if ( Boundary > 0.8 ) {
      if ( Mode > 0.8 ) {
      String += CString( "? " );
      } else {
      String += CString( ". " );
      }
      } else if ( Boundary > 0.4 ) {
      String += CString( ", " );
      }
    */
    
    // Boundary
    //Boundary = GetBoundary( (CLatticeEdgeTrans *)Edges[ e ] );
    //if ( Boundary.GetLength() > 0 ) {
    //  cout << " " << Boundary;
    //  }
  }

  Trace( 2, "Target String " << String );

  return String;
}



CString
CTranslationMemory::Translation( CArray< CLatticeEdge* > &Path )
{
  CString String, SourceString, TargetString;
  CLatticeEdgeTrans * pEdgeTrans;
  //CWordInfo * pWordInfo;

  Trace( 1, "Translation" );

  for ( int i = Path.MinIndex();
	i<= Path.MaxIndex();
	i++ ) {
 
    pEdgeTrans = (CLatticeEdgeTrans *)(Path[ i ]);
    SourceString = pEdgeTrans->WordInfo()->SourceString();
    TargetString = pEdgeTrans->WordInfo()->TargetString();
    Trace( 1, "Source " << SourceString );
    Trace( 1, "Target " << TargetString );
    if ( TargetString.GetLength() > 0 ) {
      String += " ";
      String += TargetString;
      //    } else if ( SourceString.GetLength() > 0 ) {
      //String += " ";
      ///String += SourceString;
    } else {
      String += " ";
      String += pEdgeTrans->Word();
    }
      

    /*
      // get sentence boundary and mode
      Boundary = pEdgeTrans->Prosody()->GetBoundary2();
      Mode = pEdgeTrans->Prosody()->GetMode();
      
      if ( Boundary > 0.8 ) {
      if ( Mode > 0.8 ) {
      String += CString( "? " );
      } else {
      String += CString( ". " );
      }
      } else if ( Boundary > 0.4 ) {
      String += CString( ", " );
      }
    */
    
    // Boundary
    //Boundary = GetBoundary( (CLatticeEdgeTrans *)Edges[ e ] );
    //if ( Boundary.GetLength() > 0 ) {
    //  cout << " " << Boundary;
    //  }
  }

  Trace( 2, "Target String " << String );

  return String;
}



void 
CTranslationMemory::DisplayLatticeStatistics( const CLattice *pLattice )
{
  // some statistics of the lattice
  cout << endl;

  int Edges = pLattice->NumberOfEdges();
  int MaxLength = pLattice->MaxPathLength();

  cout << "Number of Nodes =     " << pLattice->NumberOfNodes() << endl;
  cout << "Number of Edges =     " << Edges << endl;

  cout << "Maximal Path Lenght = " << MaxLength << endl;
  cout << "Minimal Path Length = " << pLattice->MinPathLength() << endl;
  cout << "Average Path Length = " << pLattice->AveragePathLength() << endl;

  float Density = (float)Edges / (float)MaxLength;

  cout << "Lattice Density =     " << Density << endl;

  //cout.setf(ios::right | ios::fixed);
  //cout.width(10);
  //cout.precision(6);

  float NumPathes = pLattice->NumberOfPathes();
  cout << "Number of Pathes =    " << NumPathes << endl;

  cout << endl;
}



void
CTranslationMemory::ConvertToLatticeRepresentation( const CSentence *pSentence, const CVocabulary *pVoc,
                                                    CArray< CString > &Words )
{
  Words.Clear();
  int l = 0;      // we start at node 0, changed from 1 on 20/05/03 sv
  int s;
  CWordIndex w;
  CString Word;
  CWordInfo *pWordInfo = NULL;

  //commented by Joy, joy@cs.cmu.edu
  //why do we need to display sentence here?
  //pSentence->Display( pVoc, True );
  //pSentence->Display( pVoc );

  for ( s = pSentence->MinIndex(); s <= pSentence->MaxIndex(); s++ ) {
    ostringstream Stream;

    w = pSentence->GetAt( s );
    Trace( 2, s << " " << w << " " << pVoc->GetWord( w ) );
    Stream << l << " " << ( l + 1 ) << " " 
		   << pVoc->GetWord( w ) << " " 
		   << "0.0 "
		   << l << " " << ( l + 1 ) << " " 
		   << String << " " ;


    pWordInfo = pSentence->WordInfo( s );
    if ( pWordInfo ) {
	  //cout << "Word with WordInfo" << endl;

      Stream << "(ST " << *pWordInfo << " )";
      //Word += pWordInfo->AsString();
    }


    Word = CString( l ) + CString( " " ) + CString( l + 1 ) 
      + CString( " " ) + pVoc->GetWord( w )
      + CString( " " ) + CString ( l ) + CString( " " ) + CString( l + 1 );
   
    if ( m_UseProsody ) {
      CProsody Prosody;
      
      // is next word a sentence mark
      if (s < pSentence->MaxIndex() ) {
		CWordIndex next = pSentence->GetAt( s + 1);
		CString NextWord = pVoc->GetWord( next );
	
		if ( CString( ".,;:-!?" ).Pos( NextWord ) >= 0 ) {
		  if ( CString( ".;:!?" ).Pos( NextWord ) >= 0 ) {
			//cout << "SetBoundaries( 0, 0, 0.999, 0 )" << endl;
			Prosody.SetBoundaries( 0.0, 0.0, 0.0, 0.0, 1.0 );
			Prosody.SetMode( 1.0, 0.0, 0.0 );
		  } else { 
			//cout << "SetBoundaries( 0, 0, 0.5, 0 )" << endl;
			Prosody.SetBoundaries( 0.0, 0.0, 0.0, 1.0, 0.0 );
			Prosody.SetMode( 1.0, 0.0, 0.0 );
		  }
		  if ( NextWord == CString( "?" ) ) {
			Prosody.SetMode( 0.0, 0.0, 1.0 );
		  }
		  s++;
		}
      }
      
      CString String;
      Prosody.ToString( String );
      Stream << " " << String;
      Word += CString( " " ) + String;
      Trace( 3, w << " " << Word );
      //cout<<w << " " << Word<<endl;

      
    }

    // ARV - might cause problems by doing this
    //char *pBuffer = Stream.str();
    // problem with terminating 0-byte -> set one
    //pBuffer[ Stream.pcount() ] = 0;
    //Trace( 1, pBuffer );
    //CString Line( pBuffer );
    //delete []pBuffer;
    CString Line ((Stream.str()).c_str());
    Words.SetAt( Words.MaxIndex() + 1, Line );
    l++;

  }
}



void
CTranslationMemory::Rescore( CLattice *pLattice, CLM * pLM )
{
  Trace( 1, "Rescoring (with LM)" );

  int NumPathes;

  multimap< float, CArray< CLatticeEdge* > > BestPathes;
  NumPathes = pLattice->BestPathes( m_NumOfNBest, BestPathes );
  
  //cout << "Number of Pathes: " << NumPathes << endl;
  //cout << "Number of Pathes: " << BestPathes.size() << endl;

  multimap< float, CArray< CLatticeEdge* > >::iterator Iter2;
  for ( Iter2 = BestPathes.begin();
	Iter2 != BestPathes.end();
	++Iter2 ) {
    
    Trace(0, "Score: " <<  Iter2->first << "  " );
    cout << "Translation: ";
    DisplayPath( Iter2->second );
    cout << endl;
  }
}



void 
CTranslationMemory::SetDisplayLattice( int DisplayLattice )
{
  m_DisplayLattice = DisplayLattice;
}

