// Header file for CTranslationMemory

// Translation memory manages transducers, i.e. a list of transducers
// which can be on different levels, to allow for hierarchical transducers


// File ID:
// static char ID_TransMem_hh[] = "$Id: TransMem.hh 1759 2006-07-02 18:12:58Z vogel $"

// Programming: S. Vogel
// Last Update: March 16, 2006


/*
History:
060701: Switch default behaviour back to do transducer matching over all lattice edges

060316: Added functionality to directly load a transducer into the translation memory
        i.e. not having to have all the parameter files required when loading 
        a variable number of transducer files

*/


/* ToDo:
- Need to check that calculation of neg log probs when loading the transducers works
  under all conditions.

*/



#ifndef TransMem_hh
#define TransMem_hh

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


//#include <Bool.hh>
#include <param.hh>

#include <Lattice.hh>
#include <LatticeEdgeTrans.hh>
#include <LatticeLM.hh>

#include <LatticeSearch.hh>

//#include "phrasetree.hh"
//#include "PhraseWhg.hh"


#include <LMInterface.hh>
#include <LMwordSA.hh>
#include <SriLM.hh>


#include <Transducer.hh>


class CTranslationMemory
{
public:
  CTranslationMemory();
  ~CTranslationMemory();
  
  void Init( const char *ParamFileName, CVocabulary *pVocabulary );

  // only init with vocabulary (used when loading transducers directly)
  void Init( CVocabulary *pVocabulary );

  // match source phrase only against first edge connected to a node (=0)
  // or match against all (=1)
  // hierarchical transducer required mode=1
  void SetMatchingMode( int MatchingMode );
  
  // load a transducer directly (avoid seeting up all the parameter files)
  CTransducer * LoadTransducer( const char *TransducerFile );

  // remove transducers, transducer source, and transducer target
  void ClearTransducer();

  // set the direction in which the transducers should be used 
  // in the next call of TranslateSentence or TranslateLattice
  // 0 = source -> target, 1 = target -> source
  void SetDirection( int Direction );

  int FirstLevel();
  int LastLevel();

  void ApplyTransducer( CLatticeLM *pLattice, int Level );

  CString TranslateSentence( const CSentence *pSentence );
  CString TranslateLattice( CLatticeLM *pLattice );

  void ConvertToLatticeRepresentation( const CSentence *pSentence,
                                       const CVocabulary *pVoc,
                                       CArray< CString > &Words );
  
  void BuildTranslationLattice( CLatticeLM *pLattice );
  CString ApplyLanguageModel( CLatticeLM *pLattice );
  
  void DisplayPath( CArray< CLatticeEdge* > &Path );
  void WritePath( ofstream &Stream, CArray< CLatticeEdge* > &Path );

  CTransducer *Transducer( int Number );
  void DisplayTransducer( int Number = -1);
  void RenormalizeTransducer( int Number = -1 );
  void ClearTransducerCounts( int Number = -1 );
  void SetScalingFactors( float scalingFactor, int Number );
  void SetScalingFactors( CNumericalArray< float >* pScalingFactors );
  void SetLatticePruning( float LatticePruneBeamSize,
                          float LatticePruneBeamSizeEpsilon,
                          int LatticePruneMaxTranslations );
  
  void StoreTransducer( int Number = -1 );

  CLM * GetLM();
  CLatticeLM *Lattice();

  void SetDisplayLattice( int DisplayLattice = 1 );

  void SetBeamFactor( float BeamFactor );
  

private:
  void InitParams( const char *ParamFileName );
  
  CTransducer * InitTransducer( CString &TransducerParamFile,
				CVocabulary *pVocabulary );

  void ApplyTransducer( const CTransducerAccess *pTransducer );
  
  void Rescore( CLattice *pLattice, CLM * pLM );
  
  CString TargetString( CArray< CLatticeEdge* > &Path );
  CString Translation( CArray< CLatticeEdge* > &Path );


  void DisplayBestPath(  const CLattice *pLattice );
  
  void DisplayTargetString( const CLattice *pLattice );
  
  void DisplayLatticeStatistics( const CLattice *pLattice );

  //void
  //ShortestPath( const CLattice *pLattice, CArray< CString > &Lattice );
  
  CString
  GetBoundary( CLatticeEdgeTrans *pEdge );
  
  

  // member variables  
  CLatticeLM *m_pLattice;
  CLatticeEdgeTrans *m_pLatticeEdgeTransPrototype;

  CLatticeSearch m_LatticeSearch;
  
  //! the scaling factors for the different models to propagate to LatticeSearch
  CNumericalArray< float >* m_pScalingFactors;

  char *m_TransducerList;
  CArray< CString > m_TransducerParamFiles;


  CNumericalArray< CTransducer * > m_Transducers;
  CNumericalArray< CTransducerAccess * > m_TransducersSource;
  CNumericalArray< CTransducerAccess * > m_TransducersTarget;

  int m_Direction; // 0 = source -> target, 1 = target -> source

  CLM *m_pLM;

  int m_UseProsody;
  int m_NumOfNBest;

  float m_BeamFactor;

  char *m_VocFile;
  CVocabulary *m_pVocabulary;

  
  char *m_BigramLMParamFile;
  char *m_TrigramLMParamFile;
  char *m_WordLM;
  // the name of the language model to be run
  char *m_LMName;
  int m_HistoryLength;

  float m_LMScale;
  float m_TMScale;
  float m_ERScale;
  float m_CCScale;

  //char *m_WhgList;
  //char *m_SentenceList;
  //char *m_InCorpFile;

  int m_LatticeStatistics;
  int m_DisplayLattice;

  //added by Joy, joy@cs.cmu.edu
  //Mar 29, 2003
  //as a switch to control how to apply transducers
  int m_TransducerMatchingMode;

  CParam m_Param;
  
};


#endif
