// Header file for CTransducer, CTransducerItem, CTransducerAccess

// File ID:
// static char ID_Transducer_hh[] = "$Id: Transducer.hh 1677 2006-06-14 14:45:55Z vogel $"

// Programming: S. Vogel
// Last Update: June 03, 2006


/*
History:
060316: double to float;
        added flag if values are probabilities or costs
        when probabilities, then calcuate logs at loading time
060603: output file name when loading
        some formatting
*/




#ifndef Transducer_hh
#define Transducer_hh

// include autoconf defines
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif



#ifdef HAVE_EXT_HASH_MAP
#include <ext/hash_map>
namespace sgi {
  using __gnu_cxx::hash;
  using __gnu_cxx::hash_map;
}
#define SGI_HASH_NAMESPACE __gnu_cxx
#else
#include <hash_map>

namespace sgi {
     using std::hash;
     using std::hash_map;
}
#define SGI_HASH_NAMESPACE std
#endif


#include <map>
#include <string>

#include <param.hh>
#include <Text.hh>
#include <NumericalArray.hh>
#include <Trace.hh>



using namespace std;


class CTransducer;
class CTransducerItem;

typedef int CTransducerState;



class CStateWord
{
public:
  CStateWord( CTransducerState State, CWordIndex WordIndex );
  CTransducerState m_State;
  CWordIndex       m_WordIndex;
};



struct CompareStateWord
{
  bool operator()(const CStateWord & StateWord1, const CStateWord & StateWord2 ) const
  {
    return 
      (StateWord1.m_State < StateWord2.m_State) 
      || ( (StateWord1.m_State == StateWord2.m_State) && (StateWord1.m_WordIndex < StateWord2.m_WordIndex) );
  }
};



class CTransducerAccess
{
public:
  // access to transducer in direct (=0) or reverted (=1) modus
  CTransducerAccess( CTransducer *pTransducer, int Modus );
  ~CTransducerAccess();

  const CTransducer *Transducer() const;

  //static CTransducerState Start();

  int Next( CTransducerState State, CWordIndex Word, 
            CTransducerState &NextState ) const;

  bool IsFinal( CTransducerState State ) const;
  bool IsExpandable( CTransducerState State ) const;

  float Score( const CTransducerItem *pTransducerItem ) const;

  CSentence * TargetSentence( const CTransducerItem *pTransducerItem ) const;
  void Display( const CTransducerItem *pTransducerItem ) const;

  CArray< CTransducerItem *> *GetOutput( CTransducerState State ) const;

  void DisplayTransitionTable() const;
  void DisplayEmissionTable() const;

  // Display number of states, final states, transitions and emissions
  void Statistics() const;


private:
  int m_Direction;

  //allows scaling of each individual transducer after the log cost have been calculated

  CTransducer *m_pTransducer;

  CVocabulary *m_pInputVoc;
  CVocabulary *m_pOutputVoc;
  
  //sgi::hash_map< CTransducerState, sgi::hash_map < CWordIndex, CTransducerState > > *m_pTransitionTable;
  std::map< CStateWord, CTransducerState, CompareStateWord > *m_pTransitionTable;

  // for each final state we have a list of transducer item indices
  // EmissionTableSource is the emissions from the final states in the source tree
  // i.e. the emission of target sentences
  sgi::hash_map< CTransducerState , CNumericalArray< int > > *m_pEmissionTable;

  // and here are the actual transducer items, i.e. the emmision items
  CNumericalArray< CTransducerItem * > *m_pTransducerItems;

  int m_NumberOfStates;
};



struct EqStr
{
  bool operator()( const char * s1, const char *s2 ) const
  {
    //std::cout << "EqualString " << s1 << " " << s2;
    /*
    if ( strcmp( s1, s2 ) == 0 )
      std::cout << " equal" << std::endl;
    else 
      std::cout << " not equal " << std::endl;
    */

    return strcmp( s1, s2 ) == 0;;
  }
};




class CTransducerItem
{
public:
  CTransducerItem();
  ~CTransducerItem();

  int Number();

  float &Count();
  // scores are renormalized emission probabilities
  // ScoreTargetGivenSource = p( Target | Source ),
  // ScoreSourceGivenTarget = p( Source | Target )
  float ScoreTargetGivenSource() const;
  float ScoreSourceGivenTarget() const;
  
  float SetScoreTargetGivenSource(float value);
  float SetScoreSourceGivenTarget(float value);


  float ScalingFactor() const;
  void SetScalingFactor(float value);

  int SourceLength() const;
  int TargetLength() const;

  CSentence * SourceSentence();
  CSentence * TargetSentence();

  string Source();
  string Target();
  string Label();

  void Display() const;
  void DisplayReverted() const;

  // This field is filled when the transducer is loaded
  CNumericalArray< float > m_ModelScores;
  

private:
  int m_Number;

  //scaling factor to be used after -log transformation has occurred
  float m_ScalingFactor;

  float m_Count;
  float m_ScoreTargetGivenSource;
  float m_ScoreSourceGivenTarget;

  int m_SourceLength;
  int m_TargetLength;

  CSentence *m_pSource;
  CSentence *m_pTarget;

  // sv 060603: hmm, this means that source and target side are stored double, what for?
  string m_Source;
  string m_Target;
  string m_Label;

  static CVocabulary *s_pInputVoc;
  static CVocabulary *s_pOutputVoc;

  friend class CTransducer;
  friend class CTransducerAccess;
};




class CTransducer
{
public:
  CTransducer();
  ~CTransducer();

  // give the vocabularies to the transducer
  bool Init( CVocabulary *pInputVocabulary, CVocabulary *pOutputVocabulary );

  // are values in file probs or costs, i.e. neg log probs
  void SetValuesInFileAreProbabilities( bool ValuesAreProbs = True );

  // name of the transducer
  const char * Name() const;

  // level of transducer ( if in a cascade of transducers )
  int Level() const;
  void SetLevel( int Level );

  // one transducer may consist of several named transducers
  // however, those are not compiled into one transducer
  // therefore it is neccessary to apply the transducer
  // iteratively, until no more modifications take place
  bool ShouldBeIterated() const;

  // probabilities are smoothed with 1/VocabularySize
  void SetLambdaSmooth( float LambdaSmooth );
  float LambdaSmooth();

  // maximum of allowed error cost
  float MaxCost() const;
  float MaxCostPerWord() const;

  // load and store transducer
  bool Load( const char *pFileName );
  bool LoadParam( const char *pFileName );
  bool Store( const char *pFileName = NULL );

  // for each final state calculate probability as relative
  // frequency of the emitted items
  // from counts
  void Renormalize();
  // from probs
  void RenormalizePr();

  // init transducer to uniform emission probabilities
  void Init();

  void Display();
  void Report() const;

  static CTransducerState Start();

  //CTransducerState StartSource();
  //CTransducerState StartTarget();

  int NextSource( CTransducerState State, CWordIndex Word, 
		  CTransducerState &NextState );

  int NextTarget( CTransducerState State, CWordIndex Word, 
		  CTransducerState &NextState );

  //int Next( CTransducerState State, CWordIndex Word, 
  //          CArray<CTransducerState > &NextStates );

  bool IsFinal( CTransducerState State ) const;
  bool IsExpandable( CTransducerState State );

  CArray< CTransducerItem *> *GetOutput( CTransducerState State ) const;
  CNumericalArray< CTransducerItem * > TransducerItems();


  void ClearCounts();
  void SetScalingFactors( float Factor );
  //void UpdateCount( CTransducerItem Item, float Count );

  // LoadCosts can only be executed after Load !!!!!!!!
  bool LoadCostFiles( const char *SubFile,
                      const char *DelFile,
                      const char *InsFile );


  void DisplayTransitionTable();
  void DisplayEmissionTable();


protected:

  void BuildSourceTables();
  void BuildTargetTables();

  void RenormalizeSource();        // from counts
  void RenormalizeTarget();
  void RenormalizePrSource();      // form probs
  void RenormalizePrTarget();
  void CopyCountToPrSource();
  void CopyCountToPrTarget();
  
  void Trim( string &S );

  CParam Param;

  // name of transducer
  char *m_Name;

  // level of transducer
  int m_Level;

  // name of the file from which transducer description is read
  char *m_File;

  // are values in file probs or costs, i.e. neg log probs
  // if yes, then calculate neg log probs while loading
  bool m_ValuesInFileAreProbabilities;

  bool m_BuildMode;
  int m_NumberOfStatesSource;
  int m_NumberOfStatesTarget;

  CVocabulary *m_pInputVoc;
  CVocabulary *m_pOutputVoc;

  int m_AllocPatterns;
  int m_AllocStates;
  //int m_AllocVocSource;
  //int m_AllocVocTarget;

  bool m_Revert;
  int m_Renormalize;
  bool m_ShouldBeIterated;

  float m_LambdaSmooth;

  char *m_CostInsFile;
  char *m_CostDelFile;
  char *m_CostSubstFile;

  float m_MaxCostPerWord;

  // here are the actual transducer items, i.e. the emmision items
  CNumericalArray< CTransducerItem * > m_TransducerItems;

  // for each state we have a list of (word index, next state) pairs
  //sgi::hash_map< CTransducerState, sgi::hash_map < CWordIndex, CTransducerState > > m_TransitionTableSource;
  //sgi::hash_map< CTransducerState, sgi::hash_map < CWordIndex, CTransducerState > > m_TransitionTableTarget;

  map< CStateWord, CTransducerState, CompareStateWord > m_TransitionTableSource;
  map< CStateWord, CTransducerState, CompareStateWord > m_TransitionTableTarget;

  // for each final state we have a list of transducer item indices
  // EmissionTableSource is the emissions from the final states in the source tree
  // i.e. the emission of target sentences
  sgi::hash_map< CTransducerState , CNumericalArray< int > > m_EmissionTableSource;
  sgi::hash_map< CTransducerState , CNumericalArray< int > > m_EmissionTableTarget;
  // replace by 
  //multimap< CTransducerState, int > m_EmissionTableSource;
  //multimap< CTransducerState, int > m_EmissionTableTarget;

  // to store counts according to label in renormalization
  sgi::hash_map< const char *, float, sgi::hash<const char*>, eqstr > m_CountsForLabels;

  static CTransducerState s_NotFound;

  friend class CTransducerAccess;

};






// -----------------------------------------
// INLINES
// -----------------------------------------

inline
CStateWord::CStateWord( CTransducerState State, CWordIndex WordIndex )
{
  m_State = State;
  m_WordIndex = WordIndex;
}



inline int
CTransducerItem::Number()
{
  return m_Number;
}



inline string
CTransducerItem::Source()
{
  return m_Source;
}


inline string
CTransducerItem::Target()
{
  return m_Target;
}


inline int
CTransducerItem::SourceLength() const
{
  return m_SourceLength;
}


inline CSentence *
CTransducerItem::SourceSentence()
{
  // source as CSentence
  if ( m_pSource == NULL ) {
    m_pSource = new CSentence;
    m_pSource->ReadFromBuffer( m_Source.c_str(), s_pInputVoc );
  }
  return m_pSource;
}


inline CSentence *
CTransducerItem::TargetSentence()
{
  // target as CSentence
  if ( m_pTarget == NULL ) {
    m_pTarget = new CSentence;
    m_pTarget->ReadFromBuffer( m_Target.c_str(), s_pOutputVoc );
  }
  return m_pTarget;
}


inline string
CTransducerItem::Label()
{
  return m_Label;
}


inline float &
CTransducerItem::Count()
{
  return m_Count;
}



inline float
CTransducerItem::ScoreTargetGivenSource() const
{
  return m_ScoreTargetGivenSource;
}



inline float
CTransducerItem::ScoreSourceGivenTarget() const
{
  return m_ScoreSourceGivenTarget;
}


inline float
CTransducerItem::SetScoreTargetGivenSource( float Score ) 
{
  m_ScoreTargetGivenSource = Score;
}


inline float
CTransducerItem::SetScoreSourceGivenTarget( float Score )
{
  m_ScoreSourceGivenTarget = Score;
}


inline float 
CTransducerItem::ScalingFactor() const{
  return m_ScalingFactor;
}


inline void 
CTransducerItem::SetScalingFactor( float Score )
{
  m_ScalingFactor = Score;
}


inline const CTransducer *
CTransducerAccess::Transducer() const
{
  return m_pTransducer;
}



inline bool
CTransducerAccess::IsFinal( CTransducerState State ) const
{  
  // states in the emission table are final

  sgi::hash_map< CTransducerState , CNumericalArray< int > >::const_iterator FindIter;
  FindIter = m_pEmissionTable->find( State );

  if ( FindIter != m_pEmissionTable->end() ) {
    //Trace( 0, State << " is final" );
    return TRUE;
  } else {
    //Trace( 0, State << " is not final" );
    return FALSE;
  }
}



inline float
CTransducerAccess::Score( const CTransducerItem *pTransducerItem ) const
{
  if ( m_Direction == 0 ) {
    return pTransducerItem->m_ScoreSourceGivenTarget;
  } else {
    return pTransducerItem->m_ScoreTargetGivenSource;
  }
}



inline CSentence *
CTransducerAccess::TargetSentence( const CTransducerItem *pTransducerItem ) const
{
  Trace( 1, "CTransducerAccess::TargetSentence" );
  if ( m_Direction == 0 ) {
    //cout << "Direct: ";
    //((CTransducerItem *)pTransducerItem)->TargetSentence()->Display( m_pOutputVoc );
    return ((CTransducerItem *)pTransducerItem)->TargetSentence();
  } else {
    //cout << "Reverted: ";
    //((CTransducerItem *)pTransducerItem)->SourceSentence()->Display( m_pInputVoc );
    return ((CTransducerItem *)pTransducerItem)->SourceSentence();
  }

}



inline void 
CTransducerAccess::Display( const CTransducerItem *pTransducerItem ) const
{
  if ( m_Direction == 0 ) {
    pTransducerItem->Display();
  } else {
    pTransducerItem->DisplayReverted();
  }
}



inline CArray< CTransducerItem *> *
CTransducerAccess::GetOutput( CTransducerState State ) const
{
  // first make shure that State is really a final state
  
  Trace( 1, "GetOutput" );

  CArray< CTransducerItem *> *pTransducerItems = new  CArray< CTransducerItem *>;
  //  sgi::hash_map< CTransducerState , CNumericalArray< int > > m_EmissionTable;

  sgi::hash_map< CTransducerState , CNumericalArray< int > >::const_iterator Iter;
  Iter = m_pEmissionTable->find( State );

  CNumericalArray< int > EmissionItems = Iter->second;

  for ( int i = EmissionItems.MinIndex(); i <= EmissionItems.MaxIndex(); i++ ) {
    //m_TransducerItems[ EmissionItems[ i ] ]->Display();
    pTransducerItems->Append( (*m_pTransducerItems)[ EmissionItems[ i ] ] );
    
  }

  //cout << endl;

  return pTransducerItems;
}



inline bool
CTransducer::ShouldBeIterated() const
{
  return m_ShouldBeIterated;
}



//inline void
//CTransducer::SetName( const CString &Name ) const
//{
//  m_Name = Name;
//}



inline const char *
CTransducer::Name() const
{
  return m_Name;
}



inline void
CTransducer::SetValuesInFileAreProbabilities( bool ValuesAreProbs )
{
  m_ValuesInFileAreProbabilities = ValuesAreProbs;
}



inline int
CTransducer::Level() const
{
  return m_Level;
}



inline void
CTransducer::SetLevel( int Level )
{
  m_Level = Level;
}



inline void 
CTransducer::SetLambdaSmooth( float LambdaSmooth )
{
  m_LambdaSmooth = LambdaSmooth;
}



inline float CTransducer::LambdaSmooth()
{
  return m_LambdaSmooth;
}



inline float
CTransducer::MaxCostPerWord() const
{
  return m_MaxCostPerWord;
}



inline CNumericalArray< CTransducerItem * >
CTransducer::TransducerItems()
{
  return m_TransducerItems;
}



#endif

