/**
 * \file Format.h 
 * \author Qin Gao
 * \date  07-24-09
 * \brief Provide the format interface, defines the default operations of format
 * A format consists of three parts:
 *   - IO, read from streams, and provide a object
 *   - Serialize, write to a stream
 */

/***************************************************************************
 *            Format.h
 *
 *  Sat Jul 25 00:55:34 2009
 *  Copyright  2009  qing
 *  <qing@<host>>
 ****************************************************************************/

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor Boston, MA 02110-1301,  USA
 */

#ifndef __QMT_FORMAT_H__
#define __QMT_FORMAT_H__

#include <set>
#include <iostream>
#include <fstream>
#include <string>
#include <boost/shared_ptr.hpp>


namespace qformat{

	typedef unsigned int TQSerialNumber;

	// Serial numbers for different objects
	const TQSerialNumber TS_SwirlSRLSentence = 0x53574952; // 'SWIR'
	
	/*! Base class of CQObject, be able to be read by CQFormat
	    Use GetClassSerial() to determine what kind of object it is
	 */
	class CQObject{
	public:
		/// Get the serial number of the class
		virtual TQSerialNumber GetClassSerial() const = 0;
	};

	/*!
	 The interface of CQFormat, that can read/write from a file
	 */
	class CQFormat{
	public:
		/// Get all supported objects (usually only one)
		virtual const std::set<TQSerialNumber>& GetSupportTypes() const = 0;
			
		/*! Read next CQObject, returned through shared pointer
		 	Note that the caller must call "HasNext()" function to determine
		 	whether next reading is possible
		*/
		virtual boost::shared_ptr<CQObject> ReadNext() = 0;

		/*!
		 	The implementation will return true if next call to ReadNext()
		    is possible, otherwise should return false
		 */
		virtual bool HasNext() = 0;

	};

	/*!
	 CQFileFormat provides basic functionality of read objects from files
	 */
	class CQFileFormat : public CQFormat{
	private:
		// The stream used in reading files
		std::ifstream is_stream;
		std::string   filename;	
	protected:
		inline std::istream& getStream() {return is_stream;};
	public:
		/*!
			 Open the file given the name
		*/
		virtual bool OpenFile(const char* fname);

		/*!
			 Close the file (if openned)
		*/
		virtual void CloseFile();

		/*!
			 Rewind the input stream, if possible.
		*/
		virtual bool Rewind();

	};
}
#endif
