/**
 * \file PrasalSRLInfo.h 
 * \author Qin Gao
 * \date  07-24-09
 * \brief Provide interface to get SRL information in a phrase
 */


/***************************************************************************
 *            PhrasalSRLInfo.h
 *
 *  Sat Jul 25 19:59:57 2009
 *  Copyright  2009  qing
 *  <qing@<host>>
 ****************************************************************************/

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor Boston, MA 02110-1301,  USA
 */

#ifndef __PHRASAL_SRL_INFO__
#define __PHRASAL_SRL_INFO__

#include <string>
#include "SwirlSRLFormat.h"

namespace qformat{

	/*!
	 The phrasal SRL information

	 Provides the following information:

	 - In the phrase, provided how many predicates, what are them
	 - In the prhase, provided how many (may be nested, or incomplete 
	   arguments, what are their labels, what are their predicate, and
	   what is the relationship between it and its predicates) 
	 - Are all the arguments complete?
	 */
	class CQPhrasalSRLInfo{
		public:
			/// Predicate information (of the phrase)
			struct _QPredicateInfo{
				/// Head word of the predicate
				std::string Predicate;
				/// Number of arguments
				size_t NumArguments; 
				/// Number of completed arguments (in the phrase)
				size_t NumCompleteArguments;
				/// Number of incomplete arguments (partially in the phrase)
				size_t NumIncompleteArguments;
				/// Number of missing arguments (not in the prhase)
				size_t NumMissingArguments;
				/// Of the incomplete arguments before the phrase
				size_t NumIncompleteArugmentsBefore;
				/// Of the incomplete arguments before the phrase
				size_t NumIncompleteArugmentsAfter;
				/// Of the missing arguments before the phrase
				size_t NumMissingArgumentsBefore;
				/// Of the missing arguments after the phrase
				size_t NumMissingArgumentsAfter;
				///Count of the happenings
				mutable size_t Count;

				_QPredicateInfo():
					NumArguments(0),NumCompleteArguments(0),
					NumIncompleteArguments(0),
					NumMissingArguments(0),NumIncompleteArugmentsBefore(0),
					NumIncompleteArugmentsAfter(0),NumMissingArgumentsBefore(0),
					NumMissingArgumentsAfter(0),Count(0)
				{}
				inline bool operator==(const _QPredicateInfo& ret)const{
					return NumArguments == ret.NumArguments &&
   				    NumCompleteArguments == ret.NumCompleteArguments &&
					NumIncompleteArguments == ret.NumIncompleteArguments &&
					NumMissingArguments == ret.NumMissingArguments &&
					NumIncompleteArugmentsBefore == 
						ret.NumIncompleteArugmentsBefore &&
					NumIncompleteArugmentsAfter == 
						ret.NumIncompleteArugmentsAfter &&
					NumMissingArgumentsBefore == 
						ret.NumMissingArgumentsBefore &&
					NumMissingArgumentsAfter ==
						ret.NumMissingArgumentsAfter &&
					Predicate == ret.Predicate;
				};
			};

			struct _QArgumentInfo{
				/// Three possible values of Predicate locations
				static const int PL_PRESENT     = 0;
				static const int PL_MISS_BEFORE = 1;
				static const int PL_MISS_AFTER  = 2;

				/// Possible values of completeness
				static const int CP_COMPLETE    = 0;
				static const int CP_INCPL_BEFORE= 1;
				static const int CP_INCPL_AFTER = 2;
				static const int CP_INCPL_BOTH  = 3;
				static const int CP_NOT_OVLP    = 4;
				
				
				/// The head of the argument 
				std::string Predicate;

				/// The label of the argument
				std::string Label;
				/*! The location of the predicate
				 */
				int PredicateLocation;

				/// The completeness of the argument
				int Completeness;

				/// Count of the happenings
				size_t Count;
			};


		private:
			std::vector<_QPredicateInfo> predicates;
			std::vector<_QArgumentInfo>  arguments;
			
		public:
			/// Get the number of provided arguments
			inline size_t GetProvidedArgumentCount() const 
					{return arguments.size();};


			/// Get the number of provided predicates
			inline size_t GetProvidedPredicateCount() const
					{return predicates.size();};


			/// Get the predicate information 
			const _QPredicateInfo& GetPredicateInfo(int index) const;

			/// Get the argument information
			const _QArgumentInfo&  GetArgumentInfo(int index) const;


		public:
			/*!
			 Build the phrasal information given sentence, and the start/end
			 of the phrase, note that the start is inclusive and the end is
			 exclusive, i.e.
			 W0 (W1 W2 W3 W5 W6) W7
			 will have start = 1, end = 7
			 */
			CQPhrasalSRLInfo(const CQSRLSentence& sent, int start, int end);
			~CQPhrasalSRLInfo(){};

		private:
			/*!
			 Build the phrase information
			 */
			void build_phrase_info(
			              const CQSRLSentence& sent, int start, int end);

			_QArgumentInfo* add_augument_info(
			                       const char* pred,
			                       const char* label,
			                       size_t predidx, size_t argstart, 
			                       size_t argend, size_t pstart, size_t pend
			                       );
			
	};
}



#endif






 