/**
 * \file SwirlSRLFormat.h 
 * \author Qin Gao
 * \date  07-24-09
 * \brief Read and write SwirlSRLFormat
 * The SwirlSRLFormat is empty-line breaking format. The file provides two
 * classes:
 *    - CQSwirlSRLFormat is the general format class, provide read/write
 *    - CQSwirlSRLSentence represents a sentence in SwirlSRL format
 */

/***************************************************************************
 *            SwirlSRLFormat.h
 *
 *  Sat Jul 25 00:43:08 2009
 *  Copyright  2009  qing
 *  <qing@<host>>
 ****************************************************************************/

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor Boston, MA 02110-1301,  USA
 */

#ifndef __SWIRL_SRL_FORMAT_H_
#define __SWIRL_SRL_FORMAT_H_

#include "Format.h"
#include <set>
#include <string>
#include <vector>


namespace qformat{


	/*!
	 A predicate in the SWirl format
	 Use GetArgumentCount() to get the number of arguments
	 The GetArgumentBegin() and GetArgumentEnd() functions get the begin/end
	 indice of arguments, given the index of the argument.
	 The begin indice are inclusive, zero based, and the end indice are exclusive
	 for example, an argument starts at the first word, ends in the third word 
	 will have
	 GetArgumentBegin() = 0
	 GetArgumentEnd() = 3
	 */
	class CQSRLPredicate{
		private:
			std::vector<std::string> argument_types;
			std::vector<size_t> argument_begins;
			std::vector<size_t> argument_ends;
			std::string predicate;
			size_t index;

		public:
			inline const std::string& GetPredicate()const {return predicate;};
			/// Number of arguments
			inline size_t GetArgumentCount() const 
						{return argument_types.size();}; 
			/// Get Argument type, assertion is made on index
			const std::string& GetArgumentType(size_t idx) const;
			/// Get the argument begin index
			size_t GetArgumentBegin(size_t idx) const;
			/// Get the argument end index
			size_t GetArgumentEnd(size_t idx) const;
			/// Get the argument length
			size_t GetArgumentLength(size_t idx) const;

			inline size_t GetIndex()const {return index;};
			
		public: // Constructor/Destructors

			/*! Create a predicate given the representation of the predicate
			 And the index in the sentence
			*/
			CQSRLPredicate(const char* pred,size_t idx);
			CQSRLPredicate(const CQSRLPredicate& r);
			~CQSRLPredicate(){};

		public: // Modifications
			/*!
			 Add a new argument, given type, begin and end
			 */
			void AddArgument(const char* type, size_t begin, size_t end);

			/*!
			 Remove an argument, given the index of the argument
			 */
			void RemoveArgument(size_t idx);

		public: // Operators/Comparisons
			inline bool operator<(const CQSRLPredicate& r) const
									{return index<r.index;};

			inline bool operator==(const CQSRLPredicate& r) const
									{return index==r.index;};

			const CQSRLPredicate& operator=(const CQSRLPredicate &r);
			
	};
	
	class CQSRLSentence : public CQObject {
		private:
			std::vector<CQSRLPredicate> predicates;
			size_t slength;

		public:
			virtual inline TQSerialNumber GetClassSerial() const 
									{return TS_SwirlSRLSentence;};
			/// Get sentence length
			virtual inline size_t GetLength() const {return slength;};

			/// Get number of predicates
			virtual inline size_t GetPredicateCount() const 
								{return predicates.size();};

			/// Get the const predicate given index
			virtual const CQSRLPredicate& GetPredicate(size_t idx) const;

			/// Get the non-const predicate
			virtual CQSRLPredicate& GetPredicate(size_t idx);

		public:
			/*!
			 Add a new predicate to the sentence, given predicate and index
			 \return The newly created predicate object
			*/	
			virtual CQSRLPredicate& AddPredicate(const char* pred,size_t idx);

		public:
			CQSRLSentence(size_t length):slength(length){};
			~CQSRLSentence(){};
	};
	
	/*! Swirl SRL Format
	    Every sentence is separated by empty lines, and 
	    each line is one word
	 */
	class CQSwirlSRLFormat : public CQFileFormat {
	private:
		static std::set<TQSerialNumber> supported_types;
		boost::shared_ptr<CQObject> next;
	public:
				/// Get all supported objects (usually only one)
		virtual const std::set<TQSerialNumber>& GetSupportTypes() const;
			
		/*! Read next CQObject, returned through shared pointer
		 	Note that the caller must call "HasNext()" function to determine
		 	whether next reading is possible
		*/
		virtual boost::shared_ptr<CQObject> ReadNext();

		/*!
		 	The implementation will return true if next call to ReadNext()
		    is possible, otherwise should return false
		 */
		virtual bool HasNext();
			
	};
}


#endif

