#include <string.h>
#include <stdlib.h>
//
// TranslateMM.cc
// Copy of TranslateBatchLM.cc to modify decoder for using multiple LMs, DMs, SLMs
// Copy of TranslateOptmizeMER.cc to modify decoder for accessing the LM in batch mode
//
// from TranslateOptimize.cc:
// Main function for the translation program which allows to use
// multiple models, i.e. multiple TM scores, LM, Distorion, Fertility, etc. 
//
// Programming: S. Vogel
//
// History
// Started from a copy of TranslateTS.cc
// 2005-10-29: restructure code, put loading and initialization into functions
// 2005-11-01: split translation part into bits and pieces, esp. separate optimization
// 2005-11-05: optimization now in COptimizer
// 2005-11-11: moved param defs into main instead of having function call with long parameter list
// 2005-12-15: changes in lexicon to make the wordindex 4 bytes on 32bit and 64bit machines


// 2006-04-04: C++ implementation of MER
// 2006-05-14: Added the 4-gram capability, i.e. copied over from TanslateOptimize.cc
// 2006-05-25: Added NULL word translation
// 2006-06-11: Integrated new suffix array LM
// 2006-06-13: Reactivated n-best list rescoring
// 2006-06-18: Started on batch mode for LM
// 2006-06-26: suffix array LM does now Good-Turing smoothing

// 2006-07-16: now up to 4 DMs can be loaded
// 2006-07-24: now up to 4 LMs, each with its own LMSentenceState, LMState, and vocabulary

// 2006-12-03: added interactive mode with TranslationServerSocket
// 2006-12:    added pruning of translation lattices.
//             new params: LatticePruneBeamSize, LatticePruneMaxTranslations, 
//                         LatticePruneBeamSizeEpsilon
//             Warning: Source Lattice edge score propagation is disabled temporarily
//
// 2007-01-06: added internal regex normalization via TextFilterBoostRegex
//             use param NormalizingRegexFileName with path to file containing perl-style
//             regular expressions.
// 2007:       added sentence-level feature min, max output to log (param: DisplayFeaturesRange)
// 2007-05-24: added sentence-level TD functionality (param: PhraseTableSentenceLevelFilePrefix)
// 2007-12-12: SriLM extended to allow long histories, i.e. 5-gram, 6-gram, etc.
//             4-gram is now treated the same way as 5-gram to allow for large vocabularies
//
// ToDo:
// 2006-03-15: (Done) make sure that for NIST optimization 5-grams are used



#include <assert.h>
#include <iostream>
#include <string.h>
#include <stdlib.h>
#include <time.h>

//Nguyen, Aug 6, 2008: Cohesive Decoding
#include <tokens.hh>

// float limits constants
#include <float.h>

//! Worker Bee uses string stream
#include <iostream>
//! Worker Bee writes to disk
#include <fstream>
//! Worker Bee checks files
#include <sys/types.h>
#include <sys/stat.h> 
#include <unistd.h>
//! Worker Bee for file locking
#include <fcntl.h> // for open() 
#include <cerrno> // for errno 
#include <cstdio> // for perror()

#include <set>

#define NO_TRACE 1
#include <Trace.hh>
#include <utils.hh>

#include <Text.hh>
#include <TextFilterBase.hh>
#include <TextFilter.hh>
#include <TextFilterBoostRegex.hh>

#include <DecoderBLM.hh>
#include <Alignment.hh>
#include <LexiconTrain.hh>

#include <LMInterface.hh>
#include <_SingleCorpusSALM.h>
#include <ClassBasedSALM.h>

//! Discriminative Language Model
//#include <_DiscriminativeLM.h>

#include <_NgramDBLM.h>
#include <_CharacterDBLM.h>

#include <BFLM.hh>
#include <BoolBFLM.hh>

#include <_SuffixArraySearch.h>
#include <PhraseAlignMS.hh>
#include <PhraseSearchMS.hh>

#include <NBestListRescorer.hh>
#include <LatticeRescorer.hh>

//#include <CompactNBest.hh>
#include <NBestList.hh>

//! for text normalization
#include <pstream.h>

//! for opimization
#include <OptimizeMER.hh>
//#include <EditSearch.hh>
#include <BleuSearch.hh>
#include <TERSearch.hh>

//! new MER optimization, matlab replacement
#include <OptimizeMER.hh>

#include <EmptyWordTranslation.hh>

#include <DistortionModelJump.hh>

#include <FutureEstimator.hh>

#include <SentenceToLattice.hh>

//! server interactive mode for translation over network
#include <TranslationServerSocket.hh>

// ====================
// typedefs
// ====================

typedef map< float, vector< float > > TScoreScalingFactorMap; //use in MER optimization



// ====================
// prototypes
// ====================

CVocabulary *         LoadVocabulary       ( CParam &Param, char * VocFile );
CTranslationMemory *  LoadTranslationMemory( CParam &Param, CVocabulary *pVoc );

CLM *                 LoadLanguageModel    ( CParam &Param, char *LMName, char *LMFile );
void                  LoadLanguageModels   ( CParam &Param, CNumericalArray< CLM * > &LMs );
CLM *                 LoadLanguageModel    ( char *pLMName, char *pFileName , unsigned NGramLength, char LMInterpolationMode,
                                             int LMMaxFrequencyForDiscounting, char *pTagsConsideredAsWords, int LMUnkIsWord,
                                             int RestrictVocabulary, int BFSize, int BFNumberOfHashfunctions );


CLexiconForTraining * LoadLexicon          ( CParam &Param, char *LexiconFile, CVocabulary *pVoc );
CText *               LoadText             ( CParam &Param, char *FileName, CVocabulary *pVoc, bool ReadOneByOne = false );
CNumericalArray< CLatticeLM * >*  LoadLattices( CParam &Param, char *LatticeList, 
                                                char *LatticeFormat, CVocabulary *pVoc );
C_SuffixArraySearch * LoadSuffixArray      ( CParam &Param, char *SourcFileSA );

CDistortionModel *    LoadDistortionModel  ( CParam &Param, char *DistortionModelFile );
void                  LoadDistortionModels ( CParam &Param, CNumericalArray< CDistortionModel * > &DMs );
CDistortionModel *    LoadDistortionModel  ( char *pDistortionModelFile, float DMSmoothingFactor );

CFertilityModel *     LoadFertilityModel   ( CParam &Param, char *FertilityFile );

//COptimizer*           InitOptimizer( CParam &Param );
MetricSearch *        InitOptimizerMER     ( CParam &Param );
CTextFilterBase *     InitTextFilter       ( CParam &Param );
void                  InitPhraseLengthModel( CFertilityModel *pFertilityFE, CFertilityModel *pFertilityEF );
CTextFilter *         InitWordExpander     ( CParam &Param );

CDecoderBLM *         InitDecoder          ( CParam &Param );
CDecoderBLM *         InitTrueCaser        ( CParam &Param );


void PrepareScalingFactors    ( float LMScale, float DMScale, float WCScale, float PLScale, char *TMScaleList );
void PrepareScalingFactors    ( char *pScaleList );

void InitPhraseSearcher       ( char *pPhraseAlignScaleList );
void InitSourceWordModels     ( CNumericalArray< double >& srcModels, SriLM* pLM, int numModels );
void InitFutureCostEstimator  ( CSentence *pPartialSentence, CLattice *pLattice);

void TranslateText            ( CText *pTest );
void TranslateSentence        ( CSentence *pSentence, CLatticeLM *&pLattices,
                                CArray< CDecoderResult* > &TranslationResults );
                                
//Nguyen, Aug 6, 2008: Source-side dependency trees
void TranslateSentenceWithDependencyTree ( CSentence *pSentence, CLatticeLM *&pLattice,
                   CArray< CDecoderResult* > &TranslationResult, vector<tokens>& sentDtree, int algorithm  );
                                
void TranslateServerStart     ( int ServerPort );

// Worker Bee methods
void   TranslateTextResumable ( CText *pTest );
time_t BeeFileExists          ( const char* FileName );
int    BeeTakeUnitLock         ( const char* filename);
bool   BeeIsUnitInProgress    ( int SentenceNumber,
                                string &Sent_LockFile, string &Sent_FeatsFile, 
                                string &Sent_NBestTranslationsFile,
                                int &lockHandle,
                                bool &workerBeeAtLeastOneSentenceUnprocessed, 
                                bool &workerBeeCompletedFirstPass,
                                bool &workerBeeGetThemAllNow);

void TranslateLattices        ();
void TranslateLatticesResumable();
void TranslateLattice         ( CLatticeLM *&pLattice,
                                CArray<CDecoderResult*> &TranslationBank,
                                bool buildTranslationLattice = true);

void BuildTranslationLattice  ( CSentence *pSentence, CLatticeLM *pLattice );
void BuildTranslationLattice  ( CLatticeLM *pLattice );

int  SourceSentenceIsInCorpus ( CSentence *pSentence );
void AddFertilityScores       ( CLattice *pLattice );

void ApplyTrueCasing          ( CArray< CDecoderResult* > TranslationResults );

void RescoreNBestList         ( CSentence *pSentence, CArray< CDecoderResult* > &TranslationResults,
                                vector< float > &ScalingFactors );

void ScoreTranslation         ( RawCounts &modelBestCounts, TScoreScalingFactorMap &trackScalingFactorScores, int MERiteration );

void DisplayScalingFactors    ( TScoreScalingFactorMap ScalingTactors );
void DisplayFeaturesMinMax    ( CLatticeLM *pLattice );

void UpdatePhraseAlignerScalingFactors();

void ExpandSourceSentence     ( CLattice *pLattice );
void NormalizeTranslations    (  CArray<CDecoderResult*>& TranslationResults );
void ShowTranslationResults   ( int SentenceNumber, CArray<CDecoderResult*>& TranslationResults, CLatticeLM *pLattice );


RawCounts BuildNBest          ( CArray< CDecoderResult* > TranslationResults, NBestList* oneNBestList,
                                MetricSearch* pMetric, int translationCount, int sentenceNumber,
                                CSentence* pSourceSentence, int& TotalNewTranslations );

float OptimizeParameters      ( vector< float > &ScalingFactors, vector< NBestList * > &nBestLists,
                                MetricSearch *pMetric,
                                TScoreScalingFactorMap& trackScalingFactorScores, string timeAdded );


// ====================
//   global variables
// ====================
CParam             *pParam             = NULL;
CDecoderBLM        *pDecoder           = NULL;
CLexiconForTraining*pLexiconFE         = NULL;
CLexiconForTraining*pLexiconEF         = NULL;
CVocabulary        *pVoc               = NULL;
C_SuffixArraySearch*pSA                = NULL;
CPhraseAlignerMS   *pPhraseAligner     = NULL;
CPhraseSearchMS    *pPhraseSearcher    = NULL;
CLatticeRescorer   *pLatticeRescorer   = NULL;
CText              *pTextF             = NULL;
CText              *pTextE             = NULL;

CLM                *pLM                = NULL;
CDistortionModel   *pDistortionModel   = NULL;
CFertilityModel    *pFertilityFE       = NULL;
CFertilityModel    *pFertilityEF       = NULL;

CSentenceLengthModel *pPhraseLengthModelFE = NULL;
CSentenceLengthModel *pPhraseLengthModelEF = NULL;

CNumericalArray< CLM * > LMs;
CNumericalArray< CDistortionModel * > DMs;

CTranslationMemory *pTranslationMemory = NULL;



CNBestListRescorerTM_IBM1   *pRescorerSourceTargetIBM1 = NULL;
CNBestListRescorerTM_IBM1   *pRescorerTargetSourceIBM1 = NULL;

char               *SplitCharacters    = NULL;
int                 RescoreTransducer  = 0;

vector< float >      ScalingFactors;
vector< NBestList* > *pNBestLists      = NULL;

CTextFilterBase    *pTextFilter        = NULL;
CTextFilter        *pWordExpander      = NULL;
CSentenceToLattice *pSentenceToLattice = NULL;

MetricSearch       *pMetricSearch      = NULL;
COptimizeMER       *pOptimizerMER      = NULL;


int                 NGramsConsidered   = 1;

// we keep track of the feature weights and the resulting 1-best score
CNumericalArray< float > MetricScores;
CNumericalArray< float > OracleScores;
CNumericalArray< float > FinalScores;
CNumericalMatrix< float > FeatureWeights;

int SentenceNumber = 0;           // the number of the sentence currently translated

//Nguyen, Aug 6, 2008: Source-side dependency trees
//whole corpus: a 2D matrix, each cell is a token, a row is a sentence
vector < vector<tokens> > dv;
int    InterruptionCheckAlgorithm;
//int    InterruptionCheckAlgorithmCombination;
bool   CohesionCheck = false; 

// =================================================================================
// main() - entry into the program
// =================================================================================

int main(int argc, char *argv[])
{
  cout << "TRANSLATION WITH PHRASE TABLES, ONLINE SEARCH FOR PHRASES, AND OPTIMIZATION" << endl;
  cout << "===========================================================================" << endl << endl;

  cout << "Compiled on " << __DATE__ << " " << __TIME__ << " with g++ version: " << __VERSION__ ;
#ifdef __OPTIMIZE__
  cout << " in optimization mode" << endl;
#else
  cout << " in debug mode" << endl;
#endif

#ifdef COVERAGE_BIT_VECTOR_SIZE
  cout <<" Coverage bit vector: " << COVERAGE_BIT_VECTOR_SIZE << endl;
#endif
  cout << endl;


  PrintTime( "Begin of execution " );

  TraceLevel = 0;

  char*  TransMemParams = NULL;
  char*  SentenceList   = NULL;
  
  //Nguyen, Aug 6, 2008: Source-side dependency trees
  char*  SentenceDependencyTrees    = NULL;
  
  
  char*  LatticeList    = NULL;
  char*  LatticeFormat  = NULL;
  float  LatticeEdgeThreshold = 0.0;
  int    LatticeRemoveEpsilonEdges = 0;

  char*  VocFile        = NULL;

  char*  SaveHypothesesToFile          = NULL;

  
  char*  PreProcessingScript           = NULL;
  char*  SourceSentenceExpandingScript = NULL;
  char*  PostProcessingScript          = NULL;

  char*  StringForEmptyWord = NULL;

  int    LMBatchMode     = 0;

  char*  pLMType;
  char*  pLMFile = NULL;
  int    LMNgramLength   = 2;
  char   LMInterpolationMode = 'e';
  int    LMMaxFrequencyForDiscounting = -1;
  char*  pLMTagsConsideredAsWords = NULL;
  int    LMUnkIsWord = 1;
  int    LMRestrictVocabulary = 0;

  char*  pLM2Type;
  char*  pLM2File = NULL;
  int    LM2NgramLength   = 2;
  char   LM2InterpolationMode = 'e';
  int    LM2MaxFrequencyForDiscounting = -1;
  char*  pLM2TagsConsideredAsWords = NULL;
  int    LM2UnkIsWord = 1;
  int    LM2RestrictVocabulary = 0;

  char*  pLM3Type;
  char*  pLM3File = NULL;
  int    LM3NgramLength   = 2;
  char   LM3InterpolationMode = 'e';
  int    LM3MaxFrequencyForDiscounting = -1;
  char*  pLM3TagsConsideredAsWords = NULL;
  int    LM3UnkIsWord = 1;
  int    LM3RestrictVocabulary = 0;

  char*  pLM4Type;
  char*  pLM4File = NULL;
  int    LM4NgramLength   = 2;
  char   LM4InterpolationMode = 'e';
  int    LM4MaxFrequencyForDiscounting = -1;
  char*  pLM4TagsConsideredAsWords = NULL;
  int    LM4UnkIsWord = 1;
  int    LM4RestrictVocabulary = 0;

  // Bloom Filter params
  int  BFSize = 0;
	int  BFNumberOfHashfunctions = 0;
	int  BF2Size = 0;
	int  BF2NumberOfHashfunctions = 0;
	int  BF3Size = 0;
	int  BF3NumberOfHashfunctions = 0;
	int  BF4Size = 0;
	int  BF4NumberOfHashfunctions = 0;
  
  int    LatticeStatistics = 0;
  int    NBest             = 1;
  int    RescoreNBest      = 0;

  int    LengthOfReorderingWindow = 1;
  double ReorderingsPerWord    = 0.0;

  int    CompareHypFeatures = 7;
  int    PruneHypFeatures   = 3;

  //char*  pLMScaleList = NULL;
  char*  pScalingFactors = NULL;

  char*  pTMScaleList = NULL;

  double LMScale  = 1.0;
  double LM2Scale = 0.0;
  double LM3Scale = 0.0;
  double LM4Scale = 0.0;
  double DMScale  = 1.0;
  double DM2Scale = 1.0;
  double DM3Scale = 1.0;
  double DM4Scale = 1.0;
  double WCScale  = 1.0;
  double PCScale  = 1.0;

  int    UseFutureCostEstimates = 0;

  ulong  BeamSize           = 10000;
  ulong  MaxHypsPerCoverage = 10000000;
  int    MaxHyps            = 1000;



  char*  pPhraseTableFile = NULL;
  char*  pPhraseTableSentLevelFile = NULL;
  int    PhraseTableScoresAreCosts = 0;
  int    PhraseTableMatchingMode   = 1;

  // suffix array stuff
  char*  SourceFileSA = NULL;
  char*  TargetFileSA = NULL;
  char*  SourceCorpus = NULL;
  char*  TargetCorpus = NULL;

  int    SA_HighestFrequency     = 0;
  int    SA_MaxOccurrences       = 0;
  int    SA_ShortestPhraseLength = 0;
  int    SA_LongestPhraseLength  = 0;

  // lexicon stuff
  char*  LexiconFileFE = NULL;
  char*  LexiconFileEF = NULL;
  int    LexiconFormat = 0;

  // multi score phrase extraction parameters
  int    PhraseAlignMaxAlignments    = 10;
  float  PhraseAlignBeamSize         = 1.0;
  int    PhraseAlignMaxTranslations  = 10;
  int    PhraseAlignFindCenterMethod = 0;
  char*  pPhraseAlignScaleList       = NULL;
  float  PhraseAlignDynamicFactors   = 0.0;
  float  PhraseAlignStretchFactor    = 1.0;
  float  PhraseAlignFactorShorter    = 0.5;
  float  PhraseAlignFactorLonger     = 2.0;
  float  PhraseAlignAddConstShorter  = -2.0;
  float  PhraseAlignAddConstLonger   = 2.0;
  
  // pruning during apply transducer
  float LatticePruneBeamSize        = 0;
  float LatticePruneBeamSizeEpsilon = 0;
  int   LatticePruneMaxTranslations = 0;

  // default values for optimization parameters
  int    MEROptimize    = 0;
  int    IterationLimit = 1;
  char  *ScoringMetric  = NULL;
  char  *GainsFile      = NULL;

  // MER specific parameters see param def for descriptions
  char*  OptimizationLowerBounds           = NULL;
  char*  OptimizationUpperBounds           = NULL;
  int    OptimizationNumberOfStarts        = 5;
  int    OptimizationNumberOfPermutations  = 3;
  int    OptimizationNumberOfModifications = 2;
  float  OptimizationEpsilon               = 0.0;
  float  OptiEpsilonScale                  = 1.0;
  int    OptiTwoStage                      = 0;
  int    OptiDoNorm                        = 0;

  char*  NormalizingScript        = NULL;
  char*  NormalizingRegexFileName = NULL;
  char*  ReferenceList            = NULL;
  int    NumReferences            = 0;
  int    GenOracleScores          = 0;
  int    LoadUnique               = 0;

  // fertility and distortion model tables
  char*  FertilityFileFE = NULL;
  char*  FertilityFileEF = NULL;

  char*  DistortionModelFile = NULL;
  float  DistortionModelSmoothingFactor = 0.0;
  char*  DistortionModel2File = NULL;
  float  DistortionModel2SmoothingFactor = 0.0;
  char*  DistortionModel3File = NULL;
  float  DistortionModel3SmoothingFactor = 0.0;
  char*  DistortionModel4File = NULL;
  float  DistortionModel4SmoothingFactor = 0.0;

  // server interactive mode options
  int    ServerPort      = 4444;
  int    InteractiveMode = 0;
  
  // Worker Bee options
  char*  WorkerBeeDir = NULL;
  int    WorkerBeeSaveSentenceFeatures = 0;
  double MaxSentenceProcessingTime = 480;
  
  // display options
  int    DisplayLattice                   = 0;
  int    DisplayLatticeStatistics         = 0;
  int    DisplayNBestData                 = -1;
  int    DisplayTranslations              = 0;
  int    DisplayTranslationsNormalized    = 0;
  int    DisplayTranslationPath           = 0;
  int    DisplayFeaturesRange             = 0;


  pParam = new CParam;
  bool Optional = True;
  bool Required = False;


  pParam->AddParamDef( "TraceLevel", 'T', &TraceLevel, 0,
                       "Trace level",
                       True );
  
  //Nguyen, Aug 6, 2008: source-side dependency trees
  pParam->AddParamDef( "SentenceDependencyTrees", 'D' , &SentenceDependencyTrees, NULL , 
                       "name of file that contains source-side dependency trees in CoNLL text format", 
                       Optional );
  pParam->AddParamDef( "InterruptionCheckAlgorithm", '\0' , &InterruptionCheckAlgorithm, 0 , 
                       "Interruption Check Algorithm: 1, 2, 3, 4 ; 5: contains 4 features 1, 3, and NounCount, VerbCount", 
                       Optional );                      
  //end of Nguyen
                       
  pParam->AddParamDef( "SentenceList", 'B' , &SentenceList, NULL , 
                       "name of file that contains a list of sentences in plain text format", 
                       Optional ); // either SentenceList or LatticeList should be required

  pParam->AddParamDef( "LatticeList", 'C' , &LatticeList, NULL , 
                       "name of file that contains a list of lattices in ASR, SRI,  ASR-MS (multiscore), or VM (Verbmobil) format", 
                       Optional );
  
  pParam->AddParamDef( "LatticeFormat", 'c' , &LatticeFormat, "SRI" ,
                       "format in which lattices are stored on disk (ASR, ASR-MS (multiscore), SRI, VM (Verbmobil))",
                       Optional );
  
  pParam->AddParamDef( "LatticeEdgeThreshold", '\0', &LatticeEdgeThreshold, 0,
                       "Threshold for acoustic score of word edge. No Threshold applied: 0",
                       Optional);
  
  pParam->AddParamDef( "LatticeRemoveEpsilonEdges", '\0', &LatticeRemoveEpsilonEdges, 0,
                       "Remove edges with <epsilon>, span incoming edges at fromNode to toNode.",
                       Optional);
  
  pParam->AddParamDef( "PreProcessingScript", '\0', &PreProcessingScript, NULL,
                       "Preprocess test sentences",
                       Optional );
  
  pParam->AddParamDef( "SourceSentenceExpandingScript", '\0', &SourceSentenceExpandingScript, NULL,
                       "Analyze source words and add alternative word forms to translation lattice",
                       Optional );
  
  pParam->AddParamDef( "PostProcessingScript", '\0', &PostProcessingScript, NULL,
                       "Postprocess translations (currentlty not used, use NormalizingScript)",
                       Optional );
  
  pParam->AddParamDef( "NormalizingScript", '\0', &NormalizingScript, NULL,
                       "Normalizing text before rescoring ",
                       Optional );

  pParam->AddParamDef( "NormalizingRegexFileName", '\0', &NormalizingRegexFileName, NULL,
                       "Normalizing perl-style regular expressions used before rescoring",
                       Optional );
  
  pParam->AddParamDef( "Vocabulary", 'V' , &VocFile, NULL , 
                       "name of file that contains the vocabulary", 
                       Optional );
  
  pParam->AddParamDef( "StringForEmptyWord", '\0', &StringForEmptyWord, NULL, 
                       "string which represents the socalled Empty or Null word", 
                       Optional );

  pParam->AddParamDef( "SplitCharacters", '\0', &SplitCharacters, "|",
                       "characters at which to split the input ",
                       Optional );
  

  pParam->AddParamDef( "Lexicon", 'X' , &LexiconFileFE, NULL , 
                       "Name of file that contains the lexicon p( f | e )", 
                       Optional );
  
  pParam->AddParamDef( "ReverseLexicon", 'Y' , &LexiconFileEF, NULL , 
                       "Name of file that contains the lexicon p( e | f )", 
                       Optional );
  
  pParam->AddParamDef( "LexiconFormat", '\0' , &LexiconFormat, 0 , 
                       "LexiconFormat:  0 = standard STTK, 1 = STTK with words, 2 = GIZA with words", 
                       Optional );
  
  

  pParam->AddParamDef( "TransMemParams", 't', &TransMemParams, NULL,
                       "Param-File for translation memory",
                       Optional );

  pParam->AddParamDef( "PhraseTableFileName", '\0', &pPhraseTableFile, NULL,
                       "File with phrase pairs in the format: @Label # Src # Tgt # Scores",
                       Optional );
  
  pParam->AddParamDef( "PhraseTableSentenceLevelFilePrefix", '\0', &pPhraseTableSentLevelFile, NULL,
                       "Prefix for files, 0-indexed by source sentence number, with phrase pairs in PhraseTable format",
                       Optional );

  pParam->AddParamDef( "PhraseTableScoresAreCost", '\0', &PhraseTableScoresAreCosts, 0,
                       "Scores in phrase table are probabilities (0) or cost (1): default 0",
                       Optional );

  // Must be on for hierarchical TD and lattice translation
  pParam->AddParamDef( "PhraseTableMatchingMode", '\0', &PhraseTableMatchingMode, 1,
                       "Apply the source phrase match only to the first word per position (=0), \
                        or apply to all words per position (=1)",
                       Optional );



  pParam->AddParamDef( "LMBatchMode", '\0', &LMBatchMode, 0,
                       "Apply LM in normal mode (=0) or in batch mode (=1), default = 0",
                       Optional );

  pParam->AddParamDef( "LMFile", '\0', &pLMFile, NULL,
                       "File from which the LM is loaded",
                       Required );

  pParam->AddParamDef( "LMType", 'M', &pLMType, "SA-LM",
                       "The kind of language model that you will use, currently: SriLM, SA-LM, Class-SA-LM, LMWordSA, DB-LM",
                       Optional );

  pParam->AddParamDef( "LMNgramLength", 'H', &LMNgramLength, 3,
                       "Length of ngrams in LM, default = 3",
                       Optional );
  
  pParam->AddParamDef( "LMInterpolationMode", '\0', &LMInterpolationMode, 'e',
                       "Interpolation mode for Suffix Array LM: Equal (=e) or IBM style (=i), default = e",
                       Optional );
  
  pParam->AddParamDef( "LMMaxFrequencyForDiscounting", '\0', &LMMaxFrequencyForDiscounting, -1,
                       "Good-Turing discounting applied in SA LM for ngrams with frequency of to this number",
                       Optional );
  
  pParam->AddParamDef( "LMTagsConsideredAsWords", '\0', &pLMTagsConsideredAsWords, NULL,
                       "List of Tags (separated by '|', which are treated as normal words",
                       Optional );

  pParam->AddParamDef( "LMScale", '\0', &LMScale, 1.0,
                       "Language model scaling factor, default = 1.0",
                       Optional );

  pParam->AddParamDef( "LMUnkIsWord", '\0', &LMUnkIsWord, 1,
                       "Consider <unk> to be a word",
                       Optional );

  pParam->AddParamDef( "LMRestrictVocabulary", '\0', &LMRestrictVocabulary, 0,
                       "Restrict the vocabulary to the one loaded into the decoder",
                       Optional );


  pParam->AddParamDef( "LM2File", '\0', &pLM2File, NULL,
                       "File from which the LM is loaded",
                       Optional );

  pParam->AddParamDef( "LM2Type", '\0', &pLM2Type, "SALM",
                       "The kind of language model that you will use, currently: SriLM, SA-LM, Class-SA-LM, LMWordSA",
                       Optional );

  pParam->AddParamDef( "LM2NgramLength", '\0', &LM2NgramLength, 3,
                       "Length of ngrams in LM, default = 3",
                       Optional );
  
  pParam->AddParamDef( "LM2InterpolationMode", '\0', &LM2InterpolationMode, 'e',
                       "Interpolation mode for Suffix Array LM: Equal (=e) or IBM style (=i), default = e",
                       Optional );
  
  pParam->AddParamDef( "LM2MaxFrequencyForDiscounting", '\0', &LM2MaxFrequencyForDiscounting, -1,
                       "Good-Turing discounting applied in SA LM for ngrams with frequency of to this number",
                       Optional );
  
  pParam->AddParamDef( "LM2TagsConsideredAsWords", '\0', &pLM2TagsConsideredAsWords, NULL,
                       "List of Tags (separated by '|', which are treated as normal words",
                       Optional );

  pParam->AddParamDef( "LM2Scale", '\0', &LM2Scale, 1.0,
                       "Language model scaling factor, default = 1.0",
                       Optional );

  pParam->AddParamDef( "LM2UnkIsWord", '\0', &LM2UnkIsWord, 1,
                       "Consider <unk> to be a word",
                       Optional );

  pParam->AddParamDef( "LM2RestrictVocabulary", '\0', &LM2RestrictVocabulary, 0,
                       "Restrict the vocabulary to the one loaded into the decoder",
                       Optional );

  pParam->AddParamDef( "LM3File", '\0', &pLM3File, NULL,
                       "File from which the LM is loaded",
                       Optional );

  pParam->AddParamDef( "LM3Type", '\0', &pLM3Type, "SALM",
                       "The kind of language model that you will use, currently: SriLM, SA-LM, Class-SA-LM, LMWordSA",
                       Optional );

  pParam->AddParamDef( "LM3NgramLength", '\0', &LM3NgramLength, 3,
                       "Length of ngrams in LM, default = 3",
                       Optional );
  
  pParam->AddParamDef( "LM3InterpolationMode", '\0', &LM3InterpolationMode, 'e',
                       "Interpolation mode for Suffix Array LM: Equal (=e) or IBM style (=i), default = e",
                       Optional );
  
  pParam->AddParamDef( "LM3MaxFrequencyForDiscounting", '\0', &LM3MaxFrequencyForDiscounting, -1,
                       "Good-Turing discounting applied in SA LM for ngrams with frequency of to this number",
                       Optional );
  
  pParam->AddParamDef( "LM3TagsConsideredAsWords", '\0', &pLM3TagsConsideredAsWords, NULL,
                       "List of Tags (separated by '|', which are treated as normal words",
                       Optional );

  pParam->AddParamDef( "LM3Scale", '\0', &LM2Scale, 1.0,
                       "Language model scaling factor, default = 1.0",
                       Optional );

  pParam->AddParamDef( "LM3UnkIsWord", '\0', &LM3UnkIsWord, 1,
                       "Consider <unk> to be a word",
                       Optional );

  pParam->AddParamDef( "LM3RestrictVocabulary", '\0', &LM3RestrictVocabulary, 0,
                       "Restrict the vocabulary to the one loaded into the decoder",
                       Optional );


  pParam->AddParamDef( "LM4File", '\0', &pLM4File, NULL,
                       "File from which the LM is loaded",
                       Optional );

  pParam->AddParamDef( "LM4Type", '\0', &pLM4Type, "SALM",
                       "The kind of language model that you will use, currently: SriLM, SA-LM, Class-SA-LM, LMWordSA",
                       Optional );

  pParam->AddParamDef( "LM4NgramLength", '\0', &LM4NgramLength, 3,
                       "Length of ngrams in LM, default = 3",
                       Optional );
  
  pParam->AddParamDef( "LM4InterpolationMode", '\0', &LM4InterpolationMode, 'e',
                       "Interpolation mode for Suffix Array LM: Equal (=e) or IBM style (=i), default = e",
                       Optional );
  
  pParam->AddParamDef( "LM4MaxFrequencyForDiscounting", '\0', &LM4MaxFrequencyForDiscounting, -1,
                       "Good-Turing discounting applied in SA LM for ngrams with frequency of to this number",
                       Optional );
  
  pParam->AddParamDef( "LM4TagsConsideredAsWords", '\0', &pLM4TagsConsideredAsWords, NULL,
                       "List of Tags (separated by '|', which are treated as normal words",
                       Optional );

  pParam->AddParamDef( "LM4Scale", '\0', &LM2Scale, 1.0,
                       "Language model scaling factor, default = 1.0",
                       Optional );

  pParam->AddParamDef( "LM4UnkIsWord", '\0', &LM4UnkIsWord, 1,
                       "Consider <unk> to be a word",
                       Optional );

  pParam->AddParamDef( "LM4RestrictVocabulary", '\0', &LM4RestrictVocabulary, 0,
                       "Restrict the vocabulary to the one loaded into the decoder",
                       Optional );


  // Bloom Filter params
  
  pParam->AddParamDef( "BFNumberOfHashfunctions", '\0', &BFNumberOfHashfunctions, 5,
                       "Number of Hash functions for the BFLM",
                       Optional );
  
  
  pParam->AddParamDef( "BFSize", '\0', &BFSize, 10000,
                       "Size of BFLM in bit",
                       Optional );
  
  pParam->AddParamDef( "BF2NumberOfHashfunctions", '\0', &BF2NumberOfHashfunctions, 5,
                       "Number of Hash functions for the BFLM",
                       Optional );
  
  
  pParam->AddParamDef( "BF2Size", '\0', &BF2Size, 10000,
                       "Size of BFLM in bit",
                       Optional );
  
	pParam->AddParamDef( "BF3NumberOfHashfunctions", '\0', &BF3NumberOfHashfunctions, 5,
                       "Number of Hash functions for the BFLM",
                       Optional );
  
  
  pParam->AddParamDef( "BF3Size", '\0', &BF3Size, 10000,
                       "Size of BFLM in bit",
                       Optional );
  
	pParam->AddParamDef( "BF4NumberOfHashfunctions", '\0', &BF4NumberOfHashfunctions, 5,
                       "Number of Hash functions for the BFLM",
                       Optional );
  
  
  pParam->AddParamDef( "BF4Size", '\0', &BF4Size, 10000,
                       "Size of BFLM in bit",
                       Optional );  

  pParam->AddParamDef( "SourceFileSA", '\0', &SourceFileSA, NULL,
                       "File name for source side of corpus with suffix array",
                       Optional );
  
  pParam->AddParamDef( "TargetFileSA", '\0', &TargetFileSA, NULL,
                       "File name for target side of corpus with suffix array",
                       Optional );
  
  pParam->AddParamDef( "SourceFile", 'F' , &SourceCorpus, NULL , 
                       "File name for source side of the corpus", 
                       Optional );

  pParam->AddParamDef( "TargetFile", 'E' , &TargetCorpus, NULL , 
                       "File name for target side of the corpus", 
                       Optional );
  
  pParam->AddParamDef( "SaveHypothesesToFile", '\0', &SaveHypothesesToFile, NULL ,
                       "File name for hypothesis output",
                       True );


  pParam->AddParamDef( "DistortionModel", '\0', &DistortionModelFile, NULL,
                       "Distortions used in decoding ",
                       Optional );

  pParam->AddParamDef( "DMSmoothingFactor", '\0', &DistortionModelSmoothingFactor, 0.0,
                       "Smoothing factor:  1 = use only background model, 0 = use word dependent model ",
                       Optional );

  pParam->AddParamDef( "DMScale", '\0', &DMScale, 1.0,
                       "Distortion model scaling factor, default = 1.0",
                       Optional );

  pParam->AddParamDef( "DistortionModel2", '\0', &DistortionModel2File, NULL,
                       "Distortions used in decoding ",
                       Optional );

  pParam->AddParamDef( "DM2SmoothingFactor", '\0', &DistortionModel2SmoothingFactor, 0.0,
                       "Smoothing factor:  1 = use only background model, 0 = use word dependent model ",
                       Optional );

  pParam->AddParamDef( "DM2Scale", 'a', &DM2Scale, 1.0,
                       "Distortion model scaling factor, default = 1.0",
                       Optional );


  pParam->AddParamDef( "DistortionModel3", '\0', &DistortionModel3File, NULL,
                       "Distortions used in decoding ",
                       Optional );

  pParam->AddParamDef( "DM3SmoothingFactor", '\0', &DistortionModel3SmoothingFactor, 0.0,
                       "Smoothing factor:  1 = use only background model, 0 = use word dependent model ",
                       Optional );

  pParam->AddParamDef( "DM3Scale", '\0', &DM3Scale, 1.0,
                       "Distortion model scaling factor, default = 1.0",
                       Optional );

  pParam->AddParamDef( "DistortionModel4", '\0', &DistortionModel4File, NULL,
                       "Distortions used in decoding ",
                       Optional );

  pParam->AddParamDef( "DM4SmoothingFactor", '\0', &DistortionModel4SmoothingFactor, 0.0,
                       "Smoothing factor:  1 = use only background model, 0 = use word dependent model ",
                       Optional );

  pParam->AddParamDef( "DM4Scale", '\0', &DM4Scale, 1.0,
                       "Distortion model scaling factor, default = 1.0",
                       Optional );


  pParam->AddParamDef( "FertilityFE", '\0' , &FertilityFileFE, NULL ,
                       "Name of file with fertilities for target(?) words: word fert0 ... fert4",
                       Optional );

  pParam->AddParamDef( "FertilityEF", '\0' , &FertilityFileEF, NULL ,
                       "Name of file with fertilities for source(?) words: word fert0 ... fert4",
                       True );
	 


  pParam->AddParamDef( "RescoreNBest", '\0', &RescoreNBest, 0,
                       "Rescore n-best list, default = False",
                       True );
  
  pParam->AddParamDef( "RescoreTransducer", '\0', &RescoreTransducer, 0,
                       "Rescore transducer, default = False",
                       True );
  

  pParam->AddParamDef( "ScalingFactors", '\0', &pScalingFactors, "1.0_1.0_-1.0_1.0_1.0",
                       "List of scaling factor for all models: LMs, DMs, WC, PC, TMs",
                       Optional );


  /*
    pParam->AddParamDef( "TMScaleList", '\0', &pTMScaleList, "0.0_0.0_0.0_0.0_1.0_0.0",
    "List of scaling factor for phrases",
    Optional );

    pParam->AddParamDef( "WCScale", '\0', &WCScale, 1.0,
    "Word count model scaling factor, default = 1.0",
    Optional );

    pParam->AddParamDef( "PCScale", '\0', &PCScale, 1.0,
    "Phrase count  model scaling factor, default = 1.0",
    Optional );
  */


  pParam->AddParamDef( "LengthOfReorderingWindow", 'L', &LengthOfReorderingWindow, 1,
                       "Length of window for reordering, default = 1, i.e. monoton",
                       Optional );
  
  pParam->AddParamDef( "ReorderingsPerWord", 'R', &ReorderingsPerWord, 1.0,
                       "Number of reorderings per word, default = 1, no restrictions",
                       Optional );
  
  pParam->AddParamDef( "CompareHypFeatures", '\0', &CompareHypFeatures, 7,
                       "Add: 1 - Total Coverage, 2 - Coverage Pattern, 4 - LM History, 8 - Target Sentence Length",
                       Optional );

  pParam->AddParamDef( "PruneHypFeatures", '\0', &PruneHypFeatures, 3,
                       "Add: 1 - Total Coverage, 2 - Coverage Pattern, 4 - LM History, 8 - Target Sentence Length",
                       Optional );

  pParam->AddParamDef( "BeamSize", 'b', &BeamSize, 10000,
                       "Beam size >= 1, Number of expanded recombined hyps within similarity class, the larger the less pruning",
                       Optional );

  pParam->AddParamDef( "MaxHypsPerCoverage", '\0', &MaxHypsPerCoverage, 10000000,
                       "Maximal number of hypotheses to be expanded per coverage",
                       Optional );

  pParam->AddParamDef( "MaxHyps", 'm', &MaxHyps, 100,
                       "??",
                       Optional );

  pParam->AddParamDef( "NBest", 'N', &NBest, 1,
                       "Size of n-best list, default = 1",
                       Optional );


  pParam->AddParamDef( "SA_HighestFrequency", '\0', &SA_HighestFrequency, 1000000,
                       "Consider only words with frequence up to this upper limit, default = 1000000",
                       Optional );

  pParam->AddParamDef( "SA_MaxOccurrences", '\0', &SA_MaxOccurrences, 100,
                       "Return at most that many occurences in the corpus: default = 100",
                       Optional );

  pParam->AddParamDef( "SA_ShortestPhraseLength", '\0', &SA_ShortestPhraseLength, 1,
                       "Consider only phrases equal or longer than this number, -1 = exact match",
                       Optional );
  
  pParam->AddParamDef( "SA_LongestPhraseLength", '\0', &SA_LongestPhraseLength, 1000,
                       "Consider only phrases shorter or equal than this number, default = 1000",
                       Optional );
  

  pParam->AddParamDef( "PhraseAlignMaxAlignments", '\0', &PhraseAlignMaxAlignments, 10,
                       "Maximal number of alignments per source phrase",
                       Optional );

  pParam->AddParamDef( "PhraseAlignBeamSize", '\0', &PhraseAlignBeamSize, 2.0,
                       "Beam size used in phrase alignment, default = 2.0",
                       Optional );

  pParam->AddParamDef( "PhraseAlignMaxTranslations", '\0', &PhraseAlignMaxTranslations, 10,
                       "Maximal number of translations per source phrase, default = 100",
                       Optional );

  pParam->AddParamDef( "PhraseAlignFindCenterMethod", '\0', &PhraseAlignFindCenterMethod, 0,
                       "Method to find expected center of target phrase: 0 = max prob, 1 = sum prob, 2 = block prob",
                       Optional );

  pParam->AddParamDef( "PhraseAlignStretchFactor", '\0', &PhraseAlignStretchFactor, 1.0,
                       "Factor to calculate expected target phrase length, default = 1.0",
                       Optional );

  pParam->AddParamDef( "PhraseAlignFactorShorter", '\0', &PhraseAlignFactorShorter, 0.5,
                       "Adjust minimum length of target phrase relative to source phrase, default = 0.5",
                       Optional );
  
  pParam->AddParamDef( "PhraseAlignFactorLonger", '\0', &PhraseAlignFactorLonger, 2.0,
                       "Adjust maximum length of target phrase relative to source phrase, default = 2.0",
                       True );
  
  pParam->AddParamDef( "PhraseAlignAddConstShorter", '\0', &PhraseAlignAddConstShorter, -2.0,
                       "Adjust minimum length of target phrase relative to source phrase, default = -2.0",
                       True );
  
  pParam->AddParamDef( "PhraseAlignAddConstLonger", '\0', &PhraseAlignAddConstLonger, 2.0,
                       "Adjust maximum length of target phrase relative to source phrase, default = 2.0",
                       True );
            
  pParam->AddParamDef( "PhraseAlignScaleList", '\0', &pPhraseAlignScaleList, "1.0_1.0_0.0_0.0_0.0_0.0",
                       "List of scaling factors for phrase alignment",
                       True );
  
  pParam->AddParamDef( "PhraseAlignDynamicFactors", '\0', &PhraseAlignDynamicFactors, 0.0,
                       "Interpolate phrase align scale factors in optimization with TM scale factors",
                       True );
  
  pParam->AddParamDef( "LatticePruneBeamSize", '\0', &LatticePruneBeamSize, 0.0,
                       "Beam size for lattice pruning when transducers are being applied",
                       True );
  
  pParam->AddParamDef( "LatticePruneBeamSizeEpsilon", '\0', &LatticePruneBeamSizeEpsilon, 0.0,
                       "Epsilon for multiplicative beam in lattice pruning when edge score = 0",
                       True );
    
  pParam->AddParamDef( "LatticePruneMaxTranslations", '\0', &LatticePruneMaxTranslations, 0,
                       "Max number of translation edges added between two adjacent nodes",
                       True );

  pParam->AddParamDef( "UseFutureCostEstimates", '\0', &UseFutureCostEstimates, 0,
                       "Use future cost estimates when PruneHypFeatures==1 ",
                       True );
  

  pParam->AddParamDef( "MEROptimize", '\0', &MEROptimize, 0,
                       "Runs iterations of Minimum Error rate training on the nbest lists generated in each iteration",
                       True );
  
  pParam->AddParamDef( "Opti_LeftRange", '\0', &OptimizationLowerBounds, "0.0_0.0_-10.0_0.0_0.0_0.0_0.0_0.0_0.0_0.0",
                       "Left range for MER optimization ",
                       True );
  
  pParam->AddParamDef( "Opti_RightRange", '\0', &OptimizationUpperBounds, "10.0_10.0_10.0_10.0_10.0_10.0_10.0_10.0_10.0_10.0",
                       "Right range for MER optimization ",
                       True );
  
  pParam->AddParamDef( "Opti_NumStarts", '\0', &OptimizationNumberOfStarts, 5,
                       "Number of times optimization restarts from new parameters ",
                       True );
  
  pParam->AddParamDef( "Opti_NumPermutes", '\0', &OptimizationNumberOfPermutations, 3,
                       "Number of time we randomly modify a few values after convergence ",
                       True );
  
  pParam->AddParamDef( "Opti_NumModify", '\0', &OptimizationNumberOfModifications, 2,
                       "Number of dimensions modified after convergence ",
                       True );
  
  pParam->AddParamDef( "Opti_Epsilon", '\0', &OptimizationEpsilon, 0.001,
                       "Difference in error to consider optimization converged ",
                       True );
  
  pParam->AddParamDef( "Opti_TwoStage", '\0', &OptiTwoStage, 0,
                       "Run 2 stage optimization - nulling out certain fields and re-optimizing. Dont use this option AND Opti_NumPermutes > 1",
                       True );

  pParam->AddParamDef( "Opti_DoNorm", '\0', &OptiDoNorm, 0,
                       "Normalize scaling factors after each iteration ",
                       Optional );

  pParam->AddParamDef( "Opti_EpsilonScale", '\0', &OptiEpsilonScale, 1.0,
                       "Scale up the threshold when using two stage optimization to save time",
                       Optional );

  pParam->AddParamDef( "ScoringMetric", '\0', &ScoringMetric, NULL,
                       "Scoring metric for model score ",
                       True );
  
  pParam->AddParamDef( "GainFile", '\0', &GainsFile, NULL,
                       "Only used for NIST scoring  ",
                       True );
  
  pParam->AddParamDef( "ReferenceList", '\0', &ReferenceList, NULL,
                       "Reference sentences for scoring ",
                       True );
  
  pParam->AddParamDef( "NumReferences", '\0', &NumReferences, 1,
                       "Number of references in data ",
                       True );
  
  pParam->AddParamDef( "GenOracleScores", '\0', &GenOracleScores, 0,
                       "Saves nbest lists across iterations for oracle",
                       True );
  
  pParam->AddParamDef( "IterationLimit", '\0', &IterationLimit, 5,
                       "Limit to the number of iterations that MER will be run for",
                       True );
  
  pParam->AddParamDef( "LoadUnique", '\0', &LoadUnique, 1,
                       "Load Only Unique elements into the n-best list ",
                       True );
  
  
  pParam->AddParamDef( "DisplayLattice", '\0', &DisplayLattice, 0,
                       "Display translation lattice ",
                       True );
  
  pParam->AddParamDef( "DisplayLatticeStatistics", '\0', &DisplayLatticeStatistics, 0,
                       "Display statistics about the translation lattice",
                       True );
  
  pParam->AddParamDef( "DisplayTranslations", '\0', &DisplayTranslations, 1,
                       "Display n-best translations before text normalization",
                       True );
  
  pParam->AddParamDef( "DisplayTranslationsNormalized", '\0', &DisplayTranslationsNormalized, 0,
                       "Display n-best translations after text normalization",
                       True );
  
  pParam->AddParamDef( "DisplayTranslationPath", '\0', &DisplayTranslationPath, 0,
                       "Display phrase information for translations",
                       True );
  
  pParam->AddParamDef( "DisplayNBestData", '\0', &DisplayNBestData, -1,
                       "Write out translation, model scores and path",
                       True );
  
  pParam->AddParamDef( "DisplayFeaturesRange", '\0', &DisplayFeaturesRange, 0,
                       "Display each feature score's min, max for each source sentence.",
                       True );
  
  // server for interactive mode
  pParam->AddParamDef( "InteractiveMode", '\0', &InteractiveMode, 0, 
                       "Read source sentences from a port",
                       True );
  pParam->AddParamDef( "ServerPort", '\0', &ServerPort, 4444,
                       "TranslateMM server port for interactive mode",
                       True );
  
  // Worker Bee params
  pParam->AddParamDef( "WorkerBeeDir", '\0', &WorkerBeeDir, NULL, 
                       "If specified we use the worker-bee mode. This is the directory in which Worker Bee files, like locks, per-sentence n-best lists and MER params are stored.",
                       True );
  pParam->AddParamDef( "WorkerBeeSaveSentenceFeatures", '\0', &WorkerBeeSaveSentenceFeatures, 0,
                       "1=save n-best hyp feature scores to .feats file.",
                       True );

  pParam->AddParamDef( "MaxSentenceProcessingTime", '\0', &MaxSentenceProcessingTime, 480,
                       "Maximum time in seconds that processing one sentence takes. Relevant only for Worker-bee mode: The worker bee will assume that a process got killed if nothing happened on a sentence for more than MaxSentenceProcessingTime seconds.",
                       True );
  
  // truecaser stuff
  char *pTrueCaserLMFile = NULL;
  char *pTrueCaserLMType = NULL;
  int   TrueCaserLMNgramLength;
  char  TrueCaserLMInterpolationMode;
  int   TrueCaserLMMaxFrequencyForDiscounting;

  pParam->AddParamDef( "TrueCaser.LMFile", '\0', &pTrueCaserLMFile, NULL,
                       "File from which the LM is loaded",
                       Optional );

  pParam->AddParamDef( "TrueCaser.LMType", '\0', &pTrueCaserLMType, "SALM",
                       "The kind of language model that you will use, currently: SriLM, SALM, LMWordSA",
                       Optional );

  pParam->AddParamDef( "TrueCaser.LMNgramLength", '\0', &TrueCaserLMNgramLength, 3,
                       "Length of ngrams in LM, default = 3",
                       Optional );
  
  pParam->AddParamDef( "TrueCaser.LMInterpolationMode", '\0', &TrueCaserLMInterpolationMode, 'e',
                       "Interpolation mode for Suffix Array LM: Equal (=e) or IBM style (=i), default = e",
                       Optional );
  
  pParam->AddParamDef( "TrueCaser.LMMaxFrequencyForDiscounting", '\0', &TrueCaserLMMaxFrequencyForDiscounting, -1,
                       "Good-Turing discounting applied in SA LM for ngrams with frequency of to this number",
                       Optional );
  


  if (! pParam->Interpret( argv, argc) ) {
    cerr << "\nParameters could not be interpreted correctly\n ... exiting!!\n";
    exit(1); 
  }




  cout << "Parameter Settings for this translation run:" << endl;
  cout << "--------------------------------------------" << endl;
  cout << endl;

  if( LatticeList ) {
    cout << "Test lattice set to translated    " << LatticeList << endl;
  } else if ( SentenceList ) {
    cout << "Test set to be translated         " << SentenceList << endl;
  } else if ( InteractiveMode ) {
    cout << "Entering Interactive Mode on port " << ServerPort << endl;
  } else {
    cout << "No test set provided\n ... exiting!!\n";
    exit(1);
  }
  cout << endl;
  
  //Nguyen, Aug 6, 2008: source-side dependency trees
  if ( SentenceDependencyTrees ){
  	CohesionCheck = true;
  	cout << "Source-side dependency trees      " << SentenceDependencyTrees << endl;
  	if ( InterruptionCheckAlgorithm ) { 
  		cout << "Interruption Check Algorithm      " << InterruptionCheckAlgorithm << endl;
  	} 
//  	else if ( InterruptionCheckAlgorithmCombination ){
//  		cout << "Interruption Check Algorithm Combination     " << InterruptionCheckAlgorithmCombination << endl;
//  	}
  }

  cout << "Models used: " << endl;
  cout << "Language model 1                  "; if ( pLMFile )  cout << pLMFile << endl; else cout << "NONE" << endl;
  cout << "Language model 2                  "; if ( pLM2File ) cout << pLM2File << endl; else cout << "NONE" << endl;
  cout << "Language model 3                  "; if ( pLM3File ) cout << pLM3File << endl; else cout << "NONE" << endl;
  cout << "Language model 4                  "; if ( pLM4File ) cout << pLM4File << endl; else cout << "NONE" << endl;

  cout << "Distortion model 1                "; if ( DistortionModelFile ) cout << DistortionModelFile << endl; else cout << "NONE" << endl;
  cout << "Distortion model 2                "; if ( DistortionModel2File ) cout << DistortionModel2File << endl; else cout << "NONE" << endl;
  cout << "Distortion model 3                "; if ( DistortionModel3File ) cout << DistortionModel3File << endl; else cout << "NONE" << endl;
  cout << "Distortion model 4                "; if ( DistortionModel4File ) cout << DistortionModel4File << endl; else cout << "NONE" << endl;

  cout << "Lexicon source-target             "; if ( LexiconFileFE ) cout << LexiconFileFE << endl; else cout << "NONE" << endl;
  cout << "Lexicon target-source             "; if ( LexiconFileEF ) cout << LexiconFileEF << endl; else cout << "NONE" << endl;
  cout << "Phrase table                      "; if ( pPhraseTableFile ) cout << pPhraseTableFile << endl; else cout << "NONE" << endl;
  cout << "Phrase table sentence-level       "; if ( pPhraseTableSentLevelFile ) cout << pPhraseTableSentLevelFile << endl; else cout << "NONE" << endl;
  cout << "Translation memory                "; if ( TransMemParams ) cout << TransMemParams << endl; else cout << "NONE" << endl;
  cout << "Fertility model source-target     "; if ( FertilityFileFE ) cout << FertilityFileFE << endl; else cout << "NONE" << endl;
  cout << "Fertility model target-source     "; if ( FertilityFileEF ) cout << FertilityFileEF << endl; else cout << "NONE" << endl;

  cout << endl;

  cout << "Model scaling factors:            " << pScalingFactors<< endl;

  //cout << "Language model scaling factor      " << LMScale << endl;
  //cout << "Translation model scaling factor   " << pTMScaleList << endl;
  //cout << "Distortion model scaling factor    " << DMScale << endl;
  //cout << "Word count model scaling factor    " << WCScale << endl;
  //cout << "Phrase count model scaling factor  " << PCScale << endl; 
  cout << endl;

  cout << "Size of reordering window          " << LengthOfReorderingWindow << endl;
  cout << "Reorderings per word               " << ReorderingsPerWord << endl;
  cout << endl;

  cout << "CompareHypFeatures                 " << CompareHypFeatures << endl;
  cout << "PruneHypFeatures                   " << PruneHypFeatures << endl;
  cout << "Beam size (num of hyps)            " << BeamSize << endl;
  cout << "Max. hyps per coverage             " << MaxHypsPerCoverage << endl;
  cout << "Max. hyps                          " << MaxHyps << endl;
  cout << "NBest                              " << NBest << endl;
  cout << endl;

  cout << "Parameter settings for phrase alignment:" << endl;
  cout << "Highest word frequency allowed     " << SA_HighestFrequency << endl;
  cout << "Maximum occurence considered       " << SA_MaxOccurrences << endl;
  cout << "Shortest source phrase             " << SA_ShortestPhraseLength << endl;
  cout << "Longest source phrase              " << SA_LongestPhraseLength << endl;
  cout << "Max. number of alignments          " << PhraseAlignMaxAlignments << endl;
  cout << "Beam size in phrase alignment      " << PhraseAlignBeamSize << endl;
  cout << "Max. number of translations        " << PhraseAlignMaxTranslations << endl;
  cout << "Method to find target center       " << PhraseAlignFindCenterMethod  << endl;
  cout << "Scaling factors                    " << pPhraseAlignScaleList << endl;
  cout << "Adjust scaling factores in opt.    " << PhraseAlignDynamicFactors << endl;
  cout << endl;

  cout << "LatticePrune beam size             " << LatticePruneBeamSize << endl;
  cout << "LatticePrune epsilon               " << LatticePruneBeamSizeEpsilon << endl;
  cout << "LatticePrune max tran              " << LatticePruneMaxTranslations << endl;
  
  cout << "Parameter settings for optimization:" << endl;
  cout << "Apply MER optimization             " << MEROptimize << endl;
  if ( MEROptimize == 1 ) {
    cout << "Initial model scaling factors      " << pScalingFactors<< endl;
    cout << "Optimziation lower bounds          " << *((char**)pParam->getValueFromKey( "Opti_LeftRange" )) << endl;
    cout << "Optimization upper bounds          " << *((char**)pParam->getValueFromKey( "Opti_RightRange" )) << endl;
  }
  cout << "Scoring metric                     " << (ScoringMetric == NULL ? "None" : ScoringMetric ) << endl;

  cout << "Max. number of iterations          " << IterationLimit << endl;
  cout << "File with information gains        " << (GainsFile == NULL ? "None" : GainsFile ) << endl;
  cout << "Text normalization script          " << (NormalizingScript == NULL ? "None" :NormalizingScript ) << endl;
  cout << "Text normalization regex file      " << (NormalizingRegexFileName == NULL ? 
                                                    "None" : NormalizingRegexFileName ) << endl;
  cout << "File with reference translations   " << (ReferenceList == NULL ? "None" : ReferenceList) << endl;
  cout << "Number of reference translations   " << NumReferences << endl;
  cout << "Display nbest translations         " << DisplayNBestData << endl;
  cout << "Generate oracle score              " << GenOracleScores << endl;
  cout << endl;

  cout << "WorkerBee directory                "; if ( WorkerBeeDir ) cout << WorkerBeeDir << endl; else cout << "NONE" << endl; 
  cout << "WorkerBeeSaveSentenceFeatures      " << WorkerBeeSaveSentenceFeatures << endl;
  cout << "MaxSentenceProcessingTime          " << MaxSentenceProcessingTime << endl << endl;
  
  cout << "Parameter setting for displaying information: " << endl;
  cout << "Display lattice                    " << DisplayLattice << endl;
  cout << "Display lattice statistics         " << LatticeStatistics << endl;
  cout << "Display NBest translations         " << DisplayNBestData << endl;
  cout << "Display translations unnormalized  " << DisplayTranslations << endl;
  cout << "Display translations normalized    " << DisplayTranslationsNormalized << endl;
  cout << "Display translation path           " << DisplayTranslationPath << endl;
  cout << "Display feature function min,max   " << DisplayFeaturesRange << endl;
  cout << endl << flush;



  // =========================================================================================
  // initialize TextFilter
  pTextFilter = InitTextFilter( *pParam );
  //assert( pTextFilter );

  pWordExpander = InitWordExpander( *pParam );


  // =========================================================================================
  cout << endl;
  cout << "Load model data" << endl;
  cout << "---------------" << endl;

  // vocabulary, combined for source and target words
  cout << "Load Vocabulary" << endl;
  pVoc = LoadVocabulary( *pParam, VocFile );
  if ( StringForEmptyWord ) {
    pVoc->SetStringForEmptyWord( StringForEmptyWord );
  }
  cout << endl;

  // language model
  cout << "Load Language Models" << endl;
  //pLM = LoadLanguageModel( *pParam, pLMType, pLMFile );
  LoadLanguageModels( *pParam, LMs );
  cout << endl;

  // phrase tables
  pTranslationMemory = LoadTranslationMemory( *pParam, pVoc );

  // load suffix array for online phrase alignment
  cout << "Load Suffix Array" << endl;
  pSA = LoadSuffixArray( *pParam, SourceFileSA );
  cout << endl;

  // load corpus, also needed for online phrase alignment
  cout << "Load Corpus" << endl;
  pTextF = LoadText( *pParam, SourceCorpus, pVoc );
  pTextE = LoadText( *pParam, TargetCorpus, pVoc );
  cout << endl;

  // lexicons, needen for online phrase alignment, also for rescoring
  cout << "Load Lexicons" << endl;
  pLexiconFE = LoadLexicon( *pParam, LexiconFileFE, pVoc );
  pLexiconEF = LoadLexicon( *pParam, LexiconFileEF, pVoc );
  cout << endl;

  // distortion model
  cout << "Load Distortion Models" << endl;
  //pDistortionModel = LoadDistortionModel( *pParam, DistortionModelFile );
  LoadDistortionModels( *pParam, DMs );

  // fertility models
  cout << "Load Fertility Tables" << endl;
  pFertilityFE = LoadFertilityModel( *pParam, FertilityFileFE );
  pFertilityEF = LoadFertilityModel( *pParam, FertilityFileEF );
  cout << endl;

  InitPhraseLengthModel( pFertilityFE, pFertilityEF );

  // =========================================================================================
  // create phrase aligner, rescorer, decoder

  // create and initialize the phrase searcher
  InitPhraseSearcher( pPhraseAlignScaleList );

  // create Lexicon Rescorer
  if ( LexiconFileFE ) {
    pLatticeRescorer = new CLatticeRescorer;
  }

  // n-best list rescorer
  if ( RescoreNBest ) {
    if ( pLexiconFE ) {
      pRescorerSourceTargetIBM1 = new CNBestListRescorerTM_IBM1;
      pRescorerSourceTargetIBM1->Init( pLexiconFE );
    } else {
      cout << "No source-target lexicon availabe, can not initialize the IBM1 rescorer for source to target" << endl;
    }
 
    if ( pLexiconEF ) {
      pRescorerTargetSourceIBM1 = new CNBestListRescorerTM_IBM1;
      pRescorerTargetSourceIBM1->Init( pLexiconEF );
    } else {
      cout << "No target-source lexicon availabe, can not initialize the IBM1 rescorer for target to source" << endl;
    }
      
  }

  // deal with the model scaling factors; there can be multiple TMs, LMs, DMs, SMs
  // scaling factors are provided as one connected string of parameters
  // need to be separated and put into one array
  //PrepareScalingFactors( LMScale, DMScale, WCScale, PCScale, pTMScaleList );
  PrepareScalingFactors( pScalingFactors );


  // create sentence to lattice converter;
  pSentenceToLattice = new CSentenceToLattice;


  // create and init the decoder
  cout << endl;
  cout << "Initialize Decoder" << endl;
  cout << "------------------" << endl;
  pDecoder = InitDecoder( *pParam );

  //cout << "Initialize TrueCaser" << endl;
  //pDecoder = InitTrueCaser( *pParam );

  // =========================================================================
  // finally, we are ready to go
  
  if ( LatticeList ) {

    if ( WorkerBeeDir ) {
      TranslateLatticesResumable();
    } else {
      TranslateLattices();      
    }

  } else if ( InteractiveMode ) {

      TranslateServerStart( ServerPort );
      
  } else {
    
    // load and translate sentences
    // sentences will be read one by one
    CText* pTest = LoadText( *pParam, SentenceList, pVoc, true );
    
        
    //Nguyen, Aug 6, 2008: source-side dependency
    if ( SentenceDependencyTrees ) {
    	
    	cout << "Loading source-side dependency trees ..." << endl; 
    	
    	//read dependency tree CoNLL text file into a 2D matrix   	
    	readCoNLLformat(SentenceDependencyTrees, dv);
    	
    	//print out source-side dependency trees
    	if (TraceLevel > 0){
    		for(vector<tokens>::size_type i = 0; i < dv.size(); ++i){
    			for(vector<tokens>::size_type j = 0; j < dv[i].size(); ++j){
    				int p = dv[i][j].parentpost;
    				string cp;
    				if (p == -1) {cp = "root";} else { cp = dv[i][p].name; }
    				
    				cout << dv[i][j].name << " | Post= " << dv[i][j].post << " | POS= " << dv[i][j].pos <<
    				" | ParentPost= " << dv[i][j].parentpost << " | Parent= " <<  cp <<
    				" | DType= " << dv[i][j].dtype << endl;
    			}
    			cout << "========" << endl;
    		}
    	}
    	
    }
    //end of Nguyen
    
    if ( WorkerBeeDir ) {
      // Resumable Worker Bee model translation
      TranslateTextResumable( pTest );
      
    } else {    
      TranslateText( pTest );
    }
    
    delete pTest;
  }
  
  pDecoder->Clear();

  delete pDecoder;

  delete pLexiconFE;
  delete pLexiconEF;
  delete pTranslationMemory;
  delete pPhraseSearcher;
  delete pPhraseAligner;
  delete pLM;
  delete pVoc;
  delete pParam;


  PrintTime( "End of execution " );


  cout<<"===================================" << endl;
  cout << "Translation done, exiting program" << endl;
  cout<<"===================================" << endl;

  return 0;
}




// ====================================================================
// Functions to load data
// ====================================================================

CVocabulary *
LoadVocabulary( CParam &Param, char * VocFile )
{
  CVocabulary *pVoc = new CVocabulary;

  pVoc->Allocate( 200000 );

  if (VocFile != NULL && strcmp(VocFile,"") != 0) {
    // vocfile specified
    //cout << "Loading vocabulary..." << endl;
    cout << "Load vocabulary from " << VocFile << endl;

    if ( !pVoc->Load( VocFile ) ) {
      cerr << "main(): Could not load " << VocFile;
      const char *source = pParam->getSourceFromKey( "Vocabulary" );
      if (source != NULL) {
        // param file associated to param
        cerr << ", referred to by param file " << source;
      } // endif: param file associated to parameter
      cerr<< ". Exiting." << endl << flush;
      exit(1);
    } // endif: error loading voc file
  } else {
    cout << "No vocabulary file specified." << endl;
  } // endif
  
  Trace( 3, "Display vocabulary" );
  if ( TraceLevel >= 3 ) {
    pVoc->Display();
  } 
  
  return pVoc;
}



void
LoadLanguageModels( CParam &Param, CNumericalArray< CLM * > &LMs )
{
  // get the information from the param object

  char *   pLMFile                      = *((char**)    pParam->getValueFromKey( "LMFile"                       ));
  char *   pLMType                      = *((char**)    pParam->getValueFromKey( "LMType"                       ));
  unsigned NGramLength                  = *((unsigned*) pParam->getValueFromKey( "LMNgramLength"                ));
  char     LMInterpolationMode          = *((char*)     pParam->getValueFromKey( "LMInterpolationMode"          ));
  int      LMMaxFrequencyForDiscounting = *((int*)      pParam->getValueFromKey( "LMMaxFrequencyForDiscounting" ));
  char *   pLMTagsConsideredAsWords     = *((char**)    pParam->getValueFromKey( "LMTagsConsideredAsWords"      ));
  int      LMUnkIsWord                  = *((int*)      pParam->getValueFromKey( "LMUnkIsWord"                  ));
  int      LMRestrictVocabulary         = *((int*)      pParam->getValueFromKey( "LMRestrictVocabulary"         ));

  // Bloom filter params
  int      BFSize                       = *((int*)      pParam->getValueFromKey( "BFSize"                       ));
	int      BFNumberOfHashfunctions      = *((int*)      pParam->getValueFromKey( "BFNumberOfHashfunctions"      ));

  
  CLM *pLM;
  int NumOfLM = 0;

  if ( pLMFile ) {
    pLM = LoadLanguageModel( pLMFile, pLMType, NGramLength, LMInterpolationMode, LMMaxFrequencyForDiscounting,
                             pLMTagsConsideredAsWords, LMUnkIsWord, LMRestrictVocabulary,
                             BFSize, BFNumberOfHashfunctions );
    if ( pLM ) {
      NumOfLM++;
      LMs[ NumOfLM ] = pLM;
    }
  }

  pLMFile                      = *((char**)    pParam->getValueFromKey( "LM2File"                       ));
  pLMType                      = *((char**)    pParam->getValueFromKey( "LM2Type"                       ));
  NGramLength                  = *((unsigned*) pParam->getValueFromKey( "LM2NgramLength"                ));
  LMInterpolationMode          = *((char*)     pParam->getValueFromKey( "LM2InterpolationMode"          ));
  LMMaxFrequencyForDiscounting = *((int*)      pParam->getValueFromKey( "LM2MaxFrequencyForDiscounting" ));
  pLMTagsConsideredAsWords     = *((char**)    pParam->getValueFromKey( "LM2TagsConsideredAsWords"      ));
  LMUnkIsWord                  = *((int*)      pParam->getValueFromKey( "LM2UnkIsWord"                  ));
  LMRestrictVocabulary         = *((int*)      pParam->getValueFromKey( "LM2RestrictVocabulary"         ));
  pLMTagsConsideredAsWords     = *((char**)    pParam->getValueFromKey( "LM2TagsConsideredAsWords"      ));
  LMUnkIsWord                  = *((int*)      pParam->getValueFromKey( "LM2UnkIsWord"                  ));
  LMRestrictVocabulary         = *((int*)      pParam->getValueFromKey( "LM2RestrictVocabulary"         ));

  // Bloom filter params
  BFSize                       = *((int*)      pParam->getValueFromKey( "BF2Size"                       ));
	BFNumberOfHashfunctions      = *((int*)      pParam->getValueFromKey( "BF2NumberOfHashfunctions"      ));

  if ( pLMFile ) {
    pLM = LoadLanguageModel( pLMFile, pLMType, NGramLength, LMInterpolationMode, LMMaxFrequencyForDiscounting,
                             pLMTagsConsideredAsWords, LMUnkIsWord, LMRestrictVocabulary,
                             BFSize, BFNumberOfHashfunctions );
    if ( pLM ) {
      NumOfLM++;
      LMs[ NumOfLM ] = pLM;
    }
  }

  pLMFile                      = *((char**)    pParam->getValueFromKey( "LM3File"                       ));
  pLMType                      = *((char**)    pParam->getValueFromKey( "LM3Type"                       ));
  NGramLength                  = *((unsigned*) pParam->getValueFromKey( "LM3NgramLength"                ));
  LMInterpolationMode          = *((char*)     pParam->getValueFromKey( "LM3InterpolationMode"          ));
  LMMaxFrequencyForDiscounting = *((int*)      pParam->getValueFromKey( "LM3MaxFrequencyForDiscounting" ));
  pLMTagsConsideredAsWords     = *((char**)    pParam->getValueFromKey( "LM3TagsConsideredAsWords"      ));
  LMUnkIsWord                  = *((int*)      pParam->getValueFromKey( "LM3UnkIsWord"                  ));
  LMRestrictVocabulary         = *((int*)      pParam->getValueFromKey( "LM3RestrictVocabulary"         ));

  // Bloom filter params
  BFSize                       = *((int*)      pParam->getValueFromKey( "BF3Size"                       ));
	BFNumberOfHashfunctions      = *((int*)      pParam->getValueFromKey( "BF3NumberOfHashfunctions"      ));
  
  if ( pLMFile ) {
    pLM = LoadLanguageModel( pLMFile, pLMType, NGramLength, LMInterpolationMode, LMMaxFrequencyForDiscounting,
                             pLMTagsConsideredAsWords, LMUnkIsWord, LMRestrictVocabulary,
                             BFSize, BFNumberOfHashfunctions );
    if ( pLM ) {
      NumOfLM++;
      LMs[ NumOfLM ] = pLM;
    }
  }

  pLMFile                      = *((char**)    pParam->getValueFromKey( "LM4File"                       ));
  pLMType                      = *((char**)    pParam->getValueFromKey( "LM4Type"                       ));
  NGramLength                  = *((unsigned*) pParam->getValueFromKey( "LM4NgramLength"                ));
  LMInterpolationMode          = *((char*)     pParam->getValueFromKey( "LM4InterpolationMode"          ));
  LMMaxFrequencyForDiscounting = *((int*)      pParam->getValueFromKey( "LM4MaxFrequencyForDiscounting" ));
  pLMTagsConsideredAsWords     = *((char**)    pParam->getValueFromKey( "LM4TagsConsideredAsWords"      ));
  LMUnkIsWord                  = *((int*)      pParam->getValueFromKey( "LM4UnkIsWord"                  ));
  LMRestrictVocabulary         = *((int*)      pParam->getValueFromKey( "LM4RestrictVocabulary"         ));

  // Bloom filter params
  BFSize                       = *((int*)      pParam->getValueFromKey( "BF4Size"                       ));
	BFNumberOfHashfunctions      = *((int*)      pParam->getValueFromKey( "BF4NumberOfHashfunctions"      ));

  if ( pLMFile ) {
    pLM = LoadLanguageModel( pLMFile, pLMType, NGramLength, LMInterpolationMode, LMMaxFrequencyForDiscounting,
                             pLMTagsConsideredAsWords, LMUnkIsWord, LMRestrictVocabulary,
                             BFSize, BFNumberOfHashfunctions );
    if ( pLM ) {
      NumOfLM++;
      LMs[ NumOfLM ] = pLM;
    }
  }

  if ( LMs.Size() == 0 ) {
    cout << "ERROR: No language model loaded, check your parameters or paramter file" << endl << flush;
    exit( -4 );
  }
  
}


CLM *
LoadLanguageModel( char *pFileName, char *pLMName, unsigned NGramLength, char LMInterpolationMode,
                   int LMMaxFrequencyForDiscounting, char *pTagsConsideredAsWords, int LMUnkIsWord,
                   int RestrictVocabulary, int BFSize, int BFNumberOfHashfunctions )
{
  // load language models
  // base language model used for translation
  CLM *pLM = NULL;

  if ( pFileName ) {
    cout << "Loading Language Model in TranslateMM ..." << endl;
    cout <<  "Load LM" << " "<< pLMName << " = " << pFileName << endl;

    if ( strcmp( pLMName, "LMWordSA" ) == 0 ) {
      cout << "Old suffix array LM is currently not supported" << endl;
      exit ( -7 );
      pLM = new CLMwordSA(  );

    } else if ( strcmp( pLMName, "SriLM" ) == 0 ) {
      //unsigned NGramLength = *((unsigned*)   pParam->getValueFromKey( "LMNgramOrder" ));

      if ( NGramLength > 10 ) {
        cout << "invalid LM n-gram order (too high), setting LM n-gram order 10." << endl;
        NGramLength = (unsigned)4;
      } else if ( NGramLength < 3 ) {
        cout << "invalid LM n-gram order (too low), setting LM n-gram order 3." << endl;
        NGramLength = (unsigned)3;
      }

      // cout << "Sri LM n-gram order " << NGramLength << endl;
      Trace( 1, "Loading SriLM" );
      if ( RestrictVocabulary )
        pLM = new SriLM( pFileName, NGramLength, *pVoc, (bool)RestrictVocabulary, (bool)LMUnkIsWord );
      else
        pLM = new SriLM( pFileName, NGramLength, (bool)LMUnkIsWord );

      assert( pLM );
      Trace( 1, "SriLM loaded");
      //cout << "Sri LM loaded and asserted" << endl;

    } else if ( strcmp( pLMName, "SA-LM" ) == 0 ) {

      cout << "Using Suffix Array Language Model" << endl;

      // create a SA LM, set NGramLength, and discounting parameter -- removed by apd 1/31/07
      //pLM = new C_SingleCorpusSALM( pFileName, NGramLength, LMMaxFrequencyForDiscounting );
      pLM = new C_SingleCorpusSALM( pFileName );

      // set interpolation method --- removed by apd 1/31/07
      // ((C_SingleCorpusSALM *)pLM)->setParam_interpolationStrategy( LMInterpolationMode );
      assert( pLM );
      cout << "SA LM loaded and asserted" << endl;
      //C_SingleCorpusSuffixArrayLM lm(argv[1], 2, 4);

    } else if ( strcmp( pLMName, "Class-SA-LM" ) == 0 ) {

      cout << "Using Class-based Suffix Array Language Model" << endl;
      cout << "Ngram Length         = " << NGramLength << endl;
      cout << "Interpolation Mode   = " << LMInterpolationMode << endl;
      cout << "Max Freq Discounting = " << LMMaxFrequencyForDiscounting << endl;


      // create a SA LM, set NGramLength, and discounting parameter
      pLM = new  ClassBasedSALM( pFileName, NGramLength, LMMaxFrequencyForDiscounting );

      // set interpolation method
      ((C_SingleCorpusSALM *)pLM)->setParam_interpolationStrategy( LMInterpolationMode );
      assert( pLM );
      cout << "SA LM loaded and asserted" << endl;
      //C_SingleCorpusSuffixArrayLM lm(argv[1], 2, 4);

    } else if ( strcmp( pLMName, "DB-LM" ) == 0 ) {

      cout << "Using DB Language Model" << endl;

      // create a DB LM. Pass config file
      pLM = new C_NgramDBLM( pFileName );
      assert( pLM );
      cout << "DB LM loaded and asserted" << endl;

    } else if ( strcmp( pLMName, "CDB-LM" ) == 0 ) {

      cout << "Using CharacterDB Language Model" << endl;

      // create a DB LM. Pass config file
      pLM = new C_CharacterDBLM( pFileName );
      assert( pLM );
      cout << "CDB LM loaded and asserted" << endl;

    }   else if ( strcmp( pLMName, "BFLM" ) == 0 ) {

      cout << "Using Bloom Filter Language Model" << endl;

      Trace( 1, "Loading BFLM" );
      pLM = new BFLM( pFileName, NGramLength, BFSize, BFNumberOfHashfunctions );

      assert( pLM );
      Trace( 1, "BFLM loaded");

    } else if ( strcmp( pLMName, "BoolBFLM" ) == 0 ) {

      cout << "Using Bool Bloom Filter Language Model" << endl;

      Trace( 1, "Loading BoolBFLM" );
      pLM = new BoolBFLM( pFileName, NGramLength, BFSize, BFNumberOfHashfunctions );

      assert( pLM );
      Trace( 1, "BoolBFLM loaded");

    } else if ( strcmp( pLMName, "DiscrimLM" ) == 0 ) {

      cout << "Using Discriminative Language Model : REMOVED" << endl;

//      Trace( 1, "Loading DiscrimLM" );
//      pLM = new C_DiscriminativeLM( pFileName );

      assert( pLM );
      Trace( 1, "DiscrimLM loaded");

    } else {
      cout << "ERROR: Unknown LM Type " << pLMName << endl << flush;
      exit( - 4 );
    }

  }

  return pLM;
}



CTranslationMemory * 
LoadTranslationMemory( CParam &Param, CVocabulary *pVoc )
{
	char * pTransMemParams          = *( (char**)pParam->getValueFromKey( "TransMemParams" )          );
	char * pPhraseTableFile         = *( (char**)pParam->getValueFromKey( "PhraseTableFileName" )     );
	int    PhraseTableMatchingMode  = *( (int*  )pParam->getValueFromKey( "PhraseTableMatchingMode" ) );

  float LatticePruneBeamSize         = *( (float*)pParam->getValueFromKey( "LatticePruneBeamSize" )  );
  float LatticePruneBeamSizeEpsilon  = *( (float*)pParam->getValueFromKey( "LatticePruneBeamSizeEpsilon" )  );
  int   LatticePruneMaxTranslations  = *( (int*)pParam->getValueFromKey( "LatticePruneMaxTranslations" )); 
  
  cout << "Load Phrase Tables" << endl;

  if ( pTransMemParams ) {
    cout << "Loading translation memory with Param-File " << pTransMemParams << endl;
    pTranslationMemory = new CTranslationMemory;
    pTranslationMemory->Init( pTransMemParams, pVoc );
    pTranslationMemory->RenormalizeTransducer();
  } else if ( pPhraseTableFile ) {
    cout << "Load one phrase table " << pPhraseTableFile << endl;
    pTranslationMemory = new CTranslationMemory;
    pTranslationMemory->Init( pVoc );
    pTranslationMemory->LoadTransducer( pPhraseTableFile );
  } else {
    cout << "No phrase table specified for loading" << endl;

    // empty pTranslationMemory is needed by sentence-level phrase table code
    pTranslationMemory = new CTranslationMemory;
    pTranslationMemory->Init( pVoc );
  }

  // set the phrase table matching mode, i.e. all edges in lattice or first edge only
  pTranslationMemory->SetMatchingMode( PhraseTableMatchingMode );
  
  pTranslationMemory->SetLatticePruning( LatticePruneBeamSize, 
                                         LatticePruneBeamSizeEpsilon,
                                         LatticePruneMaxTranslations );

  cout << endl;
  
  return pTranslationMemory;
}



CLexiconForTraining *
LoadLexicon( CParam &Param, char *LexiconFile, CVocabulary *pVoc )
{
	int LexiconFormat     = *((int*)pParam->getValueFromKey( "LexiconFormat" ));

  bool Result = false;

  // load Lexicon
  CLexiconForTraining *pLexicon = NULL;
  if ( LexiconFile ) {
    pLexicon = new CLexiconForTraining;
    pLexicon->Init( pVoc, pVoc );

    cout << "Loading lexicon from " << LexiconFile;

    switch ( LexiconFormat ) {
    case 0:
      cout << "  Format is STTK binary" << endl << flush;
      Result = pLexicon->Load( LexiconFile );
      break;
      
    case 1:
      cout << "  Format is STTK with words" << endl << flush;
      Result = pLexicon->LoadWithWords( LexiconFile );
      break;

    case 2:
      cout << "  Format is GIZA" << endl << flush;
      Result = pLexicon->LoadWithWordsGIZA( LexiconFile );
      break;

    default:
      cout << "ERROR: Unknow lexicon format requested, exit!" << endl << flush;
      exit( -8 );
      break;
    }

    if ( Result = false ) {
      cout << "Error loading lexicon from " << LexiconFile << ". Exiting.\n";
      exit( 1 );
    } // endif: error loading lexicon
    //	pLexicon->Log();
  }

  return pLexicon;
}



CText *
LoadText( CParam &Param, char *FileName, CVocabulary *pVoc, bool ReadOneByOne )
{
  // load target corpus
  CText *pText = new CText;
  pText->Allocate( 50000 );
  pText->SetVocabulary( pVoc );
  pText->ReadOneByOne( ReadOneByOne );
  
  // load target corpus
  if ( FileName ) {
    cout << "Loading text corpus from " << FileName << endl;

    //if (pText->LoadAsciiText( FileName ) <= 0) {
    if (pText->LoadAsciiText( FileName ) < 0) {
      cerr << "Error loading text from " << FileName << ". Exiting.\n";
      exit(1);
    }
  }
  return pText;
}


CNumericalArray< CLatticeLM* >*
LoadLattices( CParam &Param, char *LatticeList, char *LatticeFormat, CVocabulary *pVoc)
{
  
  CNumericalArray< CLatticeLM* >* pTestLattices = new CNumericalArray< CLatticeLM* >;
  
  Trace( 0, "Lattice-List: " << LatticeList << endl );
  
  ifstream in( LatticeList );
  char buffer[ 200 ];
  
  if (!in) {
    cerr << "main(): Couldn't find '" << LatticeList <<"', exiting" << endl << flush;
    exit(1);
  } // endif: error opening input stream
  
  in.getline( buffer, 200 );
  //Trace(0, "Buffer: " << buffer << endl << flush);
  
  //int i = 0;

  // create lattice edge prototype, which is given to the lattices
  CLatticeEdgeMS *pEdgePrototype = new CLatticeEdgeMS;
  
  while ( !in.eof() ) {
    CLatticeLM *pLattice = new CLatticeLM( pEdgePrototype, NULL );
    
    // set lattice type
    if (strcmp(LatticeFormat, "ASR") == 0) {
      pLattice->SetLatticeFileFormat(CLattice::ASR);
    } else if (strcmp(LatticeFormat, "SRI") == 0) {
      pLattice->SetLatticeFileFormat(CLattice::SRI);
    } else if (strcmp(LatticeFormat, "ASR-MS") == 0) {
      pLattice->SetLatticeFileFormat( CLattice::ASRMS );
    } else if (strcmp(LatticeFormat, "VM") == 0) {        
      // read in same way as ASR
      pLattice->SetLatticeFileFormat(CLattice::ASR);
    } else {
      // invalid lattice format given
      cout << "main(): Invalid lattice format given: '" << LatticeFormat 
           << "'. Options are SRI, ASR, ASR-MS, VM. Exiting." << endl << flush;
      cerr << "main(): Invalid lattice format given: '" << LatticeFormat
           << "'. Options are SRI, ASR, ASR-MS, VM. Exiting." << endl << flush;
      exit(1);
    }

    //i++;
    
    Trace( 1, "Start reading lattice from file" );
    pLattice->ReadFromFile( buffer );
    pTestLattices->Append( pLattice );
    //pLattice->RemoveSpecialEdges();

    /*
        Trace( 1, "Start search phrases in lattice" );
        //TraceLevel = 1;
     
        Translation = pDecoder->TranslateLattice( pLattice );
        Trace( 1, "Translation = " << Translation );
     
        if ( LatticeStatistics ) {
        pLattice->Display();
        DisplayLatticeStatistics( pLattice );
        }
        delete pLattice;
    */
     
    
    in.getline( buffer, 200 );
    // Already at end?
    
    if (!in.eof()) {
      Trace(0, "Buffer: " << buffer << endl << flush);
    } // endif: end of file
    
  } // endwhile: read Lattice-In
  
  delete pEdgePrototype;

  cout << "Loaded Lattices: " << pTestLattices->Size() << endl;
  return pTestLattices;
}





C_SuffixArraySearch *
LoadSuffixArray( CParam &Param, char *SourceFileSA )
{
  // load source corpus, with corresponding suffix array and vocabulary
  
  if ( SourceFileSA ) {
    cout << "Loading source corpus with suffix array from " << SourceFileSA << endl;
    C_SuffixArraySearch *pSA = new C_SuffixArraySearch();
    pSA->loadData_v6( SourceFileSA, false, false, false);

    int SA_HighestFrequency     = *( (int*)pParam->getValueFromKey( "SA_HighestFrequency" )      );
    int SA_MaxOccurrences       = *( (int*)pParam->getValueFromKey( "SA_MaxOccurrences" )        );
    int SA_ShortestPhraseLength = *( (int*)pParam->getValueFromKey( "SA_ShortestPhraseLength" )  );
    int SA_LongestPhraseLength  = *( (int*)pParam->getValueFromKey( "SA_LongestPhraseLength" )   );
    
    //pSA->setParameters( SA_HighestFrequency, SA_MaxOccurrences, SA_ShortestPhraseLength );

    pSA->setParam_reportMaxOccurrenceOfOneNgram( SA_MaxOccurrences );
    pSA->setParam_highestFreqThresholdForReport( SA_HighestFrequency );
    pSA->setParam_shortestUnitToReport(SA_ShortestPhraseLength );
    pSA->setParam_longestUnitToReport( -1 );

    return pSA;

    cout << "Source corpus with suffix array loaded" << endl << flush;

  } else {
    return NULL;
  }
}




void
LoadDistortionModels( CParam &Param, CNumericalArray< CDistortionModel * > &DMs )
{
  // get the information from the param object

  cout << "Load distortion models:" << endl;

  char *  pDMFile           = *((char**)pParam->getValueFromKey( "DistortionModel"   ));
  float   DMSmoothingFactor = *((float*)pParam->getValueFromKey( "DMSmoothingFactor" ));
  CDistortionModel *pDM = NULL;

  if ( pDMFile )
    DMs[ 1 ] = LoadDistortionModel( pDMFile, DMSmoothingFactor );

  pDMFile           = *((char**)pParam->getValueFromKey( "DistortionModel2"   ));
  DMSmoothingFactor = *((float*)pParam->getValueFromKey( "DM2SmoothingFactor" ));

  if ( pDMFile )
    DMs[ 2 ] = LoadDistortionModel( pDMFile, DMSmoothingFactor );

  pDMFile           = *((char**)pParam->getValueFromKey( "DistortionModel3"            ));
  DMSmoothingFactor = *((float*)pParam->getValueFromKey( "DM3SmoothingFactor" ));

  if ( pDMFile )
    DMs[ 3 ] = LoadDistortionModel( pDMFile, DMSmoothingFactor );

  pDMFile           = *((char**)pParam->getValueFromKey( "DistortionModel4"            ));
  DMSmoothingFactor = *((float*)pParam->getValueFromKey( "DM4SmoothingFactor" ));

  if ( pDMFile )
    DMs[ 4 ] = LoadDistortionModel( pDMFile, DMSmoothingFactor );

  cout << "DMs loaded " << DMs.Size() << endl;
}



CDistortionModel *
LoadDistortionModel( char *pDistortionModelFile, float DMSmoothingFactor )
{

  CDistortionModel *pDistortionModel = NULL;

  if( pDistortionModelFile != NULL ) {
    cout << "Loading distortion model " << pDistortionModelFile << endl;
    pDistortionModel = new CDistortionModelJump( pVoc );
    cout << "Distortion model smooting factor = " << DMSmoothingFactor << endl;
    pDistortionModel->SetSmoothingFactor( DMSmoothingFactor );
    pDistortionModel->Load( pDistortionModelFile );
    cout << "Done" << endl;
  } else {
    cout << "No distortion model specified" << endl;
  }

  return pDistortionModel;
}



// old function to load just one DM
CDistortionModel *
LoadDistortionModel( CParam &Param, char *pDistortionModelFile )
{
  float SmoothingFactor = *( (float*)pParam->getValueFromKey( "DistortionModelSmoothingFactor" ) );

  CDistortionModel *pDistortionModel = NULL;

  if( pDistortionModelFile != NULL ) {
    cout << "Loading distortion model " << pDistortionModelFile << endl;
    pDistortionModel = new CDistortionModelJump( pVoc );
    cout << "Distortion model smooting factor = " << SmoothingFactor << endl;
    pDistortionModel->SetSmoothingFactor( SmoothingFactor );
    pDistortionModel->Load( pDistortionModelFile );
    cout << "Done" << endl;
  } else {
    cout << "No distortion model specified" << endl;
  }

  return pDistortionModel;
}



CFertilityModel *
LoadFertilityModel( CParam &Param, char * FertilityFile )
{
  //Loads fertilility table (from Giza training)
  //F-Table is word based, and the current code requires that its is 
  //trained with a limit of top 5 probabilities
  //unseen words are given a uniform probability (1/5) for each possible length
  //5 includes the NULL option

  CFertilityModel *pFertility = NULL;
  if ( FertilityFile != NULL ) {
    pFertility = new CFertilityModel;
    pFertility->Init( pVoc );
    cout << "Loading fertility table from " << FertilityFile << endl;
    if ( !pFertility->Load( FertilityFile ) ) {
      cerr << "main() : cannot load fertility file. Exiting" << endl;
      exit(1);
    }
  }

  return pFertility;
}




// ====================================================================
// Init functions
// ====================================================================


// is this the old one???
void
PrepareScalingFactors( float LMScale, float DMScale, float WCScale, float PLScale, char *TMScaleList )
{
  cout << "Prepare Translation Model Scaling Factors:  " << TMScaleList << endl << flush;

  // vector<double> ScalingFactors is global

  ScalingFactors.push_back( LMScale );  // reaches accross phrases
  ScalingFactors.push_back( DMScale );  // reaches accross phrases
  ScalingFactors.push_back( WCScale );  // can be treated as phrase internal score
  ScalingFactors.push_back( PLScale );  // can be treated as phrase internal score
  
  //read the space separated arguments of the TMScaleList
  // strtok modifies string, therefore work on copy
  char *TMScalingFactor = strdup( TMScaleList );
  TMScalingFactor = strtok( TMScalingFactor, "_" );
  while ( TMScalingFactor != NULL ) {
    cout << "  TM scaling factor " << TMScalingFactor << endl << flush;
    ScalingFactors.push_back( atof( TMScalingFactor ) );
    TMScalingFactor = strtok(NULL, "_");
  }
  //strdup returns a new'ed pointer, it needs to be freed
  free( TMScalingFactor );

  cout << endl;
}



void
PrepareScalingFactors( char *pScaleList )
{
  cout << "Prepare Scaling Factors:  " << pScaleList << endl << flush;

  // vector<double> ScalingFactors is global
  
  //read the underscore separated arguments of the ScaleList
  // strtok modifies string, therefore work on copy
  char *pScalingFactor = strdup( pScaleList );
  pScalingFactor = strtok( pScalingFactor, "_" );
  while ( pScalingFactor != NULL ) {
    cout << "  Scaling factor " << pScalingFactor << endl << flush;
    ScalingFactors.push_back( atof( pScalingFactor ) );
    pScalingFactor = strtok(NULL, "_");
  }
  //strdup returns a new'ed pointer, it needs to be freed
  free( pScalingFactor );

  cout << endl;
}



void
InitPhraseSearcher( char *pPhraseAlignScaleList )
{

  int   PhraseAlignMaxAlignments    = *( (int*)  pParam->getValueFromKey( "PhraseAlignMaxAlignments" )    );
  float PhraseAlignBeamSize         = *( (float*)pParam->getValueFromKey( "PhraseAlignBeamSize" )         );
  int   PhraseAlignMaxTranslations  = *( (int*)  pParam->getValueFromKey( "PhraseAlignMaxTranslations" )  );
  int   PhraseAlignFindCenterMethod = *( (int*)  pParam->getValueFromKey( "PhraseAlignFindCenterMethod" ) );
  float PhraseAlignStretchFactor    = *( (float*)pParam->getValueFromKey( "PhraseAlignStretchFactor" )    );
  float PhraseAlignFactorShorter    = *( (float*)pParam->getValueFromKey( "PhraseAlignFactorShorter" )    );
  float PhraseAlignFactorLonger     = *( (float*)pParam->getValueFromKey( "PhraseAlignFactorLonger" )     );
  float PhraseAlignAddConstShorter  = *( (float*)pParam->getValueFromKey( "PhraseAlignAddConstShorter" )  );
  float PhraseAlignAddConstLonger   = *( (float*)pParam->getValueFromKey( "PhraseAlignAddConstLonger" )   );

  cout << "Init Phrase Search:           " << endl;
  cout << "  Stretch Factor            = " << PhraseAlignStretchFactor << endl;
  cout << "  Factor Shorter            = " << PhraseAlignFactorShorter << endl;
  cout << "  Factor Longer             = " << PhraseAlignFactorLonger << endl;
  cout << "  Additive Cosntant Shorter = " << PhraseAlignAddConstShorter << endl;
  cout << "  Additive Constant Longer  = " << PhraseAlignAddConstLonger << endl;



  //create a phrase aligner and initialize it => give it to phrase Searcher
  pPhraseAligner = new CPhraseAlignerMS;
  pPhraseAligner->SetFindCenterMethod( PhraseAlignFindCenterMethod );
  //pPhraseAligner->SetStretchFactors( PhraseAlignStretchFactor, PhraseAlignFactorShorter, PhraseAlignFactorLonger );
  pPhraseAligner->SetStretchFactors( PhraseAlignStretchFactor, PhraseAlignFactorShorter, PhraseAlignFactorLonger,
                                     PhraseAlignAddConstShorter, PhraseAlignAddConstLonger );

  // sv 060318 outdated: now it is allways bi-directional alignment
  /*
    if( pLexiconEF ) {
    Trace( 0, "CPhraseSearchMS::Init with two lexicons" );
    pPhraseAligner->Init( pLexiconFE, pLexiconEF );
    pPhraseAligner->SetAlignBidirectional( true );
    } else {
    Trace( 0, "CPhraseSearchMS::Init with one lexicons" );
    pPhraseAligner->Init( pLexiconFE, NULL );
    pPhraseAligner->SetAlignBidirectional( false );
    }
  */

  CPhraseAlignProbCalcIBM1 *pCalcProbs = new CPhraseAlignProbCalcIBM1;

  pPhraseAligner->Init( pLexiconFE, pLexiconEF, pCalcProbs );

  // create a phrase searcher and initialize it
  pPhraseSearcher = new CPhraseSearchMS;
  //tell the PhraseSearcher, which PhraseAligner it should use
  pPhraseSearcher->SetPhraseAligner( pPhraseAligner );
  // sv: !!! LambdaLexicon no longer needed, should be eliminated, for now, dummy value
  float LambdaLexicon = 0.5;
  pPhraseSearcher->Init( pSA, pTextF, pTextE, pVoc, NULL, pLexiconFE, pLexiconEF, LambdaLexicon );
  //pPhraseSearcher->Init( pSA, pTextF, pTextE, pVoc, NULL, pLexiconFE, pLexiconEF, LambdaLexicon, pPhraseAligner );

  pPhraseSearcher->SetMaxAlignments( PhraseAlignMaxAlignments );
  pPhraseSearcher->SetBeam( PhraseAlignBeamSize , PhraseAlignMaxTranslations );


  // split the string which contains the phrase align scaling factors
  // and give those scaling factors to the phrase searcher
  // (currently, opperate with 6 scores, needs to be made flexible)
  CNumericalArray< float > PhraseAlignFactors;
  //vector< float > PhraseAlignFactors;

  cout << "  Phrase align scaling factors " << pPhraseAlignScaleList << endl << flush;

  // strtok modifies string, therefore work on copy
  char* PhraseAlignFactor = strdup( pPhraseAlignScaleList );
  int i = 1;
  PhraseAlignFactor = strtok( PhraseAlignFactor, "_" );
  while ( PhraseAlignFactor != NULL ){
    //cout << "Token is " << OneToken << endl << flush;
    PhraseAlignFactors[ i ] = atof( PhraseAlignFactor );
    //cout << "Phrase Align Factor " << i << " is: " << PhraseAlignFactors[ i ] << endl;
    PhraseAlignFactor = strtok( NULL, "_" );
    i++;
  }

  if ( PhraseAlignFactors.MaxIndex() < pPhraseAligner->GetNumberOfScores() ) {
    cout << "Warning, setting phrase align scaling factors == 0 for last "
         << pPhraseAligner->GetNumberOfScores() - PhraseAlignFactors.MaxIndex() << " factors" << endl;
    PhraseAlignFactors.AssureBoundaries( 1, pPhraseAligner->GetNumberOfScores() );
  }
  if ( PhraseAlignFactors.MaxIndex() > pPhraseAligner->GetNumberOfScores() ) {
    cout << "Warning: additional scaling factors set for phrase extraction, only first "
         << pPhraseAligner->GetNumberOfScores() << " will be used" << endl;
    PhraseAlignFactors.Remove( pPhraseAligner->GetNumberOfScores() + 1,
                               PhraseAlignFactors.MaxIndex() - pPhraseAligner->GetNumberOfScores() );
  }
  
  pPhraseAligner->SetScalingFactors( PhraseAlignFactors );
  
  //pPhraseAligner->SetFindCenterWithSum( false );
  pPhraseAligner->SetFindCenterMethod( 2 );

  cout << endl;

}




void
InitPhraseLengthModel( CFertilityModel *pFertilityFE, CFertilityModel *pFertilityEF )
{
  if ( pFertilityFE ) {
    pPhraseLengthModelFE = new CSentenceLengthModel;
    pPhraseLengthModelFE->Init( pVoc, pFertilityFE );
  }

  if ( pFertilityEF ) {
    pPhraseLengthModelEF = new CSentenceLengthModel;
    pPhraseLengthModelEF->Init( pVoc, pFertilityEF );
  }

}




CDecoderBLM *
InitDecoder( CParam &Param )
{
  int    LengthOfReorderingWindow = *( (int*)   pParam->getValueFromKey( "LengthOfReorderingWindow" ) );
  double ReorderingsPerWord       = *( (double*)pParam->getValueFromKey( "ReorderingsPerWord" )       );
  int    NBest                    = *( (int*)   pParam->getValueFromKey( "NBest" )                    );
  int    BeamSize                 = *( (int*)   pParam->getValueFromKey( "BeamSize" )                 );
  int    MaxHypsPerCoverage       = *( (int*)   pParam->getValueFromKey( "MaxHypsPerCoverage" )       );
  int    MaxHyps                  = *( (int*)   pParam->getValueFromKey( "MaxHyps" )                  );
  int    CompareHypFeatures       = *( (int*)   pParam->getValueFromKey( "CompareHypFeatures" )       );
  int    PruneHypFeatures         = *( (int*)   pParam->getValueFromKey( "PruneHypFeatures" )         );
  char  *pTagsConsideredAsWords   = *( (char**) pParam->getValueFromKey( "LMTagsConsideredAsWords" )  );
  int    LMBatchMode              = *( (int*)   pParam->getValueFromKey( "LMBatchMode" )              );

  //int PruneHypFeatures = 3;

  // and now the decoder object
  pDecoder = new CDecoderBLM;

  // set the models
  assert( pTranslationMemory );
  
  //pLM = LMs[ 1 ];
  //assert( pLM );

  //pDistortionModel = DMs[ 1 ];

  pDecoder->SetVocabulary( pVoc );
  //pDecoder->SetLanguageModel( pLM );
  //pDecoder->SetDistortionModel( pDistortionModel );

  // important:  set LMs before DMs as the position of the models scores
  // in the model scores arrays depends on this order
  pDecoder->SetLanguageModels( LMs );
  pDecoder->SetDistortionModels( DMs );

  pDecoder->SetTranslationMemory( pTranslationMemory );
  
  //Nguyen, Oct 1, 2008
  cout << "DEBUG InterruptionCheckAlgorithm = " << InterruptionCheckAlgorithm << endl;
  if ( InterruptionCheckAlgorithm == 5 ){ // cohesion decoding with multiple constraints  	
  	pDecoder->SetStartPointsOfScalingFactorsWithCohesionCombination();
  }
  else if ( InterruptionCheckAlgorithm == 0 ) { // no cohesion decoding
  	pDecoder->SetStartPointsOfScalingFactors();
  } 
  else { // cohesion decoding with 1 constraints  	
  	pDecoder->SetStartPointsOfScalingFactorsWithCohesion();
  } 
  
  
  
  pDecoder->SetScalingFactors( ScalingFactors );


  pDecoder->SetLengthOfReorderingWindow( LengthOfReorderingWindow );
  pDecoder->SetReorderingsPerWord( ReorderingsPerWord );
  pDecoder->SetNumOfNBest( NBest );
  pDecoder->SetBeamSize( BeamSize );
  pDecoder->SetMaxHypsPerCoverage( MaxHypsPerCoverage );
  pDecoder->SetMaxHyps( MaxHyps );
  pDecoder->SetLMBatchMode( LMBatchMode );

  //pDecoder->SetRescore( Rescore );
  //pDecoder->SetRescorer( pRescorer );
  //cout << "Compare Hyp Features " << CompareHypFeatures << endl;

  pDecoder->SetCompareHypFeatures( CompareHypFeatures & 1, CompareHypFeatures & 2,
                                   CompareHypFeatures & 4, CompareHypFeatures & 8 );

  pDecoder->SetPruneHypFeatures( PruneHypFeatures & 1, PruneHypFeatures & 2,
                                 PruneHypFeatures & 4, PruneHypFeatures & 8 );

  pDecoder->SetTagsConsideredAsWords( pTagsConsideredAsWords, ';' );

  return pDecoder;
}



CDecoderBLM *
InitTrueCaser( CParam &Param )
{
  int    LengthOfReorderingWindow = 1; // monotone decoding
  double ReorderingsPerWord       = 0.0;
  int    NBest                    = 1;
  //int    MaxHypsPerCoverage       = 10000000;
  //int    MaxHyps                  = 10000000;
  //int    CompareHypFeatures       = *( (int*)   pParam->getValueFromKey( "CompareHypFeatures" )       );
  int    CompareHypFeatures       = 5;  // number of words plus LM
  //int    PruneHypFeatures         = *( (int*)   pParam->getValueFromKey( "PruneHypFeatures" )         );
  int    PruneHypFeatures         = 1;
  //int    BeamSize                 = *( (int*)   pParam->getValueFromKey( "TrueCaser.BeamSize" )       );
  int    BeamSize                 = 20;

  char *   pLMFile                      = *((char**)    pParam->getValueFromKey( "TrueCaser.LMFile"                       ));
  char *   pLMType                      = *((char**)    pParam->getValueFromKey( "TrueCaser.LMType"                       ));
  unsigned NGramLength                  = *((unsigned*) pParam->getValueFromKey( "TrueCaser.LMNgramLength"                ));
  char     LMInterpolationMode          = *((char*)     pParam->getValueFromKey( "TrueCaser.LMInterpolationMode"          ));
  int      LMMaxFrequencyForDiscounting = *((int*)      pParam->getValueFromKey( "TrueCaser.LMMaxFrequencyForDiscounting" ));

	char *   pPhraseTableFile         = *( (char**)pParam->getValueFromKey( "TrueCaser.PhraseTableFileName" )     );
	int      PhraseTableMatchingMode  = *( (int*  )pParam->getValueFromKey( "TrueCaser.PhraseTableMatchingMode" ) );

  // and now the decoder object
  pDecoder = new CDecoderBLM;

  // set the models
  assert( pTranslationMemory );

  //pLM = LMs[ 1 ];
  //assert( pLM );

  //pDistortionModel = DMs[ 1 ];

  pDecoder->SetVocabulary( pVoc );
  //pDecoder->SetLanguageModel( pLM );
  //pDecoder->SetDistortionModel( pDistortionModel );

  // important:  set LMs before DMs as the position of the models scores
  // in the model scores arrays depends on this order
  pDecoder->SetLanguageModels( LMs );
  pDecoder->SetDistortionModels( DMs );

  pDecoder->SetTranslationMemory( pTranslationMemory );

  pDecoder->SetScalingFactors( ScalingFactors );
  pDecoder->SetStartPointsOfScalingFactors();

  pDecoder->SetLengthOfReorderingWindow( LengthOfReorderingWindow );
  pDecoder->SetReorderingsPerWord( ReorderingsPerWord );
  pDecoder->SetNumOfNBest( NBest );
  pDecoder->SetBeamSize( BeamSize );
  //pDecoder->SetMaxHypsPerCoverage( MaxHypsPerCoverage );
  //pDecoder->SetMaxHyps( MaxHyps );
  //  pDecoder->SetLMBatchMode( LMBatchMode );

  //pDecoder->SetRescore( Rescore );
  //pDecoder->SetRescorer( pRescorer );
  //cout << "Compare Hyp Features " << CompareHypFeatures << endl;

  pDecoder->SetCompareHypFeatures( CompareHypFeatures & 1, CompareHypFeatures & 2,
                                   CompareHypFeatures & 4, CompareHypFeatures & 8 );

  pDecoder->SetPruneHypFeatures( PruneHypFeatures & 1, PruneHypFeatures & 2,
                                 PruneHypFeatures & 4, PruneHypFeatures & 8 );

  //pDecoder->SetTagsConsideredAsWords( pTagsConsideredAsWords, ';' );


  CTranslationMemory *pTranslationMemory = NULL;

  if ( pPhraseTableFile ) {
    cout << "Load phrase table for true-caser" << pPhraseTableFile << endl;
    pTranslationMemory = new CTranslationMemory;
    pTranslationMemory->Init( pVoc );
    pTranslationMemory->LoadTransducer( pPhraseTableFile );
  } else {
    cout << "No phrase table specified for true-caser" << endl;
  }

  // set the phrase table matching mode, i.e. all edges in lattice or first edge only
  //pTranslationMemory->SetMatchingMode( PhraseTableMatchingMode );

  //cout << endl;
  
  return pDecoder;
}



void 
InitSourceWordModels( CNumericalArray< double > &srcFeatures, SriLM *pLM, int numModels )
{
  //numModels is total number of models used

  //default LM score
  srcFeatures.Append(-log10(1.0/pLM->VocabularySize()));

  //default DM score = 0.0 since this is assigned due to reordering
  srcFeatures.Append(0.0);

  //default WC model = 1.0 since one additional target word gets generated
  srcFeatures.Append(1.0);

  //default PL model = 1.0 since one additional phrase is generated
  srcFeatures.Append(1.0);

  // default TM scores for source words, i.e. untranslated word
  // sv: ??? unclear what effect this has, and if this is the right way to do it
  for (int i=4; i<numModels; i++){
    srcFeatures.Append(10.0);
  }

}



void
InitFutureCostEstimator( CSentence *pPartialSentence, CLattice *pLattice )
{
  //ARV 07/11 future cost estimates
  /*
    if (PruneHypFeatures==1 && UseFutureCostEstimates==1){
    futures.Init(pPartialSentence, pLattice);
    //futures.Display();
    futures.ParseMinCost();
    //futures.Display();
    }
  */
}


/*
  COptimizer *
  InitOptimizer( CParam &Param )
  {
  // first, we need to access the parameters for optimization
  char *ReferenceList = *((char**)pParam->getValueFromKey( "ReferenceList" ));
  cout << "Reference List is:       " << ReferenceList << endl;

  int NumReferences = *((int*)pParam->getValueFromKey( "NumReferences" ));
  cout << "Number of References is: " << NumReferences << endl;

  char *MatlabCommand = *((char**)pParam->getValueFromKey( "MatlabCommand" ));
  cout << "Matlab Command is:       " << MatlabCommand << endl;

  char *NormalizingScript = *((char**)pParam->getValueFromKey( "NormalizingScript" ));
  cout << "Normalizing Script is:   " << NormalizingScript << endl;

  char *ScoringMetric = *((char**)pParam->getValueFromKey( "ScoringMetric" ));
  cout << "Scoring Metric is:       " << ScoringMetric << endl;

  char *GainsFile = *((char**)pParam->getValueFromKey( "GainFile" ));
  cout << "Gains File is:           " << ReferenceList << endl;

  pOptimizer = new COptimizerMER;
  pMetricSearch = pOptimizer->Init( MatlabCommand, ScoringMetric, NormalizingScript, 
  ReferenceList, NumReferences, GainsFile, pVoc );

  return pOptimizer;
  }
*/


CTextFilterBase *
InitTextFilter( CParam &Param )
{
  char *NormalizingScript        = *((char**)pParam->getValueFromKey( "NormalizingScript"        ));
  char *NormalizingRegexFileName = *((char**)pParam->getValueFromKey( "NormalizingRegexFileName" ));
  
  CTextFilterBase *pTextFilter = NULL;
  
  if ( NormalizingScript ) 
    {
      cout << "Normalizing Script is:   " << NormalizingScript << endl;
      pTextFilter = new CTextFilter();
      ((CTextFilter*)pTextFilter)->SetScript( NormalizingScript );
    
    } else if ( NormalizingRegexFileName ) {
    
    cout << "Normalizing Regex File:  " << NormalizingRegexFileName << endl;
    pTextFilter = new CTextFilterBoostRegex();
    pTextFilter->LoadPatterns( NormalizingRegexFileName );
    
  } else {
    
    cout << "No Normalizing Script or Regex File specified" << endl;
    
  }

  return pTextFilter;
}


// this cannot be used if we're going to use ApplyTransducerSimple (apd 1/5/07)
CTextFilter *
InitWordExpander( CParam &Param )
{
  char *pSourceSentenceExpandingScript = *((char**)pParam->getValueFromKey( "SourceSentenceExpandingScript" ));

  if ( pSourceSentenceExpandingScript ) {
    cout << "Script to find alternative word forms is:   " << pSourceSentenceExpandingScript << endl;

    CTextFilter *pWordExpander = new CTextFilter;
    pWordExpander->SetScript( pSourceSentenceExpandingScript );

    return pWordExpander;
  }

  return NULL;
}




MetricSearch *
InitOptimizerMER( CParam &Param )
{
  char *ScoringMetric = *( (char**)pParam->getValueFromKey( "ScoringMetric" ) );
  char *ReferenceList = *( (char**)pParam->getValueFromKey( "ReferenceList" ) );
  int   NumReferences = *( (int*)  pParam->getValueFromKey( "NumReferences" ) );
  char *GainsFile     = *( (char**)pParam->getValueFromKey( "GainFile" )      );

  cout << "Scoring Metric is:       " << ScoringMetric << endl;
  cout << "Reference List is:       " << ReferenceList << endl;
  cout << "Number of References is: " << NumReferences << endl;
  if ( GainsFile )
    cout << "Gains File is:           " << GainsFile     << endl;


  int NGramsConsidered = 1;
  pMetricSearch = NULL;

  //if there is a reference list
  if ( ReferenceList != NULL ) 
  {
    if ( ScoringMetric == NULL ) {
      cerr << "You must provide a ScoringMetric" << endl;
      exit(2);
    }

    cout << "Metric Initialized for: " << ScoringMetric << endl;

    if ( !pTextFilter )
      cout << "NO TEXT FILTER " << endl;

    if ( strcmp( ScoringMetric, "NIST" ) == 0 ) 
    {
      NGramsConsidered = 5;
      if ( NSZ < 5 ) 
      {
        NGramsConsidered = NSZ;
        cout << "WARNING: Dropping NGramsConsidered for ScoringMetric to: " << NGramsConsidered
          << " since NSZ = " << NSZ << endl;
      }
      
      pMetricSearch = new BleuSearch();
      pMetricSearch->SetTextFilter( pTextFilter );
      ((BleuSearch*) pMetricSearch)->Init( ReferenceList, GainsFile, 
                                           string( ScoringMetric ), 
                                           NumReferences, 
                                           NGramsConsidered, pVoc );
      
    } else if ( strcmp( ScoringMetric, "NISTBLEU" ) == 0 
                || strcmp( ScoringMetric, "IBMBLEU" ) == 0 ) 
    {
      NGramsConsidered = 4;
      if ( NSZ < 4 ) 
      {
        NGramsConsidered = NSZ;
        cout << "WARNING: Dropping NGramsConsidered for ScoringMetric to: " << NGramsConsidered
          << " since NSZ = " << NSZ << endl;
      }
      
      pMetricSearch = new BleuSearch();
      pMetricSearch->SetTextFilter( pTextFilter );
      ((BleuSearch*) pMetricSearch)->Init( ReferenceList, GainsFile, 
                                           string( ScoringMetric ), 
                                           NumReferences, 
                                           NGramsConsidered, pVoc );
    
    } else if ( strcmp( ScoringMetric, "TER" ) == 0 ) {

      pMetricSearch = new TERSearch( ReferenceList, NumReferences, pTextFilter, pVoc);

    } else {
      pMetricSearch = NULL;
      cout << "Unkown metric=" << ScoringMetric << " - evaluation will not be available" << endl;
      cout << endl;
    }
    
  }
  



  
  //initialize with a scoring metric
  pOptimizerMER = new COptimizeMER( pMetricSearch );

  // initialize the optimizer with the user given or the default parameters
  pOptimizerMER->Init( pParam );
  

  return pMetricSearch;

}



// ====================================================================
// Translation functions
// ====================================================================

void
TranslateText( CText *pTest )
{
  TraceLevel = 0;
  //  CParam Param;

  if ( pTest == NULL ) {
    // no srcset given!
    cerr << "main(): No srcset (Sentence List) given! Exiting.\n";
    exit(1);
  }

  //char *PreprocessingScript           = *( (char**)pParam->getValueFromKey( "PreprocessingScript" )        );
  char *SourceSentenceExpandingScript = *( (char**)pParam->getValueFromKey( "SourceSentenceExpandingScript" ));
  char *NormalizingScript             = *( (char**)pParam->getValueFromKey( "NormalizingScript" )            );
  char *NormalizingRegexFileName      = *( (char**)pParam->getValueFromKey( "NormalizingRegexFileName" )     );

  // first, we need to access the parameters for optimization
  int   MEROptimize                   = *( (int*)  pParam->getValueFromKey( "MEROptimize" )                  );
  int   NBest                         = *( (int*)  pParam->getValueFromKey( "NBest" )                        );
  char *ScoringMetric                 = *( (char**)pParam->getValueFromKey( "ScoringMetric" )                );
  char *ReferenceList                 = *( (char**)pParam->getValueFromKey( "ReferenceList" )                );
  int   NumReferences                 = *( (int*)  pParam->getValueFromKey( "NumReferences" )                );
  char *GainsFile                     = *( (char**)pParam->getValueFromKey( "GainFile" )                     );
  int   LoadUnique                    = *( (int*)  pParam->getValueFromKey( "LoadUnique" )                   );
  int   GenOracleScores               = *( (int*)  pParam->getValueFromKey( "GenOracleScores" )              );
  float PhraseAlignDynamicFactors     = *( (float*)pParam->getValueFromKey( "PhraseAlignDynamicFactors" )    );
  int   IterationLimit                = *( (int*)  pParam->getValueFromKey( "IterationLimit" )               );

  int   DisplayLattice                = *( (int*)  pParam->getValueFromKey( "DisplayLattice" )               );
  int   DisplayLatticeStatistics      = *( (int*)  pParam->getValueFromKey( "DisplayLatticeStatistics" )     );
  int   DisplayTranslations           = *( (int*)  pParam->getValueFromKey( "DisplayTranslations" )          );
  int   DisplayTranslationsNormalized = *( (int*)  pParam->getValueFromKey( "DisplayTranslationsNormalized" ));

  int   DisplayFeaturesRange          = *( (int*)  pParam->getValueFromKey( "DisplayFeaturesRange" ));
  
  char *pSentLevelPhraseTableFilePrefix = *( (char**)pParam->getValueFromKey( "PhraseTableSentenceLevelFilePrefix" ));

  // prepare to write hypotheses to file, if parameter was set
  char *SaveHypothesesToFile = *((char**)pParam->getValueFromKey( "SaveHypothesesToFile" ));
  ofstream outHypFile;
  if ( SaveHypothesesToFile != NULL ) {
    outHypFile.open( SaveHypothesesToFile );
  }

  if ( MEROptimize ) {
    cout << "-------------------------------" << endl;
    cout << "TRANSLATION (WITH OPTIMIZATION)" << endl;
    cout << "-------------------------------" << endl;
  } else {
    cout << "----------------------------------" << endl;
    cout << "TRANSLATION (WITHOUT OPTIMIZATION)" << endl;
    cout << "----------------------------------" << endl;
  }


  CLatticeLM *pLattice;
  
  //char* DistortionModelFile = NULL;
  int UseFutureCostEstimates = 0;

  Trace(1, "SplitCharacters: " << SplitCharacters << endl);
  
  // sv 060628 stuff relating to future cost estimation removed

  //  we initialize the Optimzer even if now optimization is requested, as this creates the evaluator
  if ( ReferenceList ) {
    pMetricSearch = InitOptimizerMER( *pParam );
  } else {
    cout << "No references provided. No evaluation or optimization will be performed." << endl;
  }

  // Store lattices to reuse them when running several translations for optimization
  CNumericalArray< CLatticeLM * > AllLattices;


  CString Translation;
  CSentence *pSentence;

  //initialize vector of nbest lists
  pNBestLists = new vector< NBestList * >;

  
  //keep track of scaling factors
  TScoreScalingFactorMap trackScalingFactorScores;


  //set to true if TotalNewTranslations = 0 twice
  bool MERComplete = false;
  int SavedNewTranslations = 0;
  int MERiteration = 1;
  int TotalNewTranslations = 0;
  int SumNumberOfEdges = 0;
  
  //save the start time of the whole process
  string timeAdded;
  stringstream sst(stringstream::in | stringstream::out);
  sst<<time(NULL);
  timeAdded = sst.str();


  do {
    
    SumNumberOfEdges = 0;
    TotalNewTranslations = 0;

    if ( MEROptimize )
      cout << "Start iteration " << MERiteration << " of full translations" << endl;


    PrintTime( "Start of Translation" );
    
	
    //========================================
    // first the translation
    //========================================

    RawCounts modelBestCounts;
    if (pMetricSearch != NULL){
      //pMetricSearch->initRawCounts( NGramsConsidered );
      cout << "Ngrams considered in evaluation metric = " << pMetricSearch->GetNGramConsidered() << endl;
    }


    // we store the current scaling factors for display at the end of the translation run
    CNumericalArray< float > &CurrentFactors = FeatureWeights.GetRow( MERiteration );
    for ( int m = 0; m < ScalingFactors.size(); m++ ) {
      CurrentFactors.Append( ScalingFactors[ m ] );
    }


    TPosition S;
    CWordIndex Sf;
	
    Sf = pTest->FirstSentence( S );
    SentenceNumber = 0;

    // Loop over all sentences
    while ( Sf != pTest->Vocabulary()->EOT ) 
    {
	  
      // Load sentence-level phrase table, if necessary
      // Decoder already has the pointer pTranslationMemory
      // On first sentence, pTranslationMemory has no transducers, which are then removed after translation
      if ( pSentLevelPhraseTableFilePrefix ) {
        char SentLevelPhraseTableFile[ strlen(pSentLevelPhraseTableFilePrefix) + 6 ];
        sprintf( SentLevelPhraseTableFile, "%s%d", pSentLevelPhraseTableFilePrefix, SentenceNumber );
        cout << "Attempting to load transducer file \"" << SentLevelPhraseTableFile << "\"" << endl;
        pTranslationMemory->LoadTransducer( SentLevelPhraseTableFile );
      }
      
      //Nguyen, Aug 7, 2008: Cohesive Decoding
      vector<tokens> SentDependencyTree;
      if ( CohesionCheck ){      	
      	SentDependencyTree = dv[SentenceNumber];
      }
      
      SentenceNumber++;
      
      cout << "TRANSLATION_RESULT_SENTENCE " << SentenceNumber << " ";
      pSentence = pTest->GetSentence( S );
      pSentence->Display( pVoc, True ); cout << endl;
	  
      //ignore the empty sentences
      if( pSentence->IsEmpty() ) {
		
        cout << "Translation is: NONE - SOURCE SENTENCE WAS EMPTY" << Translation << endl << endl << flush;

        // insert placeholder for empty sentence into vector of nbest lists
        if ( MERiteration == 1 )
          pNBestLists->push_back( new NBestList );
        
      } else {
        
        // NBest list object for sentence. Stored for MER training or evaluation
        NBestList* pNB = NULL;
        if ( MERiteration > 1 && ReferenceList != NULL ) {
          pNB = (*pNBestLists)[ SentenceNumber - 1 ];
        } else {
          pNB = new NBestList;
          pNB->Init( true, true );
          
          // store nbest list object which will be reused during evaluation or MER iterations
          if ( ReferenceList != NULL ) {
            pNBestLists->push_back( pNB );
          }
          
        }
        
        // array to hold translation results
        CArray< CDecoderResult* > TranslationResults;
        
		//////////////////////////////////////////////////////////////////////////
		//Nguyen, Aug 6, 2008: Cohesive decoding with source-side DependencyTree
		
		if ( CohesionCheck ){			
			TranslateSentenceWithDependencyTree( pSentence, AllLattices[ SentenceNumber ], TranslationResults, SentDependencyTree, InterruptionCheckAlgorithm );		
		} else {
			// translate the sentence
			TranslateSentence( pSentence, AllLattices[ SentenceNumber ], TranslationResults );
		}        

        // keep track of how many edges were generated
        pLattice = AllLattices[ SentenceNumber ];
        SumNumberOfEdges += pLattice->NumberOfEdges();
        //pLattice->DisplayEdges();

        // display translation results before text normalization
        if ( DisplayTranslations )
          ShowTranslationResults( SentenceNumber, TranslationResults, pLattice );
        
        if ( NormalizingScript || NormalizingRegexFileName ) 
          NormalizeTranslations( TranslationResults );

        // display translation results after text normalization
        if ( DisplayTranslationsNormalized )
          ShowTranslationResults( SentenceNumber, TranslationResults, pLattice );
        
        // and do the processing needed for optimization
        //TotalNewTranslations += 
        //pOptimizer->ProcessTranslationBank( pLattice, ScalingFactors, TranslationResults, 
        //								  nBestLists[ Snum - 1 ], Snum, modelBestCounts, 
        //								  *pSentence, LoadUnique );
        
        // true casing all the translations in the n-best list
        //ApplyTrueCasing( TranslationResults );


        RawCounts topTranslationRaw;
        
        topTranslationRaw = BuildNBest( TranslationResults, pNB,
                                        pMetricSearch, NBest, SentenceNumber - 1, pSentence, TotalNewTranslations );
        cout << "topTranslationRaw: " << topTranslationRaw << endl;
        
        modelBestCounts = modelBestCounts + topTranslationRaw;

        // sv ??? cout << "modelBestCounts: " << modelBestCounts << endl;
        
        // display translation results after text normalization
        //if ( DisplayTranslationsNormalized )
        //ShowTranslationResults( SentenceNumber, TranslationResults, pLattice );
        
        // get the 1best translation
        Translation = TranslationResults[ TranslationResults.MinIndex() ]->m_Translation;
        cout << "Translation is: " << Translation << endl << endl << flush;
        
        if( SaveHypothesesToFile != NULL && MERiteration == 1) {
          outHypFile << Translation << endl;
        }

        // Display min, max of features used to score translations
        if ( DisplayFeaturesRange ) {
          DisplayFeaturesMinMax( AllLattices[ SentenceNumber ] );
        }
        
        // Clean Memory after sentence translation
        
        // clear decoder
        pDecoder->Clear();

        //clear translation results
        for ( int t = TranslationResults.MinIndex(); t <= TranslationResults.MaxIndex(); t++ ) {
          delete TranslationResults[ t ];
        }
        TranslationResults.Clear();
        
        // clear NBestList if no references are used
        if ( ReferenceList == NULL ) {
          cout << "No references -> delete NBestList" << endl;
          delete pNB;
        }
        
        // delete translation lattice if we're rebuilding it every time
        if ( PhraseAlignDynamicFactors || ReferenceList == NULL ) {
          delete AllLattices[ SentenceNumber ];
          AllLattices[ SentenceNumber ] = NULL;
        }
 
      }
	  
      // clean up sentence-level phrase table, if used
      if ( pSentLevelPhraseTableFilePrefix ) {
        pTranslationMemory->ClearTransducer();
      }
      
      Sf = pTest->NextSentence( S );
      
    }//end all source sentences
	
    PrintTime( "End of Translation" );

    if( SaveHypothesesToFile != NULL && MERiteration == 1) {
      outHypFile.close();    
      cerr << "Wrote translations to: " << SaveHypothesesToFile << endl;
    }
    
    cout << "TotalNewEdges:       " << SumNumberOfEdges << endl;
    cout << "TotalNewTranslations:" << TotalNewTranslations << endl;


    // now the scoring
    ScoreTranslation( modelBestCounts, trackScalingFactorScores, MERiteration );
      
    cout << "=============================================" << endl;

    // and here comes the optimization part
    if ( MEROptimize ) {

      // after the last translation we don't optimize again, as we would not
      // have a verification if the resulting scaling factors are really improving the translation
      if ( MERiteration < IterationLimit ) {

        //this criteria is probably a bit too strict on its own. so we also check to make sure the metric score hasnt stagnated. we can
        //check this by looking at the score tracker
        if ( TotalNewTranslations == 0 ) {
          MERComplete = true;
          cout << "Optimization: Early termination triggered since TotalNewTranslations == 0" << endl;
        }
    
        PrintTime( "Start of Optimization" );

        float FinalScore = OptimizeParameters( ScalingFactors, *pNBestLists, pMetricSearch,
                                               trackScalingFactorScores, timeAdded );
        FinalScores[ MERiteration + 1 ] = FinalScore;

        PrintTime( "End of Optimization" );
    
        // now we have new scaling factors and give them back to the decoder and the phrase aligner
        pDecoder->SetScalingFactors( ScalingFactors );
        UpdatePhraseAlignerScalingFactors();

        cout << "End of Iteration " << MERiteration << " out of " << IterationLimit << endl;
      } // if ( MERiteration < IterationLimit )

    } else {
      // no optimization, i.e. set flag to end the loop
      MERComplete = True;
    }

    MERiteration++;
    
    //as long as there is optimization, its not finished and 
    //we have more iterations left to go
    
  } while ( MERiteration <= IterationLimit && !MERComplete );
	
  DisplayScalingFactors( trackScalingFactorScores );

  //========================================
  // we are done - cleanup
  //========================================
  
  Trace( 1, "Cleanup of data structures" );
  
  /*
    for( vector< CNBest* >::iterator iter=nBestLists.begin(); iter!=nBestLists.end(); ++iter ) {
    CNBest* oneNB = (*iter);
    oneNB->ClearAll();
    delete oneNB;
    oneNB=NULL;
    }
  */
  
  
  // delete the Lattices
  for ( int i = AllLattices.MinIndex(); i <= AllLattices.MaxIndex(); i++ ) {
    delete AllLattices[ i ];
  }
  
  // todo: delete pNBestLists
  
}



/**
 * @brief Starts up translation server on server port provided by parameter
 *        ServerPort.
 *
 * @todo  Needs to be tested for leaks and buffer overlflows.
 *
 * @author Last edited by Anthony D'Auria (apd)
 */
void
TranslateServerStart( int ServerPort )
{
  TranslationServerSocket* serverSocket = new TranslationServerSocket();
  
  if ( serverSocket->start( "0.0.0.0", ServerPort ) == -1 ) {
    cerr << "> Could not start TranslationServerSocket on port " 
         << ServerPort << endl;
    exit(-1);
  }
  cerr << "\nTranslation Server started, listening on port " << ServerPort << endl << endl;
  
  
  char source[TRANSLATION_BUFFER_SIZE], target[TRANSLATION_BUFFER_SIZE];
  int sentenceCnt = 0;
  
  CSentence* pSentence = NULL;
  CLatticeLM* pLattice = NULL;
  CArray< CDecoderResult* > TranslationResults;
  CString Translation;
  
  while ( true ) {
    sentenceCnt++;
    
    if ( serverSocket->getSource( source ) < 0 )
      break;
    
    // translation
    cout << "> Translation Input: " << source << endl;
    
    // convert char* source into CSentence
    pSentence = new CSentence();
    pSentence->ReadFromBuffer( source, pVoc );
    
    cout << "CSentence Length: " << pSentence->Size() << endl;
    
    try {
      pLattice = 0;
      cout << "CSentence Length: " << pSentence->Size() << endl;
      
      TranslateSentence ( pSentence, pLattice, TranslationResults );
      
      //translate(source, target, pVoc, pDecoder, pLexicon, pPhraseSearcher, pLatticeRescorer, RescoreTransducer, SourceFileSA, LeaveOneOutPhrases, SplitCharacters);
      
      Translation = TranslationResults[ TranslationResults.MinIndex() ]->m_Translation;
      cout << "Translation is: " << Translation << endl << endl << flush;
      strcpy(target,((std::string)Translation).c_str());
      
      // clean up translation results
      pDecoder->Clear();
      for(int t = TranslationResults.MinIndex(); t <= TranslationResults.MaxIndex(); t++) {
        delete TranslationResults[ t ];
        TranslationResults[ t ] = NULL;                    
      }
      TranslationResults.Clear();
      
    } catch (...) {
      cerr << "> Exception while translating: " << source << endl;
      cout << "> Exception while translating: " << source << endl;
      try {
        
        pLattice = NULL;
        CArray< CDecoderResult* > TranslationResults;
        
        TranslateSentence ( pSentence, pLattice, TranslationResults );
        
        //translate(source, target, pVoc, pDecoder, pLexicon, pPhraseSearcher, pLatticeRescorer, RescoreTransducer, SourceFileSA, LeaveOneOutPhrases, SplitCharacters);
        
        CString Translation = TranslationResults[ TranslationResults.MinIndex() ]->m_Translation;
        cout << "Translation is: " << Translation << endl << endl << flush;
        strcpy(target, ((std::string)Translation).c_str());
        
        // clean up translation results
        pDecoder->Clear();
        for(int t = TranslationResults.MinIndex(); t <= TranslationResults.MaxIndex(); t++) {
          delete TranslationResults[ t ];
          TranslationResults[ t ] = NULL;                    
        }
        TranslationResults.Clear();
        
      } catch (...) {
        cerr << "> Fatal exception, sentence " << sentenceCnt << endl;
        cout << "> Fatal exception, sentence " << sentenceCnt << endl;
        throw;
      }
    }
    cout << "> Translation Output: " << target << endl;
    
    // clean up source sentence items
    delete pSentence;
    if( pLattice ) {
      delete pLattice;
      pLattice = NULL;
    }
    
    if ( serverSocket->sendTarget( target ) < 0 )
      break;
  }
  
  cerr << "> TranslationServerSocket closed.\n";
  
}



/*!
TranslateTextResumable keeps track of completed sentence translations via file
 locks. NBest lists of successfully translated sentences are written to disk. 

\note 

\param pText CText pointer to test set.

\author Anthony D'Auria (apd@cs.cmu.edu)
 adapted from the Worker Bee system written by Ashish Venugopal (ashishv@cs.cmu.edu)

\todo Enable MER
*/

void
TranslateTextResumable( CText *pTest )
{

  
  TraceLevel = 0;
  //  CParam Param;
  
  if ( pTest == NULL ) {
    // no srcset given!
    cerr << "main(): No srcset (Sentence List) given! Exiting.\n";
    exit(1);
  }
  
  //char *PreprocessingScript           = *( (char**)pParam->getValueFromKey( "PreprocessingScript" )        );
  char *SourceSentenceExpandingScript = *( (char**)pParam->getValueFromKey( "SourceSentenceExpandingScript" ));
  char *NormalizingScript             = *( (char**)pParam->getValueFromKey( "NormalizingScript" )            );
  char *NormalizingRegexFileName      = *( (char**)pParam->getValueFromKey( "NormalizingRegexFileName" )     );
  
  // first, we need to access the parameters for optimization
  int   MEROptimize                   = *( (int*)  pParam->getValueFromKey( "MEROptimize" )                  );
  int   NBest                         = *( (int*)  pParam->getValueFromKey( "NBest" )                        );
  char *ScoringMetric                 = *( (char**)pParam->getValueFromKey( "ScoringMetric" )                );
  char *ReferenceList                 = *( (char**)pParam->getValueFromKey( "ReferenceList" )                );
  int   NumReferences                 = *( (int*)  pParam->getValueFromKey( "NumReferences" )                );
  char *GainsFile                     = *( (char**)pParam->getValueFromKey( "GainFile" )                     );
  int   LoadUnique                    = *( (int*)  pParam->getValueFromKey( "LoadUnique" )                   );
  int   GenOracleScores               = *( (int*)  pParam->getValueFromKey( "GenOracleScores" )              );
  float PhraseAlignDynamicFactors     = *( (float*)pParam->getValueFromKey( "PhraseAlignDynamicFactors" )    );
  int   IterationLimit                = *( (int*)  pParam->getValueFromKey( "IterationLimit" )               );
  
  int   DisplayLattice                = *( (int*)  pParam->getValueFromKey( "DisplayLattice" )               );
  int   DisplayLatticeStatistics      = *( (int*)  pParam->getValueFromKey( "DisplayLatticeStatistics" )     );
  int   DisplayTranslations           = *( (int*)  pParam->getValueFromKey( "DisplayTranslations" )          );
  int   DisplayTranslationsNormalized = *( (int*)  pParam->getValueFromKey( "DisplayTranslationsNormalized" ));
  
  int   DisplayFeaturesRange          = *( (int*)  pParam->getValueFromKey( "DisplayFeaturesRange" ));
  
  char *pSentLevelPhraseTableFilePrefix = *( (char**)pParam->getValueFromKey( "PhraseTableSentenceLevelFilePrefix" ));
  
  char *WorkerBeeDir                  =  *( (char**)pParam->getValueFromKey( "WorkerBeeDir" ));
  int   WorkerBeeSaveSentenceFeatures =  *( (int*)  pParam->getValueFromKey( "WorkerBeeSaveSentenceFeatures" ));
  double MaxSentenceProcessingTime    =  *( (double*)pParam->getValueFromKey( "DisplayLatticeStatistics" )     );
  
  // prepare to write hypotheses to file, if parameter was set
  char *SaveHypothesesToFile = *((char**)pParam->getValueFromKey( "SaveHypothesesToFile" ));
  ofstream outHypFile;
  if ( SaveHypothesesToFile != NULL ) {
    outHypFile.open( SaveHypothesesToFile );
  }
  
  if ( MEROptimize ) {
    cout << "-------------------------------" << endl;
    cout << "TRANSLATION (WITH OPTIMIZATION)" << endl;
    cout << "-------------------------------" << endl;
  } else {
    cout << "----------------------------------" << endl;
    cout << "TRANSLATION (WITHOUT OPTIMIZATION)" << endl;
    cout << "----------------------------------" << endl;
  }
  
  
  CLatticeLM *pLattice;
  
  //char* DistortionModelFile = NULL;
  int UseFutureCostEstimates = 0;
  
  Trace(1, "SplitCharacters: " << SplitCharacters << endl);
  
  // sv 060628 stuff relating to future cost estimation removed
  
  //  we initialize the Optimzer even if now optimization is requested, as this creates the evaluator
  if ( ReferenceList ) {
    pMetricSearch = InitOptimizerMER( *pParam );
  } else {
    cout << "No references provided. No evaluation or optimization will be performed." << endl;
  }
  
  // Store lattices to reuse them when running several translations for optimization
  CNumericalArray< CLatticeLM * > AllLattices;
  
  
  CString Translation;
  CSentence *pSentence;
  
  //initialize vector of nbest lists
  pNBestLists = new vector< NBestList * >;
  
  
  //keep track of scaling factors
  TScoreScalingFactorMap trackScalingFactorScores;
  
  
  //set to true if TotalNewTranslations = 0 twice
  //bool MERComplete = false;
  bool MERComplete = true; // DISABLE MER for now
  int SavedNewTranslations = 0;
  int MERiteration = 1;
  int TotalNewTranslations = 0;
  int SumNumberOfEdges = 0;
  
  //save the start time of the whole process
  string timeAdded;
  stringstream sst(stringstream::in | stringstream::out);
  sst<<time(NULL);
  timeAdded = sst.str();
  
  do {
    
    // Worker-Bee controls
    // arv sez: true if at least one sentence is still unprocessed
    bool workerBeeAtLeastOneSentenceUnprocessed = false;
    // true if at least one sentence was processed in current loop
    bool workerBeeAtLeastOneSentenceProcessed = false;
    // true if we've already gone through all sentences once
    bool workerBeeCompletedFirstPass = false;
    // true if we have to try to process all sentences now, even recently locked ones
    bool workerBeeGetThemAllNow = false;
    
    
    SumNumberOfEdges = 0;
    TotalNewTranslations = 0;
    
    if ( MEROptimize )
      cout << "Start iteration " << MERiteration << " of full translations" << endl;
    
    
    PrintTime( "Start of Translation" );
    
    
    //========================================
    // first the translation
    //========================================
    
    RawCounts modelBestCounts;
    if (pMetricSearch != NULL){
      //pMetricSearch->initRawCounts( NGramsConsidered );
      cout << "Ngrams considered in evaluation metric = " << pMetricSearch->GetNGramConsidered() << endl;
    }
    
    
    // we store the current scaling factors for display at the end of the translation run
    CNumericalArray< float > &CurrentFactors = FeatureWeights.GetRow( MERiteration );
    for ( int m = 0; m < ScalingFactors.size(); m++ ) {
      CurrentFactors.Append( ScalingFactors[ m ] );
    }
    
    
    TPosition S;
    CWordIndex Sf;
    
    do { // Worker Bee loop that makes sure this bee picks up unifinished translations
         // arv sez: true if at least one sentence is still unprocessed
      workerBeeAtLeastOneSentenceUnprocessed = false;
      // true if at least one sentence was processed in current loop
      workerBeeAtLeastOneSentenceProcessed = false;
      
      Sf = pTest->FirstSentence( S );
      SentenceNumber = 0;
      
      // Loop over all sentences
      while ( Sf != pTest->Vocabulary()->EOT ) 
      {
        SentenceNumber++;
       
        // safely concatenate strings and integers with ostringstream
        ostringstream filePrefixStream;
        filePrefixStream << WorkerBeeDir << "/" << SentenceNumber;
        
        // WorkerBee files for locking and completion checks
        string Sent_LockFile = filePrefixStream.str() + ".lock";
        int lockHandle;
        
        //string Sent_NBestCandidatesFile = filePrefixStream.str() + ".cands";
        //string Sent_NBestFeaturesFile = filePrefixStream.str() + ".feats";
        string Sent_FeatsFile = filePrefixStream.str() + ".feats";
        string Sent_NBestTranslationsFile = filePrefixStream.str() + ".translations";

        bool Sent_Skip = BeeIsUnitInProgress(SentenceNumber, 
                                             Sent_LockFile, Sent_FeatsFile, 
                                             Sent_NBestTranslationsFile,
                                             lockHandle,
                                             workerBeeAtLeastOneSentenceUnprocessed, 
                                             workerBeeCompletedFirstPass,
                                             workerBeeGetThemAllNow);
        
        
        
        if ( ! Sent_Skip )
        {

          // Load sentence-level phrase table, if necessary
          // Decoder already has the pointer pTranslationMemory
          // On first sentence, pTranslationMemory has no transducers, which are then removed after translation
          if ( pSentLevelPhraseTableFilePrefix ) {
            char SentLevelPhraseTableFile[ strlen(pSentLevelPhraseTableFilePrefix) + 6 ];
            sprintf( SentLevelPhraseTableFile, "%s%d", pSentLevelPhraseTableFilePrefix, SentenceNumber-1 );
            cout << "Attempting to load transducer file \"" << SentLevelPhraseTableFile << "\"" << endl;
            pTranslationMemory->LoadTransducer( SentLevelPhraseTableFile );
          }
          
          
          
          cout << "TRANSLATION_RESULT_SENTENCE " << SentenceNumber << " ";
          pSentence = pTest->GetSentence( S );
          pSentence->Display( pVoc, True ); cout << endl;
          
          //ignore the empty sentences
          if( pSentence->IsEmpty() ) {
            
            cout << "Translation is: NONE - SOURCE SENTENCE WAS EMPTY" << Translation << endl << endl << flush;
            
            // insert placeholder for empty sentence into vector of nbest lists
            if ( MERiteration == 1 )
              pNBestLists->push_back( new NBestList );
            
          } else {
            
            // NOT STORING N-BEST LISTS in memory
            
            
            // NBest list object for sentence. Stored for MER training or evaluation
            NBestList* pNB = NULL;
            if ( MERiteration > 1 && ReferenceList != NULL ) {
              pNB = (*pNBestLists)[ SentenceNumber - 1 ];
            } else {
              pNB = new NBestList;
              pNB->Init( true, true );
              
              // store nbest list object which will be reused during evaluation or MER iterations
              if ( ReferenceList != NULL ) {
                pNBestLists->push_back( pNB );
              }
              
            }
            
            // array to hold translation results
            CArray< CDecoderResult* > TranslationResults;
            
            // translate the sentence
            TranslateSentence( pSentence, AllLattices[ SentenceNumber ], TranslationResults );
            
            // keep track of how many edges were generated
            pLattice = AllLattices[ SentenceNumber ];
            SumNumberOfEdges += pLattice->NumberOfEdges();
            //pLattice->DisplayEdges();
            
            // display translation results before text normalization
            if ( DisplayTranslations )
              ShowTranslationResults( SentenceNumber, TranslationResults, pLattice );
            
            if ( NormalizingScript || NormalizingRegexFileName ) 
              NormalizeTranslations( TranslationResults );
            
            // display translation results after text normalization
            if ( DisplayTranslationsNormalized )
              ShowTranslationResults( SentenceNumber, TranslationResults, pLattice );
            
            // and do the processing needed for optimization
            //TotalNewTranslations += 
            //pOptimizer->ProcessTranslationBank( pLattice, ScalingFactors, TranslationResults, 
            //								  nBestLists[ Snum - 1 ], Snum, modelBestCounts, 
            //								  *pSentence, LoadUnique );
            
            // true casing all the translations in the n-best list
            //ApplyTrueCasing( TranslationResults );
            
            
            RawCounts topTranslationRaw;
            
            topTranslationRaw = BuildNBest( TranslationResults, pNB,
                                            pMetricSearch, NBest, SentenceNumber - 1, pSentence, TotalNewTranslations );
            cout << "topTranslationRaw: " << topTranslationRaw << endl;
            
            modelBestCounts = modelBestCounts + topTranslationRaw;
            
            // sv ??? cout << "modelBestCounts: " << modelBestCounts << endl;
            
            // display translation results after text normalization
            //if ( DisplayTranslationsNormalized )
            //ShowTranslationResults( SentenceNumber, TranslationResults, pLattice );
            
            // get the 1best translation
            Translation = TranslationResults[ TranslationResults.MinIndex() ]->m_Translation;
            cout << "Translation is: " << Translation << endl << endl << flush;
            
            if( SaveHypothesesToFile != NULL && MERiteration == 1) {
              outHypFile << Translation << endl;
            }
            
            // Display min, max of features used to score translations
            if ( DisplayFeaturesRange ) {
              DisplayFeaturesMinMax( AllLattices[ SentenceNumber ] );
            }
            
            // Output Bee information to file
            
            // output NBest list here. First entry is final translation.
            ofstream Sent_NBestTranslations_stream (Sent_NBestTranslationsFile.c_str());
            ofstream Sent_Feats_stream;
            if ( WorkerBeeSaveSentenceFeatures ) {
              Sent_Feats_stream.open( Sent_FeatsFile.c_str() );
            }
            // output Translation here
            int nBestListSize = TranslationResults.Size();
            for ( int s = 0; s < nBestListSize; s++ ) 
            {
              Sent_NBestTranslations_stream << s+1 << " " << TranslationResults[ s ]->m_Translation << endl;
              
              // Output raw sentence feature scores if option is selected
              if ( WorkerBeeSaveSentenceFeatures ) 
              {
                Sent_Feats_stream << s+1 << " "; 
                vector<float>::iterator i_feats = TranslationResults[ s ]->m_Features.begin();
                vector<float>::iterator i_scalingFactors = ScalingFactors.begin();
                for ( ; i_feats !=  TranslationResults[ s ]->m_Features.end() &&
                      i_scalingFactors != ScalingFactors.end(); 
                      i_feats++, i_scalingFactors++ ) {
                  Sent_Feats_stream << (*i_feats) / (*i_scalingFactors) << " ";
                }
                
                Sent_Feats_stream << endl;
              }
            }
            // Close NBest List and Features List files
            if ( WorkerBeeSaveSentenceFeatures ) {
              Sent_Feats_stream.close();
            }
            Sent_NBestTranslations_stream.close();
            
            // Unlock sentence with remove(), which deletes sentence's .lock file
            remove( Sent_LockFile.c_str() );
            // Free OS's file descriptor
            if ( close(lockHandle) < 0 ) {
              cout << "WorkerBee[ " << SentenceNumber << " ]: Unable to deallocate lock file descriptor!" << endl;
            } else {
              cout << "WorkerBee[ " << SentenceNumber << " ]: Successfully deallocated lock file descriptor." << endl;
            }
              
            // Clean Memory after sentence translation
            
            // clear decoder
            pDecoder->Clear();
            
            //clear translation results
            for ( int t = TranslationResults.MinIndex(); t <= TranslationResults.MaxIndex(); t++ ) {
              delete TranslationResults[ t ];
            }
            TranslationResults.Clear();
            
            // clear NBestList if no references are used
            if ( ReferenceList == NULL ) {
              delete pNB;
            }
            
            // delete translation lattice if we're rebuilding it every time
            if ( PhraseAlignDynamicFactors || ReferenceList == NULL ) {
              delete AllLattices[ SentenceNumber ];
              AllLattices[ SentenceNumber ] = NULL;
            }
            
          }
          
          // clean up sentence-level phrase table, if used
          if ( pSentLevelPhraseTableFilePrefix ) {
            pTranslationMemory->ClearTransducer();
          }
          
        
        }// Bee translation. Skipped if sentence was translated.
        
        
        Sf = pTest->NextSentence( S );

      }//end all source sentences
      
      
      // This bee still needs to make sure that everything has been translated
      workerBeeCompletedFirstPass = true;     
      workerBeeGetThemAllNow = workerBeeCompletedFirstPass && !workerBeeAtLeastOneSentenceProcessed;
      cout << "WorkerBee[ PASS_COMPLETED ]: Status after pass: any unprocessed: "
        << workerBeeAtLeastOneSentenceUnprocessed << ", any processed: " << workerBeeAtLeastOneSentenceProcessed << endl;

    } while ( workerBeeAtLeastOneSentenceUnprocessed );
    
    
    PrintTime( "End of Translation" );

    /* removed for bee version
    if( SaveHypothesesToFile != NULL && MERiteration == 1) {
      outHypFile.close();    
      cerr << "Wrote translations to: " << SaveHypothesesToFile << endl;
    }
    */
    cout << "TotalNewEdges:       " << SumNumberOfEdges << endl;
    cout << "TotalNewTranslations:" << TotalNewTranslations << endl;
    
    
    // now the scoring
    //ScoreTranslation( modelBestCounts, trackScalingFactorScores, MERiteration );
    
    cout << "=============================================" << endl;
    
    // and here comes the optimization part
    if ( MEROptimize ) {
      
      // after the last translation we don't optimize again, as we would not
      // have a verification if the resulting scaling factors are really improving the translation
      if ( MERiteration < IterationLimit ) {
        
        //this criteria is probably a bit too strict on its own. so we also check to make sure the metric score hasnt stagnated. we can
        //check this by looking at the score tracker
        if ( TotalNewTranslations == 0 ) {
          MERComplete = true;
          cout << "Optimization: Early termination triggered since TotalNewTranslations == 0" << endl;
        }
        
        PrintTime( "Start of Optimization" );
        
        float FinalScore = OptimizeParameters( ScalingFactors, *pNBestLists, pMetricSearch,
                                               trackScalingFactorScores, timeAdded );
        FinalScores[ MERiteration + 1 ] = FinalScore;
        
        PrintTime( "End of Optimization" );
        
        // now we have new scaling factors and give them back to the decoder and the phrase aligner
        pDecoder->SetScalingFactors( ScalingFactors );
        UpdatePhraseAlignerScalingFactors();
        
        cout << "End of Iteration " << MERiteration << " out of " << IterationLimit << endl;
      } // if ( MERiteration < IterationLimit )
      
    } else {
      // no optimization, i.e. set flag to end the loop
      MERComplete = True;
    }
    
    MERiteration++;
    
    //as long as there is optimization, its not finished and 
    //we have more iterations left to go
    
  } while ( MERiteration <= IterationLimit && !MERComplete );
	
  DisplayScalingFactors( trackScalingFactorScores );
  
  //========================================
  // we are done - cleanup
  //========================================
  
  Trace( 1, "Cleanup of data structures" );
  
  /*
   for( vector< CNBest* >::iterator iter=nBestLists.begin(); iter!=nBestLists.end(); ++iter ) {
     CNBest* oneNB = (*iter);
     oneNB->ClearAll();
     delete oneNB;
     oneNB=NULL;
   }
   */
  
  
  // delete the Lattices
  for ( int i = AllLattices.MinIndex(); i <= AllLattices.MaxIndex(); i++ ) {
    delete AllLattices[ i ];
  }
  
  // todo: delete pNBestLists
  
}

void
TranslateLattices()
{
  TraceLevel = 0;
  
  char *LatticeList                   = *( (char**)pParam->getValueFromKey( "LatticeList" )                  );
  char *LatticeFormat                 = *( (char**)pParam->getValueFromKey( "LatticeFormat" )                );
  float LatticeEdgeThreshold          = *( (float*)pParam->getValueFromKey( "LatticeEdgeThreshold" )         );
  int   LatticeRemoveEpsilonEdges     = *( (int*)  pParam->getValueFromKey( "LatticeRemoveEpsilonEdges" )    );


  char *NormalizingScript             = *( (char**)pParam->getValueFromKey( "NormalizingScript" )            );
  char *NormalizingRegexFileName      = *( (char**)pParam->getValueFromKey( "NormalizingRegexFileName" )     );

  // first, we need to access the parameters for optimization
  int   MEROptimize                   = *( (int*)  pParam->getValueFromKey( "MEROptimize" )                  );
  int   NBest                         = *( (int*)  pParam->getValueFromKey( "NBest" )                        );
  char *ScoringMetric                 = *( (char**)pParam->getValueFromKey( "ScoringMetric" )                );
  char *ReferenceList                 = *( (char**)pParam->getValueFromKey( "ReferenceList" )                );
  int   NumReferences                 = *( (int*)  pParam->getValueFromKey( "NumReferences" )                );
  char *GainsFile                     = *( (char**)pParam->getValueFromKey( "GainFile" )                     );
  int   LoadUnique                    = *( (int*)  pParam->getValueFromKey( "LoadUnique" )                   );
  int   GenOracleScores               = *( (int*)  pParam->getValueFromKey( "GenOracleScores" )              );
  float PhraseAlignDynamicFactors     = *( (float*)pParam->getValueFromKey( "PhraseAlignDynamicFactors" )    );
  int   IterationLimit                = *( (int*)  pParam->getValueFromKey( "IterationLimit" )               );
  
  int   DisplayLattice                = *( (int*)  pParam->getValueFromKey( "DisplayLattice" )               );
  int   DisplayLatticeStatistics      = *( (int*)  pParam->getValueFromKey( "DisplayLatticeStatistics" )     );
  int   DisplayTranslations           = *( (int*)  pParam->getValueFromKey( "DisplayTranslations" )          );
  int   DisplayTranslationsNormalized = *( (int*)  pParam->getValueFromKey( "DisplayTranslationsNormalized" ));
  
  int   DisplayFeaturesRange = *((int*)pParam->getValueFromKey( "DisplayFeaturesRange" ));

  char *pSentLevelPhraseTableFilePrefix = *( (char**)pParam->getValueFromKey( "PhraseTableSentenceLevelFilePrefix" ));

  // prepare to write hypotheses to file, if parameter was set
  char *SaveHypothesesToFile = *((char**)pParam->getValueFromKey( "SaveHypothesesToFile" ));
  ofstream outHypFile;
  if ( SaveHypothesesToFile != NULL ) {
    outHypFile.open( SaveHypothesesToFile );
  }
  
  
  if ( MEROptimize ) {
    cout << "-------------------------------------------" << endl;
    cout << "TRANSLATION OF LATTICES (WITH OPTIMIZATION)" << endl;
    cout << "-------------------------------------------" << endl;
  } else {
    cout << "----------------------------------------------" << endl;
    cout << "TRANSLATION OF LATTICES (WITHOUT OPTIMIZATION)" << endl;
    cout << "----------------------------------------------" << endl;
  }
  
  
  //char* DistortionModelFile = NULL;
  int UseFutureCostEstimates = 0;
  
  
  
  Trace(1, "SplitCharacters: " << SplitCharacters << endl);
  
  // sv 060628 stuff relating to future cost estimation removed
  
  //  we initialize the Optimzer even if now optimization is requested, as this creates the evaluator
  
  if ( ReferenceList ) {
    pMetricSearch = InitOptimizerMER( *pParam );
  } else {
    cout << "No references provided. No evaluation or optimization will be performed." << endl;
  }
  
  
  CString Translation;

  //keep track of scaling factors
  TScoreScalingFactorMap trackScalingFactorScores;
  
  
  //set to true if TotalNewTranslations = 0 twice
  bool MERComplete = false;
  int SavedNewTranslations = 0;
  int MERiteration = 1;
  int TotalNewTranslations = 0;
  int SumNumberOfEdges = 0;
  
  //save the start time of the whole process
  string timeAdded;
  stringstream sst(stringstream::in | stringstream::out);
  sst<<time(NULL);
  timeAdded = sst.str();
  
  // store TranslationResults for every source sentence
  CArray< CArray< CDecoderResult* > > TranslationResultsArray;
  
  CNumericalArray< CLatticeLM* >* pTestLattices;

  
  do {  // loop over translate - optimize
    
    SumNumberOfEdges = 0;
    TotalNewTranslations = 0;
    
    if ( MEROptimize ) {
      cout << "Start iteration " << MERiteration << " of full translations" << endl;
    }
    
    PrintTime( "Start of Translation" );

    // load and translate lattices
  
    // currently we delete lattices if PhraseAlignDynamicFactors > 0.0
    // therefore, we need to reload them
    if ( PhraseAlignDynamicFactors > 0.0 ) {
      pTestLattices = LoadLattices( *pParam, LatticeList, LatticeFormat, pVoc );
    }

    // initialize nbest lists for each lattice if this hasn't been done yet
    if ( pNBestLists == NULL ){
      pNBestLists = new vector< NBestList * >;
      for (int i = 0; i < pTestLattices->Size(); i++ ) 
      {
        NBestList* pNB = new NBestList;
        pNB->Init( true, true );         // store translations and counts
        pNBestLists->push_back( pNB );
      }
    }
  
    // starting lattice number
    SentenceNumber = pTestLattices->MinIndex();

    // pre-process lattices
    for ( int l = pTestLattices->MinIndex(); l <= pTestLattices->MaxIndex(); l++ ) 
    {
      CLattice* pLattice = (*pTestLattices)[ l ];
      
      if( LatticeEdgeThreshold > 0 ) {
        pLattice->RemoveEdgesBelowThreshold(LatticeEdgeThreshold);
        cout << "Finished removing edges below threshold: " << LatticeEdgeThreshold << endl;
      }
    
      if( LatticeRemoveEpsilonEdges > 0 ) {
        pLattice->RemoveSpecialEdgesConnectLattice();
        cout << "Finished removing <epsilon> edges" << endl;
      }
    }



  
    
    //========================================
    // translation
    //========================================
    
    // Set up MER 
    RawCounts modelBestCounts;
    if (pMetricSearch != NULL){
      //pMetricSearch->initRawCounts( NGramsConsidered );
      cout << "Ngrams considered in evaluation metric = " 
           << pMetricSearch->GetNGramConsidered() << endl;
    }
    
    // Get Scaling Factors
    // we store the current scaling factors for display at the end of the translation run
    CNumericalArray< float > & CurrentFactors = FeatureWeights.GetRow( MERiteration );

    // Propagate Scaling factors
    for ( int m = 0; m < ScalingFactors.size(); m++ ) {
      CurrentFactors.Append( ScalingFactors[ m ] );
    }
    
    
    // Loop over all test lattices
    for ( int l = pTestLattices->MinIndex(); l <= pTestLattices->MaxIndex(); l++ ) 
    {
      cout << "TRANSLATION_RESULT_SENTENCE " << l << " ";
      // pSentence = pTest->GetSentence( S );
      CLatticeLM* pTestLattice = (*pTestLattices)[ l ];
      
      // ignore empty lattice
      if ( pTestLattice->NumberOfEdges() == 0 ) {
        cout << "Translation is: NONE - SOURCE LATTICE WAS EMPTY" 
             << Translation << endl << endl << flush;
      
      } else {
      
        // Load sentence-level phrase table, if necessary
        // Decoder already has the pointer pTranslationMemory
        // On first sentence, pTranslationMemory has no transducers, which are then removed after translation
        if ( pSentLevelPhraseTableFilePrefix ) {
          char SentLevelPhraseTableFile[ strlen(pSentLevelPhraseTableFilePrefix) + 6 ];
          sprintf( SentLevelPhraseTableFile, "%s%d", pSentLevelPhraseTableFilePrefix, l );
          cout << "Attempting to load transducer file \"" << SentLevelPhraseTableFile << "\"" << endl;
          pTranslationMemory->LoadTransducer( SentLevelPhraseTableFile );
        }  
        
        // array to hold translation results
        // array of arrays 
        
        CArray< CDecoderResult* > TranslationResults;
    
        int OrigNumberOfEdges = pTestLattice->NumberOfEdges();

        //bool buildTranslationLattice = ( MERiteration > 1 ) ? false : true;
        bool buildTranslationLattice = true;

        TranslateLattice( pTestLattice, TranslationResults, buildTranslationLattice );
        
        // keep track of how many edges were generated
        // pLattice = AllLattices[ SentenceNumber ]; // apd rm
        SumNumberOfEdges += pTestLattice->NumberOfEdges() - OrigNumberOfEdges;

        //pLattice->DisplayEdges();
        
        // display translation results before text normalization
        if ( DisplayTranslations ) {
          ShowTranslationResults( l, TranslationResults, pTestLattice );
        }     
        
        if ( NormalizingScript || NormalizingRegexFileName ) { 
          NormalizeTranslations( TranslationResults );
        }     
        
        // display translation results after text normalization
        if ( DisplayTranslationsNormalized ) {
          ShowTranslationResults( l, TranslationResults, pTestLattice );
        }                         
        
        RawCounts topTranslationRaw;
          
        if ( MEROptimize ) {
          topTranslationRaw = BuildNBest( TranslationResults, (*pNBestLists)[ l ],
                                          pMetricSearch, NBest, l, NULL, 
                                          TotalNewTranslations );
          cout << "topTranslationRaw: " << topTranslationRaw << endl;
          
          modelBestCounts = modelBestCounts + topTranslationRaw; // apd rm
        }
        // sv ??? cout << "modelBestCounts: " << modelBestCounts << endl;
        
        // display translation results after text normalization
        //if ( DisplayTranslationsNormalized )
        //ShowTranslationResults( SentenceNumber, TranslationResults, pLattice );
        
        // get the 1best translation
        Translation = TranslationResults[ TranslationResults.MinIndex() ]->m_Translation;
        cout << "Translation is: " << Translation << endl << endl << flush;

        if( SaveHypothesesToFile != NULL && MERiteration == 1) {
          outHypFile << Translation << endl;
        }

        // delete TranslationResults after 1-lattice test
        for ( int t = TranslationResults.MinIndex(); t <= TranslationResults.MaxIndex(); t++ ) {
          delete TranslationResults[ t ];
          TranslationResults[ t ] = NULL;                    
        }
        TranslationResults.Clear();
        
        // clear memory
        pDecoder->Clear();

      } // if not empty

      if ( DisplayFeaturesRange ) {
        DisplayFeaturesMinMax( (*pTestLattices)[ SentenceNumber ] );
      }
      
      if ( PhraseAlignDynamicFactors ) {
        delete (*pTestLattices)[ SentenceNumber ];
        (*pTestLattices)[ SentenceNumber ] = NULL;
      }

      // clean up sentence-level phrase table, if used
      if ( pSentLevelPhraseTableFilePrefix ) {
        pTranslationMemory->ClearTransducer();
      }
      
      SentenceNumber++;
      
    }//end all source lattices
    
    PrintTime( "End of Translation" );    
    
    if( SaveHypothesesToFile != NULL && MERiteration == 1) {
      outHypFile.close();    
      cerr << "Wrote translations to: " << SaveHypothesesToFile << endl;
    }
    
    cout << "TotalNewEdges:       " << SumNumberOfEdges << endl;
    cout << "TotalNewTranslations:" << TotalNewTranslations << endl;
    
    // now the scoring
    if ( MEROptimize ) {
      ScoreTranslation( modelBestCounts, trackScalingFactorScores, MERiteration );
    }
    cout << "=============================================" << endl;
    
    // and here comes the optimization part
    if ( MEROptimize ) {
      
      // after the last translation we don't optimize again, as we would not
      // have a verification if the resulting scaling factors are really improving the translation
      if ( MERiteration < IterationLimit ) {
        
        //this criteria is probably a bit too strict on its own. so we also check to make sure the metric score hasnt stagnated. we can
        //check this by looking at the score tracker
        if ( TotalNewTranslations == 0 ) {
          MERComplete = true;
          cout << "Optimization: Early termination triggered since TotalNewTranslations == 0" << endl;
        }
        
        PrintTime( "Start of Optimization" );
        
        float FinalScore = OptimizeParameters( ScalingFactors, *pNBestLists, 
                                               pMetricSearch,
                                               trackScalingFactorScores, timeAdded );
        FinalScores[ MERiteration + 1 ] = FinalScore;
        
        PrintTime( "End of Optimization" );
        
        // now we have new scaling factors and give them back to the decoder and the phrase aligner
        pDecoder->SetScalingFactors( ScalingFactors );
        UpdatePhraseAlignerScalingFactors();
        
        cout << "End of Iteration " << MERiteration << " out of " << IterationLimit << endl;
      } // if ( MERiteration < IterationLimit )
      
    } else {
      // no optimization, i.e. set flag to end the loop
      MERComplete = True;
    }
    
    MERiteration++;
    
    //as long as there is optimization, its not finished and 
    //we have more iterations left to go
    
  } while ( MERiteration <= IterationLimit && !MERComplete );
	
  DisplayScalingFactors( trackScalingFactorScores );
  
  //========================================
  // we are done - cleanup
  //========================================
  
  Trace( 1, "Cleanup of data structures" );
  
  cout << "delete all lattices" << endl;
  delete pTestLattices;

  cout << "CLatticeEdge   " << CLatticeEdge::s_NumberOfObjects << endl;
}


/**
 * @brief Worker-bee version of lattice translation that uses file locks to
 *        mark translations in progress and to select untranslated lattices.
 *
 * @author Anthony D'Auria
 * @return Void.
 */
void
TranslateLatticesResumable()
{
  
  
  TraceLevel = 0;

  char *LatticeList                   = *( (char**)pParam->getValueFromKey( "LatticeList" )                  );
  char *LatticeFormat                 = *( (char**)pParam->getValueFromKey( "LatticeFormat" )                );
  float LatticeEdgeThreshold          = *( (float*)pParam->getValueFromKey( "LatticeEdgeThreshold" )         );
  int   LatticeRemoveEpsilonEdges     = *( (int*)  pParam->getValueFromKey( "LatticeRemoveEpsilonEdges" )    );
  
  
  char *SourceSentenceExpandingScript = *( (char**)pParam->getValueFromKey( "SourceSentenceExpandingScript" ));
  char *NormalizingScript             = *( (char**)pParam->getValueFromKey( "NormalizingScript" )            );
  char *NormalizingRegexFileName      = *( (char**)pParam->getValueFromKey( "NormalizingRegexFileName" )     );
  
  // first, we need to access the parameters for optimization
  int   MEROptimize                   = *( (int*)  pParam->getValueFromKey( "MEROptimize" )                  );
  int   NBest                         = *( (int*)  pParam->getValueFromKey( "NBest" )                        );
  char *ScoringMetric                 = *( (char**)pParam->getValueFromKey( "ScoringMetric" )                );
  char *ReferenceList                 = *( (char**)pParam->getValueFromKey( "ReferenceList" )                );
  int   NumReferences                 = *( (int*)  pParam->getValueFromKey( "NumReferences" )                );
  char *GainsFile                     = *( (char**)pParam->getValueFromKey( "GainFile" )                     );
  int   LoadUnique                    = *( (int*)  pParam->getValueFromKey( "LoadUnique" )                   );
  int   GenOracleScores               = *( (int*)  pParam->getValueFromKey( "GenOracleScores" )              );
  float PhraseAlignDynamicFactors     = *( (float*)pParam->getValueFromKey( "PhraseAlignDynamicFactors" )    );
  int   IterationLimit                = *( (int*)  pParam->getValueFromKey( "IterationLimit" )               );
  
  int   DisplayLattice                = *( (int*)  pParam->getValueFromKey( "DisplayLattice" )               );
  int   DisplayLatticeStatistics      = *( (int*)  pParam->getValueFromKey( "DisplayLatticeStatistics" )     );
  int   DisplayTranslations           = *( (int*)  pParam->getValueFromKey( "DisplayTranslations" )          );
  int   DisplayTranslationsNormalized = *( (int*)  pParam->getValueFromKey( "DisplayTranslationsNormalized" ));
  
  int   DisplayFeaturesRange          = *( (int*)  pParam->getValueFromKey( "DisplayFeaturesRange" ));
  
  char *pSentLevelPhraseTableFilePrefix = *( (char**)pParam->getValueFromKey( "PhraseTableSentenceLevelFilePrefix" ));
  
  char *WorkerBeeDir                  =  *( (char**)pParam->getValueFromKey( "WorkerBeeDir" ));
  int   WorkerBeeSaveSentenceFeatures =  *( (int*)  pParam->getValueFromKey( "WorkerBeeSaveSentenceFeatures" ));
  double MaxSentenceProcessingTime    =  *( (double*)pParam->getValueFromKey( "DisplayLatticeStatistics" )     );
  
  // prepare to write hypotheses to file, if parameter was set
  char *SaveHypothesesToFile = *((char**)pParam->getValueFromKey( "SaveHypothesesToFile" ));
  ofstream outHypFile;
  if ( SaveHypothesesToFile != NULL ) {
    outHypFile.open( SaveHypothesesToFile );
  }
  
  if ( MEROptimize ) {
    cout << "-------------------------------" << endl;
    cout << "TRANSLATION (WITH OPTIMIZATION)" << endl;
    cout << "-------------------------------" << endl;
  } else {
    cout << "----------------------------------" << endl;
    cout << "TRANSLATION (WITHOUT OPTIMIZATION)" << endl;
    cout << "----------------------------------" << endl;
  }
  
  //char* DistortionModelFile = NULL;
  int UseFutureCostEstimates = 0;
  
  Trace(1, "SplitCharacters: " << SplitCharacters << endl);
  
  //  we initialize the Optimzer even if now optimization is requested, as this creates the evaluator
  if ( ReferenceList ) {
    pMetricSearch = InitOptimizerMER( *pParam );
  } else {
    cout << "No references provided. No evaluation or optimization will be performed." << endl;
  }
  
  CString Translation;
  
  //initialize vector of nbest lists
  pNBestLists = new vector< NBestList * >;
  
  
  //keep track of scaling factors
  TScoreScalingFactorMap trackScalingFactorScores;
  
  
  //set to true if TotalNewTranslations = 0 twice
  //bool MERComplete = false;
  bool MERComplete = true; // DISABLE MER for now
  int SavedNewTranslations = 0;
  int MERiteration = 1;
  int TotalNewTranslations = 0;
  int SumNumberOfEdges = 0;
  
  //save the start time of the whole process
  string timeAdded;
  stringstream sst(stringstream::in | stringstream::out);
  sst<<time(NULL);
  timeAdded = sst.str();
  
  do {
    
    // Worker-Bee controls
    // arv sez: true if at least one sentence is still unprocessed
    bool workerBeeAtLeastOneSentenceUnprocessed = false;
    // true if at least one sentence was processed in current loop
    bool workerBeeAtLeastOneSentenceProcessed = false;
    // true if we've already gone through all sentences once
    bool workerBeeCompletedFirstPass = false;
    // true if we have to try to process all sentences now, even recently locked ones
    bool workerBeeGetThemAllNow = false;
    
    
    SumNumberOfEdges = 0;
    TotalNewTranslations = 0;
    
    if ( MEROptimize )
      cout << "Start iteration " << MERiteration << " of full translations" << endl;
    
    
    PrintTime( "Start of Translation" );
    
    
    //========================================
    // first the translation
    //========================================
    
    RawCounts modelBestCounts;
    if (pMetricSearch != NULL){
      //pMetricSearch->initRawCounts( NGramsConsidered );
      cout << "Ngrams considered in evaluation metric = " << pMetricSearch->GetNGramConsidered() << endl;
    }
    
    
    // we store the current scaling factors for display at the end of the translation run
    CNumericalArray< float > &CurrentFactors = FeatureWeights.GetRow( MERiteration );
    for ( int m = 0; m < ScalingFactors.size(); m++ ) {
      CurrentFactors.Append( ScalingFactors[ m ] );
    }
    
    
    TPosition S;
    CWordIndex Sf;
    
    // Worker Bee loop that makes sure this bee picks up unifinished translations
    do { 

      // arv sez: true if at least one sentence is still unprocessed
      workerBeeAtLeastOneSentenceUnprocessed = false;
      // true if at least one sentence was processed in current loop
      workerBeeAtLeastOneSentenceProcessed = false;
      
      SentenceNumber = 0;
      
      // We will loop over the same test set multiple times, as long as there
      // are untranslated lattices
      ifstream LatticeListStream( LatticeList );
      
      if ( ! LatticeListStream.is_open() ) {
        cerr << "Unable to open LatticeList file: " << LatticeList << endl
          << "Exiting...";
        exit(-1);
      }
      
      while ( ! LatticeListStream.eof() ) 
      {
        // Validate contents of lattice path file
        string latticePathname;
        getline( LatticeListStream, latticePathname );
				
				// Check for blank line
        if ( latticePathname.length() == 0 ) {
					if ( ! LatticeListStream.eof() ) {
						// Blank line not at end of file
						cerr << "Warning, blank line NOT at end of file in LatticeList file: '" 
								 << LatticeList << "'" << endl;
					}
          break;
        }

        SentenceNumber++;

				cout << "WorkerBee[ " << SentenceNumber << " ] lattice file is " << latticePathname << " of size " << latticePathname.length() << endl;

        
        // safely concatenate strings and integers with ostringstream
        ostringstream filePrefixStream;
        filePrefixStream << WorkerBeeDir << "/" << SentenceNumber;
        
        // WorkerBee files for locking and completion checks
        string Sent_LockFile = filePrefixStream.str() + ".lock";
        int lockHandle;
        
        //string Sent_NBestCandidatesFile = filePrefixStream.str() + ".cands";
        //string Sent_NBestFeaturesFile = filePrefixStream.str() + ".feats";
        string Sent_FeatsFile = filePrefixStream.str() + ".feats";
        string Sent_NBestTranslationsFile = filePrefixStream.str() + ".translations";
        
        bool Sent_Skip = BeeIsUnitInProgress(SentenceNumber, 
                                             Sent_LockFile, Sent_FeatsFile, 
                                             Sent_NBestTranslationsFile,
                                             lockHandle,
                                             workerBeeAtLeastOneSentenceUnprocessed, 
                                             workerBeeCompletedFirstPass,
                                             workerBeeGetThemAllNow);
        
        
        
        if ( ! Sent_Skip )
        {
          
          // Not skipping lattice, so load it
          CLatticeEdgeMS *pEdgePrototype = new CLatticeEdgeMS;   
          CLatticeLM *pTestLattice = new CLatticeLM( pEdgePrototype, NULL );
          delete pEdgePrototype; // the constructor clones it. who would have known?
          
          pTestLattice->SetLatticeFileFormat( LatticeFormat );
          pTestLattice->ReadFromFile( CString(latticePathname.c_str()) );
          
          // Load sentence-level phrase table, if necessary
          // Decoder already has the pointer pTranslationMemory
          // On first sentence, pTranslationMemory has no transducers, which are then removed after translation
          if ( pSentLevelPhraseTableFilePrefix ) {
            char SentLevelPhraseTableFile[ strlen(pSentLevelPhraseTableFilePrefix) + 6 ];
            sprintf( SentLevelPhraseTableFile, "%s%d", pSentLevelPhraseTableFilePrefix, SentenceNumber-1 );
            cout << "Attempting to load transducer file \"" << SentLevelPhraseTableFile << "\"" << endl;
            pTranslationMemory->LoadTransducer( SentLevelPhraseTableFile );
          }
          
          // NBest list object for sentence. Stored for MER training or evaluation
          NBestList* pNB = NULL;
          if ( MERiteration > 1 && ReferenceList != NULL ) {
            pNB = (*pNBestLists)[ SentenceNumber - 1 ];
          } else {
            pNB = new NBestList;
            pNB->Init( true, true );
            
            // store nbest list object which will be reused during evaluation or MER iterations
            if ( ReferenceList != NULL ) {
              pNBestLists->push_back( pNB );
            }
            
          }
          
          // array to hold translation results
          CArray< CDecoderResult* > TranslationResults;
          bool buildTranslationLattice = true;
          
          TranslateLattice( pTestLattice, TranslationResults, buildTranslationLattice );
          
          // keep track of how many edges were generated
          SumNumberOfEdges += pTestLattice->NumberOfEdges();
          //pLattice->DisplayEdges();
          
          // display translation results before text normalization
          if ( DisplayTranslations )
            ShowTranslationResults( SentenceNumber, TranslationResults, pTestLattice );
          
          if ( NormalizingScript || NormalizingRegexFileName ) 
            NormalizeTranslations( TranslationResults );
          
          // display translation results after text normalization
          if ( DisplayTranslationsNormalized )
            ShowTranslationResults( SentenceNumber, TranslationResults, pTestLattice );
          
          RawCounts topTranslationRaw;
          
          topTranslationRaw = BuildNBest( TranslationResults, pNB,
                                         pMetricSearch, NBest, SentenceNumber - 1, NULL, TotalNewTranslations );
          cout << "topTranslationRaw: " << topTranslationRaw << endl;
          
          modelBestCounts = modelBestCounts + topTranslationRaw;
          
          // sv ??? cout << "modelBestCounts: " << modelBestCounts << endl;
          
          // display translation results after text normalization
          //if ( DisplayTranslationsNormalized )
          //ShowTranslationResults( SentenceNumber, TranslationResults, pLattice );
          
          // get the 1best translation
          Translation = TranslationResults[ TranslationResults.MinIndex() ]->m_Translation;
          cout << "Translation is: " << Translation << endl << endl << flush;
          
          if( SaveHypothesesToFile != NULL && MERiteration == 1) {
            outHypFile << Translation << endl;
          }
          
          // Display min, max of features used to score translations
          if ( DisplayFeaturesRange ) {
            DisplayFeaturesMinMax( pTestLattice );
          }
          
          // Output Bee information to file
          
          // output NBest list here. First entry is final translation.
          ofstream Sent_NBestTranslations_stream (Sent_NBestTranslationsFile.c_str());
          ofstream Sent_Feats_stream;
          if ( WorkerBeeSaveSentenceFeatures ) {
            Sent_Feats_stream.open( Sent_FeatsFile.c_str() );
          }
          // output Translation here
          int nBestListSize = TranslationResults.Size();
          for ( int s = 0; s < nBestListSize; s++ ) 
          {
            Sent_NBestTranslations_stream << s+1 << " " << TranslationResults[ s ]->m_Translation << endl;
            
            // Output raw sentence feature scores if option is selected
            if ( WorkerBeeSaveSentenceFeatures ) 
            {
              Sent_Feats_stream << s+1 << " "; 
              vector<float>::iterator i_feats = TranslationResults[ s ]->m_Features.begin();
              vector<float>::iterator i_scalingFactors = ScalingFactors.begin();
              for ( ; i_feats !=  TranslationResults[ s ]->m_Features.end() &&
                   i_scalingFactors != ScalingFactors.end(); 
                   i_feats++, i_scalingFactors++ ) {
                Sent_Feats_stream << (*i_feats) / (*i_scalingFactors) << " ";
              }
              
              Sent_Feats_stream << endl;
            }
          }
          // Close NBest List and Features List files
          if ( WorkerBeeSaveSentenceFeatures ) {
            Sent_Feats_stream.close();
          }
          Sent_NBestTranslations_stream.close();
          
          // Unlock sentence with remove(), which deletes sentence's .lock file
          remove( Sent_LockFile.c_str() );
          // Free OS's file descriptor
          if ( close(lockHandle) < 0 ) {
            cout << "WorkerBee[ " << SentenceNumber << " ]: Unable to deallocate lock file descriptor!" << endl;
          } else {
            cout << "WorkerBee[ " << SentenceNumber << " ]: Successfully deallocated lock file descriptor." << endl;
          }
          
          // Clean Memory after sentence translation
          
          // clear decoder
          pDecoder->Clear();
          
          //clear translation results
          for ( int t = TranslationResults.MinIndex(); t <= TranslationResults.MaxIndex(); t++ ) {
            delete TranslationResults[ t ];
          }
          TranslationResults.Clear();
          
          // clear NBestList if no references are used
          if ( ReferenceList == NULL ) {
            delete pNB;
          }
          
          // delete translation lattice if we're rebuilding it every time
          delete pTestLattice;
          
          // clean up sentence-level phrase table, if used
          if ( pSentLevelPhraseTableFilePrefix ) {
            pTranslationMemory->ClearTransducer();
          }
          
          
        }// Bee translation. Skipped if sentence was translated.
        
      }//end all source sentences
      
      // Close ifstream to file of lattice paths
      LatticeListStream.close();
      
      // This bee still needs to make sure that everything has been translated
      workerBeeCompletedFirstPass = true;     
      workerBeeGetThemAllNow = workerBeeCompletedFirstPass && !workerBeeAtLeastOneSentenceProcessed;
      cout << "WorkerBee[ PASS_COMPLETED ]: Status after pass: any unprocessed: "
      << workerBeeAtLeastOneSentenceUnprocessed << ", any processed: " << workerBeeAtLeastOneSentenceProcessed << endl;
      
    } while ( workerBeeAtLeastOneSentenceUnprocessed );
    
    
    PrintTime( "End of Translation" );
    
    /* removed for bee version
     if( SaveHypothesesToFile != NULL && MERiteration == 1) {
     outHypFile.close();    
     cerr << "Wrote translations to: " << SaveHypothesesToFile << endl;
     }
     */
    cout << "TotalNewEdges:       " << SumNumberOfEdges << endl;
    cout << "TotalNewTranslations:" << TotalNewTranslations << endl;
    
    
    // now the scoring
    //ScoreTranslation( modelBestCounts, trackScalingFactorScores, MERiteration );
    
    cout << "=============================================" << endl;
    
    // and here comes the optimization part
    if ( MEROptimize ) {
      
      // after the last translation we don't optimize again, as we would not
      // have a verification if the resulting scaling factors are really improving the translation
      if ( MERiteration < IterationLimit ) {
        
        //this criteria is probably a bit too strict on its own. so we also check to make sure the metric score hasnt stagnated. we can
        //check this by looking at the score tracker
        if ( TotalNewTranslations == 0 ) {
          MERComplete = true;
          cout << "Optimization: Early termination triggered since TotalNewTranslations == 0" << endl;
        }
        
        PrintTime( "Start of Optimization" );
        
        float FinalScore = OptimizeParameters( ScalingFactors, *pNBestLists, pMetricSearch,
                                              trackScalingFactorScores, timeAdded );
        FinalScores[ MERiteration + 1 ] = FinalScore;
        
        PrintTime( "End of Optimization" );
        
        // now we have new scaling factors and give them back to the decoder and the phrase aligner
        pDecoder->SetScalingFactors( ScalingFactors );
        UpdatePhraseAlignerScalingFactors();
        
        cout << "End of Iteration " << MERiteration << " out of " << IterationLimit << endl;
      } // if ( MERiteration < IterationLimit )
      
    } else {
      // no optimization, i.e. set flag to end the loop
      MERComplete = True;
    }
    
    MERiteration++;
    
    //as long as there is optimization, its not finished and 
    //we have more iterations left to go
    
  } while ( MERiteration <= IterationLimit && !MERComplete );
	
  DisplayScalingFactors( trackScalingFactorScores );
  
  //========================================
  // we are done - cleanup
  //========================================
  
  Trace( 1, "Cleanup of data structures" );
  
}




void
TranslateLattice( CLatticeLM *&pLattice,
                  CArray<CDecoderResult*> &TranslationBank,
                  bool buildTranslationLattice )
{
  int DisplayLattice           = *((int*)pParam->getValueFromKey( "DisplayLattice" ));
  int DisplayLatticeStatistics = *((int*)pParam->getValueFromKey( "DisplayLatticeStatistics" ));


  cout << "TranslateLattice" << endl;
  
  if( buildTranslationLattice ) { 
    BuildTranslationLattice( pLattice ); 
  }

  if ( DisplayLattice ) {
    pLattice->DisplayEdges();
  }
  
  if ( DisplayLatticeStatistics ) {
    pLattice->DisplayStatistics();
  }
  
  // give source sentence to decoder
  //pDecoder->SetSourceSentence( pPartialSentence );
  
  // finally, apply language model; results are stored in TranslationBank
  TranslationBank = pDecoder->ApplyLanguageModel( pLattice );
  
  // line below is in new TranlsateSentence; how should this be treated?
  //RescoreNBestList( pSentence, TranslationResult, ScalingFactors );
  
}

/////////////////////////////////////////////////////////////
//Nguyen, Aug 6, 2008: Cohesive Decoding

void
TranslateSentenceWithDependencyTree( CSentence *pSentence, CLatticeLM *&pLattice,
                   CArray< CDecoderResult* > &TranslationResult, vector<tokens>& sentDtree, int algorithm )
{
  int DisplayLattice           = *((int*)pParam->getValueFromKey( "DisplayLattice" ));
  int DisplayLatticeStatistics = *((int*)pParam->getValueFromKey( "DisplayLatticeStatistics" ));

  CString W;
  
  cout << "DEBUG - TranslateSentenceWithDependencyTree: " <<  getNameFromVector(sentDtree) << endl;

  // if we have the lattice from a previous iteration, we use it
  if ( pLattice ) {
    Trace( 1, "Reuse lattice from previous iteration" );
  } else {
    // convert sentence into a lattice
    Trace( 1, "Create new lattice" );
    
    //pLattice = pDecoder->SentenceToLattice( pPartialSentence, pVoc );
    //pLattice = pSentenceToLattice->CreateLatticeFromSentence( pPartialSentence, pVoc, CSentenceToLattice::Flat );

    // create a lattice were tags are added as edge with annotation, but no edges are generated for the source words
    // under the tag
    pLattice = pSentenceToLattice->CreateLatticeFromSentence( pSentence, pVoc, CSentenceToLattice::Flat );
    
    // find additional word form and add them into the lattice
    ExpandSourceSentence( pLattice );

    // build translation lattice:
    // apply transducer, then search for phrase translations online
    //BuildTranslationLattice(  pPartialSentence, pLattice );
    
    BuildTranslationLattice(  pSentence, pLattice );
    
    //pLattice->Display();
  } // if ( lattice needs to be created )
  
  
  if ( DisplayLattice ) {  	
    pLattice->DisplayEdges();
  }
  
  if ( DisplayLatticeStatistics ) {
    pLattice->DisplayStatistics();
  }
  
  
  pDecoder->SetSourceSentence( pSentence );
  
  // sv: deactivated for now
  // InitFutureCostEstimator( pPartialSentence, pLattice);
	
  // finally, apply language model; results are stored in TranslationBank
  
  cout << "DEBUG - ApplyLanguageModel" << endl;
  //TranslationResult = pDecoder->ApplyLanguageModel( pLattice );
  
  //Nguyen, Aug 7, 2008: Cohesive Decoding
  TranslationResult = pDecoder->ApplyLanguageModelWithDependencyTree( pLattice, sentDtree, algorithm );
  
  
  cout << "DEBUG - RescoreNBestList" << endl;
  RescoreNBestList( pSentence, TranslationResult, ScalingFactors );

  //delete pPartialSentence;
}

/////////////////////// End of Nguyen ////////////////////////////


void
TranslateSentence( CSentence *pSentence, CLatticeLM *&pLattice,
                   CArray< CDecoderResult* > &TranslationResult  )
{
  int DisplayLattice           = *((int*)pParam->getValueFromKey( "DisplayLattice" ));
  int DisplayLatticeStatistics = *((int*)pParam->getValueFromKey( "DisplayLatticeStatistics" ));

  CString W;
  //CSentence *pPartialSentence = new CSentence;

  
  // originally, we could split a long sentence into shorter segments for translation
  // however, this does not work with storing lattices and nbest lists
  // we still need to copy the sentence, as this copy operation shifts the MinIndex by -1
  // and the phrase aligner is sensitive to this MinIndex.
  /*
    for ( int i = pSentence->MinIndex(); i <= pSentence->MaxIndex(); i++ ) {
    W = pVoc->GetWord( pSentence->GetAt( i ) );
    pPartialSentence->Append( pSentence->GetAt( i ) );
    CWordInfo *pWordInfo = pSentence->WordInfo( i );
    if ( pWordInfo ) {
    // cout << (*pWordInfo) << endl;
    pPartialSentence->SetWordInfo( pWordInfo, pPartialSentence->MaxIndex() );
    }
    }
    pPartialSentence->Display( pVoc, True );
    cout << endl;
  */

  /*
    int SentenceNumberInCorpus = SourceSentenceIsInCorpus( pSentence );
    if ( SentenceNumberInCorpus > 0 ) {
    cout << "TRANSLATIONRESULT_EXACTMATCH_SOURCETEST " << SentenceNumber << " ";
    pSentence->Display( pVoc ); cout << endl;
    cout << "TRANSLATIONRESULT_EXACTMATCH_SOURCECORPUS " << SentenceNumber << " ";
    pTextF->DisplaySentence( SentenceNumberInCorpus ); cout << endl;
    cout << "TRANSLATIONRESULT_EXACTMATCH_TARGETCORUP " << SentenceNumber << " ";
    pTextE->DisplaySentence( SentenceNumberInCorpus ); cout << endl;
    
    }
  */



  // if we have the lattice from a previous iteration, we use it
  if ( pLattice ) {
    Trace( 1, "Reuse lattice from previous iteration" );
  } else {
    // convert sentence into a lattice
    Trace( 1, "Create new lattice" );
    
    //pLattice = pDecoder->SentenceToLattice( pPartialSentence, pVoc );
    //pLattice = pSentenceToLattice->CreateLatticeFromSentence( pPartialSentence, pVoc, CSentenceToLattice::Flat );

    // create a lattice were tags are added as edge with annotation, but no edges are generated for the source words
    // under the tag
    pLattice = pSentenceToLattice->CreateLatticeFromSentence( pSentence, pVoc, CSentenceToLattice::Flat );
    
    // find additional word form and add them into the lattice
    ExpandSourceSentence( pLattice );

    // build translation lattice:
    // apply transducer, then search for phrase translations online
    //BuildTranslationLattice(  pPartialSentence, pLattice );
    
    BuildTranslationLattice(  pSentence, pLattice );
    
    //pLattice->Display();
  } // if ( lattice needs to be created )
  
  
  if ( DisplayLattice ) {  	
    pLattice->DisplayEdges();
  }
  
  if ( DisplayLatticeStatistics ) {
    pLattice->DisplayStatistics();
  }
  
  // add translation edges for words which have high probability to 
  // align to the null word, i.e. are not translated at all
  //CEmptyWordTranslation EWT;
  //EWT.SetLexicon( pLexiconFE );
  //EWT.Generate( pLattice );

  // give source sentence to decoder
  //pDecoder->SetSourceSentence( pPartialSentence );
  
  pDecoder->SetSourceSentence( pSentence );
  
  // sv: deactivated for now
  // InitFutureCostEstimator( pPartialSentence, pLattice);
	
  // finally, apply language model; results are stored in TranslationBank
  //cout<<"Decoding starting now"<<endl;
  cout << "DEBUG - ApplyLanguageModel" << endl;
  TranslationResult = pDecoder->ApplyLanguageModel( pLattice );
  
  cout << "DEBUG - RescoreNBestList" << endl;
  RescoreNBestList( pSentence, TranslationResult, ScalingFactors );

  //delete pPartialSentence;
}



// ====================================================================
// Translation support functions
// ====================================================================
void
BuildTranslationLattice( CLatticeLM *pLattice )
{
  cout << "Build Translation Lattice" << endl;
  
  if ( pTranslationMemory ) {
    cout << "Apply Transducer" << endl;
    pDecoder->BuildTranslationLattice( pLattice );
  }  
  
  /* //not doing online phrase search 
     // and finally, the online phrase search
     if ( pSA ) {
     cout << "Search phrase pairs from corpus" << endl;
     pPhraseSearcher->FindTranslations( pSentence, pLattice );
     }
  */
  
  AddFertilityScores( pLattice );
}

void
BuildTranslationLattice( CSentence *pSentence, CLatticeLM *pLattice )
{
  cout << "Build Translation Lattice" << endl;

  int SA_ShortestPhraseLength = *( (int*)pParam->getValueFromKey( "SA_ShortestPhraseLength" )  );

  bool FullMatchOnly = false;   // search for all ngrams

  if ( SA_ShortestPhraseLength < 0 ) {
    FullMatchOnly = true;
  }


  if ( pTranslationMemory ) {
    cout << "- Apply Transducer" << endl;
    pDecoder->BuildTranslationLattice( pLattice );
  }
  
  // we use the SMT transducers for which scores will be calculated anew
  /*
    if ( RescoreTransducer ) {
    cout << "Apply Transducer" << endl;
    if ( !pLexiconFE ) {
	  cout << "Rescoring of tranducer requested, but no lexicon loaded" << endl;
	  exit( -1 );
    } else {
	  cout << "Use Transducers - probabilities calculated from loaded lexicon" << endl;
	  pDecoder->BuildTranslationLattice( pLattice );
    }
    }
  */


  /*
    if ( RescoreTransducer && pLexiconFE ) {
    pLatticeRescorer->Rescore( pLattice, pLexiconFE );
    }
  */
  
  // now we check which partial translations we have
  // as we don't want to search for them online
  // a translation is indicated by eigther Word starting with '@'
  // or by having a WordAnnotation structure

  // BUG: boundaries are too low by 1
  // has to do with mismatch between original sentence (starting from index 1)
  // and partial sentence after splitting (starting from index 0)

  cout << "check existing translations from phrase table" << endl;

  CMatrix< short > *pExistingTranslations = pPhraseSearcher->GetExistingTranslationsTable();
  pExistingTranslations->SetBoundaries( pSentence->MinIndex() - 1, pSentence->MaxIndex(),
                                        pSentence->MinIndex() - 1, pSentence->MaxIndex() );
  pExistingTranslations->SetAll( 0 );
  
  const CArray< CLatticeEdge * > &ExistingEdges = pLattice->Edges();
  CLatticeEdgeMS *pEdge;
  
  if ( ExistingEdges.Size() > 0 ) {
    for ( int i = ExistingEdges.MinIndex(); i <= ExistingEdges.MaxIndex(); i++ ) {
      pEdge = (CLatticeEdgeMS *)ExistingEdges[ i ];
      //pEdge->Display();

      // if this is not the source word, but actually a translation
      // increment the counter for the translation between the two nodes
      /*
        if ( pEdge->WordInfo() ) {
        if ( pEdge->WordInfo()->HasTargetString() ) {
        cout << " is translation" << endl;
        } else {
        cout << " is source word" << endl;
        }
        } else {
        cout << " is source word" << endl;
        }
      */

      int FromNode = pEdge->FromNode()->Number();
      int ToNode = pEdge->ToNode()->Number();
      short Count = pExistingTranslations->GetAt( FromNode, ToNode );
      pExistingTranslations->SetAt( FromNode, ToNode, Count + 1 );
    }
  } else {
    ; //cout << "No edges in lattice" << endl;
  }
  
  
  // and finally, the online phrase search
  if ( pSA && pLexiconFE ) {
    cout << "Search phrase pairs from corpus" << endl;

    if ( FullMatchOnly ) {
      if ( pSentence->Length() < 250 )
        pSA->setParam_shortestUnitToReport( pSentence->Length() );
      else
        pSA->setParam_shortestUnitToReport( 250 );
      cout << "Translate exactly matching phrases or sentences" << endl;
    }

    pPhraseSearcher->FindTranslations( pSentence, pLattice );

  }
  //pLattice->Display();
  
  
  AddFertilityScores( pLattice );
}



int
SourceSentenceIsInCorpus( CSentence *pSentence )
{
  // need suffix array to do this
  if ( !pSA ) 
    return -1;


  cout << "Search for exactly matching sentences" << endl;

  ostringstream StrStream;
  pSentence->Write( StrStream, pVoc );
  StrStream << '\0';
  //cout << StrStream.str() << endl;
  
  string tmp = StrStream.str();
  cout << "String for SA: " << tmp << endl;
  
  vector< S_phraseLocationElement > PhraseLocations;
  
  if ( pSentence->Length() >= 255 ) {
    cout << "sentence is too long for giving it the the SA" << endl;
    return -1;
  } else {
    cout << "Shortest and longest phase to be reported by SA is now " << pSentence->Length() << endl;
    pSA->setParam_shortestUnitToReport( pSentence->Length() );
    //pSA->setParam_longestUnitToReport( pSentence->Length() );
    cout << "Translate exactly matching phrases or sentences" << endl;
    
    PhraseLocations = pSA->findPhrasesInASentence( (char *)tmp.c_str() );
    cout << PhraseLocations.size() << " exact matching sentences in training corpus found" << endl;

    for ( int n = 0; n < PhraseLocations.size(); n++ ) {
      //DisplayPhraseLocation( PhraseLocation[ n ] );
      
      // search phrase translations only for segments for which
      // we don't have translations from transducer
      short StartPosPhrase = PhraseLocations[ n ].posStartInSrcSent;  // FromNode is before first word
      short EndPosPhrase   = PhraseLocations[ n ].posEndInSrcSent;
      long  SentenceID     = PhraseLocations[ n ].sentIdInCorpus;
      short PosInSentence  = PhraseLocations[ n ].posInSentInCorpus;
      
      return SentenceID - 1;
      // is it matching the entire sentence ?
      /*
        CSentence *pSentenceF = pTextF->GetSentence( SentenceID - 1 );
        if ( pSentenceF->Length() == pSentence->Length() ) {
        cout << "Found sentence in corpus: " << SentenceID << " " << StartPosPhrase << " " << EndPosPhrase << endl;
        return SentenceID - 1;
        } else {
        cout << "matching partial sentence" << endl;
        }
      */

    }
  }

  return -1;
}




void
ExpandSourceSentence( CLattice *pLattice )
{

  /* 
     for each source word, which is in the lattice (this function is called
     right after the source sentence has been converted into a lattice)
     we call an external script which returns the word together with alternative
     word forms.  For the additional word forms additional edges are added
     to the lattice.

  */


  // if a WordExpander has been initialized, use it
  if ( pWordExpander ) {

    cout << "Expand source sentence: add additional word forms" << endl;

    const CArray< CLatticeEdge * > &AllEdges =  pLattice->Edges();
    CLatticeEdgeMS *pEdge, *pNewEdge;

    int n = AllEdges.MinIndex();  
    int MaxIndex = AllEdges.MaxIndex();
    
    while ( n <= MaxIndex ) {
      pEdge = (CLatticeEdgeMS *)AllEdges[ n ];
      CString Word = pEdge->Word();
      CString AdditionalWordForms = pWordExpander->Filter( Word );
      //cout << "Original Word = " << Word << endl;
      //cout << "Additional Word Forms = " << AdditionalWordForms << endl;

      CArray< CString > OtherWordForms;
      AdditionalWordForms.Split( OtherWordForms, ' ' );

      int i = OtherWordForms.MinIndex() + 1;  // we assume the word itself has been return, don't use it
      while ( i <= OtherWordForms.MaxIndex() ) {
        pNewEdge = (CLatticeEdgeMS *)pEdge->Clone();
        pNewEdge->SetWord( OtherWordForms[ i ] );
        //cout << "additional word form " << OtherWordForms[ i ] << endl;
        //cout << "Additional edge: ";
        //pNewEdge->Display();
        //cout << endl;
        pLattice->AddEdgeUniq( pNewEdge );
        i++;
      }

      n++;
    }

  }

}



void
ScoreTranslation( RawCounts &modelBestCounts, TScoreScalingFactorMap &trackScalingFactorScores, int MERiteration )
{
  //========================================
  // now the scoring
  //========================================

  cout << "==============================================="<<endl;
  cout << "Scoring this translation run" << endl;

  int   GenOracleScores               = *( (int*)  pParam->getValueFromKey( "GenOracleScores" ) );


  // sv ??? cout << "modelBestCounts = " << modelBestCounts << endl;

  if ( pMetricSearch != NULL ) {
    float TranslationScore = pMetricSearch->getMetricScore( modelBestCounts );
      
    trackScalingFactorScores.insert( pair< float, vector< float > >( TranslationScore, ScalingFactors ) );
    cout << "TrainTranslationScore: " << TranslationScore << endl;
	  MetricScores[ MERiteration ] = TranslationScore;
    
    if ( GenOracleScores ) {
      // not available any more?
      //Configuration config = pMetricSearch->GetOracleConfiguration( nBestLists );
      //cout << "OracleTranslationScore: " << pMetricSearch->ScoreConfiguration( nBestLists, config )[ 0 ] << endl;
      //build up RawCounts vectors
      AllSentRawCountsVector accumRawCounts;
      for( vector< NBestList* >::iterator iter=pNBestLists->begin(); iter!=pNBestLists->end(); ++iter ) {
        NBestList* pNB = (*iter);
        accumRawCounts.push_back( pNB->GetRawCounts() );
      }
      //pass this over the metric scorer
      Configuration oracleConf = pMetricSearch->GetOracleConfiguration( accumRawCounts );
      //the ScoreConfiguration return an AbstractScore - a vector of scores, if we wanted to
      //we could get the length penalty for example, but this is not implemented
      float OracleScore = pMetricSearch->ScoreConfiguration( accumRawCounts, oracleConf )[ 0 ];
      cout << "OracleScore = " << OracleScore << endl;
      delete [] oracleConf;

      OracleScores[ MERiteration ] = OracleScore;
    
    }
  } else {
    cout << "No scoring possible since metric search object was not initialized" << endl;
    cout << "probably because no references were provided" << endl;
  }
    
  
}




void
RescoreNBestList( CSentence *pSourceSentence, CArray< CDecoderResult * > &TranslationResults,
                  vector< float > &ScalingFactors )
{
  if ( !pRescorerSourceTargetIBM1 || !pRescorerTargetSourceIBM1 )
    return;


  cout << "Rescore NBest List" << endl;
  cout << "Use scaling factors " << ScalingFactors[ ScalingFactors.size() - 2 ] << " and  "
       << ScalingFactors[ ScalingFactors.size() - 1 ] << endl;

  cout << "Best Translation is: " << TranslationResults[ TranslationResults.MinIndex() ]->m_Translation 
       << " " << TranslationResults[ TranslationResults.MinIndex() ]->m_TotalScore << endl;

  // a multiset to temporarily store the update translation results
  multiset< CDecoderResult *, DecoderResultLess > TranslationsRescored;


  CSentence *pTranslation = new CSentence;

  for ( int n = TranslationResults.MinIndex(); n <= TranslationResults.MaxIndex(); n++ ) {
    //CString Translation = TranslationResults[ n ]->m_translation;
    //CString FilteredTranslation = pTextFilter->Filter( Translation );
    //TranslationResults[ n ]->m_translation = FilteredTranslation;
    CString Translation = TranslationResults[ n ]->m_Translation;
    pTranslation = new CSentence( Translation, pVoc );
    //TranslationResults[ n ]->GetTranslation( pTranslation, pVoc );
    
    /*
      cout << "Rescore with IBM1 lexicons:" << endl;
      cout << "Source Sentence: "; pSourceSentence->Display( pVoc ); cout << endl;
      cout << "Target Sentence: "; pTranslation->Display( pVoc ); cout << endl;
      cout << "TotalScore now:  " << TranslationResults[ n ]->m_TotalScore << endl;
    */

    float LogProbSourceTarget = -pRescorerSourceTargetIBM1->LogProb( *pSourceSentence, *pTranslation );
    float LogProbTargetSource = -pRescorerTargetSourceIBM1->LogProb( *pTranslation, *pSourceSentence );

    // multiply with scaling factors (how do we know which to use?)
    LogProbSourceTarget *= ScalingFactors[ ScalingFactors.size() - 2 ];
    LogProbTargetSource *= ScalingFactors[ ScalingFactors.size() - 1 ];

    //cout << "LogProbSourceTarget = " << LogProbSourceTarget << endl;
    //cout << "LogProbTargetSource = " << LogProbTargetSource << endl;

    // add it the the scores of the translation
    // need to scale it !!!
    int i = TranslationResults[ n ]->m_Features.size();
    TranslationResults[ n ]->m_Features[ i - 2 ] = LogProbSourceTarget;
    TranslationResults[ n ]->m_Features[ i - 1 ] = LogProbTargetSource;
    TranslationResults[ n ]->m_TotalScore += (LogProbSourceTarget + LogProbTargetSource);

    // insert into map for sorting
    TranslationsRescored.insert( TranslationResults[ n ] );


    delete pTranslation;
    //cout << "Normalized: " << FilteredTranslation << endl;
  }

  cout << "After rescoring:     " << (*TranslationsRescored.begin())->m_Translation 
       << " " << (*TranslationsRescored.begin())->m_TotalScore << endl;

  // write back into the TranslationResults array
  
  multiset< CDecoderResult *>::iterator Iter;
  int n = TranslationResults.MinIndex();
  for ( Iter = TranslationsRescored.begin(); Iter != TranslationsRescored.end(); ++Iter ) {
    //cout << (*Iter)->m_TotalScore << " " << (*Iter)->m_Translation << endl;
    TranslationResults[ n ] = *Iter;
    //TranslationResults[ n ]->Display(); cout << endl;
  }
  


}





void
NormalizeTranslations( CArray< CDecoderResult* > &TranslationResults )
{
  cout << "Normalize Translation Results" << endl;


  for ( int n = TranslationResults.MinIndex(); n <= TranslationResults.MaxIndex(); n++ ) {
    CString Translation = TranslationResults[ n ]->m_Translation;
    CString FilteredTranslation = pTextFilter->Filter( Translation );
    TranslationResults[ n ]->m_Translation = FilteredTranslation;

    //cout << "Normalized: " << FilteredTranslation << endl;
  }
}



void
UpdatePhraseAlignerScalingFactors()
{
  float PhraseAlignDynamicFactors = *((float*)pParam->getValueFromKey( "PhraseAlignDynamicFactors" ));
  if ( PhraseAlignDynamicFactors > 1.0 ) {
    PhraseAlignDynamicFactors = 1.0;
  }

  cout << "Update phrase align scaling factors in optimization: " << PhraseAlignDynamicFactors << endl;

  if ( PhraseAlignDynamicFactors ) {
    CNumericalArray< float > PhraseAlignFactors = pPhraseAligner->GetScalingFactors();

    for ( int i = 1; i <= pPhraseAligner->GetNumberOfScores(); i++) {
      cout << "Interpolate phrase align scaling factor " << i 
           << " from " <<  PhraseAlignFactors[ i ]
           << " and " << ScalingFactors[ i + 3 ];

      PhraseAlignFactors[ i ] 
        = ( 1 - PhraseAlignDynamicFactors ) * PhraseAlignFactors[ i ]
        + PhraseAlignDynamicFactors * ScalingFactors[ i + 3 ];

      cout << " to " <<  PhraseAlignFactors[ i ] << endl;
    }
    cout << endl;

    pPhraseAligner->SetScalingFactors( PhraseAlignFactors );
  }
}	  



void
AddFertilityScores( CLattice *pLattice )
{
  // add fertility scores
  if ( pFertilityFE && pFertilityEF ) {
	
    const CArray< CLatticeEdge * > &AllEdges =  pLattice->Edges();
    for ( int n = AllEdges.MinIndex(); n <= AllEdges.MaxIndex(); n++ ) {
      CLatticeEdgeMS *pEdge = (CLatticeEdgeMS *)AllEdges[ n ];
      CWordInfo *pWordInfo = pEdge->WordInfo();
      CSentence *pSourcePhrase = &(pWordInfo->SourceWords());
      CSentence *pTargetPhrase = &(pWordInfo->TargetWords());
      int SourceLength = pSourcePhrase->Length();
      int TargetLength = pTargetPhrase->Length();
	  
      if ( TargetLength > 0 ) {
        CNumericalArray< float > &Scores = pEdge->Scores();
		
		
        // phrase length prob F->E
        pPhraseLengthModelFE->GenerateTargetLengthProbabilities( pSourcePhrase );
        float TargetLengthProb = pPhraseLengthModelFE->GetPr( TargetLength );
        pEdge->Scores().Append( -log ( TargetLengthProb ) );
        
        // phrase length prob E->F
        pPhraseLengthModelEF->GenerateTargetLengthProbabilities( pTargetPhrase );
        float SourceLengthProb = pPhraseLengthModelEF->GetPr( SourceLength );
        pEdge->Scores().Append( -log ( SourceLengthProb ) );
		
		
        // QUICK HACK: now we add a place holder for lexicon rescoring
        //pEdge->Scores().Append( 0.0 );
		
        /*
          pSourcePhrase->Display( pVoc ); 
          cout << " ";
          pTargetPhrase->Display( pVoc );
          cout << " LengthProbs = " << TargetLengthProb << " " << SourceLengthProb;
          cout << Scores << endl;
        */
      }
    } 
  } // if fertility models have been loaded
  
}




// ====================================================================
// Display functions
// ====================================================================

void
ShowTranslationResults( int SentenceNumber, CArray<CDecoderResult*>& TranslationResults, CLatticeLM *pLattice )
{
  int DisplayNBestData       = *((int*)pParam->getValueFromKey( "DisplayNBestData" ));
  int DisplayTranslationPath = *((int*)pParam->getValueFromKey( "DisplayTranslationPath" ));
  
  //cout << "DisplayNBestData is: " << DisplayNBestData << endl;
  
  //TranslationBank[ 0 ]->Display( "Top Translation Is:", 0, ScalingFactors, pLattice );
  //cout<<"Top Translation is:"<<TranslationBank[TranslationBank.MinIndex()]->m_translation<<endl;
  
  //cout << "DisplayNBestData " << DisplayNBestData << endl;
  if ( DisplayNBestData > 0 ) {
    cout << "Display Nbest Translations" << endl;
    // nbest list is stored 0-based, therefore the -1
    DisplayNBestData--;
    int MaxN = DisplayNBestData < TranslationResults.MaxIndex() ? DisplayNBestData : TranslationResults.MaxIndex();
    for ( int NBestNumber = TranslationResults.MinIndex(); NBestNumber <= MaxN; NBestNumber++ ) {
      TranslationResults[ NBestNumber ]->Display( SentenceNumber, NBestNumber, ScalingFactors, pLattice, DisplayTranslationPath );
    }
  }
  
}


void
ApplyTrueCasing( CArray< CDecoderResult* > TranslationResults )
{
  cout << "ApplyTrueCasing" << endl;

  // loop over all translations
  CDecoderResult *pTranslationResult;

  int i = TranslationResults.MinIndex();
  while ( i <= TranslationResults.MaxIndex() ) {

    // we need the model scores
    pTranslationResult = TranslationResults[ i ];
    CString Translation = pTranslationResult->m_Translation;

    // split the string and convert into array of word IDs, into a CSentence
    CSentence LowerCasedTranslation = CSentence( Translation, pVoc );

    cout << "Lower cased: "; LowerCasedTranslation.Display( pVoc ); cout << endl;

    // 'translate' this with the true caser

    // replace the original translation by the true-cased translation

    // add the true-casing scores to the scores of the translation


    i++;
  }

}


RawCounts
BuildNBest( CArray< CDecoderResult* > TranslationResults, NBestList* oneNBestList,  MetricSearch* pMetric,
            int translationCount, int sentenceNumber, CSentence* pSourceSentence, int &TotalNewTranslations )
{

  cout << "BuildNBest list from TranslationResults, also calculate ngram matches" << endl;
  cout << "Number of translations = " << TranslationResults.Size() << endl;
  cout << "translationCount = " << translationCount << " sentenceNumber = " << sentenceNumber << endl;
  cout << "NBestListSize = " << oneNBestList->Size() << endl;


  int numFeatures;

  // we return the results for the best translations, as this will be used to caclulate the translation score
  // over the entire test set in the current optimization iteration
  RawCounts topTranslationRaw;

  // loop over all translations
  CDecoderResult *pTranslationResult;

  int numTranslationsAdded = 0;
  int i = TranslationResults.MinIndex();
  while ( i <= TranslationResults.MaxIndex() ) {

    // we need the model scores
    pTranslationResult = TranslationResults[ i ];
    CString Translation = pTranslationResult->m_Translation;
  
    /*
      if ( normalizeTranslations ) {
      CString NormalizedTranslation = pTextFilter->Filter( Translation );
      //cout << "Raw translaton:        " << Translation << endl;
      //cout << "Normalized translaton: " << NormalizedTranslation << endl;

      pTranslationResult->m_translation = NormalizedTranslation;
      cout << "Normalized translation: " << TranslationResults[ i ]->m_translation << endl;

      } else {
      //cout << "WARNING: No Text Normalization" << endl;
      }
    */


    /*
      cout << i << " Translation:  " << Translation
      << " Model scores: ";
      for ( int n = 0; n < pTranslationResult->m_features.size(); n++ ) {
      cout << pTranslationResult->m_features[ n ] << " " ;
      }
      cout << endl;
      
      cout << "Ngrams considered = " << pMetric->GetNGramConsidered() << endl;
    */
    
    //RawCounts raw = pMetric->initRawCounts( pMetric->GetNGramConsidered() );
    RawCounts raw; //it should be initialized immediately
    
	
    // here, sentence numbering starts from 1, in metric it starts from 0, therefore -1
    if ( pMetric )
      pMetric->ScoreSingleHyp( sentenceNumber, (const char *)(Translation), raw );
    
    /*
      for ( int j = 0; j < NSZ; j++ ) {
      cout << "ngram =" << j << " Score=" << raw.numCorrect[ j ] << "/" << raw.numSuggested[ j ] << "  ";
      }
      cout << endl;
    */
    
    // store the raw counts for the first translation, as we want to return it
    if ( i == TranslationResults.MinIndex() ) {
      topTranslationRaw = raw;
    }
    

    // Insert one candidate into the current nbest list: model scores, translation as word ids, and metric scores
    // model scores from pTranslationResult->m_features
    // translation from pTranslationResult->m_translation (is string)
    // metric scores from raw counts

	
    CSentence TranslationAsWordIDs;
    TranslationAsWordIDs.ReadFromBuffer( (const char *)Translation, pVoc );
    //cout << Translation << endl;
    //TranslationAsWordIDs.CArray< CWordIndex >::Display();
    
    // now, we need to convert the CArray into a vector (this programming all over the place has to stop!)
    vector< CWordIndex > TranslationAsWordIndexVector;
    for ( int j = TranslationAsWordIDs.MinIndex(); j <= TranslationAsWordIDs.MaxIndex(); j++ ) {
      TranslationAsWordIndexVector.push_back( TranslationAsWordIDs[ j ] );
    }
	
    //SCALE IT (divide)
    pTranslationResult->ScaleFeatures( ScalingFactors, true );

    numTranslationsAdded += oneNBestList->InsertCandidate( pTranslationResult->m_Features,
                                                           TranslationAsWordIndexVector, raw );
    //cout << numTranslationsAdded << "translations added to the global nbest list, size is now: " << 
    //          oneNBestList->Size() << endl;
    
    i++;
  }
  
  cout << "numTranslationsAdded: "  << numTranslationsAdded 
       << " NBestListSize is now: " << oneNBestList->Size() << endl;
    
  //increment the number of new translations
  TotalNewTranslations += numTranslationsAdded;

  // display n-best list
  //oneNBestList->PrintOptimizationData( sentenceNumber, 5, cout, cout );
  

  return topTranslationRaw;
}



float
OptimizeParameters( vector< float >& ScalingFactors, vector< NBestList* >& nBestLists, 
                    MetricSearch* pMetric, 
                    TScoreScalingFactorMap& trackScalingFactorScores, string timeAdded )
{
  cout << "Internal MER Optimization" << endl;

  // harvest the feats data
  vector< LogProbList* > allFeats;
  vector< RawCounts* > allScores;
  vector< int > allCands;

  // copy all necessary information from n-best lists to those tables used by the optimizer
  for ( vector< NBestList* >::iterator iter = nBestLists.begin(); iter != nBestLists.end(); ++iter ) {
    NBestList* oneSent = (*iter);
    // misnomer: not the nbest list is loading data; nbest list is providing that data
    oneSent->LoadOptimizationData( allFeats, allScores, allCands );
    //cout << "after LoadOptimizationData we have sizes of arrays: " << allFeats->size() 
    //<< " " << allScores->size() << " " << allCands->size() << endl;
  }
  
  
  cout << "Running optimization" << endl;

  // we initialize the optimization with those scaling factors which so far had given
  // the best scores; might not be the one from the last iteration
  pair< float, vector< float > > pr = (*(trackScalingFactorScores.rbegin()));
  cout << "Initializing with scaling factors scoring: " << pr.first << endl;
  
  vector< double > initialValuesDoubles( pr.second.size(), 0 );
  copy( pr.second.begin(), pr.second.end(), initialValuesDoubles.begin() );

  // the actual optimization; caclualte the optimimal feature weights: 
  // minCost = minimal model scores (these are costs)
  // MaxScore = maximal metric cost
  float FinalScore;
  Lambdas returnLambdas = pOptimizerMER->minModelCostMaxMetricScore( allFeats, allScores, allCands, initialValuesDoubles, FinalScore, 100000 );

  // these are the new scaling factors used in the next iterations
  cout << "Got Params: ";
  // output to cout
  copy( returnLambdas.begin(), returnLambdas.end(), ostream_iterator< float >( cout, " " ) );
  cout << endl;

  // copying it to the scaling factor array, to use them in the next iteration
  // notice, this may not be the best scaling factors seen so far; but using the best
  // configuration would not generate new translations, i.e. the optimization would then stop
  // Actually, this should be handled by the optimization; the optimization should never return
  // a configuration, which is worth then the previous one

  //ScalingFactors = returnLambdas;
  copy( returnLambdas.begin(), returnLambdas.end(), ScalingFactors.begin() );

  return FinalScore;
}


/*!
 Display min and max of each feature score used in the sentence, adjusted
 by scaling factors. Formats and prints scores to std out.

 Iterates over all edges. If edge has a pointer to a transducer item, verified
 with pEdge->GetTransducerItem() != NULL, then it is a translation edge. Its 
 Score() array is then traversed, and the min,max for each score is updated in
 the score vector.
 
 \note Uses global array ScalingFactors
 \param pLattice of type CLatticeLM* representing one translated sentence.
 \author Anthony D'Auria (apd@cs.cmu.edu)

 */
void
DisplayFeaturesMinMax( CLatticeLM *pLattice )
{
  
  int TMStartIndex = pDecoder->GetStartingPointTMScalingFactor();
  int TMFeaturesSize = ScalingFactors.size()-(TMStartIndex);
  
  // Initialize MinMaxArray to be at most the size of ScalingFactors.
  CArray< std::pair<float,float> > MinMaxArray;
  for ( int s = 0; s < TMFeaturesSize; s++ ) {
    MinMaxArray[ s ].first  = FLT_MAX; // large number for min
    MinMaxArray[ s ].second = FLT_MIN; // small number for max
  }

  // Get scores from all translation edges of the lattice
  CArray< CLatticeEdge* > Edges = pLattice->Edges();
  for ( int e = Edges.MinIndex(); e <= Edges.MaxIndex(); e++ )
  {
    CLatticeEdge* pEdge = Edges[e];
    // Test if edge was generated by a transducer. Then it has scores.
    if ( ((CLatticeEdgeMS*)pEdge)->TransducerItem() != NULL ) 
    {
      CNumericalArray<float> EdgeScores = ((CLatticeEdgeMS*) pEdge)->Scores();
      //Iterate over scores and update min, max for each feature

      for ( int s = 0; s <= EdgeScores.MaxIndex(); s++ ) 
      {
        float score = EdgeScores[ s ] * ScalingFactors[ s+TMStartIndex ];
        if ( MinMaxArray[ s ].first > score )     // min > score
          MinMaxArray[ s ].first = score;
        if ( MinMaxArray[ s ].second < score )    // max < score
          MinMaxArray[ s ].second = score;
      }
      
    } else {
      // Not a translation edge
    }
  } // edge loop
  
  cout << endl << "FeatureScores(Min,Max): ";

  for ( int s = 0; s < TMFeaturesSize; s++ ) {
    cout << MinMaxArray[ s ].first << "#" << MinMaxArray[ s ].second << "#";
  }
  cout << endl;
  
}


void
DisplayScalingFactors    ( TScoreScalingFactorMap AllScalingFactors )
{

  TScoreScalingFactorMap::iterator iter;
  
  cout << endl;
  cout << "===========================================================================" << endl;
  cout << "All scaling factors from the different MER iterations" << endl;
  cout << endl;
  cout << "  Predicted 1best Oracle ScalingFactors" << endl;


  for ( int i = MetricScores.MinIndex(); i <= MetricScores.MaxIndex(); i++ ) {
    cout.setf( ios::fixed );
    cout.precision( 4 );
    cout << i << ": " <<  FinalScores[ i ] << " " << MetricScores[ i ] << " " << OracleScores[ i ] << "  ";

    cout.setf( ios::fixed );
    cout.precision( 6 );

    CNumericalArray< float > &Factors = FeatureWeights.GetRow( i );
    
    int n = Factors.MinIndex();
    cout << Factors[ n ];
    n++;
    while ( n <= Factors.MaxIndex() ) {
      cout << "_" << Factors[ n ];
      n++;
    }
    cout << endl;
  }



  /*
    for ( iter = AllScalingFactors.begin(); 
    iter != AllScalingFactors.end(); 
    ++iter ) {

    cout << iter->first << " : ";

    vector< float > &ScalingFactors = iter->second;

    int n = 0;
    cout << ScalingFactors[ n ];
    while ( n < ScalingFactors.size() ) {
    cout << "_" << ScalingFactors[ n ];
    n++;
    }
    cout << endl;
    }
  */

  cout << "===========================================================================" << endl;

}


//////
// Worker Bee helper functions
/////

//! checks if the file exists and returns the last-modified time
time_t BeeFileExists(const char* strFilename) {
  struct stat stFileInfo;
  int intStat;
  // Attempt to get the file attributes
  intStat = stat(strFilename,&stFileInfo);
  if(intStat == 0) {
    // We were able to get the file attributes
    // so the file obviously exists.
    time_t tmpt = time(NULL);
    cout << "File " << strFilename << " exists and was last modified at " << ctime(&stFileInfo.st_mtime);
    cout << " (current time: "
      << ctime(&tmpt) << ", difftime: " << difftime(tmpt,stFileInfo.st_mtime) << ")" << endl;
    return stFileInfo.st_mtime;
  } else {
    // We were not able to get the file attributes.
    // This may mean that we don't have permission to
    // access the folder which contains this file. If you
    // need to do that level of checking, lookup the
    // return values of stat which will give you
    // more details on why stat failed.
    cout << "File " << strFilename << " doesn't exist" << endl;
    return 0;
  }
}

/*!
  Attempts to claim a sentence for the current bee using open() system call.

 \note 
 
 \param const char* filename
 
 \return int file descriptor if lock was created successfully. 0 if lock already 
  exists.
 
 \author Andreas Zollman zollmann@cs.cmu.edu
 
 \todo 
 */
//! check the return code of open() as follows:
int BeeTakeUnitLock(const char* filename) // returns the lock's descriptor on success
{
  int fd; 
  fd=open(filename, O_WRONLY | O_CREAT | O_EXCL , NULL);
  if (fd<0 && errno==EEXIST)
  {
    // the file already exist; another process is 
    // holding the lock
    // cout<<"the file is currently locked; try again later";
    return -1;
  }
  else if (fd < 0)
  {
    // perror() appends a verbal description of the current
    // errno value after the user-supplied string
    perror("Error: Locking failed for the following reason: ");
    cerr << "Exit due to locking problem." << endl;
    exit(2);
    // return -1;
  }
  // if we got here, we own the lock
  return fd;
}


/*!
 Determines whether a sentence has already been translated. If it hasn't, tests 
 for presence of lock. If lock exists, tests if it is too old and should be taken 
 by the current bee.
 
 \note 
 
 \param string Sent_LockFile name of lock file. File exists only when sentence 
        translation hasn't been completed.
 \param string Sent_FeatsFile name of features file.
 \param bool workerBeeAtLeastOneSentenceUnprocessed if true, forces the bee to 
        recheck if all translations were successfully completed.
 \param bool workerBeeCompletedFirstPass if true, will attempt to reclaim locked 
        sentences. If false, will skip locked sentences.
 
 \return bool true if sentence has been fully translated and all data was written
 
 \author Initial design: Andreas Zollmann (zollmann@cs.cmu.edu)
         Integration and updates: Anthony D'Auria (apd@cs.cmu.edu) 7/25/07
 
 \todo 
 */

bool BeeIsUnitInProgress( int SentenceNumber, 
                          string &Sent_LockFile, string &Sent_FeatsFile, 
                          string &Sent_NBestTranslationsFile,
                          int &lockHandle,
                          bool &workerBeeAtLeastOneSentenceUnprocessed, 
                          bool &workerBeeCompletedFirstPass,
                          bool &workerBeeGetThemAllNow)
{
  
  bool Sent_Skip = false;
  
  // Get parameters
  int   WorkerBeeSaveSentenceFeatures =  *( (int*)  pParam->getValueFromKey( "WorkerBeeSaveSentenceFeatures" ));
  double MaxSentenceProcessingTime    =  *( (double*)pParam->getValueFromKey( "DisplayLatticeStatistics" )     );
  
  // Time lock file was last touched
  time_t lockFileModTime = BeeFileExists( Sent_LockFile.c_str() );
  
  // Check if feats file must be completed
  bool featsIncomplete = false;
  if ( WorkerBeeSaveSentenceFeatures 
       && ( BeeFileExists( Sent_FeatsFile.c_str() ) == 0 ) )
  {
    featsIncomplete = true;
  }
  
  //arv sez: you can only pass thru if you see both the feats and cands there AND the lock file has been properly removed
  //this picks up cases where the process was terminated during writing the files
  if (   BeeFileExists( Sent_NBestTranslationsFile.c_str() ) 
         && !featsIncomplete
         && !lockFileModTime) {
    cout << "WorkerBee[ " << SentenceNumber << " ]: Skipping completed sentence." << endl;
    Sent_Skip = true;
    
  } else {
    // Current sentence was not translated or translation was incomplete
    //---- LOCK CODE ----
    // If Sentence has lock, we can skip it altogether or sleep and re-check
    lockHandle = BeeTakeUnitLock(Sent_LockFile.c_str());
    if ( lockHandle < 0 )
    {
      cout << "WorkerBee[ " << SentenceNumber << " ]: Sentence lock found." << endl;
      
      // If on first pass, do not attempt to translate or wait for locked sentence
      if ( !workerBeeCompletedFirstPass ) 
      {      // In first pass: simply skip locked sentence
        
        Sent_Skip = true;
        workerBeeAtLeastOneSentenceUnprocessed = true;
        cout << "WorkerBee[ " << SentenceNumber << " ]: On first pass. Skipping locked sentence." << endl;
        
      } else { // Not in first pass, so attempt to process locked sentence
        
        // 1) Wait for lock to go away by sleeping for n sec,
        // 2) Decide to re-translate sentence if lock is too old.
        
        // arv sez: get the difference in time between lock file timestamp and current time
        // and sleep until that difference is greater than MaxSentenceProcessingTime
        
        // if the lock file didn't exist (lockFileModTime == 0), then we should never have had trouble getting the lock
        if (!lockFileModTime) {
          cerr << "WorkerBee[ " << SentenceNumber << " ]: Error: couldn't get lock even though the lock file doesn't exist." << endl;
          exit(2);
        }
        
        time_t timenow;
        time(&timenow);
        double time_to_sleep = (double)MaxSentenceProcessingTime - difftime(timenow, lockFileModTime);
        if ( time_to_sleep < 0 ) time_to_sleep = 0; 
        //MAX(0.0, (double) MaxSentenceProcessingTime - difftime(timenow, lockFileModTime));
        
        cout << "WorkerBee[ " << SentenceNumber << " ]: Has been locked for " 
          << (double) difftime(timenow,lockFileModTime) << " seconds; MaxSentenceProcessingTime="
          << MaxSentenceProcessingTime << "; MaxSentenceProcessingTime-difftime=" 
          << (double) MaxSentenceProcessingTime - difftime(timenow,lockFileModTime) 
          << "; time_to_sleep=" << time_to_sleep << endl;
        
        if ( time_to_sleep > 0 && !workerBeeGetThemAllNow ) 
        {
          cout << "WorkerBee[ " << SentenceNumber << " ]: Sentence locked too recently. Skipping." << endl;
          Sent_Skip = true;
          workerBeeAtLeastOneSentenceUnprocessed = true;
          
        } else { // Lock is old
          
          // Will re-translate. First, wait for lock to go away by sleeping.
          
          // Sleep
          cout << "WorkerBee[ " << SentenceNumber << " ]: Waiting for lock. Going to sleep for " << (time_to_sleep) << "s";
          usleep((long) (time_to_sleep * 1000000.0));
          
          // Wake up and check if the lock file still exists and its modification time hasn't changed
          time_t lockFileModTimeNew = BeeFileExists(Sent_LockFile.c_str());
          if (!lockFileModTimeNew) {
            
            // Lock went away on its own
            
            Sent_Skip = true;
          } else {
            
            if (lockFileModTimeNew == lockFileModTime) 
            {
              
              // After sleeping, lock still exists with old time stamp
              
              cout << "WorkerBee[ " << SentenceNumber << " ]: Removing lock of dead process." << endl;
              
              // Removing old lock by deleting file with system call
              remove( Sent_LockFile.c_str() );
              
              // Create new lock
              lockHandle = BeeTakeUnitLock( Sent_LockFile.c_str() );
              if ( lockHandle < 0 ) {
                cerr << "Error: Deleted lock file but it's still locked!" << endl;
                exit(2);
              }
              
              // Not skipping sentence
              
            } else {
              // Lock still exists but another Bee has taken it over (its time stamp changed)
              Sent_Skip = true;
              workerBeeAtLeastOneSentenceUnprocessed = true;
            }
          } 
        } 
      } // Attempt to retranslate locked sentence
    } // Lock found
  } // Translation not found, so Sent_Skip = false;

  return Sent_Skip;
}
